#!/usr/bin/perl
#
#  Analyze dependencies between functions/global/static variables in programs
#   Joerg Schoen 2005-2008
#

use strict;
use English;

use Data::Dumper;
use IO::Handle; # for progress bar

my $MAXLINELEN = 78;

my @COptions;
my $Verbose = 0;
my $Debug = 0;
while(scalar @ARGV > 0) {
  if($ARGV[0] eq "-D") {
    ++$Debug;
  } elsif($ARGV[0] =~ /^-D(\d+)/) {
    $Debug = $1;
  } elsif($ARGV[0] =~ /^-[eID]/) {
    push @COptions, $ARGV[0];
  } elsif($ARGV[0] eq "-L") {
    $MAXLINELEN = -1;
  } elsif($ARGV[0] =~ /^-v/) {
    my $nr = $POSTMATCH;
    if($nr eq "") {
      ++$Verbose;
    } else {
      $Verbose = $nr;
    }
  } else {
    last;
  }
  shift @ARGV;
}

my $cmdmode = shift @ARGV;
if($cmdmode eq "generate" && scalar @ARGV > 1) {
  my (%fctdep, %fct2src);

  my $depfile = shift @ARGV;

  my $count = 0;
  foreach my $srcfile (@ARGV) {
    &show_progress("", $count / scalar @ARGV) if $Verbose;
    #printf "%s              \r", $srcfile if $Verbose;
    &genfdep(\%fctdep, \%fct2src, $srcfile, \@COptions);
    ++$count;
  }
  print "\n" if $Verbose;

  &dumpfdep(\%fctdep, \%fct2src, $depfile);
} elsif($cmdmode eq "update" && scalar @ARGV >= 1) {
  my $depfile = shift @ARGV;
  &updatedeps($depfile, \@ARGV, \@COptions);
} elsif($cmdmode eq "filedep" && scalar @ARGV >= 1) {
  my ($depfile) = @ARGV;
  &fct2filedep($depfile);
} elsif($cmdmode eq "findunused" && scalar @ARGV >= 2) {
  my $depfile = shift @ARGV;
  &findunused($depfile, \@ARGV);
} elsif($cmdmode eq "findundefined" && scalar @ARGV == 1) {
  &findundefined($ARGV[0]);
} elsif($cmdmode eq "findstatic" && scalar @ARGV == 1) {
  &findstatic($ARGV[0], 0);
} elsif($cmdmode eq "findstatic2" && scalar @ARGV == 1) {
  &findstatic($ARGV[0], 1);
} elsif($cmdmode eq "fileseparate" && scalar @ARGV == 1) {
  &fileseparate($ARGV[0]);
} else {
  print "Usage: jsgendeps.pl <opts> generate <depfile> <srcfile> ...\n";
  print "  Generate a dependency file from the given sources.\n";
  print "  This requires utility 'jscpp'. C-options like '-I' and\n";
  print "  and '-D' are understood.\n";
  print "Usage: jsgendeps.pl <opts> update <depfile> <srcfile> ...\n";
  print "  Update an existing dependency file, only reanalyzing\n";
  print "  sources that are newer than.\n";
  print "Usage: jsgendeps.pl <opts> findunused <depfile> <startsymbol> ...\n";
  print "  Find unused symbols.\n";
  print "Usage: jsgendeps.pl <opts> findundefined <depfile>\n";
  print "  Find undefined symbols that are referenced.\n";
  print "Usage: jsgendeps.pl <opts> findstatic|findstatic2 <depfile>\n";
  print "  Find symbols that could be made static. 'findstatic2' finds\n";
  print "  symbols that could be made static after relocating them to\n";
  print "  another source.\n";
  print "Usage: jsgendeps.pl <opts> fileseparate <depfile>\n";
  print "  Show for each file in which interdependent groups of symbols\n";
  print "  they could be separated.\n";
  exit 5;
}

# ***********************************************************************
sub genfdep
{
  my ($fctdep, $fct2src, $file, $coptions) = @_;
  local *CMD;

  my $cmd = &genfdepcmd($file, $coptions);

  printf "  Executing: %s\n", $cmd if $Verbose >= 2;
  open(CMD, "-|", $cmd) || die "Cannot execute '$cmd': $!";
  while(my $line = <CMD>)  {
    chomp $line;

    my ($function, $depstring) = ($line =~ /^(\S+):\s*(.*)/);
    die "Invalid line $line!" unless defined $depstring;

    my @deps = split /\s+/, $depstring;

    $$fct2src{$function} = $file;
    push @{$$fctdep{$function}}, @deps;
  }
  close CMD || die "Error executing '$cmd': $!";
}

sub genfdepcmd
{
  my ($srcfile, $coptions) = @_;

  my $cmd = "jscpp -m fdeps -pp ";
  $cmd .= join " ", @$coptions;
  $cmd .= " " . $srcfile;

  return $cmd;
}

# ***********************************************************************
sub dumpfdep
{
  my ($fctdep, $fct2src, $outfile) = @_;

  local *OUT;
  open(OUT, ">", $outfile) || die "Cannot write to $outfile: $!";
  print OUT "# Generated by jsgendeps.pl\n";
  foreach my $function (sort { $$fct2src{$a} cmp $$fct2src{$b} }
	keys %$fctdep) {
    my $line = $$fct2src{$function} . " " . $function . ":";

    foreach my $dep (sort @{$$fctdep{$function}}) {
      $line .= " " . $dep;
    }

    &printlongline(\*OUT, $line);
  }
  close OUT;
}

# ***********************************************************************
sub updatedeps
{
  my ($depfile, $filearr, $coptions) = @_;
  local *DEP, *DEPNEW;

  my $tempfile = $depfile . "~";

  my $deptime;
  if(-r $depfile) {
    $deptime = (stat($depfile))[9];
  } else {
    $deptime = 0;
  }

  my %updated;
  my $changed = 0;
  open(DEP, "<", $depfile) || die "Cannot read $depfile: $!";
  open(DEPNEW, ">", $tempfile) || die "Cannot write to $tempfile: $!";
  my $line;
  while(my $linepart = <DEP>) {
    chomp $linepart;

    if($linepart =~ /\\$/) {
      $line .= $PREMATCH;
      next;
    }

    if($linepart =~ /^#/) {
      print DEPNEW $linepart . "\n";
      next;
    }

    $line .= $linepart;

    if($line !~ /^(\S+) (\S+):(.*)/) {
      die "Invalid line '$line'!";
    }
    my $srcfile = $1;

    if(exists $updated{$srcfile} && $updated{$srcfile} != 2) {
      $line = "";
      next;
    }

    if(!-r $srcfile) {
      $updated{$srcfile} = 1;
      ++$changed;
      printf "  Delete dependencies for deleted source file '%s'\n",
	 $srcfile if $Verbose >= 1;
    } elsif((stat($srcfile))[9] >= $deptime) {
      local *CMD;

      my $cmd = &genfdepcmd($srcfile, $coptions);
      open(CMD, "-|", $cmd) || die "Cannot execute $cmd: $!";
      while(my $line2 = <CMD>) {
        printf DEPNEW "%s %s", $srcfile, $line2;
      }
      close CMD || die "Error executing '$cmd': $!";

      printf "  Updated dependencies for source file '%s'\n",
	 $srcfile if $Verbose >= 2;

      $updated{$srcfile} = 3;
      ++$changed;
    } else {
      $updated{$srcfile} = 2;
      print DEPNEW $line . "\n";
    }

    $line = "";
  }
  close DEP;

  foreach my $srcfile (@$filearr) {
    next if exists $updated{$srcfile};

    local *CMD;

    my $cmd = &genfdepcmd($srcfile, $coptions);
    open(CMD, "-|", $cmd) || die "Cannot execute $cmd: $!";
    while(my $line = <CMD>) {
      printf DEPNEW "%s %s", $srcfile, $line;
    }
    close CMD || die "Error executing '$cmd': $!";

    printf "  Added dependencies for source file '%s'\n",
	 $srcfile if $Verbose >= 2;

    $updated{$srcfile} = 4;
    ++$changed;
  }
  close DEPNEW;

  if($changed) {
    unlink $depfile;
    rename $tempfile, $depfile;
  } else {
    unlink $tempfile;
  }
}

# ***********************************************************************
sub printlongline
{
  my ($out, $line) = @_;

  my @words = split /\s+/, $line;

  my $start = shift @words;
  print $out $start;
  my $loff = length($start);
  foreach my $word (@words) {
    my $len = length($word);

    if($MAXLINELEN > 0 && ($loff + $len) > $MAXLINELEN) {
      print $out " \\\n\t";
      $loff = 8;
    } else {
      print $out " ";
      ++$loff;
    }

    printf $out $word;
    $loff += $len;
  }
  print $out "\n";
}

# ***********************************************************************
sub readfdep
{
  my ($infile) = @_;
  local *IN;

  my %fctdep;
  my %fct2src;

  open(IN, "<", $infile) || die("Cannot read $infile: $!");
  my $line;
  while(my $linepart = <IN>) {
    chomp $linepart;
    next if $linepart =~ /^#/;

    if($linepart =~ /\\$/) {
      $line .= $PREMATCH;
      next;
    }
    $line .= $linepart;

    if($line !~ /^(\S+) (\S+):\s*(.*)/) {
      die("Invalid line '$line'!");
    }
    my $srcfile = $1;
    my $function = $2;
    my @deps = split /\s+/, $3;

    $fct2src{$function} = $srcfile;
    $fctdep{$function} = \@deps;

    $line = "";
  }
  close IN;

  return(\%fctdep, \%fct2src);
}

# ***********************************************************************
sub invertsymbdep
{
  my $fctdep = $_[0];

  my %invfctdep;
  while(my ($symbol, $depsymbs) = each(%$fctdep)) {
    foreach my $depsymbol (@$depsymbs) {
      push @{$invfctdep{$depsymbol}}, $symbol;
    }
  }

  return \%invfctdep;
}

# ***********************************************************************
sub findunused
{
  my ($depfile, $startsymbols) = @_;

  my ($fctdep, $fct2src) = &readfdep($depfile);

  my %used;
  foreach my $startsymbol (@$startsymbols) {
    die "Unknown start symbol '" . $startsymbol . "'"
	unless defined $$fctdep{$startsymbol};

    &markclosure($startsymbol, $fctdep, \%used, 1);
  }

  foreach my $symbol (sort keys %$fctdep) {
    next if $used{$symbol} == 1;

    printf "%-60s %s\n", $symbol, $$fct2src{$symbol};
  }
}

sub markclosure
{
  my ($startsymbol, $fctdep, $desthash, $flag) = @_;

  $$desthash{$startsymbol} = $flag;

  my @todosymbols;
  push @todosymbols, $startsymbol;
  while(@todosymbols > 0) {
    my $symbol = shift @todosymbols;
    printf "# SYMB '%s' '%s'\n", $symbol, $flag if $Debug >= 1;

    my $depsymbols = $$fctdep{$symbol};
    foreach my $depsymbol (@$depsymbols) {
      next if $$desthash{$depsymbol} == $flag;

      printf "#   %s '%s'\n", $depsymbol, $$desthash{$depsymbol} if $Debug >= 2;

      push @todosymbols, $depsymbol;
      $$desthash{$depsymbol} = $flag;
    }
  }
}

# ***********************************************************************
sub findundefined
{
  my ($depfile) = @_;

  my ($fctdep, $fct2src) = &readfdep($depfile);

  my %undefsymbs;
  foreach my $depsymbols (values %$fctdep) {
    foreach my $depsymbol (@$depsymbols) {
      next if defined $$fctdep{$depsymbol};

      $undefsymbs{$depsymbol}++;
    }
  }

  foreach my $symbol (sort keys %undefsymbs) {
    printf "%-20s %d\n", $symbol, $undefsymbs{$symbol};
  }
}

# ***********************************************************************
#  If $mode == 0, we find symbols that could be made static.
#  If $mode == 1, we find symbols that could be made static if they
#  were relocated to the source they are referenced from.
#
sub findstatic
{
  my ($depfile, $mode) = @_;

  my ($fctdep, $fct2src) = &readfdep($depfile);
  my $invfctdep = &invertsymbdep($fctdep);

  my %staticcand;
  while(my ($depsymbol, $refsymbs) = each(%$invfctdep)) {
    next if !defined $$fctdep{$depsymbol} || $depsymbol =~ /.+@.+/;

    my %sources = ();
    $sources{$$fct2src{$depsymbol}} = undef if $mode == 0;
    foreach my $refsymb (@$refsymbs) {
      $sources{$$fct2src{$refsymb}} = undef;
    }

    my @sourcelist = keys %sources;
    if(scalar @sourcelist == 1) {
      push @{$staticcand{$sourcelist[0]}}, $depsymbol;
    }
  }

  foreach my $source (sort keys %staticcand) {
    printf "%s:\n", $source;
    foreach my $symbol (@{$staticcand{$source}}) {
      if($mode != 0) {
        my $symbsource = $$fct2src{$symbol};
        if($symbsource ne $source) {
          printf "\t%-40s (%s)\n", $symbol, $symbsource;
          next;
        }
      }

      printf "\t%s\n", $symbol;
    }
  }
}

sub fileseparate
{
  my ($depfile) = @_;

  my ($fctdep, $fct2src) = &readfdep($depfile);

  #  Do grouping by files
  my $currfile;
  my %symbdeps;
  foreach my $symbol (sort { $$fct2src{$a} cmp $$fct2src{$b} } keys %$fctdep) {
    #printf "%-30s %s\n", $symbol, $$fct2src{$symbol};

    my $srcfile = $$fct2src{$symbol};
    if($srcfile ne $currfile) {
      if(scalar keys %symbdeps > 0) {
	my $array = &separateone(\%symbdeps);

	if(scalar @$array > 1) {
	  printf "\n%s\n", $currfile;
	  foreach my $block (@$array) {
	    printf ">%s\n", join " ", @$block;
	  }
	}
      }

      %symbdeps = ();
      $currfile = $srcfile;
    }

    my $depsymbols = $$fctdep{$symbol};
    foreach my $depsymbol (@$depsymbols) {
      next if $$fct2src{$depsymbol} ne $srcfile;

      printf "%s: %s <-> %s\n", $srcfile, $symbol, $depsymbol
	  if $Debug > 0;
      push @{$symbdeps{$symbol}}, $depsymbol;
      push @{$symbdeps{$depsymbol}}, $symbol;
    }
  }
}

sub separateone
{
  my ($fctdeps) = @_;

  my %symbhash = map { $_ => undef } keys %$fctdeps;

  for(my $compnr = 1 ; ; ++$compnr) {
    my ($startsymbol) = grep !defined $symbhash{$_}, keys %symbhash;

    last unless defined $startsymbol;

    &markclosure($startsymbol, $fctdeps, \%symbhash, $compnr);
  }

  my @result;
  foreach my $symbol (sort { $symbhash{$a} <=> $symbhash{$b} } keys %symbhash) {
    printf "%3d %s\n", $symbhash{$symbol}, $symbol if $Debug >= 1;
    push @{$result[$symbhash{$symbol} - 1]}, $symbol;
  }

  return \@result;
}

# ***********************************************************************
#  Convert function-function dependencies into file-file
#   dependencies.
sub fct2filedep
{
  my ($depfile) = @_;

  my ($fctdep, $fct2src) = &readfdep($depfile);

  my %filedep;
  while(my ($symbol, $depsymbs) = each(%$fctdep)) {
    foreach my $depsymbol (@$depsymbs) {
      my $srcfile = $$fct2src{$symbol};
      my $depsrcfile = $$fct2src{$depsymbol};
      $filedep{$srcfile} = $depsrcfile;
    }
  }

  foreach my $srcfile (sort keys %filedep) {
    my $depsrcfile = $filedep{$srcfile};
    printf "%s %s\n", $srcfile, $depsrcfile;
  }
}

# ----------------------------------------------------------------------------
#  Show internal data structures for debugging
sub showdata
{
    my ($object) = @_;

    my $data = Data::Dumper->new([ $object ]);

    # No newlines and no variable names
    #printf "%s\n", $data->Terse(1)->Indent(0)->Dump;
    printf "%s\n", $data->Terse(1)->Dump;
}

# ----------------------------------------------------------------------------
#  Progress bar
INIT
{
  my $linesize = 78;
  my @windmill = ('|', '/', '-', '\\');
  my $wmill = 0;

  sub show_progress
  {
    my ($prefix, $perc) = @_;

    my $lsize = $linesize - length($prefix);

    my $bar = sprintf "%-*s", $lsize, "#" x int($perc * $lsize);
    my $pos = $lsize / 2 - 2;
    $bar = substr($bar, 0, $pos) . sprintf("%3d%%", 100 * $perc) . substr($bar, $pos + 4);
    printf "%s%s[%s]\r", $prefix, $windmill[$wmill++ % @windmill], $bar;
    STDOUT->flush;
  }
}
