/*JS*********************************************************************
*
*    File    : GENSORT.H
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : To generate automatically code for a sorting routine.
*
*              Copyright (C) 1994-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*
*  You have to define the following preprocessor symbols:
*  GENSORT_NAME       will be the name of the generated routine.
*  GENSORT_TYPE       is the type of the array to be sorted.
*  GENSORT_KEYTYPE    is the type of the key of the array elements.
*  GENSORT_GETKEY(a)  is a macro to get the key element of the array
*                     element a.
*  GENSORT_COMPAREKEYS(k1,k2) compares the two key elements k1 and k2 with
*                     condition "lower than" and yields TRUE / FALSE values.
*                     If the two keys match, it must return FALSE!
*
*  With the line
*     #include "gensort.h"
*  this file will automatically generate a sort routine.
*
*  The generated routine will have the prototype
*     void GENSORT_NAME(GENSORT_TYPE *array, long n)
*   and sorts the n array entries array[0] to array[n - 1].
*
*  At the end of this file the symbols GENSORT_* are undefined for reuse.
*
*  Flags:
*   If the flag "GENSORT_USEPOINTERS" is defined, the routine uses pointer
*    gymnastics instead of indices, which is faster on some machines.
*
*   If the flag "GENSORT_NISINT" is defined, the size argument n to the
*    generated routine is of type "unsigned int" instead of the default
*    "unsigned long".
*
*   If the flag "GENSORT_STATIC" is defined, the routines are generated
*    as static ones.
*
*   If the flag "GENSORT_NOSMALL" is defined, the quicksort routines do
*    not use a straight insertion algorithm for small n. The threshold
*    for n can be changed by defining "GENSORT_LIMIT" as the threshold,
*    the default is given below. Straight insertion is not used if the
*    macro "GENSORT_SWAP" is defined.
*
*   If the flag "GENSORT_NO_RECURSIVE" is defined, the quicksort routines
*    are non-recursive. If "GENSORT_RECURSIVE" is defined, they are. The
*    default is currently to generate recursive routines, since they are
*    faster (at least on a pentium).
*
*   If the flag "GENSORT_HEAPSORT" is defined, the sort routine uses heap-
*    sort instead of the default quicksort algorithm.
*
*   If the symbol "GENSORT_ARGSPROTO" and "GENSORT_ARGS" is defined, the
*    function has additional arguments.
*
*   If the symbol "GENSORT_SWAP" is defined, exchange of two elements is
*    done with the macro GENSORT_SWAP instead with a temporary variable.
*    The macro is called via "GENSORT_SWAP(a,i,j);" enclosed in
*    curly brackets and is expected to exchange the elements with index i
*    and j in the array a. Usage of "GENSORT_SWAP" together with
*    "GENSORT_HEAPSORT" is not possible!
*
*   If the symbol "GENSORT_MERGESORT" is defined, a mergesort routine for
*    linked lists is generated. The routines only expects one argument,
*    namely the pointer to the head of the linked list and returns the head
*    sorted list.
*    The additional macros "GENSORT_GETNEXT(a)" and "GENSORT_SETNEXT(a,b)"
*    are required to access the pointer to the next element of the list
*    and to set it.
*    If the additional flag "GENSORT_MERGESORT_NONATURAL" is set, the
*    optimization to look for natural runs, i. e. take advantage of
*    sequences already sorted is disabled.
*
*  If the symbol "GENSORT_DEBUG" is defined, the macro GENSORT_COMPAREKEYS
*   will be tested with two identical keys, checking if it returns false.
*   Otherwise the program will stop with an appropriate error message.
*
*  The quicksort is a recursive routine which uses at most
*      log_2(n) * (3 * sizeof(GENSORT_TYPE*) + sizeof(GENSORT_KEYTYPE)
*                  sizeof(GENSORT_TYPE))
*   bytes of stack (may be smaller if optimizer is good).
*
*  Example 1: To generate a quicksort routine which sorts an integer array
*   in increasing order, use the following lines:
*      | Left margin
*      V
*      #define GENSORT_NAME                 intquicksort
*      #define GENSORT_TYPE                 int
*      #define GENSORT_KEYTYPE              int
*      #define GENSORT_GETKEY(a)            a
*      #define GENSORT_COMPAREKEYS(k1,k2)   k1 < k2
*
*      #include "gensort.h"
*
*  Example 2: To generate a pointer quicksort routine that sorts a structure
*   according to a string in the key field (<string.h> is included to get
*   the prototype for the strcmp function):
*      | Left margin
*      V
*      #include <string.h>
*
*      struct Test {
*        char *KeyString;
*        int   OtherFields;
*         ....
*      };
*
*      #define GENSORT_NAME                 namequicksort
*      #define GENSORT_TYPE                 struct Test
*      #define GENSORT_KEYTYPE              char *
*      #define GENSORT_GETKEY(a)            a.KeyString
*      #define GENSORT_COMPAREKEYS(k1,k2)   strcmp(k1,k2) < 0
*
*      #define GENSORT_USEPOINTERS
*
*      #include "gensort.h"
*
*  Example 3: Generate a routine to sort an indirection table 'ind' for
*   a string array 'table', i. e. 'table' stays unsorted, but 'table[ind[i]]'
*   will access the original array in sorted order
*      | Left margin
*      V
*      #include <string.h>
*
*      ...
*
*      #define GENSORT_NAME                 indquicksort
*      #define GENSORT_ARGS                 ,strings
*      #define GENSORT_ARGSPROTO            ,char **strings
*      #define GENSORT_TYPE                 int
*      #define GENSORT_KEYTYPE              char *
*      #define GENSORT_GETKEY(a)            strings[a]
*      #define GENSORT_COMPAREKEYS(k1,k2)   strcmp(k1,k2) < 0
*
*      #define GENSORT_NISINT
*
*      #include "gensort.h"
*
*      ...
*
*      {
*        char **strings;
*        unsigned int i, n, *ind;
*
*        ...
*
*        for(i = 0 ; i < n ; i++) ind[i] = i;
*
*        indquicksort(ind, n, strings);
*
*        for(i = 0 ; i < n ; i++)
*          printf("%s\n", strings[ind[i]]);
*
*      }
*
*  Example 4: Generate a mergesort routine for a linked list
*
*      struct linkedlist {
*        struct linkedlist *Next;
*        int Key;
*      };
*
*      #define GENSORT_NAME                 mergesort
*      #define GENSORT_TYPE                 struct linkedlist
*      #define GENSORT_KEYTYPE              int
*      #define GENSORT_GETKEY(a)            a.Key
*      #define GENSORT_COMPAREKEYS(k1, k2)  k1 < k2
*      #define GENSORT_MERGESORT
*      #define GENSORT_GETNEXT(a)           a->Next
*      #define GENSORT_SETNEXT(a, b)        a->Next = b
*      #define GENSORT_STATIC
*      #include "gensort.h"
*
*      {
*        linkedlist *list;
*
*        ...
*
*        list = mergesort(list);
*
*      }
*
*************************************************************************/

/*  Check if all symbols we need are defined, so that user gets a
 *   message from the preprocessor in case he forgot something.
 */
#ifndef GENSORT_NAME
# error "gensort.h": Symbol "GENSORT_NAME" undefined!
#endif

#ifndef GENSORT_TYPE
# error "gensort.h": Symbol "GENSORT_TYPE" undefined!
#endif

#ifndef GENSORT_KEYTYPE
# error "gensort.h": Symbol "GENSORT_KEYTYPE" undefined!
#endif

#ifndef GENSORT_GETKEY
# error "gensort.h": Macro "GENSORT_GETKEY"  undefined!
#endif

#ifndef GENSORT_COMPAREKEYS
# error "gensort.h": Macro "GENSORT_COMPAREKEYS"  undefined!
#endif

#ifndef GENSORT_NO_RECURSIVE
/*  Recursion is faster on my machine!  */
# define GENSORT_RECURSIVE
#endif

#ifndef GENSORT_ARGSPROTO
/*  If not given, set them to nothing  */
# define GENSORT_ARGSPROTO
# define GENSORT_ARGS
#endif

/*  It's better to cast the result of GENSORT_GETKEY to the expected type,
 *   otherwise we might encounter bug (i. e. GENSORT_TYPE is "signed char *",
 *   but GENSORT_GETKEY actually returns "unsigned char *". Since we define
 *   a variable of type GENSORT_KEYTYPE, the assignment to it will do a cast
 *   implicitly, but when we later compare this variable with the result of
 *   macro GENSORT_GETKEY, there will be no cast, leading to wrong
 *   comparisons which actually might result in a core dump.
 *   Since some compilers don't like the cast on the variable, we don't use
 *   it if we apply macro GENSORT_GETKEY to the variable itself.
 */
#ifndef GENSORT_NOCAST_GETKEY
# define _intern_GENSORT_GETKEY(a)   ((GENSORT_KEYTYPE)(GENSORT_GETKEY(a)))
#else
# define _intern_GENSORT_GETKEY(a)   GENSORT_GETKEY(a)
#endif

#ifndef GENSORT_LIMIT
/*  If array size is GENSORT_LIMIT or below, use straight insertion  */
# define GENSORT_LIMIT   14  /*  value is hand optimized  */
#endif

#ifndef GENSORT_INTTYPE
# ifdef GENSORT_NISINT
#  define GENSORT_INTTYPE     unsigned int
# else
#  define GENSORT_INTTYPE     unsigned long
# endif
#endif

#if defined(GENSORT_HEAPSORT) || defined(GENSORT_XDEBUG)
/*   Preprocessor macro to generate temporary names  */
# define GENSORT_CONCAT(name, ext)   GENSORT_CONCAT2(name, ext)
# define GENSORT_CONCAT2(name, ext)  name##ext
#endif

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#if !defined GENSORT_HEAPSORT && !defined GENSORT_MERGESORT
# ifndef GENSORT_USEPOINTERS

#  ifndef GENSORT_GETELEMENT
#   define GENSORT_GETELEMENT(a,i)    a[i]
#  endif

/* ******   Quicksort without pointers     *********** */
#  ifdef GENSORT_XDEBUG
#   include <stdio.h> /*  for stderr + fprintf  */

static int GENSORT_CONCAT(GENSORT_NAME, Level);
#  endif

#  ifdef GENSORT_STATIC
static
#  endif
void GENSORT_NAME(GENSORT_TYPE *_array,GENSORT_INTTYPE _n GENSORT_ARGSPROTO)
{
#  ifndef GENSORT_RECURSIVE
  struct {
    GENSORT_TYPE *_stkarray;
    GENSORT_INTTYPE _stkn;
  } _stack[sizeof(GENSORT_INTTYPE) * 8],*_sptr;

  _sptr = _stack;
#  endif

#  ifdef GENSORT_DEBUG
  if(_n > 0 && (GENSORT_COMPAREKEYS(GENSORT_GETELEMENT(_array,0),
				     GENSORT_GETELEMENT(_array,0)))) {
#   define GENSORT_STRINGIZE2(s)   #s
#   define GENSORT_STRINGIZE(s)    GENSORT_STRINGIZE2(s)
    fprintf(stderr,"\
ERROR routine %s: comparison of identical elements yields true -- Break!\n",
	    GENSORT_STRINGIZE(GENSORT_NAME));
#   undef GENSORT_STRINGIZE
#   undef GENSORT_STRINGIZE2
    exit(30);
  }
#  endif

#  ifdef GENSORT_XDEBUG
  ++GENSORT_CONCAT(GENSORT_NAME, Level);
#  endif

  /*   We only need __n - 1 in most cases    */
  if(_n) /*  protect against GENSORT_INTTYPE being "unsigned"  */
    _n--;

#  ifndef GENSORT_RECURSIVE
  for(;;) {
#  endif

#  ifdef GENSORT_XDEBUG
    fprintf(stderr, "GENSORT_XDEBUG: Level=%d, N=%d\n",
	    GENSORT_CONCAT(GENSORT_NAME, Level), _n + 1);
#  endif

    /*    Only sort array of at least two elements, all other are sorted  */
    while(_n > 0) {
      register GENSORT_INTTYPE _i,_j;
      GENSORT_KEYTYPE _x;

#  if !defined(GENSORT_NOSMALL) && !defined(GENSORT_SWAP)
      /*  If __n is small enough, use straight insertion  */
      if(_n < GENSORT_LIMIT) {
	for(_j = 1 ; _j <= _n ; _j++) {
	  GENSORT_TYPE _save;

	  _save = GENSORT_GETELEMENT(_array,_j);

	  for(_i = _j - 1 ; _i != (GENSORT_INTTYPE)-1 &&
		(GENSORT_COMPAREKEYS(GENSORT_GETKEY(_save),
				     _intern_GENSORT_GETKEY((GENSORT_GETELEMENT(_array,_i)))))
		; _i--) {
	    GENSORT_GETELEMENT(_array,(_i + 1)) =
	      GENSORT_GETELEMENT(_array,_i);
	  }

	  GENSORT_GETELEMENT(_array,(_i + 1)) = _save;
	}

	break;
      }
#  endif

      /*   Preset pointers   */
      _i = 0;    /*  Index to first element  */
      _j = _n;  /*  Index to last element   */

#  ifndef GENSORT_RANDMIDDLE
      /*      Get middle element      */
      _x = _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array,((_n + 1) >> 1)));
#  else
      {
	static unsigned long _rand;

	_rand = ((_rand * 7621) + 1) % 32768;
	switch(_rand % 3) {
	case 0:
	  _x = _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array, 0));
	  break;
	case 1:
	  _x = _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array,((_n + 1) >> 1)));
	  break;
	case 2:
	  _x = _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array, _n + 1));
	  break;
	}
      }
#  endif

      do {
	while(GENSORT_COMPAREKEYS(_intern_GENSORT_GETKEY((GENSORT_GETELEMENT(_array,_i))),_x))
	  _i++;

	while(GENSORT_COMPAREKEYS(_x,_intern_GENSORT_GETKEY((GENSORT_GETELEMENT(_array,_j)))))
	  _j--;

	if(_i > _j) break;

#  ifdef GENSORT_SWAP
	{
	  /*    Use user supplied macro   */
	  GENSORT_SWAP(_array,_i,_j);
	}
	_i++; _j--;
#  else
	{
	  GENSORT_TYPE _temp; /*   Only needed temporary  */

	  /*     Exchange elements   */
	  _temp = _array[_i];  _array[_i++] = _array[_j];
	  _array[_j--] = _temp;
	}
#  endif
      } while(_i <= _j);

      if(_j < (_n - _i)) {
#  ifndef GENSORT_RECURSIVE
	/*  Call recursive with right part  */
	if(_i < _n) {
	  _sptr->_stkarray = &GENSORT_GETELEMENT(_array,_i);
	  _sptr->_stkn = _n - _i;
	  _sptr++;
	}

	/*  Loop again with left part   */
	_n = _j;
#  else
	/*  Call recursive with left part  */
	if(0 < _j)
	  GENSORT_NAME(_array,(_j + 1) GENSORT_ARGS);

	/*  Loop again with right part   */
	_array = &GENSORT_GETELEMENT(_array,_i);
	_n -= _i;
#  endif
      } else {
#  ifndef GENSORT_RECURSIVE
	/*  Call recursive with left part  */
	if(0 < _j) {
	  _sptr->_stkarray = _array;
	  _sptr->_stkn = _j;
	  _sptr++;
	}

	/*  Loop again with right part   */
	_array = &GENSORT_GETELEMENT(_array,_i);
	_n -= _i;
#  else
	/*  Call recursive with right part  */
	if(_i < _n)
	  GENSORT_NAME(&GENSORT_GETELEMENT(_array,_i),
		       (_n + 1 - _i) GENSORT_ARGS);

	/*  Loop again with left part   */
	_n = _j;
#  endif
      }
    }

#  ifndef GENSORT_RECURSIVE
    if(--_sptr < _stack) break;
    _array = _sptr->_stkarray;
    _n = _sptr->_stkn;
  }
#  endif

#  ifdef GENSORT_XDEBUG
  --GENSORT_CONCAT(GENSORT_NAME, Level);
#  endif
}

# else /*  #ifndef GENSORT_USEPOINTERS  */

#  ifdef GENSORT_GETELEMENT
#   error GENSORT.H: Cannot use GENSORT_GETELEMENT with pointers!
#  endif

/* ******   Quicksort with pointers     *********** */
#  ifdef GENSORT_STATIC
static
#  endif
void GENSORT_NAME(GENSORT_TYPE *_array,GENSORT_INTTYPE _n GENSORT_ARGSPROTO)
{
#  ifndef GENSORT_RECURSIVE
  struct {
    GENSORT_TYPE *_stkarray;
    GENSORT_INTTYPE _stkn;
  } _stack[sizeof(GENSORT_INTTYPE) * 8],*_sptr;

  _sptr = _stack;
#  endif

  /*   We only need __n - 1 in most cases    */
  if(_n) /*  protect against GENSORT_INTTYPE being "unsigned"  */
    _n--;

#  ifndef GENSORT_RECURSIVE
  for(;;) {
#  endif
    /*    Only sort array of at least two elements, all other are sorted  */
    while(_n > 0) {
      register GENSORT_TYPE *_arrayi;
      register GENSORT_TYPE *_arrayj;
      GENSORT_KEYTYPE _x;

#  if !defined(GENSORT_NOSMALL) && !defined(GENSORT_SWAP)
      /*  If __n is small enough, use straight insertion  */
      if(_n < GENSORT_LIMIT) {
	for(_arrayj = &_array[1] ; _arrayj <= &_array[_n] ; _arrayj++) {
	  GENSORT_TYPE _save;

	  _save = *_arrayj;

	  for(_arrayi = _arrayj - 1 ; _arrayi >= _array &&
		(GENSORT_COMPAREKEYS(GENSORT_GETKEY(_save),
				     _intern_GENSORT_GETKEY((*_arrayi))))
		; _arrayi--) {
	    _arrayi[1] = *_arrayi;
	  }

	  _arrayi[1] = _save;
	}

	break;
      }
#  endif

      /*   Preset pointers   */
      _arrayi = _array;        /*  Pointer to first element  */
      _arrayj = _array + _n;  /*  Pointer to last element   */

#  ifndef GENSORT_RANDMIDDLE
      /*      Get middle element      */
      _x = _intern_GENSORT_GETKEY((_array[(_n + 1) >> 1]));
#  else
      {
	static unsigned long _rand;

	_rand = ((_rand * 7621) + 1) % 32768;
	fprintf(stderr, "RAND %d %d\n",_rand,_rand%3);
	switch(_rand % 3) {
	case 0:
	  _x = _intern_GENSORT_GETKEY((_array[0]));
	  break;
	case 1:
	  _x = _intern_GENSORT_GETKEY((_array[(_n + 1) >> 1]));
	  break;
	case 2:
	  _x = _intern_GENSORT_GETKEY((_array[_n]));
	  break;
	}
      }
#  endif

      do {
	while(GENSORT_COMPAREKEYS(_intern_GENSORT_GETKEY((*_arrayi)),_x))
	  _arrayi++;

	while(GENSORT_COMPAREKEYS(_x,_intern_GENSORT_GETKEY((*_arrayj))))
	  _arrayj--;

	if(_arrayi > _arrayj)
	  break;

#  ifdef GENSORT_SWAP
	{
	  /*    Use user supplied macro   */
	  GENSORT_SWAP(_array,(_arrayi - _array),(_arrayj - _array));
	}
	_arrayi++; _arrayj--;
#  else
	{
	  GENSORT_TYPE _temp; /*   Only needed temporary  */

	  /*     Exchange elements   */
	  _temp = *_arrayi;  *_arrayi++ = *_arrayj;  *_arrayj-- = _temp;
	}
#  endif
      } while(_arrayi <= _arrayj);

      if( (GENSORT_INTTYPE)(_arrayj - _array) < (_array - _arrayi + _n)) {
#  ifndef GENSORT_RECURSIVE
	/*  Call recursive with right part  */
	if(_arrayi < (_array + _n)) {
	  _sptr->_stkarray = _arrayi;
	  _sptr->_stkn = (_array - _arrayi) + _n;
	  _sptr++;
	}

	/*  Loop again with left part   */
	_n = _arrayj - _array;
#  else
	/*  Call recursive with left part  */
	if(_array < _arrayj)
	  GENSORT_NAME(_array,_arrayj - _array + 1 GENSORT_ARGS);

	/*  Loop again with right part   */
	_n -= _arrayi - _array;
	_array = _arrayi;
#  endif
      } else {
#  ifndef GENSORT_RECURSIVE
	/*  Call recursive with left part  */
	if(_array < _arrayj) {
	  _sptr->_stkarray = _array;
	  _sptr->_stkn = _arrayj - _array;
	  _sptr++;
	}

	/*  Loop again with right part   */
	_n -= _arrayi - _array;
	_array = _arrayi;
#  else
	/*  Call recursive with right part  */
	if(_arrayi < (_array + _n))
	  GENSORT_NAME(_arrayi,_array + _n + 1 - _arrayi GENSORT_ARGS);

	/*  Loop again with left part   */
	_n = _arrayj - _array;
#  endif
      }
    }

#  ifndef GENSORT_RECURSIVE
    if(--_sptr < _stack) break;
    _array = _sptr->_stkarray;
    _n = _sptr->_stkn;
  }
#  endif
}
# endif /*  #ifndef GENSORT_USEPOINTERS  */

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#elif defined GENSORT_MERGESORT

# ifndef GENSORT_GETNEXT
#  error "gensort.h": Macro "GENSORT_GETNEXT" required for mergesort!
# endif
# ifndef GENSORT_SETNEXT
#  error "gensort.h": Macro "GENSORT_SETNEXT" required for mergesort!
# endif

#  ifndef GENSORT_MERGESORT_NONLOCAL

/* ******   Localized mergesort routine              *********** */

/*  The following algorithm is my own idea - at least I don't know if anybody
 *   else has come up with this before.
 *   The non-local algorithm below goes through the whole list repeatedly and
 *   doubles the sorting length on every iteration.
 *   This one maintains a stack of sorting length and tries to merge equal
 *   lengths as soon as possible. This takes advantage of a CPU caches.
 */

#   ifdef GENSORT_STATIC
static
#   endif
GENSORT_TYPE *GENSORT_NAME(GENSORT_TYPE *_list GENSORT_ARGSPROTO)
{
  struct {
    GENSORT_TYPE *_list;
    unsigned int _level;
  } _stack[sizeof(GENSORT_INTTYPE) * 8];
  int _nstack;

  for(_nstack = -1 ; ; ) {
    register GENSORT_TYPE *_curr, *_plist, *_qlist;

    /*  Can we or do we have to merge lists from the stack?  */
    if(_nstack < 1 || (_stack[_nstack - 1]._level != _stack[_nstack]._level &&
		       _list != NULL)) {
      /* **  Push element(s) on the stack  ** */
      if(_list == NULL)
	/*  This is where we break the loop. We have _nstack == 0 now  */
	break;

      /* **  Sort two entries manually to push at least lists of length 2  ** */
      _curr = _list;
      _list = GENSORT_GETNEXT(_list);

      if(_list != NULL) {
	_plist = _list;
	_list = GENSORT_GETNEXT(_list);

	if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_plist)), GENSORT_GETKEY((*_curr)))) {
	  /*  Reverse the two entries  */
	  GENSORT_SETNEXT(_curr, _list);
	  GENSORT_SETNEXT(_plist, _curr);
	  _qlist = _curr;
	  _curr = _plist;
	  _plist = _qlist;
	}

#   if !defined GENSORT_MERGESORT_NONATURAL
	/*  Check for a natural run  */
	while(_list != NULL) {
	  if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_list)), GENSORT_GETKEY((*_plist))))
	    break;

	  _plist = _list;
	  _list = GENSORT_GETNEXT(_list);
	}
#   endif
      } else {
	_plist = _curr;
      }

      /*  Properly terminate the run  */
      GENSORT_SETNEXT(_plist, NULL);

      /* **  Push this run on the stack  ** */
      ++_nstack;
      _stack[_nstack]._list = _curr;
      _stack[_nstack]._level = 0;

      continue;
    }

    /* **  Merge top two entries from stack  ** */
    _plist = _stack[_nstack]._list;
    --_nstack;
    _qlist = _stack[_nstack]._list;

    /* **  Set up new stack element. This also simplifies the main loop below ** */
    if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_qlist)), GENSORT_GETKEY((*_plist)))) {
      _curr = _qlist;
      _qlist = GENSORT_GETNEXT(_qlist);

    } else {
      _curr = _plist;
      _plist = GENSORT_GETNEXT(_plist);
    }

    _stack[_nstack]._list = _curr;
    _stack[_nstack]._level++;

    /* **  Here is the main merging loop  ** */
    while(_plist != NULL && _qlist != NULL) {
      if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_qlist)), GENSORT_GETKEY((*_plist)))) {
	GENSORT_SETNEXT(_curr, _qlist); _curr = _qlist;
	_qlist = GENSORT_GETNEXT(_qlist);
      } else {
	GENSORT_SETNEXT(_curr, _plist); _curr = _plist;
	_plist = GENSORT_GETNEXT(_plist);
      }
    }

    /*  Add remainder to result list  */
    if(_plist != NULL) {
      GENSORT_SETNEXT(_curr, _plist);
    } else {
      GENSORT_SETNEXT(_curr, _qlist);
    }

#   if defined GENSORT_XDEBUG
    {
      GENSORT_INTTYPE _i;

      printf("MERGE #%d:", 1 + _nstack);
      for(_i = 0 ; _i <= _nstack ; ++_i)
	printf(" %d", _stack[_i]._level);
      printf("\n");
    }
#   endif
  }

  /*  Here we treat the case of list == NULL  */
  return _nstack == 0 ? _stack[_nstack]._list : NULL;
}
#  else /*  GENSORT_MERGESORT_NONLOCAL  */

/* ******   Non-localized mergesort routine              *********** */

/*  Version of mergesort that is non-local, i. e. it runs through
 *   the whole list repeatedly and doubles the length of the runs
 *   on each iteration. This is usually slower than the localized
 *   version above due to cache misses.
 *   As it also turns out, this algorithm is more complex than the
 *   localized above, i. e. more effort has to be taken to treat
 *   special cases.
 */

#   ifdef GENSORT_STATIC
static
#   endif
GENSORT_TYPE *GENSORT_NAME(GENSORT_TYPE *_list GENSORT_ARGSPROTO)
{
  GENSORT_INTTYPE _k;

  for(_k = 1 ;; _k <<= 1) {
    register GENSORT_TYPE *_plist, *_qlist, *_newlist, *_curr;
    unsigned int _nmerged;

    /*  Initialize result list  */
    _newlist = _curr = NULL;
    _nmerged = 0;
    for(_plist = _list ; _plist != NULL ; ) {
      register GENSORT_INTTYPE _psize, _qsize;

      ++_nmerged;

      /*  We are at 'p' and want to merge lists of length 'k'  */
      _qlist = _plist;
      for(_psize = _k ; _qlist != NULL && _psize-- > 0 ; )
	_qlist = GENSORT_GETNEXT(_qlist);

      if(_qlist == NULL) {
	/*  We reached the trailer, so add remainder to the result  */
	if(_curr == NULL)
	  _newlist = _plist;
	else
	  GENSORT_SETNEXT(_curr, _plist);
	break;
      }

      /* ***  Merge two lists at 'p' and 'q'  *** */
      _psize = _qsize = _k;
      /*  Do the first element separately to simplify the main loop below  */
      if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_qlist)), GENSORT_GETKEY((*_plist)))) {
	if(_curr == NULL)
	  _newlist = _curr = _qlist;
	else
	  GENSORT_SETNEXT(_curr, _qlist);
	_curr = _qlist;

	_qlist = GENSORT_GETNEXT(_qlist);
	--_qsize;
      } else {
	if(_curr == NULL)
	  _newlist = _curr = _plist;
	else
	  GENSORT_SETNEXT(_curr, _plist);
	_curr = _plist;

	_plist = GENSORT_GETNEXT(_plist);
	--_psize;
      }

      if(_qlist != NULL) {
	/* **  Here is the main merging loop  ** */
	while(_psize > 0 && _qsize > 0) {
	  if(GENSORT_COMPAREKEYS(GENSORT_GETKEY((*_qlist)), GENSORT_GETKEY((*_plist)))) {
	    GENSORT_SETNEXT(_curr, _qlist);
	    _curr = _qlist;
	    --_qsize;
	    _qlist = GENSORT_GETNEXT(_qlist);

	    if(_qlist == NULL)
	      break;
	  } else {
	    GENSORT_SETNEXT(_curr, _plist);
	    _curr = _plist;
	    --_psize;
	    _plist = GENSORT_GETNEXT(_plist);
	  }
	}
      }

      if(_psize > 0) {
	GENSORT_SETNEXT(_curr, _plist);

	do
	  _curr = GENSORT_GETNEXT(_curr);
	while(--_psize > 0);
      } else {
	GENSORT_SETNEXT(_curr, _qlist);

	while(_qsize-- > 0 && GENSORT_GETNEXT(_curr) != NULL)
	  _curr = GENSORT_GETNEXT(_curr);

	_qlist = GENSORT_GETNEXT(_curr);
      }

      GENSORT_SETNEXT(_curr, NULL);

      _plist = _qlist;
    }

    _list = _newlist;

    if(_nmerged == 1)
      break;
  }

  return _list;
}
#  endif

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
#else

/* ******   Heapsort routine              *********** */

# ifdef GENSORT_SWAP
#  error "gensort.h": Cannot use GENSORT_SWAP and GENSORT_HEAPSORT!
# endif

# ifndef GENSORT_GETELEMENT
#  define GENSORT_GETELEMENT(a,i)    a[i]
# endif

/* ****   Sift routine (always static)  **** */
static void GENSORT_CONCAT(_sift, GENSORT_NAME) (GENSORT_TYPE *_array,GENSORT_INTTYPE _i,
						 GENSORT_INTTYPE _n GENSORT_ARGSPROTO)
{
  register GENSORT_INTTYPE _j;
  GENSORT_TYPE _temp;

  /*   Get first element to sink in heap   */
  _temp = GENSORT_GETELEMENT(_array,_i);

  while( (_j = ((_i << 1) + 1)) < _n) {
    if( (_j < (_n - 1)) &&
	(GENSORT_COMPAREKEYS(_intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array,_j)),
			     _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array,(_j + 1))))) )
      ++_j;

    if( !(GENSORT_COMPAREKEYS(_intern_GENSORT_GETKEY(_temp),
			      _intern_GENSORT_GETKEY(GENSORT_GETELEMENT(_array,_j)))) )
      break;

    /*   Copy lower element to top    */
    GENSORT_GETELEMENT(_array,_i) = GENSORT_GETELEMENT(_array,_j);
    _i = _j;
  }

  /*   Now the first element has reached its place   */
  GENSORT_GETELEMENT(_array,_i) = _temp;
}

/* ****    Now the main heapsort routine  **** */
# ifdef GENSORT_STATIC
static
# endif
void GENSORT_NAME(GENSORT_TYPE *_array,GENSORT_INTTYPE _n GENSORT_ARGSPROTO)
{
  register GENSORT_INTTYPE _i;

  /*  Build up heap    */
  for(_i = (_n >> 1) ; _i > 0 ; )
    GENSORT_CONCAT(_sift, GENSORT_NAME) (_array, --_i, _n GENSORT_ARGS);

  for(_i = _n - 1 ; _i > 0 ; ) {
    {
      GENSORT_TYPE _temp;

      /*     Exchange elements   */
      _temp = GENSORT_GETELEMENT(_array, 0);
      GENSORT_GETELEMENT(_array, 0) = GENSORT_GETELEMENT(_array,_i);
      GENSORT_GETELEMENT(_array, _i) = _temp;
    }

    /*   Call sift routine    */
    GENSORT_CONCAT(_sift, GENSORT_NAME) (_array, 0, _i-- GENSORT_ARGS);
  }
}

#endif
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

# undef GENSORT_CONCAT
# undef GENSORT_CONCAT2

/* *******  Undefine all symbols for reuse    ****** */
#undef GENSORT_NAME
#undef GENSORT_TYPE
#undef GENSORT_KEYTYPE
#undef GENSORT_GETKEY
#undef _intern_GENSORT_GETKEY
#undef GENSORT_COMPAREKEYS
#undef GENSORT_HEAPSORT
#undef GENSORT_MERGESORT
#undef GENSORT_MERGESORT_NONLOCAL
#undef GENSORT_MERGESORT_NONATURAL
#undef GENSORT_GETNEXT
#undef GENSORT_SETNEXT
#undef GENSORT_NO_RECURSIVE
#undef GENSORT_RECURSIVE
#undef GENSORT_ARGSPROTO
#undef GENSORT_ARGS
#undef GENSORT_LIMIT
#undef GENSORT_NOSMALL
#undef GENSORT_SWAP
#undef GENSORT_GETELEMENT
#undef GENSORT_NISINT
#undef GENSORT_INTTYPE
#undef GENSORT_STATIC
#undef GENSORT_RANDMIDDLE
#undef GENSORT_NOCAST_GETKEY
