/*JS*********************************************************************
*
*    Program : JSCPPLIB
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Include file for C preprocessor library "jscpp".
*
*              Copyright (C) 1994-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef JSCPPLIB_H /*   File must not be included more than once!   */
#define JSCPPLIB_H 1

/*********     INCLUDES 					*********/
/*  Needed when parsing this file without access to system headers  */
#ifndef NO_STD_INCLUDES
# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <ctype.h>

# include <jsconfig.h>

# include <jssubs.h>
#endif

/*********     DEFINES						*********/

/*  Add debugging code and routines  */
#define CONFIG_JSCPPLIB_DEBUG      8

/*
 *  ********  Explanation of configuration options:  *********
 *
 *  1. Preprocessor configuration
 *
 *    CONFIG_JSCPPLIB_LANGUAGELEVEL
 *      Default language level
 *
 *  1.1 Additional symbols
 *
 *    CONFIG_JSCPP_INCLEVEL:
 *      The preprocessor macro "__INCLUDE_LEVEL__" evaluates to the current
 *      depth of included files (0 in base file).
 *      Optional.
 *
 *    CONFIG_JSCPP_BASEFILE:
 *      The preprocessor macro "__BASE_FILE__" evaluates to the file name of
 *      the base file.
 *      Optional.
 *
 *    CONFIG_JSCPP_TIMESTAMP:
 *      The preprocessor macro "__TIMESTAMP__" evaluates to the modification
 *      date of the source file or the current time if reading from stdin.
 *      Optional.
 *
 *  1.2 Additional commands
 *
 *    CONFIG_JSCPP_ASSERTION:
 *      Allow assertions: Constructions like "#system(i386)". Only valid in
 *      "#if.." lines. Also, preprocessor commands "#assert" and "#unassert".
 *      Optional.
 *
 *    CONFIG_JSCPP_INCNEXT:
 *      Preprocessor command "#include_next": Includes the NEXT file found
 *      with the given name in one of the include directories that is not
 *      currently being included.
 *      Required.
 *
 *    CONFIG_JSCPP_SINCLUDE:
 *      Preprocessor command "#sinclude": Keep quiet if the include file
 *      doesn't exist.
 *      Optional.
 *
 *    CONFIG_JSCPP_DEFIMM:
 *      Preprocessor command "#define_imm": Expands macros while reading
 *      the "#define" line, so the newly defined macro gets its definition
 *      from the current setting of macros on the line.
 *      Required.
 *
 *    CONFIG_JSCPP_SYSVCMD:
 *      Preprocessor command "#ident" and "#sccs", used on System V: Is
 *      read and discarded. Exists simply to support some system headers
 *      that contain them.
 *      Optional.
 *
 *    CONFIG_JSCPP_WARNING:
 *      Preprocessor command "#warning": Similar to "#error", but does not
 *      break.
 *      Required for "jsmcheck".
 *
 *    CONFIG_JSCPP_LAZY_LINECMD:
 *      Accept lines like
 *        # 4711 "file"
 *      to be identical to a '#line' directive.
 *      Optional.
 *
 *  1.3 Behaviour
 *
 *    CONFIG_JSCPP_OPTLINES:
 *      Optimize preprocessed (or parsed) output by substituting more than
 *      10 empty lines with a appropriate '#line' directive.
 *      Optional.
 *
 *    CONFIG_JSCPP_OPTINC:
 *      Optimize multiple inclusion of header files if they appear bracketized
 *      in '#ifndef SYMBOL ... #endif' pairs. Avoid multiple inclusion of the
 *      file if 'SYMBOL' is still defined.
 *
 *  1.4 Environment
 *
 *    CONFIG_JSCPP_ENVINCPATH:
 *      Name of the environment variable that contains a list of include
 *      directories, separated by ':'.
 *
 */

#ifndef CONFIG_JSCPPLIB_LANGUAGELEVEL
#define CONFIG_JSCPPLIB_LANGUAGELEVEL      3 /*  default is extended ANSI  */
#endif

/* ***  1.1 Additional symbols  *** */
#if defined(__GNUC__)
/*  These are GNU extensions  */
#define CONFIG_JSCPP_INCLEVEL  /*  optional   */
#define CONFIG_JSCPP_BASEFILE  /*  required!  */
#endif

#if defined(__TIMESTAMP__)
#define CONFIG_JSCPP_TIMESTAMP  /*  optional   */
#endif

/* ***  1.2 Additional commands  *** */
#if defined(__GNUC__) || defined(__sgi) || defined(__sun)
#define CONFIG_JSCPP_ASSERTION   /*  optional   */
#endif

#define CONFIG_JSCPP_INCNEXT    /*  required!  */

/*  Maybe needed in the future...  */
#define CONFIG_JSCPP_SINCLUDE       /*  optional   */

#define CONFIG_JSCPP_DEFIMM      /*  required!  */

#if defined(__sgi)
#define CONFIG_JSCPP_SYSVCMD        /*  optional   */
#endif

#define CONFIG_JSCPP_WARNING        /*  required!  */

#if defined(__GNUC__)
#define CONFIG_JSCPP_LAZY_LINECMD       /*  optional   */
#endif

/* ***  1.3 Behaviour  *** */
#define CONFIG_JSCPP_OPTLINES    /*  optional   */

#define CONFIG_JSCPP_OPTINC   /*  optional   */

/*  Compile the simple includes into the executable  */
#define CONFIG_JSCPP_PIGGYBACKINCLUDES     "~STDINC"

/* **************************************************************** */
#ifndef CONFIG_JSCPP_INCLUDEDIR
/*  For function dependency analysis, some very crude include files
 *   are sufficient to keep things going, e. g. the define "FILE" as
 *   a typedef etc. This makes us independent of the system include
 *   which normally don't work unless one has a whole bunch of
 *   predefined macros.
 */
#ifdef CONFIG_JSCPP_PIGGYBACKINCLUDES
#define CONFIG_JSCPP_INCLUDEDIR     CONFIG_JSCPP_PIGGYBACKINCLUDES /* ":/usr/include" */
#else
/*  Well, this at least looks like some reasonable standard otherwise  */
#define CONFIG_JSCPP_INCLUDEDIR       "/usr/include"
#endif
#endif

#ifndef CONFIG_JSCPP_ENVINCPATH
/*  We check this environment variable each time we open a stream. If set,
 *   it is assume to contain a colon separated list of directories to be
 *   used as standard includes instead of CONFIG_JSCPP_INCLUDEDIR
 */
#define CONFIG_JSCPP_ENVINCPATH       "JSCC_INCPATH"
#endif

#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

/* ***  Symbol to recognize it's *me* that is running  *** */
#define JSCPP_MACRO      "__JSCPP__"

/* ***********	 Extracted definitions of library code  ********* */
/* ----------- FILE: jsppcmds.c           ------------ */
#ifdef JSCPP_INTERNAL
#define PP_OP_ELLIPSIS   (PTK_USERTOKEN1 +  1)
#endif

/* ----------- FILE: jsppmacdef.c         ------------ */
/* ***  Additional tokens for routines  *** */
#define PTK_EMPTY        (CTK_EOF - 1) /*  Empty token -- Ignore  */

#define PTK_MACROARG0    (CTK_EOF - 2) /* This and below for macro arguments */

#define PTK_ISMACROARG(tokValue)        ((tokValue) <= PTK_MACROARG0)
#define PTK_GENMACROARG(nr)             (PTK_MACROARG0 - (nr))
#define PTK_GETMACROARG(tokValue)       (PTK_MACROARG0 - (tokValue))

/*  Special preprocessor's token  */
#define PTK_STRINGIZE    (CTK_USERTOKEN + 0)  /*  "# arg"  */
#define PTK_CONCAT       (CTK_USERTOKEN + 1)  /*  "arg1 ## arg2"  */

#define PTK_IDENTNOEXP   (CTK_USERTOKEN + 2)

#define PTK_LINEDIRSTART (CTK_USERTOKEN + 3) /*  Starts a "#line" directive */
#define PTK_SUBTOKENS    (CTK_USERTOKEN + 4) /* Use TK_Ptr field for tokens */

/* ****************************************** */
#define PTK_FIRSTINTERN  (CTK_USERTOKEN + 10)
/*
 *  To recognize a token sequence that should not
 *   be interpreted from the next stage (parser).
 */
#define PTK_STARTIGN     (CTK_USERTOKEN + 10)
#define PTK_ENDIGN       (CTK_USERTOKEN + 11)

/*  Ignore the next block in curly brackets  */
#define PTK_IGNBLOCK     (CTK_USERTOKEN + 12)

/*   To mark a special token sequence to the parser  */
#define PTK_STARTSPC     (CTK_USERTOKEN + 15)
#define PTK_ENDSPC       (CTK_USERTOKEN + 16)

#define PTK_LASTINTERN   (CTK_USERTOKEN + 19)
/* ****************************************** */

#define PTK___LINE       (CTK_USERTOKEN + 20)  /* "__LINE__"  */
#define PTK___FILE       (CTK_USERTOKEN + 21)  /* "__FILE__"  */
#ifdef CONFIG_JSCPP_INCLEVEL
# define PTK___INCLEVEL  (CTK_USERTOKEN + 22)  /* "__INCLUDE_LEVEL__"  */
#endif
#ifdef CONFIG_JSCPP_BASEFILE
# define PTK___BASEFILE  (CTK_USERTOKEN + 23)  /* "__BASE_FILE__"  */
#endif

/* ****************************************** */
#ifdef JSCPP_INTERNAL
/*  Is used in jsppcmds.c to define the preprocessor operators  */
#define PTK_USERTOKEN1   (CTK_USERTOKEN + 30)
#endif

/* ****************************************** */
/*   This is the first token value to be used outside  */
#define PTK_USERTOKEN    (CTK_USERTOKEN + 50)

/* ----------- FILE: jsppmacro.c          ------------ */
#ifdef JSCPP_INTERNAL
/*  Flags for routine jspptkmacro and _jspptkbasic  */
#define GTK_NO_EXP           (GCTK_FLAG_USER)
#define GTK_NO_CONCAT        (GCTK_FLAG_USER << 1)
#define GTK_PREPROCOPS       (GCTK_FLAG_USER << 2)

/*  Flags for routine _jsppexpand  */
#define EXPM_DEFINED_MACRO   (GCTK_FLAG_USER << 4)
#endif
#define EXPM_DEFEREXPANSION  (GCTK_FLAG_USER << 5)
#ifdef JSCPP_INTERNAL
#define EXPM_SAVE_ORIG       (GCTK_FLAG_USER << 6)
#define EXPM_SAVE_ORIG2      (GCTK_FLAG_USER << 7)
#endif

/* ----------- FILE: jsppmacsub.c         ------------ */
/* **  For macro definitions  ** */
typedef struct _jsppmacentry_t _jsppmacentry_t;
struct _jsppmacentry_t {
    _jsppmacentry_t *ME_Next;
    char           *ME_Ident; /*  The macro identifier  */
    long            ME_Flags;
    int             ME_NArgs;

    /*  Position where it was defined  */
    char           *ME_File;
    int             ME_LineNr;

    /*  The actual tokens  */
    jsctoken_t      ME_Token[1];
};

#define MAFLAG_PROTECTED      (1 <<  0)  /*  Not undefinable  */
#define MAFLAG_INTERNAL       (1 <<  1)  /*  Internal macro   */
#define MAFLAG_SYSTEMHEADER   (1 <<  2)  /*  Defined in a system header file  */
#define MAFLAG_USERHEADER     (1 <<  3)  /*  Defined in a user header file  */
#define MAFLAG_ONEARGUMENT    (1 <<  4)  /*  Ignore ',' as argument separator  */
#define MAFLAG_DEFEREXPANSION (1 <<  5)  /*  Defer expansion till end  */
#define MAFLAG_DISABLED       (1 <<  6)  /*  Expansion disabled  */
#define MAFLAG_DISABLEDINFILE (1 <<  7)  /*  Expansion within file disabled  */
#define MAFLAG_EVALIMMEDIATE  (1 <<  8)  /*  Evaluate immediately on definition */
#define MAFLAG_EXPANDORIG     (1 <<  9)  /*  Expand macro in original stream  */
#define MAFLAG_EXPANDORIGFULL (1 << 10)  /*  Expand macro in original stream completely  */
#define MAFLAG_UNDEFINED      (1 << 11)  /*  Macro not defined, only flags are used  */
#define MAFLAG_VARARG         (1 << 12)  /*  Macro is variadic, last argument collects everything  */

#define MAFLAG_USED           (1 << 16) /*  Mark that it was used  */

/*  Size of macro hash table  */
#ifndef JSPP_MACTABSIZE /*  Allow changes on the fly  */
# define JSPP_MACTABSIZE 1024
#endif

#ifdef CONFIG_JSCPP_ASSERTION
typedef struct _jspppredanswer_t _jspppredanswer_t;
struct _jspppredanswer_t {
    _jspppredanswer_t *PA_Next;
    char               PA_Answer[1];
};

typedef struct _jspppredicate_t _jspppredicate_t;
struct _jspppredicate_t {
    _jspppredicate_t  *PR_Next;
    _jspppredanswer_t *PR_Answers;
    char               PR_Predicate[1];
};
#endif

/* ----------- FILE: jsppqueue.c          ------------ */
typedef struct jsqueuestream_t jsqueuestream_t;

/* ----------- FILE: jsppsubs.c           ------------ */
#define jspptkprint(jps, fp, token)           jspptkprint2(jps, fp, token, -2)

/* ----------- FILE: jsppsymbs.c          ------------ */
#define JSCPP_USE_PREDEFINES

/* ----------- FILE: jsppfile.c           ------------ */
/* ***  Structure to store open file  *** */
#define IF_FLAG_EOF            (1 <<  0)
#define IF_FLAG_MACRO          (1 <<  1)
#define IF_FLAG_PPSTREAM       (1 <<  2)
#define IF_FLAG_NOEXP          (1 <<  3)
#define IF_FLAG_ORIGTOK        (1 <<  4)
#define IF_FLAG_COPYTOKS       (1 <<  5)
#define IF_FLAG_EXPANDORIG     (1 <<  6)
#define IF_FLAG_EXPANDORIGFULL (1 <<  7)

#define IF_FLAG_EXPANDINGARG   (1 <<  8)
#define IF_FLAG_DISABLEEXPAND  (1 <<  9)
#define IF_FLAG_LINETOKENS     (1 << 10)

#ifdef JSCPP_INTERNAL
#ifdef CONFIG_JSCPP_OPTINC
/*  Description:
 *   NONSPC1:    Used internally to set NONSPC *after* a non-space
 *               token has been read.
 *   NONSPC:     Marks if a non-space token has been read in the file.
 *   INCONSTRUCT: Within '#ifndef' .. 'endif' construct.
 *   CONSTRUCT:  '#ifndef SYMBOL' has been seen in file with solely
 *               blanks prior to '#ifndef'.
 */
# define IF_FLAG_NONSPC1      (1 << 12)
# define IF_FLAG_NONSPC       (1 << 13)
# define IF_FLAG_INCONSTRUCT  (1 << 14)
# define IF_FLAG_CONSTRUCT    (1 << 15)
#endif
#endif

#define IF_SYSINCLUDE         (1 << 16)

#ifdef JSCPP_INTERNAL
#define IF_ISMACRO(iFile)     ((iFile)->IF_Flag & IF_FLAG_MACRO)
#define IF_ISFILE(iFile)      (!IF_ISMACRO(iFile))
#define IF_ISATEOF(iFile)     ((iFile) == NULL ||			\
                               ((iFile)->IF_Flag & IF_FLAG_EOF))
#endif

typedef struct {
    jsctoken_t *MA_Token;
    int         MA_UseCount;
} _jsppmacarg_t;

/* **  For an input channel  ** */
typedef struct jsppfile_t jsppfile_t;
struct jsppfile_t {
    jsppfile_t *IF_Next;
    int         IF_Flag;
    const char *IF_Name;   /*  The file name or the macro identifier  */
    union {
	extfile_t       *if_FFPtr;
	jsctoken_t      *if_MTokens; /*  Tokens for current macro  */
	jsqueuestream_t *if_QStream;
#define IF_FFPtr    if_c1.if_FFPtr
#define IF_MTokens  if_c1.if_MTokens
#define IF_QStream  if_c1.if_QStream
    } if_c1;
    union {
	int         if_LineNr; /*  Contains corrected line number  */
	_jsppmacarg_t *if_MArgs;    /*  Arguments of current macro  */
#define IF_LineNr   if_c2.if_LineNr
#define IF_MArgs    if_c2.if_MArgs
    } if_c2;

    int IF_CondLevel;

#ifdef CONFIG_JSCPP_OPTINC
    const char *IF_Symbol;
#endif
};

/*  For include files  */
typedef struct {
    const char *IN_Directory;
    int         IN_Flags;
} jsppincdir_t;

#define JPP_FILE_FSYSINC           (1 << 0)
#define JPP_FILE_FNODIRSEP         (1 << 1)
#define JPP_FILE_FADDTOTOP         (1 << 2) /*  is cleared after adding at the front  */
#define JPP_FILE_FALLOCATED        (1 << 3)

/* **  For routine jspppushfile  ** */
#define JPP_PUSHFILE_FNOCURRDIR    (1 << 0)
#define JPP_PUSHFILE_FINCNEXT      (1 << 1)
#define JPP_PUSHFILE_FSINCLUDE     (1 << 2)
#define JPP_PUSHFILE_FNOLINEDIR    (1 << 3)

#define jspppushtokens(jps, tokens)    _jspppushmacro(jps, NULL, tokens, NULL, 0)

/* ----------- FILE: jsppinit.c           ------------ */
#ifdef CONFIG_JSCPP_OPTINC
typedef struct _jspponcefile_t _jspponcefile_t;
struct _jspponcefile_t {
    _jspponcefile_t *OF_Next;
    const char      *OF_Name;
    const char      *OF_Symbol;
};
#endif

typedef struct {
    int JSOQ_Mode;
    jctokfifo_t JSOQ_OQueue;
    jctokfifo_t JSOQ_MacOQueue;
} _jsppoqueue_t;

typedef struct jsppstream_t jsppstream_t;
struct jsppstream_t {
    int JSPP_Mode;

    int JSPP_OLineNr;

    int JSPP_CTokMode;
    int JSPP_LastToken, JSPP_LastToken2;

    const jsctokops_t *JSPP_Ops;

    jsppincdir_t *JSPP_IncDir;
    int JSPP_NIncDir;

    /*  The current stream (file or macro expansion) we are reading from  */
    jsppfile_t *JSPP_File;
    int JSPP_NFiles;

    ifstack_t JSPP_Conds;

    /*  For all macro definitions  */
    _jsppmacentry_t *JSPP_Macros[JSPP_MACTABSIZE];

    jctokfifo_t JSPP_Queue;

    _jsppoqueue_t *JSPP_OQueue;
    int JSPP_QueueLevel, JSPP_MaxQueueLevel;

#ifdef CONFIG_JSCPP_ASSERTION
    _jspppredicate_t *JSPP_Predicates;
#endif

    /*  Called to print error messages  */
    void (*JSPP_PrintError)(void *user, const char *message);

    /*  If set, is called if pragma line encountered  */
    int (*JSPP_DoPragma)(jsppstream_t *jps, jsctoken_t *pToken, int n,
			 void *user);

    void *JSPP_User;

    jsctoken_t (*JSPP_SpecToken)(jsppstream_t *jps, int value);

    /*  When non-zero, is called every time a new file is opened  */
    int (*JSPP_OpenFile)(jsppstream_t *jps, const char *name, int incLevel,
			 int flag);

    int JSPP_NLLeftOver, JSPP_NLLeftOver2;
    int JSPP_Ignored;

    jssharestr_t     JSPP_FileNames;

#ifdef CONFIG_JSCPP_OPTINC
    _jspponcefile_t *JSPP_OnceFiles;
#endif
};

/* **  For JSPP_Mode  ** */
#define JPPM_LINEDIR               (1 <<  0) /*  Print '#line' directives in output  */
#define JPPM_PRESCOMM              (1 <<  1) /*  Preserve comments in output  */
#define JPPM_PRESSPACE             (1 <<  2) /*  Preserve spaces in output  */
#ifdef CONFIG_JSCPP_OPTLINES
# define JPPM_SAVENEWLN            (1 <<  3) /*  Save newlines by using '#line' directives */
#endif
#define JPPM_PPMACARGS             (1 <<  4) /*  Do preprocessing within macro args - not well tested  */
#define JPPM_PRWARNING             (1 <<  5)
/*  Do not accept redefinition of macros to a different token sequence  */
#define JPPM_STRICT                (1 <<  6)

/* Accept several errors in preprocessor lines and generate solely
 * warnings instead of errors (i. e. for unknown commands, garbage
 * at line end, illegal '#' and '##' tokens in macro definitions,
 * empty arguments to macro, illegal usages of '##' operators).
 * This is especially useful if the preprocessor parses things other
 * than valid C-code.
 */
#define JPPM_LAZYSTYLE             (1 <<  8)

#define JPPM_MACVARARG             (1 <<  9)

/*  Return original tokens in non-system includes  */
#define JPPM_EXPUSERFILES          (1 << 10)

/*  Expand macros in original stream in non-system includes  */
#define JPPM_EXPUSERMACROS         (1 << 11)

#define JPPM_FLAG_SEENDEFEREDMAC   (1 << 16)

#define jsppopen(jps, filename, ops, omode, mode)		\
    jsppopen2(jps, NULL, filename, NULL, ops, omode, mode)

/* ***  To simplify writing applications that parse cmdline options  *** */
typedef struct {
    const char *JSPES_String;
    int JSPES_Type;
} _jsppenvstype_t;

#define JSPE_INDEX_TINCS       0
#define JSPE_INDEX_TDEFS       1
#define JSPE_INDEX_TKEY        2
#define JSPE_INDEX_TASS        3
#define JSPE_INDEX_TSTARTFILE  4

#define JSPE_INDEX_MAX         5

#define jsppenvincdir(jppenv, dir, flag)  _jsppenvadd(jppenv, JSPE_INDEX_TINCS, dir, flag)
#define jsppenvdefine(jppenv, string)     _jsppenvadd(jppenv, JSPE_INDEX_TDEFS, string, 0)

typedef struct {
    long JSPE_Mode, JSPE_IMode;
    long JSPE_CTokMode;

    _jsppenvstype_t *JSPE_STypes[JSPE_INDEX_MAX];
    int JSPE_NSTypes[JSPE_INDEX_MAX];
} jsppenviron_t;

#ifdef CONFIG_JSCPP_ASSERTION
#define JSCPP_OPTSTRING_ADDASS  "A:"
#else
#define JSCPP_OPTSTRING_ADDASS   ""
#endif

/*  These are the standard options understood by jsppenvhandle  */
#define JSCPP_OPTSTRING       "I;i:D:U;x:PCWS:e;K:V"  JSCPP_OPTSTRING_ADDASS

#define jsppsetprinterr(jps, printError)   ((jps)->JSPP_PrintError = (printError))

/* ----------- FILE: jsppqueue.c          ------------ */
#define JSQS_TYPE_PPSTREAM    1
#define JSQS_TYPE_USER        2
#define JSQS_TYPE_FMASK       0xff

#define JSQS_TYPE_FFREEQUEUE  (1<<8)

struct jsqueuestream_t {
    int JSQS_Type;
    jctokfifo_t *JSQS_Queue;

    union {
	jsppstream_t *jsqs_PPStream;
	int (*jsqs_GetIndex)(void *user, jctokfifo_t *queue);
    } c;
#define JSQS_PPStream         c.jsqs_PPStream
#define JSQS_GetIndex         c.jsqs_GetIndex

    void *JSQS_User;
};

#define jsqsgetqueue(jsqs)     ((jsqs)->JSQS_Queue)

#define jsqsgetppstream(jsqs)  (((jsqs)->JSQS_Type & JSQS_TYPE_FMASK) == \
				JSQS_TYPE_PPSTREAM ? (jsqs)->JSQS_PPStream : \
				NULL)


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: jsppfile.c            ***** */
Prototype int            _jspppushmacro(jsppstream_t *jps, const char *name,
                                        jsctoken_t *pTokens,
                                        _jsppmacarg_t *args, int flags);
Prototype int            jspppushfile(jsppstream_t *jps, const char *name,
                                      int mode);
Prototype int            jspppushppstream(jsppstream_t *jps,
                                          jsqueuestream_t *jsqs, int mode);
Prototype int            jspppushinplevel(jsppstream_t *jps);
Prototype int            jspppopinplevel(jsppstream_t *jps);
Prototype int            jsppaddincdir(jsppstream_t *jps, const char *name,
                                       int flags);
Prototype int            jsppaddincpath(jsppstream_t *jps,
                                        const char *includepath, int flags);
Prototype void           jsppshowinc(jsppstream_t *jps, FILE *fp, int mode);
Prototype int            jsppgetfilepos(jsppstream_t *jps,
                                        const char **pFileName, int mode);
Prototype void           jsppfreeoncefl(jsppstream_t *jps);

/* ***** File: jsppget.c             ***** */
Prototype jsctoken_t	 jsppgettoken(jsppstream_t *jps, jsctoken_t *pOToken, int mode);

/* ***** File: jsppinit.c            ***** */
Prototype int            jsppinit(int mode);
Prototype jsppstream_t  *jsppopen2(jsppstream_t *jps, jsppenviron_t *jppenv,
				   const char *filename, const char *incpath,
				   const jsctokops_t *ops, int omode, int mode);
Prototype int            jsppclose(jsppstream_t *jps, int flag);
Prototype void           jsppdelmacdef(jsppstream_t *jps, int flag);
Prototype void           jsppenvclear(jsppenviron_t *jppenv);
Prototype int            jsppenvhandle(jsppenviron_t *jppenv, int opt, const char *string);
Prototype int            _jsppenvadd(jsppenviron_t *jppenv, int nr, const char *string,
				     int type);
Prototype void           jsppenvfree(jsppenviron_t * jppenv);
Prototype const char     JSPPStdOptsHelp[];
Prototype const char     JSPPStdOptsHelp2[];
Prototype int            JSPPLanguageLevel;
Prototype const char   **JSPPPragmaIntroducer;

/* ***** File: jsppmacdef.c          ***** */
Prototype int            jsppdefinemacro(jsppstream_t *jps,
					     const char *string, int mode);
Prototype int            jsppdefundefmac(jsppstream_t *jps, const char *name,
                                         const jsctoken_t *pToken, int n,
                                         long mFlags);
Prototype int            jsppsetmacflags(jsppstream_t *jps, const char *name,
                                         long mFlags);
Prototype void           jsppshowmac(jsppstream_t *jps, FILE *fp, int mode);

/* ***** File: jsppmacro.c           ***** */
Prototype jsctoken_t     jspptkmacro(jsppstream_t *jps,
                                     jsctoken_t *pOToken, int mode);

/* ***** File: jsppmacsub.c          ***** */
Prototype unsigned int   jspptkhash(const char *word, unsigned int tabSize);
Prototype int            jsppnewpred(jsppstream_t *jps,
				     const char *predicate,
				     const char *answer);
Prototype int            jsppdelpred(jsppstream_t *jps,
				     const char *predicate,
				     const char *answer);
Prototype void           jspplistpreds(jsppstream_t *jps, FILE *fp);

/* ***** File: jsppqueue.c           ***** */
Prototype jsqueuestream_t *jsqscreateppstream(jsppstream_t *jps,
					      jctokfifo_t *queue, int qmode);
Prototype jsqueuestream_t *jsqscreatequeuestream(jctokfifo_t *queue,
						 int (*routine)(void *,
								jctokfifo_t *),
						 void *user);
Prototype void           jsqsfree(jsqueuestream_t *jsqs);
Prototype int            jsqsget(jsqueuestream_t *jsqs, int mode);

/* ***** File: jsppstubinc.c         ***** */
Prototype const unsigned char JSPPStubIncludes[];

/* ***** File: jsppsubs.c            ***** */
Prototype int            jspptkprint2(jsppstream_t *jps, FILE *fp,
				      jsctoken_t token, int currLine);
Prototype void           jsppseterrcode(int errnr, int erroff, const char *string,
					jctokfifo_t *queue, int indx);
Prototype void		 jsppprinterr(jsppstream_t *jps, char *message, int mode, ...);
Prototype int            jsppflushtokens(jsppstream_t *jps, jctokfifo_t *queue,
					 FILE *fp, int mode);
Prototype int            jsppisusrinclude(jsppstream_t *jps, int mode);
Prototype const char    *jspptok2string(jsctoken_t token);

/* ***** File: jsppversion.c         ***** */
Prototype const long     JSCPPNVersion;


# undef Prototype   /* I like to be clean  */

# ifdef JSCPP_INTERNAL
/*  For C++ compatibility   */
#  ifdef __cplusplus
#   define IPrototype extern "C"
#  else
#   define IPrototype extern
#  endif
/* ***** File: jsppcmds.c            ***** */
IPrototype int		 _jsppdoline(jsppstream_t *jps, int lineNr, char *fileName);
IPrototype int           _jsppdoerror(jsppstream_t *jps, jsctoken_t *pToken, int n);
IPrototype int           _jsppdowarn(jsppstream_t *jps, jsctoken_t *pToken, int n);
IPrototype int           _jsppdoifdef(jsppstream_t *jps, char *name, int flag);
IPrototype int           _jsppdoif(jsppstream_t *jps, jsctoken_t *pToken, int n,
				   int isif, int flag);
IPrototype int		 _jsppdocondcomp(jsppstream_t *jps, int type,
					 int condition);
IPrototype int           _jsppdopragma(jsppstream_t *jps, jsctoken_t *pToken,
				       int n);
IPrototype const jsctokops_t _jsppOperators[];

/* ***** File: jsppfile.c            ***** */
IPrototype int           _jspppopistack(jsppstream_t *jps);

/* ***** File: jsppmacbas.c          ***** */
IPrototype jsctoken_t    _jspptkbasic(jsppstream_t *jps,
                                      jsctoken_t *pOToken, int mode);
IPrototype jsctoken_t    _jspptkbasic2(jsppstream_t *jps,
                                       jsctoken_t *pOToken, int mode);

/* ***** File: jsppmacro.c           ***** */
IPrototype int           _jsppexpand(jsppstream_t *jps,
                                     const char *name, int mode);

/* ***** File: jsppmacsub.c          ***** */
IPrototype void          _jsppfreemacdef(jsppstream_t *jps,
                                         _jsppmacentry_t *entry);
IPrototype _jsppmacentry_t *_jsppfindmacentry(jsppstream_t *jps,
					      const char *name);
IPrototype int           _jsppppdefass(jsppstream_t *jps, int mode,
                                       int flagAss, jctokfifo_t *oQueue);

/* ***** File: jsppsubs.c            ***** */
IPrototype jsctoken_t    _jsppstringize(jsctoken_t *pToken, int mode);
IPrototype jsctoken_t    _jsppgetnnewlines(int n);
IPrototype void          _jspprmlinecont(char *string);

/* ***** File: jsppsymbs.c           ***** */
IPrototype int           _jspppredefine(jsppstream_t *jps, int level);


#  undef IPrototype   /* I like to be clean  */
# endif

#endif
