/*JS*********************************************************************
*
*    Program : JSENDIAN
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Define some simple macros to read/write integers in
*              big/little endian.
*              Could be easily optimized for some architectures where
*              the endianess is fixed and we can read/write unaligned
*              data.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

#include <jsstdint.h>

/*  Make sure we read the values properly by casting the array to the
 *   right pointer type
 */
#define _JSECST(a)    (unsigned char *)(a)

/* ***  For 8 bits  *** */
#define JS_INT82CHAR_BE(value, array)		\
  (* _JSECST(array)      = ((value) & 0xff))

#define JS_CHAR2INT8_BE(array)		      \
  ((jsint8_t)* _JSECST(array))

#define JS_INT82CHAR_LE(value, array)		    \
  (* _JSECST(array)      = ((value) & 0xff))

#define JS_CHAR2INT8_LE(array)		     \
  ((jsint8_t)* _JSECST(array))

#define JS_INT82CHAR_LEN      1

/* ***  For 16 bits  *** */
#define JS_INT162CHAR_BE(value, array)		    \
  (* _JSECST(array)      = (((value) >> 8) & 0xff), \
   *(_JSECST(array) + 1) = ( (value)       & 0xff)  \
   )
#define JS_CHAR2INT16_BE(array)		      \
  (((jsint16_t)* _JSECST(array)      <<  8) | \
   ((jsint16_t)*(_JSECST(array) + 1)      )   \
   )

#define JS_INT162CHAR_LE(value, array)		    \
  (* _JSECST(array)      = ( (value)       & 0xff), \
   *(_JSECST(array) + 1) = (((value) >> 8) & 0xff)  \
   )
#define JS_CHAR2INT16_LE(array)		     \
  (((jsint16_t)* _JSECST(array)          ) | \
   ((jsint16_t)*(_JSECST(array) + 1) << 8)   \
   )

#define JS_INT162CHAR_LEN      2

/* ***  For 32 bits  *** */
#define JS_INT322CHAR_BE(value, array)		     \
  (* _JSECST(array)      = (((value) >> 24) & 0xff), \
   *(_JSECST(array) + 1) = (((value) >> 16) & 0xff), \
   *(_JSECST(array) + 2) = (((value) >>  8) & 0xff), \
   *(_JSECST(array) + 3) = ( (value)        & 0xff)  \
   )
#define JS_CHAR2INT32_BE(array)		       \
  (((jsint32_t)* _JSECST(array)      << 24) | \
   ((jsint32_t)*(_JSECST(array) + 1) << 16) | \
   ((jsint32_t)*(_JSECST(array) + 2) <<  8) | \
   ((jsint32_t)*(_JSECST(array) + 3)      )   \
   )

#define JS_INT322CHAR_LE(value, array)		     \
  (* _JSECST(array)      = ( (value)        & 0xff), \
   *(_JSECST(array) + 1) = (((value) >>  8) & 0xff), \
   *(_JSECST(array) + 2) = (((value) >> 16) & 0xff), \
   *(_JSECST(array) + 3) = (((value) >> 24) & 0xff)  \
   )
#define JS_CHAR2INT32_LE(array)		       \
  (((jsint32_t)* _JSECST(array)           ) | \
   ((jsint32_t)*(_JSECST(array) + 1) <<  8) | \
   ((jsint32_t)*(_JSECST(array) + 2) << 16) | \
   ((jsint32_t)*(_JSECST(array) + 3) << 24)   \
   )

#define JS_INT322CHAR_LEN       4

/* ***  For 64 bits  *** */
#define JS_INT642CHAR_BE(value, array)		     \
  (* _JSECST(array)      = (((value) >> 56) & 0xff), \
   *(_JSECST(array) + 1) = (((value) >> 48) & 0xff), \
   *(_JSECST(array) + 2) = (((value) >> 40) & 0xff), \
   *(_JSECST(array) + 3) = (((value) >> 32) & 0xff), \
   *(_JSECST(array) + 4) = (((value) >> 24) & 0xff), \
   *(_JSECST(array) + 5) = (((value) >> 16) & 0xff), \
   *(_JSECST(array) + 6) = (((value) >>  8) & 0xff), \
   *(_JSECST(array) + 7) = ( (value)        & 0xff)  \
  )
#define JS_CHAR2INT64_BE(array) \
  (((jsint64_t)* _JSECST(array)      << 56) | \
   ((jsint64_t)*(_JSECST(array) + 1) << 48) | \
   ((jsint64_t)*(_JSECST(array) + 2) << 40) | \
   ((jsint64_t)*(_JSECST(array) + 3) << 32) | \
   ((jsint64_t)*(_JSECST(array) + 4) << 24) | \
   ((jsint64_t)*(_JSECST(array) + 5) << 16) | \
   ((jsint64_t)*(_JSECST(array) + 6) <<  8) | \
   ((jsint64_t)*(_JSECST(array) + 7)      )   \
  )

#define JS_INT642CHAR_LE(value, array)		     \
  (* _JSECST(array)      = ( (value)        & 0xff), \
   *(_JSECST(array) + 1) = (((value) >>  8) & 0xff), \
   *(_JSECST(array) + 2) = (((value) >> 16) & 0xff), \
   *(_JSECST(array) + 3) = (((value) >> 24) & 0xff), \
   *(_JSECST(array) + 4) = (((value) >> 32) & 0xff), \
   *(_JSECST(array) + 5) = (((value) >> 40) & 0xff), \
   *(_JSECST(array) + 6) = (((value) >> 48) & 0xff), \
   *(_JSECST(array) + 7) = (((value) >> 56) & 0xff)  \
   )
#define JS_CHAR2INT64_LE(array)		      \
  (((jsint64_t)* _JSECST(array)           ) | \
   ((jsint64_t)*(_JSECST(array) + 1) <<  8) | \
   ((jsint64_t)*(_JSECST(array) + 2) << 16) | \
   ((jsint64_t)*(_JSECST(array) + 3) << 24) | \
   ((jsint64_t)*(_JSECST(array) + 4) << 32) | \
   ((jsint64_t)*(_JSECST(array) + 5) << 40) | \
   ((jsint64_t)*(_JSECST(array) + 6) << 48) | \
   ((jsint64_t)*(_JSECST(array) + 7) << 56)   \
   )

#define JS_INT642CHAR_LEN      8
