/*JS*********************************************************************
*
*    Program : JSPARSE.H
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Include file for C parser library "jsparse".
*
*              Copyright (C) 1994-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef JSPARSE_H /*   File must not be included more than once!   */
#define JSPARSE_H 1

/*********     INCLUDES 					*********/

#include <jscpplib.h>

/*********     DEFINES						*********/
/*
 *  Parser configuration
 *
 *  1. Language extensions
 *
 *    CONFIG_JSPARSE_TYPEOF:
 *      Allow C constructions "__typeof__(..)" that are semantically
 *      equivalent to a type name.
 *
 *    CONFIG_JPARSE_CONSTANT:
 *      New keyword "__constant__" with same syntactic value as an integer
 *      constant, i. e. the parser sees this expression as a constant and not
 *      as a variable. Can be used in conjunction with "#pragma keyword"
 *      to introduce identifiers as constant values so that parser does not
 *      have to deal with them.
 *
 *  2. Specialties
 *
 *    CONFIG_JSPARSE_IGNORE:
 *      Protect token sequences from processing by the parser by enclosing
 *      them with the "__ignore__" macro. At the end, it evaluates to nothing,
 *      but the enclosed tokens will not be seen by the parser.
 *
 *    CONFIG_JSPARSE_YNAME
 *      If defined, changes the prefix 'yy' in yacc defined names to
 *      CONFIG_JSCPP_YNAME to avoid a name clash when having multiple
 *      parsers linked in the final executable. May not work if your
 *      yacc itself introduces #define's for its symbols. Unfortunately
 *      yacc usually produces a lot more of global symbols that are visible
 *      It seems that the current solution only works reliable with bison.
 */
#define CONFIG_JSPARSE_IGNORE

#define CONFIG_JSPARSE_CONSTANT

#define CONFIG_JSPARSE_TYPEOF

/*  Other definitions  */
#define CONFIG_JSPARSE_MACRO     "__JSPARSE__"

#define CONFIG_JSPARSE_VERSION   10102

/*  Make the wrapper includes available  */
#define CONFIG_JSPARSE_PIGGYBACKINCLUDES   "~WRAPINC"

#define CONFIG_JSPARSE_INCLUDEDIR       CONFIG_JSPARSE_PIGGYBACKINCLUDES

#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

/* ***********	 Extracted definitions of library code  ********* */
/* ----------- FILE: jsparsexpr3.c        ------------ */
/*  Returned values from jsparsgetvartype* routines  */
#define JPA_VARTYPE_VOID     (1 << 0)
#define JPA_VARTYPE_CHAR     (1 << 1)
#define JPA_VARTYPE_SHORT    (1 << 2)
#define JPA_VARTYPE_INT      (1 << 3)
#define JPA_VARTYPE_LONG     (1 << 4)
#define JPA_VARTYPE_LONGLONG (1 << 5)

#define JPA_VARTYPE_FLOAT    (1 << 8)
#define JPA_VARTYPE_DOUBLE   (1 << 9)

#define JPA_VARTYPE_WCHAR    (1 << 12)

#define JPA_VARTYPE_STRUCT   (1 << 16)
#define JPA_VARTYPE_UNION    (1 << 17)
#define JPA_VARTYPE_ENUM     (1 << 18)

/* ----------- FILE: jsparsgram2.c        ------------ */
#define jsparsopen(jpa, jps, filename, omode, mode)	\
    jsparsopen2(jpa, jps, NULL, filename, omode, mode)

/* ----------- FILE: jsparsscan.c         ------------ */
typedef struct jsparsvartab_t jsparsvartab_t;
typedef struct jsparsvardef_t jsparsvardef_t;

/* ----------- FILE: jsparssymbs.c        ------------ */
#define JSPARSE_USE_PREDEFINES

/* ----------- FILE: jsparssubs.c         ------------ */
/*   Type of tree nodes  */
typedef enum {
    JPACT_EMPTY = 0,

    /* **  For declarations  ** */

    /*  Declaration types  */
    JPACT_DECLARATION, JPACT_FUNCDEF, JPACT_PARDECL, JPACT_DECLARATOR,

    /*  Storage class specifiers  */
    JPACT_AUTO, JPACT_REGISTER, JPACT_STATIC, JPACT_EXTERN, JPACT_TYPEDEF,

    /*  Type specifiers  */
    JPACT_VOID, JPACT_CHAR, JPACT_SHORT, JPACT_INT, JPACT_LONG, JPACT_SIGNED, JPACT_UNSIGNED,
    JPACT_FLOAT, JPACT_DOUBLE, JPACT_STRUCT, JPACT_UNION, JPACT_ENUM,

    /*  Type qualifiers  */
    JPACT_CONST, JPACT_VOLATILE,

    /*  Complex types  */
    JPACT_POINTER, JPACT_VECTOR, JPACT_FUNCCALL, JPACT_ELLIPSIS,
    JPACT_SMALLINT, JPACT_TYPEOF,

    /*  Comma lists and blocks  */
    JPACT_LIST, JPACT_BLOCK,

    /* **  Statements  ** */
    JPACT_COMPSTMT, JPACT_LABELEDSTMT,
    JPACT_SWITCH, JPACT_CASE, JPACT_DEFAULT, JPACT_CONTINUE, JPACT_BREAK,
    JPACT_IFELSE, JPACT_WHILE, JPACT_DOWHILE, JPACT_FOR, JPACT_RETURN,
    JPACT_LABEL, JPACT_GOTO,

    /* **  Expressions  ** */
    /*  Basic constants  */
    JPACT_IDENT, JPACT_INTCONST, JPACT_FLOATCONST,
    JPACT_CHARCONST, JPACT_WCHARCONST, JPACT_UCHARCONST,
    JPACT_STRING, JPACT_WSTRING, JPACT_USTRING,

    /*  Special node for type names and enumeration constants  */
    JPACT_TYPEDEFNAME, JPACT_ENUMCONST,

    /*  Simple expression  */
    JPACT_UNARYEXPR,

    /*  Assignment expression  */
    JPACT_ASS, JPACT_MULASS, JPACT_DIVASS, JPACT_MODASS, JPACT_ADDASS, JPACT_SUBASS,
    JPACT_LSHIFTASS, JPACT_RSHIFTASS, JPACT_ANDASS, JPACT_EXORASS, JPACT_ORASS,

    /*  Arithmetric/logical expressions  */
    JPACT_CONDEVAL, JPACT_LOR, JPACT_LAND, JPACT_EQUALS, JPACT_NEQUALS,
    JPACT_LOWER, JPACT_GREATER, JPACT_LOEQUAL, JPACT_GREQUAL,
    JPACT_OR, JPACT_EXOR, JPACT_AND, JPACT_LSHIFT, JPACT_RSHIFT,
    JPACT_ADD, JPACT_SUB, JPACT_TIMES, JPACT_DIV, JPACT_MODULO,

    /*  Unary operators  */
    JPACT_PREINC, JPACT_PREDEC, JPACT_POSTINC, JPACT_POSTDEC,
    JPACT_ADDRESS, JPACT_UPLUS, JPACT_UMINUS, JPACT_NOT, JPACT_LNOT,
    JPACT_MEMBER, JPACT_POINTTO,
    JPACT_SIZEOF, JPACT_CAST,

    /* **  Pragma lines  ** */
    JPACT_PRAGMALINE, /* general pragma line  */
    JPACT_PRAGMALINE2, /* pragma line with correct pragma introducer seen  */
    JPACT_PRAGMACONT

#ifdef CONFIG_JSCPPLIB_DEBUG
    /*  Highest allowed type  */
    , JPACT_MAXTYPE
#endif
} jsparscnodetype_e;

/* **  Tree to store C program  ** */
typedef struct jsparscnode_t jsparscnode_t;
struct jsparscnode_t {
    jsparscnode_t    *JPACN_Next;

    jsparscnodetype_e JPACN_Type;
    jsparscnode_t    *JPACN_Sub;

    /*  Start and end index in queue  */
    unsigned int      JPACN_Start, JPACN_End;

    /*  We send the original tokens first, then the preprocessed
     *   ones. So if we have a single token like IDENT where original
     *   and preprocessed one is different, JPACN_Start points to the
     *   original one, JPACN_End to the preprocessed one
     */
#define JPACN_IndexOrig  JPACN_Start
#define JPACN_IndexPP    JPACN_End
};

#define _jsparschknode(jpa, node)		\
    (((jpa)->JSPA_Mode & JPAM_FGETORIG) ?	\
     _jsparschknode2(jpa, node) : 0)

#define jsparsflushtokens(jpa, fp) jsparsflushtokens2(jpa,		\
						      (jpa)->JSPA_Stream, fp)

typedef struct {
    jsppenviron_t JPAE_PPEnviron;
    unsigned long JPAE_PAMode;
    unsigned long JPAE_WarnOpts;

} jsparsenviron_t;

#define jsparsenvgetpp(paenv)            (&((paenv)->JPAE_PPEnviron))
#define jsparsenvdefine(paenv, string)   jsppenvdefine(jsparsenvgetpp(paenv), string)

/*  The "W;" overwrites the single "W" later on  */
#define JSPARSE_OPTSTRING         "W;" JSCPP_OPTSTRING

#define jsparsseterrcode(errnr, erroff, queue, node)	\
    jsparsseterrcode2(errnr, erroff, NULL, queue, node)

/* ----------- FILE: jsparsexpr.c         ------------ */
typedef struct {
    unsigned long JSPE_Flags;

    /*  Only valid if flag JPA_VARTYPE_FLONGCONST or JPA_VARTYPE_FARRAYSIZE
     *  set
     */
    long JSPE_LongConst;

    jsparscnode_t *JSPE_Declarator; /*  without top node JPACT_DECLARATOR */
    jsparscnode_t *JSPE_Specifier;
    jsparsvartab_t *JSPE_SubTable;

    jctokfifo_t   *JSPE_Queue;
} jsparsexprtype_t;

/*  For JSPE_Flags field  */
#define JPA_VARTYPE_FALLOCDECL     1
#define JPA_VARTYPE_FCOPYDECL      2
#define JPA_VARTYPE_FFREESUBTAB    4
#define JPA_VARTYPE_FISINTNULL     8

#define JPA_VARTYPE_FLONGCONST  1024
#define JPA_VARTYPE_FARRAYSIZE  2048

/* ----------- FILE: jsparsflow.c         ------------ */
typedef struct {
  long JPAFLN_Flags;

  jsparscnode_t *JPAFLN_Node;

  unsigned int JPAFLN_Counter; /*  counter of flow nodes pointing to this node  */
  unsigned int JPAFLN_Next1, JPAFLN_Next2;

} jsparsflownode_t;

/* ***  For JPAFLN_Flags field  *** */
/*  Single statement of a if-else or while/do-while/for loop which
 *   has to be surrounded by curly brackets if to be extended to
 *   more than one statement. This helps us avoid generating too
 *   many curly brackets when patching code.
 */
#define JPA_FLOW_SINGLESTMT      (1 << 0)

/*  Statement is at the start of a block, so we can insert
 *   declarations.
 */
#define JPA_FLOW_BLOCKSTART      (1 << 1)

/*  Not a statement, only an expression (i. e. condition of an if
 *   statement or a loop).
 */
#define JPA_FLOW_EXPRESSION      (1 << 4)

/*  Only a pseudo statement to link other statements, should be
 *   skipped. This is used for statements that only connect the
 *   various switch labels.
 */
#define JPA_FLOW_PSEUDOSTMT      (1 << 5)

/*  The statement has subnodes that are recorded somewhere else as
 *   statements. Only the first sub node must be looked at.
 */
#define JPA_FLOW_FIRSTSUBNODE    (1 << 6)

/*  The statement was reached after a branch in the flow
 */
#define JPA_FLOW_BRANCHED        (1 << 7)

typedef struct {
  unsigned int JPAFL_Start;

  unsigned int JPAFL_NStatements;

  jsparsflownode_t *JPAFL_Statements;

  unsigned long **JPAFL_Bitmaps;
  unsigned int JPAFL_MBitmaps;

} jsparsflow_t;

/*  The index in the JPAFLN_NextX fields to indicate that the surrounding
 *   routine is left
 */
#define JPA_FLOW_INDEX_RETURN    -2

#define jsparsflowclear(jpa)     jsparsflowclear2((jpa)->JSPA_StmtFlow)
#define jsparsflowmark(jpa, indx, opFunction, user, mode) \
	jsparsflowmark2((jpa)->JSPA_StmtFlow, indx, opFunction, user, mode)

/* ----------- FILE: jsparsscan.c         ------------ */
/* ***  Variable definitions  *** */
struct jsparsvardef_t {
    jsparsvardef_t       *JPAVD_Next; /*  To link within hash table  */

    const char           *JPAVD_Name;
    unsigned long         JPAVD_Flags;
    jsparscnode_t        *JPAVD_Declarator;
    jsparscnode_t        *JPAVD_Specifier;

    /*  queue where tokens of definition are stored  */
    jctokfifo_t          *JPAVD_Queue;
    jsparsvartab_t       *JPAVD_SubTable; /* In case of struct/union
					   *  or a typedef with
					   *  anonymous struct
					   */

    unsigned long         JPAVD_LongValue; /*  if this is an enumeration,
					    *   the value.
					    *  If it is an automatically
					    *   sized array (flag
					    */

    /*  To let the user store some other relevant information  */
    union {
	void               *jpavd_UserPtr;
	long               *jpavd_UserLong;
    } u;
#define JPAVD_UserPtr    u.jpavd_UserPtr
#define JPAVD_UserLong   u.jpavd_UserLong
};

/*  For the flags field  */
#define JVDF_TMASK       0x1fUL  /*  Mask for type field  */

/*  Submask to distinguish different name spaces:
 *   Reject multiple definition within one namespace
 */
#define JVDF_TGMASK      0x18UL

/*  Name space 0 for struct/union/enums  */
#define JVDF_TSTRUCT     ((0UL << 3) | 1UL)
#define JVDF_TUNION      ((0UL << 3) | 2UL)
#define JVDF_TENUM       ((0UL << 3) | 3UL)

/*  Name space 1 for variable type identifiers  */
#define JVDF_TVARIABLE   ((1UL << 3) | 1UL)
#define JVDF_TTYPEDEF    ((1UL << 3) | 2UL)
#define JVDF_TENUMIDENT  ((1UL << 3) | 3UL)
#define JVDF_TFUNCTION   ((1UL << 3) | 4UL)

/*  Name space 2 for statement labels  */
#define JVDF_TLABEL      ((2UL << 3) | 1UL)

#define JVDF_FDECLARED   (1UL <<  8) /*  declared  */
#define JVDF_FDEFINED    (1UL <<  9) /*  defined  */
#define JVDF_FFREESUBTAB (1UL << 10) /*  Free subtable when freeing resources  */
#define JVDF_FFREENODES  (1UL << 11) /*  free declarator/specifier  */

#define JVDF_FGLOBAL     (1UL << 12) /*  variable in global variable table  */

/*  Some information flags about this variable  */
#define JVDF_IEXTERN     (1UL << 13) /*  external variable  */
#define JVDF_ISTATIC     (1UL << 14) /*  static variable  */
#define JVDF_IFUNCARG    (1UL << 15) /*  function argument  */
#define JVDF_ICONST      (1UL << 16) /*  constant  */

#define JVDF_FREAD       (1UL << 17) /*  has been read  */
#define JVDF_FWRITE      (1UL << 18) /*  has been written to  */
#define JVDF_FADDR       (1UL << 19) /*  address has been taken of  */

#define JVDF_FARRAYSIZE  (1UL << 20) /*  JPAVD_LongValue contains
				      *   automatically determined
				      *   array size
				      */

#define JVDF_FUSER       (1UL << 24) /*  user flags  */

/* **  Size of variable hash table  ** */
#ifndef JPA_VARIABLE_TABSIZE
# define JPA_VARIABLE_TABSIZE   251  /*  prime number  */
#endif

    struct jsparsvartab_t {
	jsparsvartab_t *JPAVT_Next;

	long            JPAVT_Flags;

	/*  Indexes into a token queue for which range this
	 *   variable table is valid
	 */
	int             JPAVT_StartIndex, JPAVT_EndIndex;

	jsparsvardef_t *JPAVT_Vars[JPA_VARIABLE_TABSIZE];
    };

#define JVDT_FSTRUCTUNION                 (1UL << 0)

#define jsparsgetglobalvartab(jpa)        ((jpa)->JSPA_VarTables[0])
#define jsparsisglobalvartab(jpa, vTab)   ((vTab) == (jpa)->JSPA_VarTables[0])

/* ----------- FILE: jsparsgram.y         ------------ */
/*  Hash table for keywords  */
#define JPA_KEYWORD_TABSIZE  251   /*  prime number  */

/*  For a keyword or an identifier  */
typedef struct _jsparsident_t _jsparsident_t;
struct _jsparsident_t {
  _jsparsident_t *JPAID_Next;
  int             JPAID_Value;

  const char     *JPAID_File;
  int             JPAID_Line;

  char	          JPAID_Name[1];
};

typedef struct _jsparstypedef_t _jsparstypedef_t;
struct _jsparstypedef_t {
  _jsparstypedef_t *JPATD_Next;
  int               JPATD_Flags;
  _jsparsident_t   *JPATD_HashTable[JPA_KEYWORD_TABSIZE];
};

typedef struct {
  int JPABC_Flag;
  const char *JPABC_Name;
  int JPABC_ArrayArg, JPABC_IndexArg;
} _jsparsboundchk_t;

/* **  For JSPA_Mode  ** */
/*  To detect empty files and thus emit a warning (ANSI forbids empty files)  */
#define JPAM_SEENNODE               (1 << 0)

/*  The parser should return original tokens along with the preprocessed ones  */
#define JPAM_FGETORIG               (1 << 1)
/*  The parser should return file position information along with the tokens  */
#define JPAM_FGETFPOS               (1 << 2)
/*  When flushing, print the original ones instead of the preprocessed ones  */
#define JPAM_FPRINTORIG             (1 << 3)
/*  When flushing, print preprocessed ones except when in system include
 *   files or when expanding tokens from system include files: then print
 *   original ones.
 */
#define JPAM_FEXPANDUSERFILES       (1 << 4)

/*  When parsing, generate variable tables  */
#define JPAM_FVARTABLES             (1 << 5)

/*  When parsing, generate a statement flow  */
#define JPAM_FSTMTFLOW              (1 << 6)

/*  Generate a complete flow for each node  */
#define JPAM_FSTMTFULLFLOW          (1 << 7)

/*  When generating a statement flow chart, treat the end of a compound
 *   statement "}" as a separate "statement". All statements that flow to
 *   the end of the block will "tip" on it instead of going directly to
 *   the follower.
 */
#define JPAM_FFLOWBLOCKEND          (1 << 8)

/*  When encountering "if", "while", "for", the statement itself
 *   should become the node instead of the expression with
 *   JPA_FLOW_EXPRESSION set
 */
#define JPAM_FFLOWREALSTMT          (1 << 9)

/*  This flag is set by the statement flow generator if the end of the
 *   block can be reached
 */
#define JPAM_FFALLTOEND             (1 << 10)

/*  Make some things a warning rather than an error  */
#define JPAM_FMIXEDDECLS            (1 << 11)
#define JPAM_FALLOWREDECL           (1 << 12)

#ifdef JSPARSE_INTERNAL
#define JPAM_FCLOSESTREAM           (1 << 16)
#endif

#define JPAM_FUSERFLAGS             (1 << 24)

/* ****  For JSPA_WarnFlags   **** */
/*  Show warnings pertaining the preprocessor  */
#define JPAM_WARNPREPROC            (1 << 0)

/*  Warn about missing 'static' in function definition where static
 *   prototype was seen and function calls without prototype.
 */
#define JPAM_WARNPROTO              (1 << 1)
/*  Show other warnings like bad code style (better use a decent compiler for that) */
#define JPAM_WARNSTYLE              (1 << 2)
/*  Check for unused variables, redefinitions etc.  */
#define JPAM_WARNVARS               (1 << 3)
/*  Do bounds checking  */
#define JPAM_WARNBOUNDS             (1 << 4)
/*  Find bad casts  */
#define JPAM_WARNBADCAST            (1 << 5)
/*  Warn about unreachable code or unused labels  */
#define JPAM_WARNFLOW               (1 << 6)
/*  Warn about functions with high usage of stack memory, i. e. local variables  */
#define JPAM_WARNSTACKUSE           (1 << 7)

typedef struct {
  long              JSPA_Mode;

  long              JSPA_WarnFlags;

  jsqueuestream_t  *JSPA_Stream;

  _jsparsident_t   *JSPA_KeyWordTable[JPA_KEYWORD_TABSIZE];

  /*  The following three are needed by the parser to handle
   *   typedefs correctly.
   */
  _jsparstypedef_t *JSPA_TypeDefs;
  _jsparstypedef_t *JSPA_DeferBlockDef;
  _jsparstypedef_t *JSPA_SavedBlockDef;

  jsparsvartab_t  **JSPA_VarTables;
  unsigned int      JSPA_NVarTables, JSPA_MVarTables;

  _jsparsboundchk_t *JSPA_BoundChkPragma;
  unsigned int JSPA_NBoundChkPragma, JSPA_MBoundChkPragma;

  jsparsflow_t     *JSPA_StmtFlow;

#define JSPA_Statements     JSPA_StmtFlow->JPAFL_Statements
#define JSPA_FlowStart      JSPA_StmtFlow->JPAFL_Start

  jctokfifo_t       JSPA_GlobalQueue;

  /*  Called to print error messages  */
  void (*JSPA_PrintError)(void *user, const char *message);

  void *JSPA_User;
} jsparsstream_t;

#define jsparsgetqueue(jpa)       jsqsgetqueue((jpa)->JSPA_Stream)

/*  First free token for parser  */
#define PTK_PARSERTOKEN   (PTK_USERTOKEN + 10)

#ifdef JSPARSE_INTERNAL
#define BTD_CMPSTMT             1
#endif


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: jsparsexpr.c          ***** */
Prototype int            jsparsgenvardef2(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparsexprtype_t *pType,
					  const char *varname);
Prototype int            jsparsgenvardef(jsparsstream_t *jpa,
					 jstrbuff_t *sbuffer,
					 jsparsvardef_t *vDef,
					 const char *varname);
Prototype int            jsparsgenvardef3(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparscnode_t *decl,
					  jsparscnode_t *spec,
					  jctokfifo_t *queue,
					  const char *varname);
Prototype int            jsparsgenwrapper(jsparsstream_t *jpa,
					  jstrbuff_t *sbuffer,
					  jsparscnode_t *node,
					  const char *newname,
					  const char *retvar,
					  const char *strings[3],
					  jctokfifo_t *queue,
					  jstrbuff_t *auxBuffer,
					  const char *(*getArgName)(void *user),
					  void *uPtr, int mode);
Prototype int            jsparsgenwrapper2(jsparsstream_t *jpa, jstrbuff_t *sbuffer,
					   jsparscnode_t *decl, jsparscnode_t *spec,
					   const char *newname, const char *retvar,
					   const char *strings[2],
					   jctokfifo_t *queue, jstrbuff_t *auxBuffer,
					   const char *(*getArgName)(void *user),
					   void *uPtr, int mode);

/* ***** File: jsparsexpr2.c         ***** */
Prototype jsparsexprtype_t *jsparsgetexprtype(jsparsexprtype_t *pType,
					      jsparsstream_t *jpa,
					      jsparsvartab_t *vTab,
					      jsparscnode_t *expr,
					      jctokfifo_t *queue,
					      int mode);
Prototype int            jsparsfreeexprtype(jsparsstream_t *jpa,
					    jsparsexprtype_t *eType,
					    int flag);
Prototype unsigned long  jsparsgetvarsize(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					  const char *varName);
Prototype unsigned long  jsparsgetvarsize2(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					   jsparsvardef_t *vDef);
Prototype unsigned long  jsparsgetvarsize3(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					   jctokfifo_t *queue, jsparscnode_t *decl,
					   jsparscnode_t *spec);

/* ***** File: jsparsexpr3.c         ***** */
Prototype jsparsexprtype_t *jsparsmakeetype(jsparsexprtype_t *pType,
					    jsparscnode_t *decl,
					    jsparscnode_t *spec,
					    jctokfifo_t *queue);
Prototype jsparsexprtype_t *jsparsnode2type(jsparsexprtype_t *pType,
					    jsparscnode_t *node,
					    jctokfifo_t *queue);
Prototype jsparsexprtype_t *jsparsvdef2type(jsparsstream_t *jpa,
					    jsparsvartab_t *vTab,
					    jsparsexprtype_t *pType,
					    const jsparsvardef_t *vDef,
					    int mode);
Prototype long           jsparsgetvartype3(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   jsparsvardef_t *vDef,
					   int *pPtrCnt, int mode);
Prototype long           jsparsgetvartype2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   jsparsexprtype_t *pType,
					   int *pPtrCnt, int mode);
Prototype long           jsparsgetvartype(jsparsstream_t *jpa,
					  jsparsvartab_t *vTab,
					  jsparscnode_t *decl,
					  jsparscnode_t *spec,
					  jctokfifo_t *queue, int *pPtrCnt,
					  int mode);
Prototype int            jsparsisvartype(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 jsparsexprtype_t *type,
					 jsparsexprtype_t *refType,
					 int mode);

/* ***** File: jsparsflow.c          ***** */
Prototype void            jsparsflowclear2(jsparsflow_t *jFlow);
Prototype int             jsparsflowmark2(jsparsflow_t *jFlow, int indx,
					  int (*opFunction)(jsparscnode_t *,
							    int, void *),
					  void *user, int mode);
Prototype int             jsparsfindflow(jsparsstream_t *jpa,
					 jsparscnode_t *node);

/* ***** File: jsparsgram.y          ***** */
Prototype jsparscnode_t *jsparsgetexpression(jsparsstream_t *jpa,
					     jsqueuestream_t *jsqs,
					     int mode);
Prototype jsparscnode_t *jsparsgettype(jsparsstream_t *jpa,
				       jsqueuestream_t *jsqs,
				       int mode);
Prototype const char    *jsparstok2string(jsctoken_t token);

/* ***** File: jsparsgram2.c         ***** */
Prototype int            jsparsinit(int mode);
Prototype jsparsstream_t *jsparsopen2(jsparsstream_t *jpa, jsppstream_t *jps,
				      jsparsenviron_t *jpaenv, const char *filename,
				      int omode, int mode);
Prototype int            jsparsclose(jsparsstream_t *jpa, int flag);
Prototype void           jsparssetprinterr(jsparsstream_t *jpa,
					   void (*printError)(void *user,
							      const char *message));
Prototype int            jsparsaddkeyword(jsparsstream_t *jpa,
					  const char *name,
					  const char *oldName);
Prototype int            jsparsget(jsparsstream_t *jpa, jsparscnode_t **pNode,
				   int mode);
Prototype int            jsparsflush(jsparsstream_t *jpa, jsparscnode_t *node,
				     FILE *fp, int mode);
Prototype const dbgopts_t JSParseDebugOptions[];

/* ***** File: jsparsscan.c          ***** */
Prototype int             jsparsinsertvtab(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab);
Prototype jsparsvardef_t *jsparsfindvariable(jsparsstream_t *jpa,
					     jctokfifo_t *queue,
					     jsparscnode_t *node,
					     int type, int mode);
Prototype jsparsvartab_t *jsparsfindvartab(jsparsstream_t *jpa,
					   jsparscnode_t *node);
Prototype jsparsvartab_t *jsparsfindvartab2(jsparsstream_t *jpa, int start,
					    int end);
Prototype jsparsvardef_t *jsparsdefinevar2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab, int type,
					   jctokfifo_t *queue,
					   const char *name,
					   jsparscnode_t *declarator,
					   jsparscnode_t *specifier,
					   int mode);
Prototype jsparsvardef_t *jsparsgetvardef(jsparsstream_t *jpa,
					  jsparsvartab_t *vTab,
					  const char *name, int type,
					  int mode);
Prototype jsparsvardef_t *jsparsgetvardef2(jsparsstream_t *jpa,
					   jsparsvartab_t *vTab,
					   const char *name, int type,
					   jsparsvartab_t **pTab,
					   int mode);
Prototype int            jsparsdelvardef(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 const char *name, int type,
					 int mode);
Prototype void           jsparsfreevartab(jsparsvartab_t *vTab);

/* ***** File: jsparsscan2.c         ***** */
Prototype void           jsparsclearvarflags(jsparsvartab_t *vTab, long flags);
Prototype int            jsparsprocessvars(jsparsstream_t *jpa,
					   jsparscnode_t *node,
					   jctokfifo_t *queue,
					   int (*doVariable)(jsparsstream_t *,
							     jsparscnode_t *,
							     jctokfifo_t *,
							     jsparsvartab_t *,
							     jsparsvardef_t *,
							     long, void *),
					   void *user, int mode);
Prototype int            jsparsgetvarname(jsparscnode_t *node,
					  jsparscnode_t **pVarNode,
					  int (*handleOps)(void *u, jsparscnode_t *op),
					  void *user);

/* ***** File: jsparssubs.c          ***** */
Prototype int            jsparsflushtokens2(jsparsstream_t *jpa,
                                            jsqueuestream_t *jsqs,
                                            FILE *fp);
Prototype void           jsparsseterrcode2(int errnr, int erroff,
                                           const char *string,
                                           jctokfifo_t *queue,
                                           jsparscnode_t *node);
Prototype void           jsparsprinterr(jsparsstream_t *jpa,
                                        jctokfifo_t *queue,
                                        jsparscnode_t *node,
                                        const char *message, int mode,
                                        ...);
Prototype void           jsparsprinterr2(jsparsstream_t *jpa,
                                         jctokfifo_t *queue,
                                         unsigned int indx,
                                         const char *message,
                                         int mode, ...);
Prototype jsparscnode_t *jsparsmkcnode(jsparscnode_t *next,
                                       jsparscnodetype_e type,
                                       jsparscnode_t *sub,
                                       unsigned int start,
                                       unsigned int end,
                                       jctokfifo_t *queue);
Prototype jsparscnode_t *jsparsfreecnode(jsparscnode_t *node);
Prototype jsparscnode_t *jsparscopycnode(jsparscnode_t *node,
                                         jctokfifo_t *queue,
                                         jctokfifo_t *destQ);
Prototype int            jsparsinserttokstr(jsparsstream_t *jpa,
                                            jsparscnode_t *tree,
                                            jctokfifo_t *queue,
                                            const char *strings[], int where,
                                            int mode);
Prototype int            jsparsinserttoks(jsparsstream_t *jpa,
                                          jsparscnode_t *tree,
                                          jctokfifo_t *queue,
                                          jsctoken_t *tokens, int n,
                                          int where, int mode);
Prototype int            jsparsmatchorigtoks(jctokfifo_t *queue, int *pStart,
                                             int *pEnd);
Prototype void           jsparsprctree(FILE *fp, jctokfifo_t *queue,
                                       jsparscnode_t *node,
                                       int ident, int mode);
Prototype void           jsparsprnode(FILE *fp, jctokfifo_t *queue,
                                      jsparscnode_t *node, int mode);
Prototype void           jsparsenvclear(jsparsenviron_t *jpaenv);
Prototype int            jsparsenvhandle(jsparsenviron_t *jpaenv, int opt, const char *string);
Prototype void           jsparsenvfree(jsparsenviron_t *jpaenv);
Prototype const char     JSParsStdOptsHelp[];

/* ***** File: jsparswrapinc.c       ***** */
Prototype const unsigned char JSParseWrapIncludes[];


# undef Prototype   /* I like to be clean  */

# ifdef JSPARSE_INTERNAL
/*  For C++ compatibility   */
#  ifdef __cplusplus
#   define IPrototype extern "C"
#  else
#   define IPrototype extern
#  endif
/* ***** File: jsparsexpr3.c         ***** */
IPrototype int           _jsparsres1type(jsparsstream_t *jpa,
					 jsparsvartab_t *vTab,
					 jsparsexprtype_t *pType);
IPrototype int           _jsparsrestype(jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparscnode_t *node,
					jctokfifo_t *queue,
					jsparsexprtype_t *pType);
IPrototype void          _jsparssetuptype(jsparsexprtype_t *pType,
					  const jsparsvardef_t *vDef);

/* ***** File: jsparsflow.c          ***** */
IPrototype jsparsflow_t  *_jsparsscanflow(jsparsstream_t *jpa,
					  jsparscnode_t *node,
					  jctokfifo_t *queue);
IPrototype void           _jsparsfreeflow(jsparsflow_t *jFlow);

/* ***** File: jsparsgram.y          ***** */
IPrototype int           _jsparsgetnode(jsparsstream_t *jpa,
					jsqueuestream_t *jsqs,
					jsparscnode_t **pNode,
					int starttoken);
IPrototype const jsctokops_t JSParserCOperators[];
IPrototype const jsctokops_t JSParserCKeyWords[];

/* ***** File: jsparsgram2.c         ***** */
IPrototype _jsparsident_t *_jsparsgetident(const char *name, _jsparsident_t *
					   hTab[JPA_KEYWORD_TABSIZE]);
IPrototype int           _jsparspushbtype(jsparsstream_t *jpa,
					  _jsparstypedef_t *def, int flags);
IPrototype int           _jsparspopbtype(jsparsstream_t *jpa);
IPrototype void          _jsparsfreebtype(_jsparstypedef_t *def);
IPrototype int           _jsparschkstrunion(jsparsstream_t *jpa,
					    jctokfifo_t *queue,
					    jsparscnode_t *node);
IPrototype jsparscnode_t *_jsparsmkunary(jsparscnode_t *node,
					 jsparscnode_t *sub,
					 jctokfifo_t *queue);
IPrototype _jsparsident_t *_jsparsputident(const char *name, int value,
					   _jsparsident_t *hTab[JPA_KEYWORD_TABSIZE]);
IPrototype jsparscnode_t  *_jsparsmkpragmaline(jctokfifo_t *queue,
					       int start, int end);

/* ***** File: jsparsscan.c          ***** */
IPrototype jsparsvartab_t *_jsparsgetvartab(jsparsvartab_t *next, long flags,
					    int start, int end);

/* ***** File: jsparsscan2.c         ***** */
IPrototype int           _jsparsfindvariables(jsparsstream_t *jpa,
					      jsparscnode_t *node);
IPrototype int           _jsparsscanvars(jsparsstream_t *jpa,
					 jsparscnode_t *node,
					 jctokfifo_t *queue,
					 jsparsvartab_t *vTab, int mode);
IPrototype int           _jsparsclearvariables(jsparsstream_t *jpa);

/* ***** File: jsparssubs.c          ***** */
IPrototype int           _jsparsdopredefine(jsparsstream_t *jpa);
IPrototype int           _jsparschknode2(jsparsstream_t *jpa,
                                         jsparscnode_t *node);
IPrototype int           _jsparsdopragma(jsppstream_t *jps,
                                         jsctoken_t *pToken, int n,
                                         void *user);
IPrototype jsparscnode_t *jsparscopyncnode(jsparscnode_t *node,
                                           jctokfifo_t *queue,
                                           jctokfifo_t *destQ);
IPrototype jsparscnode_t *_jsparslinktail(jsparscnode_t *node,
                                          jsparscnode_t *next);
IPrototype jsparscnode_t *_jsparsgettail(jsparscnode_t *node);

/* ***** File: jsparssymbs.c         ***** */
IPrototype int           _jsparspredefine(jsparsstream_t *jpa, int level);


#  undef IPrototype   /* I like to be clean  */
# endif

#endif
