/*JS*********************************************************************
*
*    Program : JSSUBS.H
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Include file for library 'libjssubs.a' and all programs
*	       that use this library.
*
*              Copyright (C) 1994-2015 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/
#ifndef JSSUBS_H /*   File must not be included more than once!   */
#define JSSUBS_H 1

/*********     GENERAL DEFINES					*********/
#ifndef FALSE
# define FALSE	0
#endif
#ifndef TRUE
# define TRUE	1
#endif

/* ****    Some useful macros (be aware of side effects because   **** */
/* ****     they treat arguments more than once).		  **** */
#ifndef MAX
# define MAX(a,b)  ((a) > (b) ? (a) :  (b))
#endif
#ifndef MIN
# define MIN(a,b)  ((b) > (a) ? (a) :  (b))
#endif
#ifndef ABS
# define ABS(x)    ((x) >= 0  ? (x) : -(x))
#endif

/* ********************************************************************** */
#ifdef __GNUC__
# ifndef likely
#  define likely(x)     (__builtin_expect(!!(x), 1))
# endif
# ifndef unlikely
#  define unlikely(x)   (__builtin_expect(!!(x), 0))
# endif
#else
# ifndef likely
#  define likely(x)     (x)
# endif
# ifndef unlikely
#  define unlikely(x)   (x)
# endif
#endif

/* ********************************************************************** */
/*  This is used to reallocate dynamic arrays in a smart way:
 *   We increase the size by 25%
 */
#define REALLOC_RESIZE(maxvar, incr)  ((maxvar) += MAX((maxvar) / 4, incr))

/* ********************************************************************** */
#ifndef EGOTO
# ifdef DEBUG_EGOTO_VERBOSE0
#  define EGOTO(label)							\
    do {								\
	fprintf(stderr, "ERROR l. %d '%s'\n", __LINE__, __FILE__);	\
	goto label;							\
    } while(0)
# elif defined(DEBUG_EGOTO_VERBOSE)
#  define EGOTO(label)							\
    do {								\
	sprintf(JSErrorStack[JSErrorStackLevel], "l. %d '%s'", __LINE__, __FILE__); \
	if(++JSErrorStackLevel >= sizeof(JSErrorStack) / sizeof(JSErrorStack[0])) \
	    JSErrorStackLevel = 0;					\
	goto label;							\
    } while(0)
# elif defined(DEBUG_EGOTO_ABORT)
#  define EGOTO(label)    abort()
# else
#  define EGOTO(label)    goto label
# endif
#endif

/*  This will initialize the error call stack (not yet)  */
#define EGOTO0(label)     EGOTO(label)

/* ********************************************************************** */
/*  Mark error if's. For GCC, we help branch prediction
 *   We could also redefine this to force errors
 */
#ifndef EIF
# define EIF(expr)         if(unlikely(expr))
#endif

/* ********************************************************************** */
/* ****   Compile time helper   ***** */
/*  Test a condition at compile time instead of runtime. Idea taken from
 *   the Linux kernel
 */
#define CASSERT(condition)        ((void)sizeof(int [1 - 2 * !(condition)]))

/* ********************************************************************** */
#define USE(param)        (void)param

/* ********************************************************************** */
/* ****   I'd like to type PI if I want PI !   ***** */
#ifndef PI
# define PI         3.1415926535897932384626433832795029
#endif

#ifndef LN2
# define LN2        0.6931471805599453094172321214581766
#endif

/* ********************************************************************** */
/* ****   Interval intersection   ***** */
/*  Check if two intervals intersect. It is easy if you have the
 *   right perspective on it: They *don't* intersect if the first
 *   intervals end is left of the second intervals start or vice
 *   versa. Reversing the condition gives the macros below.
 */

/*  Intersection including boundaries  */
#define INTERVAL_INTERSECT(astart, aend, bstart, bend) \
		((aend) >= (bstart) && (bend) >= (astart))

/*  Intersection excluding boundaries  */
#define OPEN_INTERVAL_INTERSECT(astart, aend, bstart, bend) \
		((aend) > (bstart) && (bend) > (astart))

/* ********************************************************************** */
/*  Calculate required padding for alignment  */
#define JS_ALIGN_OFFSET(offset, align)		\
    ((offset) % (align) == 0 ? (offset) :	\
     (offset) + (align) - ((offset) % (align)))

/* *******  General definitions  *********** */
/*  Some systems define FILENAME_MAX very small (14 for example)  */
#include <stdio.h>

#ifndef JS_FILENAME_MAX
# if !defined(FILENAME_MAX) || (FILENAME_MAX < 256 && !defined(__MSDOS__))
#  define JS_FILENAME_MAX  256
# else
#  define JS_FILENAME_MAX  FILENAME_MAX
# endif
#endif

/* ******  Common environment variable for options and path ***** */
#define JSENVOPT     "JSOPTIONS"
#define JSENVPATH    "JSPATH"

/* ***********	 Extracted definitions of library code  ********* */
/* ----------- FILE: crccalc.c            ------------ */
#include <jsendian.h>

/*  Should be CHAR_BIT, but we do not wan't to export this
 *   into the definition of CRC_TABSIZE.
 */
#define CRC_CHAR_BIT    8

/* **  Needed length of a CRC table  ** */
#define CRC_TABSIZE     (1 << CRC_CHAR_BIT)

/*  These macros can be used to write a calculated checksum at the
 *   end of the area to get a total checksum of zero
 */
/*  For 16 bits  */
#define CHKSUM_TO_CHARS1(value, array)  JS_INT162CHAR_BE(value, array)
#define CHKSUM_LEN1                     JS_INT162CHAR_LEN

/*  For 32 bits  */
#define CHKSUM_TO_CHARS2(value, array)  JS_INT322CHAR_BE(value, array)
#define CHKSUM_LEN2                     JS_INT322CHAR_LEN

/*  For 64 bits  */
#define CHKSUM_TO_CHARS3(value, array)  JS_INT642CHAR_BE(value, array)
#define CHKSUM_LEN3                     JS_INT642CHAR_LEN

/*  We can get the polynomial back from the generated table  */
#define CRC_GETPOLY(tab)   (tab[1])

/* ----------- FILE: cstrfprint.c         ------------ */
#include <stdio.h>   /*	 For "FILE" definition    */

/* ----------- FILE: debugopts.c          ------------ */
typedef struct {
  const char    *DBOP_Name;
  unsigned long  DBOP_Flag;
  void          *DBOP_Pointer;
  const char    *DBOP_Help;
} dbgopts_t;

/* ----------- FILE: extfile.c            ------------ */

/*  Basic definition for extended file pointer  */
typedef struct {
  int FM_UngetChar;
  int (*FM_Module)(void *uPtr, int *pC);
  void *FM_User;
} efilemod_t;

/*  Return values for modules  */
#define EMOD_IDLE  0  /*  nothing done, call module again  */
#define EMOD_IN    1  /*  input char used  */
#define EMOD_OUT   2  /*  output char produced  */
#define EMOD_THROUGH  (EMOD_IN | EMOD_OUT)

typedef struct {
  long          EF_Flags;

  /*   > 0: current linen number in file.
   *  <= 0: current negated index in string
   */
  int           EF_LineNr;
  union {
    FILE       *ef_FPtr;
    void       *ef_Ptr;
    const char *ef_String;
  } ef_Cont;
  int          *EF_UngetBuff;
  unsigned int  EF_UngetIndex, EF_UngetMax;

  efilemod_t  *EF_ModStack; /*  Stack of file modules  */
  int           EF_ModMax;
  union {
    int       (*ef_PPut)(int c, void *p);
    int       (*ef_PGet)(void *p, int flag);
  } ef_Cont2;
  void         *EF_User;
} extfile_t;

#define EF_FPtr    ef_Cont.ef_FPtr
#define EF_Ptr     ef_Cont.ef_Ptr
#define EF_String  ef_Cont.ef_String

#define EF_PPut    ef_Cont2.ef_PPut
#define EF_PGet    ef_Cont2.ef_PGet

/*  If reading from a string, get the current position  */
#define efstrpos(efp)    ((efp)->EF_String - (efp)->EF_LineNr - \
			  (efp)->EF_UngetIndex)

/*  For EF_Flags field  */
#define EFFLAG_EOF       (1<<0)
#define EFFLAG_ERROR     (1<<1)
#define EFFLAG_PIPE      (1<<2)
#define EFFLAG_FREESTR   (1<<3)
#define EFFLAG_BUFFLINE  (1<<4)
#define EFFLAG_WRITE     (1<<5)
#define EFFLAG_INPUTEOF  (1<<6)

#define EFFLAG_USER      (1<<16)

#define efeof(efp)       ((efp)->EF_PGet ? \
			  ((efp)->EF_Flags & EFFLAG_INPUTEOF) : \
			  (efp)->EF_LineNr > 0 ? \
                          (efp->EF_PGet == NULL && feof((efp)->EF_FPtr)) : \
                          ((efp)->EF_String[-(efp)->EF_LineNr]) == '\0')
#define efcleareof(efp)  ((efp)->EF_Flags &= ~(EFFLAG_EOF|EFFLAG_INPUTEOF))

#define eferror(efp)     ((efp)->EF_Flags & EFFLAG_ERROR)

#define eputc(c, efp)     (efwrite(&c, 1, efp) == 1 ? (c) : EOF)
#define efputs(s, efp)    efwrite(s, strlen(s), efp)

/*  Close an extended file descriptor without closing the underlying
 *   stdio file descriptor
 */
#define efclose2(efp)     ((efp)->EF_FPtr = NULL, efclose(efp))

/* ----------- FILE: fexists.c            ------------ */
#define FILEEXISTS_TFILE           1
#define FILEEXISTS_TDIRECTORY      2
#define FILEEXISTS_TSYMLINK        3
#define FILEEXISTS_TFIFO           4

#define fileexists(name)          fileexists2(name, 0)

/* ----------- FILE: filelen.c            ------------ */
#include <jsllio.h>  /*	 For "off_t" definition    */

/* ----------- FILE: flocks.c             ------------ */
#include <jsconfig.h>  /*  For configurational definitions  */

/*  Definitions for routine lockRegion and fopenLock  */
/*  Don't change these values!  */
#define JSLOCK_READ     0
#define JSLOCK_READW    1
#define JSLOCK_WRITE    2
#define JSLOCK_WRITEW   3
#define JSLOCK_UNLOCK   4

#ifndef CONFIG_NO_POSIX
/*  Macros for usage of routine lockRegion2  */
#define lockRegion(fp, type, offset, whence, len)  \
	lockRegion2(fileno(fp), type, offset, whence, len)
#else
extern int lockRegion(FILE *fp, int type, long offset, int whence, long len);
#endif

#define lockRead(fp, o, w, l)   lockRegion(fp, JSLOCK_READ, o, w, l)
#define lockReadW(fp, o, w, l)  lockRegion(fp, JSLOCK_READW, o, w, l)

#define lockWrite(fp, o, w, l)  lockRegion(fp, JSLOCK_WRITE, o, w, l)
#define lockWriteW(fp, o, w, l) lockRegion(fp, JSLOCK_WRITEW, o, w, l)

#define lockRead2(fd, o, w, l)   lockRegion2(fd, JSLOCK_READ, o, w, l)
#define lockReadW2(fd, o, w, l)  lockRegion2(fd, JSLOCK_READW, o, w, l)

#define lockWrite2(fd, o, w, l)  lockRegion2(fd, JSLOCK_WRITE, o, w, l)
#define lockWriteW2(fd, o, w, l) lockRegion2(fd, JSLOCK_WRITEW, o, w, l)

#define unLock(fp, o, w, l)      lockRegion(fp, JSLOCK_UNLOCK, o, w, l)
#define unLock2(fd, o, w, l)     lockRegion2(fd, JSLOCK_UNLOCK, o, w, l)

/* ----------- FILE: fopenpath.c          ------------ */
#define fopenPath(name, mode)          fopenPath2(name, mode, FOpenPath)
#define fopenPath2(name, mode, path)   fopenPath3(name, mode, path, 0)

/* ----------- FILE: fstralloc.c          ------------ */
typedef struct fstrallocchunk_t fstrallocchunk_t;

struct fstrallocchunk_t {
  fstrallocchunk_t *FSAC_Next;

  unsigned int FSAC_Index, FSAC_Max;
  char FSAC_Mem[1];
};

typedef struct {
  fstrallocchunk_t *FSA_Chain;
  unsigned int FSA_AllocSize;

} fstralloc_t;

#define fstrinit(fstr)				\
  ((fstr)->FSA_Chain = NULL,			\
   (fstr)->FSA_AllocSize = FStringAllocSize)

/* ******************************************************* */
/*  Memorize current watermark of a pool to be restored later  */
typedef struct {
  const fstrallocchunk_t *FSAM_Chunk;
  unsigned int FSAM_Index;

} fstrallocmark_t;

#define fstrmarkinvalidate(pmark)		\
  ((pmark)->FSAM_Chunk = NULL,			\
   (pmark)->FSAM_Index = -1)

#define fstrmarkisvalid(pmark)			\
  ((pmark)->FSAM_Index != (unsigned int)-1)


/* ----------- FILE: getctoken.c          ------------ */

#define GCTK_NOLINEESC     (1 <<  0) /*  Drop "\\\n" in identifiers and numbers */
#define GCTK_NOLINEESCTOT  (1 <<  1) /*  Drop "\\\n"  completely */
#define GCTK_UNIQUE        (1 <<  2) /*  Generate unique token string memory  */
#define GCTK_PPINCLUDELINE (1 <<  3) /*  Recognize <..> like strings  */
#define GCTK_CPPCOMMENT    (1 <<  4) /*  Recognize C++ comments  */
#define GCTK_READTOTERM    (1 <<  5) /*  Read till GCTKTerminator is found  */
/*  Escape character does not work in single quoted strings  */
#define GCTK_SHELLSTRING   (1 <<  6)

#define GCTK_SKIPSPACETOK  (1 <<  7) /*  Don't return blanks, just skip them  */
#define GCTK_SKIPSPACETOK2 (1 <<  8) /*  Don't return blanks, except newline */
#define GCTK_INCSTARTSTOP  (1 <<  9) /*  Return complete tokens  */
#define GCTK_CBRASTRING    (1 << 10) /*  Recognize string with curly brackets  */
#define GCTK_SQBRASTRING   (1 << 11) /*  Recognize string with square brackets  */

/*  Switch on extensions: Allow '$' in identifiers and newline in string */
#define GCTK_EXTENSIONS    (1 << 12)
/*  Extension2: Return " in case of \" sequence  */
#define GCTK_EXTENSIONS2   (1 << 13)

#define GCTK_SQLSTRING     (1 << 14) /*  '' and "" are escapes for quotes  */
#define GCTK_SQLCOMMENT    (1 << 15) /*  -- introduces comment  */

#define GCTK_FLAG_USER     (1 << 16) /*  Start of user definable flags  */

/*  Basic definition for C token  */
typedef struct {
#define CTOKEN_FALLOCATED         (1 << 0)
#define CTOKEN_FORIGTOKEN         (1 << 8)
    int TK_Flags;
    int TK_Value;

    union {
        const char  *tk_String;
        void        *tk_Ptr; /*  to store an arbitrary pointer  */
    } tk_cont;
#define TK_String  tk_cont.tk_String
#define TK_Ptr     tk_cont.tk_Ptr

} jsctoken_t;

#define CTOKEN_GENSTATIC(value, string)    { 0, value, { string } }

#define copyCToken(token, mode)     copyCToken2(NULL, token, mode)

/*  Definition for operator table  */
typedef struct {
  char *KD_String;
  int   KD_Token;
} jsctokops_t;

/*  Basic return tokens of routine getCToken  */
#define CTK_EOF            (-1) /*  My own EOF value  */
#define CTK_MINIMUM        257  /*  Lowest special token value  */

#define CTK_ERROR          (CTK_MINIMUM + 0) /* Error in getCToken  */

#define CTK_BLANK_START   (CTK_MINIMUM + 1)  /* Blank tokens  */
#define CTK_BLANKSEOL      (CTK_MINIMUM + 1)
#define CTK_BLANK_START2  (CTK_MINIMUM + 2)  /* Blank tokens without newline */
#define CTK_BLANKS         (CTK_MINIMUM + 2)
#define CTK_COMMENT        (CTK_MINIMUM + 3)
#define CTK_CPPCOMMENT     (CTK_MINIMUM + 4)
#define CTK_SQLCOMMENT     (CTK_MINIMUM + 5)
#define CTK_BLANK_END     (CTK_MINIMUM + 5)

#define CTK_STRING         (CTK_MINIMUM + 10)  /*  String tokens  */
#define CTK_WSTRING        (CTK_MINIMUM + 11)
#define CTK_USTRING        (CTK_MINIMUM + 12)
#define CTK_PPSTRING       (CTK_MINIMUM + 13)
#define CTK_CHARCONST      (CTK_MINIMUM + 14)
#define CTK_WCHARCONST     (CTK_MINIMUM + 15)
#define CTK_UCHARCONST     (CTK_MINIMUM + 16)

#define CTK_IDENT          (CTK_MINIMUM + 20) /*  Identifier  */

#define CTK_NUMBER         (CTK_MINIMUM + 25) /*  Number tokens:  */
#define CTK_INUMBER        (CTK_MINIMUM + 26)
#define CTK_DNUMBER        (CTK_MINIMUM + 27)

#define CTK_VANILLA        (CTK_MINIMUM + 30) /*  Solely string  */
#define CTK_CBRASTRING     (CTK_MINIMUM + 31) /*  String in curly brackets  */
#define CTK_SQBRASTRING    (CTK_MINIMUM + 32) /*  String in square brackets  */

#define CTK_USERTOKEN      (CTK_MINIMUM + 43) /*  First user defined token == 300  */

/*  Estimate needed space to print a token  */
#define CTK_ESTLENGTH(token)   ((token).TK_String ? \
                                (strlen((token).TK_String) + 5) : 1)

/* ----------- FILE: ifstack.c            ------------ */
typedef struct {
  int *IS_Mem, *IS_Stack, IS_Max;
} ifstack_t;

/*  For "op" argument below  */
#define IFSTK_TYPE_ASK       0 /*  Only return current condition  */
#define IFSTK_TYPE_CANBETRUE 1 /*  Only return if 'elif' or 'else' could switch to true  */

#define IFSTK_TYPE_IF        5 /*  Handle 'if <cond>'  */
#define IFSTK_TYPE_ELIF      6 /*  Handle 'elif <cond>'  */
#define IFSTK_TYPE_ELSE      7 /*  Handle 'else'   */
#define IFSTK_TYPE_ENDIF     8 /*  Handle 'endif'  */

#define IDSTK_ELSESEEN   (1 << 2) /*  'else' part already encountered  */

#define ifstacklevel(istk)     ((istk)->IS_Mem == NULL ? 0 : \
				1 + ((istk)->IS_Stack - (istk)->IS_Mem))
#define ifstackhadelse(istk)   ((istk)->IS_Stack && \
				(*((istk)->IS_Stack) & IDSTK_ELSESEEN))

/* ----------- FILE: jsgetline.c          ------------ */

/*  For parameter 'mode' of jsgetline  */
#define GETLINE_RMEOL     (1<<0)  /*  Remove '\n'  */
#define GETLINE_CONCAT    (1<<1)  /*  Allow line concatenation  */
#define GETLINE_RMCONCAT  (1<<2)  /*  Remove "\\\n" sequences  */
#define GETLINE_COMMENT   (1<<3)  /*  Allow shell like comments (" .. # ..") */
#define GETLINE_UNIQUE    (1<<4)  /*  Make returned line unique  */
#define GETLINE_RMBSLASH  (1<<5)  /*  Leave '\n', but remove '\\'  */

/* ----------- FILE: microtime.c          ------------ */
#include <time.h>

/* ----------- FILE: mkbackup.c           ------------ */
#define mkBackupFile(fileName, mode)   mkBackupFile2(fileName, NULL, mode)

/* ----------- FILE: saxparser.c          ------------ */
/*  Trim heading and trailing spaces from character sequences  */
#define SAXTYPE_FTRIMSPACES        1
/*  Do not return character sequences consisting solely of spaces  */
#define SAXTYPE_FSKIPSPACES        2
/*  Keep escape sequences like &lt; in attribute values and characters  */
#define SAXTYPE_FKEEPESCCHAR       4
/*  Convert <element/> into <element></element>  */
#define SAXTYPE_FNOSINGLEEL        8
/*  Use pLineNr[1] to update current column  */
#define SAXTYPE_FUPDCOLUMN        16

typedef enum {
  SAXTYPE_EOF,

  SAXTYPE_XMLVERSION,    /*  <?xml version="1.0" ... ?>  */
  SAXTYPE_PROCINSTR,     /*  <?process ... >  */
  SAXTYPE_DOCTYPE,       /*  <!DOCTYPE ... >  */
  SAXTYPE_CHARACTER,
  SAXTYPE_COMMENT,       /*  <!-- ... -->  */

  SAXTYPE_STARTELEMENT,  /*  <element ...>  */
  SAXTYPE_STOPELEMENT,   /*  </element>  */
  SAXTYPE_ELEMENT        /*  <element .../>  */
} saxtype_e;

typedef struct {
  char *SAPA_Buffer;
  const char **SAPA_Attributes;
  unsigned int SAPA_NAttributes;

  unsigned int SAPA_Mode;
  unsigned int SAPA_State;
  unsigned int SAPA_MAttributes, SAPA_MBuffer;

} saxparse_t;

#define saxparsegetname(sapa)    ((sapa)->SAPA_Buffer)
#define saxparsegetattrs(sapa)   ((sapa)->SAPA_Attributes)
#define saxparsegetnattrs(sapa)  ((sapa)->SAPA_NAttributes)

/* ----------- FILE: sharestrdup.c        ------------ */
#define JSHST_TABSIZE  251   /*  prime number  */

typedef struct _jsshareident_t _jsshareident_t;
struct _jsshareident_t {
  _jsshareident_t *JSHID_Next;
  char             JSHID_String[1];
};

typedef struct {
  _jsshareident_t *JSHST_HashTable[JSHST_TABSIZE];
} jssharestr_t;

/* ----------- FILE: strbuffer.c          ------------ */
#include <stdarg.h>  /*  For "va_list" definition */

typedef struct {
  char *SB_Buffer;
  int SB_MBuffer, SB_NBuffer;
} jstrbuff_t;

#define jstrbuffer(sb)         ((sb)->SB_Buffer)
#define jstrbuffmaxlen(sb)     ((sb)->SB_MBuffer)
#define jstrbufflen(sb)        ((sb)->SB_NBuffer)
#define jstrbuffputc(sb, c)    ((sb)->SB_Buffer[((sb)->SB_NBuffer)++] = (c))
#define jstrbuffclose(sb)      ((sb)->SB_Buffer[(sb)->SB_NBuffer] = '\0')

#define jstrbuffinc(sb, len)    ((sb)->SB_NBuffer += (len))
#define jstrbuffset(sb, len)    ((sb)->SB_NBuffer = (len))

#define jstrbuffclear(sb)      ((sb)->SB_Buffer = NULL, \
			       (sb)->SB_MBuffer = (sb)->SB_NBuffer = 0)
#define jstrbuffempty(sb)      (jstrbuffset(sb, 0), (sb)->SB_Buffer ? \
				(sb)->SB_Buffer[0] = '\0' : 0)
#define jstrbufffree(sb)       free(((sb)->SB_Buffer))

/* ----------- FILE: wildcmp.c            ------------ */
#define WILD_STRICT	  (1<<0)
#define WILD_STARTNAME	  (1<<1)
#define WILD_IGNORECASE   (1<<2)
#define WILD_PARTIALLY	  (1<<3)

/* ----------- FILE: ctokfifo.c           ------------ */
/* **  To defer a sequence of tokens for later output  ** */
typedef struct {
    const char        *TKFPM_String;
    unsigned int       TKFPM_Offset;
} jctokfifoptrmap_t;

typedef struct {
    unsigned int       TKFF_Mode, TKFF_Step;

    unsigned int       TKFF_Start, TKFF_Index, TKFF_Max;
    jsctoken_t        *TKFF_Token;

    char              *TKFF_Mem;
    unsigned int       TKFF_MIndex, TKFF_MMax;

    jctokfifoptrmap_t *TKFF_PtrMap;
    unsigned int       TKFF_NPtrMap, TKFF_MPtrMap;

} jctokfifo_t;

#define JCTOK_TYPE_FORIGTOK       1
#define JCTOK_TYPE_FFILEPOS       2

/*  TK_String field is not monotonically increasing  */
#define JCTOK_TYPE_FNOSTRORDER    256

#define JCTOK_MAXSTEP             3

#define jctokgetstep(queue)      ((queue)->TKFF_Step)
#define jctokindorig(queue)      (((queue)->TKFF_Mode &			\
                                   JCTOK_TYPE_FORIGTOK) ? 1 : -1U)
#define jctokindfpos(queue)      (((queue)->TKFF_Mode &		\
                                   JCTOK_TYPE_FFILEPOS) ?	\
                                  (queue)->TKFF_Step - 1 : -1U)

/*  Empty the queue  */
#define jctokclear(queue)   ((queue)->TKFF_Start =	\
			     (queue)->TKFF_Index = 0,	\
			     jctokclearptrmap(queue))

#define jctokclearptrmap(queue)   ((queue)->TKFF_NPtrMap = 0)

#define jctoktostring(queue, start, end)		\
    jctoktostring2(queue, start, end, NULL, NULL)

/*  Modes for routine jctokwriten  */
#define JCTOK_MODE_INSERT    (1 << 0)
#define JCTOK_MODE_CHKPTR    (1 << 1)

/* ----------- FILE: jserror.c            ------------ */
#include <errno.h>

#define JSCLEARSYSERROR()       (errno = 0)

#define JSERROR_MAXLENGTH       1000

#ifndef _ERRMODE_STRING

# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()     (JSErrorStackLevel = 0,	\
			      JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()     (JSErrNo = 0,		\
			      JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrNo = (code) + (off), JSErrLine = 0,				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#else
# ifdef DEBUG_EGOTO_VERBOSE
#  define JSCLEARERROR()						\
  (JSErrorStackLevel = 0,						\
   JSErrNo = JSErrLine = 0, JSErrLabel = NULL,				\
   JSErrString[0] = '\0')
# else
#  define JSCLEARERROR()					\
  (JSErrNo = JSErrLine = 0, JSErrLabel = NULL,			\
   JSErrString[0] = '\0')
# endif

# define JSSETERROR(code, off)						\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   JSErrLine = 0, JSErrString[0] = '\0')

# define JSSETERROR2X(code, off, string)				\
  (JSCLEARSYSERROR(),							\
   JSErrLabel = #code, JSErrNo = 1 + (off),				\
   strncpy(JSErrString, string, sizeof(JSErrString) - 1))
#endif

# define JSSETERRSTRING(string)					\
  strncpy(JSErrString, string, sizeof(JSErrString) - 1)

# define JSSETSYSERRSTRING(string)			\
  (JSCLEARERROR(), JSSETERRSTRING(string))

#define JSSETERROR2(code, off, string)				\
  JSSETERROR2X(code, off, (string) ? (string) : "<NULL>")

#define JSSETERRFILE(file, line)		\
  JSSETERRFILE2(file, line, 1)

#define JSSETERRFILE2(file, line, column)				\
  (strncpy(JSErrFile, (file) ? (file) : "<NULL>", sizeof(JSErrFile) - 1), \
   JSErrLine = (line), JSErrColumn = (column))

/*  These are internal error codes for debugging  */
#define JSMINIERROR             10000
#define JSSETIERROR(code, off)   JSSETERROR(JSMINIERROR + (code), off)
#define JSSETIERROR2(code, off, s) JSSETERROR2(JSMINIERROR + (code), off, s)

/* ----------- FILE: piggyback.c          ------------ */
#define PGBACK_MAGIC        "PGO1"
#define PGBACK_MAGIC_INV    "1OGP"

#define PGBACK_MAGIC_LEN    4

#define PGBACK_FINAL        PGBACK_MAGIC "\377\377"

#define PGBACK_FLEN4BYTE       (1 << 0)
#define PGBACK_FBWCOMP1        (1 << 1)
#define PGBACK_FBWENCRYPT      (1 << 2)

#ifdef CONFIG_HAS_BWCOMPRESSION
#define CONFIG_PIGGYBACK_COMPRESS
/*  This is rather big, but should get the best out of it  */
#define BWCOMP1_BLOCKSIZE        500000
#endif

/* ----------- FILE: tokpatmat.c          ------------ */
/* ***  A single state description of the finite state machine  *** */
typedef struct tomastatedescr_t tomastatedescr_t;
struct tomastatedescr_t {
    unsigned int TOMASD_Type;
#define TOMASD_TYPE_TOKENS           1
#define TOMASD_TYPE_CHOICE           2
#define TOMASD_TYPE_REPEAT           3

#define TOMASD_TYPE_EXPR             4
#define TOMASD_TYPE_EXPRSTART        5
#define TOMASD_TYPE_EXPREND          6

#define TOMASD_TYPE_STMT             7
#define TOMASD_TYPE_STMTSTART        8
#define TOMASD_TYPE_STMTEND          9

#define TOMASD_TYPE_IDENT           10
#define TOMASD_TYPE_NUMBER          11
#define TOMASD_TYPE_STRING          12

#define TOMASD_TYPE_LINENR          13
#define TOMASD_TYPE_COUNTER         14

    tomastatedescr_t *TOMASD_Next;
    union {
	/*  For TOMASD_TYPE_TOKENS. The indexes refer to TOPA_StateQueue  */
	struct {
	    unsigned int tomasd_TokStart, tomasd_TokEnd;

	} tomasd_us1;
#define TOMASD_TokStart     tomasd_u.tomasd_us1.tomasd_TokStart
#define TOMASD_TokEnd       tomasd_u.tomasd_us1.tomasd_TokEnd

	/*  For TOMASD_TYPE_CHOICE and TYPE_REPEAT  */
	struct {
	    tomastatedescr_t *tomasd_Sub;
	    unsigned int tomasd_ILoopVar;
	    unsigned int tomasd_MinLoop, tomasd_MaxLoop;
	} tomasd_us2;
#define TOMASD_Sub          tomasd_u.tomasd_us2.tomasd_Sub
#define TOMASD_ILoopVar     tomasd_u.tomasd_us2.tomasd_ILoopVar
#define TOMASD_MinLoop      tomasd_u.tomasd_us2.tomasd_MinLoop
#define TOMASD_MaxLoop      tomasd_u.tomasd_us2.tomasd_MaxLoop

	/*  For TOMASD_TYPE_EXPR + TOMASD_TYPE_STMT + TOMASD_TYPE_IDENT +
	 *   TOMASD_TYPE_NUMBER + TOMASD_TYPE_STRING
	 */
	unsigned int tomasd_ExprIndex;
#define TOMASD_ExprIndex     tomasd_u.tomasd_ExprIndex

    } tomasd_u;

    /*  Used for subordinates of TYPE_CHOICE to connect them  */
    tomastatedescr_t *TOMASD_Connect;
};

/* ***  Runtime structure to describe a state where we are  *** */
/*  To keep what a single variable matches  */
typedef struct {
    unsigned int MASTV_Start, MASTV_End;

} topavarmatch_t;

typedef struct matcherstate_t matcherstate_t;
struct matcherstate_t {
    matcherstate_t *MAST_Next;

    unsigned int MAST_TokenIndex; /*  refers to TOPA_InputQueue  */
    tomastatedescr_t *MAST_StateDescription;

    unsigned int *MAST_LoopCounter;
    topavarmatch_t MAST_VariableMatch[1];
};

/* ***  The overall structure  *** */
typedef struct {
    const char *TOPD_Name;

    tomastatedescr_t *TOPD_StartState;

    /*  Indexes into TOPA_StateQueue that contain the replacement or -1 if none  */
    unsigned int TOPD_Start, TOPD_End;

} tokpattdef_t;

#define TOPA_MEMORY_SIZE      200
typedef struct topamemory_t topamemory_t;
struct topamemory_t {
    topamemory_t *TOPM_Next;

    unsigned int TOPM_Index;
    tomastatedescr_t TOPM_Memory[TOPA_MEMORY_SIZE];
};

typedef struct {
    extfile_t *TOPA_Input;
    jctokfifo_t TOPA_InputQueue;

    unsigned int TOPA_NVariables;
    unsigned int TOPA_NCounters;
    unsigned int TOPA_MaxLoopCounter;

    tokpattdef_t *TOPA_Definitions;
    unsigned int TOPA_NDefinitions;

    jctokfifo_t TOPA_StateQueue; /*  for TOMASD_TYPE_TOKENS  */

    /*  Linked list to recycle matcher states  */
    matcherstate_t *TOPA_MatcherStates;

    /*  Automatically incremented counters  */
    unsigned long *TOPA_Counters;

    /*  Memory chunks for states  */
    topamemory_t *TOPA_StateMemory;

} tokpattmatch_t;

/* ----------- FILE: ErrCodes.c           ------------ */
#define _ERR_GETCTOKEN              1
#define _ERR_IFSTACK                6
#define _ERR_MKBACKUPNAME          10
#define _ERR_PIGGYBACKOP           11
#define _ERR_SAXPARSE              15
#define _ERR_TOPACREATE            28
#define _ERR_WILDCMP               40
#define _ERR_EDITFORM              42
#define _ERR_ADDMENU               43
#define _ERR_FINDMENU              44
#define _ERR_EDITTREE              46
#define _ERR_JSPPENVHANDLE         48
#define _ERR_JSPARSGETEXPRTYPE     49
#define _ERR_JSPARSGETUEXPRTYPE    50
#define _ERR_JSPARSMERGETYPES      57
#define _ERR_JSPARSGETSIZEOFTYPE   62
#define _ERR_JSPARSGETVARTYPE      65
#define _ERR_JSPARSGENVARDEF       66
#define _ERR_JSPARSSCANFLOW        67
#define _ERR_JSPARSPUTIDENT        71
#define _ERR_JSPARSSCANENUM        72
#define _ERR_JSPARSCHECKARRBOUND   73
#define _ERR_JSPARSFINDVARIABLE    74
#define _ERR_JSPARSGETVARDEF       75
#define _ERR_JSPARSDOPRAGMA        80

#define _ERR_MAXERROR              83

/* ----------- FILE: TraceCodes.c         ------------ */
#ifndef CONFIG_JSTRACE_NOTRACES
# define TRACETYPE_MAIN              0
# define TRACETYPE_TOKPATMAT         1
# define TRACETYPE_JSCPP             2
#endif


/* *******   Prototypes of library functions  ***** */
/*  For C++ compatibility   */
# ifdef __cplusplus
#  define Prototype extern "C"
# else
#  define Prototype extern
# endif
/* ***** File: ansisym.c             ***** */
Prototype const char * const ANSISymbols[];

/* ***** File: crccalc.c             ***** */
Prototype void           crcGenTab(unsigned long *table, unsigned long poly,
				   int flag);
Prototype unsigned long  crcCalc(unsigned long start, const unsigned char *buf,
				 long bufLen, const unsigned long *crcTab,
				 int flag);

/* ***** File: crctab.c              ***** */
Prototype const unsigned long JCrcTab[];

/* ***** File: critical.c            ***** */
Prototype void           criticalStart(void);
Prototype void           criticalEnd(void);

/* ***** File: cstrcompile.c         ***** */
Prototype long           cStrCompile(const char *src, char *dest, int mode);

/* ***** File: cstrfprint.c          ***** */
Prototype int            cStrFPrint(FILE *fp, const char *src, long len);

/* ***** File: cstrsprint.c          ***** */
Prototype int            cStrSPrint(char *dest, const char *src, long len);

/* ***** File: ctokfifo.c            ***** */
Prototype jctokfifo_t   *jctokcreate(jctokfifo_t *queue, int mode);
Prototype void           jctokfree(jctokfifo_t *queue, int flag);
Prototype int            jctokinsert(jctokfifo_t *queue, jsctoken_t token, int mode);
Prototype int            jctokwriten(jctokfifo_t *queue, jsctoken_t *tokens,
                                     int n, int where, int mode);
Prototype int            jctokdelete(jctokfifo_t *queue, int start, int end,
                                     int mode);
Prototype jsctoken_t     jctokget(jctokfifo_t *queue);
Prototype int            jctokgetfpos(jctokfifo_t *queue, const char **pName);
Prototype int            jctokgetindex(jctokfifo_t *queue);
Prototype void           jctokblank(jctokfifo_t *queue, int last);
Prototype int            jctoksubstitute(jctokfifo_t *queue, int start,
                                         int end, const char *format, ...);
Prototype int            jctoksubststring(jctokfifo_t *queue, int start,
                                          int end, const char *string);
Prototype char          *jctoktostring2(jctokfifo_t *queue, int start, int end,
                                        jsctoken_t (*converter)(void *user,
                                                                jsctoken_t token),
                                        void *user);
Prototype int            jctoktostrbuff1(jstrbuff_t *sbuffer,
                                         jsctoken_t token);
Prototype int            jctoktostrbuff(jstrbuff_t *sbuffer, jctokfifo_t *queue,
                                        int start, int end,
                                        jsctoken_t (*converter)(void *user,
                                                                jsctoken_t token),
                                        void *user);
Prototype int            jctokgetcolumn(jctokfifo_t *queue, int indx,
                                        int mode);
Prototype void           jctokshowqueue(FILE *fp, jctokfifo_t *queue);
Prototype void           jctokshowqueue2(FILE *fp, jctokfifo_t *queue,
                                         int start, int end);

/* ***** File: debug.c               ***** */
Prototype long           Debug;

/* ***** File: debugopts.c           ***** */
Prototype int            parsedbgopts(const dbgopts_t *optdef,
				      const char *options, long *pvalue,
				      int mode);

/* ***** File: dsleep.c              ***** */
Prototype double         dsleep(double secs);

/* ***** File: extfile.c             ***** */
Prototype extfile_t     *efopen(const char *name, const char *mode);
Prototype extfile_t     *effopen(FILE *fp, int iswrite);
Prototype extfile_t     *efmkdest(int (*pPut)(int c, void *p), void *p);
Prototype extfile_t     *efmksrc(int (*pGet)(void *, int), void *p);
Prototype int            efclose(extfile_t *efp);
Prototype FILE          *efileptr(extfile_t *efp);
Prototype int            efread(void *ptr, long size, extfile_t *efp);
Prototype int            efwrite(const void *ptr, long size, extfile_t *efp);
Prototype int            efgetc(extfile_t *efp);
Prototype int            eungetc(int c, extfile_t *efp);
Prototype int            efputc(int c, extfile_t *efp);
Prototype int            epushmod(extfile_t *efp, int (*module)(void *, int *pC),
				  void *user, int where);
Prototype int            epopmod(extfile_t *efp, int (*module)(void *, int *pC),
				 void **puser);
Prototype int            epushopener(const char *prefix,
				     extfile_t *(*opener)(void *user,
							  const char *postfix,
							  const char *mode),
				     void *user);
Prototype int            epopopener(const char *prefix);
Prototype const char    *EFilePath;

/* ***** File: fexists.c             ***** */
Prototype int            fileexists2(const char *name, int mode);

/* ***** File: file2buff.c           ***** */
Prototype char          *file2buffer(const char *name, long *pLen);
Prototype char          *ffile2buffer(FILE *fp, long *pLen);
Prototype char          *efile2buffer(extfile_t *efp, long *pLen);

/* ***** File: filelen.c             ***** */
Prototype off_t          fileLength(const char *name);
Prototype off_t          fpFileLength(FILE *fp);
Prototype off_t          fdFileLength(int fd);

/* ***** File: flocks.c              ***** */
Prototype int            lockRegion2(int fd, int type, long offset, int whence,
				     long len);
Prototype FILE          *fopenLock(const char *name, const char *mode, int type);

/* ***** File: fopencreate.c         ***** */
Prototype FILE          *fopenPermit(const char *name, const char *mode);
Prototype FILE          *fopenCreate(const char *name, const char *mode);
Prototype int            statPermit(const char *name, void *bufPtr);
Prototype int            lstatPermit(const char *name, void *bufPtr);
Prototype int            utimePermit(const char *name, void *bufPtr);
Prototype int            symlinkPermit(const char *oldpath,
				       const char *newpath);
Prototype int            chmodPermit(const char *name, unsigned long mode);
Prototype int            FOpCreateVerbose;

/* ***** File: fopenpath.c           ***** */
Prototype FILE		*fopenPath3(const char *name, const char *mode,
				    const char *path, int flag);
Prototype FILE		*fopenScan(const char *name, const char *mode);
Prototype const char	*FOpenName, *FOpenPath;

/* ***** File: fstralloc.c           ***** */
Prototype char          *fstralloc(fstralloc_t *fstr, unsigned int len);
Prototype char          *fstrdup(fstralloc_t *fstr, const char *string);
Prototype void           fstrclear(fstralloc_t *fstr);
Prototype void           fstrfree(fstralloc_t *fstr);
Prototype void           fstrmarksave(const fstralloc_t *fstr,
				      fstrallocmark_t *pmark);
Prototype void           fstrmarkrestore(fstralloc_t *fstr,
					 const fstrallocmark_t *pmark);
Prototype unsigned int   FStringAllocSize;

/* ***** File: getctoken.c           ***** */
Prototype jsctoken_t     getCToken(extfile_t *eFp, const jsctokops_t *opDef,
                                   int mode);
Prototype int            printCToken(FILE *fp, jsctoken_t token);
Prototype int            sprintCToken(char *s, jsctoken_t token);
Prototype jsctoken_t     copyCToken2(fstralloc_t *fstr, jsctoken_t token, int mode);
Prototype void           freeCToken(jsctoken_t token);
Prototype const char    *tkStrcompare(const char *input, const char *string);
Prototype int            tkStrcmp(const char *input, const char *string);
Prototype int            countnlCToken(const char *string);
Prototype int            updcolCToken(jsctoken_t token, int column);
Prototype void           tkDelNewLines(char *string);
Prototype const char    *dbgNameCToken(jsctoken_t token);
Prototype int            GCTKTerminator;

/* ***** File: ifstack.c             ***** */
Prototype ifstack_t     *ifstackcreate(ifstack_t *istk);
Prototype int            ifstackfree(ifstack_t *istk, int flag);
Prototype int            ifstackop(ifstack_t *istk, int op, int cond);

/* ***** File: jserror.c             ***** */
Prototype int            JSErrNo;
Prototype char           JSErrString[JSERROR_MAXLENGTH + 1];
Prototype const char    *JSErrLabel;
Prototype int            JSErrLine, JSErrColumn;
Prototype char           JSErrFile[JS_FILENAME_MAX + 1];
Prototype int            JSErrorStackLevel;
Prototype char           JSErrorStack[100][JSERROR_MAXLENGTH];

/* ***** File: jsgetline.c           ***** */
Prototype char          *jsgetline(FILE *fp, int mode);
Prototype void           jsgetlineclear(void);

/* ***** File: jsgetopt.c            ***** */
Prototype int		 jsGetOpt(int *pargc, void *argv, const char *optstring,
				  const char **poptarg);
Prototype int            jsFirstArg(char *argv[], const char *optstring);
Prototype int		 JSOptMode, JSOptIndex;

/* ***** File: jsperror.c            ***** */
Prototype void           jsperror(const char *progname);

/* ***** File: jsstrerr.c            ***** */
Prototype const char    *jsStrError(int err);
Prototype int            JSErrorMode;

/* ***** File: jstrace.c             ***** */
Prototype void           jstrace(int type, int level,
				 const char *format, ...);
Prototype FILE          *jstracegetfp(int type, int level);
Prototype int            jstracelevel(int type);
Prototype int            jstracemask(int type);
Prototype int            jstraceset(const char *type, int level);
Prototype int            jstraceseti(int ttype, int level);
Prototype int            jstracesets(const char *string);
Prototype char           JSTraceDirectory[JS_FILENAME_MAX];
Prototype int            JSTraceFlags;

/* ***** File: microtime.c           ***** */
Prototype int            getMicroTime(time_t *pTime, long *pMicroTime);

/* ***** File: mkbackup.c            ***** */
Prototype const char    *mkBackupFile2(const char *fileName, char *oName,
				       int mode);
Prototype int            mkBackupName(char *name, int mlen, const char *ext);

/* ***** File: mystrdup.c            ***** */
Prototype char          *mystrdup(const char *s);

/* ***** File: piggyback.c           ***** */
Prototype extfile_t     *piggybackopener(void *data, const char *name,
 					 const char *mode);
Prototype void           piggybackcrypt(long pwd, unsigned char *data,
					unsigned long len);
Prototype unsigned long  piggystring2pwd(const char *name);
Prototype int            piggybackscan(unsigned char *data,
				       int (*routine)(void *user,
						      int flags, long len,
						      const char *name,
						      unsigned char *d),
				       void *user,
				       unsigned char **pend, int mode);
Prototype unsigned long (*PiggyBackGetPwd)(const char *name);

/* ***** File: random.c              ***** */
Prototype double         ran0(const long *pseed);
Prototype long           lran0(const long *pseed);
Prototype double         ran1(const long *pseed);
Prototype long           lran1(const long *pseed);

/* ***** File: saxparser.c           ***** */
Prototype saxparse_t    *saxparsecreate(int mode);
Prototype void           saxparsefree(saxparse_t *sapa);
Prototype int            saxparse(saxparse_t *sapa, FILE *fp,
				  unsigned int *pLineNr);

/* ***** File: sharestrdup.c         ***** */
Prototype jssharestr_t  *sharestrinit(jssharestr_t *ss);
Prototype void           sharestrdeinit(jssharestr_t *ss, int flag);
Prototype const char    *sharestrdup(jssharestr_t *ss, const char *string);

/* ***** File: strbuffer.c           ***** */
Prototype int            jstrbuffstore(jstrbuff_t *sb, const char *s, long len);
Prototype int            jstrbuffprint(jstrbuff_t *sb, const char *s);
Prototype int            jstrbuffprintf(jstrbuff_t *sb, const char *format, ...);
Prototype int            jstrbuffvprintf(jstrbuff_t *sb, const char *format,
					va_list ap1, va_list ap2);
Prototype int            jstrbuffins(jstrbuff_t *sb, const char *s, long off,
				    long len);
Prototype void           jstrbuffdel(jstrbuff_t *sb, long off, long len);

/* ***** File: strcomp.c             ***** */
Prototype const char    *jstrstartswith(const char *input, const char *string);
Prototype const char    *jstristartswith(const char *input, const char *string);
Prototype int            jstrendswith(const char *input, const char *string);

/* ***** File: stricmp.c             ***** */
Prototype int            jsstricmp(const char *s1, const char *s2);
Prototype int            jsstrincmp(const char *s1, const char *s2, size_t len);

/* ***** File: strtoid.c             ***** */
Prototype long           jstrtoid(const char *input);

/* ***** File: strtol2.c             ***** */
Prototype long           jstrtol2(const char *input, char **endp, int base);
Prototype unsigned long  jstrtoul2(const char *input, char **endp, int base);

/* ***** File: tokpatmat.c           ***** */
Prototype tokpattmatch_t *topacreate(const char *fileName);
Prototype void           topadelete(tokpattmatch_t *topa);
Prototype int            topaprocess(tokpattmatch_t *topa, extfile_t *efp,
				     const char *fileName, FILE *fpOut,
				     const char **matchers, int mode);

/* ***** File: updatecol.c           ***** */
Prototype int            updateColumn(int column, const char *string);

/* ***** File: vaprintf.c            ***** */
Prototype char          *aprintf(const char *format, ...);
Prototype char          *vaprintf(const char *format, va_list ap1,
				  va_list ap2);
Prototype long           vlprintf(const char *format, va_list ap);

/* ***** File: wildcmp.c             ***** */
Prototype int		 wildcmp(const char *name, const char *wild, int mode);
Prototype int		 jiswildcard(const char *name);

/* ***** File: ErrCodes.c            ***** */
Prototype const char * const JSErrors[];
Prototype const int          JSErrorsI[];

/* ***** File: TraceCodes.c          ***** */
Prototype const char * const JSTraceTypes[];
Prototype int                JSTraceLevels[];
Prototype int                JSNTraceTypes;


# undef Prototype   /* I like to be clean  */

#endif /*  "#ifdef JSSUBS_H"  */
