/*JS*********************************************************************
*
*    Program : JEDFRAME
*    Language: ANSI-C + CURSES routines.
*    Author  : Joerg Schoen
*    Purpose : Edit text frames (like an editor).
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <jsconfig.h>
#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
#ifndef TFRAME_INSMODE
/*  START-DEFINITIONS  */
/* ***  Definitions for text frames  *** */
#define TFRAME_INSMODE         (1<<0) /*  Insert characters, not overwrite  */
#define TFRAME_SCROLLBAR       (1<<1) /*  Display scroll bar  */
#define TFRAME_ALLOCATE        (1<<2) /*  Reallocate text region if growing  */
#define TFRAME_CHANGED         (1<<3) /*  Mark that something was changed  */
#define TFRAME_FILLBLANK       (1<<4) /*  Fill with blanks (instead of
				       *   'TFFillCharacter').
				       */
#define TFRAME_NONEWLINE       (1<<5) /*  Do not treat newline, return
				       *   'unknown'.
				       */

typedef struct {
  int TF1_LineNr;
  long TF1_LineOff;
} jedtextline_t;

typedef struct {
  int TF_Mode;

  /*  The text to display, its length, maximum length and number of lines  */
  char *TF_Text;
  long TF_Length, TF_MaxLen, TF_NLines;

  int TF_Line, TF_Column, TF_Height, TF_Width;
  int TF_TopL, TF_BottL;
  int TF_CuLine, TF_CuColumn; /*  cursor position  */
  long TF_CuOffset; /*  cursor position as offset  */
  long TF_CuMark; /*  current text mark  */

  int TF_ScrollBarS, TF_ScrollBarE; /*  start and end of scrollbar  */

  long TF_StartOff; /*  offset of start of first screen line  */
  jedtextline_t TF_Lines[1]; /*  line numbers/offset pairs  */
} jedtext_t;

#define jedtextdelete(tf)     free(tf)
/*  END-DEFINITIONS  */
#endif

/*
 *  Empty lines on the screen, i. e. lines outside the text, are shown empty
 *   (like in emacs), not like vi. We can edit a text that does not end in a
 *   newline and can go to the last+1 line to continue writing (like emacs),
 *   so that makes sense. The vi like display is a bit confusing if going on
 *   the last+1 line.
 */
#if 1
# define EMPTY_LINE             ""
# define NEMPTY_LINE            0
#else
# define EMPTY_LINE             "~"
# define NEMPTY_LINE            1
#endif

#if 0 /*  Show TABS as arrows  */
#define SHOWTABS
#endif

/*  How many lines to scroll if the cursor moves
 *   out of the current window.
 */
#define FRAME_SCROLL_SKIP           5

/* Update start and end of scrollbar  */
#define tframesetscbar(tf)    _jedsetscrollbar((tf)->TF_Lines[0].TF1_LineNr, \
                                            (tf)->TF_NLines, (tf)->TF_Height, \
                                            &((tf)->TF_ScrollBarS), \
                                            &((tf)->TF_ScrollBarE))

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype jedtext_t     *jedtextcreate(char *text, long maxlen, int line,
				       int column, int height, int width,
				       int mode);
static void              tframeinit(jedtext_t *tf);

Prototype int            jedtextdokey(WINDOW *win, jedtext_t *tf, int key,
				      int mode);

static int               tframemove(jedtext_t *tf, int d);
static int               tframescroll(jedtext_t *tf, int d);
static int               tframealign(jedtext_t *tf);

static int               tframeinsert(jedtext_t *tf, const char *s, long len,
				      int insmode);
static int               tframeredraw(jedtext_t *tf, WINDOW *win, int l1, int l2);
static int               tframesetoffs(jedtext_t *tf, int l1, int l2, int flag);
static int               tframeo2lc(jedtext_t *tf);
static void              tframelc2o(jedtext_t *tf);
static int               tframelstart(jedtext_t *tf, int line);

static long              getnlines(const char *s);
static int               getlinelen(const char *s, int width);
static int               getfulllinelen(const char *s);
static int               findlstart(const char *text, long offset);
static int               findnextl(const char *text, long offset);
#ifdef DEBUG
static void              checkframe(jedtext_t *tf, int key);
#endif

/*JS*********************************************************************
*   Set up a TextFrame structure for the given text at the given position
*    in a curses position. maxlen gives the amount of available space at
*    text.
*************************************************************************/

jedtext_t *jedtextcreate(char *text, long maxlen, int line, int column,
			 int height, int width, int mode)

/************************************************************************/
{
  jedtext_t *tf;

  if((tf = (jedtext_t *)malloc(sizeof(*tf) + (height + 1)
			       * sizeof(*(tf->TF_Lines)))) == NULL)
    return(NULL);

  tf->TF_Mode   = mode;
  tf->TF_Text   = text;
  tf->TF_Length = strlen(text);
  tf->TF_MaxLen = maxlen;
  tf->TF_NLines = getnlines(text);

  tf->TF_Line   = line;
  tf->TF_Column = column;
  tf->TF_Height = height;
  tf->TF_Width  = width;

  tf->TF_TopL   = tf->TF_Line;
  tf->TF_BottL  = tf->TF_Line + tf->TF_Height - 1;

  tframeinit(tf);

  return(tf);
}

/*JS*********************************************************************
*   Initialize the text frame, going to the first line/column.
*************************************************************************/

static void tframeinit(jedtext_t *tf)

/************************************************************************/
{
  tf->TF_Mode |= TFRAME_INSMODE;
  tf->TF_CuLine = tf->TF_CuColumn = 0;
  tf->TF_CuOffset = tf->TF_StartOff = 0;
  tf->TF_CuMark = -1;

  tf->TF_Lines[0].TF1_LineNr = 0;
  tf->TF_Lines[0].TF1_LineOff = 0;
  tframesetoffs(tf, 0, tf->TF_Height, 0);
  tframesetscbar(tf);
}

/*  Action flags for _tframedokey  */
#define FLAG_REDRAW      (1<<0)
#define FLAG_MOVECURSOR  (1<<1)
#define FLAG_REFRESH     (1<<2)
#define FLAG_ILLEGAL     (1<<3)
#define FLAG_UNKNOWN     (1<<4)
#define FLAG_DOALIGN     (1<<5)

/*JS*********************************************************************
*   Handle the keystroke for the text frame, doing all required updates
*    in the window. If Bit 0 of mode is set, don't refresh the window.
*    If Bit 1 is set, don't move the curses cursor to its new position.
*    Returns -1 on error, 0 if keystroke was handled. If the keystroke
*    was illegal, returns 1.
*************************************************************************/

int jedtextdokey(WINDOW *win, jedtext_t *tf, int key, int mode)

/************************************************************************/
{
  int flag, height, l1, l2, c;

  height = tf->TF_Height;
  flag = 0;
  l1 = 0;
  l2 = height - 1;

#ifdef DEBUG
  if(key != EDKEY_REDRAW) checkframe(tf, key);
#endif

  switch(key) {
  case EDKEY_REDRAW:
    flag |= FLAG_REDRAW;
    break;
  case EDKEY_LEFT:
    if(tf->TF_CuOffset == 0) {
      flag |= FLAG_ILLEGAL;
    } else {
      tf->TF_CuOffset--;
      flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
      tframeo2lc(tf);
    }
    break;
  case EDKEY_RIGHT:
    if(tf->TF_CuOffset >= tf->TF_Length) {
      flag |= FLAG_ILLEGAL;
    } else {
      tf->TF_CuOffset++;
      flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
      tframeo2lc(tf);
    }
    break;
  case EDKEY_UP:
    if((c = tframemove(tf, -1)) == -1)
      flag |= FLAG_ILLEGAL;
    else
      flag |= c ? FLAG_REDRAW : FLAG_MOVECURSOR;
    break;
  case EDKEY_DOWN:
    if((c = tframemove(tf, 1)) == -1)
      flag |= FLAG_ILLEGAL;
    else
      flag |= c ? FLAG_REDRAW : FLAG_MOVECURSOR;
    break;
  case EDKEY_PUP:
    if(tframescroll(tf, -height) == 0)
      flag |= FLAG_ILLEGAL;
    else {
      tframelc2o(tf);
      flag |= FLAG_REDRAW;
    }
    break;
  case EDKEY_PDOWN:
    if(tf->TF_Lines[height].TF1_LineNr == -1 || tframescroll(tf, height) == 0)
      flag |= FLAG_ILLEGAL;
    else {
      tframelc2o(tf);
      flag |= FLAG_REDRAW;
    }
    break;
  case EDKEY_LSTART:
    tf->TF_CuOffset = findlstart(tf->TF_Text, tf->TF_CuOffset);
    flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
    tframeo2lc(tf);
    break;
  case EDKEY_LEND:
    if((c = findnextl(tf->TF_Text, tf->TF_CuOffset)) == -1)
      c = tf->TF_Length;
    else if(c > 0)
      c--;

    tf->TF_CuOffset = c;

    flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
    tframeo2lc(tf);
    break;
  case EDKEY_START:
    tf->TF_CuOffset = 0;
    flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
    tframeo2lc(tf);
    break;
  case EDKEY_END:
    tf->TF_CuOffset = tf->TF_Length;
    flag |= tframealign(tf) ? FLAG_REDRAW : FLAG_MOVECURSOR;
    tframeo2lc(tf);
    break;
  case EDKEY_BS:
    if(tf->TF_CuOffset == 0) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    tf->TF_CuOffset--;
    if(tframealign(tf)) flag |= FLAG_REDRAW;
    tframeo2lc(tf);
  case EDKEY_DELETE:
    if(tf->TF_CuOffset == tf->TF_Length) {
      flag |= FLAG_ILLEGAL;
      break;
    }
    l1 = tf->TF_CuLine;
    if((l2 = tframeinsert(tf, NULL, 1, 0)) == -1) EGOTO(error);
    if(l2 == -2) {
      flag |= FLAG_ILLEGAL;
      break;
    }

    if(flag & FLAG_REDRAW) {/*  already set by EDKEY_BS? */
      l1 = 0;
      l2 = height - 1;
    }

    flag |= FLAG_REDRAW;
    break;
  case EDKEY_KILLEOL:
    if((c = findnextl(tf->TF_Text, tf->TF_CuOffset)) == -1) c = tf->TF_Length;

    if(c > (tf->TF_CuOffset + 1) && tf->TF_Text[c - 1] == '\n') c--;

    if((c -= tf->TF_CuOffset) == 0) {
      flag |= FLAG_ILLEGAL;
      break;
    }

    if((l2 = tframeinsert(tf, NULL, c, 1)) == -1) EGOTO(error);
    l1 = tf->TF_CuLine;

    flag |= FLAG_REDRAW;
    tframeo2lc(tf);
    break;
  case EDKEY_SETMARK:
    tf->TF_CuMark = tf->TF_CuOffset;
    break;
  case EDKEY_KILLREGION:
    if(tf->TF_CuMark == -1 || tf->TF_CuMark == tf->TF_CuOffset) {
      flag |= FLAG_ILLEGAL;
      break;
    }

    if(tf->TF_CuMark < tf->TF_CuOffset) {
      long t = tf->TF_CuMark;
      tf->TF_CuMark = tf->TF_CuOffset;
      tf->TF_CuOffset = t;
      if(tframealign(tf) == -1)
	EGOTO(error);

      tframeo2lc(tf);
    }

    c = tf->TF_CuMark - tf->TF_CuOffset;
    if((l2 = tframeinsert(tf, NULL, c, 1)) == -1) EGOTO(error);

    /*tframeo2lc(tf);*/
    l1 = tf->TF_CuLine;

    flag |= FLAG_REDRAW;
    break;
  case EDKEY_YANK:
    {
      const char *s;
      int len;

      if((s = _jedgetyank())) {
	len = strlen(s);

	l1 = tf->TF_CuLine;
	if((l2 = tframeinsert(tf, s, len, 1)) == -1) EGOTO(error);
	if(l2 == -2) {
	  flag |= FLAG_ILLEGAL;
	  break;
	}

	tf->TF_CuOffset += len;
	flag |= FLAG_REDRAW | FLAG_DOALIGN;
      } else
	flag |= FLAG_ILLEGAL;
    }
    break;
  case EDKEY_ROTYANK:
    _jedrotkillring(-1);
    break;
  case EDKEY_TOGGLEINS:
    tf->TF_Mode ^= TFRAME_INSMODE;
    break;
#ifdef NCURSES_MOUSE_VERSION
  case KEY_MOUSE:
    {
      int ml, mc;
      MEVENT mev;

      if(getmouse(&mev) == ERR) EGOTO(error);
      getbegyx(win, ml, mc);
      ml = mev.y - tf->TF_Line - ml;
      mc = mev.x - tf->TF_Column - mc;

      if(ml < 0 || ml >= tf->TF_Height || mc < 0 || mc >= tf->TF_Width) {
	flag |= FLAG_ILLEGAL;
	break;
      }

      if(mev.bstate & BUTTON1_CLICKED) {
	tf->TF_CuLine = ml;
	tf->TF_CuColumn = mc;
	tframelc2o(tf);
	flag |= FLAG_MOVECURSOR;
      }
    }
    break;
#endif
  default:
    /*  Literal key?  */
    if((c = key) < 0) key = -key;

    if(!(key < 256 &&
	 ((tf->TF_Mode & TFRAME_NONEWLINE) ?
	  (isprint(key) || (c < 0 && (key == '\n' || key == '\t'))) :
	  (isprint(key) || (c > 0 && (key == '\n' || key == '\t')))))) {
      flag |= FLAG_UNKNOWN;
      break;
    }

    l1 = tf->TF_CuLine;
    {
      char cc;

      cc = key;
      /*  The most common case of inserting a single character
       *   is also handled by the more general routine (because
       *   of all the possible complications that might arise)
       */
      if((l2 = tframeinsert(tf, &cc, 1, 0)) == -1) EGOTO(error);
      if(l2 == -2) {
	flag |= FLAG_ILLEGAL;
	break;
      }
    }

    tf->TF_CuOffset++;
    flag |= FLAG_REDRAW | FLAG_DOALIGN;
    break;
  }

#ifdef XDEBUG
  dbgprintf("KEY %d [%d %d] [%d %d %d] [%d] %d\n", key,
	    l1, l2, tf->TF_CuLine, tf->TF_CuColumn,
	    tf->TF_CuOffset, tf->TF_StartOff, tf->TF_Length);
  for(c = 0 ; c <= height ; c++)
    dbgprintf("  %d: %d %d\n", c, tf->TF_Lines[c].TF1_LineNr,
	      tf->TF_Lines[c].TF1_LineOff);
#endif

  if(flag & FLAG_DOALIGN) {
    if(tframealign(tf)) {
      l1 = 0;
      l2 = height - 1;
    }
    tframeo2lc(tf);
  }

  if(flag & FLAG_REDRAW) {
    tframeredraw(tf, win, l1, l2);
    flag |= FLAG_MOVECURSOR;
  }

  if((flag & FLAG_MOVECURSOR) && !(mode & 2)) {
    wmove(win, tf->TF_Line + tf->TF_CuLine, tf->TF_Column + tf->TF_CuColumn);
#ifdef CONFIG_HAS_CURSSET
    curs_set(1);
#endif
  }

  if((flag & FLAG_REFRESH) && !(mode & 1)) jedwstackrefresh(win);

  if(flag & FLAG_UNKNOWN) return(-2);
  if(flag & FLAG_ILLEGAL) return(-3);

  return 0;
error:
  return -1;
}

#undef FLAG_REDRAW
#undef FLAG_MOVECURSOR
#undef FLAG_REFRESH
#undef FLAG_ILLEGAL
#undef FLAG_UNKNOWN
#undef FLAG_DOALIGN

/*JS*********************************************************************
*   Move the cursor one line down (d = +1) or up (d = -1), aligning the
*    visible window if necessary.
*************************************************************************/

static int tframemove(jedtext_t *tf, int d)

/************************************************************************/
{
  long offset, startoff;
  int line, col;

  line = tframelstart(tf, tf->TF_CuLine);
  startoff = line > 0 ? tf->TF_Lines[line].TF1_LineOff : tf->TF_StartOff;

  offset = tf->TF_CuOffset;

  /*  The column in the current line  */
  col = offset - startoff;

  if(d == -1) {
    if(startoff == 0) return(-1);

    offset = findlstart(tf->TF_Text, startoff - 1);
    if((offset += col) >= startoff)
      /* The previous line is shorter than our current, so go to last column */      offset = startoff - 1;
  } else if(d == 1) {
    if((startoff = findnextl(tf->TF_Text, offset)) == -1)
      /*  We are already in the last line  */
      return(-1);

    if((line = getfulllinelen(tf->TF_Text + startoff)) <= col &&
       (col = line - 1) < 0)
      col = 0;

    offset = startoff + col;
  }
  tf->TF_CuOffset = offset;

  col = tframealign(tf);
  tframeo2lc(tf);

  return(col);
}

/*JS*********************************************************************
*   Scroll the text frame by d screenlines (a text line can be comprised
*    of several screen lines). d can be positive/negative. Returns the
*    number of lines actually scrolled with the same sign as d.
*************************************************************************/

static int tframescroll(jedtext_t *tf, int d)

/************************************************************************/
{
  int c, height;

  height = tf->TF_Height;
  if(d < 0) { /* ***   Scroll window down  *** */
    int i, nr, offset;

    if(d < -tf->TF_Height) d = -tf->TF_Height;
    c = d;
    d = -d;
    /*  First move lines downward to generate space for the new ones  */
    memmove(&(tf->TF_Lines[d]), &(tf->TF_Lines[0]), (height - d + 1) *
	    sizeof(*(tf->TF_Lines)));

    if((nr = tf->TF_Lines[0].TF1_LineNr) == -1) nr = tf->TF_NLines;
    offset = tf->TF_StartOff;
    while(d > 0) {
      long end;

      end = tf->TF_Lines[d].TF1_LineOff;
      if(offset < end) {
	long off2;

	for(off2 = offset, i = 0 ; off2 < end ; i++) {
	  long l;

	  if(i == d) {
	    if(--i > 0)
	      /*  This line has too many screenlines, so scroll  */
	      memmove(&(tf->TF_Lines[0]), &(tf->TF_Lines[1]), i *
		      sizeof(*(tf->TF_Lines)));
	  }

	  tf->TF_Lines[i].TF1_LineNr  = nr;
	  tf->TF_Lines[i].TF1_LineOff = off2;
	  if((l = getlinelen(tf->TF_Text + off2, tf->TF_Width)) == 0) break;
	  off2 += l;
	  if(tf->TF_Text[off2 - 1] == '\n') nr++;
	}

	if(i > 0 && i < d) {
	  memmove(&(tf->TF_Lines[d - i]), &(tf->TF_Lines[0]), i *
		  sizeof(*(tf->TF_Lines)));
	}
	d -= i;
      } else if(offset > 0) {
	offset = tf->TF_StartOff = findlstart(tf->TF_Text, offset - 1);
	nr = tf->TF_Lines[0].TF1_LineNr - 1;
      } else
	/*  We cannot scroll enough lines down since we are already at the
	 *   the start of the text
	 */
	break;
    }

    /*  Did we successfully scroll?  */
    if(d > 0) {
      memmove(&(tf->TF_Lines[0]), &(tf->TF_Lines[d]), (height - d + 1) *
	      sizeof(*(tf->TF_Lines)));
      tframesetoffs(tf, height - d, height, 0);
      c += d;
    }
  } else { /* ***   Scroll window up  *** */
    if(d > tf->TF_Height) d = tf->TF_Height;

    for(c = 0 ; c < d ; c++) {
      if(tf->TF_Lines[c].TF1_LineNr == -1) {
	if(c > 0) c--;
	break;
      }
      if(tf->TF_Lines[c].TF1_LineNr != tf->TF_Lines[c + 1].TF1_LineNr &&
	 tf->TF_Lines[c + 1].TF1_LineNr != -1)
	tf->TF_StartOff = tf->TF_Lines[c + 1].TF1_LineOff;
    }

    if(c > 0) {
      if(tf->TF_Lines[c].TF1_LineNr == -1) c--;

      memmove(&(tf->TF_Lines[0]), &(tf->TF_Lines[c]), (height - c + 1) *
	      sizeof(*(tf->TF_Lines)));
      tframesetoffs(tf, height - c, height, 0);
    }
  }

  if(c) tframesetscbar(tf);

  return(c);
}

/*JS*********************************************************************
*   Scroll the window up/down to make the current cursor position
*    visible. Returns -1 on internal error, 1 if any scrolling was done,
*    otherwise 0.
*************************************************************************/

static int tframealign(jedtext_t *tf)

/************************************************************************/
{
  long offset;
  int n, ret;

  offset = tf->TF_CuOffset;
  n = tf->TF_Height;
  ret = 0;

  if(offset < tf->TF_Lines[0].TF1_LineOff) {
    /*  offset before first line, so scroll down  */
    do {
      if(tframescroll(tf, -FRAME_SCROLL_SKIP) == 0)
	return -1;

      ret = 1;
    } while(offset < tf->TF_Lines[0].TF1_LineOff);

  } else if(offset >= tf->TF_Lines[n].TF1_LineOff &&
	    offset > tf->TF_Lines[n-1].TF1_LineOff) {
    /*  offset after last line, so scroll up  */
    do {
      if(tframescroll(tf, FRAME_SCROLL_SKIP) == 0)
	return -1;

      ret = 1;
    } while(offset >= tf->TF_Lines[n].TF1_LineOff &&
	    offset > tf->TF_Lines[n-1].TF1_LineOff);
  }

  if(ret) tframesetscbar(tf);

  return(ret);
}

/*JS*********************************************************************
*   If s != NULL, insert the len characters of the string at the current
*    cursor position. If s == NULL, delete len characters at the current
*    cursor position. On growth, reallocate the field if marked so.
*    Returns -1 on no memory, -2 if no reallocation is possible and the
*    field run out of space. Otherwise, recalculate the offsets and
*    return the last line where changes will be visible.
*************************************************************************/

static int tframeinsert(jedtext_t *tf, const char *s, long len, int insmode)

/************************************************************************/
{
  char *text;
  long off, i, j;

  text = tf->TF_Text;
  off = tf->TF_CuOffset;

  if(s) { /*  INSERT  */
    long end;

    /*  We insert at the end of the line (that is what most editors do)  */
    if(!insmode)
      insmode = (tf->TF_Mode & TFRAME_INSMODE) || text[off] == '\n';

    end = insmode ? (tf->TF_Length + len) : (off + len);

    if(end >= tf->TF_MaxLen) {
      /*  Are we allowed to reallocate?  */
      if(!(tf->TF_Mode & TFRAME_ALLOCATE)) return(-2);
      tf->TF_MaxLen = MAX(tf->TF_MaxLen + 200, end + 1);
      if((tf->TF_Text = text = (char *)realloc(text, tf->TF_MaxLen)) == NULL)
	return(-1);
    }

    if(insmode) {
      end = off + len;
      memmove(text + end, text + off, tf->TF_Length + 1 - off);
      tf->TF_Length += len;
    } else {
      for(i = off ; i < end /* == off+len */ && text[i] ; i++)
	if(text[i] == '\n') tf->TF_NLines--;

      /*  Maybe we appended at the end of text?  */
      if(end > tf->TF_Length) {
	tf->TF_Length = end;
	text[end] = '\0';
      }
    }

    /*  Now insert the string, adjusting the number of lines  */
    for(i = off ; i < end ; i++, s++) {
      if(*s == '\n') tf->TF_NLines++;
      text[i] = *s;
    }
    off = insmode ? len : 0;
  } else { /*  DELETE  */
    long end;

    if((end = off + len) > tf->TF_Length) {
      len = tf->TF_Length - off;
      end = tf->TF_Length;
    }

    for(i = off ; i < end ; i++)
      if(text[i] == '\n') tf->TF_NLines--;

    if(insmode && _jedinskillring(text + off, len) == -1) return(-1);

    memmove(text + off, text + end, tf->TF_Length + 1 - end);
    tf->TF_Length -= len;
    off = -len;
  }

  /*  Adjust the offsets of all following lines  */
  j = tf->TF_CuLine;
  for(i = j + 1 ; i <= tf->TF_Height ; i++) tf->TF_Lines[i].TF1_LineOff += off;

  tf->TF_Mode |= TFRAME_CHANGED;
  tframesetscbar(tf);
  return(tframesetoffs(tf, j, tf->TF_Height, 1));
}

/*JS*********************************************************************
*   Redraw the text frame from line l1 to l2.
*************************************************************************/

static int tframeredraw(jedtext_t *tf, WINDOW *win, int l1, int l2)

/************************************************************************/
{
  const char *text;
  int line, column, width;
  long o1, o2;

  line = tf->TF_Line;
  column = tf->TF_Column;

  text = tf->TF_Text;
  width = tf->TF_Width;
  for(o1 = tf->TF_Lines[l1].TF1_LineOff ; l1 <= l2 ; l1++, o1 = o2) {
    const char *s;
    int col;
    long l;

    o2 = tf->TF_Lines[l1 + 1].TF1_LineOff;
    if(tf->TF_Lines[l1].TF1_LineNr == -1) {
      s = EMPTY_LINE;
      l = NEMPTY_LINE;
    } else {
      s = text + o1;
      l = o2 - o1;

      /*  Don't print the actual newline character, since it
       *   will mess up the screen.
       */
      if(l > 0 && s[l - 1] == '\n') l--;
    }
    if((line + l1) < tf->TF_TopL || (line + l1) > tf->TF_BottL) continue;

    if(wmove(win, line + l1, column) == ERR) EGOTO(error);

    for(col = 0 ; l-- > 0 ; s++) {
      if(*s == '\t') {
	int col2;

	col2 = (col + 8) & ~7;
#ifdef SHOWTABS
	if(waddch(win, (JEditCursesMode & CURSMODE_NOACSGRAPH) ? '>' :
		  ACS_RARROW) == ERR)
	  EGOTO(error);
	col++;
#endif
	for( ; col < col2 ; col++)
	  if(waddch(win, ' ') == ERR)
	    EGOTO(error);

      } else if(*s) {
	if(waddch(win, *(unsigned char *)s) == ERR)
	  EGOTO(error);

	col++;
      } else
	break;
    }

    l = (tf->TF_Mode & TFRAME_FILLBLANK) ? ' ' : JEditTFFillChars;
    while(col++ < width) if(waddch(win, l) == ERR) EGOTO(error);

    if(tf->TF_Mode & TFRAME_SCROLLBAR) {
      if(tf->TF_NLines < tf->TF_Height) {
	if(waddch(win, ' ') == ERR) EGOTO(error);
      } else if(waddch(win, JEDScrollChars[tf->TF_ScrollBarS <= l1 && l1 < tf->TF_ScrollBarE ? 1 : 0]) == ERR)
	EGOTO(error);
    }
  }

  return(0);
error:
  return(-1);
}

/*JS*********************************************************************
*   Set the line number and offsets of TF_Lines[l1..l2]. Assume that
*    the values for l1 are already correct. Returns the highest index
*    where TF_LineOff changed. If flag is true, stop if values no longer
*    change.
*************************************************************************/

static int tframesetoffs(jedtext_t *tf, int l1, int l2, int flag)

/************************************************************************/
{
  int nr, last;
  long off;

  nr  = tf->TF_Lines[l1].TF1_LineNr;
  off = tf->TF_Lines[l1].TF1_LineOff;

  for(last = l1++ ; l1 <= l2 ; l1++) {
    long l;

    if((l = getlinelen(tf->TF_Text + off, tf->TF_Width)) == 0) break;
    off += l;

    if(tf->TF_Text[off - 1] == '\n') nr++;
    else if(tf->TF_Text[off] == '\0') break;

    if(tf->TF_Lines[l1].TF1_LineOff != off ||
       tf->TF_Lines[l1].TF1_LineNr != nr) {
      last = l1;
      tf->TF_Lines[l1].TF1_LineNr = nr;
      tf->TF_Lines[l1].TF1_LineOff = off;
    } else if(flag) {
      goto ende;
    }
  }

  for( ; l1 <= l2 ; l1++)
    if(tf->TF_Lines[l1].TF1_LineOff != off ||
       tf->TF_Lines[l1].TF1_LineNr != -1) {
      last = l1;
      tf->TF_Lines[l1].TF1_LineNr = -1;
      tf->TF_Lines[l1].TF1_LineOff = off;
    }

ende:
  return(last == tf->TF_Height ? last - 1 : last);
}

/*JS*********************************************************************
*   Recalculate TF_CuLine and TF_CuColumn from TF_CuOffset.
*************************************************************************/

static int tframeo2lc(jedtext_t *tf)

/************************************************************************/
{
  int i, j;
  long offset;

  offset = tf->TF_CuOffset;
  j = tf->TF_Height;
  for(i = 0 ; i < j ; i++)
    if(offset < tf->TF_Lines[i + 1].TF1_LineOff ||
       tf->TF_Lines[i + 1].TF1_LineNr == -1)
      break;

  tf->TF_CuLine = i;

  for(i = tf->TF_Lines[i].TF1_LineOff, j = 0 ; i < offset ; i++)
    if(tf->TF_Text[i] == '\t')
      j = (j + 8) & ~7;
    else
      j++;

  /*if(tf->TF_Text[i] == '\t') j |= 7;*/
  tf->TF_CuColumn = j;

  return 0;
}

/*JS*********************************************************************
*   Recalculate TF_CuOffset from TF_CuLine and TF_CuColumn. Adjust
*    TF_CuColumn (in case of '\t' characters).
*************************************************************************/

static void tframelc2o(jedtext_t *tf)

/************************************************************************/
{
  long offset;
  int col;

  while(tf->TF_CuLine > 0 && tf->TF_Lines[tf->TF_CuLine].TF1_LineNr == -1)
    tf->TF_CuLine--;

  offset = tf->TF_Lines[tf->TF_CuLine].TF1_LineOff;
  for(col = 0 ; col < tf->TF_CuColumn ; offset++) {
    int c;

    c = tf->TF_Text[offset];
    if(c == '\t') col = (col + 8) & ~7;
    else if(c && c != '\n') col++;
    else break;
  }

  /*  Because of TAB's, re-adjust column as well  */
  tf->TF_CuColumn = col;
  tf->TF_CuOffset = offset;
}

/*JS*********************************************************************
*   Return the index in the TF_Lines vector which contains the start of
*    the line. If the return value is zero, the start is in TF_StartOff.
*************************************************************************/

static int tframelstart(jedtext_t *tf, int line)

/************************************************************************/
{
  int nr;

  nr = tf->TF_Lines[line].TF1_LineNr;

  /*  Find offset for beginning of this line  */
  for( ; line > 0 ; line--)
    if(tf->TF_Lines[line - 1].TF1_LineNr != nr) break;

  /*  The screen line that contains the start of the current line  */
  return(line);
}

/*JS*********************************************************************
*   Get the number of lines of the text at s.
*************************************************************************/

static long getnlines(const char *s)

/************************************************************************/
{
  long n;

  for(n = 0 ; *s ; s++) if(*s == '\n') n++;

  return(n);
}

/*JS*********************************************************************
*   Get the line length at s, up the the maximum given width.
*************************************************************************/

static int getlinelen(const char *s, int width)

/************************************************************************/
{
  const char *s2;
  int col;

  for(s2 = s, col = 0 ; *s2 && col <= width ; s2++) {
    if(*s2 == '\n') {
      s2++;
      break;
    }

    if(*s2 == '\t') col = (col + 8) & ~7;
    else col++;
  }

  if(col > width) s2--;

  return(s2 - s);
}

/*JS*********************************************************************
*   Get the length of the line starting at s (includes the '\n' character)
*************************************************************************/

static int getfulllinelen(const char *s)

/************************************************************************/
{
  const char *s2;

  for(s2 = s ; *s2 ; s2++)
    if(*s2 == '\n') return(s2 - s + 1);

  return(s2 - s);
}

/*JS*********************************************************************
*   Find the start of the current line, given by offset.
*************************************************************************/

static int findlstart(const char *text, long offset)

/************************************************************************/
{
  while(offset > 0 && text[offset - 1] != '\n') offset--;

  return(offset);
}

/*JS*********************************************************************
*   Find the start of the next line, starting from offset. Return -1 if
*    this is the last line.
*************************************************************************/

static int findnextl(const char *text, long offset)

/************************************************************************/
{
  for( ; text[offset] ; offset++)
    if(text[offset] == '\n') return(offset + 1);

  return(-1);
}

#ifdef DEBUG
/*JS*********************************************************************
*   Find the start of the next line, starting from offset. Return -1 if
*    this is the last line.
*************************************************************************/

static void checkframe(jedtext_t *tf, int key)

/************************************************************************/
{
  const char *s;
  long i, nr, off;

  s = tf->TF_Text;
  i = strlen(s);
  if(i != tf->TF_Length) {
    dbgprintf("ERROR TextFrame: structure length %ld != real length %ld\n",
	      tf->TF_Length, i);
    tf->TF_Length = i;
  }

  if((tf->TF_Mode & TFRAME_ALLOCATE) && i > tf->TF_MaxLen)
    dbgprintf("ERROR TextFrame: max len %ld < len %ld\n", tf->TF_MaxLen, i);

  i = getnlines(s);
  if(i != tf->TF_NLines) {
    dbgprintf("ERROR TextFrame: structure nr lines %ld != real %ld\n",
	      tf->TF_NLines, i);
    tf->TF_NLines = i;
  }

  if(tf->TF_CuLine < 0 || tf->TF_CuLine >= tf->TF_Height ||
     tf->TF_CuColumn < 0 || tf->TF_CuColumn >= tf->TF_Width) {
    dbgprintf("ERROR TextFrame: Cursor position %d,%d out of bounds.\n",
	      tf->TF_CuLine, tf->TF_CuColumn);
  }

  s = tf->TF_Text;
  off = tf->TF_StartOff;
  if(off < 0 || off > tf->TF_Length || (off > 0 && s[off - 1] != '\n'))
    dbgprintf("ERROR TextFrame: Illegal start offset %ld.\n", off);

  for(nr = 0, i = 0 ; i < off ; i++) if(s[i] == '\n') nr++;

  if(tf->TF_Lines[0].TF1_LineNr != nr)
    dbgprintf("ERROR TextFrame: First line number %d not %d.\n",
	      tf->TF_Lines[0].TF1_LineNr, nr);

  for(i = 0 ; off <= tf->TF_Length && i <= tf->TF_Height ; ) {
    if(tf->TF_Lines[i].TF1_LineOff == off) {
      if(tf->TF_Lines[i].TF1_LineNr != nr &&
	 !(tf->TF_Lines[i].TF1_LineNr == -1 && off == tf->TF_Length))
	dbgprintf("ERROR TextFrame: Line %d number %d != %d\n", i,
		  tf->TF_Lines[i].TF1_LineNr, nr);
      i++;
      continue;
    }
    if(tf->TF_Lines[i].TF1_LineOff < off)
      dbgprintf("ERROR TextFrame: Line %d offset %d < %d.\n", i,
		tf->TF_Lines[i].TF1_LineOff, off);

    if(s[off] == '\n') nr++;
    off++;
  }

  if(i <= tf->TF_Height)
    dbgprintf("ERROR TextFrame: Line %d (nr %d off %d) not in text.\n",
	      i, tf->TF_Lines[i].TF1_LineNr, tf->TF_Lines[i].TF1_LineOff);
}
#endif

#ifdef TEST
/* ***  A little test program for text frames  *** */
#include <unistd.h>

#define LINE    2
#define COLUMN  1
#define HEIGHT  24
#define WIDTH   72

#define DO_SAVE    KEY_F(1)

int main(int argc, char *argv[])
{
  char *name, *text;
  jedtext_t *tf;
  long len;
  int key;

  if(argc != 2) {
    fprintf(stderr, "Usage: testframe <file>\n");
    return(5);
  }

  name = argv[1];

  if(access(name, F_OK) == 0) {
    if((text = file2buffer(name, &len)) == NULL) EGOTO(error);
  } else {
    len = 1;
    if((text = malloc(len)) == NULL) EGOTO(error);
    text[0] = '\0';
  }

  jedinitialize(0);
  JEditTFFillChars = ' ';

  if((tf = jedtextcreate(text, len, LINE, COLUMN, HEIGHT, WIDTH, TFRAME_SCROLLBAR|
			TFRAME_ALLOCATE|TFRAME_INSMODE)) == NULL)
    EGOTO(error);

  _jedprintframe(stdscr, LINE-1, COLUMN-1, HEIGHT+2, WIDTH+3, 0);
  mvprintw(0, 14, "Press F1 to save text, ESC to quit without saving");

  for(key = EDKEY_REDRAW ; ; ) {
    int ret;

    if((ret = jedtextdokey(stdscr, tf, jeddokeytrans(key, NULL), 0))
       == -1)
      EGOTO(error);

    if(ret == -2 || ret == -3) {
      beep();
      refresh();
    }
    if((key = wgetch(stdscr)) == EDKEY_QUIT || key == DO_SAVE) break;
  }

  jeddeinitialize();

  if(key == DO_SAVE) {
    FILE *fp;

    if(mkBackupFile(name, 1) == NULL || (fp = fopen(name, "w")) == NULL ||
       fwrite(tf->TF_Text, 1, tf->TF_Length, fp) != tf->TF_Length ||
       fclose(fp) == EOF)
      EGOTO(error);

    printf("Saved file.\n");
  } else
    printf("Not saved.\n");

  return(0);
error:
  perror("testframe");
  return(30);
}
#endif
