/*JS*********************************************************************
*
*    Program : JEDHIST
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Remember input values
*
*              Copyright (C) 2006 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <jsconfig.h>
#include <jssubs.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS */
#define JEDLIB_HASHISTORY
/*  END-DEFINITIONS  */

#define JED_HISTORYFILE      ".jedhistory"

typedef struct {
  const char *JEHI_FieldName;

  const char **JEHI_Values;
  int JEHI_NValues, JEHI_MValues;
} jedhistentry_t;

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
IPrototype int           _jedhistload(void);
static int               mergeEntries(FILE *fp, const char *name);
IPrototype int           _jedhistdump(void);
static int               dumpEntries(FILE *fp);

static const char        *getHistFile(void);

IPrototype const char  **_jedhistgetvalues(const char *fldname,
					   const char *content,
					   int *pNEntries);
IPrototype int           _jedhistremember(const char *fldname,
					  const char *content,
					  int len);

static jedhistentry_t   *gethistentry(const char *fldname, int create);
static int               addcontent(jedhistentry_t *entry, const char *content,
				    int len);

Prototype char           JEDHistFile[];

/*********     GLOBAL/STATIC VARIABLES                          *********/
char JEDHistFile[JS_FILENAME_MAX];

static jedhistentry_t *JEDHistory;
static int JEDNHistory, JEDMHistory;

/*JS*********************************************************************
*
*************************************************************************/

int _jedhistload(void)

/************************************************************************/
{
  FILE *fp;
  const char *name;

  name = getHistFile();

  if(fileexists(name) == 0)
    return 0;

  if((fp = fopenLock(name, "r", JSLOCK_READW)) == NULL ||
     mergeEntries(fp, name) == -1)
    EGOTO(error);

  if(fclose(fp) == EOF)
    EGOTO(error0);

  return 0;
error:
  if(fp) fclose(fp);
error0:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int mergeEntries(FILE *fp, const char *name)

/************************************************************************/
{
  char *line;
  int linenr;

  for(linenr = 1 ; (line = jsgetline(fp, GETLINE_RMEOL |
				     GETLINE_COMMENT)) != NULL ; ++linenr) {
    jedhistentry_t *entry;
    char *s;

    for(s = line ; *s ; ++s)
      if(*s == ':')
	break;

    if(*s != ':') EGOTO(invalid);

    *s++ = '\0';

    if((entry = gethistentry(line, 1)) == NULL)
      EGOTO(error);

    while(*s) {
      char *string;
      int l;

      while(isspace(*s)) ++s;

      if(*s++ != '"')
	EGOTO(invalid);

      for(string = s ; *s ; ++s)
	if(*s == '\\' && s[1])
	  ++s;
	else if(*s == '"')
	  break;

      if(*s != '"')
	EGOTO(invalid);

      *s++ = '\0';

      l = cStrCompile(string, string, 0);
      string[l] = '\0';

#if 0
      printf("GOT '%s'\n", string);
#endif

      if(addcontent(entry, string, strlen(string)) == -1)
	EGOTO(error);
    }

    continue;

invalid:
    fprintf(stderr, "WARNING '%s', l. %d: Ignored invalid line '%s'!\n",
	    name, linenr, line);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedhistdump(void)

/************************************************************************/
{
  FILE *fp;
  const char *name;

  name = getHistFile();

  if(fileexists(name) == 0) {
    if((fp = fopenLock(name, "w", JSLOCK_WRITE)) == NULL)
      EGOTO(error);

  } else {
    if((fp = fopenLock(name, "r+", JSLOCK_WRITE)) == NULL)
      EGOTO(error);

    /*  File might have changed meanwhile, first merge, then dump  */
    if(mergeEntries(fp, name) == -1)
      EGOTO(error);

    rewind(fp);
  }

  if(dumpEntries(fp) == -1)
    EGOTO(error);

  if(fclose(fp) == EOF)
    EGOTO(error0);

  return 0;
error:
  if(fp) fclose(fp);
error0:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int dumpEntries(FILE *fp)

/************************************************************************/
{
static const char InEsc[] = "\"\'\n\t\r\f\v\b\a\033\\";
static const char OutEsc[]  = "\"\'ntrfvbae\\";
  int i;

  for(i = 0 ; i < JEDNHistory ; ++i) {
    jedhistentry_t *entry;
    int j;

    entry = &JEDHistory[i];
    if(fprintf(fp, "%s:", entry->JEHI_FieldName) == -1)
      EGOTO(error);

    for(j = 0 ; j < entry->JEHI_NValues ; ++j) {
      const char *string;

      if(fprintf(fp, " \"") == EOF)
	EGOTO(error);

      for(string = entry->JEHI_Values[j] ; *string ; ++string) {
	const char *t;
	char c;

	c = *string;
	if((t = strchr(InEsc, c)) != NULL) {
	  if(fputc('\\', fp) == EOF)
	    EGOTO(error);
	  c = OutEsc[t - InEsc];
	}

	if(fputc(c, fp) == EOF)
	  EGOTO(error);
      }

      if(fputc('"', fp) == EOF)
	EGOTO(error);
    }

    if(fprintf(fp, "\n") == -1)
      EGOTO(error);
  }


  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static const char *getHistFile(void)

/************************************************************************/
{
  if(JEDHistFile[0] == '\0') {
    const char *home, *name;

    name = JED_HISTORYFILE;

    if((home = getenv("HOME")) != NULL)
      sprintf(JEDHistFile, "%s/%s", home, name);
    else
      strcpy(JEDHistFile, name);
  }

  return JEDHistFile;
}

/*JS*********************************************************************
*
*************************************************************************/

const char **_jedhistgetvalues(const char *fldname, const char *content,
			       int *pNEntries)

/************************************************************************/
{
  static const char **Values;
  static int MValues;
  jedhistentry_t *entry;
  int i, cnt, n;

  if((entry = gethistentry(fldname, 0)) != NULL)
    n = entry->JEHI_NValues;
  else
    n = 0;

  if(n >= MValues) {
    MValues = n + 1; /*  treat n==0 properly  */
    if((Values = (const char **)
	realloc(Values, MValues * sizeof(*Values))) == NULL)
      return NULL;
  }

  cnt = 0;
  for(i = 0 ; i < n ; ++i)
    if(jstrstartswith(entry->JEHI_Values[i], content) != NULL) {
      const char *value;
      int k;

      /*  Keep sorted. We use insertion sort, assuming that the
       *   number of entries is rather small
       */
      value = entry->JEHI_Values[i];
      for(k = cnt ; k > 0 ; --k) {
	if(strcmp(Values[k - 1], value) < 0)
	  break;

	Values[k] = Values[k - 1];
      }
      Values[k] = value;

      ++cnt;
    }

  *pNEntries = cnt;

  return Values;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedhistremember(const char *fldname, const char *content,
		     int len)

/************************************************************************/
{
  jedhistentry_t *entry;

  if((entry = gethistentry(fldname, 1)) == NULL)
    return -1;

  return addcontent(entry, content, len);
}

/*JS*********************************************************************
*
*************************************************************************/

static jedhistentry_t *gethistentry(const char *fldname, int create)

/************************************************************************/
{
  jedhistentry_t *entry;
  int i;

  for(i = 0 ; i < JEDNHistory ; ++i)
    if(strcmp(JEDHistory[i].JEHI_FieldName, fldname) == 0)
      return &JEDHistory[i];

  if(!create)
    return NULL;

  if(JEDNHistory >= JEDMHistory) {
    JEDMHistory += 20;
    if((JEDHistory = (jedhistentry_t *)
	realloc(JEDHistory, JEDMHistory * sizeof(*JEDHistory))) == NULL)
      return NULL;
  }

  entry = &JEDHistory[JEDNHistory];

  if((entry->JEHI_FieldName = strdup(fldname)) == NULL)
    return NULL;

  entry->JEHI_Values = NULL;
  entry->JEHI_MValues = entry->JEHI_NValues = 0;

  JEDNHistory++;

  return entry;
}

/*JS*********************************************************************
*
*************************************************************************/

static int addcontent(jedhistentry_t *entry, const char *content, int len)

/************************************************************************/
{
  char *dest;
  int i;

  for(i = 0 ; i < entry->JEHI_NValues ; ++i)
    /*  If the entry already appears exactly, skip adding it  */
    if(strncmp(entry->JEHI_Values[i], content, len) == 0 &&
       entry->JEHI_Values[i][len] == '\0')
      return 0;

  if(entry->JEHI_NValues >= entry->JEHI_MValues) {
    entry->JEHI_MValues += 20;
    if((entry->JEHI_Values = (const char **)
	realloc(entry->JEHI_Values, entry->JEHI_MValues *
		sizeof(*(entry->JEHI_Values)))) == NULL)
      EGOTO(error);
  }

  if((dest = (char *)malloc(len + 1)) == NULL)
    EGOTO(error);

  memcpy(dest, content, len);
  dest[len] = '\0';

  entry->JEHI_Values[i] = dest;

  entry->JEHI_NValues++;

  return i;
error:
  return -1;
}

#ifdef TEST
int main(int argc, char *argv[])
{
  _jedhistload();
  dumpEntries(stdout);

  return 0;
}
#endif
