/*JS*********************************************************************
*
*    Program : JEDWSTACK
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose :
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

#define JEDLIB_INTERNAL
#include <jedlib.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
typedef struct {
  WINDOW *Window;
  long Flag;
} jedwinstack_t;

/*  Check if two windows intersect  */
#define WININTERSECT(x1l, y1l, x1r, y1r, x2l, y2l, x2r, y2r) \
	(OPEN_INTERVAL_INTERSECT(x1l, x1r, x2l, x2r) && \
	 OPEN_INTERVAL_INTERSECT(y1l, y1r, y2l, y2r))

#define GETWINCOORD(win, xl, yl, xr, yr)  do { \
					getbegyx(win, (xl), (yl)); \
					getmaxyx(win, (xr), (yr)); \
					(xr) += (xl); (yr) += (yl); \
				      } while(0)

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
Prototype WINDOW        *jedwstacknew(int nlines, int ncols, int y, int x);
Prototype int            jedwstackdel(WINDOW *win);

IPrototype int           _jedwstackpush(WINDOW *win, int pos);
IPrototype int           _jedwstackpop(WINDOW *win);
IPrototype int           _jedwstackflag(WINDOW *win, int flag);

Prototype int            jedwstackrefresh(WINDOW *win);
Prototype int            wnoutrefresh_s(WINDOW *win);
Prototype int            jedwstacktouch(WINDOW *win);

Prototype int            jedwstackdoupdate(int flag);
static void              _jedwstacktouch(int ind);

Prototype int            jedwstackgetch(WINDOW *win);

/*********     GLOBAL/STATIC VARIABLES                          *********/
static jedwinstack_t *WindowStack;
static int NWindowStack, MaxWindowStack;

/*JS*********************************************************************
*
*************************************************************************/

WINDOW *jedwstacknew(int nlines, int ncols, int y, int x)

/************************************************************************/
{
  WINDOW *win;

  if((win = newwin(nlines, ncols, y, x)) && _jedwstackpush(win, -1) == -1) {
    delwin(win);
    win = NULL;
  }

  return win;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedwstackdel(WINDOW *win)

/************************************************************************/
{
  int ret;

  /*  Make sure the whole stack of windows gets repainted  */
  WindowStack[0].Flag |= 1;
  ret = delwin(win);
  if(_jedwstackpop(win) == -1)
    return ERR;

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedwstackpush(WINDOW *win, int pos)

/************************************************************************/
{
  if(pos < 0 && (pos += 1 + NWindowStack) < 0) pos = 0;

  if(NWindowStack >= MaxWindowStack) {
    jedwinstack_t *nStack;

    if((nStack = (jedwinstack_t *)
	realloc(WindowStack, (MaxWindowStack + 10) * sizeof(*nStack))) /*DBG*/
       == NULL)
      return -1;

    WindowStack = nStack;
    MaxWindowStack += 10;
  }

  if(pos < NWindowStack)
    memmove(&WindowStack[pos + 1], &WindowStack[pos],
	    (NWindowStack - pos) * sizeof(*WindowStack));

  WindowStack[pos].Window = win;
  WindowStack[pos].Flag = 0;
  NWindowStack++;

  return pos;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedwstackpop(WINDOW *win)

/************************************************************************/
{
  int pos;

  if(win) {
    for(pos = 0 ; pos < NWindowStack ; pos++)
      if(WindowStack[pos].Window == win) break;

    if(pos == NWindowStack) EGOTO(error);
  } else
    pos = NWindowStack - 1;

  NWindowStack--;
  if(pos < NWindowStack)
    memmove(&WindowStack[pos], &WindowStack[pos + 1],
	    (NWindowStack - pos) * sizeof(*WindowStack));

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int _jedwstackflag(WINDOW *win, int flag)

/************************************************************************/
{
  int pos;

  for(pos = 0 ; pos < NWindowStack ; pos++)
    if(WindowStack[pos].Window == win) {
      if((flag & (flag - 1)) == 0)
	WindowStack[pos].Flag |= flag;
      else
	WindowStack[pos].Flag &= ~flag;

      return 0;
    }

  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedwstackrefresh(WINDOW *win)

/************************************************************************/
{
  if(wnoutrefresh_s(win) == ERR)
    return wrefresh(win);

  return jedwstackdoupdate(1);
}

/*JS*********************************************************************
*
*************************************************************************/

int wnoutrefresh_s(WINDOW *win)

/************************************************************************/
{
  int i;

  for(i = NWindowStack - 1 ; i >= 0 ; i--)
    if(win == WindowStack[i].Window)
      break;

  if(i < 0)
    return ERR;

  WindowStack[i].Flag |= 2;
  return OK;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedwstacktouch(WINDOW *win)

/************************************************************************/
{
  int i;

  for(i = NWindowStack - 1 ; i >= 0 ; i--)
    if(win == WindowStack[i].Window)
      break;

  if(i < 0) return ERR;

  WindowStack[i].Flag |= 1;
  return OK;
}

/*JS*********************************************************************
*
*************************************************************************/

int jedwstackdoupdate(int flag)

/************************************************************************/
{
  int i;

  _jedwstacktouch(0);
  for(i = 0 ; i < NWindowStack ; i++) {
    int fl;

    fl = WindowStack[i].Flag;
    WindowStack[i].Flag = 0;
    if(fl & 4)
      continue;

    if(((fl & 1) && touchwin(WindowStack[i].Window) == ERR) ||
       ((fl & 3) && wnoutrefresh(WindowStack[i].Window) == ERR))
      return ERR;
  }

  if(flag) return doupdate();

  return OK;
}

/*JS*********************************************************************
*
*************************************************************************/

static void _jedwstacktouch(int ind)

/************************************************************************/
{
  for(; ind < NWindowStack ; ind++) {
    WINDOW *win;
    int j, x1l, y1l, x1r, y1r;

    if(WindowStack[ind].Flag & 4) continue;

    if(is_wintouched(WindowStack[ind].Window))
      WindowStack[ind].Flag |= 2;
    else if(WindowStack[ind].Flag == 0)
      continue;

    /*  Get upper left and lower right corner of window  */
    win = WindowStack[ind].Window;
    GETWINCOORD(win, x1l, y1l, x1r, y1r);

    for(j = ind + 1 ; j < NWindowStack ; j++) {
      int x2l, y2l, x2r, y2r;

      win = WindowStack[j].Window;
      GETWINCOORD(win, x2l, y2l, x2r, y2r);
      if(WININTERSECT(x1l, y1l, x1r, y1r, x2l, y2l, x2r, y2r))
	WindowStack[j].Flag |= 1;
    }
  }
}

/*JS*********************************************************************
*
*************************************************************************/

int jedwstackgetch(WINDOW *win)

/************************************************************************/
{
  int key, i;

  if(jedwstackrefresh(win) == ERR)
    return ERR;

  for(i = 0 ; ; i++) {
    if((key = wgetch(win)) == '\r'
#ifdef KEY_ENTER
       || key == KEY_ENTER
#endif
       )
      key = '\n';

    if(i)
      return -key;

    if(key != CNTRL('V'))
      return key;
  }

  /*NOTREACHED*/
}
