/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Main program and functions that only need the
 *              preprocessor.
 *
 *              Copyright (C) 2003-2018 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

static const char chksumid[] = "$CK: AAAAAA $";

/*********     INCLUDES                                         *********/
#include "jscpp.h"

/*********     DEFINES                                          *********/
#ifndef INDEX_MAX
/*  START-DEFINITIONS  */
/* ***  Close streams  *** */
#define closePreProcStream(jps, flag)     jsppclose(jps, flag)
#define closeParserStream(jpa, flag)      jsparsclose(jpa, flag)

#define EXTFLAG_DOUBLEASSIGN              (JPAM_FUSERFLAGS << 0)
#define EXTFLAG_INSERTCAST                (JPAM_FUSERFLAGS << 1)
#define EXTFLAG_INSERTCAST2               (JPAM_FUSERFLAGS << 2)

/*  END-DEFINITIONS  */
#endif

#define JSCPP_MODE_PREPROC      "cpp"
#define JSCPP_MODE_OPREPROC     "opp"

#define JSCPP_NAME_IDEPS        "jsideps"
#define JSCPP_MODE_IDEPS        "ideps"

#ifdef CONFIG_JSCPPLIB_DEBUG
#define OPTSTRING_ADDDBG  "X:"
#else
#define OPTSTRING_ADDDBG  ""
#endif

#define OPTSTRING         JSPARSE_OPTSTRING "hqm:o:d:pLl;jtvasf" OPTSTRING_ADDDBG

#define PRGMODE_PREPROC      1
#define PRGMODE_IDEPS        2
#define PRGMODE_FDEPS        3
#define PRGMODE_OPREPROC     4
#define PRGMODE_PPUNUSED     5

#define PRGMODE_GLOBVARS    10
#define PRGMODE_FCTSEP      11
#define PRGMODE_PROTO       12
#define PRGMODE_FCTTOC      13

#define PRGMODE_FTRACE      20
#define PRGMODE_TRACE       21
#define PRGMODE_STMTSTEP    22
#define PRGMODE_STMTSTEP0   23
#define PRGMODE_CODECOVER   24
#define PRGMODE_FCTWRAP     25

#define PRGMODE_PTRDEPS     30
#define PRGMODE_PTRDEPS0    31
#define PRGMODE_PTRDEPSF    32

#define DEFAULT_INDENTATION        2

#ifdef CONFIG_JSCPPLIB_DEBUG
static const dbgopts_t DebugOptions[] = {
#if 0
    { "showpragma",      1, NULL,
      "In mode cpp: Show preprocessing directly\n\
In mode ftrace: Show internal representation of pragma lines\n\
In mode fctsep:\n" },
    { "",      2, NULL,
      "" },
    { "",      4, NULL,
      "In mode ftrace: " },
#endif

    { "showsourcemac",     0x08, NULL,
      "Show macro definitions from the source file itself" },
    { "showusermac",       0x18, NULL,
      "Show macro definitions from source and user header files" },
    { "showfilemac",       0x38, NULL,
      "Show macro definitions from source, user, and system header" },
    { "showallmac",        0x78, NULL,
      "Show all macro definitions" },
    { "showassert",        0x80, NULL,
      "Show defined assertions" },

    { "oncefiles",        1 << 8, NULL,
      "Print files recognized as include once" },

    { NULL,          0,   (void *)&JSParseDebugOptions, NULL }
};
#endif

static const dbgopts_t ExtensionOptions[] = {
    { "doubleassign", EXTFLAG_DOUBLEASSIGN, NULL,
      "Patch assignments in declarations like\n\
\tint i = 0;\n\
as\n\
\tint i = (<tracecall>, i = 0);\n\
instead of the shorter form\n\
\tint i = (<tracecall>, 0);\n\
This works around some deficient compilers." },

    { "insertcast", EXTFLAG_INSERTCAST, NULL,
      "Patch pointer assignments in declarations like\n\
\tint *i = 0;\n\
with an additional cast as\n\
\tint *i = (<tracecall>, (int *)0);\n\
This works around some deficient compilers." },

    { "insertcast2", EXTFLAG_INSERTCAST2, NULL,
      "Patch all assignments in declarations like\n\
\tint i = 0;\n\
with an additional cast as\n\
\tint i = (<tracecall>, (int)0);\n" },

    { NULL,          0, NULL,  NULL }
};

#define Prototype extern
/*********     PROTOTYPES                                       *********/
static void              topline(void);
static void              help(void);

Prototype FILE          *openFile(const char *directory, const char *name);

Prototype jstrbuff_t     TempBuffer;

/*********     GLOBAL / STATIC VARIABLES                        *********/
jstrbuff_t TempBuffer;

/*JS*********************************************************************
 *
 *************************************************************************/

static void topline(void)

/************************************************************************/
{
    fprintf(stderr, "\
JSCPP C-Preprocessor and Parser V%ld.%ld.%ld             (C) Joerg Schoen 1996-2018\n",
	    JSCPPNVersion / 10000, (JSCPPNVersion / 100) % 100,
	    JSCPPNVersion % 100);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static void help(void)

/************************************************************************/
{
    topline();

    fprintf(stderr, "\
Usage: jscpp [<opts>] <infile> [...]\n\
  Default mode is to preprocess <infile> writing to stdout.\n\
  Other modes may be chosen by option '-m'.\n\
  General options:\n\
    -h        Show this help information.\n\
    -o<outfile> Write to <outfile> instead of stdout.\n\
    -d<outdir> Write each input file to a file with the same name in\n\
              the directory. Mutually exclusive with option '-o'.\n");

    fprintf(stderr,"\
    -m<mode>  Choose different main modes:\n\
              '"JSCPP_MODE_PREPROC
	    "':   Preprocess the file. This is the default mode.\n\
              '"JSCPP_MODE_IDEPS
	    "': Print lines in make style that show the dependency of\n\
                       each input file on include files.\n\
                       This mode is the default when the executable is\n\
                       named '" JSCPP_NAME_IDEPS "'.\n");

    fprintf(stderr, "              '"JSCPP_MODE_FDEPS
	    "': Print for each function and global/static variables the\n\
                       functions it calls and the global/static variables\n\
                       used. Result looks like:\n\
                         <function_name>: <called_function> ...\n\
                       Option '-f' only considers functions, not variables.\n\
                       Option '-p' adds to static functions and variables the\n\
                       postfix '@<sourcefile>' to make the names globally unique.\n\
                       If option '-p' is given twice, variable names are\n\
                       prefixed with '$' to distinguish them from functions.\n\
                       Different output style may be chosen with option '-j'\n\
                       or '-t'.\n\
                       This mode is the default when the executable is\n\
                       named '"JSCPP_NAME_FDEPS"'.\n");
    fprintf(stderr, "              '" JSCPP_MODE_OPREPROC
	    "':   Preprocess the file, expanding only macros from 'expandorig'\n\
                       or 'expandorigfull' pragma lines.\n");

    fprintf(stderr, "              '" JSCPP_MODE_FTRACE
	    "': Trace call and return of all functions.\n\
                       This mode is the default when the executable is\n\
                       named '"JSCPP_NAME_FTRACE"'.\n\
              '" JSCPP_MODE_TRACE
	    "': Same as '"JSCPP_MODE_FTRACE "' but nothing is traced by default.\n\
                       This mode is the default when the executable is\n\
                       named '"JSCPP_NAME_TRACE"'.\n");

    fprintf(stderr, "              '" JSCPP_MODE_FCTWRAP
	    "': Trace call and return of all functions by generating wrapper\n\
                       functions. This also works for functions without definition,\n\
                       i. e. library functions.\n\
                       This mode is the default when the executable is\n\
                       named '"JSCPP_NAME_FCTWRAP"'.\n");

    fprintf(stderr, "              '"JSCPP_MODE_PPUNUSED
	    "': Find preprocessor macros that are never used in any\n\
                       of the given input files.\n");

    fprintf(stderr, "              '"JSCPP_MODE_GLOBVARS
	    "': Determine which global variables are used in at\n\
                       least one input file.\n");

    fprintf(stderr, "              '"JSCPP_MODE_PROTO
	    "': Print prototype for all functions defined in the sources.\n\
                       Option '-a' prints also static prototypes and option '-j'\n\
                       enables additional warnings (see below).\n");

    fprintf(stderr, "              '"JSCPP_MODE_FCTTOC
	    "': Print table of contents for all functions defined in the\n\
                       sources. Each line consists of the source file name, the\n\
                       start and end line of the function definition, the function\n\
                       name, and the prototype of the function with the name\n\
                       substituted by a '@'.\n\
                       Option '-a' includes also static functions and option '-j'\n\
                       enables additional warnings (see below).\n");

    fprintf(stderr, "              '"JSCPP_MODE_PTRDEPS
	    "' or '" JSCPP_MODE_PTRDEPS "0' or '" JSCPP_MODE_PTRDEPS "f':\n\
                       Do some (limited) flow analysis to detect function\n\
                       parameters that can have 'const' modifiers.\n\
                       The modified source is written as output. Option '-v'\n\
                       shows on stdout which function parameters were\n\
                       modified.\n\
                       Option '-j' prints the output in a numeric format.\n\
                       The second mode only adds 'const' to the argument\n\
                       specifiers, the third mode also checks if the\n\
                       parameter itself is never modified.\n");

    fprintf(stderr, "              '"JSCPP_MODE_STMTSTEP
	    "' or '" JSCPP_MODE_STMTSTEP0
	    "': Trace each statement as executed.\n\
                       The second form only generates ONE trace statement\n\
                       for statements that execute in a row without a branch.\n\
                       This reduces the amount of traces produced.\n");

    fprintf(stderr, "              '"JSCPP_MODE_FCTSEP
	    "': Separate blocks from function if preceded by lines\n\
                           #pragma separate FUNCNAME\n\
                       See program docu for more details.\n");

    fprintf(stderr, "\
    -L        Print 'line' directive at the start of the output file\n\
              to recognize it uniquely.\n\
    -l[<indent>] Format trace statements in multiple lines. Only useful\n\
              if output is to be kept. Formatting inserts additional lines\n\
              in instrumented code, so current line information differs.\n\
              <indent> defaults to %d.\n", DEFAULT_INDENTATION);

    fprintf(stderr, "  Standard options:\n%s%s",
	    JSPPStdOptsHelp,
	    JSParsStdOptsHelp);

    fprintf(stderr, "\
  Other options:\n\
    -p        In mode '" JSCPP_MODE_FTRACE "' or '"
	    JSCPP_MODE_TRACE
	    "', yield the partially preprocessed\n\
              source instead of the original one. \"Partially\" means that\n\
              only user includes are expanded.\n\
              If given twice, also user macros (i. e. defined in the source\n\
              itself or in user includes are expanded.\n\
              This would not work if user macros appear as arguments to\n\
              system macros. See the docs for more details.\n\
              If given thrice, normal preprocessing is done.\n\
              In mode '" JSCPP_MODE_FDEPS
	    "', prefix static names with the source file they\n\
              appear in. If given twice, prefix variables with '$'.\n\
              to distinguish them from functions.\n\
    -j        In mode '"JSCPP_MODE_FDEPS"', print output in style suitable for\n\
              'jsmakedeps' utility. Unless option '-p' is also given, static\n\
              functions and variables are not printed, though they are scanned\n\
              for usage of global ones.\n\
              In mode '"JSCPP_MODE_PTRDEPS"', print argument position instead\n\
              of argument name.\n\
              In mode '"JSCPP_MODE_FCTTOC"' and '"JSCPP_MODE_PROTO
	    "' and if prototype warnings are\n\
              enabled ('-Wproto'), check that the prototypes found in the\n\
              source's accompanying header (i. e. \"file.c\" -> \"file.h\")\n\
              matches the functions defined in the source. This can be used\n\
              to check that prototypes appear in the proper header file.\n\
    -f        In mode '" JSCPP_MODE_FDEPS
	    "' only show dependencies on/of functions.\n\
    -t        In mode '"JSCPP_MODE_FDEPS"', print output suitable for topological\n\
              sorting, i. e. only one pair of identifiers per line.\n");

    fprintf(stderr, "    -a        In mode '" JSCPP_MODE_PPUNUSED
	    "', also report unused identifiers\n\
              from user include files. If given twice, also from system\n\
              include files.\n");
    fprintf(stderr, "              In mode '"JSCPP_MODE_FDEPS
	    "', do not remove duplicate function calls and\n\
              keep the sequence in which they were called.\n");
    fprintf(stderr, "              In mode '"JSCPP_MODE_FCTTOC"' and '"JSCPP_MODE_PROTO
	    "' also print prototypes for static\n\
              functions.\n");

    fprintf(stderr, "    -s        In mode '" JSCPP_MODE_FDEPS
	    "', also show dependencies on symbols from the\n\
              standard library like 'printf' etc.\n");
}

/*JS*********************************************************************
 *
 *************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
    jsparsenviron_t ppenv;
    const char *outName;
    char *string;
    FILE *fpOut = NULL;
    int c, prgMode, ppMode, traceMode, depsMode;
    int unusedMode, ptrDepMode, tocMode;

    /* **  Preset mode variables  ** */
    ppMode = 0;
    traceMode = JSCPP_TRACE_PROTOTYPEFCT |
	JSCPP_TRACE_PROTOTYPEVAR | JSCPP_TRACE_PROTOTYPESTMT |
	JSCPP_TRACE_PROTOTYPECOV | JSCPP_TRACE_INTTYPEMAPPING;
    depsMode = JSCPP_FDEPS_STATICDEF | JSCPP_FDEPS_VARDEPS;
    unusedMode = 0;
    ptrDepMode = 0;
    tocMode = 0;

    outName = NULL;

    jsparsenvclear(&ppenv);

    /* **  Determine program mode from executable name ** */
    if((string = strrchr(argv[0], CONFIG_DIR_DELIMITER)))
	++string;
    else
	string = argv[0];

    if(strcmp(string, JSCPP_NAME_IDEPS) == 0)
	prgMode = PRGMODE_IDEPS;
    else if(strcmp(string, JSCPP_NAME_FDEPS) == 0)
	prgMode = PRGMODE_FDEPS;
    else if(strcmp(string, JSCPP_NAME_FTRACE) == 0)
	prgMode = PRGMODE_FTRACE;
    else if(strcmp(string, JSCPP_NAME_TRACE) == 0)
	prgMode = PRGMODE_TRACE;
    else if(strcmp(string, JSCPP_NAME_FCTWRAP) == 0)
	prgMode = PRGMODE_FCTWRAP;
    else
	prgMode = PRGMODE_PREPROC;

    if(jsparsinit(0) == -1)
	EGOTO(error);

    while((c = jsGetOpt(&argc, argv, OPTSTRING, (const char **)&string))
	  != EOF) {
	switch(c) {
	case '?':
	    EGOTO(error);

	case 'e': /*  Switch on extensions  */
	    if(string != NULL &&
	       parsedbgopts(ExtensionOptions, string,
			    (long *)&(ppenv.JPAE_PAMode), 7) == -1)
		EGOTO(error);

	    /*FALLTHROUGH*/

	case 'D':
	    if(isdigit(*string)) {
		Debug = strtol(string, NULL, 0);
		break;
	    }

	default:
	    {
		int ret;

		/*  Generic option handling for preprocessor  */
		if((ret = jsparsenvhandle(&ppenv, c, string)) == -1)
		    EGOTO(error);

		if(ret != -2)
		    /*  OK, option handled  */
		    break;
	    }
	    /*FALLTHROUGH*/

	case 'h':
	    help();
	    return 5;

	case 'o':
	    outName = string;
	    ppMode &= ~1;
	    traceMode &= ~JSCPP_TRACE_DIRECTORY;
	    break;

	case 'd':
	    outName = string;
	    ppMode |= 1;
	    traceMode |= JSCPP_TRACE_DIRECTORY;
	    break;

	case 'm': /*  Set program mode  */
	    if(strcmp(string, JSCPP_MODE_PREPROC) == 0)
		prgMode = PRGMODE_PREPROC;
	    else if(strcmp(string, JSCPP_MODE_IDEPS) == 0)
		prgMode = PRGMODE_IDEPS;
	    else if(strcmp(string, JSCPP_MODE_PPUNUSED) == 0)
		prgMode = PRGMODE_PPUNUSED;
	    else if(strcmp(string, JSCPP_MODE_FDEPS) == 0)
		prgMode = PRGMODE_FDEPS;
	    else if(strcmp(string, JSCPP_MODE_OPREPROC) == 0)
		prgMode = PRGMODE_OPREPROC;
	    else if(strcmp(string, JSCPP_MODE_FTRACE) == 0)
		prgMode = PRGMODE_FTRACE;
	    else if(strcmp(string, JSCPP_MODE_TRACE) == 0)
		prgMode = PRGMODE_TRACE;
	    else if(strcmp(string, JSCPP_MODE_FCTWRAP) == 0)
		prgMode = PRGMODE_FCTWRAP;
	    else if(strcmp(string, JSCPP_MODE_GLOBVARS) == 0)
		prgMode = PRGMODE_GLOBVARS;
	    else if(strcmp(string, JSCPP_MODE_PROTO) == 0)
		prgMode = PRGMODE_PROTO;
	    else if(strcmp(string, JSCPP_MODE_FCTTOC) == 0)
		prgMode = PRGMODE_FCTTOC;
	    else if(strcmp(string, JSCPP_MODE_PTRDEPS) == 0)
		prgMode = PRGMODE_PTRDEPS;
	    else if(strcmp(string, JSCPP_MODE_PTRDEPS "0") == 0)
		prgMode = PRGMODE_PTRDEPS0;
	    else if(strcmp(string, JSCPP_MODE_PTRDEPS "f") == 0)
		prgMode = PRGMODE_PTRDEPSF;
	    else if(strcmp(string, JSCPP_MODE_STMTSTEP) == 0)
		prgMode = PRGMODE_STMTSTEP;
	    else if(strcmp(string, JSCPP_MODE_STMTSTEP0) == 0)
		prgMode = PRGMODE_STMTSTEP0;
	    else if(strcmp(string, JSCPP_MODE_CODECOVER) == 0)
		prgMode = PRGMODE_CODECOVER;
	    else if(strcmp(string, JSCPP_MODE_FCTSEP) == 0)
		prgMode = PRGMODE_FCTSEP;
	    else {
		fprintf(stderr, "ERROR: Unknown mode '%s'!\n", string);
		EGOTO(error2);
	    }

	    break;

#ifdef CONFIG_JSCPPLIB_DEBUG
	case 'X':
	    if(parsedbgopts(DebugOptions, string, &Debug, 0) == -1)
		EGOTO(error);
	    break;
#endif

	case 'v':
	    ptrDepMode |= JSCPP_PTRDEP_VERBOSE;
	    break;

	case 'p':
	    if(traceMode & JSCPP_TRACE_EXPUSERMACS)
		traceMode = (traceMode & ~(JSCPP_TRACE_EXPUSERINCS | JSCPP_TRACE_EXPUSERMACS)) |
		    JSCPP_TRACE_OUTPREPROC;
	    else if(traceMode & JSCPP_TRACE_EXPUSERINCS)
		traceMode = (traceMode & ~JSCPP_TRACE_OUTPREPROC) |
		    JSCPP_TRACE_EXPUSERMACS;
	    else
		traceMode = (traceMode & ~JSCPP_TRACE_OUTPREPROC) |
		    JSCPP_TRACE_EXPUSERINCS;

	    if(depsMode & JSCPP_FDEPS_STATICPREFIX)
		depsMode |= JSCPP_FDEPS_VARPREFIX;
	    else
		depsMode |= JSCPP_FDEPS_STATICPREFIX;
	    break;

	case 'f':
	    depsMode &= ~JSCPP_FDEPS_VARDEPS;
	    break;

	case 'L':
	    traceMode |= JSCPP_TRACE_LINEDIRECTIVE;
	    break;

	case 'l':
	    DefaultIndentation = string ? strtol(string, NULL, 10) :
	    DEFAULT_INDENTATION;
	    break;

	case 'j':
	    depsMode = (depsMode & ~JSCPP_FDEPS_STATICDEF) | JSCPP_FDEPS_STYLEDEPS;
	    ptrDepMode |= JSCPP_PTRDEP_NUMFORMAT | JSCPP_PTRDEP_VERBOSE;
	    tocMode |= 256;
	    break;

	case 't':
	    depsMode |= JSCPP_FDEPS_STYLETSORT;
	    break;

	case 'a':
	    if(unusedMode & 1)
		unusedMode |= 2;
	    else
		unusedMode |= 1;
	    depsMode |= JSCPP_FDEPS_DUPLICATES;
	    tocMode |= 2;
	    break;

	case 's':
	    depsMode |= JSCPP_FDEPS_ANSISYMS;
	    break;
	}
    }

    /*  Add some special symbols depending on the program mode  */
    if(prgMode == PRGMODE_FTRACE || prgMode == PRGMODE_TRACE || prgMode == PRGMODE_FCTWRAP) {
	if(jsparsenvdefine(&ppenv, JSCPP_FTRACE_SYMBOL "=1") == -1)
	    EGOTO(error);
    }

    if(argc < 2) {
	topline();
	fprintf(stderr, "  Enter 'jscpp -h' for more information.\n");
	return 10;
    }

    if(prgMode != PRGMODE_PREPROC && !(prgMode == PRGMODE_FTRACE ||
				       prgMode == PRGMODE_TRACE  ||
				       prgMode == PRGMODE_FCTWRAP)) {
	if(outName == NULL) {
	    fpOut = stdout;
	} else if((fpOut = fopen(outName, "w")) == NULL) {
	    fprintf(stderr, "ERROR: Cannot open \"%s\" -- Break.\n", outName);
	    goto error;
	}
    }

    if(prgMode == PRGMODE_PREPROC) {
	if(doPreProcess(jsparsenvgetpp(&ppenv), (const char **)&argv[1], argc - 1,
			outName, ppMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_OPREPROC) {
	if(doPreProcess(jsparsenvgetpp(&ppenv), (const char **)&argv[1], argc - 1,
			outName, 2 | ppMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_FDEPS) {
	/*  If user specified "-p", we ought to print the static definitions, even
	 *   if he specified "-j"
	 */
	if(depsMode & JSCPP_FDEPS_STATICPREFIX)
	    depsMode |= JSCPP_FDEPS_STATICDEF;

	if(doFDeps(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		   depsMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_PPUNUSED) {
	if(doPPUnUsed(jsparsenvgetpp(&ppenv), (const char **)&argv[1], argc - 1, fpOut,
		      unusedMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_FTRACE) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    JSCPP_TRACE_FCTCALL | JSCPP_TRACE_FCTCALLARGS |
		    JSCPP_TRACE_FCTRET | JSCPP_TRACE_FCTRETARG |
		    JSCPP_TRACE_FCTRETLINE | traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_TRACE) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_FCTWRAP) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    JSCPP_TRACE_FCTWRAP | traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_GLOBVARS) {
	if(doGlobVars(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		      unusedMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_PROTO) {
	if(doFctToc(&ppenv, (const char **)&argv[1], argc - 1, fpOut, tocMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_FCTTOC) {
	if(doFctToc(&ppenv, (const char **)&argv[1], argc - 1, fpOut, tocMode | 1) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_PTRDEPS) {
	if(doPtrDeps(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		     ptrDepMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_PTRDEPS0) {
	if(doPtrDeps(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		     JSCPP_PTRDEP_SPECONLY | ptrDepMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_PTRDEPSF) {
	if(doPtrDeps(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		     JSCPP_PTRDEP_ARGCONST | ptrDepMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_STMTSTEP) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    JSCPP_TRACE_STMTTRACE | traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_STMTSTEP0) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    JSCPP_TRACE_STMTTRACE_BRANCH | traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_CODECOVER) {
	if(doFTrace(&ppenv, (const char **)&argv[1], argc - 1, outName,
		    JSCPP_TRACE_CODECOVERAGE | traceMode) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_FCTSEP) {
	if(doFctSep(&ppenv, (const char **)&argv[1], argc - 1, fpOut,
		    0) == -1)
	    EGOTO(error);

    } else if(prgMode == PRGMODE_IDEPS) {
	if(doIDeps(jsparsenvgetpp(&ppenv), (const char **)&argv[1], argc - 1, fpOut) == -1)
	    EGOTO(error);
    }

    if(prgMode != PRGMODE_PREPROC && fpOut != NULL && fpOut != stdout)
	fclose(fpOut);

    jsparsenvfree(&ppenv);

    return 0;
 error:
    jsperror("jscpp");
 error2:
    return 40;
}

/*JS*********************************************************************
 *
 *************************************************************************/

FILE *openFile(const char *directory, const char *name)

/************************************************************************/
{
    FILE *fp;
    char fileName[JS_FILENAME_MAX];
    const char *s;

    if((s = strrchr(name, '/')))
	++s;
    else
	s = name;

    sprintf(fileName, "%s/%s", directory, s);
    if((fp = fopen(fileName, "w")) == NULL)
	fprintf(stderr, "ERROR: Cannot open '%s' - Break.\n", fileName);

    return fp;
}
