/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Routines to handle the actual preprocessor commands.
 *              Return value:
 *                <0: Error occured.
 *                 0: Preprocessor command interpreted, do not write to output.
 *                 1: Preserve in output, bracketized in PTK_STARTIGN,
 *                    PTK_ENDIGN.
 *                 2: Same as 1, but avoid macro expansion in output.
 *                 3: Preserve in output, bracketized in PTK_STARTSPC,
 *                    PTK_ENDSPC.
 *                 4: Same as 3, but avoid macro expansion in output.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#include <assert.h>

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#ifdef JSCPP_INTERNAL
#define PP_OP_ELLIPSIS   (PTK_USERTOKEN1 +  1)
#endif
/*  END-DEFINITIONS  */

/*  Defines for preprocessor operators  */
#define PP_OP_ADD        '+'
#define PP_OP_SUB        '-'
#define PP_OP_TIM        '*'
#define PP_OP_DIV        '/'
#define PP_OP_IDV        '%'

#define PP_OP_EQU        (PTK_USERTOKEN1 +  2)
#define PP_OP_NEQ        (PTK_USERTOKEN1 +  3)
#define PP_OP_LEQ        (PTK_USERTOKEN1 +  4)
#define PP_OP_GEQ        (PTK_USERTOKEN1 +  5)
#define PP_OP_LOW        '<'
#define PP_OP_GRT        '>'

#define PP_OP_LSH        (PTK_USERTOKEN1 +  6)
#define PP_OP_RSH        (PTK_USERTOKEN1 +  7)
#define PP_OP_LAN        (PTK_USERTOKEN1 +  8)
#define PP_OP_LOR        (PTK_USERTOKEN1 +  9)

#define PP_OP_AND        '&'
#define PP_OP_OR         '|'
#define PP_OP_EXO        '^'

/*  Triadic operators  */
#define PP_OP_QUEST      '?'
#define PP_OP_COLON      ':'

/*  Unary operators:  */
#define PP_OP_UNEG       (PTK_USERTOKEN1 + 10)
#define PP_OP_UNOT       '~'
#define PP_OP_ULNOT      '!'

typedef struct {
    int OPD_Op;
    int OPD_Prio;
    int OPD_NArgs;
} ppoperatordef_t;

#ifdef CONFIG_HAS_LONGLONG
/*  use 64 bit integers  */
typedef long long preprocint_t;
#define preprocintgetvalue(string)   strtoull(string, NULL, 0);
#else
/*  use at least long :)  */
typedef long preprocint_t;
#define preprocintgetvalue(string)   strtoul(string, NULL, 0);
#endif

#define PRAGMA_NAME_DEFER       "defer"
#define PRAGMA_NAME_EXPORIG     "expandorig"
#define PRAGMA_NAME_EXPORIGFULL "expandorigfull"

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
IPrototype int		 _jsppdoline(jsppstream_t *jps, int lineNr, char *fileName);
IPrototype int           _jsppdoerror(jsppstream_t *jps, jsctoken_t *pToken, int n);
#ifdef CONFIG_JSCPP_WARNING
IPrototype int           _jsppdowarn(jsppstream_t *jps, jsctoken_t *pToken, int n);
#endif

IPrototype int           _jsppdoifdef(jsppstream_t *jps, char *name, int flag);
IPrototype int           _jsppdoif(jsppstream_t *jps, jsctoken_t *pToken, int n,
				   int isif, int flag);
static int               evalIExpression(jsppstream_t *jps, jsctoken_t *pToken,
					 int n);
static int               popOps(jsppstream_t *jps, preprocint_t *vStk,
				unsigned int *pvStkI,
				ppoperatordef_t *oStk,
				unsigned int *poStkI,
				int prior);
static preprocint_t      evaluate(jsppstream_t *jps, int type, preprocint_t val1,
				  preprocint_t val2);

IPrototype int		 _jsppdocondcomp(jsppstream_t *jps, int type,
					 int condition);

IPrototype int           _jsppdopragma(jsppstream_t *jps, jsctoken_t *pToken,
				       int n);


IPrototype const jsctokops_t _jsppOperators[];

/*********     GLOBAL VARIABLES                                 *********/
/* **  Operators for the preprocessor (for "#if" lines) ** */
const jsctokops_t _jsppOperators[] = {
    { "==",   PP_OP_EQU },
    { "!=",   PP_OP_NEQ },
    { "<=",   PP_OP_LEQ },
    { ">=",   PP_OP_GEQ },

    { "<<",   PP_OP_LSH },
    { ">>",   PP_OP_RSH },
    { "&&",   PP_OP_LAN },
    { "||",   PP_OP_LOR },

    { "...",  PP_OP_ELLIPSIS }, /*  for C99 variadic macros  */

    { NULL, 0 }
};

static const ppoperatordef_t OperatorDefs[] = {
    { PP_OP_QUEST,  2, 3 },
    { PP_OP_COLON,  3, 3 },

    { PP_OP_LOR,    4, 2 },
    { PP_OP_LAN,    6, 2 },

    { PP_OP_OR,     8, 2 },
    { PP_OP_EXO,   10, 2 },
    { PP_OP_AND,   12, 2 },

    { PP_OP_EQU,   14, 2 },
    { PP_OP_NEQ,   14, 2 },

    { PP_OP_LOW,   16, 2 },
    { PP_OP_GRT,   16, 2 },
    { PP_OP_LEQ,   16, 2 },
    { PP_OP_GEQ,   16, 2 },

    { PP_OP_LSH,   18, 2 },
    { PP_OP_RSH,   18, 2 },

    { PP_OP_ADD,   20, 2 },
    { PP_OP_SUB,   20, 2 },

    { PP_OP_TIM,   22, 2 },
    { PP_OP_DIV,   22, 2 },
    { PP_OP_IDV,   22, 2 },

    { PP_OP_UNEG,  30, 1 },
    { PP_OP_UNOT,  30, 1 },
    { PP_OP_ULNOT, 30, 1 }
};

/*JS*********************************************************************
 *   Handles '#line' preprocessor commands.
 *************************************************************************/

int _jsppdoline(jsppstream_t *jps, int lineNr, char *fileName)

/************************************************************************/
{
    static char FileNameBuffer[JS_FILENAME_MAX + 1];

    if(fileName) {
	/*	Copy the file name to the static buffer. Reusage is OK  */
	strcpy(FileNameBuffer, fileName);
	jps->JSPP_File->IF_Name = FileNameBuffer;
    }

    /*  We may be some lines forwarded if some empty lines were
     *   after the line command, so correct this
     */
    jps->JSPP_File->IF_LineNr += lineNr - jps->JSPP_OLineNr - 1;
    jps->JSPP_File->IF_FFPtr->EF_LineNr += lineNr - jps->JSPP_OLineNr - 1;

    /*  Preserve line in output, and macro expansion may be skipped  */
    return 2;
}

/*JS*********************************************************************
 *   Handles '#error' lines.
 *************************************************************************/

int _jsppdoerror(jsppstream_t *jps, jsctoken_t *pToken, int n)

/************************************************************************/
{
    int i;

    jsppprinterr(jps, "\'#error\' directive encountered", 1);

    for(i = 0 ; i < n ; i++) printCToken(stderr, pToken[i]);
    fprintf(stderr, "\n");

    return -1;
}

#ifdef CONFIG_JSCPP_WARNING
/*JS*********************************************************************
 *   Handles '#warning' lines.
 *************************************************************************/

int _jsppdowarn(jsppstream_t *jps, jsctoken_t *pToken, int n)

/************************************************************************/
{
    int i;

    jsppprinterr(jps, "", 2);

    for(i = 0 ; i < n ; i++) printCToken(stderr, pToken[i]);
    fprintf(stderr, "\n");

    return 0;
}
#endif

/*JS*********************************************************************
 *   Handles '#ifdef' (flag true) and '#ifndef' (flag false) preprocessor
 *    lines.
 *************************************************************************/

int _jsppdoifdef(jsppstream_t *jps, char *name, int flag)

/************************************************************************/
{
    _jsppmacentry_t *entry;

    /*  Search for macro entry  */
    if((entry = _jsppfindmacentry(jps, name)) != NULL &&
       !(entry->ME_Flags & MAFLAG_UNDEFINED))
	entry->ME_Flags |= MAFLAG_USED;
    else
	flag = !flag;

    if(_jsppdocondcomp(jps, IFSTK_TYPE_IF, flag) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Handles '#if' and '#elif' preprocessor lines.
 *************************************************************************/

int _jsppdoif(jsppstream_t *jps, jsctoken_t *pToken, int n, int type, int flag)

/************************************************************************/
{
    int res;

    /*  Don't bother to get the value if we are within an FALSE part  */
    if(!flag)
	res = 0;
    else if((res = evalIExpression(jps, pToken, n)) == -1)
	EGOTO(error);

    if(_jsppdocondcomp(jps, type, res) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int evalIExpression(jsppstream_t *jps, jsctoken_t *pToken, int n)

/************************************************************************/
{
    static ppoperatordef_t *OStk;
    static preprocint_t *VStk;
    static unsigned int VStkM, OStkM;
    unsigned int opFlag, oStkI, vStkI;

    vStkI = oStkI = 0;
    for(opFlag = 0 ; n > 0 ; pToken++, n--) {
	int tokVal, opi;

	tokVal = pToken->TK_Value;

	if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
	    continue;

	/*  Increase memory if needed  */
	if(oStkI >= OStkM) {
	    OStkM += 10;
	    if((OStk = (ppoperatordef_t *)
		realloc(OStk, OStkM * sizeof(*OStk))) == NULL) /*DBG*/
		EGOTO(error);
	}

	if(vStkI >= VStkM) {
	    VStkM += 10;
	    if((VStk = (preprocint_t *)
		realloc(VStk, VStkM * sizeof(*VStk))) == NULL) /*DBG*/
		EGOTO(error);
	}

	if(opFlag) { /* ***  Looking for operator  *** */
	    if(tokVal == ')') {
		if(popOps(jps, VStk, &vStkI, OStk, &oStkI, 0) == -1)
		    EGOTO(error);

		/*  Remove matching opening bracket  */
		--oStkI;

		if(oStkI < 0 || OStk[oStkI].OPD_Op != '(') {
		    jsppprinterr(jps, "Unmatched closing bracket", 1);
		    EGOTO(error);
		}
		continue;
	    }

	    opFlag = FALSE; /*  Now look again for variables  */

	} else { /* ***  Looking for variable  *** */
	    if(tokVal == '(') {
		/*  Push an operator with a very low priority  */
		OStk[oStkI].OPD_Op   = tokVal;
		OStk[oStkI].OPD_Prio = -1;
		++oStkI;
		continue;
	    }

	    if(tokVal == PP_OP_SUB) {
		tokVal = PP_OP_UNEG;

	    } else if(tokVal == PP_OP_ADD) {
		/*  '+' can be ignored  */
		continue;

	    } else if(tokVal != PP_OP_ULNOT && tokVal != PP_OP_UNOT) {
		preprocint_t value;

		if(tokVal == CTK_NUMBER || tokVal == CTK_INUMBER) {
		    value = preprocintgetvalue(pToken->TK_String);

		} else if(tokVal == CTK_CHARCONST || tokVal == CTK_WCHARCONST) {
		    value = pToken->TK_String[0];

		} else if(tokVal == CTK_IDENT) {
		    /*  As required by the C standard, identifiers that are not
		     *   defined (and thus are no longer visible here) are equal
		     *   to zero
		     */
		    value = 0;

		} else {
		    jsppprinterr(jps, "Missing value in preprocessor line: \"%s\"", 0,
				 pToken->TK_String ? pToken->TK_String : "???");

		    value = 0;
		}

		/*  Push value on stack  */
#ifdef XDEBUG
		fprintf(stderr, "Pushing value %d (%d,%d)\n", value, vStkI, oStkI);
#endif
		VStk[vStkI++] = value;
		opFlag = TRUE; /*  Now look for operator  */
		continue;
	    }
	}

	/*  Look for priority  */
	for(opi = sizeof(OperatorDefs) / sizeof(OperatorDefs[0]) - 1 ;
	    opi >= 0 ; --opi)
	    if(OperatorDefs[opi].OPD_Op == tokVal)
		break;

	if(opi < 0) {
	    jsppprinterr(jps, "Invalid operator %d \"%s\" in preprocessor line", 1,
			 pToken->TK_Value, pToken->TK_String);
	    EGOTO(error);
	}

	if(OperatorDefs[opi].OPD_NArgs == 1) {
	    /*  Push unary operator on stack and continue  */
	    OStk[oStkI++] = OperatorDefs[opi];
	    continue;
	}

	if(popOps(jps, VStk, &vStkI, OStk, &oStkI, OperatorDefs[opi].OPD_Prio) == -1)
	    EGOTO(error);

#ifdef XDEBUG
	fprintf(stderr, "Pushing operator <%d> (%d,%d)\n", tokVal, vStkI, oStkI);
#endif

	/*  Push operator on stack  */
	OStk[oStkI++] = OperatorDefs[opi];
    }

    if(popOps(jps, VStk, &vStkI, OStk, &oStkI, 0) == -1)
	EGOTO(error);

    /*  Check correctness of expression  */
    if(oStkI != 0 || vStkI != 1) {
	jsppprinterr(jps, "Invalid expression", 1);
	EGOTO(error);
    }

#ifdef XDEBUG
    fprintf(stderr, "RESULT is %d\n", VStk[0]);
#endif

    return VStk[0] ? TRUE : FALSE;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int popOps(jsppstream_t *jps, preprocint_t *vStk,
		  unsigned int *pvStkI,
		  ppoperatordef_t *oStk,
		  unsigned int *poStkI,
		  int prior)

/************************************************************************/
{
    unsigned int oStkI, vStkI;

    oStkI = *poStkI;
    vStkI = *pvStkI;

    while(oStkI > 0 && prior <= oStk[oStkI - 1].OPD_Prio) {
#ifdef XDEBUG
	fprintf(stderr, "  Popping %d %d\n", oStkI, vStkI);
#endif

	oStkI--;
	vStkI -= oStk[oStkI].OPD_NArgs - 1;

	if(oStk[oStkI].OPD_NArgs == 3) {
	    /*  Check for correct usage of "?:" operator  */
	    if(oStk[oStkI].OPD_Op != PP_OP_COLON ||
	       oStk[oStkI - 1].OPD_Op != PP_OP_QUEST) {
		jsppprinterr(jps, "Incorrect usage of '?:' operator", 1);
		EGOTO(error);
	    }
	    oStkI--;
	    vStk[vStkI - 1] = vStk[vStkI - 1] ? vStk[vStkI] : vStk[vStkI + 1];

	} else
	    vStk[vStkI - 1] = evaluate(jps, oStk[oStkI].OPD_Op, vStk[vStkI - 1], vStk[vStkI]);
    }

    *poStkI = oStkI;
    *pvStkI = vStkI;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static preprocint_t evaluate(jsppstream_t *jps, int type, preprocint_t val1, preprocint_t val2)

/************************************************************************/
{
    preprocint_t res;

    switch(type) {
    case PP_OP_ADD:   res = (val1 +  val2);  break;
    case PP_OP_SUB:   res = (val1 -  val2);  break;
    case PP_OP_TIM:   res = (val1 *  val2);  break;
    case PP_OP_DIV:
	if(val2 == 0) EGOTO(divzero);
	res = (val1 /  val2);
	break;
    case PP_OP_IDV:
	if(val2 == 0) EGOTO(divzero);
	res = (val1 %  val2);
	break;

    case PP_OP_EQU:   res = (val1 == val2);  break;
    case PP_OP_NEQ:   res = (val1 != val2);  break;
    case PP_OP_LEQ:   res = (val1 <= val2);  break;
    case PP_OP_GEQ:   res = (val1 >= val2);  break;
    case PP_OP_LOW:   res = (val1 <  val2);  break;
    case PP_OP_GRT:   res = (val1 >  val2);  break;

    case PP_OP_LSH:   res = (val1 << val2);  break;
    case PP_OP_RSH:   res = (val1 >> val2);  break;
    case PP_OP_LAN:   res = (val1 && val2);  break;
    case PP_OP_LOR:   res = (val1 || val2);  break;

    case PP_OP_AND:   res = (val1 &  val2);  break;
    case PP_OP_OR :   res = (val1 |  val2);  break;
    case PP_OP_EXO:   res = (val1 ^  val2);  break;

    case PP_OP_UNEG:  res = -val1;  break;
    case PP_OP_UNOT:  res = ~val1;  break;
    case PP_OP_ULNOT: res = !val1;  break;
    default:
	assert(FALSE);
	break;
    }

#ifdef XDEBUG
    fprintf(stderr, "  EVAL %d <%d> %d -> %d\n", val1, type, val2, res);
#endif

    return res;
 divzero:
    jsppprinterr(jps, "Division by zero in preprocessor line", 0);
    return 1;
}

/*JS*********************************************************************
 *   Do conditional compilation handling.
 *************************************************************************/

int _jsppdocondcomp(jsppstream_t *jps, int op, int cond)

/************************************************************************/
{
    int ret, level;

    if((ret = ifstackop(&(jps->JSPP_Conds), op, cond)) == -1)
	return -1;

    level = ifstacklevel(&(jps->JSPP_Conds));

    if(level < jps->JSPP_File->IF_CondLevel) {
	jsppprinterr(jps, "Unmatched '%s'", 1,
		     op == IFSTK_TYPE_ELSE ? "#else" :
		     op == IFSTK_TYPE_ELIF ? "#elif" :
		     op == IFSTK_TYPE_ENDIF ? "#endif" : "???");
	return -1;
    }

    return ret;
}

/*JS*********************************************************************
 *   Recognize internal pragmas.
 *************************************************************************/

int _jsppdopragma(jsppstream_t *jps, jsctoken_t *pToken, int n)

/************************************************************************/
{
    int flag, mFlags, commaFlag;

    /*  Look for pragma introducer  */
    for(flag = 0 ; ; ) {
	int i;

	/*  Skip blank token  */
	while(n > 0 && CTK_BLANK_START <= pToken->TK_Value &&
	      pToken->TK_Value <= CTK_BLANK_END) {
	    ++pToken;
	    --n;
	}

	if(n <= 0 || pToken->TK_Value != CTK_IDENT)
	    goto invalid;

	if(flag)
	    break;

	/*  Does any of the pragma introducer match?  */
	for(i = 0 ; JSPPPragmaIntroducer[i] ; ++i)
	    if(tkStrcmp(pToken->TK_String, JSPPPragmaIntroducer[i]) == 0)
		break;

	/*  None matched, so stop  */
	if(JSPPPragmaIntroducer[i] == NULL)
	    break;

	flag = 1;
	++pToken;
	--n;
    }

    if(tkStrcmp(pToken->TK_String, PRAGMA_NAME_DEFER) == 0) {
	mFlags = MAFLAG_DEFEREXPANSION;

    } else if(tkStrcmp(pToken->TK_String, PRAGMA_NAME_EXPORIG) == 0) {
	mFlags = MAFLAG_EXPANDORIG;

    } else if(tkStrcmp(pToken->TK_String, PRAGMA_NAME_EXPORIGFULL) == 0) {
	mFlags = MAFLAG_EXPANDORIGFULL;

    } else {
	if(jps->JSPP_DoPragma == NULL) {
	    /*  Print message if pragma started with our introducer  */
	    if(flag)
		jsppprinterr(jps, "Unknown pragma \"%s\"", 0, pToken->TK_String);

	    goto invalid;
	}

	return jps->JSPP_DoPragma(jps, pToken, n, jps->JSPP_User);
    }

    for(commaFlag = FALSE ; ; ) {
	/*  Skip blank token  */
	for( ; ; ) {
	    do {
		++pToken;
		--n;
	    } while(n > 0 && CTK_BLANK_START <= pToken->TK_Value &&
		    pToken->TK_Value <= CTK_BLANK_END);

	    if(!commaFlag)
		/*  Expect identifier  */
		break;

	    /*  Expect ','  */
	    if(n <= 0)
		break;

	    if(pToken->TK_Value != ',')
		EGOTO(syntaxErr);

	    commaFlag = FALSE;
	}

	if(n <= 0) {
	    if(!commaFlag)
		EGOTO(syntaxErr);

	    break;
	}

	if(pToken->TK_Value != CTK_IDENT)
	    goto invalid;

	if(jsppsetmacflags(jps, pToken->TK_String, mFlags) == -1)
	    EGOTO(error);

	commaFlag = TRUE;
    }

    /*  Eat up  */
    return 0;
 invalid:
    /*  Invalid pragma recognized, preserve in output  */
    if(flag)
	jsppprinterr(jps, "Syntax error in pragma line", 0);

    return 2;
 syntaxErr:
    if(flag)
	jsppprinterr(jps, "Syntax error in pragma line", 0);

    return 2;
 error:
    return -1;
}
