/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Get next token.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES						*********/

/*  Internal modes for routine doPreProcLine  */
#define DPP_DO_COMMAND	   (1 << 0) /*  Do all preprocessor commands  */
#define DPP_DO_IFCOMMAND   (1 << 1) /*  Do if command (might change something  */
#define DPP_SAVE_ORIG	   (1 << 2)

/*  Internal modes for routine doPreProcLine  */
#define DPP_EXPAND_MACROS  (1 << 4)

#define ID_IS_OUTPUT(jps)  ifstackop(&((jps)->JSPP_Conds), IFSTK_TYPE_ASK, 0)
#define ID_MIGHT_OUT(jps)  ifstackop(&((jps)->JSPP_Conds), IFSTK_TYPE_CANBETRUE, 0)

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
Prototype jsctoken_t	 jsppgettoken(jsppstream_t *jps, jsctoken_t *pOToken, int mode);
static int		 doPreProcLine(jsppstream_t *jps, int mode);

/*JS*********************************************************************
 *   Get next token where preprocessor commands and macro expansion has
 *    already been done. In case of a preprocessor's command, the sequence
 *    of original tokens, i.e. the preprocessor line itself is returned
 *    first in pOToken with an empty token in the return value. If you
 *    don't want to get the original one's, simply pass NULL for pOToken.
 *************************************************************************/

jsctoken_t jsppgettoken(jsppstream_t *jps, jsctoken_t *pOToken, int mode)

/************************************************************************/
{
    jsctoken_t token;

 redo:
    /*  Any defered tokens in FIFO queue?  */
    if(pOToken) {
	*pOToken = jctokget(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_OQueue));

	if(pOToken->TK_Value != CTK_EOF) {
	    if(pOToken->TK_Value == CTK_ERROR)
		EGOTO(error);

	    /*  Return only original token, no preprocessed one  */
	    token.TK_Flags = 0; token.TK_Value = PTK_EMPTY; token.TK_String = NULL;
	    return token;
	}

	/*  Clear "#define"-line-mode after exhausting orig queue  */
	jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_Mode = 0;
    }

    /*  Get next token  */
    for( ; ; ) {
	int ret;
	int mode2;

	mode2 = mode;
	if(!ID_IS_OUTPUT(jps))
	    mode2 |= GTK_NO_EXP;

	if(jps->JSPP_NFiles > 1 &&
	   /*  We can request to see original tokens in user include files  */
	   !((jps->JSPP_Mode & JPPM_EXPUSERFILES) && jsppisusrinclude(jps, 0))) {
	    /*  No original tokens within include files  */
	    token = (mode & GTK_NO_EXP) ?
		_jspptkbasic2(jps, NULL, mode2) : jspptkmacro(jps, NULL, mode2);

	    if(pOToken) {
		pOToken->TK_Flags = 0;
		pOToken->TK_Value = PTK_EMPTY;
		pOToken->TK_String = NULL;
	    }
	} else {
	    token = (mode & GTK_NO_EXP) ?
		_jspptkbasic2(jps, pOToken, mode2) :
		jspptkmacro(jps, pOToken, mode2);
	}

	if(token.TK_Value != CTK_EOF)
	    break;

	/*	Pop finished file from stack  */
	if(jps->JSPP_NFiles == 0)
	    goto ende;

	if((ret = _jspppopistack(jps)) == -1)
	    EGOTO(error);

	if(ret) /*  Is really EOF  */
	    goto ende;
    }

    /*  Check for preprocessor line, but not while macro expansion or
     *   in QStream!
     */
    if(token.TK_Value == '#' && IF_ISFILE(jps->JSPP_File) &&
       jps->JSPP_LastToken == CTK_BLANKSEOL) {
	int ppMode;

	/*	Only do macro def's and undef's, error, line and pragma commands
	 *	 when output is active (depends on conditional compilation).
	 */
	if(ID_IS_OUTPUT(jps))
	    ppMode = DPP_DO_COMMAND;
	else if(ID_MIGHT_OUT(jps))
	    ppMode = DPP_DO_IFCOMMAND;
	else
	    ppMode = 0;

	/*	Save original tokens if caller wants and we are not within
	 *	 an include file.
	 */
	if(pOToken && (jps->JSPP_NFiles <= 1 ||
		       ((jps->JSPP_Mode & JPPM_EXPUSERMACROS) &&
			jsppisusrinclude(jps, 0))))
	    ppMode |= DPP_SAVE_ORIG;

	/*	Save the line number where preprocessor command started  */
	jps->JSPP_OLineNr = jps->JSPP_File->IF_LineNr;

	/*	Read and execute preprocessor line  */
	if(doPreProcLine(jps, ppMode) == -1)
	    EGOTO(error);

	goto redo;
    }

 ende:
    /*  Only newlines survive during conditional compilation  */
    if(!ID_IS_OUTPUT(jps) && token.TK_Value != CTK_BLANKSEOL &&
       token.TK_Value != CTK_EOF) {
	int n;

	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	if(token.TK_String && (n = countnlCToken(token.TK_String)) > 0) {
	    token = _jsppgetnnewlines(n);
	} else {
	    /*  Empty token  */
	    token.TK_Flags  = 0;
	    token.TK_Value  = PTK_EMPTY;
	    token.TK_String = NULL;
	}
    }

    return token;
 error:
    token.TK_Flags  = 0;
    token.TK_Value  = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   The '#' token has already been recognized, now read the whole
 *    preprocessor line and do the command.
 *************************************************************************/

static int doPreProcLine(jsppstream_t *jps, int mode)

/************************************************************************/
{
    jsctoken_t token;
    int state, getMode, retValue;
    int auxInt, auxSt;
    int startIndex, tokenIndex;
#ifdef CONFIG_JSCPP_OPTINC
    int spaceFlag;
#endif
    static const struct PPKey {
	const char *PP_Key;
	int	PP_State;
    } PPKeyWords[] = {
	/*  Reserved words of preprocessor and next state to go to  */
	{ "include", 10 },
	{ "define",  20 }, { "undef",   30 },
	{ "if",      40 }, { "else",    50 }, { "endif",   60 },
	{ "ifdef",   70 }, { "ifndef",  80 }, { "elif",    90 },
	{ "line",   100 }, { "error",  110 }, { "pragma", 120 },
	/*  Some extensions  */
#ifdef CONFIG_JSCPP_SYSVCMD
	{ "ident",  302 },
	{ "sccs",   302 },
#endif
#ifdef CONFIG_JSCPP_INCNEXT
	{ "include_next", 130 },
#endif
#ifdef CONFIG_JSCPP_SINCLUDE
	{ "sinclude", 190 },
#endif
#ifdef CONFIG_JSCPP_DEFIMM
	{ "define_imm", 170 },
#endif
#ifdef CONFIG_JSCPP_WARNING
	{ "warning", 140 },
#endif
#ifdef CONFIG_JSCPP_ASSERTION
	{ "assert", 150 },
	{ "unassert", 160 },
#endif
	{ NULL,     200 }
    };

    /*  Clear queue and save first token	*/
    jctokclear(&(jps->JSPP_Queue));
    token.TK_Flags  = 0;
    token.TK_String = NULL;

    token.TK_Value = PTK_STARTIGN;
    if((startIndex = jctokinsert(&(jps->JSPP_Queue), token, 0)) == -1)
	EGOTO(error);

    token.TK_Value = '#';
    if(jctokinsert(&(jps->JSPP_Queue), token, 0) == -1)
	EGOTO(error);

    /*  Save '#' token in original queue too  */
    if((mode & DPP_SAVE_ORIG) &&
       jctokinsert(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_OQueue), token, 0) == -1)
	EGOTO(error);

#ifdef CONFIG_JSCPP_OPTINC
    if(jps->JSPP_File->IF_Flag & IF_FLAG_NONSPC)
	spaceFlag = FALSE;
    else
	/*  Mark that there was solely space prior to preprocessor command  */
	spaceFlag = TRUE;
#endif

    retValue = 0;
    getMode = 0;
    for(state = 0 ; (state & 1) == 0 ; ) {
	do {
	    /*  Get next token and save it  */
	    if(mode & DPP_SAVE_ORIG) {
		jsctoken_t oToken;

		token = mode & DPP_EXPAND_MACROS ?
		    jspptkmacro(jps, &oToken, getMode) :
		    _jspptkbasic(jps, &oToken, getMode);

		if(oToken.TK_Value != PTK_EMPTY &&
		   jctokinsert(&(jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_OQueue), oToken, 0) == -1)
		    EGOTO(error);

	    } else {
		token = mode & DPP_EXPAND_MACROS ?
		    jspptkmacro(jps, NULL, getMode) :
		    _jspptkbasic(jps, NULL, getMode);
	    }

	    /*  Skip empty tokens from macro expansion  */
	} while(token.TK_Value == PTK_EMPTY);

	if(token.TK_Value == CTK_EOF) {
	    token.TK_Value = CTK_BLANKSEOL;
	    token.TK_String = "\n";
	}

	/*  Insert token in queue and check for end of preprocessor line  */
	if((tokenIndex = jctokinsert(&(jps->JSPP_Queue), token, 0)) == -1)
	    EGOTO(error);

	/*	Skip non-newline blanks  */
	if(CTK_BLANK_START2 <= token.TK_Value && token.TK_Value <= CTK_BLANK_END)
	    continue;

	if(token.TK_Value == CTK_ERROR) EGOTO(error);

	if(token.TK_Value == CTK_BLANKSEOL)
	    /*  Allow different cases to see whole line before break  */
	    state |= 1;

    redocurrent:
#ifdef XDEBUG
	printf("state %3d, index %2d, mode %d token %d '%s'\n",
	       state, tokenIndex, mode, token.TK_Value,
	       token.TK_String ? token.TK_String : "<NULL>");
	/*jctokshowqueue(stdout, &(jps->JSPP_Queue));*/
#endif

	switch(state) {
	case 0:
	    /*  Look at preprocessor command	*/
	    if(token.TK_Value == CTK_IDENT) {
		int i;

		/*  Look in table of preprocessor keywords  */
		for(i = 0 ; PPKeyWords[i].PP_Key ; i++)
		    if(tkStrcmp(PPKeyWords[i].PP_Key, token.TK_String) == 0) break;

		state = PPKeyWords[i].PP_State;
		goto redocurrent;
#ifdef CONFIG_JSCPP_LAZY_LINECMD
	    } else if(token.TK_Value == CTK_NUMBER ||
		      token.TK_Value == CTK_INUMBER) {
		state = 102; /*  Behave as if "line" was seen  */
		goto redocurrent;
#endif
	    } else {
		if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		    jsppprinterr(jps, "Wrong syntax in preprocessor line", 0);
		} else {
		    jsppprinterr(jps, "Wrong syntax in preprocessor line", 1);
		    retValue = -1;
		}

		state = 302; /*  read until end of line  */
	    }
	    break;
	case 10:   /* **  INCLUDE  ** */
#ifdef CONFIG_JSCPP_INCNEXT
	case 130:  /* **  INCLUDE_NEXT  ** */
#endif
#ifdef CONFIG_JSCPP_SINCLUDE
	case 190:  /* **  SINCLUDE  ** */
#endif
	    if(!(mode & DPP_DO_COMMAND)) {
		state = 302; /*  Just ignore anything on that line  */
		break;
	    }

	    getMode |= GCTK_PPINCLUDELINE; /*  Get <..> correctly  */
	    mode |= DPP_EXPAND_MACROS;     /*  Macro expansion is allowed  */

#ifdef CONFIG_JSCPP_INCNEXT
	    /*  Distinguish between "include" and "include_next"  */
	    if(state == 130) auxInt = JPP_PUSHFILE_FINCNEXT;
	    else
#endif
#ifdef CONFIG_JSCPP_SINCLUDE
		/*  Distinguish between "include" and "sinclude"  */
		if(state == 190) auxInt = JPP_PUSHFILE_FSINCLUDE;
		else
#endif
		    auxInt = 0;

	    state = 12;
	    break;
	case 12:   /*  Read file name for inclusion  */
	    state = 14; /*  read trailer of line  */

	    auxSt = tokenIndex;

	    if(token.TK_Value == CTK_PPSTRING)
		auxInt |= JPP_PUSHFILE_FNOCURRDIR; /*  Look only in list of include directories  */
	    else if(token.TK_Value == '<') {
		/*  Macro expansion yields '<' ... '>'  */
		auxSt++;
		state = 18;
	    } else if(token.TK_Value != CTK_STRING) {
		jsppprinterr(jps, "Invalid include directive", 1);
		retValue = -1;
		state = 302; /*  read until end of line  */
	    }
	    break;
	case 15: case 17: /*  Include line finished, now push include file  */
	    if(jspppushfile(jps, jps->JSPP_Queue.TKFF_Token[auxSt].TK_String, auxInt))
		retValue = -1;
	    break;
	case 18: /*  Look for '>'  */
	    if(token.TK_Value == '>') {
		/*  Delimit with CTK_EOF for routine _jsppstringize  */
		jps->JSPP_Queue.TKFF_Token[tokenIndex].TK_Value = CTK_EOF;
		jps->JSPP_Queue.TKFF_Token[tokenIndex].TK_String = NULL;

		token = _jsppstringize(&(jps->JSPP_Queue).TKFF_Token[auxSt], 0);

		/*  Put on queue  */
		jps->JSPP_Queue.TKFF_Index = auxSt; /*  forget original ones  */
		if((tokenIndex = jctokinsert(&(jps->JSPP_Queue), token, 0)) == -1)
		    EGOTO(error);

		/*  Normal behaviour  */
		auxSt = tokenIndex;
		auxInt = 3;
		state = 14;
	    }
	    break;
	case 20:   /* **  DEFINE   ** */
	    /*  Remember that this is a "#define" line  */
	    if(mode & DPP_DO_COMMAND)
		jps->JSPP_OQueue[jps->JSPP_QueueLevel].JSOQ_Mode |= 1;

	case 30:   /* **  UNDEF    ** */
#ifdef CONFIG_JSCPP_DEFIMM
	case 170:   /* **  DEFINE_IMM   ** */
#endif
	    if(!(mode & DPP_DO_COMMAND)) {
		state = 302; /*  Just ignore anything on that line  */
		break;
	    }
	case 70:   /* **  IFDEF    ** */
	case 80:   /* **  IFNDEF   ** */
	    state += 2;
	    break;
	case 22: case 32: case 72: case 82: /*  Get identifier and position  */
#ifdef CONFIG_JSCPP_DEFIMM
	case 172:
#endif
	    if(token.TK_Value != CTK_IDENT) {
		jsppprinterr(jps, "Expected identifier in preprocessor line", 1);
		retValue= -1;
		state = 302; /*  read until end of line  */
		break;
	    }
	    auxInt = tokenIndex + 1; /*  start of definition  */
	    auxSt = auxInt - 1;	 /*  index of macro's name  */

	    state += 2;
	    break;
	case 24: /* **  Wait till end of line  ** */
	    break;
#ifdef CONFIG_JSCPP_DEFIMM
	case 174: /* **  Wait till end of line  ** */
	    break;
	case 175:
#endif
	case 25: case 35: case 37: /*  Now do line  */
	    if(mode & DPP_DO_COMMAND) {
		long mFlags;

		if(jps->JSPP_NFiles > 1)
		    mFlags = (jps->JSPP_File->IF_Flag & IF_SYSINCLUDE) ?
			MAFLAG_SYSTEMHEADER : MAFLAG_USERHEADER;
		else
		    mFlags = 0;

#ifdef CONFIG_JSCPP_DEFIMM
		/*  Handle "define_imm"  */
		if(state == 175) mFlags |= MAFLAG_EVALIMMEDIATE;
#endif

		retValue = jsppdefundefmac(jps, jps->JSPP_Queue.TKFF_Token[auxSt].
					   TK_String,
					   &(jps->JSPP_Queue.TKFF_Token[auxInt]),
					   (state != 35 && state != 37) ?
					   (tokenIndex - auxInt) : -1, mFlags);
	    }
	    break;
	case 85: case 87:
#ifdef CONFIG_JSCPP_OPTINC
	    /*  Must have been no space up to now and this '#ifndef' must
	     *   be outermost and first construct like this to recognize.
	     */
	    if(spaceFlag &&
	       ifstacklevel(&(jps->JSPP_Conds)) == jps->JSPP_File->IF_CondLevel) {
		/*  We are at the very first '#ifndef' in this file  */
		if(!(jps->JSPP_File->IF_Flag & IF_FLAG_CONSTRUCT)) {
		    /*  Looks good, save symbol and set flags  */
		    if((jps->JSPP_File->IF_Symbol =
			strdup(jps->JSPP_Queue.TKFF_Token[auxSt].TK_String)) == NULL)
			EGOTO(error);

		    jps->JSPP_File->IF_Flag = (jps->JSPP_File->IF_Flag &
					       ~(IF_FLAG_NONSPC | IF_FLAG_NONSPC1)) |
			IF_FLAG_INCONSTRUCT;
		} else {
		    /*  This is the second time we have a construct like this, so
		     *   give up.
		     */
		    jps->JSPP_File->IF_Flag |= IF_FLAG_NONSPC; /* is set anyway  */
		}
	    }
#endif
	case 75: case 77:
	    /*  Now do '#ifdef' / '#ifndef' line  */
	    retValue = _jsppdoifdef(jps, jps->JSPP_Queue.TKFF_Token[auxSt].TK_String,
				    state == 75 || state == 77);
	    break;
	case 40:   /* **  IF       ** */
	case 90:   /* **  ELIF     ** */
	    getMode |= EXPM_DEFINED_MACRO | GTK_PREPROCOPS;
	    mode |= DPP_EXPAND_MACROS;
	    state += 2;
	    break;
	case 42: case 92:
	    auxInt = tokenIndex;  /*  start of expression  */
	    state += 2;
	    break;
	case 44: case 94: /* **  Wait till end of line  ** */
	    break;
	case 45: /*  Now do "#if" line  */
	    retValue = _jsppdoif(jps, &(jps->JSPP_Queue.TKFF_Token[auxInt]),
				 tokenIndex - auxInt, IFSTK_TYPE_IF,
				 mode & DPP_DO_COMMAND);
	    break;
	case 95: /*  Now do "#elif" line  */
	    retValue = _jsppdoif(jps, &(jps->JSPP_Queue.TKFF_Token[auxInt]),
				 tokenIndex - auxInt, IFSTK_TYPE_ELIF,
				 mode & (DPP_DO_COMMAND | DPP_DO_IFCOMMAND));
	    break;
	case 50:   /* **  ELSE     ** */
	case 60:   /* **  ENDIF    ** */
	    state += 2;
	    break;
	case 53: case 55:   /*  Now do '#else' line  */
	    if(_jsppdocondcomp(jps, IFSTK_TYPE_ELSE, FALSE) == -1)
		retValue = -1;
	    break;
	case 63: case 65:  /*  Now do '#endif' line  */
#ifdef CONFIG_JSCPP_OPTINC
	    {
		/*  Remember if this if-construct had an else or elif part  */
		int hadElse = ifstackhadelse(&(jps->JSPP_Conds));
#endif

		if(_jsppdocondcomp(jps, IFSTK_TYPE_ENDIF, FALSE) == -1)
		    retValue = -1;

#ifdef CONFIG_JSCPP_OPTINC
		/*  Check for recognition of '#ifndef' .. '#endif' construct  */
		if(spaceFlag &&
		   ifstacklevel(&(jps->JSPP_Conds)) == jps->JSPP_File->IF_CondLevel) {
		    if(!hadElse && (jps->JSPP_File->IF_Flag & IF_FLAG_INCONSTRUCT)) {
			/*  Looks better, delete NONSPC flag from this '#endif'  */
			jps->JSPP_File->IF_Flag = (jps->JSPP_File->IF_Flag &
						   ~(IF_FLAG_NONSPC | IF_FLAG_NONSPC1 |
						     IF_FLAG_INCONSTRUCT)) | IF_FLAG_CONSTRUCT;
		    } else {
			/*  Something's wrong, so give up  */
			jps->JSPP_File->IF_Flag |= IF_FLAG_NONSPC; /* is set anyway  */
		    }
		}
	    }
#endif
	    break;
	case 100:  /* **  LINE     ** */
	    if(!(mode & DPP_DO_COMMAND))
		state = 302; /*  Just ignore anything on that line  */
	    else
		state = 102;
	    break;
	case 102:
	    if(token.TK_Value == CTK_NUMBER || token.TK_Value == CTK_INUMBER) {
		auxInt = strtol(token.TK_String, NULL, 0);

	    } else {
		if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		    jsppprinterr(jps, "Invalid '#line' directive", 0);
		} else {
		    jsppprinterr(jps, "Invalid '#line' directive", 1);
		    retValue = -1;
		}

		state = 302; /*  read until end of line  */
		break;
	    }
	    auxSt = -1; /*  Preset file name	*/
	    state = 104;
	    break;
	case 104:
	    state = 106;
	    if(token.TK_Value == CTK_STRING) {
		auxSt = tokenIndex;
	    } else if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		jsppprinterr(jps, "Invalid '#line' directive", 0);
	    } else {
		jsppprinterr(jps, "Invalid '#line' directive", 1);
		retValue = -1;
		state = 302; /*  read until end of line  */
	    }
	    break;
	case 105: case 107: case 109:
	    retValue = _jsppdoline(jps, auxInt, auxSt >= 0 ?
				   jps->JSPP_Queue.TKFF_Token[auxSt].TK_String : NULL);
	    break;
#ifdef CONFIG_JSCPP_WARNING
	case 140:  /* **  WARNING  ** */
#endif
	case 110:  /* **  ERROR    ** */
	    getMode |= GCTK_READTOTERM; /*  Suppress token recognition  */
	case 120:  /* **  PRAGMA   ** */
	    auxInt = tokenIndex + 1;
	    state += 2;
	    break;
	case 112: case 122:
#ifdef CONFIG_JSCPP_WARNING
	case 142:
#endif
	    state += 2;
	    break;
	case 114: case 124: /* **  Wait till end  of line  ** */
#ifdef CONFIG_JSCPP_WARNING
	case 144:
#endif
	    break;
#ifdef CONFIG_JSCPP_WARNING
	case 143: case 145: /*  Do warning command  */
	    if(mode & DPP_DO_COMMAND)
		retValue = _jsppdowarn(jps, &(jps->JSPP_Queue.TKFF_Token[auxInt]), tokenIndex - auxInt);
	    break;
#endif
	case 113: case 115: /*  Do error command  */
	    if(mode & DPP_DO_COMMAND)
		retValue = _jsppdoerror(jps, &(jps->JSPP_Queue.TKFF_Token[auxInt]), tokenIndex - auxInt);
	    break;
	case 125: /*  Do pragma command  */
	    if(mode & DPP_DO_COMMAND)
		retValue = _jsppdopragma(jps, &(jps->JSPP_Queue.TKFF_Token[auxInt]),
					 tokenIndex - auxInt);
	    break;
#ifdef CONFIG_JSCPP_ASSERTION
	case 150:  /* **  ASSERT ** */
	case 160:  /* **  UNASSERT ** */
	    if(!(mode & DPP_DO_COMMAND)) {
		state = 302; /*  Just ignore anything on that line  */
		break;
	    }
	    state += 2;
	    break;
	case 152: case 162: /*  Look for predicate  */
	    auxInt = tokenIndex; /*  predicate  */
	    state += 2;
	    break;
	case 154: case 164: /*  look for opening brace  */
	    if(token.TK_Value == '(') {
		auxSt = tokenIndex + 1;
		state += 2;
	    }
	    break;
	case 165: /*  do unassert whole predicate  */
	    retValue = jsppdelpred(jps, jps->JSPP_Queue.TKFF_Token[auxInt].TK_String,
				   NULL);
	    break;
	case 156: case 166: /*  wait for ')'  */
	    if(token.TK_Value == ')') {
		jps->JSPP_Queue.TKFF_Token[tokenIndex].TK_Value = CTK_EOF;
		jps->JSPP_Queue.TKFF_Token[tokenIndex].TK_String = NULL;

		token = _jsppstringize(&(jps->JSPP_Queue.TKFF_Token[auxSt]), 3);

		/*  Put on queue  */
		jps->JSPP_Queue.TKFF_Index = auxSt; /*  forget original ones  */
		if((tokenIndex = jctokinsert(&(jps->JSPP_Queue), token, 0)) == -1)
		    EGOTO(error);

		auxSt = tokenIndex;
		state += 2;
	    }
	    break;
	case 159: /*  Do assert  */
	    retValue = jsppnewpred(jps, jps->JSPP_Queue.TKFF_Token[auxInt].TK_String,
				   jps->JSPP_Queue.TKFF_Token[auxSt].TK_String);
	    break;
	case 169: /*  Do unassert  */
	    retValue = jsppdelpred(jps, jps->JSPP_Queue.TKFF_Token[auxInt].TK_String,
				   jps->JSPP_Queue.TKFF_Token[auxSt].TK_String);
	    break;
#endif
	case 200:  /* **  Unknown directive  ** */
	    if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		jsppprinterr(jps, "Unknown preprocessor command \"%s\"", 0, token.TK_String);
	    } else {
		jsppprinterr(jps, "Unknown preprocessor command \"%s\"", 1, token.TK_String);
		retValue = -1;
	    }

	    state = 302;
	    break;
	case 14:   /* **  Check trailing garbage (include line)  ** */
	case 34:   /* **  Check trailing garbage (undef line)    ** */
	case 74:   /* **  Check trailing garbage (ifdef line)    ** */
	case 84:   /* **  Check trailing garbage (ifndef line)   ** */
	case 52:   /* **  Check trailing garbage (else line)     ** */
	case 62:   /* **  Check trailing garbage (endif line)    ** */
	case 106:  /* **  Check trailing garbage (line line)     ** */
#ifdef CONFIG_JSCPP_ASSERTION
	case 158:  /* **  Check trailing garbage (assert line)   ** */
#endif
	case 300:  /* **  Check trailing garbage  ** */
	    if(token.TK_Value < CTK_BLANK_START || CTK_BLANK_END < token.TK_Value) {
		if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		    if(jps->JSPP_Mode & JPPM_PRWARNING)
			jsppprinterr(jps, "Garbage in preprocessor line", 0);
		} else {
		    jsppprinterr(jps, "Garbage in preprocessor line", 1);
		    retValue = -1;
		}

		state += 2;
	    }
	    break;
	case 1: /*  Empty preprocessor line  */
	case 16: case 36: /* **  Ignore trailing garbage  ** */
	case 76: case 86: /* **  Ignore trailing garbage  ** */
	case 54: case 64: /* **  Ignore trailing garbage  ** */
	case 108:
	case 301: case 302: case 303:        /* **  Ignore trailing garbage  ** */
	    break;
	default:
	    if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		jsppprinterr(jps, "Unfinished preprocessor line", 0);
	    } else {
		jsppprinterr(jps, "Unfinished preprocessor line", 1);
		retValue = -1;
	    }

	    break;
	}
    }

    /*  Only when no output then blank completely  */
    if(!ID_IS_OUTPUT(jps)) tokenIndex = -1;

    if(retValue == 3 || retValue == 4) {
	/*  Change start token to special  */
	jps->JSPP_Queue.TKFF_Token[startIndex].TK_Value = PTK_STARTSPC;

	token.TK_Value = PTK_ENDSPC;

    } else {
	token.TK_Value = PTK_ENDIGN;
    }

    token.TK_Flags  = 0;
    token.TK_String = NULL;

    if(jctokinsert(&(jps->JSPP_Queue), token, 0) == -1)
	EGOTO(error);

    token.TK_Value = CTK_EOF;

    /*  Put trailing token in queue  */
    if(jctokinsert(&(jps->JSPP_Queue), token, 0) == -1)
	EGOTO(error);

    if(retValue == 0)
	/*  Blank out so that only newlines will survive  */
	jctokblank(&(jps->JSPP_Queue), tokenIndex);

    /*  Preserve original/blanked preprocessor line.
     *   If bit 1 in retValue is set, avoid macro expansion there.
     */
    if(_jspppushmacro(jps, NULL, jps->JSPP_Queue.TKFF_Token, NULL,
		      (retValue == 2 || retValue == 4) ? IF_FLAG_DISABLEEXPAND : 0) == -1)
	EGOTO(error);

    return retValue;
 error:
    return -1;
}
