/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Define/ undefine a macro.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#define JSCPP_INTERNAL

#include <jscpplib.h>

/*********     DEFINES                                          *********/
#ifndef PTK_EMPTY
/*  START-DEFINITIONS  */
/* ***  Additional tokens for routines  *** */
#define PTK_EMPTY        (CTK_EOF - 1) /*  Empty token -- Ignore  */

#define PTK_MACROARG0    (CTK_EOF - 2) /* This and below for macro arguments */

#define PTK_ISMACROARG(tokValue)        ((tokValue) <= PTK_MACROARG0)
#define PTK_GENMACROARG(nr)             (PTK_MACROARG0 - (nr))
#define PTK_GETMACROARG(tokValue)       (PTK_MACROARG0 - (tokValue))

/*  Special preprocessor's token  */
#define PTK_STRINGIZE    (CTK_USERTOKEN + 0)  /*  "# arg"  */
#define PTK_CONCAT       (CTK_USERTOKEN + 1)  /*  "arg1 ## arg2"  */

#define PTK_IDENTNOEXP   (CTK_USERTOKEN + 2)

#define PTK_LINEDIRSTART (CTK_USERTOKEN + 3) /*  Starts a "#line" directive */
#define PTK_SUBTOKENS    (CTK_USERTOKEN + 4) /* Use TK_Ptr field for tokens */

/* ****************************************** */
#define PTK_FIRSTINTERN  (CTK_USERTOKEN + 10)
/*
 *  To recognize a token sequence that should not
 *   be interpreted from the next stage (parser).
 */
#define PTK_STARTIGN     (CTK_USERTOKEN + 10)
#define PTK_ENDIGN       (CTK_USERTOKEN + 11)

/*  Ignore the next block in curly brackets  */
#define PTK_IGNBLOCK     (CTK_USERTOKEN + 12)

/*   To mark a special token sequence to the parser  */
#define PTK_STARTSPC     (CTK_USERTOKEN + 15)
#define PTK_ENDSPC       (CTK_USERTOKEN + 16)

#define PTK_LASTINTERN   (CTK_USERTOKEN + 19)
/* ****************************************** */

#define PTK___LINE       (CTK_USERTOKEN + 20)  /* "__LINE__"  */
#define PTK___FILE       (CTK_USERTOKEN + 21)  /* "__FILE__"  */
#ifdef CONFIG_JSCPP_INCLEVEL
# define PTK___INCLEVEL  (CTK_USERTOKEN + 22)  /* "__INCLUDE_LEVEL__"  */
#endif
#ifdef CONFIG_JSCPP_BASEFILE
# define PTK___BASEFILE  (CTK_USERTOKEN + 23)  /* "__BASE_FILE__"  */
#endif

/* ****************************************** */
#ifdef JSCPP_INTERNAL
/*  Is used in jsppcmds.c to define the preprocessor operators  */
#define PTK_USERTOKEN1   (CTK_USERTOKEN + 30)
#endif

/* ****************************************** */
/*   This is the first token value to be used outside  */
#define PTK_USERTOKEN    (CTK_USERTOKEN + 50)
/*  END-DEFINITIONS  */
#endif

#define VARIADIC_ARGNAME     "__VA_ARGS__"

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            jsppdefinemacro(jsppstream_t *jps,
					     const char *string, int mode);
Prototype int            jsppdefundefmac(jsppstream_t *jps, const char *name,
                                         const jsctoken_t *pToken, int n,
                                         long mFlags);
Prototype int            jsppsetmacflags(jsppstream_t *jps, const char *name,
                                         long mFlags);
static _jsppmacentry_t  *createMacroEntry(jsppstream_t *jps, const char *name,
                                          int nArgs, int nTokens, long mFlags);

Prototype void           jsppshowmac(jsppstream_t *jps, FILE *fp, int mode);
static void              sortMacros(_jsppmacentry_t **table, unsigned long n);
static void              printTokenInt(FILE *fp, jsctoken_t token);

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppdefinemacro(jsppstream_t *jps, const char *string, int mode)

/************************************************************************/
{
    extfile_t *eFp;
    jsctoken_t *ptr;
    const char *name;
    unsigned int start, ptrMax, ptrIndex, state, brackets, flag;

    if((eFp = efopen(string, NULL)) == NULL) {
	jsppprinterr(jps, "Unable to read definition", 1);
	return -1;
    }

    /*  Transfer string to tokens  */
    start = -1;
    ptr = NULL;
    ptrMax = ptrIndex = 0;
    name = NULL;
    brackets = 0;
    flag = FALSE;  /*  Flag for equal sign  */
    for(state = 0 ; ; ++ptrIndex) {
	jsctoken_t token;

	if(ptrIndex >= ptrMax) {
	    /*  Increase memory  */
	    ptrMax += 100;
	    if((ptr = (jsctoken_t *)realloc(ptr, ptrMax * sizeof(*ptr))) == NULL)
		EGOTO(error);
	}

	token = getCToken(eFp, jps->JSPP_Ops, 0);
	if(token.TK_Value == CTK_EOF) break;

	ptr[ptrIndex] = copyCToken(token, 0);

	switch(state) {
	case 0: /* ***  Look for macros name  *** */
	    if(token.TK_Value != CTK_IDENT)
		EGOTO(illegal);

	    name = ptr[ptrIndex].TK_String;
	    start = ptrIndex + 1;

	    state = 20;
	    break;
	case 10: /* ***  Skip macros arguments  *** */
	    if(token.TK_Value == '(')
		brackets++;
	    else if(token.TK_Value == ')')
		brackets--;

	    if(brackets == 0) state = 30;
	    break;
	case 20: /* ***  Check if macro with arguments  *** */
	    if(token.TK_Value == '(') {
		brackets++;
		state = 10;
		break;
	    }
	case 30:
	    /*   Cast equal sign to blank  */
	    if(token.TK_Value == '=') {
		/*  Cast '=' to blank  */
		ptr[ptrIndex].TK_Value = CTK_BLANKS;
		ptr[ptrIndex].TK_String = strdup(" ");
		flag = TRUE;
	    }
	    state = 50;
	    break;
	case 50: /* ***  Read until end of definition  *** */
	    break;
	}
    }

    if(name == NULL || brackets != 0) {
	jsppprinterr(jps, "Invalid definition", 0);
	goto ende;
    }

    if(ptrIndex == start && !flag) {
	static jsctoken_t OneTok = { 0, CTK_NUMBER, { "1" } };

	/*  Define at least to "1"  */
	if(jsppdefundefmac(jps, name, &OneTok, 1, MAFLAG_SYSTEMHEADER | mode))
	    EGOTO(error);

    } else {
	/*  Define this macro  */
	if(jsppdefundefmac(jps, name, &ptr[start], ptrIndex - start,
			   MAFLAG_SYSTEMHEADER | mode))
	    EGOTO(error);
    }

    /*  Free the memory  */
    for(start = 0 ; start < ptrIndex ; start++)
	freeCToken(ptr[start]);
    free(ptr);

 ende:
    if(efclose(eFp) == -1)
	EGOTO(error);

    return 0;
 illegal:
    jsppprinterr(jps, "Illegal definition", 1);
 error:
    efclose(eFp);
    return -1;
}

/*JS*********************************************************************
 *   Define the macro 'name' to the token sequence pToken[0] .. pToken[n-1].
 *    If pToken[0]=='(' the start of the pToken array are parameters for
 *    the macro. Returns 0 on success, < 0 on error.
 *************************************************************************/

int jsppdefundefmac(jsppstream_t *jps, const char *name,
                    const jsctoken_t *pToken, int n, long mFlags)

/************************************************************************/
{
    _jsppmacentry_t *entry, *oEntry;
    unsigned int hash;

    hash = jspptkhash(name, JSPP_MACTABSIZE);

    /*  Find entry where to insert macro  */
    for(oEntry = NULL, entry = jps->JSPP_Macros[hash] ; entry ;
	oEntry = entry, entry = entry->ME_Next)
	if(tkStrcmp(entry->ME_Ident, name) == 0) break;

    if(n >= 0) { /* ***  DEFINE MACRO  *** */
	const char **argL;
	_jsppmacentry_t *newEntry;
	int i, j, flag, nArgs;

	argL = NULL;
	if(n > 0 && pToken->TK_Value == '(') {
	    int argLMax, state;

	    pToken++; n--; /*  Skip '('  */

	    /*  Token with arguments, so generate argument list  */
	    argLMax = nArgs = 0;
	    for(state = 0 ; state >= 0 ; pToken++, n--) {
		if(n <= 0) EGOTO(syntError);

		/*  Skip blanks  */
		if(CTK_BLANK_START <= pToken->TK_Value &&
		   pToken->TK_Value <= CTK_BLANK_END) continue;

		if(nArgs >= argLMax) {
		    argLMax += 5;
		    if((argL = (const char **)realloc(argL, argLMax * sizeof(*argL))) == NULL)
			EGOTO(error);
		}

		switch(state) {
		case 0: /* * Look for argument, but allow ')' in case of "empty()" * */
		    if(pToken->TK_Value == ')') {
			state = -1; /*  finished, no args  */
			break;
		    } else if(pToken->TK_Value == '(') {
			/*  We have "name((arg))"  */
			mFlags |= MAFLAG_ONEARGUMENT;
			state = 4;
			break;
		    }
		    state = 2;
		case 2: /* ***  Look for argument  *** */
		case 4: /* ***  Look for lonely argument in case of "((arg))"  *** */
		    if(pToken->TK_Value != CTK_IDENT) {
			/*  Check for variadic macro (C99)  */
			if((jps->JSPP_Mode & JPPM_MACVARARG) &&
			   (pToken->TK_Value == PP_OP_ELLIPSIS ||
			    (pToken->TK_String != NULL && tkStrcmp(pToken->TK_String, "...") == 0))) {
			    mFlags |= MAFLAG_VARARG;
			    argL[nArgs++] = VARIADIC_ARGNAME;
			    state = 1;
			    break;
			}

			EGOTO(syntError);
		    }

		    argL[nArgs++] = pToken->TK_String;
		    state ^= 1; /*  flip bit 0  */

		    break;
		case 3: /* ***  Look for ',' or ')'  *** */
		    /*  Check for variadic macro (C99)  */
		    if((jps->JSPP_Mode & JPPM_MACVARARG) &&
		       (pToken->TK_Value == PP_OP_ELLIPSIS ||
			(pToken->TK_String != NULL && tkStrcmp(pToken->TK_String, "...") == 0))) {
			mFlags |= MAFLAG_VARARG;
			argL[nArgs++] = VARIADIC_ARGNAME;
			state = 1;
			break;
		    }

		    if(pToken->TK_Value == ',') {
			state ^= 1; /*  flip bit 0  */
			break;
		    }
		case 5: /* **  Look solely for ')' ** */
		    if(pToken->TK_Value == ')') {
			/*  state == 3: stop
			 *  state == 5: goto state 1
			 */
			state -= 4;
		    } else
			EGOTO(syntError);
		    break;
		case 1: /* ***  Look for final ')' in case of "((arg))" or variadic macro *** */
		    if(pToken->TK_Value != ')')
			EGOTO(syntError);

		    state = -1; /*  stop  */
		    break;
		}
	    }
	} else
	    nArgs = -1; /*  Default: No arguments  */

	/*  Remove blank tokens at start of definition  */
	while(n > 0 && CTK_BLANK_START <= pToken->TK_Value &&
	      pToken->TK_Value <= CTK_BLANK_END) {
	    pToken++;
	    n--;
	}

	if((newEntry = createMacroEntry(jps, name, nArgs,n, mFlags)) == NULL)
	    EGOTO(error);

	/* **  Copy replacement tokens  ** */
	/* Meaning of flag:
	 *   flag < 0: position -(1+flag) was a concat token '##'.
	 *   flag > 0: position  (flag-1) was a stringize token '#'.
	 */
	flag = 0;
	for(i = j = 0 ; i < n ; i++, j++) {
	    if(nArgs > 0 && pToken[i].TK_Value == CTK_IDENT) {
		int k;
		const char *idName = pToken[i].TK_String;

		/*  Scan replacement tokens and change occurences of arguments  */
		for(k = 0 ; k < nArgs ; k++)
		    if(tkStrcmp(idName, argL[k]) == 0)
			break;

		/*  If argument found, change token to replacement token  */
		if(k < nArgs) {
		    newEntry->ME_Token[j].TK_Value = PTK_GENMACROARG(k);
		    newEntry->ME_Token[j].TK_String = NULL;
		    flag = 0; /*  reset flag, since argument after '#' or '##' is ok  */
		    continue;
		}
	    }

	    /* **  Handle non blank tokens  ** */
	    if(pToken[i].TK_Value < CTK_BLANK_START ||
	       CTK_BLANK_END < pToken[i].TK_Value) {
		if(flag > 0) {
		    /*  Next token after '#' wasn't a macro argument, so restore to '#' */
		    newEntry->ME_Token[flag - 1].TK_Value = '#';
		    flag = 0;
		}

		if(pToken[i].TK_Value != '#') {
		    /*  Reset flag, we are maybe at something like "arg ## token"  */
		    flag = 0;

		} else if(flag == 0) {
		    /*  flag > 0 cannot occur, treat here only case
		     *   of '##' followed by another '#'.
		     */

		    /*  Check for string concatentation  */
		    if(pToken[i + 1].TK_Value == '#') {
			/*  Check for '##' at start of macro definition  */
			if(j != 0) {
			    newEntry->ME_Token[j].TK_Value = PTK_CONCAT;
			    newEntry->ME_Token[j].TK_String = NULL;

			    i++; /*  Delete second '#' token  */
			    flag = -1 - j; /*  Mark last token was concat token  */
			    continue; /*  skip copy section below  */

			} else if(!(jps->JSPP_Mode & JPPM_LAZYSTYLE)) {
			    jsppprinterr(jps, "\"##\" token at start of macro definition", 1);
			    EGOTO(error);
			}

		    } else {
			/*  Change token to stringize token  */
			newEntry->ME_Token[j].TK_Value = PTK_STRINGIZE;
			newEntry->ME_Token[j].TK_String = NULL;

			flag = 1 + j; /*  Mark last token was stringize token  */
			continue; /*  skip copy section below  */
		    }
		} else if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		    flag = 0;
		} else {
		    jsppprinterr(jps, "Illegal sequence of '#' and '##' tokens", 1);
		    EGOTO(error);
		}
	    }

	    /*  Copy the token and remove newlines in the token  */
	    newEntry->ME_Token[j] = copyCToken(pToken[i], 0);
	    if(newEntry->ME_Token[j].TK_String) {
		if(newEntry->ME_Token[j].TK_Value == CTK_BLANKSEOL)
		    newEntry->ME_Token[j].TK_Value = CTK_BLANKS;
		tkDelNewLines((char *)newEntry->ME_Token[j].TK_String);
	    }
	}

	if(!(jps->JSPP_Mode & JPPM_PRESSPACE) &&
	   (j == 0 ||
	    /*  We know that a final ')' doesn't need a space after it  */
	    !(newEntry->ME_Token[j - 1].TK_Value == ')'  ||
	      (CTK_BLANK_START <= newEntry->ME_Token[j - 1].TK_Value &&
	       newEntry->ME_Token[j - 1].TK_Value <= CTK_BLANK_END)))) {
	    /*  Must have at least one space after macro expansion
	     *   to avoid unwanted concatenation.
	     */
	    newEntry->ME_Token[j].TK_Value = CTK_BLANKS;
	    if((newEntry->ME_Token[j].TK_String = strdup(" ")) == NULL)
		EGOTO(error);
	    j++;
	}

	/*  Mark end of replacement tokens  */
	newEntry->ME_Token[j].TK_Value = CTK_EOF;
	newEntry->ME_Token[j].TK_String = NULL;

	/*  Sanity check  */
	if(flag) {
	    if(jps->JSPP_Mode & JPPM_LAZYSTYLE) {
		if(flag > 0) {
		    /*  Restore single '#' token  */
		    newEntry->ME_Token[flag - 1].TK_Value = '#';
		} else {
		    /*  Restore first '#' token and insert lost '#' behind it  */
		    newEntry->ME_Token[-flag - 1].TK_Value = '#';

		    for(i = j ; i >= -flag ; i--)
			newEntry->ME_Token[i + 1] = newEntry->ME_Token[i];

		    newEntry->ME_Token[-flag].TK_Value = '#';
		    newEntry->ME_Token[-flag].TK_String = NULL;
		    j++;
		}
	    } else {
		jsppprinterr(jps, "\"%s\" token at end of macro definition", 1,
			     flag > 0 ? "#" : "##");
		EGOTO(error);
	    }
	}

	if(argL) free(argL);

	/*  Run immediately through preprocessor to get actual definition  */
	if(mFlags & MAFLAG_EVALIMMEDIATE) {
	    jsctoken_t *newList;
	    int maxList;

	    jps->JSPP_File->IF_Flag |= IF_FLAG_EOF;

	    /*  Push tokens to output as a macro expansion  */
	    if(_jspppushmacro(jps, NULL, newEntry->ME_Token, NULL, 0) == -1)
		EGOTO(error);

	    newList = NULL;  maxList = 0;
	    for(j = 0 ; ; j++) {
		/*  Save new tokens, including final EOF  */
		if(j >= maxList) {
		    REALLOC_RESIZE(maxList, 50);
		    if((newList = (jsctoken_t *)realloc(newList, maxList * sizeof(*newList))) == NULL)
			EGOTO(error);
		}

		newList[j] = copyCToken(jspptkmacro(jps, NULL, 0), 0);

		if(newList[j].TK_Value == CTK_EOF) break;
	    }

	    jps->JSPP_File->IF_Flag &= ~IF_FLAG_EOF;

	    /*  Free old tokens  */
	    for(i = 0 ; newEntry->ME_Token[i].TK_Value != CTK_EOF ; i++)
		freeCToken(newEntry->ME_Token[i]);

	    /*  Resize original entry to new length (== (j+1)) and copy tokens  */
	    if((newEntry = (_jsppmacentry_t *)realloc(newEntry, sizeof(*newEntry) + j *
						      sizeof(newEntry->ME_Token[0])))
	       == NULL) EGOTO(error);

	    for(i = 0 ; i <= j ; i++) newEntry->ME_Token[i] = newList[i];

	    free(newList);
	}

#ifdef TRACETYPE_JSCPP
	if(jstracelevel(TRACETYPE_JSCPP) >= 6) {
	    FILE *tfp;

	    jstrace(TRACETYPE_JSCPP, 6,
		    "Define \"%s\" (#%d) to: ==>", newEntry->ME_Ident, nArgs);

	    tfp = jstracegetfp(TRACETYPE_JSCPP, 6);
	    for(i = 0 ; newEntry->ME_Token[i].TK_Value != CTK_EOF ; i++)
		printTokenInt(tfp, newEntry->ME_Token[i]);

	    jstrace(TRACETYPE_JSCPP, 6,
		    "<==\n");
	}
#endif

	if(entry != NULL && (entry->ME_Flags & MAFLAG_UNDEFINED)) {
	    /*  Just memorize any possible flags  */
	    newEntry->ME_Flags |= (entry->ME_Flags & ~MAFLAG_UNDEFINED);

	    /*  Substitute entry  */
	    newEntry->ME_Next = entry->ME_Next;

	    if(oEntry) oEntry->ME_Next = newEntry;
	    else       jps->JSPP_Macros[hash] = newEntry;

	    _jsppfreemacdef(jps, entry);
	} else if(entry == NULL) {
	    /*  Put new definition in global list  */
	    newEntry->ME_Next = jps->JSPP_Macros[hash];
	    jps->JSPP_Macros[hash] = newEntry;
	} else {
	    jsctoken_t *oPtr, *nPtr;

	    /*  Compare old with new definition and skip blanks  */
	    for(oPtr = entry->ME_Token, nPtr = newEntry->ME_Token ; ;
		oPtr++, nPtr++) {
		/*  Skip for next token to compare  */
		while(CTK_BLANK_START <= oPtr->TK_Value &&
		      oPtr->TK_Value <= CTK_BLANK_END) oPtr++;
		while(CTK_BLANK_START <= nPtr->TK_Value &&
		      nPtr->TK_Value <= CTK_BLANK_END) nPtr++;

		/*  Check if end of token sequence found  */
		if(oPtr->TK_Value == CTK_EOF && nPtr->TK_Value == CTK_EOF) break;

		if(nPtr->TK_Value != oPtr->TK_Value ||
		   (nPtr->TK_String && tkStrcmp(nPtr->TK_String, oPtr->TK_String))) {
		    jsppprinterr(jps, "Macro \"%s\" from '%s' l. %d redefined to different token sequence",
				 (jps->JSPP_Mode & JPPM_STRICT) ? 1 : 0,
				 name, entry->ME_File ? entry->ME_File : "",
				 entry->ME_LineNr);
		    if(jps->JSPP_Mode & JPPM_STRICT)
			/*  Abort  */
			return -1;

		    /*  Ignore  */
		    break;
		}
	    }

	    /*  Remove newly generated entry because old definition was ok  */
	    _jsppfreemacdef(jps, newEntry);
	}
    } else if(entry) { /* ***  UNDEFINE MACRO  *** */
	if(entry->ME_Flags & MAFLAG_PROTECTED) {
	    jsppprinterr(jps, "Tried to undefine protected macro \"%s\"", 1, name);
	    EGOTO(error);
	}

	/*  Remove it from the table  */
	if(oEntry) oEntry->ME_Next = entry->ME_Next;
	else       jps->JSPP_Macros[hash] = entry->ME_Next;

	_jsppfreemacdef(jps, entry);
    }

    return 0;
 syntError:
    jsppprinterr(jps, "Wrong syntax in macro definition \"%s\"", 1, name);
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppsetmacflags(jsppstream_t *jps, const char *name, long mFlags)

/************************************************************************/
{
    _jsppmacentry_t *entry;
    unsigned int hash;

    hash = jspptkhash(name, JSPP_MACTABSIZE);

    /*  Find entry where to insert macro  */
    for(entry = jps->JSPP_Macros[hash] ; entry ; entry = entry->ME_Next)
	if(tkStrcmp(entry->ME_Ident, name) == 0) break;

    if(entry == NULL) {
	/*  Create an entry that will not be used  */
	if((entry = createMacroEntry(jps, name, 0, 0,
				     MAFLAG_UNDEFINED | mFlags)) == NULL)
	    EGOTO(error);

	entry->ME_NArgs = 0;
	entry->ME_Token[0].TK_Value = CTK_EOF;
	entry->ME_Token[0].TK_String = NULL;

	entry->ME_Next = jps->JSPP_Macros[hash];
	jps->JSPP_Macros[hash] = entry;
    } else
	entry->ME_Flags |= mFlags;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static _jsppmacentry_t *createMacroEntry(jsppstream_t *jps, const char *name,
                                         int nArgs, int nTokens, long mFlags)

/************************************************************************/
{
    _jsppmacentry_t *entry;

    if((entry = (_jsppmacentry_t*)malloc(sizeof(*entry) + (nTokens + 1) *
					 sizeof(entry->ME_Token[0]))) == NULL)
	EGOTO(error);

    if((entry->ME_Ident = strdup(name)) == NULL)
	EGOTO(error);

    if(!IF_ISATEOF(jps->JSPP_File) && jps->JSPP_File->IF_Name != NULL) {
	/*  Memorize where it was defined  */
	if((entry->ME_File = strdup(jps->JSPP_File->IF_Name)) == NULL)
	    EGOTO(error);
	entry->ME_LineNr = jps->JSPP_File->IF_LineNr;
    } else {
	entry->ME_File = NULL;
	entry->ME_LineNr = -1;
    }

    entry->ME_Flags = mFlags;
    entry->ME_NArgs = nArgs;

    return entry;
 error:
    return NULL;
}

/*JS*********************************************************************
 *   Variable mode:
 *    Bit 0: Show definitions from the source file.
 *    Bit 1: Show definitions from user header files.
 *    Bit 2: Show definitions from system header files.
 *    Bit 3: Show internal definitions.
 *    Bit 4: Show assertions.
 *    Bit 5: Prefix lines with 'MACRO'.
 *************************************************************************/

void jsppshowmac(jsppstream_t *jps, FILE *fp, int mode)

/************************************************************************/
{
    _jsppmacentry_t *entry, **table;
    int maxTab, indx, i;

    /*  Collect all macros in a table  */
    table = NULL;
    maxTab = 0;
    indx = 0;
    for(i = 0 ; i < JSPP_MACTABSIZE ; i++) {
	for(entry = jps->JSPP_Macros[i] ; entry ; entry = entry->ME_Next) {
	    /*  Skip some kind of macros  */
	    if((!(entry->ME_Flags & (MAFLAG_USERHEADER | MAFLAG_SYSTEMHEADER |
				     MAFLAG_INTERNAL)) && !(mode & 1)) ||
	       ((entry->ME_Flags & MAFLAG_USERHEADER) && !(mode & 2)) ||
	       ((entry->ME_Flags & MAFLAG_SYSTEMHEADER) && !(mode & 4)) ||
	       ((entry->ME_Flags & MAFLAG_INTERNAL) && !(mode & 8)))
		continue;

	    /*  Increase memory  */
	    if(maxTab >= indx) {
		maxTab += 20;
		if((table = (_jsppmacentry_t **)
		    realloc(table, maxTab * sizeof(*table))) == NULL)
		    goto ende;
	    }
	    table[indx++] = entry;
	}
    }

    if(indx > 0) {
	/*  Now sort alphabetically  */
	sortMacros(table, indx);

	/*  And now output of table  */
	for(i = 0 ; i < indx ; i++) {
	    int l, j;

	    entry = table[i];

	    l = (mode & 32) ? fprintf(fp, "MACRO ") : 0;

	    l += fprintf(fp, "%s ", entry->ME_Ident);

	    if(entry->ME_NArgs >= 0)
		l += fprintf(fp, (entry->ME_Flags & MAFLAG_ONEARGUMENT) ?
			     "((#%d))" : "(#%d)", entry->ME_NArgs);

	    if(l < 28)
		fprintf(fp, "%*s", 25 - l, "");

	    if(entry->ME_File)
		fprintf(fp, " (l. %d \"%s\")", entry->ME_LineNr, entry->ME_File);

	    for(j = 0 ; entry->ME_Token[j].TK_Value != CTK_EOF ; j++)
		printTokenInt(fp, entry->ME_Token[j]);

	    fprintf(fp, "\n");
	}
    }

    /*printf(" #%d Macros\n",indx);*/

#ifdef CONFIG_JSCPP_ASSERTION
    if(mode & 16) jspplistpreds(jps, fp);
#endif

 ende:
    return;
}

/*JS*********************************************************************
 *
 *************************************************************************/

/*  Generated automatically with "gensort.h"  */
#define GENSORT_NAME                 sortMacros
#define GENSORT_TYPE                 _jsppmacentry_t *
#define GENSORT_KEYTYPE              char *
#define GENSORT_GETKEY(a)            a->ME_Ident
#define GENSORT_COMPAREKEYS(k1, k2)   strcmp(k1, k2) < 0
#define GENSORT_STATIC

#include <gensort.h>

/*JS*********************************************************************
 *   Prints the token, recognizing internal ones.
 *************************************************************************/

static void printTokenInt(FILE *fp, jsctoken_t token)

/************************************************************************/
{
    if(PTK_ISMACROARG(token.TK_Value))
	fprintf(fp, "$Arg_%d$", 1 + PTK_GETMACROARG(token.TK_Value));
    else if(token.TK_Value == PTK_STRINGIZE)
	fprintf(fp, "$Stringize$");
    else if(token.TK_Value == PTK_CONCAT)
	fprintf(fp, "$Concat$");
    else if(token.TK_Value == PTK___FILE)
	fprintf(fp, "$__FILE__$");
    else if(token.TK_Value == PTK___LINE)
	fprintf(fp, "$__LINE__$");
#ifdef CONFIG_JSCPP_INCLEVEL
    else if(token.TK_Value == PTK___INCLEVEL)
	fprintf(fp, "$__INCLUDE_LEVEL__$");
#endif
#ifdef CONFIG_JSCPP_BASEFILE
    else if(token.TK_Value == PTK___BASEFILE)
	fprintf(fp, "$__BASE_FILE__$");
#endif
    else if(token.TK_Value == PTK_STARTIGN)
	fprintf(fp, "$StartIgnore$");
    else if(token.TK_Value == PTK_ENDIGN)
	fprintf(fp, "$EndIgnore$");
    else if(token.TK_Value == PTK_IGNBLOCK)
	fprintf(fp, "$IgnoreBlock$");
    else if(token.TK_Value == PTK_LINEDIRSTART)
	fprintf(fp, "$StartLineDir$");
    else if(token.TK_Value == CTK_EOF)
	fprintf(fp, "$EOF$");
    else
	printCToken(fp, token);
}
