/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Routines for macro definitions and handling of "defined()"
 *              and assertions.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES                                          *********/
#ifndef MAFLAG_PROTECTED
/*  START-DEFINITIONS  */
/* **  For macro definitions  ** */
typedef struct _jsppmacentry_t _jsppmacentry_t;
struct _jsppmacentry_t {
    _jsppmacentry_t *ME_Next;
    char           *ME_Ident; /*  The macro identifier  */
    long            ME_Flags;
    int             ME_NArgs;

    /*  Position where it was defined  */
    char           *ME_File;
    int             ME_LineNr;

    /*  The actual tokens  */
    jsctoken_t      ME_Token[1];
};

#define MAFLAG_PROTECTED      (1 <<  0)  /*  Not undefinable  */
#define MAFLAG_INTERNAL       (1 <<  1)  /*  Internal macro   */
#define MAFLAG_SYSTEMHEADER   (1 <<  2)  /*  Defined in a system header file  */
#define MAFLAG_USERHEADER     (1 <<  3)  /*  Defined in a user header file  */
#define MAFLAG_ONEARGUMENT    (1 <<  4)  /*  Ignore ',' as argument separator  */
#define MAFLAG_DEFEREXPANSION (1 <<  5)  /*  Defer expansion till end  */
#define MAFLAG_DISABLED       (1 <<  6)  /*  Expansion disabled  */
#define MAFLAG_DISABLEDINFILE (1 <<  7)  /*  Expansion within file disabled  */
#define MAFLAG_EVALIMMEDIATE  (1 <<  8)  /*  Evaluate immediately on definition */
#define MAFLAG_EXPANDORIG     (1 <<  9)  /*  Expand macro in original stream  */
#define MAFLAG_EXPANDORIGFULL (1 << 10)  /*  Expand macro in original stream completely  */
#define MAFLAG_UNDEFINED      (1 << 11)  /*  Macro not defined, only flags are used  */
#define MAFLAG_VARARG         (1 << 12)  /*  Macro is variadic, last argument collects everything  */

#define MAFLAG_USED           (1 << 16) /*  Mark that it was used  */

/*  Size of macro hash table  */
#ifndef JSPP_MACTABSIZE /*  Allow changes on the fly  */
# define JSPP_MACTABSIZE 1024
#endif

#ifdef CONFIG_JSCPP_ASSERTION
typedef struct _jspppredanswer_t _jspppredanswer_t;
struct _jspppredanswer_t {
    _jspppredanswer_t *PA_Next;
    char               PA_Answer[1];
};

typedef struct _jspppredicate_t _jspppredicate_t;
struct _jspppredicate_t {
    _jspppredicate_t  *PR_Next;
    _jspppredanswer_t *PR_Answers;
    char               PR_Predicate[1];
};
#endif
/*  END-DEFINITIONS  */
#endif

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
IPrototype void          _jsppfreemacdef(jsppstream_t *jps,
                                         _jsppmacentry_t *entry);
IPrototype _jsppmacentry_t *_jsppfindmacentry(jsppstream_t *jps,
					      const char *name);

Prototype unsigned int   jspptkhash(const char *word, unsigned int tabSize);

IPrototype int           _jsppppdefass(jsppstream_t *jps, int mode,
                                       int flagAss, jctokfifo_t *oQueue);

#ifdef CONFIG_JSCPP_ASSERTION
Prototype int            jsppnewpred(jsppstream_t *jps,
				     const char *predicate,
				     const char *answer);
Prototype int            jsppdelpred(jsppstream_t *jps,
				     const char *predicate,
				     const char *answer);
static _jspppredicate_t *findPredicate(jsppstream_t *jps,
                                       const char *predicate);

Prototype void           jspplistpreds(jsppstream_t *jps, FILE *fp);
#endif

/*JS*********************************************************************
 *   Free a macro definition.
 *************************************************************************/

void _jsppfreemacdef(jsppstream_t *jps, _jsppmacentry_t *entry)

/************************************************************************/
{
    int i;

    free(entry->ME_Ident);

    if(entry->ME_File) free(entry->ME_File);

    for(i = 0 ; entry->ME_Token[i].TK_Value != CTK_EOF ; i++)
	freeCToken(entry->ME_Token[i]);

    free(entry);
}

/*JS*********************************************************************
 *   Find the corresponding macro entry for macro 'name'.
 *************************************************************************/

_jsppmacentry_t *_jsppfindmacentry(jsppstream_t *jps, const char *name)

/************************************************************************/
{
    _jsppmacentry_t *entry;

    /*  Search for macro entry  */
    for(entry = jps->JSPP_Macros[jspptkhash(name, JSPP_MACTABSIZE)] ;
	entry ; entry = entry->ME_Next)
	if(tkStrcmp(entry->ME_Ident, name) == 0)
	    break;

    return entry;
}

/*JS*********************************************************************
 *   Get hash value for word. Ignore "\\\n" sequences.
 *************************************************************************/

unsigned int jspptkhash(const char *word, unsigned int tabSize)

/************************************************************************/
{
    unsigned int hash, count;

#define HASH_BITS   32
#define HASH_SHIFT  3

    /*  Generate polynomial   */
    for(hash = count = 0 ; *word ; word++) {
	/*  Skip "\\\n"  */
	if(*word == '\\' && word[1] == '\n') {
	    word++;
	    continue;
	}
	hash = (hash << HASH_SHIFT) | (hash >> (HASH_BITS - HASH_SHIFT));
	hash += *(const unsigned char *)word;
	count++;
    }

    return (hash ^ count) % tabSize;
}
#undef HASH_BITS
#undef HASH_SHIFT

/*JS*********************************************************************
 *   Handle special preprocessor macro "defined(name)" (flagAss FALSE) or
 *    assertion (flagAss TRUE): Expand it to "0" or "1".
 *************************************************************************/

int _jsppppdefass(jsppstream_t *jps, int mode, int flagAss, jctokfifo_t *oQueue)

/************************************************************************/
{
    jsctoken_t token;
    static jsctoken_t TrueFalseTk[] = {
	CTOKEN_GENSTATIC(CTK_NUMBER, NULL), /*  Will be filled in  */
#define INDEX  0
	CTOKEN_GENSTATIC(CTK_EOF, NULL)
    };
    int state;
    char *string;
#ifdef CONFIG_JSCPP_ASSERTION
    static jctokfifo_t AssertQueue;

    if(flagAss) jctokclear(&AssertQueue);
#endif

    if(mode & EXPM_SAVE_ORIG) {
#ifdef CONFIG_JSCPP_ASSERTION
	if(flagAss) {
	    /*  Put introducer in FIFO queue  */
	    token.TK_Value = '#'; token.TK_String = NULL;
	} else
#endif
	    {
		/*  Put macro's name in FIFO queue  */
		token.TK_Value = CTK_IDENT; token.TK_String = "defined";
	    }

	if(jctokinsert(oQueue, token, 0) == -1)
	    EGOTO(error);
    }

    string = "0";
    state =
#ifdef CONFIG_JSCPP_ASSERTION
	flagAss ? 0 :
#endif
	2;

    while(state >= 0) {
#ifdef CONFIG_JSCPP_ASSERTION
	_jspppredicate_t *pred;
#endif
	int tokenIndex;

	/*  Put token in FIFO queue  */
	if(mode & EXPM_SAVE_ORIG) {
	    jsctoken_t oToken;

	    token = _jspptkbasic2(jps, &oToken, mode);
	    if(oToken.TK_Value != PTK_EMPTY &&
	       jctokinsert(oQueue, oToken, 0) == -1)
		EGOTO(error);

	} else {
	    token = _jspptkbasic2(jps, NULL, mode);
	}

	/*  Skip non-newline blanks  */
	if(CTK_BLANK_START2 <= token.TK_Value && token.TK_Value <= CTK_BLANK_END
#ifdef CONFIG_JSCPP_ASSERTION
	   && state != 8
#endif
	   )
	    continue;

	switch(state) {
#ifdef CONFIG_JSCPP_ASSERTION
	case 0:
	    /*  Find the predicate  */
	    pred = findPredicate(jps, token.TK_String);
	    state = 2;
	    break;
#endif
	case 2:
	    if(token.TK_Value == '(') {
#ifdef CONFIG_JSCPP_ASSERTION
		if(flagAss) {
		    state = 8;
		} else
#endif
		    state = 4;
		break;
#ifdef CONFIG_JSCPP_ASSERTION
	    } else if(flagAss) {
		jsppprinterr(jps, "Wrong syntax of assertion: miss opening brace", 1);
		EGOTO(error);
#endif
	    } else {
		/*  We have to break immediately after we check the identifier  */
		state = -3;
	    }
	case 4:
	    if(token.TK_Value != CTK_IDENT) {
		jsppprinterr(jps, "Wrong syntax of defined(name): no identifier", 1);
		EGOTO(error);
	    }

	    /*  Check if identifier is defined  */
	    {
		_jsppmacentry_t *entry;

		if((entry = _jsppfindmacentry(jps, token.TK_String)) != NULL &&
		   !(entry->ME_Flags & MAFLAG_UNDEFINED)) {
		    entry->ME_Flags |= MAFLAG_USED;
		    string = "1";
		}
	    }

	    state += 2;
	    break;
	case 6:
	    if(token.TK_Value != ')') {
		jsppprinterr(jps, "Wrong syntax of defined(name): miss closing bracket", 1);
		EGOTO(error);
	    }
	    state = -1;
	    break;
#ifdef CONFIG_JSCPP_ASSERTION
	case 8:
	    if(token.TK_Value == ')') {
		if(pred) {
		    _jspppredanswer_t *ans;

		    AssertQueue.TKFF_Token[tokenIndex + 1].TK_Value = CTK_EOF;
		    AssertQueue.TKFF_Token[tokenIndex + 1].TK_String = NULL;

		    token = _jsppstringize(&AssertQueue.TKFF_Token[0], 3);

		    for(ans = pred->PR_Answers ; ans ; ans = ans->PA_Next)
			if(tkStrcmp(token.TK_String, ans->PA_Answer) == 0) {
			    string = "1";
			    break;
			}
		}

#ifdef TRACETYPE_JSCPP
		jstrace(TRACETYPE_JSCPP, 6,
			"Using a predicate (=%s)", 0, string);
#endif
		state = -1;

	    } else if(token.TK_Value == CTK_BLANKSEOL) {
		jsppprinterr(jps, "Wrong syntax of assertion: no closing brace", 1);
		EGOTO(error);

	    } else if((tokenIndex = jctokinsert(&AssertQueue, token, 0)) == -1)
		EGOTO(error);

	    break;
#endif
	}
    }

    /*  Push simple token sequence  */
    TrueFalseTk[INDEX].TK_String = string;
    if(_jspppushmacro(jps, NULL, TrueFalseTk, NULL, 0) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}
#undef INDEX

#ifdef CONFIG_JSCPP_ASSERTION
/*JS*********************************************************************
 *
 *************************************************************************/

int jsppnewpred(jsppstream_t *jps, const char *predicate,
                const char *answer)

/************************************************************************/
{
    _jspppredicate_t *pred;
    _jspppredanswer_t *ans;

    if((pred = findPredicate(jps, predicate)) == NULL) {
	if((pred = (_jspppredicate_t *)
	    malloc(sizeof(*pred) + strlen(predicate))) == NULL)
	    return -1;

	strcpy(pred->PR_Predicate, predicate);
	pred->PR_Next = jps->JSPP_Predicates;
	jps->JSPP_Predicates = pred;
	pred->PR_Answers = NULL;
    }

    if((ans = (_jspppredanswer_t *)malloc(sizeof(*ans) + strlen(answer))) == NULL)
	return -1;

    strcpy(ans->PA_Answer, answer);
    ans->PA_Next = pred->PR_Answers;
    pred->PR_Answers = ans;

#ifdef TRACETYPE_JSCPP
    jstrace(TRACETYPE_JSCPP, 6,
	    "New assertion %s(%s)", 0, predicate, answer);
#endif

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsppdelpred(jsppstream_t *jps, const char *predicate, const char *answer)

/************************************************************************/
{
    _jspppredicate_t *pred, *old, *next;

    if(predicate == NULL) {
	/*  Delete all predicates  */
	while((pred = jps->JSPP_Predicates))
	    if(jsppdelpred(jps, pred->PR_Predicate, NULL))
		return -1;

	return 0;
    }

    for(old = NULL, pred = jps->JSPP_Predicates ; pred ; old = pred,
	    pred = next) {
	next = pred->PR_Next;
	if(tkStrcmp(pred->PR_Predicate, predicate) == 0) {
	    _jspppredanswer_t *ans, *ans2;

	    if(answer == NULL) {
		/*  Delete whole predicate  */
		if(old)
		    old->PR_Next = pred->PR_Next;
		else
		    jps->JSPP_Predicates = pred->PR_Next;

		for(ans = pred->PR_Answers ; ans ; ans = ans2) {
		    ans2 = ans->PA_Next;
		    free(ans);
		}

		free(pred);
	    } else {
		for(ans2 = NULL, ans = pred->PR_Answers ; ans ; ans2 = ans,
			ans = ans->PA_Next) {
		    if(tkStrcmp(ans->PA_Answer, answer) == 0) {
			if(ans2)
			    ans2->PA_Next = ans->PA_Next;
			else
			    pred->PR_Answers = ans->PA_Next;
			free(ans);
			break;
		    }
		}
	    }

#ifdef TRACETYPE_JSCPP
	    if(answer)
		jstrace(TRACETYPE_JSCPP, 6,
			"Deleted predicate %s(%s)", 0, predicate, answer);
	    else
		jstrace(TRACETYPE_JSCPP, 6,
			"Deleted predicate %s", 0, predicate);
#endif

	    if(answer == NULL)
		break;
	}
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static _jspppredicate_t *findPredicate(jsppstream_t *jps,
                                       const char *predicate)

/************************************************************************/
{
    _jspppredicate_t *pred;

    for(pred = jps->JSPP_Predicates ; pred ; pred = pred->PR_Next)
	if(tkStrcmp(pred->PR_Predicate, predicate) == 0) break;

    return pred;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jspplistpreds(jsppstream_t *jps, FILE *fp)

/************************************************************************/
{
    _jspppredicate_t *pred;

    for(pred = jps->JSPP_Predicates ; pred ; pred = pred->PR_Next) {
	_jspppredanswer_t *ans;

	for(ans = pred->PR_Answers ; ans ; ans = ans->PA_Next)
	    fprintf(fp, "#%s(%s)\n", pred->PR_Predicate, ans->PA_Answer);
    }
}
#endif
