/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : General subroutines.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
#include <stdarg.h>

#define JSCPP_INTERNAL
#include <jscpplib.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define jspptkprint(jps, fp, token)           jspptkprint2(jps, fp, token, -2)
/*  END-DEFINITIONS  */

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
Prototype int            jspptkprint2(jsppstream_t *jps, FILE *fp,
				      jsctoken_t token, int currLine);

Prototype void           jsppseterrcode(int errnr, int erroff, const char *string,
					jctokfifo_t *queue, int indx);

Prototype void		 jsppprinterr(jsppstream_t *jps, char *message, int mode, ...);

IPrototype jsctoken_t    _jsppstringize(jsctoken_t *pToken, int mode);
static int               doBackSlash(char *buffer, const char *input, int mode);

IPrototype jsctoken_t    _jsppgetnnewlines(int n);

IPrototype void          _jspprmlinecont(char *string);

Prototype int            jsppflushtokens(jsppstream_t *jps, jctokfifo_t *queue,
					 FILE *fp, int mode);

Prototype int            jsppisusrinclude(jsppstream_t *jps, int mode);

Prototype const char    *jspptok2string(jsctoken_t token);

/*JS*********************************************************************
 *   Print token, probably avoiding a lot of space ones. If currLine == -2,
 *    we use jsppgetfilepos to obtain our current position. If it is
 *    positive, we use it directly. If it is -1, we refrain from trying
 *    to obtain the file position.
 *************************************************************************/

int jspptkprint2(jsppstream_t *jps, FILE *fp, jsctoken_t token,
		 int currLine)

/************************************************************************/
{
    int len;

    if(token.TK_Value == PTK_LINEDIRSTART) {
	/*  Start with a '\n' to avoid line clash in case an
	 *   include file had no final '\n'.
	 */
	len = fprintf(fp, (jps->JSPP_Ignored >= 0 &&
			   !(jps->JSPP_Mode & JPPM_PRESSPACE)) ?
		      "\n#line " : "#line ");

	/*  Any collected newlines are now obsolete  */
	jps->JSPP_NLLeftOver = jps->JSPP_NLLeftOver2 = jps->JSPP_Ignored = 0;

	return len;
    }

    /*  If requested, don't do any optimization  */
    if(jps->JSPP_Mode & JPPM_PRESSPACE)
	return printCToken(fp, token);

    /*  If token contains solely spaces, collect first newlines  */
    if(CTK_BLANK_START <= token.TK_Value && token.TK_Value <= CTK_BLANK_END &&
       (!(token.TK_Value == CTK_COMMENT || token.TK_Value == CTK_CPPCOMMENT) ||
	!(jps->JSPP_Mode & JPPM_PRESCOMM))) {
	const char *s;

	if((s = token.TK_String)) {
	    while(*s) {
		if(*s == '\\' && s[1] == '\n') {
		    jps->JSPP_NLLeftOver2++;
		    s += 2;
		} else {
		    if(*s == '\n')
			jps->JSPP_NLLeftOver++;
		    s++;
		}
	    }

	    if(jps->JSPP_NLLeftOver && jps->JSPP_NLLeftOver2) {
		/*  "\\\n" obsolete if token contains real '\n'  */
		jps->JSPP_NLLeftOver += jps->JSPP_NLLeftOver2;
		jps->JSPP_NLLeftOver2 = 0;
	    }
	}

	if(jps->JSPP_Ignored >= 0)
	    jps->JSPP_Ignored++; /*  count number of skipped tokens  */

	return 0;
#ifdef CONFIG_JSCPP_OPTLINES
    } else if((jps->JSPP_Mode & (JPPM_LINEDIR|JPPM_SAVENEWLN)) ==
	      (JPPM_LINEDIR|JPPM_SAVENEWLN) && jps->JSPP_NLLeftOver >= 10 &&
	      /*  Either we have it already or we are allowed to use getfilepos  */
	      (currLine >= 0 ||
	       (currLine == -2 &&
		(currLine = jsppgetfilepos(jps, NULL, 0)) >= 0))) {
	/* Print line statement for line before current token (is this correct?) */
	fprintf(fp, "\n#line %d\n", currLine);
#endif
    } else if(jps->JSPP_NLLeftOver2) { /*  Print missing escaped newlines  */
	int i = jps->JSPP_NLLeftOver2;

	while(i-- > 0) {
	    putc('\\', fp);
	    putc('\n', fp);
	}
	putc(' ', fp); /*  Make sure we don't get a token clash accidentally  */
    } else if(jps->JSPP_NLLeftOver) { /*  Print missing newlines  */
	int i = jps->JSPP_NLLeftOver;

	while(i-- > 0)
	    putc('\n', fp);
    } else if(jps->JSPP_Ignored > 0) {
	/*  Prevent accidental concatenation of tokens, but avoid
	 *   space at start of file.
	 */
	putc(' ', fp);
    }

    if((len = printCToken(fp, token)) == -1)
	return -1;

    len += jps->JSPP_NLLeftOver;
    jps->JSPP_NLLeftOver = jps->JSPP_NLLeftOver2 = jps->JSPP_Ignored = 0;

    return len;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsppseterrcode(int errnr, int erroff, const char *string,
		    jctokfifo_t *queue, int indx)

/************************************************************************/
{
    int pos;

    if(string != NULL)
	JSSETERROR2(errnr, erroff, string);
    else
	JSSETERROR(errnr, erroff);

    if(indx == -1 || queue == NULL)
	return;

    if((pos = jctokindfpos(queue)) == -1)
	return;

    indx += pos;
    JSSETERRFILE(queue->TKFF_Token[indx].TK_String,
		 queue->TKFF_Token[indx].TK_Value);
}

/*JS*********************************************************************
 *   Print an error message to stderr if mode is != 0, otherwise print
 *    only a warning message.
 *************************************************************************/

void jsppprinterr(jsppstream_t *jps, char *message, int mode, ...)

/************************************************************************/
{
    va_list ap;
    const char *file;
    char buffer[2048];
    int off, lineNr;

    lineNr = jsppgetfilepos(jps, &file, 0);

    off = sprintf(buffer, "%s", (mode & 1) ? "ERROR" : "WARNING");
    if(file)
	off += sprintf(buffer + off, " l. %d \"%s\"", lineNr, file);
    buffer[off++] = ':';
    buffer[off++] = ' ';

    /*  Print error message  */
    va_start(ap, mode);
    off += vsprintf(buffer + off, message, ap);
    va_end(ap);

    buffer[off++] = '\n';
    buffer[off++] = '\0';

    if(jps->JSPP_PrintError)
	jps->JSPP_PrintError(jps->JSPP_User, buffer);

    return;
}

/*JS*********************************************************************
 *   pToken is a sequence of tokens ended with a CTK_EOF. Handle ANSI's
 *    stringize operator '#': Convert the tokens to a string. Remove
 *    spaces from start and end of the sequence.
 *     Bit 0 of mode: Cast sequences of newlines to one blank.
 *     Bit 1 of mode: Delete start and trailing blanks.
 *     Bit 2 of mode: Precede quotes in strings by backslash.
 *************************************************************************/

jsctoken_t _jsppstringize(jsctoken_t *pToken, int mode)

/************************************************************************/
{
    jsctoken_t token;
    static char *Buffer;
    static int MaxB;
    static const char *Blank = " ", *Empty = "";
    const char *string;
    int indB;

    /*  Print all tokens to a buffer  */
    string = (mode & 2) ? Blank : Empty;
    for(indB = 0 ; pToken->TK_Value != CTK_EOF ; pToken++) {
	int len;

	/*	Skip special tokens  */
	if(PTK_FIRSTINTERN <= pToken->TK_Value &&
	   pToken->TK_Value <= PTK_LASTINTERN) continue;

	/*	Handle blank tokens  */
	if(CTK_BLANK_START <= pToken->TK_Value &&
	   pToken->TK_Value <= CTK_BLANK_END) {
	    /*  Sequence of blanks are cast to one and
	     *   start and trailers are removed.
	     */
	    if(string == Blank)
		string = Empty;
	    else if(string != Empty && (mode & 1))
		string = Blank;
	    else if(string != Empty)
		string = pToken->TK_String;
	} else
	    string = pToken->TK_String;

	if(string) {
	    const char *s;

	    len = strlen(string);

	    /*  Count '\\' in string  */
	    for(s = string ; (s = strchr(s, '\\')) ; s++) ++len;
	} else
	    len = 1;

	/*  Allocate enough space for worst case  */
	if((indB + len + 5) > MaxB) {
	    MaxB = indB + len + 5;
	    if((Buffer = (char *) realloc(Buffer, MaxB)) == NULL)
		EGOTO(error);
	}

	if(pToken->TK_Value == CTK_STRING || pToken->TK_Value == CTK_WSTRING) {
	    indB += doBackSlash(Buffer + indB, string, (mode & 4) | (8|1));
	} else if(pToken->TK_Value == CTK_CHARCONST ||
		  pToken->TK_Value == CTK_WCHARCONST) {
	    /*  We don't need a '\\' in front of '\''  */
	    indB += doBackSlash(Buffer + indB, string, (8|2));
	} else if(string) {
	    indB += doBackSlash(Buffer + indB, string, 0);
	} else {
	    Buffer[indB++] = pToken->TK_Value;
	}
    }

    if(mode & 2) {
	/*  Remove blanks from trailer of string  */
	while(indB > 0 && Buffer[indB - 1] == ' ')
	    indB--;
    }

    /*  Return string token  */
    token.TK_Value = CTK_STRING;

    if(Buffer) {
	Buffer[indB] = '\0';
	token.TK_String = Buffer;
    } else {
	token.TK_String = "";
    }

    return token;
 error:
    token.TK_Value = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   Copy input to buffer, return number of copied characters.
 *    Parameter mode:
 *      Bit 1/0 : 00  No start/end character.
 *                01  Use double quotes as start and end character.
 *                10  Use single quotes as start and end character.
 *      Bit 2: If set, precede start and end character by \, too.
 *      Bit 3: Precede each \ by \.
 *    Backslash-newline is removed completely.
 *************************************************************************/

static int doBackSlash(char *buffer, const char *input, int mode)

/************************************************************************/
{
    char *s;

    s = buffer;

    if(mode & 3) {
	if(mode & 4) *s++ = '\\';
	*s++ = (mode & 1) ? '"' : '\'';
    }

    while(*input) {
	if(*input == '\\') {
	    if(input[1] == '\n') {
		input += 2;
		continue;
	    }
	    if(mode & 8) *s++ = '\\'; /*  precede it  */
	}
	*s++ = *input++;
    }

    if(mode & 3) {
	if(mode & 4) *s++ = '\\';
	*s++ = (mode & 1) ? '"' : '\'';
    }

    *s = '\0';

    return s - buffer;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsctoken_t _jsppgetnnewlines(int n)

/************************************************************************/
{
    jsctoken_t token;
    static char *BlankBuffer;
    static int   BlankMax;

    /*  Increase buffer  */
    if(n >= BlankMax) {
	BlankMax = n + 1;
	if((BlankBuffer = (char *)realloc(BlankBuffer, BlankMax)) == NULL) /*DBG*/
	    EGOTO(error);
    }

    memset(BlankBuffer, '\n', n);
    BlankBuffer[n] = '\0';
    token.TK_Value  = CTK_BLANKSEOL;
    token.TK_String = BlankBuffer;

    return token;
 error:
    token.TK_Value = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
 *   Removes "\\\n" in the string.
 *************************************************************************/

void _jspprmlinecont(char *string)

/************************************************************************/
{
    char *dest;

    for(dest = string ; *string ; ) {
	if(*string == '\\' && string[1] == '\n') {
	    string += 2;
	} else
	    *dest++ = *string++;
    }
    *dest = '\0';
}

/*JS*********************************************************************
 *   Flush all tokens from the queue. If bit 0 of mode is set, flush
 *    original tokens. If bit 1 is set, do not clear the queue at the
 *    end.
 *    jps might be NULL.
 *************************************************************************/

int jsppflushtokens(jsppstream_t *jps, jctokfifo_t *queue, FILE *fp, int mode)

/************************************************************************/
{
    jsctoken_t token;
    static jsctoken_t EOFTok = CTOKEN_GENSTATIC(CTK_EOF, NULL);
    int ignLevel, indorig, step;

#ifdef TRACETYPE_JSCPP
    {
	FILE *tfp;

	if((tfp = jstracegetfp(TRACETYPE_JSCPP, 6)))
	    jctokshowqueue(tfp, queue);
    }
#endif

    step = jctokgetstep(queue);

    indorig = jctokindorig(queue);
    if((mode & 1) && indorig != -1) {
	for(;;) {
	    int line, ind;

	    line = jctokgetfpos(queue, NULL);
	    if((ind = jctokgetindex(queue)) == -1)
		EGOTO(error);

	    if(ind == -2)
		break;

	    token = queue->TKFF_Token[ind + indorig];

	    /*  Print original tokens to file  */
	    if(token.TK_Value == PTK_EMPTY ||
	       (PTK_FIRSTINTERN <= token.TK_Value &&
		token.TK_Value <= PTK_LASTINTERN))
		continue;

	    if(jps) {
		if(jspptkprint2(jps, fp, token, line) == -1)
		    EGOTO(error);
	    } else {
		if(printCToken(fp, token) == -1)
		    EGOTO(error);
	    }
	}
    } else if(jps && (jps->JSPP_Mode & JPPM_FLAG_SEENDEFEREDMAC)) {
#if 0 /*  FIXME: Test it before use  */
	jsqueuestream_t *jsqs;

	/*   Process all tokens through preprocessor again, since
	 *   there was a deferred token around.
	 */

	if((jsqs = jsqscreatequeuestream(queue, NULL, NULL)) == NULL ||
	   jspppushppstream(jps, jsqs, 1) == -1)
	    EGOTO(error);

#else
	jps->JSPP_File->IF_Flag |= IF_FLAG_EOF;

	if(step > 1) {
	    int i, j;

	    i = queue->TKFF_Start + step;
	    j = queue->TKFF_Start + 1;
	    for( ; i < queue->TKFF_Index ; i += step, ++j)
		queue->TKFF_Token[j] = queue->TKFF_Token[i];
	    queue->TKFF_Index = j;
	}

	/*  Put trailing token on queue and push
	 *   tokens to output as a macro expansion
	 */
	if(jctokinsert(queue, EOFTok, 0) == -1 ||
	   jspppushtokens(jps, &queue->TKFF_Token[queue->TKFF_Start]) == -1)
	    EGOTO(error);
#endif

	ignLevel = 0;
	while((token = jspptkmacro(jps, NULL, EXPM_DEFEREXPANSION)).TK_Value
	      != CTK_EOF) {
	    /*  Don't print special token sequence  */
	    if(token.TK_Value == PTK_STARTSPC) {
		ignLevel++;
		continue;
	    }

	    if(token.TK_Value == PTK_ENDSPC) {
		ignLevel--;
		continue;
	    }

	    /*  Print preprocessed tokens to file  */
	    if(token.TK_Value == PTK_EMPTY || ignLevel > 0 ||
	       (PTK_FIRSTINTERN <= token.TK_Value &&
		token.TK_Value <= PTK_LASTINTERN))
		continue;

	    /*  Unfortunately we have lost our file position information  */
	    (void)jspptkprint2(jps, fp, token, -1);
	}

#if 0 /*  FIXME: see above  */
	jsqsfree(jsqs);
#else
	jps->JSPP_File->IF_Flag &= ~IF_FLAG_EOF;
#endif

	jps->JSPP_Mode &= ~JPPM_FLAG_SEENDEFEREDMAC;

    } else {
	/* **  Simply print tokens from queue  ** */
	if(jctokinsert(queue, EOFTok, 0) == -1)
	    EGOTO(error);

	for(ignLevel = 0 ; ; ) {
	    int line;

	    line = jctokgetfpos(queue, NULL);
	    if((token = jctokget(queue)).TK_Value == CTK_EOF)
		break;

	    /*  Don't print special token sequence  */
	    if(token.TK_Value == PTK_STARTSPC) {
		ignLevel++;
		continue;
	    } else if(token.TK_Value == PTK_ENDSPC) {
		ignLevel--;
		continue;
	    }

	    /*  Print preprocessed tokens to file  */
	    if(token.TK_Value == PTK_EMPTY || ignLevel > 0 ||
	       (PTK_FIRSTINTERN <= token.TK_Value &&
		token.TK_Value <= PTK_LASTINTERN))
		continue;

	    if(jps)
		(void)jspptkprint2(jps, fp, token, line);
	    else
		printCToken(fp, token);
	}
    }

    /*  Clear the queue  */
    if(!(mode & 2)) jctokclear(queue);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Determine if we are currently within a user include file or if
 *    we currently expand a macro seen therein.
 *************************************************************************/

int jsppisusrinclude(jsppstream_t *jps, int mode)

/************************************************************************/
{
    jsppfile_t *iPtr;

    for(iPtr = jps->JSPP_File ; iPtr ; iPtr = iPtr->IF_Next) {
	/*  Doesn't work, but just keep it  */
	if((mode & 2) && (iPtr->IF_Flag & IF_FLAG_ORIGTOK))
	    break;

	if((mode & 1) && (iPtr->IF_Flag & IF_FLAG_LINETOKENS))
	    return 0;

	if(iPtr->IF_Flag & IF_SYSINCLUDE)
	    return 0;
    }

    return 1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

const char *jspptok2string(jsctoken_t token)

/************************************************************************/
{
    if(PTK_ISMACROARG(token.TK_Value)) {
	static char Buffer[20];

	sprintf(Buffer, "PTK_MACRARG%d", PTK_GETMACROARG(token.TK_Value));
	return Buffer;
    }

#define GEN_CASEVALUE(name)    case name: return #name
    switch(token.TK_Value) {
	GEN_CASEVALUE(PTK_EMPTY);

	GEN_CASEVALUE(PTK_STRINGIZE);
	GEN_CASEVALUE(PTK_CONCAT);

	GEN_CASEVALUE(PTK_IDENTNOEXP);
	GEN_CASEVALUE(PTK_LINEDIRSTART);
	GEN_CASEVALUE(PTK_SUBTOKENS);

	GEN_CASEVALUE(PTK_STARTIGN);
	GEN_CASEVALUE(PTK_ENDIGN);
	GEN_CASEVALUE(PTK_IGNBLOCK);
	GEN_CASEVALUE(PTK_STARTSPC);
	GEN_CASEVALUE(PTK_ENDSPC);

	GEN_CASEVALUE(PTK___LINE);
	GEN_CASEVALUE(PTK___FILE);
#ifdef PTK___INCLEVEL
	GEN_CASEVALUE(PTK___INCLEVEL);
#endif
#ifdef PTK___BASEFILE
	GEN_CASEVALUE(PTK___BASEFILE);
#endif
    }
#undef GEN_CASEVALUE

    return dbgNameCToken(token);
}
