/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Separate blocks from functions
*
*              Copyright (C) 2004-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

#include <jsbitutil.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_FCTSEP      "fctsep"
/*  END-DEFINITIONS  */

typedef struct {
  int Index;
  int Type;
  const char *OrigStmt;
  int Value;
} blockentry_t;

#define VARTYPE_IGNORE       0
#define VARTYPE_BLOCKLOCAL   1
#define VARTYPE_PASSVALUE    2
#define VARTYPE_PASSREF      3

typedef struct {
  int Type;
  const char *Name;
  const char *NewName;
} fctsepvar_t;

#define SAUXMODE_SCANLEAVES    1
#define SAUXMODE_NEEDCOMMA     2 /* when generating the function call/header */

typedef struct {
  int Mode;

  jsparsstream_t *JPA;
  jctokfifo_t *Queue;
  FILE *Fp;
  const char *FunctionName;
  jsparscnode_t *Function, *Body;

  int Start, End; /*  Start and end index of block to separate  */

  blockentry_t *BlockEntries;
  int NBlockEntries, MBlockEntries;

  fctsepvar_t *VarDefs;
  int NVarDefs, MVarDefs;
} fctsepaux_t;

#define FLAG_RD_BEFORE       JVDF_FUSER
#define FLAG_WR_BEFORE      (JVDF_FUSER << 1)
#define FLAG_RD_INSIDE      (JVDF_FUSER << 2)
#define FLAG_WR_INSIDE      (JVDF_FUSER << 3)
#define FLAG_RD_AFTER       (JVDF_FUSER << 4)
#define FLAG_WR_AFTER       (JVDF_FUSER << 5)

#define FLAG_ALL            (FLAG_RD_BEFORE | FLAG_WR_BEFORE | \
			     FLAG_RD_INSIDE | FLAG_WR_INSIDE | \
			     FLAG_RD_AFTER  | FLAG_WR_AFTER)

#define PRAGMA_NAME_SEPARATE     "separate"

#define PRAGMA_CLAUSE_LOCAL      "local"
#define PRAGMA_CLAUSE_VALUE      "value"
#define PRAGMA_CLAUSE_REFERENCE  "reference"

#define PATTERN_GENHEADER    "/* ***  Generated by JSCPP  *** */"
#define PATTERN_GENHEADER2   "/* ***  Generated by JSCPP, separated from '%s'  *** */"

#define PATTERN_SELTYPE         "int"
#define PATTERN_SELPARAM        "sel_%s"
#define PATTERN_RETVAR          "ret_%s"
#define PATTERN_FUNCLABEL       "label_%s"

#define PATTERN_RETSTMT         "return;"
#define PATTERN_RETSTMTVAL      "return %d;"

#define PATTERN_PTRNAME         "p%s"
#define PATTERN_PTRNAME_TOK     "p%o"

#define ERRORLABEL_START        "err"
#define ERRORLABEL_START_LEN    3

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doFctSep(jsparsenviron_t *ppenv,
				  const char *names[],
				  int nnames, FILE *fpOut, int mode);

static int               parseFctSep(jsparsstream_t *jpa, FILE *fpOut);
static int               doProcFctSep(jsparsstream_t *jpa,
				      jsparscnode_t *node, FILE *fp,
				      jctokfifo_t *queue);
static void              freesaux(fctsepaux_t *saux, int flag);
Prototype const char    *getFunctionName(jsparscnode_t *node,
					 jctokfifo_t *queue);
static int               doProcStmt(fctsepaux_t *saux, jsparscnode_t *node);
static int               doSeparateBlock(fctsepaux_t *saux, int start, int end,
					 jsparscnode_t *block);
static int               readIdents(fctsepaux_t *saux, int type,
				    int start, int end);
static int               addVarDef(fctsepaux_t *saux, int type,
				   const char *name,
				   const char *newName);
static int               scan4blockentryexit(fctsepaux_t *saux, int start);
static int               addblockentryexit(fctsepaux_t *saux, int indx,
					   int type);
static int               findblockentryexit(fctsepaux_t *saux, int indx);

static int               doSepVariable(jsparsstream_t *jpa, jsparscnode_t *node,
				       jctokfifo_t *queue, jsparsvartab_t *vTab,
				       jsparsvardef_t *vDef, long flags,
				       void *user);
static int               checkReturnType(fctsepaux_t *saux,
					 jsparscnode_t *node);
static int               genCalls(fctsepaux_t *saux,
				  jstrbuff_t *headerBuff,
				  jstrbuff_t *callBuff, jstrbuff_t *varBuff);
static int               getVarType(fctsepaux_t *saux, jsparsvardef_t *vDef,
				    const char **pName);
static int               substVariable(jsparsstream_t *jpa,
				       jsparscnode_t *node,
				       jctokfifo_t *queue,
				       jsparsvartab_t *vTab,
				       jsparsvardef_t *vDef,
				       long flags, void *user);

static int               printBlock(jsparsstream_t *jpa, FILE *fp,
				    jctokfifo_t *queue, jsparscnode_t *node,
				    const char *switchStmt,
				    int offset);

/*********     GLOBAL / STATIC VARIABLES                        *********/
jstrbuff_t HeaderBuffer, CallBuffer, BlockVarBuffer, SwitchBuffer;

/*JS*********************************************************************
*
*************************************************************************/

int doFctSep(jsparsenviron_t *ppenv, const char *names[], int nnames,
	     FILE *fpOut, int mode)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < nnames ; ++i) {
    jsparsstream_t *jpa;
    jsppstream_t *jps;
    int ret, pmode;

    pmode = JPAM_FVARTABLES | JPAM_FSTMTFLOW | JPAM_FFLOWBLOCKEND;
    pmode |= JPAM_FGETORIG | JPAM_FPRINTORIG;

    /*  Get meaningful warning messages  */
    pmode |= JPAM_FGETFPOS;

    if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
      EGOTO(error);

    jps = jsqsgetppstream(jpa->JSPA_Stream);
    jps->JSPP_Mode |= JPPM_PRESSPACE;

    ret = parseFctSep(jpa, fpOut);

    if(closeParserStream(jpa, ret != 0 ? 2 : 0) == -1)
      EGOTO(error);

    if(ret) EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int parseFctSep(jsparsstream_t *jpa, FILE *fpOut)

/************************************************************************/
{
  jctokfifo_t *queue;

  queue = jsparsgetqueue(jpa);
  for(;;) {
    jsparscnode_t *node;

    if(jsparsget(jpa, &node, 0) == -1)
      EGOTO(error);

    if(node == NULL)
      break;

    if(doProcFctSep(jpa, node, fpOut, queue) == -1)
      EGOTO(error);

    if(jsparsflush(jpa, node, fpOut, 0) == -1)
      EGOTO(error);
  }

  /*  Flush any trailing tokens from last jsparsget
   *   (e. g. trailing comments)
   */
  if(jsparsflushtokens(jpa, fpOut) == -1)
    EGOTO(error);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doProcFctSep(jsparsstream_t *jpa, jsparscnode_t *node,
			FILE *fp, jctokfifo_t *queue)

/************************************************************************/
{
  jsparscnode_t *head, *body;
  fctsepaux_t saux;

  if(node->JPACN_Type != JPACT_FUNCDEF)
    /*  We don't bother about global typedefs, variable defs etc.  */
    return 0;

  head = node->JPACN_Sub;
  body = head->JPACN_Next;
  if(body->JPACN_Type == JPACT_LIST ||
     body->JPACN_Type == JPACT_DECLARATION) {
    /*  K&R-type function definition  */
    body = body->JPACN_Next;
  }

  saux.Mode = 0;
  saux.JPA = jpa;
  saux.Queue = queue;
  saux.Fp = fp;
  saux.FunctionName = getFunctionName(node, queue);
  saux.Function = node;
  saux.Body     = body;
  saux.BlockEntries = NULL;
  saux.MBlockEntries = saux.NBlockEntries = 0;
  saux.VarDefs = NULL;
  saux.MVarDefs = saux.NVarDefs = 0;

  if(doProcStmt(&saux, node) == -1)
    return -1;

  /*  Free allocate memory  */
  freesaux(&saux, 1);

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static void freesaux(fctsepaux_t *saux, int flag)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < saux->NBlockEntries ; ++i)
    if(saux->BlockEntries[i].OrigStmt)
      free((void *)saux->BlockEntries[i].OrigStmt);
  saux->NBlockEntries = 0;

  for(i = 0 ; i < saux->NVarDefs ; ++i) {
    free((void *)saux->VarDefs[i].Name);
    if(saux->VarDefs[i].NewName)
      free((void *)saux->VarDefs[i].NewName);
  }
  saux->NVarDefs = 0;

  if(flag) {
    if(saux->BlockEntries)
      free(saux->BlockEntries);

    if(saux->VarDefs)
      free(saux->VarDefs);
  }
}

/*JS*********************************************************************
*
*************************************************************************/

const char *getFunctionName(jsparscnode_t *node, jctokfifo_t *queue)

/************************************************************************/
{
  for(node = node->JPACN_Sub->JPACN_Sub ; node ; node = node->JPACN_Next)
    if(node->JPACN_Type == JPACT_DECLARATOR) {
      node = node->JPACN_Sub;
      assert(node->JPACN_Type == JPACT_IDENT);

      return queue->TKFF_Token[node->JPACN_IndexPP].TK_String;
    }

  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doProcStmt(fctsepaux_t *saux, jsparscnode_t *node)

/************************************************************************/
{
  jsparscnode_t *sub;

  if(node->JPACN_Type == JPACT_PRAGMALINE ||
     node->JPACN_Type == JPACT_PRAGMALINE2) {
    jctokfifo_t *queue;
    const char *name;
    int start, end, step;

    queue = saux->Queue;
    step = jctokgetstep(queue);

    /*  The actual content is in the child node of the pragma node  */
    assert(node->JPACN_Sub && node->JPACN_Sub->JPACN_Type == JPACT_PRAGMACONT);
    start = node->JPACN_Sub->JPACN_Start;
    end = node->JPACN_Sub->JPACN_End;

    if(queue->TKFF_Token[start].TK_Value != CTK_IDENT) {
      if(saux->JPA->JSPA_WarnFlags != 0)
	jsparsprinterr(saux->JPA, queue, node->JPACN_Sub,
		       "Syntax error in pragma line", 0);

      EGOTO(error);
    }

    name = queue->TKFF_Token[start].TK_String;
    for(start += step ; start <= end ; start += step)
      if(!(CTK_BLANK_START <= queue->TKFF_Token[start].TK_Value &&
	   queue->TKFF_Token[start].TK_Value <= CTK_BLANK_END))
	break;

    if(strcmp(name, PRAGMA_NAME_SEPARATE) == 0) {
      int ret;

      sub = node->JPACN_Next;

      if(sub->JPACN_Type != JPACT_COMPSTMT) {
	if(saux->JPA->JSPA_WarnFlags != 0)
	  jsparsprinterr(saux->JPA, queue, node,
			 "Pragma \"separate\" not prefixing block statement!",
			 0);

	return 0;
      }

      if((ret = doSeparateBlock(saux, start, end, sub)) == -1)
	EGOTO(error);

      if(ret != -2) {
	/*  Delete pragma line, since we have handled it  */
	start = node->JPACN_Start;
	end = node->JPACN_End;
	if(saux->JPA->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

	/*  Substitute it by the newlines if we care about the
	 *   original lines
	 */
	if(jctoksubstitute(queue, start, end,
			   (saux->JPA->JSPA_Mode & JPAM_FGETORIG) ? "" :
			   "%n") == -1)
	  EGOTO(error);

	/*  Somebody might have another "separate" pragma in the actual block.
	 *   That would possibly end in a mess, so we do not scan the block.
	 */
	return 0;
      }
    }

    if(node->JPACN_Type == JPACT_PRAGMALINE2)
      jsparsprinterr(saux->JPA, saux->Queue, node, "Unknown pragma line", 0);
  }

  /*  Scan recursively  */
  for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next)
    if(doProcStmt(saux, sub) == -1)
      EGOTO(error);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doSeparateBlock(fctsepaux_t *saux, int start, int end,
			   jsparscnode_t *block)

/************************************************************************/
{
  jsparsflownode_t *stmt;
  const char *funcName;
  jctokfifo_t *queue;
  int i, step, flags, indent, queueOff;
  int inds[3];

  /*  Clear previous stuff  */
  saux->NBlockEntries = 0;
  saux->NVarDefs = 0;

  queue = saux->Queue;
  step = jctokgetstep(queue);

  if(start > end)
    goto invalid;

  funcName = queue->TKFF_Token[start].TK_String;

  for(start += step ; start <= end ; ) {
    const char *string;
    int tokVal;

    tokVal = queue->TKFF_Token[start].TK_Value;
    string = queue->TKFF_Token[start].TK_String;
    start += step;

    if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
      continue;

    if(tokVal != CTK_IDENT)
      goto invalid;

    if(strcmp(string, PRAGMA_CLAUSE_LOCAL) == 0) {
      start = readIdents(saux, VARTYPE_BLOCKLOCAL, start, end);
    } else if(strcmp(string, PRAGMA_CLAUSE_VALUE) == 0) {
      start = readIdents(saux, VARTYPE_PASSVALUE, start, end);
    } else if(strcmp(string, PRAGMA_CLAUSE_REFERENCE) == 0) {
      start = readIdents(saux, VARTYPE_PASSREF, start, end);
    } else
      goto invalid;

    if(start == -1)
      EGOTO(error);

    if(start == -2)
      goto invalid;
  }

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & 1) {
    for(i = 0 ; i < saux->NVarDefs ; ++i)
      printf("VARTYPE %d '%s' '%s'\n",
	     saux->VarDefs[i].Type, saux->VarDefs[i].Name,
	     saux->VarDefs[i].NewName != NULL ?
	     saux->VarDefs[i].NewName : "<null>");
  }
#endif

  /* ************************************************* */
  queueOff = (saux->JPA->JSPA_Mode & JPAM_FGETORIG) ? 1 : 0;

  saux->Start = block->JPACN_Start;
  saux->End   = block->JPACN_End;

  /* ***  Find transitions of entering and leaving the block  *** */
  saux->Mode &= ~(SAUXMODE_SCANLEAVES | SAUXMODE_NEEDCOMMA);
  jsparsflowclear(saux->JPA);
  if(scan4blockentryexit(saux, saux->JPA->JSPA_FlowStart) == -1)
    EGOTO(error);

  /* ***  Scan block leaves for variable usage  *** */
  jsparsflowclear(saux->JPA);
  saux->Mode |= SAUXMODE_SCANLEAVES;

  stmt = saux->JPA->JSPA_StmtFlow->JPAFL_Statements;

  /*  bit 0: Block has more than one entry point.
   *  bit 1: Block has more than one exit point.
   */
  flags = 0;

  inds[0] = inds[1] = -1;
  for(i = 0 ; i < saux->NBlockEntries ; ++i) {
    int type, j;

    type = saux->BlockEntries[i].Type;
    start = (type & 1) ? stmt[saux->BlockEntries[i].Index].JPAFLN_Next1 :
      stmt[saux->BlockEntries[i].Index].JPAFLN_Next2;

    j = (type - 1) / 2; /*  1 & 2: 0   3 & 4: 1  */
    if(inds[j] == -1)
      inds[j] = start;
    else if(inds[j] != start)
      flags |= (1 << j); /*  there is more than one entry/exit point  */

    if(j == 1 && start >= 0 && scan4blockentryexit(saux, start) == -1)
      EGOTO(error);
  }

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & 1) {
    printf("Flag: %s, %s\n",
	   (flags & 1) ? "multiple entry points" : "only one entry point",
	   (flags & 2) ? "multiple exit points" : "only one exit point");

    for(i = 0 ; i < saux->NBlockEntries ; ++i)
      printf("%d: BLOCK%s indx %2d -> %d; type %d\n", i,
	     saux->BlockEntries[i].Type < 3 ? "ENTRY" : "EXIT ",
	     saux->BlockEntries[i].Index,
	     (saux->BlockEntries[i].Type & 1) ?
	     stmt[saux->BlockEntries[i].Index].JPAFLN_Next1 :
	     stmt[saux->BlockEntries[i].Index].JPAFLN_Next2,
	     saux->BlockEntries[i].Type);
  }
#endif

  jstrbuffempty(&HeaderBuffer);
  jstrbuffempty(&CallBuffer);
  jstrbuffempty(&BlockVarBuffer);
  jstrbuffempty(&SwitchBuffer);

  /*  Generate function definition  */
  if(jstrbuffprintf(&HeaderBuffer, "\n\n"
		    PATTERN_GENHEADER2 "\nstatic %s %s(",
		    saux->FunctionName,
		    (flags & 2) ? PATTERN_SELTYPE : "void",
		    funcName) == -1 ||
     ((flags & 1) && jstrbuffprintf(&HeaderBuffer,
				    PATTERN_SELTYPE " " PATTERN_SELPARAM ", ",
				    funcName) == -1))
    EGOTO(error);

  indent = jctokgetcolumn(queue, block->JPACN_Start + queueOff, 0);

  /*  Generate function call  */
  if(flags & 2) {
    if(jstrbuffprintf(&CallBuffer, "{\n%*s  "
		      PATTERN_SELTYPE " " PATTERN_RETVAR ";\n\n",
		      indent, "", funcName) == -1)
      EGOTO(error);
  }

  if(flags & 1) {
    if(jstrbuffprintf(&CallBuffer, PATTERN_FUNCLABEL ":\n",
		      funcName) == -1)
      EGOTO(error);
  }

  if((flags & 3) &&
     jstrbuffprintf(&CallBuffer, "%*s", indent, "") == -1)
    EGOTO(error);

  if(jstrbuffprintf(&CallBuffer, PATTERN_GENHEADER "\n") == -1)
    EGOTO(error);

  if(flags & 2) {
    if(jstrbuffprintf(&CallBuffer, "%*s  " PATTERN_RETVAR " = %s(",
		      indent, "", funcName, funcName) == -1)
      EGOTO(error);
  } else {
    if(jstrbuffprintf(&CallBuffer, "%*s%s(", indent, "", funcName) == -1)
      EGOTO(error);
  }

  if(flags & 1) {
    if(jstrbuffprintf(&CallBuffer, PATTERN_SELPARAM ", ", funcName) == -1)
      EGOTO(error);
  }

  /*  Add all other arguments  */
  if(genCalls(saux, &HeaderBuffer, &CallBuffer, &BlockVarBuffer) == -1)
    EGOTO(error);

  if(jstrbuffprint(&CallBuffer, ");") == -1)
    EGOTO(error);

  /*  inds[0]: return values for positive return.
   *  inds[1]: return values for negative return.
   *  inds[2]: parameter values for entry points.
   */
  inds[0] = 1;   inds[1] = -1;  inds[2] = 1;
  for(i = 0 ; i < saux->NBlockEntries ; ++i) {
    jsparscnode_t *node;
    int type, stmtFlags, indx, rstart, rend, next;

    type = saux->BlockEntries[i].Type;
    indx = saux->BlockEntries[i].Index;
    stmtFlags = stmt[indx].JPAFLN_Flags;
    if(stmtFlags & JPA_FLOW_PSEUDOSTMT)
      continue;

    node = stmt[indx].JPAFLN_Node;

    next = (type & 1) ? stmt[indx].JPAFLN_Next1 : stmt[indx].JPAFLN_Next2;

    rstart = node->JPACN_Start + queueOff;
    rend   = node->JPACN_End   + queueOff;

    /*  Get original statement  */
    if((saux->BlockEntries[i].OrigStmt =
	jctoktostring(queue, rstart, rend)) == NULL)
      EGOTO(error);

    if(type == 1 || type == 2) {
      if(next != -1 && next == jsparsfindflow(saux->JPA, block->JPACN_Sub))
	continue;

      assert(!(stmtFlags & JPA_FLOW_EXPRESSION));

      if(flags & 1) {
	int j;

	for(j = 0 ; j < i ; ++j)
	  if(saux->BlockEntries[i].Type < 3 &&
	     saux->BlockEntries[j].OrigStmt != NULL &&
	     strcmp(saux->BlockEntries[i].OrigStmt,
		    saux->BlockEntries[j].OrigStmt) == 0)
	    break;

	if(j < i)
	  indx = saux->BlockEntries[j].Value;
	else
	  indx = inds[2]++;

	saux->BlockEntries[i].Value = indx;

	if(jctoksubstitute(queue, rstart, rend,
			   "{%I  " PATTERN_SELPARAM " = %d;"
			   "%I  goto " PATTERN_FUNCLABEL ";"
			   "%I}", funcName, indx, funcName) == -1)
	  EGOTO(error);

	if(j == i)
	  if(jstrbuffprintf(&SwitchBuffer,
			    "%sif(" PATTERN_SELPARAM " == %d) %s\n",
			    jstrbufflen(&SwitchBuffer) > 0 ? "  " : "",
			    funcName, indx,
			    saux->BlockEntries[i].OrigStmt) == -1)
	    EGOTO(error);
      } else {
	if(jctoksubstitute(queue, rstart, rend,
			   "goto " PATTERN_FUNCLABEL ";",
			   funcName) == -1)
	  EGOTO(error);
      }
    } else {
      int j;

      assert(!(stmtFlags & JPA_FLOW_EXPRESSION));

      for(j = 0 ; j < i ; ++j)
	if(saux->BlockEntries[i].Type >= 3 &&
	   saux->BlockEntries[j].OrigStmt != NULL &&
	   strcmp(saux->BlockEntries[i].OrigStmt,
		  saux->BlockEntries[j].OrigStmt) == 0)
	  break;

      if(j < i)
	indx = saux->BlockEntries[j].Value;
      else if(checkReturnType(saux, node) == 1)
	indx = inds[0]++;
      else
	indx = inds[1]--;

      saux->BlockEntries[i].Value = indx;

      if(j == i)
	if(jstrbuffprintf(&CallBuffer, "\n%*s  if(" PATTERN_RETVAR " == %d)\n"
			  "%*s    %s",
			  indent, "", funcName, indx,
			  indent, "", saux->BlockEntries[i].OrigStmt) == -1)
	  EGOTO(error);

      if(jctoksubstitute(queue, rstart, rend,
			 (flags & 2) ? PATTERN_RETSTMTVAL :
			 PATTERN_RETSTMT, indx) == -1)
	EGOTO(error);
    }
  }

  if((flags & 2) && jstrbuffprintf(&CallBuffer, "\n%*s}\n", indent, "") == -1)
    EGOTO(error);

  start = block->JPACN_Start + queueOff;
  end   = block->JPACN_End   + queueOff;

  /* ***  Print separated function  *** */
  if(jsparsprocessvars(saux->JPA, block, queue, substVariable,
		       saux, 1) == -1)
    EGOTO(error);

  if(fprintf(saux->Fp, "%s%s)\n{", jstrbuffer(&HeaderBuffer),
	     (saux->Mode & SAUXMODE_NEEDCOMMA) ? "" : "void") == -1)
    EGOTO(error);

  if(jstrbufflen(&BlockVarBuffer) > 0 &&
     fprintf(saux->Fp, "%s", jstrbuffer(&BlockVarBuffer)) == -1)
    EGOTO(error);

  if(printBlock(saux->JPA, saux->Fp, queue, block,
		jstrbufflen(&SwitchBuffer) > 0 ?
		jstrbuffer(&SwitchBuffer) : NULL,
		jctokgetcolumn(queue, start, 0)) == -1)
    EGOTO(error);

  if((flags & 2) &&
     fprintf(saux->Fp, "\n  " PATTERN_RETSTMTVAL "\n", 0) == -1)
    EGOTO(error);

  if(fprintf(saux->Fp, "}\n\n") == -1)
    EGOTO(error);

  /* ***  Substitute original  *** */
  if(jctoksubstitute(queue, start, end,
		     (saux->JPA->JSPA_Mode & JPAM_FGETORIG) ?
		     "%I%s" : "%s", jstrbuffer(&CallBuffer)) == -1)
    EGOTO(error);

  if(flags & 1) {
    start = saux->Body->JPACN_Start + queueOff;
    if(jctoksubstitute(queue, start, start,
		       "{\n  " PATTERN_SELTYPE " " PATTERN_SELPARAM " = 0;",
		       funcName) == -1)
      EGOTO(error);
  }

  /* ***  Free memory  *** */
  freesaux(saux, 0);

  return 0;
error:
  freesaux(saux, 0);
  return -1;
invalid:
  freesaux(saux, 0);
  return -2;
}

/*JS*********************************************************************
*
*************************************************************************/

static int readIdents(fctsepaux_t *saux, int type, int start, int end)

/************************************************************************/
{
  const char *name;
  jctokfifo_t *queue;
  int step, state;

  name = NULL;
  queue = saux->Queue;
  step = jctokgetstep(queue);

  for(state = 0 ; start <= end && state >= 0 ; start += step) {
    int tokVal;

    tokVal = queue->TKFF_Token[start].TK_Value;
    if(CTK_BLANK_START <= tokVal && tokVal <= CTK_BLANK_END)
      continue;

    switch(state) {
    case 0: /*  look for '('  */
      if(tokVal != '(')
	goto invalid;

      state = 10;
      break;

    case 10: /*  look for name  */
      if(tokVal != CTK_IDENT)
	goto invalid;

      name = queue->TKFF_Token[start].TK_String;
      state = 20;
      break;

    case 20: /*  look for name or separator  */
      if(tokVal == CTK_IDENT) {
	if(addVarDef(saux, type, name,
		     queue->TKFF_Token[start].TK_String) == -1)
	  EGOTO(error);
	state = 5;
	break;
      }

      if(addVarDef(saux, type, name, NULL) == -1)
	EGOTO(error);

      /* FALLTHROUGH */
    case 5: /*  look for separator  */
      if(tokVal == ',')
	state = 10;
      else if(tokVal == ')')
	state = -1;
      break;
    }
  }

  return start;
error:
  return -1;
invalid:
  return -2;
}

/*JS*********************************************************************
*
*************************************************************************/

static int addVarDef(fctsepaux_t *saux, int type,
		     const char *name, const char *newName)

/************************************************************************/
{
  if(saux->NVarDefs >= saux->MVarDefs) {
    REALLOC_RESIZE(saux->MVarDefs, 10);
    if((saux->VarDefs = (fctsepvar_t *)
	realloc(saux->VarDefs, saux->MVarDefs * sizeof(*(saux->VarDefs))))
       == NULL)
      return -1;
  }

  saux->VarDefs[saux->NVarDefs].Type = type;
  if((saux->VarDefs[saux->NVarDefs].Name = strdup(name)) == NULL)
    return -1;

  if(newName == NULL)
    saux->VarDefs[saux->NVarDefs].NewName = NULL;
  else if((saux->VarDefs[saux->NVarDefs].NewName = strdup(newName)) == NULL)
    return -1;

  saux->NVarDefs++;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int scan4blockentryexit(fctsepaux_t *saux, int start)

/************************************************************************/
{
  jsparsflow_t *jFlow;
  jsparsflownode_t *stmt;
  int nexts[2];

  jFlow = saux->JPA->JSPA_StmtFlow;
  stmt = jFlow->JPAFL_Statements;

  do {
    jsparscnode_t *node;
    int i, isInBlock;

    if(BIT_ISSET(jFlow->JPAFL_Bitmaps[0], start))
      return 0;

    BIT_SET(jFlow->JPAFL_Bitmaps[0], start);

    nexts[0] = stmt[start].JPAFLN_Next1;
    nexts[1] = stmt[start].JPAFLN_Next2;

    node = stmt[start].JPAFLN_Node;

    isInBlock = saux->Start <= node->JPACN_Start &&
      node->JPACN_End <= saux->End;

    if(!(stmt[start].JPAFLN_Flags & JPA_FLOW_PSEUDOSTMT)) {
      if(jsparsprocessvars(saux->JPA, node, saux->Queue, doSepVariable,
			   saux, 1) == -1)
	EGOTO(error);
    }

    for(i = 0 ; i < 2 ; ++i) {
      int goesIntoBlock;

      if(nexts[i] == -1)
	continue;

      if(nexts[i] == JPA_FLOW_INDEX_RETURN) {
	goesIntoBlock = FALSE;
      } else {
	node = stmt[nexts[i]].JPAFLN_Node;
	goesIntoBlock = saux->Start <= node->JPACN_Start &&
	  node->JPACN_End <= saux->End;
      }

      if(!isInBlock) {
	if(goesIntoBlock) {
	  /* ***  Statement flow enters block  **** */
	  if(addblockentryexit(saux, start, 1 + i) == -1)
	    EGOTO(error);
	}
      } else { /*  isInBlock == TRUE  */
	if(!goesIntoBlock) {
	  /* ***  Statement flow leaves block  **** */
	  if(addblockentryexit(saux, start, 3 + i) == -1)
	    EGOTO(error);

	  if((saux->Mode & SAUXMODE_SCANLEAVES) == 0)
	    /*  First round: Only look for block entries and exits,
	     *   but do not follow block exits
	     */
	    break;
	}
      }
    }

    if(nexts[1] >= 0)
      if(scan4blockentryexit(saux, nexts[1]) == -1)
	EGOTO(error);
  } while((start = nexts[0]) >= 0);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   type == 1/2: JPAFLN_Next1/2 enters block
*   type == 3/4: JPAFLN_Next1/2 leaves block
*************************************************************************/

static int addblockentryexit(fctsepaux_t *saux, int indx, int type)

/************************************************************************/
{
  int pos;

  pos = findblockentryexit(saux, indx);

  if(pos < saux->NBlockEntries && saux->BlockEntries[pos].Index == indx &&
     saux->BlockEntries[pos].Type == type)
    /*  Entry already present  */
    return 0;

  if(saux->NBlockEntries >= saux->NBlockEntries) {
    REALLOC_RESIZE(saux->MBlockEntries, 10);
    if((saux->BlockEntries = (blockentry_t *)
	realloc(saux->BlockEntries, saux->MBlockEntries *
		sizeof(*(saux->BlockEntries)))) == NULL)
      return -1;
  }

  if(pos < saux->NBlockEntries)
    memmove(&(saux->BlockEntries[pos + 1]), &(saux->BlockEntries[pos]),
	    (saux->NBlockEntries - pos) * sizeof(saux->BlockEntries[pos]));

  saux->BlockEntries[pos].Index = indx;
  saux->BlockEntries[pos].Type = type;
  saux->BlockEntries[pos].OrigStmt = NULL;
  saux->BlockEntries[pos].Value = 0;

  saux->NBlockEntries++;

  return 1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int findblockentryexit(fctsepaux_t *saux, int indx)

/************************************************************************/
{
  int lower, upper;

  /*  Binary search  */
  lower = 0;   upper = saux->NBlockEntries;
  while(lower < upper) {
    int m = (lower + upper) >> 1;
    if(saux->BlockEntries[m].Index < indx)
      lower = m + 1;
    else
      upper = m;
  }

  return lower;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doSepVariable(jsparsstream_t *jpa, jsparscnode_t *node,
		         jctokfifo_t *queue, jsparsvartab_t *vTab,
		         jsparsvardef_t *vDef, long flags, void *user)

/************************************************************************/
{
  int isInBlock;

#define SAUX      ((fctsepaux_t *)user)

  /*  Check for block-local variables or global variable  */
  if((SAUX->Start <= vTab->JPAVT_StartIndex &&
      vTab->JPAVT_EndIndex <= SAUX->End) ||
     !(vTab->JPAVT_StartIndex <= SAUX->Body->JPACN_Start &&
       SAUX->Body->JPACN_End <= vTab->JPAVT_EndIndex))
    return 0;

  isInBlock = SAUX->Start <= node->JPACN_Start &&
    node->JPACN_End <= SAUX->End;

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & 4)
    printf("Queue %4d: Variable %s flags %c%c%c '%s'\n",
	   node->JPACN_Start,
	   isInBlock ? "inside " : "outside",
	   (flags & JVDF_FREAD) ?  'R' : '-',
	   (flags & JVDF_FWRITE) ? 'W' : '-',
	   (flags & JVDF_FADDR) ?  'A' : '-',
	   vDef->JPAVD_Name);
#endif

  if(flags & (JVDF_FREAD | JVDF_FADDR))
    vDef->JPAVD_Flags |= isInBlock ? FLAG_RD_INSIDE :
      (SAUX->Mode & SAUXMODE_SCANLEAVES) ? FLAG_RD_AFTER : FLAG_RD_BEFORE;

  if(flags & (JVDF_FWRITE | JVDF_FADDR))
    vDef->JPAVD_Flags |= isInBlock ? FLAG_WR_INSIDE :
      (SAUX->Mode & SAUXMODE_SCANLEAVES) ? FLAG_WR_AFTER : FLAG_WR_BEFORE;

#undef SAUX

  return 0;
}

/*JS*********************************************************************
*   1: Use positive return value
*   2: Use negative return value
*************************************************************************/

static int checkReturnType(fctsepaux_t *saux, jsparscnode_t *node)

/************************************************************************/
{
  if(node->JPACN_Type == JPACT_GOTO) {
    node = node->JPACN_Sub;
    assert(node->JPACN_Type == JPACT_IDENT);

    if(jsstrincmp(saux->Queue->TKFF_Token[node->JPACN_IndexPP].TK_String,
		  ERRORLABEL_START, ERRORLABEL_START_LEN) == 0)
      return 2;
  } else if(node->JPACN_Type == JPACT_RETURN) {
    int type;

    node = node->JPACN_Sub;
    assert(node->JPACN_Type == JPACT_UNARYEXPR);

    node = node->JPACN_Sub;
    type = node->JPACN_Type;
    if(type == JPACT_INTCONST || type == JPACT_ENUMCONST) {
      node = node->JPACN_Next;
      if(node && node->JPACN_Type == JPACT_UMINUS)
	return 2;
    }
  }

  return 1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int genCalls(fctsepaux_t *saux, jstrbuff_t *headerBuff,
		    jstrbuff_t *callBuff, jstrbuff_t *varBuff)

/************************************************************************/
{
  jsparsstream_t *jpa = saux->JPA;
  int i;

  for(i = 0 ; i < jpa->JSPA_NVarTables ; ++i) {
    jsparsvartab_t *vTab;
    int h;

    vTab = jpa->JSPA_VarTables[i];
    for(h = 0 ; h < JPA_VARIABLE_TABSIZE ; ++h) {
      jsparsvardef_t *vDef;

      for(vDef = vTab->JPAVT_Vars[h] ; vDef ; vDef = vDef->JPAVD_Next) {
	const char *name;
	int type;

	name = vDef->JPAVD_Name;

	if((type = getVarType(saux, vDef, &name)) == VARTYPE_IGNORE)
	  continue;

	if(type == VARTYPE_BLOCKLOCAL) {
	  /*  Not used outside block - at least we have to generate the
	   *   definition inside the block
	   */
	  if(jstrbuffprintf(varBuff, "\n  ") == -1 ||
	     jsparsgenvardef(saux->JPA, varBuff, vDef, name) == -1 ||
	     jstrbuffprint(varBuff, ";") == -1)
	    EGOTO(error);
	} else {
	  /*  Put separator in buffer  */
	  if(saux->Mode & SAUXMODE_NEEDCOMMA) {
	    if(jstrbuffprint(headerBuff, ", ") == -1 ||
	       jstrbuffprint(callBuff, ", ") == -1)
	      EGOTO(error);
	  } else
	    /*  From now on, we need it  */
	    saux->Mode |= SAUXMODE_NEEDCOMMA;

	  if(type == VARTYPE_PASSREF) {
	    jsparscnode_t *decl;

	    decl = vDef->JPAVD_Declarator->JPACN_Sub;
	    if(decl && decl->JPACN_Type == JPACT_IDENT)
	      decl = decl->JPACN_Next;

	    jstrbuffempty(&TempBuffer);
	    if(jstrbuffprintf(&TempBuffer, decl ?
			      "(*" PATTERN_PTRNAME ")" :
			      "*" PATTERN_PTRNAME,
			      name) == -1)
	      EGOTO(error);

	    name = jstrbuffer(&TempBuffer);
	  }

	  if(jsparsgenvardef(saux->JPA, headerBuff, vDef, name) == -1)
	    EGOTO(error);

	  if(jstrbuffprintf(callBuff, type == VARTYPE_PASSREF ?
			    "&%s" : "%s", vDef->JPAVD_Name) == -1)
	    EGOTO(error);
	}
      }
    }
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   VARTYPE_IGNORE:      nothing to do
*   VARTYPE_BLOCKLOCAL:  define in block
*   VARTYPE_PASSVALUE:   pass by value
*   VARTYPE_PASSREF:     pass by reference
*************************************************************************/

static int getVarType(fctsepaux_t *saux, jsparsvardef_t *vDef,
		      const char **pName)

/************************************************************************/
{
  long flags;
  int i;

  /*  Look for any definition  */
  for(i = 0 ; i < saux->NVarDefs ; ++i)
    if(strcmp(saux->VarDefs[i].Name, *pName) == 0) {
      if(saux->VarDefs[i].NewName != NULL)
	*pName = saux->VarDefs[i].NewName;

      return saux->VarDefs[i].Type;
    }

  flags = vDef->JPAVD_Flags;

  if((flags & JVDF_TMASK) != JVDF_TVARIABLE)
    return VARTYPE_IGNORE;

  if((flags & (FLAG_RD_INSIDE | FLAG_WR_INSIDE)) == 0)
    return VARTYPE_IGNORE; /*  nothing to do  */

  if((flags & (FLAG_RD_BEFORE | FLAG_WR_BEFORE | FLAG_RD_AFTER |
	       FLAG_WR_AFTER)) == 0)
    /*  Not used outside block - at least we have to generate the
     *   definition inside the block
     */
    return VARTYPE_BLOCKLOCAL;

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & 8)
    printf("VAR before %c%c inside %c%c after %c%c '%s'\n",
	   (flags & FLAG_RD_BEFORE) ? 'r' : '-',
	   (flags & FLAG_WR_BEFORE) ? 'w' : '-',
	   (flags & FLAG_RD_INSIDE) ? 'r' : '-',
	   (flags & FLAG_WR_INSIDE) ? 'w' : '-',
	   (flags & FLAG_RD_AFTER)  ? 'r' : '-',
	   (flags & FLAG_WR_AFTER)  ? 'w' : '-',
	   vDef->JPAVD_Name);
#endif

  if((flags & FLAG_WR_INSIDE) && (flags & FLAG_RD_AFTER))
    return VARTYPE_PASSREF;

  return VARTYPE_PASSVALUE;
}

/*JS*********************************************************************
*
*************************************************************************/

static int substVariable(jsparsstream_t *jpa, jsparscnode_t *node,
			 jctokfifo_t *queue, jsparsvartab_t *vTab,
			 jsparsvardef_t *vDef, long flags, void *user)

/************************************************************************/
{
  const char *name;
  int start, end, type;

#define SAUX      ((fctsepaux_t *)user)

  start = node->JPACN_Start;
  end   = node->JPACN_End;
  if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

  name = vDef->JPAVD_Name;

  type = getVarType(SAUX, vDef, &name);

  if(type == VARTYPE_PASSREF) {
    const char *format;

    node = node->JPACN_Next;

    if(node && node->JPACN_Type == JPACT_ADDRESS) {
      format = PATTERN_PTRNAME;
      start = node->JPACN_Start;
      if(jpa->JSPA_Mode & JPAM_FGETORIG) ++start;
    } else if(node != NULL)
      format = "(*" PATTERN_PTRNAME ")";
    else
      format = "*" PATTERN_PTRNAME;

    if(jctoksubstitute(SAUX->Queue, start, end, format, name) == -1)
      EGOTO(error);
  } else if(type != VARTYPE_IGNORE && name != vDef->JPAVD_Name) {
    if(jctoksubstitute(SAUX->Queue, start, end, "%s", name) == -1)
      EGOTO(error);
  }

#undef SAUX

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int printBlock(jsparsstream_t *jpa, FILE *fp, jctokfifo_t *queue,
		      jsparscnode_t *node, const char *switchstmt,
		      int offset)

/************************************************************************/
{
  int start, end, step, switchPos;

  step = jctokgetstep(queue);

  start = node->JPACN_Start;
  end   = node->JPACN_End;
  if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

  assert(queue->TKFF_Token[start].TK_Value == '{');
  assert(queue->TKFF_Token[end].TK_Value == '}');
  start += step;
  end   -= step;

  if(switchstmt) {
    jsparscnode_t *sub;
    for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next)
      if(sub->JPACN_Type != JPACT_DECLARATION &&
	 sub->JPACN_Type != JPACT_TYPEDEF)
	break;

    if(sub) {
      switchPos = sub->JPACN_Start;
      if(jpa->JSPA_Mode & JPAM_FGETORIG) ++switchPos;
    } else
      /*  That's funny - somebody tries to separate an empty block  */
      switchPos = end;
  } else
    switchPos = -1;

  for( ; start <= end ; start += step) {
    jsctoken_t token;
    int tokVal;

    token = queue->TKFF_Token[start];
    tokVal = token.TK_Value;

    if(start == switchPos &&
       fprintf(fp, "%s\n  %s\n  ", PATTERN_GENHEADER, switchstmt) == -1)
      EGOTO(error);

    if(tokVal == CTK_BLANKSEOL && offset != 0) {
      const char *string, *next;
      int col;

      for(string = token.TK_String ;
	  (next = strchr(string, '\n')) != NULL ; string = next + 1)
	if(putc('\n', fp) == EOF)
	  EGOTO(error);

      if((col = updateColumn(0, string) - offset) < 0)
	col = 0;

      if(fprintf(fp, "%*s", col, "") == -1)
	EGOTO(error);

      continue;
    }

    if(token.TK_Value < 0)
      continue;

    if(printCToken(fp, token) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}
