/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Print table of contents for functions.
 *
 *              Copyright (C) 2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_PROTO      "proto"
#define JSCPP_MODE_FCTTOC     "fcttoc"
/*  END-DEFINITIONS  */

typedef struct {
    /*
     *  Bit 0: prototype seen in proper header "file.h".
     *  Bit 1: prototype seen in some other header.
     *  Bit 2: static prototype seen.
     *  Bit 8: function definition seen.
     *  Bit 9: static function definition seen.
     */
    int Flags;
    const char *Function;
    const char *Location; /*  header where this prototype was seen  */
} prototypeaux_t;

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doFctToc(jsparsenviron_t *ppenv,
				  const char *names[],
				  int nnames, FILE *fpOut, int mode);
static int               printToc(jsparsstream_t *jpa, FILE *fpOut,
				  const char *fileName, const char *headerName,
				  int mode);
static int               handleNode(jsparsstream_t *jpa, jsparscnode_t *node,
				    FILE *fpOut,
				    const char *fileName,
				    const char *headerName, int mode);

static jsctoken_t        convertTokens(void *user, jsctoken_t token);
static char             *trimString(char *string);
static void              cleanPrototypes(void);
static int               addPrototype(int flags, const char *function,
				      const char *location);

/*********     GLOBAL VARIABLES                                 *********/
static prototypeaux_t *ProtoList;
static int NProtoList, MProtoList;

/*JS*********************************************************************
 *
 *************************************************************************/

int doFctToc(jsparsenviron_t *ppenv, const char *names[], int nnames,
	     FILE *fpOut, int mode)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < nnames ; ++i) {
	jsparsstream_t *jpa;
	jsppstream_t *jps;
	char headerName[JS_FILENAME_MAX];
	int ret, pmode;

	/*  Get meaningful warning messages  */
	pmode = JPAM_FGETFPOS;

	if(mode & 256) {
	    char *sepPtr;

	    strcpy(headerName, names[i]);
	    if((sepPtr = strrchr(headerName, '.')) != NULL)
		strcpy(sepPtr + 1, "h");
	    else
		strcat(headerName, ".h");
	}

	if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
	    EGOTO(error);

	jps = jsqsgetppstream(jpa->JSPA_Stream);
	jps->JSPP_Mode |= JPPM_PRESSPACE;

	/*  Print comment line in case of multiple sources  */
	if(nnames > 1 && !(mode & 1)) {
	    if((ret = 40 - strlen(names[i])) < 0)
		ret = 0;

	    if(fprintf(fpOut, "%s/* ********* PROTOTYPES '%s'%*s ********* */\n",
		       i > 0 ? "\n" : "", names[i], ret, "") == -1)
		EGOTO(error);
	}

	ret = printToc(jpa, fpOut, names[i],
		       (mode & 256) ? headerName : NULL, mode);

	cleanPrototypes();

	if(closeParserStream(jpa, 0) == -1)
	    EGOTO(error);

	if(ret == -1)
	    EGOTO(error);
    }

    /*  Clean up everything  */
    free(ProtoList);
    ProtoList = NULL;
    MProtoList = 0;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   mode bit 0: Print special TOC with line ranges and '@' like prototype.
 *   mode bit 1: Print also static prototypes.
 *   mode bit 8: If set, check that functions from "xxx.c" have prototypes
 *               in "xxx.h" and print warning (only if prototype warnings
 *               are enabled).
 *************************************************************************/

static int printToc(jsparsstream_t *jpa, FILE *fpOut, const char *fileName,
		    const char *headerName, int mode)

/************************************************************************/
{
    for(;;) {
	jsparscnode_t *node;

	if(jsparsget(jpa, &node, 0) == -1)
	    EGOTO(error);

	if(node == NULL)
	    break;

	if(handleNode(jpa, node, fpOut, fileName, headerName, mode) == -1)
	    EGOTO(error);

	if(jsparsflush(jpa, node, NULL, 0) == -1)
	    EGOTO(error);
    }

    if(jpa->JSPA_WarnFlags & JPAM_WARNPROTO) {
	jctokfifo_t *queue;
	int i;

	queue = jsparsgetqueue(jpa);

	for(i = 0 ; i < NProtoList ; ++i) {
	    int flags;

	    flags = ProtoList[i].Flags;

	    if((flags & 256) && !(flags & 3) &&
	       strcmp(ProtoList[i].Function, "main") != 0)
		jsparsprinterr(jpa, queue, NULL,
			       "No prototype for function '%s' seen", 0,
			       ProtoList[i].Function);

	    if((mode & 256) && headerName && (flags & 256) && !(flags & 1))
		jsparsprinterr(jpa, queue, NULL,
			       "No prototype for function '%s' in accompanying header '%s' seen", 0,
			       ProtoList[i].Function, headerName);

	    if((flags & 1) && !(flags & 256))
		jsparsprinterr(jpa, queue, NULL,
			       "Prototype for function '%s' in accompanying header '%s', but no definition", 0,
			       ProtoList[i].Function, headerName);
	}
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int handleNode(jsparsstream_t *jpa, jsparscnode_t *node,
		      FILE *fpOut, const char *fileName,
		      const char *headerName, int mode)

/************************************************************************/
{
    jctokfifo_t *queue;
    int step, indfpos, type;

    queue = jsparsgetqueue(jpa);
    step = jctokgetstep(queue);
    indfpos = jctokindfpos(queue);

    type = node->JPACN_Type;

    if(type == JPACT_DECLARATION &&
       (jpa->JSPA_WarnFlags & JPAM_WARNPROTO)) {
	jsparscnode_t *decl, *spec;
	const char *location;
	int isFuncCall;

	decl = node->JPACN_Sub;
	for(spec = decl->JPACN_Next ; spec ; spec = spec->JPACN_Next)
	    if(spec->JPACN_Type == JPACT_STATIC)
		break;

	type = decl->JPACN_Type;
	if(type == JPACT_LIST)
	    decl = decl->JPACN_Sub;
	else if(type != JPACT_DECLARATOR)
	    /*  Only specifiers, something like "struct x {};"  */
	    return 0;

	isFuncCall = FALSE;
	for( ; decl ; decl = decl->JPACN_Next) {
	    jsparscnode_t *idsub;
	    const char *funcName;

	    assert(decl->JPACN_Type == JPACT_DECLARATOR);
	    idsub = decl->JPACN_Sub;
	    assert(idsub->JPACN_Type == JPACT_IDENT);
	    funcName = queue->TKFF_Token[idsub->JPACN_IndexPP].TK_String;

	    idsub = idsub->JPACN_Next;
	    if(idsub != NULL && idsub->JPACN_Type == JPACT_FUNCCALL) {
		int flags, oflags;

		isFuncCall = TRUE;
		if(indfpos != -1) {
		    location = queue->TKFF_Token[idsub->JPACN_Start + indfpos].TK_String;
		    flags = spec != NULL ? 4 : (headerName != NULL && strcmp(location, headerName) == 0) ? 1 : 2;
		} else {
		    location = NULL;
		    flags = spec != NULL ? 4 : 2;
		}

		if((oflags = addPrototype(flags, funcName, location)) == -1)
		    EGOTO(error);

		if((flags & 1) && (oflags & 1))
		    jsparsprinterr(jpa, queue, idsub,
				   "Duplicate prototype for function '%s' in header",
				   0, funcName);
		else if((oflags & 3) &&
			/*  A prototype within the original source file is OK  */
			location != NULL && strcmp(location, fileName) != 0)
		    jsparsprinterr(jpa, queue, idsub,
				   "Prototype for function '%s' also in other header",
				   0, funcName);
	    }

	    if(type != JPACT_LIST)
		break;
	}

	if(spec != NULL && isFuncCall && indfpos != -1) {
	    location = queue->TKFF_Token[spec->JPACN_Start + indfpos].TK_String;
	    if(strcmp(location, fileName) != 0)
		jsparsprinterr(jpa, queue, spec,
			       "'static' prototype in header '%s'", 0,
			       location);
	}
    } else if(type == JPACT_FUNCDEF) { /* ***  Function definition  *** */
	jsparscnode_t *head, *sub, *decl, *spec;
	const char *location, *funcName;

	/*
	 * ===========DECLARATION=============....
	 *        =====DECLARATOR=============....
	 * ............IDENTIFIER(ARGS)......
	 * <---------->                <---->
	 *  pre                         after
	 */
#if 0
	jsparsprctree(stdout, queue, node, 0, 2);
	/*jctokshowqueue(stdout, queue);*/
#endif

	head = node->JPACN_Sub;
	assert(head->JPACN_Type == JPACT_DECLARATION);

	/*  Determine declarator and specifier  */
	for(decl = head->JPACN_Sub ; decl ; decl = decl->JPACN_Next)
	    if(decl->JPACN_Type == JPACT_DECLARATOR)
		break;

	assert(decl != NULL);
	spec = decl->JPACN_Next;
	assert(spec != NULL);

	/*  Check for static function  */
	for(spec = decl->JPACN_Next ; spec ; spec = spec->JPACN_Next)
	    if(spec->JPACN_Type == JPACT_STATIC)
		break;

	if(spec != NULL && !(mode & 2))
	    /*  Skip static prototypes  */
	    return 0;

	/*  Get function name and location  */
	sub = decl->JPACN_Sub;
	assert(sub->JPACN_Type == JPACT_IDENT);
	funcName = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;

	location = indfpos != -1 ?
	    queue->TKFF_Token[node->JPACN_Start + indfpos].TK_String :
	    NULL;

	jstrbuffempty(&TempBuffer);

	if(mode & 1) {
	    int pre0, pre1, after0, after1, hmode;
	    char *string;

	    pre0 = head->JPACN_Start;

	    after1 = decl->JPACN_End;

	    decl = decl->JPACN_Sub;
	    assert(decl->JPACN_Type == JPACT_IDENT);
	    pre1 = decl->JPACN_Start - step;
	    after0 = decl->JPACN_End + step;
#if 0
	    printf("[#%d %d %d %d %d]", step, pre0, pre1, after0, after1);
#endif

	    /*   Print function name and line range (if exists)  */
	    if(location != NULL)
		fprintf(fpOut, "\"%s\" %d %d ", location,
			queue->TKFF_Token[node->JPACN_Start + indfpos].TK_Value,
			queue->TKFF_Token[node->JPACN_End + indfpos].TK_Value);

	    if(fprintf(fpOut, "%s", funcName) == -1)
		EGOTO(error);

	    hmode = 1;
	    if(jctoktostrbuff(&TempBuffer, queue, pre0, pre1, convertTokens, &hmode) == -1 ||
	       jstrbuffprint(&TempBuffer, "@") == -1 ||
	       jctoktostrbuff(&TempBuffer, queue, after0, after1, convertTokens, &hmode) == -1)
		EGOTO(error);

	    string = trimString(jstrbuffer(&TempBuffer));
	    tkDelNewLines(string);

	    if(fprintf(fpOut, " %s\n", string) == -1)
		EGOTO(error);
	} else {
	    int hmode;
	    char *string;

	    hmode = 0;
	    if(jctoktostrbuff(&TempBuffer, queue, head->JPACN_Start, head->JPACN_End,
			      convertTokens, &hmode) == -1)
		EGOTO(error);

	    string = trimString(jstrbuffer(&TempBuffer));
	    tkDelNewLines(string);

	    if(fprintf(fpOut, "%s%s;\n", spec != NULL ? "" : "extern ", string) == -1)
		EGOTO(error);
	}

	if(jpa->JSPA_WarnFlags & JPAM_WARNPROTO) {
	    int oflags;

	    if((oflags = addPrototype(spec != NULL ? 512 : 256, funcName, location)) == -1)
		EGOTO(error);

	    if(spec == NULL && (oflags & 4))
		jsparsprinterr(jpa, queue, NULL,
			       "Static prototype for function '%s', but no 'static' keyword in definition", 0,
			       funcName);
	}
    }

    return 1;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jsctoken_t convertTokens(void *user, jsctoken_t token)

/************************************************************************/
{
    int tokval, mode;

    mode = *(int *)user;

    tokval = token.TK_Value;

    if(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END) {
	if(*(int *)user & 2) {
	    *(int *)user &= ~2;

	    token.TK_Value = CTK_BLANKS;
	    token.TK_String = " ";
	} else {
	    token.TK_Value = PTK_EMPTY;
	    token.TK_String = NULL;
	}

    } else if(tokval == PTK_STARTIGN || tokval == PTK_ENDIGN) {
	token.TK_Value = PTK_EMPTY;
	token.TK_String = NULL;
    } else {
	*(int *)user |= 2; /*  last one wasn't space  */

	if((mode & 1) && tokval == CTK_IDENT) {
	    if(tkStrcmp(token.TK_String, "static") == 0 ||
	       tkStrcmp(token.TK_String, "extern") == 0) {
		token.TK_Value = PTK_EMPTY;
		token.TK_String = NULL;
	    }
	}
    }

    return token;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static char *trimString(char *string)

/************************************************************************/
{
    int len;

    for( ; *string ; ++string)
	if(!isspace(*string))
	    break;

    len = strlen(string);
    while(len > 0 && isspace(string[len - 1]))
	--len;

    string[len] = '\0';

    return string;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static void cleanPrototypes(void)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < NProtoList ; ++i) {
	free((void *)ProtoList[i].Function);

	if(ProtoList[i].Location != NULL)
	    free((void *)ProtoList[i].Location);
    }

    NProtoList = 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addPrototype(int flags, const char *function,
			const char *location)

/************************************************************************/
{
    int lower, upper;

    /*printf("ADD %d '%s' '%s'\n", flags, function, location);*/

    /*  Binary search  */
    lower = 0;   upper = NProtoList;
    while(lower < upper) {
	int m = (lower + upper) >> 1;

	if(strcmp(ProtoList[m].Function, function) < 0)
	    lower = m + 1;
	else
	    upper = m;
    }

    if(lower < NProtoList && strcmp(ProtoList[lower].Function, function) == 0) {
	int oldFlags;

	oldFlags = ProtoList[lower].Flags;

	if(flags == 256 && (oldFlags & 4)) {
	    /*  a definition after a static prototype forces a static definition  */
	    flags = 512;
	}

	ProtoList[lower].Flags |= flags;

	if(flags & 1) {
	    free((void *)ProtoList[lower].Location);

	    if(location != NULL &&
	       (ProtoList[lower].Location = strdup(location)) == NULL)
		EGOTO(error);
	}

	return oldFlags;
    }

    if(NProtoList >= MProtoList) {
	REALLOC_RESIZE(MProtoList, 50);
	if((ProtoList = (prototypeaux_t *)realloc(ProtoList, MProtoList *
						  sizeof(*ProtoList))) == NULL)
	    EGOTO(error);
    }

    if(lower < NProtoList)
	memmove(&ProtoList[lower + 1], &ProtoList[lower],
		(NProtoList - lower) * sizeof(*ProtoList));
    ++NProtoList;

    ProtoList[lower].Flags = flags;
    if((ProtoList[lower].Function = strdup(function)) == NULL)
	EGOTO(error);

    if(location == NULL)
	ProtoList[lower].Location = NULL;
    else if((ProtoList[lower].Location = strdup(location)) == NULL)
	EGOTO(error);

    return 0;
 error:
    return -1;
}
