/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library
*
*    Part    : Print function dependencies.
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_NAME_FDEPS        "jsfdeps"
#define JSCPP_MODE_FDEPS        "fdeps"

/* ***  Various modes for doFDeps function  *** */
/*  Use jsmakedeps output style  */
#define JSCPP_FDEPS_STYLEDEPS     (1 << 0)
/*  Use topological sort output style  */
#define JSCPP_FDEPS_STYLETSORT    (1 << 1)
/*  Hide definitions from include file */
#define JSCPP_FDEPS_HIDEINC       (1 << 2)

/*  Also include dependencies on global/static variables  */
#define JSCPP_FDEPS_VARDEPS       (1 << 3)

/*  Include static functions and variables as well  */
#define JSCPP_FDEPS_STATICDEF     (1 << 4)

/*  Prefix static variables/functions with the function name to
 *   make them globally unique
 */
#define JSCPP_FDEPS_STATICPREFIX  (1 << 5)

/*  Global Variable names are prefixed with ''  */
#define JSCPP_FDEPS_VARPREFIX     (1 << 6)

/*  When showing called functions, do not avoid duplicates and do not sort  */
#define JSCPP_FDEPS_DUPLICATES    (1 << 7)

/*  Also show functions that start with __ or _[A-Z]  */
#define JSCPP_FDEPS_SHOWALL       (1 << 8)

/*  Exclude ANSI symbols from dependencies  */
#define JSCPP_FDEPS_ANSISYMS      (1 << 9)
/*  END-DEFINITIONS  */

typedef struct {
  int Mode;

  FILE *FP;

  /*  0: Defined functions or global/static variables.
   *  1: Called functions or used global/static variables.
   *  2: Identifiers to exclude in 1.
   */
  const char **Ident[3];
  int NIdent[3], MIdent[3];

  int PrintFlag;
  const char *FileName;
} fdepsaux_t;

#define STATIC_SEPARATOR         '@'
#define STATIC_SEPARATOR_LEN     1

#define VARIABLE_PREFIX          '$'

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doFDeps(jsparsenviron_t *ppenv, const char *names[],
				 int nnames, FILE *fpOut, int depsMode);
static int               parseFile(jsparsstream_t *jpa, fdepsaux_t *paux);
static int               doProcFDeps(jsparsstream_t *jpa,
				     jsparscnode_t *node, jctokfifo_t *queue,
				     fdepsaux_t *paux);
static int               doSingleDef(jsparsstream_t *jpa, const char *name,
				     jsparscnode_t *node, jctokfifo_t *queue,
				     fdepsaux_t *paux, int amode);
static int               scanFDeps(jsparsstream_t *jpa, jsparscnode_t *node,
				   jctokfifo_t *queue, fdepsaux_t *paux);
static int               addName(fdepsaux_t *paux, int nr, const char *name,
				 int mode);
static void              freeIdent(fdepsaux_t *paux, int nr, int mode);

static int               printDeps(fdepsaux_t *paux);
static void              sortIdent(const char **, unsigned long n);

/*JS*********************************************************************
*
*************************************************************************/

int doFDeps(jsparsenviron_t *ppenv, const char *names[], int nnames,
	    FILE *fpOut, int depsMode)

/************************************************************************/
{
  fdepsaux_t aux;
  int i, pmode;

  aux.Mode = depsMode;
  aux.FP = fpOut;

  aux.Ident[0] = aux.Ident[1] = aux.Ident[2] = NULL;
  aux.MIdent[0] = aux.MIdent[1] = aux.MIdent[2] = 0;
  aux.NIdent[0] = aux.NIdent[1] = aux.NIdent[2] = 0;

  pmode = JPAM_FVARTABLES;

  /*  Get meaningful warning messages  */
  pmode |= JPAM_FGETFPOS;

  if(!(depsMode & JSCPP_FDEPS_ANSISYMS)) {
    for(i = 0 ; ANSISymbols[i] != NULL ; ++i)
      if(addName(&aux, 2, ANSISymbols[i], 0) == -1)
        EGOTO(error);

    sortIdent(aux.Ident[2], aux.NIdent[2]);
  }

  for(i = 0 ; i < nnames ; ++i) {
    jsparsstream_t *jpa;
    int ret;

    if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
      EGOTO(error);

    aux.FileName = names[i];
    aux.PrintFlag = 0;

    if(depsMode & JSCPP_FDEPS_STYLEDEPS) {
      freeIdent(&aux, 0, 0);
      freeIdent(&aux, 1, 0);
    }

    ret = parseFile(jpa, &aux);

    if(depsMode & JSCPP_FDEPS_STYLEDEPS)
      printDeps(&aux);

    if(aux.PrintFlag == 0 && (depsMode & JSCPP_FDEPS_STYLEDEPS)) {
      /*  Print at least one line  */
      if(fprintf(fpOut, "%s:\n", names[i]) == -1)
	EGOTO(error);
    }

    if(closeParserStream(jpa, ret ? 2 : 0) == -1)
      EGOTO(error);

    if(ret == -1)
      EGOTO(error);
  }

  freeIdent(&aux, 0, 1);
  freeIdent(&aux, 1, 1);
  freeIdent(&aux, 2, 1);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int parseFile(jsparsstream_t *jpa, fdepsaux_t *paux)

/************************************************************************/
{
  jctokfifo_t *queue;

  queue = jsparsgetqueue(jpa);
  for(;;) {
    jsparscnode_t *node;

    if(jsparsget(jpa, &node, 0) == -1)
      EGOTO(error);

    if(node == NULL)
      break;

    if(doProcFDeps(jpa, node, queue, paux) == -1)
      EGOTO(error);

    if(jsparsflush(jpa, node, NULL, 0) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doProcFDeps(jsparsstream_t *jpa, jsparscnode_t *node,
		       jctokfifo_t *queue, fdepsaux_t *paux)

/************************************************************************/
{
  jsparscnode_t *sub;
  const char *name;
  int type;

  if((paux->Mode & JSCPP_FDEPS_HIDEINC) &&
     jsqsgetppstream(jpa->JSPA_Stream)->JSPP_NFiles != 1)
    /*  Function definitions stems from an include file  */
    return 0;

#if 0
  printf("--------------\n");
  jsparsprctree(stdout, queue, node, 0, 0);
#endif

  type = node->JPACN_Type;

  node = node->JPACN_Sub;
  if(type == JPACT_FUNCDEF) { /* ***  Function definition  *** */
    int isstatic;

    sub = node->JPACN_Sub;

    name = queue->TKFF_Token[sub->JPACN_Sub->JPACN_IndexPP].TK_String;

    node = node->JPACN_Next;

    /*  Check for static functions  */
    isstatic = FALSE;
    for(sub = sub->JPACN_Next ; sub ; sub = sub->JPACN_Next)
      if(sub->JPACN_Type == JPACT_STATIC) {
	isstatic = TRUE;
	break;
      }

    if(!isstatic) {
      jsparsvardef_t *vDef;

      /*  Even if the specifier doesn't have the "static" keyword,
       *   a preceding prototype might declare this function to be
       *   static.
       */
      if((vDef = jsparsgetvardef(jpa, jsparsgetglobalvartab(jpa),
				 name, JVDF_TFUNCTION, 0)) != NULL &&
	 (vDef->JPAVD_Flags & JVDF_ISTATIC))
	isstatic = TRUE;
    }

    if(isstatic && !(paux->Mode & JSCPP_FDEPS_STATICDEF)) {
      if(paux->Mode & JSCPP_FDEPS_STYLEDEPS)
	/*  We need the globals used by this static function, but not the
	 *   function name itself
	 */
	name = NULL;
      else
	/*  We should not scan static functions, so forget about it  */
	node = NULL;
    }

    if(node &&
       doSingleDef(jpa, name, node, queue, paux, isstatic ? 1 : 0) == -1)
      EGOTO(error);
  } else if((paux->Mode & JSCPP_FDEPS_VARDEPS) &&
	    type == JPACT_DECLARATION) { /* ***  Global definition  *** */
    int flag;

    /*  Check for "extern" in the specifier to avoid scanning declarations  */
    flag = 0;
    for(sub = node->JPACN_Next ; sub ; sub = sub->JPACN_Next)
      if(sub->JPACN_Type == JPACT_EXTERN)
	flag |= 1;
      else if(sub->JPACN_Type == JPACT_STATIC)
	flag |= 2;

    if(!(flag & 1) &&
       (!(flag & 2) || (paux->Mode & (JSCPP_FDEPS_STYLEDEPS |
				      JSCPP_FDEPS_STATICDEF)))) {
      /*  No pure declaration, so go ahead  */
      type = node->JPACN_Type;
      if(type == JPACT_LIST)
	node = node->JPACN_Sub;
      else if(type != JPACT_DECLARATOR && type != JPACT_ASS)
	/*  Maybe a global struct or union  */
	node = NULL;

      for( ; node ; node = node->JPACN_Next) {
	jsparscnode_t *decl;

	if(node->JPACN_Type == JPACT_ASS) {
	  decl = node->JPACN_Sub;
	  sub = decl->JPACN_Next;
	  decl = decl->JPACN_Sub;
	} else {
	  decl = node->JPACN_Sub;
	  /*  Only a global definition without initialization, no need to
	   *   scan this.
	   */
	  sub = NULL;
	}

	if(!(flag & 2) || (paux->Mode & JSCPP_FDEPS_STATICDEF))
	  name = queue->TKFF_Token[decl->JPACN_IndexPP].TK_String;
	else
	  name = NULL;

	/*  Check if it is only a function prototype  */
	decl = decl->JPACN_Next;
	if(decl == NULL || decl->JPACN_Type != JPACT_FUNCCALL) {
	  if(doSingleDef(jpa, name, sub, queue, paux,
			 (flag & 2) ? 3 : 2) == -1)
	    EGOTO(error);
	}

	if(type != JPACT_LIST)
	  break;
      }
    }
  }

  /*  Signal that nodes may be freed  */
  return 2;
error:
  return -1;
}

/*JS*********************************************************************
*   mode bit 0: name refers to static variable/function
*************************************************************************/

static int doSingleDef(jsparsstream_t *jpa, const char *name,
		       jsparscnode_t *node, jctokfifo_t *queue,
		       fdepsaux_t *paux, int amode)

/************************************************************************/
{
  if(!(paux->Mode & JSCPP_FDEPS_STYLEDEPS)) {
    /*  Output per function  */
    freeIdent(paux, 0, 0);
    freeIdent(paux, 1, 0);
  }

  if(name != NULL && addName(paux, 0, name, amode) == -1)
    EGOTO(error);

  /*  Recursive scan  */
  if(node && scanFDeps(jpa, node, queue, paux) == -1)
    EGOTO(error);

  if(!(paux->Mode & JSCPP_FDEPS_STYLEDEPS)) {
    /*  The recursion above has collected all called functions,
     *   so output them now.
     */
    if(printDeps(paux) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

#if defined CHECKFUNCARG_FUNCTIONS && defined CHECKFUNCARG_ARRTYPELEN
/*JS*********************************************************************
*
*************************************************************************/

static int getLength(jctokfifo_t *queue, jsparscnode_t *node)

/************************************************************************/
{
  int len;

  if(node->JPACN_Type == JPACT_INTCONST)
    return atoi(queue->TKFF_Token[node->JPACN_End].TK_String);

  if(node->JPACN_Type == JPACT_UNARYEXPR)
    return getLength(queue, node->JPACN_Sub);

  if(node->JPACN_Type != JPACT_ADD)
    return -1;

  len = 0;
  for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next) {
    int l;

    if((l = getLength(queue, node)) == -1)
	return -1;
     len += l;
  }

  return len;
}
#endif

#ifdef CHECKFUNCARG_FUNCTIONS
/*JS*********************************************************************
*
*************************************************************************/

static int checkIdent(const char *string, void *user)

/************************************************************************/
{
  if(strcmp(string, (const char *)user) == 0)
    return 1;

  return 0;
}
#endif

/*JS*********************************************************************
*
*************************************************************************/

static int scanFDeps(jsparsstream_t *jpa, jsparscnode_t *node,
		     jctokfifo_t *queue, fdepsaux_t *paux)

/************************************************************************/
{
  jsparscnode_t *next;
  int type;

  type = node->JPACN_Type;

  for(node = node->JPACN_Sub ; node ; node = next) {
    next = node->JPACN_Next;

    if(node->JPACN_Type == JPACT_IDENT) {
      const char *ident;

      ident = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

#ifdef CHECKFUNCARG_FUNCTIONS
      /*  Special coding: We check if some function(s) are called where the
       *   first argument is of a given type or a given array type with at
       *   most a given length.
       */
      if(jsstringsep(CHECKFUNCARG_FUNCTIONS, ',', checkIdent, (void *)ident) > 0) {
	jsparscnode_t *n, *n2;

	n = node->JPACN_Next;
	if(n->JPACN_Type == JPACT_FUNCCALL) {
	  jsparsexprtype_t eType;
	  jsparsvartab_t *lTab;

	  n = n->JPACN_Sub;

	  if((lTab = jsparsfindvartab(jpa, n)) == NULL)
	    EGOTO(error);

	  if(jsparsgetexprtype(&eType, jpa, lTab, n, queue, 0) == NULL)
	    EGOTO(error);

	  if((n = eType.JSPE_Specifier) != NULL && n->JPACN_Type == JPACT_TYPEDEFNAME) {
	    const char *typeName;

	    typeName = eType.JSPE_Queue->TKFF_Token[n->JPACN_IndexPP].TK_String;
#ifdef CHECKFUNCARG_TYPENAME
	    if(strcmp(typeName, CHECKFUNCARG_TYPENAME) == 0) {
	      jsparsprinterr(jpa, queue, node, "FOUND %s(%s, ...)!\n", 0,
		     CHECKFUNCARG_FUNCTIONS, CHECKFUNCARG_TYPENAME);
	    }
#endif

#ifdef CHECKFUNCARG_ARRTYPENAME
	    if(strcmp(typeName, CHECKFUNCARG_ARRTYPENAME) == 0 &&
		      (n2 = eType.JSPE_Declarator) != NULL &&
		      n2->JPACN_Type == JPACT_VECTOR) {
	      int length;

	      if((length = getLength(eType.JSPE_Queue, n2->JPACN_Sub)) == -1) {
		jsparsprinterr(jpa, queue, node, "??? %s(%s,...)\n", 0,
			       CHECKFUNCARG_FUNCTIONS, CHECKFUNCARG_ARRTYPENAME);

		jsparsprctree(stdout, eType.JSPE_Queue, n2, 0, 0);
	      } else if(length < CHECKFUNCARG_ARRTYPELEN) {
		jsparsprinterr(jpa, queue, node, "FOUND %s(%s [%d], ...)!\n", 0,
		       CHECKFUNCARG_FUNCTIONS, CHECKFUNCARG_ARRTYPENAME,
		       CHECKFUNCARG_ARRTYPELEN);
	      }
	    }
#endif
	  }

	  jsparsfreeexprtype(jpa, &eType, 1);
	}
      }
#endif

      if(type == JPACT_UNARYEXPR) {
	if((paux->Mode & JSCPP_FDEPS_SHOWALL) ||
	   !(ident[0] == '_' && (ident[1] == '_' ||
				 isupper(ident[1])))) {
	  jsparsvartab_t *vTab;
	  jsparsvardef_t *vDef;
	  int flagAdd, amode;

	  if((vTab = jsparsfindvartab(jpa, node)) == NULL)
	    EGOTO(error);

	  flagAdd = FALSE;
	  amode = 0;
	  if((vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TVARIABLE,
				     0)) != NULL) {
	    /*  The name is known as a variable, so check if
	     *   we reference a global/static variable.
	     */
	    if((paux->Mode & JSCPP_FDEPS_VARDEPS) &&
	       (vDef->JPAVD_Flags & JVDF_FGLOBAL)) {
	      if(vDef->JPAVD_Flags & JVDF_ISTATIC) {
		if(paux->Mode & JSCPP_FDEPS_STATICDEF) {
		  flagAdd = TRUE; /*  static variable  */
		  amode |= 3;
		}
	      } else {
		flagAdd = TRUE; /*  global variable  */
		amode |= 2;
	      }
	    }
	  } else if((vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TFUNCTION,
					    0)) != NULL) {
	    /*  We know a function like this  */
	    if(!(vDef->JPAVD_Flags & JVDF_ISTATIC))
	      flagAdd = TRUE; /*  global function  */
	    else if(paux->Mode & JSCPP_FDEPS_STATICDEF) {
	      flagAdd = TRUE; /*  static function  */
	      amode |= 1;
	    }

	  } else if((vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TENUMIDENT,
					    0)) == NULL) {
	    if(type == JPACT_UNARYEXPR &&
	       next && next->JPACN_Type == JPACT_FUNCCALL) {
	      /*  Missing prototype: assume global function  */
	      flagAdd = TRUE;
	    }

	    if(jpa->JSPA_WarnFlags & JPAM_WARNVARS)
	      jsparsprinterr(jpa, queue, node, "Undeclared %s '%s'",
			     0, flagAdd ? "function" : "identifier", ident);
	  }

	  if(flagAdd && addName(paux, 1, ident, amode) == -1)
	    EGOTO(error);
	}
      }
    }

    /*  Recursive call  */
    if(scanFDeps(jpa, node, queue, paux) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   mode bit 0: static identifier, postfix with '@filename'
*        bit 1: Identifier refers to variable, not function.
*************************************************************************/

static int addName(fdepsaux_t *paux, int nr, const char *name, int mode)

/************************************************************************/
{
  char *string;
  int len;

  if(nr == 1) {
    int lower, upper;

    /*  Exclude symbols only refer to referenced symbols - if we define
     *   something from the ANSI namespace, we'll show it.
     */
    lower = 0;  upper = paux->NIdent[2];
    while(lower < upper) {
      int m = (lower + upper) >> 1;
      if(strcmp(paux->Ident[2][m], name) < 0)
        lower = m + 1;
      else
        upper = m;
    }

    if(lower < paux->NIdent[2] && strcmp(paux->Ident[2][lower], name) == 0)
        return 0;
  }

  if(paux->NIdent[nr] >= paux->MIdent[nr]) {
    REALLOC_RESIZE(paux->MIdent[nr], 200);
    if((paux->Ident[nr] = (const char **)
	realloc((void *)paux->Ident[nr], paux->MIdent[nr] *
		sizeof(*(paux->Ident[nr])))) == NULL)
      EGOTO0(error);
  }

  if((mode & 1) && (paux->Mode & JSCPP_FDEPS_STATICPREFIX)) {
    len = strlen(name) + strlen(paux->FileName) + STATIC_SEPARATOR_LEN + 1;
  } else {
    len = strlen(name) + 1;
    mode &= ~1;
  }

  if((mode & 2) && (paux->Mode & JSCPP_FDEPS_VARPREFIX))
    ++len;
  else
    mode &= ~2;

  if((string = (char *)malloc(len)) == NULL)
    EGOTO(error);

  paux->Ident[nr][paux->NIdent[nr]++] = string;

  if(mode & 2)
    *string++ = VARIABLE_PREFIX;

  if(mode & 1)
    sprintf(string, "%s%c%s", name, STATIC_SEPARATOR, paux->FileName);
  else
    strcpy(string, name);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void freeIdent(fdepsaux_t *paux, int nr, int mode)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < paux->NIdent[nr] ; ++i)
    free((void *)paux->Ident[nr][i]);

  paux->NIdent[nr] = 0;

  if(mode) {
    free((void *)paux->Ident[nr]);
    paux->Ident[nr] = NULL;
    paux->MIdent[nr] = 0;
  }
}

/*JS*********************************************************************
*
*************************************************************************/

static int printDeps(fdepsaux_t *paux)

/************************************************************************/
{
  const char *caller, *called, *prev;
  int i;

  if(!(paux->Mode & JSCPP_FDEPS_DUPLICATES))
    sortIdent(paux->Ident[1], paux->NIdent[1]);

  if(paux->Mode & JSCPP_FDEPS_STYLEDEPS) {
    paux->PrintFlag++;

    if(fprintf(paux->FP, "%s:", paux->FileName) == -1)
      EGOTO(error);

    /*  Make definitions appear in sorted order  */
    sortIdent(paux->Ident[0], paux->NIdent[0]);

    for(i = 0 ; i < paux->NIdent[0] ; ++i) {
      caller = paux->Ident[0][i];

      if(fprintf(paux->FP, " >%s", caller) == -1)
	EGOTO(error);
    }
  } else
    /*  There is only *one* definition  */
    caller = paux->Ident[0][0];

  if(!(paux->Mode & (JSCPP_FDEPS_STYLEDEPS|JSCPP_FDEPS_STYLETSORT)) &&
     fprintf(paux->FP, "%s:", caller) == -1)
    EGOTO(error);

  /*  Print the dependent functions, possibly avoiding duplicates  */
  for(prev = NULL, i = 0 ; i < paux->NIdent[1] ; ++i, prev = called) {
    called = paux->Ident[1][i];
    if((paux->Mode & JSCPP_FDEPS_DUPLICATES) || prev == NULL ||
       strcmp(prev, called) != 0) {
      int ret;

      if(paux->Mode & JSCPP_FDEPS_STYLEDEPS)
	ret = fprintf(paux->FP, " <%s", called);
      else if(paux->Mode & JSCPP_FDEPS_STYLETSORT)
	ret = fprintf(paux->FP, "%s %s\n", caller, called);
      else
	ret = fprintf(paux->FP, " %s", called);

      if(ret == -1)
	EGOTO(error);
    }
  }

  if((paux->Mode & JSCPP_FDEPS_STYLEDEPS) ||
     !(paux->Mode & JSCPP_FDEPS_STYLETSORT)) {
    if(putc('\n', paux->FP) == EOF)
      EGOTO(error);
  } else if(paux->NIdent[1] == 0) {
    /*  Make sure every defined routines appears at least once  */
    if(fprintf(paux->FP, "%s\n", caller) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

/*  Generated automatically with "gensort.h"  */
#define GENSORT_NAME		     sortIdent
#define GENSORT_TYPE		     const char *
#define GENSORT_KEYTYPE 	     const char *
#define GENSORT_GETKEY(a)            a
#define GENSORT_COMPAREKEYS(k1, k2)  strcmp(k1, k2) < 0
#define GENSORT_STATIC

#include <gensort.h>
