/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Add tracing calls to functions.
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
#ifndef JSCPP_FTRACE_SYMBOL
/*  START-DEFINITIONS  */
#define JSCPP_FTRACE_SYMBOL     "__JSFTRACE__"

/*  When called like this, all functions are traced as default  */
#define JSCPP_NAME_FTRACE       "jsftrace"
#define JSCPP_MODE_FTRACE       "ftrace"

/*  When called like this, nothing is traced/changed as default  */
#define JSCPP_NAME_TRACE        "jstrace"
#define JSCPP_MODE_TRACE        "trace"

/*  When called like this, all functions are traced by wrapping them  */
#define JSCPP_NAME_FCTWRAP      "jsfctwrap"
#define JSCPP_MODE_FCTWRAP      "fctwrap"

/* ***  Various modes for doFTrace function  *** */
/*  Emit trace call on start of function  */
#define JSCPP_TRACE_FCTCALL           (1 <<  0)
/*  Start trace calls contain functions arguments  */
#define JSCPP_TRACE_FCTCALLARGS       (1 <<  1)
/*  Function arguments are labeled as default  */
#define JSCPP_TRACE_FCTCALLLBLD       (1 <<  2)

/*  Emit trace call on return statements  */
#define JSCPP_TRACE_FCTRET            (1 <<  4)
/*  The "return" trace calls contain return value  */
#define JSCPP_TRACE_FCTRETARG         (1 <<  5)
/*  Label return values as well  */
#define JSCPP_TRACE_FCTRETLBLD        (1 <<  6)
/*  The "return" trace calls contain the line  */
#define JSCPP_TRACE_FCTRETLINE        (1 <<  7)

/*  Produce trace calls for each statement  */
#define JSCPP_TRACE_STMTTRACE         (1 << 12)
/*  Produce trace calls only for statements where flow branches  */
#define JSCPP_TRACE_STMTTRACE_BRANCH  (1 << 13)

/*  Produce code coverage  */
#define JSCPP_TRACE_CODECOVERAGE      (1 << 14)

/*  Generate wrapper functions for each function call and redirect
 *   them
 */
#define JSCPP_TRACE_FCTWRAP           (1 << 16)

/*  To recognize that wrapper generator has to be called  */
#define JSCPP_TRACE_NEEDWRAPPING      (1 << 17)

/*  Use internal mapping of types to trace format  */
#define JSCPP_TRACE_INTTYPEMAPPING    (1 << 20)

/*  Print prototype of trace function at the start  */
#define JSCPP_TRACE_PROTOTYPEFCT      (1 << 21)
/*  Print prototype for variablel trace function  */
#define JSCPP_TRACE_PROTOTYPEVAR      (1 << 22)
/*  Print prototype for statement trace function  */
#define JSCPP_TRACE_PROTOTYPESTMT     (1 << 23)
/*  Print prototype for code coverage function + array  */
#define JSCPP_TRACE_PROTOTYPECOV      (1 << 24)

/*  Yield preprocessed tokens instead of original ones  */
#define JSCPP_TRACE_OUTPREPROC        (1 << 25)
/*  Expand system include files  */
#define JSCPP_TRACE_EXPUSERINCS       (1 << 26)
/*  Expand macros defined in user include files  */
#define JSCPP_TRACE_EXPUSERMACS       (1 << 27)

/*  outName points to a directory  */
#define JSCPP_TRACE_DIRECTORY         (1 << 29)

/*  Generate 'line' directive at the start to identify sources
 *   correctly
 */
#define JSCPP_TRACE_LINEDIRECTIVE     (1 << 30)

/* ********************************************************** */
/*  If this is defined and we have a trace pragma with explicit expressions
 *   and no label is given, the expression is used as a label.
 *   E. g. "call(arg * 5)" will use the label "arg * 5".
 *   If you want to suppress a label, specify an empty one like ""
 */
#define CONFIG_JSPARSE_ARGLABELS

typedef struct {
  const char *Label;
  const char *Expression;

  /*  Contains the tokens to read in the expression  */
  jsqueuestream_t *QStream;
} argdesc_t;

typedef struct {
  int Type;

  const char **Patterns;
  int NPatterns;

  argdesc_t *CallDescriptions;
  int NCallDescriptions; /*  -1: patch in the standard way  */

  argdesc_t *ReturnDescriptions;
  int NReturnDescriptions; /*  -1: patch in the standard way  */
} patchdesc_t;

typedef struct {
  const char *Pattern;
  int Level;
} tracelevel_t;

/*  Maps one expression type to an expression  */
typedef struct {
  jsparsexprtype_t ExprType;

  argdesc_t *Descriptions;
  int NDescriptions;
} typemapping_t;

typedef struct {
  const char *TypeName;
  const char *SourceName, *TargetName;

} typerename_t;

/*  Maps a given type to a character for the format string of the
 *   trace function
 */
typedef struct {
  /*
   *  Bit 0: ExprType not used, this format is for pointers
   *  Bit 1: Never use a label.
   *  Bit 2: Always use a label.
   */
  int FFlag;

  jsparsexprtype_t ExprType;

  const char *Format[2];
} typeformat_t;

typedef struct {
  const char *FuncPattern;
  const char *VarPattern;

  long Flag; /*  bit 0: trace read access.
	      *  bit 1: trace write access.
	      *  bit 8: only global variables matching the pattern.
	      *  bit 9: only local variables matching the pattern.
	      */
} vartrace_t;

typedef struct {
  const char *Name;
  int Flag;
} wrapfunc_t;

#define VARTRACE_FREAD      1
#define VARTRACE_FWRITE     2
#define VARTRACE_FGLOBAL  256
#define VARTRACE_FLOCAL   512

#define PDESC_TYPE_TRACE              1
#define PDESC_TYPE_MASK            0xff

#define PDESC_TYPE_FCALL          (1 <<  8)
#define PDESC_TYPE_FCALLLBL       (1 <<  9)

#define PDESC_TYPE_FRETURN        (1 << 12)
#define PDESC_TYPE_FRETURNLINE    (1 << 13)
#define PDESC_TYPE_FRETURNLBL     (1 << 14)

#define PDESC_TYPE_FALLSTMT       (1 << 16) /*  trace all statements  */
#define PDESC_TYPE_FBRANCHSTMT    (1 << 17) /*  trace only after branches  */

#define PDESC_TYPE_FCODECOVERAGE  (1 << 18) /*  trace only after branches  */

#define PDESC_TYPE_FFCTWRAP       (1 << 20)

typedef struct {
  int Mode;
  int PrototypeFlag; /*  marks which prototypes were generated  */

  const char *CurrentFile;

  /*  If non-zero, use proper indentation for trace statements,
   *   possibly inserting newlines
   */
  int Indentation;

  int MaxMappingLevel;
  int MaxRecursiveLevel;

  jsparscnode_t **NodeStack;
  int INodeStack, MNodeStack;

  int *ExpandMapStack;
  int IExpandMapStack, MExpandMapStack;

  patchdesc_t *Descriptions;
  int NDescriptions, MDescriptions;

  wrapfunc_t *WrappedFunctions;
  int NWrappedFunctions, MWrappedFunctions;

  tracelevel_t *TraceLevels;
  int NTraceLevels, MTraceLevels;

  typemapping_t *TypeMappings;
  int NTypeMappings, MTypeMappings;

  typerename_t *TypeRenames;
  int NTypeRenames, MTypeRenames;

  typeformat_t *TypeFormats;
  int NTypeFormats, MTypeFormats;

  const char *TraceFormatCall;
  const char *TraceFormatReturn;
  const char *TraceFormatStmt;

  const char *NameRetVariable;
  const char *TraceVarFunction;

  const char *LabelFormat;

  vartrace_t *VarTraces;
  int NVarTraces, MVarTraces;

  const char *PrefixWrapFunction;
  const char *PrefixWrapFunction2;

  int CurrentVarNumber;

  /*  For code coverage  */
  int *CovRange;
  int NCovRange, MCovRange;

} traceaux_t;

#define CALLTRACE_FUNCTION   "_jstracefunc"

/*  END-DEFINITIONS  */
#endif

#define TRACE_FORMAT_CALL    CALLTRACE_FUNCTION"(\"%m%n\",\"%f\"%a)"
#define TRACE_FORMAT_RETURN  CALLTRACE_FUNCTION"(\"%m%P%n\",\"%f\"%p%a)"
#define TRACE_FORMAT_STMT    STMTTRACE_FUNCTION"(%F,%L)"

#define NAME_RETVALUE        "_retvalue"
#define VARTRACE_FUNCTION    "_jstracevar"
#define PREFIX_WRAPFUNCTION  "_jswrap_"

#define LABEL_FORMAT         "\"%s\""

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doFTrace(jsparsenviron_t *ppenv, const char *names[],
				  int nnames, const char *outName, int mode);
static int               traceFile(jsparsstream_t *jpa, const char *fileName,
				   FILE *fpOut, int mode);

static void              freetraceaux(traceaux_t *paux);
Prototype void           freepatchdesc(patchdesc_t *pdesc);
static void              freevartrace(vartrace_t *vtrace);
static void              freeargdesc(argdesc_t *adesc);

#ifdef CONFIG_JSCPPLIB_DEBUG
void                     showpatchdesc(const patchdesc_t *pdesc);
void                     showargdesc(const argdesc_t *adesc);
void                     showexprtype(const jsparsexprtype_t *etype);
void                     showvartrace(const vartrace_t *vtrace);
#endif

Prototype jstrbuff_t      SBuffTypes, SBuffArgs;
Prototype int             DefaultIndentation;

/*********     GLOBAL / STATIC VARIABLES                        *********/
static const char *TraceIntroducer[] = {
  /*  We allow both identifiers  */
  JSCPP_MACRO, JSCPP_FTRACE_SYMBOL, NULL
};

jstrbuff_t SBuffTypes, SBuffArgs;
int DefaultIndentation;

/*JS*********************************************************************
*
*************************************************************************/

int doFTrace(jsparsenviron_t *ppenv, const char *names[], int nnames,
	     const char *outName, int mode)

/************************************************************************/
{
  FILE *fpOut;
  int pmode, i;

  if(outName == NULL)
    fpOut = stdout;
  else if(!(mode & JSCPP_TRACE_DIRECTORY) &&
	  (fpOut = fopen(outName, "w")) == NULL) {
    fprintf(stderr, "ERROR: Cannot open '%s' - Break.\n", outName);
    EGOTO(error);
  }

  JSPPPragmaIntroducer = TraceIntroducer;

  pmode = JPAM_FVARTABLES | JPAM_FSTMTFLOW;
  if(mode & (JSCPP_TRACE_EXPUSERMACS | JSCPP_TRACE_EXPUSERINCS))
    pmode |= JPAM_FGETORIG | JPAM_FPRINTORIG | JPAM_FEXPANDUSERFILES;
  else if(mode & JSCPP_TRACE_OUTPREPROC)
    pmode |= JPAM_FGETFPOS;
  else
    pmode |= JPAM_FGETORIG | JPAM_FPRINTORIG;

  /*  Get meaningful warning messages in case of syntax errors  */
  pmode |= JPAM_FGETFPOS;

  for(i = 0 ; i < nnames ; ++i) {
    jsparsstream_t *jpa;
    jsppstream_t *jps;
    int ret;

    if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
      EGOTO(error);

    jps = jsqsgetppstream(jpa->JSPA_Stream);

    if(mode & JSCPP_TRACE_EXPUSERMACS)
      jps->JSPP_Mode |= JPPM_EXPUSERMACROS;

    if(outName && (mode & JSCPP_TRACE_DIRECTORY) &&
       (fpOut = openFile(outName, names[i])) == NULL)
      EGOTO(error);

    if(mode & JSCPP_TRACE_LINEDIRECTIVE)
      fprintf(fpOut, "#line 1 \"%s\"\n", names[i]);

    ret = traceFile(jpa, names[i], fpOut, mode);

    if(closeParserStream(jpa, ret ? 2 : 0) == -1)
      EGOTO(error);

    if(ret == -1)
      EGOTO(error);

    if(outName && (mode & JSCPP_TRACE_DIRECTORY) && fclose(fpOut) == EOF)
      EGOTO(error);
  }

  if(outName && !(mode & JSCPP_TRACE_DIRECTORY) && fclose(fpOut) == EOF)
    EGOTO(error);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int traceFile(jsparsstream_t *jpa, const char *fileName,
		     FILE *fpOut, int mode)

/************************************************************************/
{
  traceaux_t aux;
  jctokfifo_t *queue;

  if(mode & JSCPP_TRACE_FCTWRAP) mode |= JSCPP_TRACE_NEEDWRAPPING;

  aux.Mode = mode;
  aux.PrototypeFlag = 0;
  aux.Indentation = DefaultIndentation;

  aux.MaxMappingLevel = -1;
  aux.MaxRecursiveLevel = 1;

  aux.NodeStack = NULL;      aux.INodeStack      = aux.MNodeStack = 0;
  aux.ExpandMapStack = NULL; aux.IExpandMapStack = aux.MExpandMapStack = 0;
  aux.Descriptions = NULL;   aux.NDescriptions   = aux.MDescriptions = 0;
  aux.WrappedFunctions = NULL; aux.NWrappedFunctions = aux.MWrappedFunctions = 0;
  aux.TraceLevels = NULL;    aux.NTraceLevels    = aux.MTraceLevels = 0;
  aux.TypeMappings = NULL;   aux.NTypeMappings   = aux.MTypeMappings = 0;
  aux.TypeRenames = NULL;    aux.NTypeRenames    = aux.MTypeRenames = 0;
  aux.TypeFormats = NULL;    aux.NTypeFormats    = aux.MTypeFormats = 0;
  aux.VarTraces = NULL;      aux.NVarTraces      = aux.MVarTraces = 0;
  aux.CurrentVarNumber = 0;

  aux.CovRange = NULL;
  aux.NCovRange = aux.MCovRange = 0;

  if((aux.TraceFormatCall     = strdup(TRACE_FORMAT_CALL)) == NULL ||
     (aux.TraceFormatReturn   = strdup(TRACE_FORMAT_RETURN)) == NULL ||
     (aux.TraceFormatStmt     = strdup(TRACE_FORMAT_STMT)) == NULL ||
     (aux.NameRetVariable     = strdup(NAME_RETVALUE)) == NULL ||
     (aux.PrefixWrapFunction  = strdup(PREFIX_WRAPFUNCTION)) == NULL ||
     (aux.PrefixWrapFunction2 = strdup(PREFIX_WRAPFUNCTION "%")) == NULL ||
     (aux.TraceVarFunction    = strdup(VARTRACE_FUNCTION)) == NULL ||
     (aux.LabelFormat         = strdup(LABEL_FORMAT)) == NULL)
    EGOTO(error);

  queue = jsparsgetqueue(jpa);
  for(;;) {
    jsparscnode_t *node;

    if(jsparsget(jpa, &node, 0) == -1)
      EGOTO(error);

    if(node == NULL)
      break;

    aux.CurrentFile = fileName;
    if(doProcFTrace(&aux, jpa, node, queue) == -1)
      EGOTO(error);

    if(aux.NTypeRenames > 0 &&
       doProcTypeRename(&aux, jpa, node, queue) == -1)
        EGOTO(error);

    if(jsparsflush(jpa, node, fpOut, 0) == -1)
      EGOTO(error);
  }

  /*  Flush any trailing tokens from last jsparsget
   *   (e. g. trailing comments)
   */
  if(jsparsflushtokens(jpa, fpOut) == -1)
    EGOTO(error);

  if(aux.Mode & JSCPP_TRACE_CODECOVERAGE) {
      int i;

      if(fprintf(fpOut, "\nlong %s[%d];\n", CODECOVER_COUNTARRAY, aux.NCovRange > 0 ?
		 aux.NCovRange : 1) == -1)
	  EGOTO(error);

      if(fprintf(fpOut, "int %s[] = {", CODECOVER_DESCRIPTARRAY) == -1)
	  EGOTO(error);

      for(i = 0 ; i < aux.NCovRange ; ++i) {
	  if((i % 10) == 0 && fprintf(fpOut, "\n  ") == -1)
	      EGOTO(error);

	  if(fprintf(fpOut, " %d, %d,", aux.CovRange[2 * i],
		     aux.CovRange[2 * i + 1]) == -1)
	      EGOTO(error);
      }

      if(fprintf(fpOut, "\n};\n") == -1)
	  EGOTO(error);
  }

  freetraceaux(&aux);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static void freetraceaux(traceaux_t *paux)

/************************************************************************/
{
  int i, j;

  free(paux->NodeStack);
  free(paux->ExpandMapStack);

  for(i = 0 ; i < paux->NDescriptions ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1)
      showpatchdesc(&(paux->Descriptions[i]));
#endif

    freepatchdesc(&(paux->Descriptions[i]));
  }
  free(paux->Descriptions);

  for(i = 0 ; i < paux->NWrappedFunctions ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1)
      printf("Wrapped function '%s'\n", paux->WrappedFunctions[i].Name);
#endif
    free((void *)paux->WrappedFunctions[i].Name);
  }
  free(paux->WrappedFunctions);

  for(i = 0 ; i < paux->NTraceLevels ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1)
      printf("Pattern '%s' level %d\n", paux->TraceLevels[i].Pattern,
	     paux->TraceLevels[i].Level);
#endif
    free((void *)paux->TraceLevels[i].Pattern);
  }

  free(paux->TraceLevels);

  for(i = 0 ; i < paux->NTypeMappings ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1) {
      printf("TYPEMAP #%d: %d\n", i, paux->TypeMappings[i].NDescriptions);
      showexprtype(&(paux->TypeMappings[i].ExprType));
      printf("  EXPRESSIONS %d:\n", paux->TypeMappings[i].NDescriptions);
    }
#endif

    for(j = 0 ; j < paux->TypeMappings[i].NDescriptions ; ++j) {
#ifdef CONFIG_JSCPPLIB_DEBUG
      if(Debug & 1)
	showargdesc(&(paux->TypeMappings[i].Descriptions[j]));
#endif

      freeargdesc(&(paux->TypeMappings[i].Descriptions[j]));
    }
  }
  free(paux->TypeMappings);

  for(i = 0 ; i < paux->NTypeFormats ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1) {
      printf("TYPEFORMAT #%d: flag %d format '%s' '%s'\n", i,
	     paux->TypeFormats[i].FFlag,
	     paux->TypeFormats[i].Format[0] ?
	     paux->TypeFormats[i].Format[0] : "<null>",
	     paux->TypeFormats[i].Format[1] ?
	     paux->TypeFormats[i].Format[1] : "<null>");

      if((paux->TypeFormats[i].FFlag & 1) == 0)
	showexprtype(&(paux->TypeFormats[i].ExprType));
    }
#endif

    for(j = 0 ; j < 2 ; ++j)
      if(paux->TypeFormats[i].Format[j])
	free((void *)(paux->TypeFormats[i].Format[j]));
  }
  free(paux->TypeFormats);

  for(i = 0 ; i < paux->NVarTraces ; ++i) {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 1)
      showvartrace(&(paux->VarTraces[i]));
#endif
    freevartrace(&(paux->VarTraces[i]));
  }

  free((void *)paux->TraceFormatCall);
  free((void *)paux->TraceFormatReturn);
  free((void *)paux->TraceFormatStmt);
  free((void *)paux->NameRetVariable);
  free((void *)paux->PrefixWrapFunction);
  free((void *)paux->PrefixWrapFunction2);
  free((void *)paux->TraceVarFunction);
  free((void *)paux->LabelFormat);

  free(paux->VarTraces);

  free(paux->CovRange);
}

/*JS*********************************************************************
*
*************************************************************************/

void freepatchdesc(patchdesc_t *pdesc)

/************************************************************************/
{
  int i;

  if(pdesc->Patterns) {
    for(i = 0 ; i < pdesc->NPatterns ; ++i)
      free((void *)pdesc->Patterns[i]);

    free((void *)pdesc->Patterns);
  }

  if(pdesc->CallDescriptions) {
    for(i = 0 ; i < pdesc->NCallDescriptions ; ++i)
      freeargdesc(&(pdesc->CallDescriptions[i]));

    free(pdesc->CallDescriptions);
  }

  if(pdesc->ReturnDescriptions) {
    for(i = 0 ; i < pdesc->NReturnDescriptions ; ++i)
      freeargdesc(&(pdesc->ReturnDescriptions[i]));

    free(pdesc->ReturnDescriptions);
  }
}

/*JS*********************************************************************
*
*************************************************************************/

static void freevartrace(vartrace_t *vtrace)

/************************************************************************/
{
  if(vtrace->FuncPattern)
    free((void *)vtrace->FuncPattern);
  else
    free((void *)vtrace->VarPattern);
}

/*JS*********************************************************************
*
*************************************************************************/

static void freeargdesc(argdesc_t *adesc)

/************************************************************************/
{
  free((void *)adesc->Label);
  free((void *)adesc->Expression);

  free(adesc->QStream->JSQS_User);

  jsqsfree(adesc->QStream);
}

#ifdef CONFIG_JSCPPLIB_DEBUG
/*JS*********************************************************************
*
*************************************************************************/

void showpatchdesc(const patchdesc_t *pdesc)

/************************************************************************/
{
  int i;

  printf("Type: %#x\n", pdesc->Type);

  if(pdesc->Patterns)
    for(i = 0 ; i < pdesc->NPatterns ; ++i)
	printf("  PATTERN #%d '%s'\n", i, pdesc->Patterns[i]);

  printf("  NCALL %d\n", pdesc->NCallDescriptions);
  if(pdesc->CallDescriptions)
    for(i = 0 ; i < pdesc->NCallDescriptions ; ++i)
      showargdesc(&(pdesc->CallDescriptions[i]));

  printf("  NRETURN %d\n", pdesc->NReturnDescriptions);
  if(pdesc->ReturnDescriptions)
    for(i = 0 ; i < pdesc->NReturnDescriptions ; ++i)
      showargdesc(&(pdesc->ReturnDescriptions[i]));
}

/*JS*********************************************************************
*
*************************************************************************/

void showargdesc(const argdesc_t *adesc)

/************************************************************************/
{
  jctokfifo_t *queue;
  int ind;

  printf("    LABEL '%s' EXPRESSION '%s' TOKENS <",
	 adesc->Label ? adesc->Label : "<no_label>",
	 adesc->Expression ? adesc->Expression : "<no_expression>");

  tokseqrewind((tokenseq_t *)adesc->QStream->JSQS_User);

  queue = jsqsgetqueue(adesc->QStream);
  while((ind = jsqsget(adesc->QStream, 0)) != -2)
    printCToken(stdout, queue->TKFF_Token[ind]);
  printf(">\n");
}


/*JS*********************************************************************
*
*************************************************************************/

void showexprtype(const jsparsexprtype_t *etype)

/************************************************************************/
{
  printf("  DECLARATOR:\n");
  jsparsprctree(stdout, etype->JSPE_Queue, etype->JSPE_Declarator, 4, 4);
  printf("  SPECIFIER:\n");
  jsparsprctree(stdout, etype->JSPE_Queue, etype->JSPE_Specifier, 4, 0);
}

/*JS*********************************************************************
*
*************************************************************************/

void showvartrace(const vartrace_t *vtrace)

/************************************************************************/
{
  printf("Variable '%s' '%s' flag 0x%lx\n",
	 vtrace->FuncPattern ? vtrace->FuncPattern : "<NULL>",
	 vtrace->VarPattern  ? vtrace->VarPattern  : "<NULL>",
	 vtrace->Flag);
}
#endif
