/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Add tracing calls to functions.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES                                          *********/
typedef struct {
    traceaux_t *Paux;
    jsparsstream_t *JPA;
    jsparscnode_t *Tree;
    jctokfifo_t *Queue;
} wrapaux_t;

#define GENPARAM_PATTERN       "p%d"

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doProcFTrace(traceaux_t *paux, jsparsstream_t *jpa,
                                      jsparscnode_t *node, jctokfifo_t *queue);

Prototype int            doProcTypeRename(traceaux_t *paux, jsparsstream_t *jpa,
                                          jsparscnode_t *node, jctokfifo_t *queue);
static int               doTypeRenameExpr(traceaux_t *paux, jsparsstream_t *jpa,
                                          jsparscnode_t *node, jctokfifo_t *queue);
static int               doTypeRenameDef(traceaux_t *paux, jsparsstream_t *jpa,
                                         jsparscnode_t *node, jctokfifo_t *queue);

static int               findFirstStatement(traceaux_t *paux,
                                            jsparsstream_t *jpa,
                                            jsparscnode_t *body,
                                            jctokfifo_t *queue,
                                            jsparscnode_t **pVarNode,
                                            jsparscnode_t **pExpression);
static int               genPrototype(traceaux_t *paux, jsparsstream_t *jpa,
                                      jctokfifo_t *queue, jsparscnode_t *node,
                                      long type, int mode);

Prototype int            doGetName(traceaux_t *paux, jsparscnode_t *node,
                                   const char **pName, jctokfifo_t *queue,
                                   int mode);
static const patchdesc_t *checkRoutine(traceaux_t *paux, const char *funcName);

Prototype int            checkPattern(const char *pattern,
                                      const char *fileName,
                                      const char *funcName);

static char             *genFctEntryCall(traceaux_t *paux, jsparsstream_t *jpa,
                                         jctokfifo_t *queue,
                                         const patchdesc_t *pdesc,
                                         jsparsvartab_t *vTab,
                                         jsparscnode_t *decl,
                                         const char *funcname,
                                         const char *prefix[],
                                         int mode);
static int               genStmtEntryCall(traceaux_t *paux,
                                          jsparsstream_t *jpa,
                                          jctokfifo_t *queue,
                                          int pos, const char *traceCall,
                                          int mode);

Prototype char           *genTraceFctCall(traceaux_t *paux,
                                          const patchdesc_t *pdesc,
                                          jctokfifo_t *queue, jsparscnode_t *node,
                                          const char *format, const char *marker,
                                          const char *funcname,
                                          const char *types, const char *args,
                                          const char *prefix[]);

static int                scanFuncCalls(wrapaux_t *waux, jsparscnode_t *node);
static int                handleFuncNode(wrapaux_t *waux, jsparscnode_t *node);
static int                insWrappedFunction(traceaux_t *paux, const char *name);
static int                genWrapperFunction(wrapaux_t *aux, const patchdesc_t *pdesc,
                                             jsparscnode_t *node, jsparsvartab_t *vTab,
                                             jsparsvardef_t *vDef);
static const char        *getArgName(void *uPtr);

/*JS*********************************************************************
 *
 *************************************************************************/

int doProcFTrace(traceaux_t *paux, jsparsstream_t *jpa,
                 jsparscnode_t *node, jctokfifo_t *queue)

/************************************************************************/
{
    jsparscnode_t *head, *body, *n, *decl, *spec, *exprNode, *exprVarNode;
    jsparsvartab_t *vTab;
    const patchdesc_t *pdesc;
    const char *funcname;
    jsparsvardef_t *fctDef;
    int callPos;

    funcname = NULL;

    /*  We don't bother about global typedefs, variable defs etc.  */
    if(node->JPACN_Type != JPACT_FUNCDEF) {
        if(node->JPACN_Type == JPACT_PRAGMALINE ||
           node->JPACN_Type == JPACT_PRAGMALINE2) {
            /*  Handle global pragma lines  */
            if(doPragma(jpa, paux, node, queue, 0) == -1)
                EGOTO(error);
        }

        return 0;
    }

    if(paux->Mode & JSCPP_TRACE_NEEDWRAPPING) {
        wrapaux_t waux;

        waux.Paux = paux;
        waux.JPA = jpa;
        waux.Tree = node;
        waux.Queue = queue;

        if(genPrototype(paux, jpa, queue, node, JSCPP_TRACE_PROTOTYPEFCT, 0) == -1)
            EGOTO(error);

        /*  Scan for functions to wrap  */
        if(scanFuncCalls(&waux, node) == -1)
            EGOTO(error);
    }

    head = node->JPACN_Sub;
    body = head->JPACN_Next;
    if(body->JPACN_Type == JPACT_LIST ||
       body->JPACN_Type == JPACT_DECLARATION) {
        /*  K&R-type function definition  */
        body = body->JPACN_Next;
    }

    /*  To be able to include the function arguments, search for the declarator,
     *   step down to the actual function arguments
     */
    for(decl = head->JPACN_Sub ; decl ; decl = decl->JPACN_Next)
        if(decl->JPACN_Type == JPACT_DECLARATOR)
            break;

    if(decl == NULL)
        return 0;

    spec = decl->JPACN_Next;

    /*  Find the variable table which determines the types of variables we
     *   are going to look at
     */
    if((vTab = jsparsfindvartab(jpa, decl)) == NULL)
        EGOTO(error);

    n = decl->JPACN_Sub;
    assert(n->JPACN_Type == JPACT_IDENT);
    if((funcname = strdup(queue->TKFF_Token[n->JPACN_IndexPP].TK_String))
       == NULL)
        EGOTO(error);

    if((fctDef = jsparsgetvardef(jpa, jsparsgetglobalvartab(jpa),
                                 funcname, JVDF_TFUNCTION, 0)) == NULL)
        EGOTO(error);

    /*  Mark that combined file + function name is not up-to-date  */
    checkPattern(NULL, NULL, NULL);

    /*  What do to for this routine  */
    if((pdesc = checkRoutine(paux, funcname)) == NULL)
        EGOTO(error);

    /* ***************************************************************** */

    /*  Handle statement tracing. This is done first, because it does not
     *   interfere with the remainder
     */

    if(pdesc->Type & (PDESC_TYPE_FALLSTMT | PDESC_TYPE_FBRANCHSTMT | PDESC_TYPE_FCODECOVERAGE)) {
        if(genPrototype(paux, jpa, queue, node,
                        (pdesc->Type & PDESC_TYPE_FCODECOVERAGE) ? JSCPP_TRACE_PROTOTYPECOV :
                        JSCPP_TRACE_PROTOTYPESTMT, 0) == -1)
            EGOTO(error);

        if(doStmtTracing(paux, pdesc, jpa, node, queue, funcname) == -1)
            EGOTO(error);
    }

    /* *********************************************************************** */
    /*  Should we generate the prototype for the trace function?  */
    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP) &&
       (pdesc->Type & (PDESC_TYPE_FCALL | PDESC_TYPE_FRETURN))) {
        if(genPrototype(paux, jpa, queue, node, JSCPP_TRACE_PROTOTYPEFCT, 0) == -1)
            EGOTO(error);
    }

    if(paux->NVarTraces > 0) {
        if(genPrototype(paux, jpa, queue, node, JSCPP_TRACE_PROTOTYPEVAR, 0) == -1)
            EGOTO(error);
    }

    /* *********************************************************************** */
    exprVarNode = exprNode = NULL;
    callPos = 0;
    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP) && (pdesc->Type & PDESC_TYPE_FCALL)) {
        /*  Find position where to insert trace statement  */

        /*  We have to do this first because we clear a flag JPA_FLOW_BLOCKSTART
         *   which might be essential for the return tracing
         */
        if((callPos = findFirstStatement(paux, jpa, body, queue,
                                         &exprVarNode, &exprNode)) == -1)
            EGOTO(error);
    }

    /*  Patch return statements  */
    head = NULL;
    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP) && (pdesc->Type & PDESC_TYPE_FRETURN)) {
        /*  Routine ends without proper 'return', so patch this  */
        if(jpa->JSPA_Mode & JPAM_FFALLTOEND) {
            int fromi, toi;

            /*  Generate and patch missing final return statement  */
            fromi = body->JPACN_End;
            toi   = node->JPACN_End;

            if(!(jpa->JSPA_Mode & JPAM_FGETORIG) ||
               jsparsmatchorigtoks(queue, &fromi, &toi)) {
                const char *traceCall;

                if((traceCall = genTraceFctCall(paux, pdesc, queue, node,
                                                paux->TraceFormatReturn,
                                                (pdesc->Type & PDESC_TYPE_FCALL) ?
                                                "-" : "", funcname, "", "",
                                                NULL)) == NULL)
                    EGOTO(error);

                if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++fromi; ++toi; }

                if(jctoksubstitute(queue, fromi, toi,
                                   paux->Indentation > 0 ?
                                   "  %s;\n%o" : "  %s; %o", traceCall) == -1)
                    EGOTO(error);
            } else if(jpa->JSPA_WarnFlags != 0) {
                jsparsprinterr(jpa, queue, node,
                               "Cannot patch final return statement in function '%s'",
                               0, funcname);
            }
        }

        if(vTab->JPAVT_Next) {
            /*  Add "_retvalue" to vTab. We get the type from the function
             *   head, but we remove the function call to get the right type.
             *   We need this when we determine the type of an expression
             *   which (most likely) contains the return variable.
             */
            node = decl->JPACN_Sub;
            assert(node->JPACN_Type == JPACT_IDENT);
            node = node->JPACN_Next;
            assert(node->JPACN_Type == JPACT_FUNCCALL);
            node = node->JPACN_Next;

            if((head = jsparsmkcnode(NULL, JPACT_DECLARATOR, node,
                                     decl->JPACN_Start, decl->JPACN_End,
                                     queue)) == NULL)
                EGOTO(error);

            if(jsparsdefinevar2(jpa, vTab, JVDF_FDEFINED | JVDF_TVARIABLE,
                                queue, paux->NameRetVariable,
                                head, spec, 0) == NULL)
                EGOTO(error);
        }
    }

    /*  First scan body and patch return statements  */
    if(doPatchStmt(paux, jpa, fctDef, body, pdesc, 0) == -1)
        EGOTO(error);

    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP) && (pdesc->Type & PDESC_TYPE_FRETURN)) {
        if(head) {
            /*  Free allocated node for the variable definition  */
            head->JPACN_Sub = NULL;
            jsparsfreecnode(head);
        }
    }

    /* *********************************************************************** */

    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP) && (pdesc->Type & PDESC_TYPE_FCALL)) {
        if(callPos != 0 || exprNode != NULL) {
            static const char *Prefixes[] = { "(", ", " };
            char *traceCall;

            /*  Patch the function entry  */
            if((traceCall = genFctEntryCall(paux, jpa, queue, pdesc, vTab, decl,
                                            funcname, Prefixes, 0)) == NULL)
                EGOTO(error);

            if(exprNode != NULL) {
                int start, end;

#if 0
                jsparsprctree(stdout, queue, exprNode, 0, 2);
#endif

                /*  We create the statement at the first expression in this function  */
                start = exprNode->JPACN_Start;
                end = exprNode->JPACN_End;

                /*  Make sure we do not mess up strings  */
                if(!(jpa->JSPA_Mode & JPAM_FGETORIG) ||
                   jsparsmatchorigtoks(queue, &start, &end)) {
                    end += jctokgetstep(queue);
                    if(tokInsertString(jpa, node, queue, end, ")", 0) == -1)
                        EGOTO(error);

                    if(jpa->JSPA_Mode & EXTFLAG_INSERTCAST) {
                        jsparsvardef_t *vDef;
                        int ptrcnt;

                        /*  The AIX compiler is deficient: Valid statements like
                         *      char *p = NULL;
                         *  will normally be patched as
                         *      char *p = (_jstracefunc(...), NULL);
                         *  and are then considered invalid because "NULL"
                         *  evaluates to "0" and the compiler seems not to be
                         *  able to recognize this as valid.
                         *
                         *  With this hook we generate the statement as
                         *      char *p = (_jstracefunc(...), (char *)NULL);
                         *                                    ^^^^^^^^
                         *  which should remedy this.
                         */
                        if((vDef = jsparsgetvardef(jpa, vTab,
                                                   queue->TKFF_Token[exprVarNode->
                                                                     JPACN_IndexPP].TK_String,
                                                   JVDF_TVARIABLE, 0)) == NULL)
                            EGOTO(error);

                        /*  Is it a pointer?  */
                        if(jpa->JSPA_Mode & EXTFLAG_INSERTCAST2)
                            ptrcnt = 1;
                        else if(jsparsgetvartype3(jpa, vTab, vDef, &ptrcnt, 0) == -1)
                            EGOTO(error);

                        if(ptrcnt > 0) {
                            jstrbuffempty(&TempBuffer);
                            if(jsparsgenvardef(jpa, &TempBuffer, vDef, NULL) == -1 ||
                               tokInsertString(jpa, node, queue, start,
                                               jstrbuffer(&TempBuffer), 0) == -1)
                                EGOTO(error);
                        }
                    }

                    if(jpa->JSPA_Mode & EXTFLAG_DOUBLEASSIGN) {
                        /*  Alternative workaround to the above EXTFLAG_INSERTCAST:
                         *  Patch assignment statements like
                         *      char *p = NULL;
                         *  as
                         *      char *p = (_jstracefunc(...), p = NULL);
                         *                                    ^^^^
                         */
                        if(tokInsertString(jpa, node, queue, start, "=", 0) == -1)
                            EGOTO(error);

                        if(tokInsertString(jpa, node, queue, start,
                                           queue->TKFF_Token[exprVarNode->JPACN_IndexPP].TK_String,
                                           0) == -1)
                            EGOTO(error);
                    }

                    if(tokInsertString(jpa, node, queue, start, traceCall, 0) == -1)
                        EGOTO(error);

                    callPos = 0;
                } else if(jpa->JSPA_WarnFlags != 0) {
                    jsparsprinterr(jpa, queue, exprNode,
                                   "Cannot patch expression in function '%s'", 0,
                                   funcname);

                    /*  We leave callPos as it is, so we generate the trace call
                     *   not in the expression itself, but after the declarations.
                     *   That would not produce the right flow in some cases but
                     *   at least it *does* generate a flow.
                     */
                }
            }

            if(callPos != 0) {
                ++traceCall; /*  actually "+= strlen(Prefixes[0]);"  */
                traceCall[strlen(traceCall) - 2] = '\0'; /*  actually "- strlen(Prefixes[1])"  */

                if(genStmtEntryCall(paux, jpa, queue, callPos, traceCall, 0) == -1)
                    EGOTO(error);
            }
        } else {
            /*  That looks like a function definition through a macro or via
             *   file inclusion. There is nothing we can do about that
             */
            if((pdesc->Type & (PDESC_TYPE_FCALL | PDESC_TYPE_FRETURN)) &&
               jpa->JSPA_WarnFlags != 0)
                jsparsprinterr(jpa, queue, body, "Cannot patch function '%s'", 0,
                               funcname);
        }
    }

    free((void *)funcname);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int doProcTypeRename(traceaux_t *paux, jsparsstream_t *jpa,
                     jsparscnode_t *node, jctokfifo_t *queue)
{
    jsparscnode_t *sub;

    /*  First handle sub nodes  */
    for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next)
        if(doProcTypeRename(paux, jpa, sub, queue) == -1)
            EGOTO(error);

    if(node->JPACN_Type == JPACT_UNARYEXPR) {
        if(doTypeRenameExpr(paux, jpa, node, queue) == -1)
            EGOTO(error);

    } else if(node->JPACN_Type == JPACT_TYPEDEF ||
              node->JPACN_Type == JPACT_STRUCT) {
        if(doTypeRenameDef(paux, jpa, node, queue) == -1)
            EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int doTypeRenameExpr(traceaux_t *paux, jsparsstream_t *jpa,
                            jsparscnode_t *node, jctokfifo_t *queue)

/************************************************************************/
{
#define MAXINDSUBST       200
    jsparscnode_t *prev, *sub;
    jsparsvartab_t *lTab = NULL;
    unsigned int iinds, ninds, indoff, start, step;
    const char  *strTypes[MAXINDSUBST], *strSource[MAXINDSUBST], *strTarget[MAXINDSUBST];

    indoff = jpa->JSPA_Mode & JPAM_FPRINTORIG ? jctokindorig(queue) : 0;
    step = jctokgetstep(queue);

    ninds = 0;
    for(prev = NULL, sub = node->JPACN_Sub ; sub ; prev = sub, sub = sub->JPACN_Next) {
        jsparsexprtype_t eType;
        jsparscnode_t *spec, *sub2;
        const char *typeName, *memberName, *targetName;
        unsigned int substInd;
        int pos;

        if(sub->JPACN_Type != JPACT_MEMBER && sub->JPACN_Type != JPACT_POINTTO)
            continue;

        assert(prev != NULL);

        if(lTab == NULL && (lTab = jsparsfindvartab(jpa, node)) == NULL)
            EGOTO(error);

        assert(prev->JPACN_Next == sub);
        prev->JPACN_Next = NULL;
        pos = jsparsgetexprtype(&eType, jpa, lTab, node, queue, 0) == NULL;
        prev->JPACN_Next = sub;

        if(pos) {
            jsparsprctree(stdout, queue, node, 0, 0);
            EGOTO(error);
        }

        spec = eType.JSPE_Specifier;

        if(spec->JPACN_Type != JPACT_TYPEDEFNAME)
            continue;

        assert(spec->JPACN_Start < eType.JSPE_Queue->TKFF_Index && spec->JPACN_End < eType.JSPE_Queue->TKFF_Index);

        typeName = eType.JSPE_Queue->TKFF_Token[spec->JPACN_IndexPP].TK_String;
        assert(typeName != NULL);

        sub2 = sub->JPACN_Sub;

        assert(sub2->JPACN_Type == JPACT_IDENT || sub2->JPACN_Type == JPACT_ENUMCONST);

        memberName = queue->TKFF_Token[sub2->JPACN_IndexPP].TK_String;

        assert(memberName != NULL);

        if((pos = findTypeRename(paux, typeName, memberName)) == -1)
            continue;

        targetName = paux->TypeRenames[pos].TargetName;

#ifdef TRACETYPE_JSCPP
        jstrace(TRACETYPE_JSCPP, 3,
                "Rename type='%s' member='%s' to '%s'\n", typeName, memberName, targetName);
#endif

        assert(ninds < sizeof(strTypes)/sizeof(strTypes[0]));

        strTypes[ninds]  = typeName;
        strSource[ninds] = memberName;
        strTarget[ninds] = targetName;
        ++ninds;
    }

    if(ninds == 0)
        return 0;

#if 0
    printf("------------------------------------ %d node start %d end %d\n", ninds, node->JPACN_Start, node->JPACN_End);
    jsparsprctree(stdout, queue, node, 0, 7);
    printf("------------------------------------\n");
#endif

    /* ******************************************************************* */
    start = node->JPACN_Start;

    if(queue->TKFF_Token[start + indoff].TK_Value == PTK_EMPTY) {
        for( ; start > 0 ; start -= step)
            if(queue->TKFF_Token[start + indoff].TK_Value != PTK_EMPTY)
                break;

        for( ; start > 0 ; start -= step)
            if(queue->TKFF_Token[start].TK_Value != PTK_EMPTY)
                break;
    }

    for(iinds = 0 ; iinds < ninds ; ++iinds) {
        const char *memberName, *targetName;
        unsigned int curr, count;

        memberName = strSource[iinds];
        targetName = strTarget[iinds];

        /*printf("MEM '%s' -> '%s'\n", memberName, targetName);*/

        count = 0;
        for(curr = start ; curr <= node->JPACN_End ; curr += step) {

#if 0
            printf("%d %d: %d  '%s'\n", curr, curr/step,
                   queue->TKFF_Token[curr + indoff].TK_Value,
                   queue->TKFF_Token[curr + indoff].TK_String);
#endif

            if(queue->TKFF_Token[curr + indoff].TK_Value != CTK_IDENT &&
               queue->TKFF_Token[curr + indoff].TK_Value != JPACT_ENUMCONST)
                continue;

            /*  By checking for pointer identity, we only replace what stems from it!  */
#if 0
            printf("HAVE %p %p\n",queue->TKFF_Token[curr + indoff].TK_String, memberName);
#endif

            if(queue->TKFF_Token[curr + indoff].TK_String != memberName)
                continue;

            queue->TKFF_Token[curr + indoff].TK_String = (char *)targetName;
            ++count;
        }

        if(count == 0 && jpa->JSPA_WarnFlags != 0)
            jsparsprinterr(jpa, queue, node, "Cannot rename type '%s' member '%s' being expanded", 0,
                           strTypes[iinds], memberName);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int doTypeRenameDef(traceaux_t *paux, jsparsstream_t *jpa,
                           jsparscnode_t *node, jctokfifo_t *queue)

/************************************************************************/
{
    jsparscnode_t *sub, *block;
    const char *typeName;
    unsigned int indoff;

    indoff = jpa->JSPA_Mode & JPAM_FPRINTORIG ? jctokindorig(queue) : 0;

#if 0
    printf("========================\n");
    jsparsprctree(stdout, queue, node, 0, 7);
#endif

    typeName = NULL;
    block = NULL;

    if(node->JPACN_Type == JPACT_TYPEDEF) {
        for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
            if(sub->JPACN_Type == JPACT_DECLARATOR) {
                jsparscnode_t *idnode;

                for(idnode = sub->JPACN_Sub ; idnode ; idnode = idnode->JPACN_Next) {
                    const char *typeName1;

                    if(idnode->JPACN_Type != JPACT_IDENT)
                        continue;

                    typeName1 = queue->TKFF_Token[idnode->JPACN_IndexPP].TK_String;
                    if(findTypeRename(paux, typeName1, NULL) != -1) {
                        assert(typeName == NULL);

                        if(queue->TKFF_Token[idnode->JPACN_IndexOrig + indoff].TK_Value == PTK_EMPTY) {
                            if(jpa->JSPA_WarnFlags != 0)
                                jsparsprinterr(jpa, queue, node, "Cannot rename typedef '%s' appearing in header file", 0,
                                               typeName1);

                            continue;
                        }

                        typeName = typeName1;
                    }
                }
            }
            else if(sub->JPACN_Type == JPACT_STRUCT) {
                jsparscnode_t *block0;

                block0 = sub->JPACN_Sub;
                if(block0->JPACN_Type == JPACT_BLOCK) {
                    assert(block == NULL);
                    block = block0;
                }
            }
        }

    } else {
        assert(node->JPACN_Type == JPACT_STRUCT);

        for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
            if(sub->JPACN_Type == JPACT_IDENT) {
                const char *typeName1;

                typeName1 = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;
                if(findTypeRename(paux, typeName1, NULL) != -1) {
                    assert(typeName == NULL);

                    if(queue->TKFF_Token[sub->JPACN_IndexOrig + indoff].TK_Value == PTK_EMPTY) {
                        if(jpa->JSPA_WarnFlags != 0)
                            jsparsprinterr(jpa, queue, sub, "Cannot rename typedef '%s' appearing in header file", 0,
                                           typeName1);

			continue;
		    }

		    typeName = typeName1;
                }
            }

            if(sub->JPACN_Type == JPACT_BLOCK) {
                assert(block == NULL);
                block = sub;
            }
        }
    }

    if(typeName == NULL || block == NULL)
        return 0;

#if 0
    printf("========================\n");
    jsparsprctree(stdout, queue, block, 0, 7);
#endif

    for(sub = block->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
        const char *memberName, *targetName;
        jsparscnode_t *decl;
        unsigned int useNext;

#if 0
        printf("----------------------------------------\n");
        jsparsprctree(stdout, queue, sub, 0, 7);
#endif

        assert(sub->JPACN_Type == JPACT_DECLARATION);
        decl = sub->JPACN_Sub;

        useNext = decl->JPACN_Type == JPACT_LIST;

        if(useNext)
            decl = decl->JPACN_Sub;

        for(;;) {
            jsparscnode_t *idnode;
            unsigned int substInd, pos;

            /*printf("++++++++++++++++++++++++++++++++++++++--\n");
              jsparsprctree(stdout, queue, decl, 0, 7);*/

            assert(decl->JPACN_Type == JPACT_DECLARATOR);
            idnode = decl->JPACN_Sub;

            assert(idnode->JPACN_Type == JPACT_IDENT);
            memberName = queue->TKFF_Token[idnode->JPACN_IndexPP].TK_String;

            if((pos = findTypeRename(paux, typeName, memberName)) != (unsigned int)-1) {
                substInd = idnode->JPACN_IndexOrig + indoff;

                targetName = paux->TypeRenames[pos].TargetName;
                queue->TKFF_Token[substInd].TK_String = (char *)targetName;
            }

            if(!useNext)
                break;

            if((decl = decl->JPACN_Next) == NULL)
                break;
        }
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int findFirstStatement(traceaux_t *paux, jsparsstream_t *jpa,
                              jsparscnode_t *body, jctokfifo_t *queue,
                              jsparscnode_t **pVarNode,
                              jsparscnode_t **pExpression)

/************************************************************************/
{
    jsparscnode_t *lastdecl, *firststmt, *n;
    int pos, step;

    *pExpression = NULL;
    *pVarNode = NULL;
    lastdecl = firststmt = NULL;
    for(n = body->JPACN_Sub ; n ; n = n->JPACN_Next) {
        if(n->JPACN_Type == JPACT_DECLARATION) {
            lastdecl = n;

            if(*pExpression == NULL) {
                jsparscnode_t *sub, *node2;
                int ret;

                /*  We check if this contains something like an assignment like in
                 *     void dummy() { int i = callee(); printf("Hello, world!\n"); }
                 */
                sub = n->JPACN_Sub;

                /*  Make sure that we do not try to patch static or const definitions like
                 *   "static int X = 1;"
                 */
                for(node2 = sub->JPACN_Next ; node2 ; node2 = node2->JPACN_Next)
                    if(node2->JPACN_Type == JPACT_CONST ||
                       node2->JPACN_Type == JPACT_STATIC)
                        break;
                if(node2 != NULL)
                    continue; /*  not suitable  */

                if(sub->JPACN_Type == JPACT_LIST) {
                    for(sub = sub->JPACN_Sub ; sub ; sub = sub->JPACN_Next)
                        if(sub->JPACN_Type == JPACT_ASS) {
                            if((ret = checkAssignment(jpa, queue, sub)) == -1)
                                EGOTO(error);

                            if(ret != 0) {
                                sub = sub->JPACN_Sub;
                                assert(sub->JPACN_Type == JPACT_DECLARATOR);

                                *pVarNode = sub->JPACN_Sub;
                                *pExpression = sub->JPACN_Next;

                                break;
                            }
                        }
                } else if(sub->JPACN_Type == JPACT_ASS) {
                    if((ret = checkAssignment(jpa, queue, sub)) == -1)
                        EGOTO(error);

                    if(ret != 0) {
                        sub = sub->JPACN_Sub;
                        assert(sub->JPACN_Type == JPACT_DECLARATOR);

                        *pVarNode = sub->JPACN_Sub;
                        *pExpression = sub->JPACN_Next;

                        /*  must not break here since we still want to find
                         *   the first statement in order to later check
                         *   if we have matching originals
                         */
                        /*break;*/
                    }
                }
            }
        } else if(n->JPACN_Type == JPACT_TYPEDEF)
            lastdecl = n;
        else if(firststmt == NULL)
            firststmt = n;
    }

    step = jctokgetstep(queue);

    pos = (lastdecl ? lastdecl->JPACN_End : body->JPACN_Start) + step;
    if(jpa->JSPA_Mode & JPAM_FGETORIG) ++pos;

    if(paux->Indentation != 0) {
        /*  Skip any intermediate blanks to make statement appear after
         *   comments behind declarations
         */
        while(CTK_BLANK_START <= queue->TKFF_Token[pos].TK_Value &&
              queue->TKFF_Token[pos].TK_Value <= CTK_BLANK_END)
            pos += step;
    }

    if(firststmt != NULL && (jpa->JSPA_Mode & JPAM_FGETORIG)) {
        int end;

        /*  Search through the statements between the last declaration and
         *   the first real statements. There should be one with an original
         *   token which we can use.
         */
        end = firststmt->JPACN_Start + 1;

        for( ; pos <= end ; pos += step)
            if(queue->TKFF_Token[pos].TK_Value != PTK_EMPTY)
                break;

        if(pos > end)
            return 0;
    }

    if(firststmt != NULL) {
        int fpos;

        /*  Make sure no one thinks that he can insert declarations
         *   after the first statement, since we are going to insert
         *   a trace call there
         */
        if((fpos = jsparsfindflow(jpa, firststmt)) == -1)
            EGOTO(error);

        jpa->JSPA_Statements[fpos].JPAFLN_Flags &= ~JPA_FLOW_BLOCKSTART;
    }

    return pos;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int genPrototype(traceaux_t *paux, jsparsstream_t *jpa,
                        jctokfifo_t *queue,
                        jsparscnode_t *node, long type, int mode)

/************************************************************************/
{
    int start, ret;

    if(!(paux->Mode & type) || (paux->PrototypeFlag & type))
        return 0;

    start = node->JPACN_Start;
    if((jpa->JSPA_Mode & JPAM_FGETORIG) &&
       !jsparsmatchorigtoks(queue, &start, NULL))
        return 0;

    paux->PrototypeFlag |= type;

    /*  We put the prototype in the same line so that any compiler errors
     *   later on can be easily located in the original file
     */
    jstrbuffempty(&TempBuffer);

    switch(type) {
    case JSCPP_TRACE_PROTOTYPESTMT:
        if(jstrbuffprintf(&TempBuffer,
                          "extern void %s(const char *,int); "
                          "static const char "STMT_FILENAME_VAR"[] = ",
                          STMTTRACE_FUNCTION) == -1)
            EGOTO(error);

        if((ret = jctokindfpos(queue)) != -1)
            ret = jstrbuffprintf(&TempBuffer, "\"%s\"; ",
                                 queue->TKFF_Token[node->JPACN_Start + ret].
                                 TK_String);
        else
            ret = jstrbuffprint(&TempBuffer, "__FILE__; ");
        break;

    case JSCPP_TRACE_PROTOTYPECOV:
        ret = jstrbuffprintf(&TempBuffer,
                             "extern void %s(long *,int *);extern long %s[]; extern int %s[];",
                             CODECOVER_REGFUNCTION,
                             CODECOVER_COUNTARRAY, CODECOVER_DESCRIPTARRAY);

        break;

    case JSCPP_TRACE_PROTOTYPEFCT:
        ret = jstrbuffprintf(&TempBuffer,
                             "extern void %s(const char *,const char *, ...);"
                             "%s", CALLTRACE_FUNCTION,
                             paux->Indentation > 0 ? "\n" : " ");
        break;

    case JSCPP_TRACE_PROTOTYPEVAR:
        ret = jstrbuffprintf(&TempBuffer,
                             "extern void %s(const char *,int,const char*, ...);"
                             "%s", paux->TraceVarFunction,
                             paux->Indentation > 0 ? "\n" : " ");
        break;
#ifdef CONFIG_JSCPPLIB_DEBUG
    default:
        EGOTO(error);
#endif
    }

    if(ret == -1) EGOTO(error);

    jstrbuffclose(&TempBuffer);
    if(tokInsertString(jpa, node, queue, start,
                       jstrbuffer(&TempBuffer), mode) == -1)
        EGOTO(error);

    return 1;
 error:
    return -1;

}

/*JS*********************************************************************
 *
 *************************************************************************/

int doGetName(traceaux_t *paux, jsparscnode_t *node, const char **pName,
              jctokfifo_t *queue, int mode)

/************************************************************************/
{
    char *name;

    if(node->JPACN_Type != JPACT_IDENT) {
        node = node->JPACN_Sub;
        if(node->JPACN_Type != JPACT_DECLARATOR) {
            if(node->JPACN_Type == JPACT_VOID ||
               node->JPACN_Type == JPACT_ELLIPSIS) {
                /*  a function without arguments  */
                *pName = NULL;
                return 0;
            }

            /*  Special mode for parameter declarations consisting only of specifiers,
             *   i. e. "int"
             */
            if(mode & 2) goto newargname;

            EGOTO(error);
        }

        node = node->JPACN_Sub;
        if(node == NULL || node->JPACN_Type != JPACT_IDENT) {
            /*  Special mode for parameter declarations without a identifier  */
            if(mode & 2) goto newargname;

            EGOTO(error);
        }
    }

    name = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

    if(mode & 1) {
        if((name = (char *)strdup(name)) == NULL)
            EGOTO(error);
    }

    *pName = name;

    return 0;
 newargname:
    if((name = strdup(getArgName(paux))) == NULL)
        EGOTO(error);

    *pName = name;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static const patchdesc_t *checkRoutine(traceaux_t *paux, const char *funcName)

/************************************************************************/
{
    static patchdesc_t DefaultDescription = {
        0, NULL, 0, NULL, -1, NULL, -1
    };
    int i;

    for(i = paux->NDescriptions - 1 ; i >= 0 ; --i) {
        const patchdesc_t *pdesc;
        int j;

        pdesc = (const patchdesc_t *) &(paux->Descriptions[i]);

        for(j = 0 ; j < pdesc->NPatterns ; ++j) {
            int ret;

            if((ret = checkPattern(pdesc->Patterns[j], paux->CurrentFile,
                                   funcName)) == -1)
                EGOTO(error);

            if(ret != 0)
                break;
        }

        if(j < pdesc->NPatterns)
            return pdesc;
    }

    /*  Return default description  */
    DefaultDescription.Type = PDESC_TYPE_TRACE;

    /* ***  Function entry/exit tracing  *** */
    if(paux->Mode & JSCPP_TRACE_FCTCALL) {
        DefaultDescription.Type |= PDESC_TYPE_FCALL;
        DefaultDescription.NCallDescriptions =
            (paux->Mode & JSCPP_TRACE_FCTCALLARGS) ? -1 : 0;

        if(paux->Mode & JSCPP_TRACE_FCTCALLLBLD)
            DefaultDescription.Type |= PDESC_TYPE_FCALLLBL;
    }

    if(paux->Mode & JSCPP_TRACE_FCTRET) {
        DefaultDescription.Type |= PDESC_TYPE_FRETURN;
        DefaultDescription.NReturnDescriptions =
            (paux->Mode & JSCPP_TRACE_FCTRETARG) ? -1 : 0;

        if(paux->Mode & JSCPP_TRACE_FCTRETLINE)
            DefaultDescription.Type |= PDESC_TYPE_FRETURNLINE;

        if(paux->Mode & JSCPP_TRACE_FCTRETLBLD)
            DefaultDescription.Type |= PDESC_TYPE_FRETURNLBL;
    }

    /* ***  Statement tracing  *** */
    if(paux->Mode & JSCPP_TRACE_STMTTRACE)
        DefaultDescription.Type |= PDESC_TYPE_FALLSTMT;
    if(paux->Mode & JSCPP_TRACE_STMTTRACE_BRANCH)
        DefaultDescription.Type |= PDESC_TYPE_FBRANCHSTMT;

    /* ***  Code coverage  *** */
    if(paux->Mode & JSCPP_TRACE_CODECOVERAGE)
        DefaultDescription.Type |= PDESC_TYPE_FCODECOVERAGE;

    /* ***  Function wrapping  *** */
    if(paux->Mode & JSCPP_TRACE_FCTWRAP)
        DefaultDescription.Type |= PDESC_TYPE_FFCTWRAP;

    return (const patchdesc_t *) &DefaultDescription;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int checkPattern(const char *pattern, const char *fileName,
                 const char *funcName)

/************************************************************************/
{
    static int NameGenerated = FALSE;
    const char *name;

    if(pattern == NULL) {
        NameGenerated = FALSE;
        return 0;
    }

    if(strchr(pattern, '@') != NULL) {
        static char *NameBuffer;
        static int NNameBuffer;

        if(!NameGenerated) {
            int len;

            if((len = strlen(fileName) + strlen(funcName) + 1) >= NNameBuffer) {
                NNameBuffer = len + 100;
                if((NameBuffer = (char *)realloc(NameBuffer, NNameBuffer)) == NULL)
                    EGOTO(error);
            }

            sprintf(NameBuffer, "%s@%s", fileName, funcName);
            NameGenerated = TRUE;
        }

        name = NameBuffer;
    } else
        name = funcName;

    return jiswildcard(pattern) ? wildcmp(name, pattern, 0) :
        strcmp(name, pattern) == 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static char *genFctEntryCall(traceaux_t *paux, jsparsstream_t *jpa,
                             jctokfifo_t *queue,
                             const patchdesc_t *pdesc,
                             jsparsvartab_t *vTab,
                             jsparscnode_t *decl,
                             const char *funcname,
                             const char *prefix[], int mode)

/************************************************************************/
{
    const char *types, *args;

    jstrbuffempty(&SBuffTypes);
    jstrbuffempty(&SBuffArgs);

    /*  Generate the tracing call  */
    if(pdesc->NCallDescriptions >= 0) {
        /*  The arguments/expressions to patch are explicitly given  */
        if(genTraceStmt(paux, jpa, vTab, &SBuffTypes, &SBuffArgs,
                        pdesc->CallDescriptions, pdesc->NCallDescriptions,
                        NULL, NULL, 0) == -1)
            EGOTO(error);
    } else {
        jsparscnode_t *node;

        /*  Find the start of the function argument list  */
        for(node = decl->JPACN_Sub ; node ; node = node->JPACN_Next)
            if(node->JPACN_Type == JPACT_FUNCCALL)
                break;

        if(node != NULL) {
            /*  Walk through all the parameters  */
            for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next) {
                jsparsexprtype_t eType;
                const char *name;
                int useLabel;

                if(doGetName(paux, node, &name, queue, mode | 1) == -1)
                    EGOTO(error);

                if(name == NULL)
                    continue;

                /*  Check for K&R type of declarator. In this case, we have to search
                 *   the variable definitions
                 */
                if(node->JPACN_Type == JPACT_IDENT) {
                    jsparsvardef_t *vDef;

                    if((vDef = jsparsgetvardef(jpa, vTab, name,
                                               JVDF_TVARIABLE, 0)) == NULL)
                        EGOTO(error);

                    if(jsparsvdef2type(jpa, vTab, &eType, vDef, 0) == NULL)
                        EGOTO(error);
                } else {
                    /*  Convert the node into declarator + specifier  */
                    if(jsparsnode2type(&eType, node, queue) == NULL)
                        EGOTO(error);
                }

                useLabel = (pdesc->Type & PDESC_TYPE_FCALLLBL) ? 1 : 0;
                if(genTracePart(paux, jpa, vTab, &SBuffTypes, &SBuffArgs,
                                name, name, &eType, &useLabel, 0) == -1)
                    EGOTO(error);

                free((void *)name);
            }
        }
    }

    types = jstrbufflen(&SBuffTypes) != 0 ? jstrbuffer(&SBuffTypes) : "";

    if((args = jstrbuffer(&SBuffArgs)) == NULL)
        args = "";

    return genTraceFctCall(paux, pdesc, queue, decl, paux->TraceFormatCall,
                           (pdesc->Type & PDESC_TYPE_FRETURN) ?
                           "+" : "", funcname, types, args, prefix);
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int genStmtEntryCall(traceaux_t *paux, jsparsstream_t *jpa,
                            jctokfifo_t *queue, int pos, const char *traceCall,
                            int mode)

/************************************************************************/
{
    char format[20];
    int offset;

    if(paux->Indentation > 0)
        offset = sprintf(format, "%%%dI", paux->Indentation);
    else {
        format[0] = ' ';
        offset = 1;
    }

    offset += sprintf(format + offset, "%%s;");

    if(paux->Indentation == 0)
        format[offset++] = ' ';
    else if(!(CTK_BLANK_START <= queue->TKFF_Token[pos].TK_Value &&
              queue->TKFF_Token[pos].TK_Value <= CTK_BLANK_END))
        offset += sprintf(format + offset, "%%I");

    offset += sprintf(format + offset, "%%o");

    assert(offset < (int)sizeof(format));

    if(jctoksubstitute(queue, pos, pos, format, traceCall) == -1)
        EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

char *genTraceFctCall(traceaux_t *paux, const patchdesc_t *pdesc,
                      jctokfifo_t *queue, jsparscnode_t *node,
                      const char *format, const char *marker,
                      const char *funcname, const char *types,
                      const char *args, const char *prefix[])

/************************************************************************/
{
    static jstrbuff_t CallBuffer;

    jstrbuffempty(&CallBuffer);
    if(prefix && prefix[0] && jstrbuffprint(&CallBuffer, prefix[0]) == -1)
        EGOTO(error);

    while(*format) {
        const char *s;
        int len;

        if((s = strchr(format, '%')) != NULL)
            len = s - format;
        else
            len = strlen(format);

        if(len > 0) {
            if(jstrbuffstore(&CallBuffer, format, len) == -1)
                EGOTO(error);

            format += len;
        }

        if(*format == '%') {
            ++format;

            switch(*format) {
            case '%':
                s = "%";
                break;
            case 'n':
                s = funcname;
                break;
            case 'f':
                s = types;
                break;
            case 'a':
                s = args;
                break;
            case 'm':
                s = marker;
                break;
            case 'p':
                if(!(pdesc->Type & PDESC_TYPE_FRETURNLINE)) {
                    s = NULL;
                    break;
                }

                /*FALLTHROUGH*/
            case 'L':
                if((len = jctokindfpos(queue)) == -1) {
                    s = *format == 'p' ? ",__LINE__" : "__LINE__";

                } else {
                    if(jstrbuffprintf(&CallBuffer, *format == 'p' ? ",%d" : "%d",
                                      queue->TKFF_Token[node->JPACN_Start + len].TK_Value) == -1)
                        EGOTO(error);

                    s = NULL;
                }
                break;

            case 'C':
                jstrbuffprintf(&CallBuffer, "%d", paux->NCovRange);
                s = NULL;
                break;

            case 'P':
                s = (pdesc->Type & PDESC_TYPE_FRETURNLINE) ? "@" : NULL;
                break;
            case 'l':
                {
                    int level;

                    if((level = findTraceLevel(paux, funcname)) == -1 ||
                       jstrbuffprintf(&CallBuffer, "%d", level) == -1)
                        EGOTO(error);
                }
                s = NULL;
                break;

            case 'F':
                s = (paux->PrototypeFlag & JSCPP_TRACE_PROTOTYPESTMT) ? STMT_FILENAME_VAR : "__FILE__";
                break;

            default:
                EGOTO(error);
            }

            if(s && jstrbuffprint(&CallBuffer, s) == -1)
                EGOTO(error);

            ++format;
        }
    }

    if(prefix && prefix[1] && jstrbuffprint(&CallBuffer, prefix[1]) == -1)
        EGOTO(error);

    jstrbuffclose(&CallBuffer);

    return jstrbuffer(&CallBuffer);
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int scanFuncCalls(wrapaux_t *waux, jsparscnode_t *node)

/************************************************************************/
{
    if(node->JPACN_Type == JPACT_UNARYEXPR && handleFuncNode(waux, node) == -1)
        EGOTO(error);

    for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next) {
        /*  Recursive call  */
        if(scanFuncCalls(waux, node) == -1)
            EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int handleFuncNode(wrapaux_t *waux, jsparscnode_t *node)

/************************************************************************/
{
    jctokfifo_t *queue;
    jsparscnode_t *sub;
    jsparsstream_t *jpa;
    jsparsvardef_t *vDef;
    jsparsvartab_t *vTab;
    const char *ident;
    const patchdesc_t *pdesc;

    jpa = waux->JPA;
    queue = waux->Queue;

    if((sub = node->JPACN_Sub) == NULL || sub->JPACN_Type != JPACT_IDENT)
        return 0;

    ident = waux->Queue->TKFF_Token[sub->JPACN_End].TK_String;

    if(jstrstartswith(ident, waux->Paux->PrefixWrapFunction) != NULL)
        /*  This is already wrapped  */
        return 0;

    if((vTab = jsparsfindvartab(jpa, node)) == NULL)
        EGOTO(error);

    if((vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TVARIABLE,
                               0)) != NULL)
        /*  The name is known as a variable (which might possibly
         *   shadow function name).
         */
        return 0;

    /*  Now check for a function with this name  */
    if((vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TFUNCTION,
                               0)) == NULL) {
        /*  Let's be puristic and check if this is know at all  */
        if((waux->JPA->JSPA_WarnFlags & JPAM_WARNVARS) &&
           (vDef = jsparsgetvardef(jpa, vTab, ident, JVDF_TENUMIDENT,
                                   0)) == NULL) {
            jsparscnode_t *sub2;

            /*  Unknown identifier. Report it  */
            sub2 = sub->JPACN_Next;
            jsparsprinterr(jpa, waux->Queue, node,
                           "Undeclared %s '%s'", 0,
                           (sub2 != NULL && sub2->JPACN_Type == JPACT_FUNCCALL) ?
                           "function" : "identifier", ident);
        }

        return 0;
    }

    pdesc = checkRoutine(waux->Paux, ident);
    if(!(pdesc->Type & PDESC_TYPE_FFCTWRAP))
        return 0;

    {
        patchdesc_t copyDesc;
        int pos, flag, toff, start, end;

        start = sub->JPACN_Start;
        end = sub->JPACN_End;
        if((jpa->JSPA_Mode & JPAM_FGETORIG) &&
           !jsparsmatchorigtoks(queue, &start, &end)) {
            if(jpa->JSPA_WarnFlags != 0)
                jsparsprinterr(jpa, queue, node,
                               "Cannot wrap preprocessed identifier '%s'", 0,
                               ident);

            return 0;
        }

        /*  Line information is not useful in our wrapper routine,
         *   so avoid it
         */
        copyDesc = *pdesc;
        copyDesc.Type &= ~PDESC_TYPE_FRETURNLINE;
        pdesc = &copyDesc;

        /*  Insert into list of already wrapped routines  */
        if((pos = insWrappedFunction(waux->Paux, ident)) == -1)
            EGOTO(error);

        if(waux->Paux->WrappedFunctions[pos].Flag == 0) {
            int ostart = sub->JPACN_Start;

            if((flag = genWrapperFunction(waux, pdesc, node,
                                          vTab, vDef)) == -1)
                EGOTO(error);

            /*  This has surely changed, so adapt  */
            if(sub->JPACN_Start != ostart) {
                start += sub->JPACN_Start - ostart;
                end += sub->JPACN_Start - ostart;
            }

            /*  This might have changed due to reallocation  */
            ident = waux->Queue->TKFF_Token[end].TK_String;

            /*  Flag == 1: Wrapper function already done.
             *  Flag == 2: Cannot wrap function
             */
            waux->Paux->WrappedFunctions[pos].Flag = flag == 0 ? 2 : 1;
        } else
            /*  Function already wrapped?  */
            flag = waux->Paux->WrappedFunctions[pos].Flag == 1;

        toff = (jpa->JSPA_Mode & JPAM_FGETORIG) ? 1 : 0;

        /*  Redirect function call to wrapper function  */
        if(flag &&
           jctoksubstitute(waux->Queue, start + toff, end + toff, "%s%s",
                           waux->Paux->PrefixWrapFunction, ident) == -1)
            EGOTO(error);
    }

    return 1;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int insWrappedFunction(traceaux_t *paux, const char *name)

/************************************************************************/
{
    wrapfunc_t *table;
    int lower, upper, n;

    table = paux->WrappedFunctions;
    n = paux->NWrappedFunctions;

    /*  Binary search  */
    lower = 0;   upper = n;
    while(lower < upper) {
        int m = (lower + upper) >> 1;
        if(strcmp(table[m].Name, name) < 0)
            lower = m + 1;
        else
            upper = m;
    }

    if(lower < n && strcmp(table[lower].Name, name) == 0)
        /*  already in the table  */
        return lower;

    if(n >= paux->MWrappedFunctions) {
        REALLOC_RESIZE(paux->MWrappedFunctions, 20);
        if((table = (wrapfunc_t *)
            realloc(table, paux->MWrappedFunctions * sizeof(*table))) == NULL)
            EGOTO(error);

        paux->WrappedFunctions = table;
    }

    if(lower < n)
        memmove(&table[lower + 1], &table[lower], (n - lower) * sizeof(*table));
    paux->NWrappedFunctions++;

    if((table[lower].Name = strdup(name)) == NULL)
        EGOTO(error);
    table[lower].Flag = 0;

    return lower;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int genWrapperFunction(wrapaux_t *waux, const patchdesc_t *pdesc,
                              jsparscnode_t *node, jsparsvartab_t *vTab,
                              jsparsvardef_t *vDef)

/************************************************************************/
{
    traceaux_t *paux;
    const char *traceCalls[3], *funcName, *prefix[2];
    int ret;

    paux = waux->Paux;

    if(paux->Indentation > 0) {
        prefix[0] = "  ";
        prefix[1] = ";\n";
    } else {
        prefix[0] = " ";
        prefix[1] = "; ";
    }

    funcName = vDef->JPAVD_Name;

    traceCalls[0] = traceCalls[1] = traceCalls[2] = NULL;

    /* ***  Generate the entry statement  *** */
    if(pdesc->Type & PDESC_TYPE_FCALL) {
        paux->CurrentVarNumber = 0;

        if((traceCalls[0] = genFctEntryCall(paux, waux->JPA, vDef->JPAVD_Queue,
                                            pdesc, vTab, vDef->JPAVD_Declarator,
                                            funcName, prefix, 2)) == NULL ||
           (traceCalls[0] = strdup(traceCalls[0])) == NULL)
            EGOTO(error);
    }

    /* ***  Generate the return statement  *** */
    if(pdesc->Type & PDESC_TYPE_FRETURN) {
        jsparscnode_t *decl;
        const char *types, *marker;
        int useLabel;

        jstrbuffempty(&SBuffArgs);
        if(jstrbuffprintf(&SBuffArgs, ",%s", paux->NameRetVariable) == -1)
            EGOTO(error);
        jstrbuffclose(&SBuffArgs);

        /*  Get proper declarator  */
        decl = vDef->JPAVD_Declarator;
        if(decl) {
            decl = decl->JPACN_Sub;
            if(decl && decl->JPACN_Type == JPACT_IDENT) decl = decl->JPACN_Next;
            if(decl && decl->JPACN_Type == JPACT_FUNCCALL) decl = decl->JPACN_Next;
        }

        useLabel = (pdesc->Type & PDESC_TYPE_FRETURNLBL) ? 1 : 0;
        if((types = getTraceFormat(paux, waux->JPA, vTab,
                                   decl, vDef->JPAVD_Specifier,
                                   vDef->JPAVD_Queue, &useLabel, 1)) == NULL)
            EGOTO(error);

        marker = (pdesc->Type & PDESC_TYPE_FCALL) ? "-" : "";
        if((traceCalls[1] = genTraceFctCall(paux, pdesc, waux->Queue,
                                            node, paux->TraceFormatReturn,
                                            marker, funcName,
                                            types, jstrbuffer(&SBuffArgs),
                                            prefix)) == NULL ||
           (traceCalls[1] = strdup(traceCalls[1])) == NULL)
            EGOTO(error);

        if((traceCalls[2] = genTraceFctCall(paux, pdesc, waux->Queue,
                                            node, paux->TraceFormatReturn,
                                            marker, funcName,
                                            "", "",
                                            prefix)) == NULL)
            EGOTO(error);
    }

    /*  Now generate the whole statement  */
    jstrbuffempty(&TempBuffer);
    paux->CurrentVarNumber = 0;
    if(jstrbuffprint(&TempBuffer, "static ") == -1 ||
       (ret = jsparsgenwrapper2(waux->JPA, &TempBuffer,
                                vDef->JPAVD_Declarator, vDef->JPAVD_Specifier,
                                paux->PrefixWrapFunction2, paux->NameRetVariable,
                                traceCalls, vDef->JPAVD_Queue, &SBuffArgs,
                                getArgName, paux,
                                paux->Indentation > 0 ? 2 : 0)) == -1)
        EGOTO(error);

    if(traceCalls[0]) free((void *)traceCalls[0]);
    if(traceCalls[1]) free((void *)traceCalls[1]);

    if(ret == 0) {
        if(waux->JPA->JSPA_WarnFlags != 0)
            jsparsprinterr(waux->JPA, waux->Queue, node,
                           "Cannot generate wrapper function for %s", 0, funcName);
        return 0;
    }

    jstrbuffclose(&TempBuffer);

    if(tokInsertString(waux->JPA, waux->Tree, waux->Queue,
                       waux->Tree->JPACN_Start, jstrbuffer(&TempBuffer),
                       0) == -1)
        EGOTO(error2);

    return 1;
 error:
    if(traceCalls[0]) free((void *)traceCalls[0]);
    if(traceCalls[1]) free((void *)traceCalls[1]);
 error2:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static const char *getArgName(void *uPtr)

/************************************************************************/
{
    static char Buffer[20];

    sprintf(Buffer, GENPARAM_PATTERN,
            ++(((traceaux_t *)uPtr)->CurrentVarNumber));

    return Buffer;
}
