/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Add tracing calls to functions.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
#define popNodeStack(paux)    ((paux)->INodeStack--)
#define popExpandStack(paux)  ((paux)->IExpandMapStack--)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doPatchStmt(traceaux_t *paux,
				     jsparsstream_t *jpa,
				     jsparsvardef_t *fctDef,
				     jsparscnode_t *node,
				     const patchdesc_t *pdesc,
				     int mode);
static int               pushNodeStack(traceaux_t *paux,
				       jsparscnode_t *node);
static int               patchReturnStmt(traceaux_t *paux,
					 jsparsstream_t *jpa,
					 jsparsvardef_t *fctDef,
					 jsparscnode_t *node,
					 jctokfifo_t *queue,
					 const patchdesc_t *pdesc);
static int               needTempVariable(jsparsstream_t *jpa,
					  jsparscnode_t *node,
					  jctokfifo_t *queue, int level);
static int               checkTempVariable(traceaux_t *paux,
					   int start, int end,
					   jctokfifo_t *queue);

static int               patchVariable(traceaux_t *paux, jsparsstream_t *jpa,
				       const char *funcName,
				       jsparscnode_t *node,
				       jctokfifo_t *queue);
static int               checkPatchableType(jsparsstream_t *jpa,
					    jsparsexprtype_t *eType,
					    int pType);
static long              checkVariable(traceaux_t *paux, const char *funcname,
				       const char *name, const char *member);

static const char       *getNameType(traceaux_t *paux, jsparsstream_t *jpa,
				     jsparsvartab_t *vTab,
				     jsparsvardef_t *vDef, int *pUseLabel);

/*********     GLOBAL / STATIC VARIABLES                        *********/
static jstrbuff_t SBuffer;

/*JS*********************************************************************
 *   mode bit 0 + 1: do not call patchVariable
 *************************************************************************/

int doPatchStmt(traceaux_t *paux, jsparsstream_t *jpa,
		jsparsvardef_t *fctDef,
		jsparscnode_t *node,
		const patchdesc_t *pdesc, int mode)

/************************************************************************/
{
    jsparscnode_t *sub;
    jctokfifo_t *queue;
    int patchVar, patchReturn;

    queue = jsparsgetqueue(jpa);

    if(pushNodeStack(paux, node) == -1)
	EGOTO(error2);

    patchVar = patchReturn = FALSE;

    switch(node->JPACN_Type) {
    case JPACT_RETURN:
	patchReturn = !(pdesc->Type & PDESC_TYPE_FFCTWRAP) &&
	    (pdesc->Type & PDESC_TYPE_FRETURN);
	break;

	/* *********************************************************************** */
    case JPACT_PRAGMALINE: case JPACT_PRAGMALINE2:
	if(doPragma(jpa, paux, node, queue, 1) == -1)
	    EGOTO(error);
	break;

	/* *********************************************************************** */
    case JPACT_UNARYEXPR:
	if(!(mode & 3))
	    patchVar = TRUE;
	break;

	/* *********************************************************************** */
    case JPACT_ASS:
    case JPACT_ADDASS: case JPACT_SUBASS:
    case JPACT_MULASS: case JPACT_DIVASS:
    case JPACT_MODASS:
    case JPACT_LSHIFTASS: case JPACT_RSHIFTASS:
    case JPACT_ANDASS: case JPACT_ORASS:
    case JPACT_EXORASS:
	/*  Shouldn't call patchVariable for first subexpression (== lhs)  */
	mode |= 1;
	patchVar = TRUE;

	sub = node->JPACN_Sub->JPACN_Next;
	if(sub->JPACN_Type == JPACT_BLOCK)
	    /*  Must not patch anything in the block  */
	    mode |= 2;

#ifdef CONFIG_JSCPPLIB_DEBUG
	if(Debug & 6) {
	    jsparsexprtype_t eType;
	    jsparsvartab_t *lTab;

	    /*  This is for testing the jsparsgetexprtype routine:
	     *   If we see an assignment, we determine the type of
	     *   the right hand side
	     */

	    if((lTab = jsparsfindvartab(jpa, node)) == NULL)
		EGOTO(error);

	    if(jsparsgetexprtype(&eType, jpa, lTab,
				 node->JPACN_Sub->JPACN_Next,
				 queue, 0) == NULL)
		EGOTO(error);

	    if(Debug & 4) {
		int ind;

		if((ind = jctokindfpos(queue)) != -1) {
		    ind += node->JPACN_Start;
		    printf("File '%s' l. %d ", queue->TKFF_Token[ind].TK_String,
			   queue->TKFF_Token[ind].TK_Value);
		}

		printf("EXPRESSION:\n");
		jsparsprctree(stdout, queue, node, 0, 0);
	    }

	    printf("TYPE (flag %ld):\n", eType.JSPE_Flags);

	    if(eType.JSPE_Declarator)
		jsparsprctree(stdout, eType.JSPE_Queue, eType.JSPE_Declarator, 2, 4);

	    if(eType.JSPE_Specifier)
		jsparsprctree(stdout, eType.JSPE_Queue, eType.JSPE_Specifier, 2, 0);

	    jsparsfreeexprtype(jpa, &eType, 1);
	}
#endif
	break;

    default:
	break;
    }

    /* *********************************************************************** */

    /*  Scan recursively  */
    for(sub = node->JPACN_Sub ; sub ; sub = sub->JPACN_Next) {
	if(doPatchStmt(paux, jpa, fctDef, sub, pdesc, mode) == -1)
	    EGOTO(error);

	mode &= ~1; /*  switch off for follow-up expressions  */
    }

    if(patchVar &&
       patchVariable(paux, jpa, fctDef->JPAVD_Name, node, queue) == -1)
	EGOTO(error);

    if(patchReturn &&
       patchReturnStmt(paux, jpa, fctDef, node, queue, pdesc) == -1)
	EGOTO(error);

    popNodeStack(paux);

    return 0;

 error:
    popNodeStack(paux);
 error2:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int pushNodeStack(traceaux_t *paux, jsparscnode_t *node)

/************************************************************************/
{
    if(paux->INodeStack >= paux->MNodeStack) {
	paux->MNodeStack += 20;
	if((paux->NodeStack = (jsparscnode_t **)
	    realloc(paux->NodeStack,
		    paux->MNodeStack * sizeof(*(paux->NodeStack)))) == NULL)
	    return -1;
    }

    paux->NodeStack[paux->INodeStack++] = node;

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int patchReturnStmt(traceaux_t *paux, jsparsstream_t *jpa,
			   jsparsvardef_t *fctDef,
			   jsparscnode_t *node, jctokfifo_t *queue,
			   const patchdesc_t *pdesc)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    jsparscnode_t *nvalue;
    const char *def, *args, *types;
    const char *funcName;
    int needTempVar, start, end;

    jstrbuffempty(&SBuffTypes);
    jstrbuffempty(&SBuffArgs);

    funcName = fctDef->JPAVD_Name;

    nvalue = node->JPACN_Sub;

    start = node->JPACN_Start;
    end = node->JPACN_End;

    /*  Check if we can patch the return statement as a whole as well as the
     *   expression (if there is one).
     */
    if(jpa->JSPA_Mode & JPAM_FGETORIG) {
	if(!jsparsmatchorigtoks(queue, &start, &end))
	    goto impossible;

	if(nvalue != NULL) {
	    int start2, end2;

	    start2 = nvalue->JPACN_Start;
	    end2 = nvalue->JPACN_End;

	    if(!jsparsmatchorigtoks(queue, &start2, &end2))
		goto impossible;
	}
    }

    if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

    /* ***  Default: Only trace the return itself without any arguments  *** */
    needTempVar = FALSE;
    def = args = types = "";

    if((vTab = jsparsfindvartab(jpa, node)) == NULL)
	EGOTO(error);

    /* *********************************************************************** */
    if(pdesc->NReturnDescriptions > 0) {
	/*  Make the return statement to nothing so '%o' evaluates to the
	 *   expression alone (but still includes the trailing ';')
	 */
	if(jctoksubstitute(queue, start, start, "") == -1)
	    EGOTO(error);

	/*  Generate the proper variable definition for the temporary
	 *   return variable
	 */
	jstrbuffempty(&SBuffer);
	if(jsparsgenvardef(jpa, &SBuffer, fctDef, paux->NameRetVariable) == -1)
	    EGOTO(error);

	if(genTraceStmt(paux, jpa, vTab, &SBuffTypes, &SBuffArgs,
			pdesc->ReturnDescriptions,
			pdesc->NReturnDescriptions, NULL, NULL, 0) == -1)
	    EGOTO(error);

	needTempVar = TRUE;
	def = jstrbuffer(&SBuffer);
	types = jstrbuffer(&SBuffTypes);

	if((args = jstrbuffer(&SBuffArgs)) == NULL)
	    args = "";

	/* *********************************************************************** */
    } else if(pdesc->NReturnDescriptions == -1 && nvalue != NULL) {
	jsparsexprtype_t fctType;
	const char *expression, *newTypes;
	int useLabel, flag;

	flag = 0;
	if((needTempVar = needTempVariable(jpa, nvalue, queue, 0))) {
	    /*  Generate the proper variable definition for the temporary
	     *   return variable
	     */
	    jstrbuffempty(&SBuffer);

	    expression = paux->NameRetVariable;

	    if(jsparsgenvardef(jpa, &SBuffer, fctDef, expression) == -1)
		EGOTO(error);
	} else {
	    int nvstart, nvend;

	    /*  This expression is simple enough, we can repeat it in the tracing
	     *   call and the final return
	     */
	    nvstart = nvalue->JPACN_Start;
	    nvend = nvalue->JPACN_End;

	    if(jpa->JSPA_Mode & JPAM_FGETORIG) {
		/*  We call this to allow adapting the range. Since we already checked
		 *   this before, we *know* that it does not fail.
		 */
		if(!jsparsmatchorigtoks(queue, &nvstart, &nvend))
		    EGOTO(error);

		++nvstart;
		++nvend;
	    }

	    if((expression = jctoktostring(queue, nvstart, nvend)) == NULL)
		EGOTO(error);

	    flag |= 1;
	}

	jstrbuffempty(&SBuffTypes);
	jstrbuffempty(&SBuffArgs);

	if(jsparsvdef2type(jpa, vTab, &fctType, fctDef, 1) == NULL)
	    EGOTO(error);

	useLabel = (pdesc->Type & PDESC_TYPE_FRETURNLBL) ? 1 : 0;
	if(genTracePart(paux, jpa, vTab, &SBuffTypes, &SBuffArgs,
			expression, expression, &fctType, &useLabel, 0) == -1)
	    EGOTO(error);

	if(flag & 1)
	    free((void *)expression);

	newTypes = jstrbuffer(&SBuffTypes);
	if(newTypes != NULL && *newTypes != '\0' && strcmp(newTypes, "?") != 0) {
	    types = newTypes;

	    /*  We have some arguments to show on function entry  */
	    if((args = jstrbuffer(&SBuffArgs)) == NULL)
		args = "";

	    if(needTempVar) {
		/*  Make the return statement to nothing so '%o' evaluates to the
		 *   expression alone (but still includes the trailing ';')
		 */
		if(jctoksubstitute(queue, start, start, "") == -1)
		    EGOTO(error);

		def = jstrbuffer(&SBuffer);
	    } else
		def = "";
	} else
	    needTempVar = FALSE;

	/* ******************************************************************** */
    }

    {
	const char *traceCall;
	char format[200];
	int i, needBraces;
	long flowFlags;

	if(needTempVar && checkTempVariable(paux, start, end, queue)) {
	    fprintf(stderr, "\
ERROR: '%s' l. %d: Cannot patch return, already contains temporary '%s'!\n",
		    queue->TKFF_Token[start + 1].TK_String,
		    queue->TKFF_Token[start + 1].TK_Value, paux->NameRetVariable);

	    EGOTO(error);
	}

	if((traceCall = genTraceFctCall(paux, pdesc, queue, node, paux->TraceFormatReturn,
					(pdesc->Type & PDESC_TYPE_FCALL) ?
					"-" : "", funcName, types, args, NULL)) == NULL)
	    EGOTO(error);

	if((i = jsparsfindflow(jpa, node)) == -1)
	    EGOTO(error);

	flowFlags = jpa->JSPA_Statements[i].JPAFLN_Flags;

	needBraces = needTempVar ? !(flowFlags & JPA_FLOW_BLOCKSTART) :
	    (flowFlags & JPA_FLOW_SINGLESTMT);

	i = 0;
	if(needBraces) {
	    format[i++] = '{';
	    if(paux->Indentation > 0)
		i += sprintf(format + i, "%%.%dI", paux->Indentation);
	    else
		format[i++] = ' ';
	}

	i += sprintf(format + i, "%%s%s", needTempVar ? " = %o " : "");

	if(*def && paux->Indentation > 0) {
	    if(needBraces)
		i += sprintf(format + i, "%%.%dI", paux->Indentation);
	    else
		i += sprintf(format + i, "%%I");
	}

	i += sprintf(format + i, "%%s;");

	if(paux->Indentation > 0) {
	    if(needBraces)
		i += sprintf(format + i, "%%.%dI", paux->Indentation);
	    else
		i += sprintf(format + i, "%%I");
	} else
	    format[i++] = ' ';

	if(needTempVar)
	    i += sprintf(format + i, "return %%s;");
	else
	    i += sprintf(format + i, "%%o");

	if(needBraces)
	    i += sprintf(format + i, "%s}",
			 (paux->Indentation > 0) ? "%i" : " ");

	assert(i < (int)sizeof(format));

	if(jctoksubstitute(queue, start, end, format,
			   def, traceCall, paux->NameRetVariable) == -1)
	    EGOTO(error);
    }

    return 0;
 impossible:
    if(jpa->JSPA_WarnFlags != 0)
	jsparsprinterr(jpa, queue, node,
		       "Cannot patch return statement in function '%s'", 0,
		       fctDef->JPAVD_Name);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int needTempVariable(jsparsstream_t *jpa, jsparscnode_t *node,
			    jctokfifo_t *queue, int level)

/************************************************************************/
{
    if(node->JPACN_Type != JPACT_UNARYEXPR) {
	/*  FIXME: This would allow something like "1 << 0", but also if
	 *   the return expression is rather big and complex. So let's
	 *   be conservative and rather use a temporary. This makes the
	 *   code more readable anyway.
	 */
#if 0
	if(level > 2)
	    /*  Expression is too deeply nested, so we better use a temporary  */
	    return TRUE;

	switch(node->JPACN_Type) {
	case JPACT_LIST: /*  comma operator  */
	case JPACT_ASS:
	case JPACT_MULASS:    case JPACT_DIVASS:    case JPACT_MODASS:
	case JPACT_ADDASS:    case JPACT_SUBASS:
	case JPACT_LSHIFTASS: case JPACT_RSHIFTASS:
	case JPACT_ANDASS:    case JPACT_EXORASS:   case JPACT_ORASS:
	    return TRUE;
	default:
	    break;
	}

	/*  Check each subexpression for side effects  */
	for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
	    if(needTempVariable(jpa, node, queue, level + 1))
		return TRUE;

	return FALSE;
#else
	return TRUE;
#endif
    }

    if((node = node->JPACN_Sub) == NULL)
	/*  What is that?  */
	return TRUE;

    switch(node->JPACN_Type) {
    case JPACT_IDENT:
	{
	    int ind;

	    /*  This might have been patched already as a variable  */
	    ind = node->JPACN_Start;
	    if(jpa->JSPA_Mode & JPAM_FGETORIG)
		++ind;

	    if(queue->TKFF_Token[ind].TK_Value == CTK_VANILLA)
		return TRUE;
	}
	break;

    case JPACT_INTCONST:  case JPACT_ENUMCONST:  case JPACT_FLOATCONST:
    case JPACT_CHARCONST: case JPACT_WCHARCONST:
    case JPACT_STRING:    case JPACT_WSTRING:
	break;

    default:
	/*  We are not taking any risk - the expression looks ugly enough,
	 *   so better use temporaries
	 */
	return TRUE;
    }

    /*  OK, the thingadoo is simple enough, now check how it is adorned  */
    for(node = node->JPACN_Next ; node ; node = node->JPACN_Next)
	switch(node->JPACN_Type) {
	case JPACT_UMINUS: case JPACT_UPLUS:
	case JPACT_NOT:    case JPACT_LNOT:
	    /*  These are *known* to be OK  */
	    break;
	default:
	    return TRUE;
	}

    return FALSE;
}

/*JS*********************************************************************
 *   Check that the temporary variable does not appear anywhere in the
 *    return expression. Otherwise we surely alter the semantics of the
 *    statement.
 *************************************************************************/

static int checkTempVariable(traceaux_t *paux, int start, int end,
			     jctokfifo_t *queue)

/************************************************************************/
{
    int step;

    step = queue->TKFF_Step;

    for( ; start <= end ; start += step)
	if(queue->TKFF_Token[start].TK_Value == CTK_IDENT &&
	   tkStrcmp(queue->TKFF_Token[start].TK_String,
		    paux->NameRetVariable) == 0)
	    return 1;

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int patchVariable(traceaux_t *paux, jsparsstream_t *jpa,
			 const char *funcName, jsparscnode_t *node,
			 jctokfifo_t *queue)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    jsparscnode_t *node2;
    jsparsexprtype_t eType;
    jsparsvardef_t *vDef;
    const char *name, *member, *vType;
    int start, end, ind, indorig;
    int mask, flag, pType, useLabel;

    /*
     *  pType:
     *   1   var
     *   2   var++ or ++var
     *   3   var-- or --var
     *   4   &var
     *   5   var = ...
     *   6   int var = ...
     */

    member = NULL;
    start = node->JPACN_Start;
    end = node->JPACN_End;
    if(node->JPACN_Type == JPACT_UNARYEXPR ||
       node->JPACN_Type == JPACT_DECLARATOR) {
	node2 = node->JPACN_Sub;
	if(node2 == NULL || node2->JPACN_Type != JPACT_IDENT)
	    return 0; /*  empty statement or constant expression  */

	ind = node2->JPACN_IndexPP;
	name = queue->TKFF_Token[ind].TK_String;

	pType = 1;
	mask = VARTRACE_FREAD;

	/*  Check for operators that modify the variable and
	 *   struct member. In cases like "a.b->c.d" only
	 *   the last member is relevant.
	 */
	for(node2 = node2->JPACN_Next ; node2 ; node2 = node2->JPACN_Next) {
	    jsparscnode_t *n;
	    int nType = node2->JPACN_Type;

	    if(nType == JPACT_FUNCCALL)
		return 0; /*  We have a function call, so no associated variable  */

	    n = node2;
	    if(nType == JPACT_PREINC || nType == JPACT_POSTINC) {
		pType = 2;
		mask = VARTRACE_FWRITE;

	    } else if(nType == JPACT_PREDEC || nType == JPACT_POSTDEC) {
		pType = 3;
		mask = VARTRACE_FWRITE;

	    } else if(nType == JPACT_ADDRESS) {
		pType = 4;
		mask = VARTRACE_FWRITE;

	    } else if(nType == JPACT_MEMBER || nType == JPACT_POINTTO) {
#if 1 /*  FIXME  */
		if(mask & VARTRACE_FWRITE) {
		    /*  If a previous part had a side effect like in
		     *     (--a)->b++->c
		     *   we better refrain from tracing it since we cannot
		     *   properly handle side effects
		     */
		    return 0;
		}
#endif

		member = queue->TKFF_Token[node2->JPACN_Sub->
					   JPACN_IndexPP].TK_String;
	    } else
		n = NULL;

	    if(n != NULL) {
		if(n->JPACN_End > end)     end   = n->JPACN_End;
		if(n->JPACN_Start < start) start = n->JPACN_Start;
	    }
	}

	node2 = node;

    } else {
	jsparscnode_t *sub, *next;

	/*  node points to an assignment  */
	sub = node->JPACN_Sub;
	next = sub->JPACN_Next;

	pType = 5;
	mask = VARTRACE_FWRITE;

	if(sub->JPACN_Type == JPACT_UNARYEXPR) {
	    sub = sub->JPACN_Sub;

	} else if(sub->JPACN_Type == JPACT_DECLARATOR) {
	    /*  Special treatment for assignments within definition, see
	     *   below
	     */
	    sub = sub->JPACN_Sub;
	    pType = 6;

	    /*  We only change the expression part in this case  */
	    start = next->JPACN_Start;
	    end = next->JPACN_End;

	    /*  Make sure that we do not try to patch definitions like
	     *   "static int X = 1;"
	     */
	    for(node2 = node->JPACN_Next ; node2 ; node2 = node2->JPACN_Next)
		if(node2->JPACN_Type == JPACT_CONST ||
		   node2->JPACN_Type == JPACT_STATIC)
		    return 0;
	}

	/*  FIXME: For now, we can only patch assignments where the left hand
	 *   side is a single variable or a struct/union member
	 */
	if(sub->JPACN_Type != JPACT_IDENT)
	    return 0;

	for(node2 = sub->JPACN_Next ; node2 ; node2 = node2->JPACN_Next) {
	    int nType = node2->JPACN_Type;

	    if(nType == JPACT_MEMBER || nType == JPACT_POINTTO) {
		member = queue->TKFF_Token[node2->JPACN_Sub->
					   JPACN_IndexPP].TK_String;
	    } else
		/*  Anything else - we better stop  */
		return 0;
	}

	if(next->JPACN_Type == JPACT_BLOCK)
	    /*  something like "TYPE s = { ... };"   */
	    return 0;

	ind = sub->JPACN_IndexPP;
	name = queue->TKFF_Token[ind].TK_String;

	node2 = node->JPACN_Sub;
    }

    /*  If it is an expression statement like in "i = 1;", the
     *   trailing semicolon is part of the expression
     */
    if(queue->TKFF_Token[end].TK_Value == ';')
	end -= queue->TKFF_Step;

    if((vTab = jsparsfindvartab(jpa, node)) == NULL)
	EGOTO(error);

    /*  Parameters are already defined in the function body,
     *   so get the definition
     */
    if(jsparsgetexprtype(&eType, jpa, vTab, node2, queue, 2) == NULL)
    {
        jsparsprctree(stdout, queue, node2, 0, 0);
	EGOTO(error);
    }

    /*  Anything to do for this variable?  */
    if((flag = checkVariable(paux, funcName, name, member)) == -1)
	EGOTO(error2);

    if((flag & mask) == 0) {
	jsparsfreeexprtype(jpa, &eType, 1);
	return 0;
    }

    {
	int mode;

	switch(flag & (VARTRACE_FGLOBAL | VARTRACE_FLOCAL)) {
	case  VARTRACE_FGLOBAL: mode = 2; break; /*  only global vars  */
	case  VARTRACE_FLOCAL:  mode = 4; break; /*  only local vars   */
	default:                mode = 6; break; /*  both  */
	}

	if((vDef = jsparsgetvardef(jpa, vTab, name, JVDF_TVARIABLE, mode)) == NULL) {
	    /*  We should not patch this  */
	    jsparsfreeexprtype(jpa, &eType, 1);
	    return 0;
	}
    }

    if((indorig = jctokindorig(queue)) != -1 &&
       !jsparsmatchorigtoks(queue, &start, &end)) {
	if(jpa->JSPA_WarnFlags & JPAM_WARNVARS)
	    jsparsprinterr(jpa, queue, node,
			   "Cannot patch usage of variable '%s'.", 0,
			   name);

	jsparsfreeexprtype(jpa, &eType, 1);
	return 0;
    }

    if((flag = checkPatchableType(jpa, &eType, pType)) == -1)
	EGOTO(error2);

    jsparsfreeexprtype(jpa, &eType, 1);

    if(flag == 0)
	return 0;

    useLabel = 0;
    if((vType = getNameType(paux, jpa, vTab, vDef, &useLabel)) == NULL)
	EGOTO(error);

    /*  name lives on the same queue we are going to modify, so make a copy
     *   beforehand
     */
    if((name = strdup(name)) == NULL)
	EGOTO(error);

    /*  FIXME: Should use 'useLabel' in what follows. Should generate
     *   the format in buffer.
     */
    /*jstrbuffempty(&SBuffer);*/

    if(indorig != -1) { start += indorig; end += indorig; }

    switch(pType) {
    case 1:
	flag = jctoksubstitute(queue, start, end,
			       "(%s(__FILE__,__LINE__,\"%s<%s\",%s), %o)",
			       paux->TraceVarFunction, vType, name, name);
	break;

    case 2:
	flag = jctoksubstitute(queue, start, end,
			       "(%s(__FILE__,__LINE__,\"%s>%s\",%s + 1), %o)",
			       paux->TraceVarFunction, vType, name, name);
	break;

    case 3:
	flag = jctoksubstitute(queue, start, end,
			       "(%s(__FILE__,__LINE__,\"%s>%s\",%s - 1), %o)",
			       paux->TraceVarFunction, vType, name, name);
	break;

    case 4:
	flag = jctoksubstitute(queue, start, end,
			       "(%s(__FILE__,__LINE__,\"%s&%s\",%s), %o)",
			       paux->TraceVarFunction, vType, name, name);
	break;

    case 5:
	/*  Patch assignment  */
	flag = jctoksubstitute(queue, start, end,
			       "(%o, %s(__FILE__,__LINE__,\"%s>%s\",%s), %s)",
			       paux->TraceVarFunction, vType,
			       name, name, name);
	break;

    case 6:
	/*  Patch assignment within variable definition. We use a
	 *   "dirty trick" here: We cannot patch "int i = 1;"
	 *   directly, so we treat it as "int i = (i = 1);" and
	 *   patch this. According to ANSI, after having seen "int i",
	 *   the variable "i" should be known, so the statement should
	 *   be correct.
	 */
	flag = jctoksubstitute(queue, start, end,
			       "(%s = %o, %s(__FILE__,__LINE__,\"%s>%s\",%s), %s)",
			       name, paux->TraceVarFunction, vType,
			       name, name, name);
	break;
    }

    free((void *)name);

    if(flag == -1)
	EGOTO(error);

    return 0;

 error2:
    jsparsfreeexprtype(jpa, &eType, 1);
 error:
    return -1;
}

/*JS*********************************************************************
 *   Return value 0: cannot be patched. 1: Can be patched.
 *************************************************************************/

static int checkPatchableType(jsparsstream_t *jpa, jsparsexprtype_t *eType,
			      int pType)

/************************************************************************/
{
    jsparscnode_t *node;

    /*  We have to check if the variable is an array that is initialized, e. g.
     *    typedef char TEXT[100];
     *    TEXT t = "hello";
     */
    if(pType == 6 && (node = eType->JSPE_Declarator) &&
       node->JPACN_Type == JPACT_VECTOR)
	/*  First node after identifier is a vector, so we can't patch  */
	return 0;

    return 1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static long checkVariable(traceaux_t *paux, const char *funcname,
			  const char *name, const char *member)

/************************************************************************/
{
    int i;

    if(member != NULL)  /*  FIXME: Disable member tracing for now  */
	return 0;

#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & 8)
	/*  pretend to trace everything  */
	return VARTRACE_FREAD | VARTRACE_FWRITE;
#endif

    for(i = paux->NVarTraces - 1 ; i >= 0 ; --i) {
	const char *pattern;

	if((pattern = paux->VarTraces[i].FuncPattern) != NULL) {
	    if(!(jiswildcard(pattern) ? wildcmp(funcname, pattern, 0) :
		 strcmp(funcname, pattern) == 0))
		continue;
	}

	if((pattern = paux->VarTraces[i].VarPattern) == NULL ||
	   (jiswildcard(pattern) ? wildcmp(name, pattern, 0) :
	    strcmp(name, pattern) == 0))
	    return paux->VarTraces[i].Flag;
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static const char *getNameType(traceaux_t *paux, jsparsstream_t *jpa,
			       jsparsvartab_t *vTab, jsparsvardef_t *vDef,
			       int *pUseLabel)

/************************************************************************/
{
    jsparscnode_t *decl, *spec;
    jctokfifo_t *queue;

    decl = vDef->JPAVD_Declarator;
    spec = vDef->JPAVD_Specifier;
    queue = vDef->JPAVD_Queue;

    if(decl->JPACN_Type != JPACT_DECLARATOR && decl->JPACN_Type != JPACT_PARDECL)
	goto unknown;

    decl = decl->JPACN_Sub;
    if(decl == NULL || decl->JPACN_Type != JPACT_IDENT)
	goto unknown;

    decl = decl->JPACN_Next;

    return getTraceFormat(paux, jpa, vTab, decl, spec, queue, pUseLabel, 0);

 unknown:
    return "";
}
