/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Add tracing calls to functions.
*
*              Copyright (C) 2003-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
#define popNodeStack(paux)    ((paux)->INodeStack--)
#define popExpandStack(paux)  ((paux)->IExpandMapStack--)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            genTraceStmt(traceaux_t *paux,
				      jsparsstream_t *jpa,
				      jsparsvartab_t *vTab,
				      jstrbuff_t *sbuffTypes,
				      jstrbuff_t *sbuffArgs,
				      argdesc_t *desc, int n,
				      const char *name,
				      const char *defLabel,
				      int mode);
Prototype int            genTracePart(traceaux_t *paux, jsparsstream_t *jpa,
				      jsparsvartab_t *vTab,
				      jstrbuff_t *sbuffTypes,
				      jstrbuff_t *sbuffArgs,
				      const char *label,
				      const char *expression,
				      jsparsexprtype_t *pType,
				      int *pUseLabel, int mode);

static int               expandMapping(traceaux_t *paux, jsparsstream_t *jpa,
				       jsparsvartab_t *vTab,
				       jstrbuff_t *sbuffTypes,
				       jstrbuff_t *sbuffArgs,
				       const char *expression,
				       jsparsexprtype_t *pType,
				       const char *defLabel, int mode);
static int               pushExpandStack(traceaux_t *paux, int indx);

static int               getExprType(jsparsstream_t *jpa, jsparsvartab_t *vTab,
				     jsqueuestream_t *jsqs,
				     jsparsexprtype_t *pType,
				     const char *name, int mode);
static jsctoken_t        nameconverter(void *expression, jsctoken_t token);

static int               getTypeMapping(traceaux_t *paux,
					jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparsexprtype_t *pType,
					argdesc_t **pDescs,
					int *pNSubDesc);

Prototype const char    *getTraceFormat(traceaux_t *paux,
					jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparscnode_t *decl,
					jsparscnode_t *spec,
					jctokfifo_t *queue,
					int *pUseLabel, int mode);
static const char       *getTypeFormat(long flag, int ptrcnt, int useLabel);

/*JS*********************************************************************
*
*************************************************************************/

int genTraceStmt(traceaux_t *paux, jsparsstream_t *jpa, jsparsvartab_t *vTab,
		 jstrbuff_t *sbuffTypes, jstrbuff_t *sbuffArgs,
		 argdesc_t *desc, int n, const char *name,
		 const char *defLabel, int mode)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < n ; ++i) {
    jsparsexprtype_t eType;
    const char *label, *expression;
    int useLabel;

    label = desc[i].Label;
    expression = desc[i].Expression;

#ifdef CONFIG_JSPARSE_ARGLABELS
    if(label == NULL)
      label = expression;
#endif

    if(label == NULL)
      label = defLabel;

    useLabel = (label != NULL && *label != '\0');

    /*  Parse the expression  */
    if(getExprType(jpa, vTab, desc[i].QStream, &eType, name, mode) == -1)
      EGOTO(error);

    if(expression == NULL) {
      jctokfifo_t *queue;
      tokenseq_t *tseq;

      if(name == NULL) {
	/*  Should not occur!  */
	EGOTO(error);
      }

      queue = jsqsgetqueue(desc[i].QStream);
      tseq = (tokenseq_t *)desc[i].QStream->JSQS_User;

      if((expression = jctoktostring2(queue, tseq->Start, tseq->End,
				      nameconverter, (void *)name)) == NULL)
	EGOTO(error);
    }

    if(genTracePart(paux, jpa, vTab, sbuffTypes, sbuffArgs,
		    label, expression, &eType, &useLabel, mode) == -1)
      EGOTO(error);

    jsparsfreeexprtype(jpa, &eType, 1);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int genTracePart(traceaux_t *paux, jsparsstream_t *jpa, jsparsvartab_t *vTab,
		 jstrbuff_t *sbuffTypes, jstrbuff_t *sbuffArgs,
		 const char *label, const char *expression,
		 jsparsexprtype_t *pType, int *pUseLabel, int mode)

/************************************************************************/
{
  const char *format;
  int ret;

  if((ret = expandMapping(paux, jpa, vTab, sbuffTypes, sbuffArgs,
			  expression, pType, *pUseLabel ? label : NULL,
			  mode)) == -1)
    EGOTO(error);

  if(ret != 0)
    return 0;

  /*  Is this type any of the registered ones?  */
  if((format = getTraceFormat(paux, jpa, vTab, pType->JSPE_Declarator,
			      pType->JSPE_Specifier,
			      pType->JSPE_Queue, pUseLabel, 0)) == NULL)
    EGOTO(error);

  if(*format != '\0') {
    if(jstrbuffprint(sbuffTypes, format) == -1)
      EGOTO(error);

    if(*pUseLabel) {
      if(jstrbuffprint(sbuffArgs, ",") == -1)
	EGOTO(error);

      format = paux->LabelFormat;

      while(*format != '\0') {
	int len;

	if(*format != '%') {
	  const char *next;

	  if((next = strchr(format, '%')) != NULL)
	    len = next - format;
	  else
	    len = strlen(format);

	  if(jstrbuffstore(sbuffArgs, format, len) == -1)
	    EGOTO(error);

	  format += len;
	  continue;
	}

	/*  Treat format specifiers  */
	++format;
	if(*format == '%')
	  len = 1;
	else if(*format == 's')
	  len = 4 * strlen(label);
	else
	  EGOTO(error);

	if(jstrbuffstore(sbuffArgs, NULL, len) == -1)
	  EGOTO(error);

	if(*format == '%')
	  jstrbuffputc(sbuffArgs, '%');
	else {
	  if((len = cStrSPrint(jstrbuffer(sbuffArgs) + jstrbufflen(sbuffArgs),
			       label, strlen(label))) == -1)
	    EGOTO(error);

	  jstrbuffinc(sbuffArgs, len);
	}

	++format;
      }

      jstrbuffclose(sbuffArgs);
    }

    if(jstrbuffprintf(sbuffArgs, ",%s", expression) == -1)
      EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int expandMapping(traceaux_t *paux, jsparsstream_t *jpa,
			 jsparsvartab_t *vTab,
			 jstrbuff_t *sbuffTypes, jstrbuff_t *sbuffArgs,
			 const char *expression,
			 jsparsexprtype_t *pType,
			 const char *defLabel, int mode)

/************************************************************************/
{
  argdesc_t *subDesc;
  jsparscnode_t decl;
  int isDefined;
  int nSubDesc, indx, j, count;

  /*  Check if your mapping expansion has gone too far  */
  if(paux->MaxMappingLevel != -1 &&
     paux->IExpandMapStack > paux->MaxMappingLevel)
    return 0;

  if((indx = getTypeMapping(paux, jpa, vTab, pType, &subDesc,
			    &nSubDesc)) == -1)
    EGOTO(error);

  if(indx == 0)
    return 0;

  /*  Check how often we already expand this  */
  count = 0;
  for(j = 0 ; j < paux->IExpandMapStack ; ++j)
    if(paux->ExpandMapStack[j] == indx)
      ++count;

  /*  If we are already expanding this, prevent infinite recursion  */
  if(count >= paux->MaxRecursiveLevel)
    return 0;

  if(jsparsgetvardef(jpa, vTab, expression, JVDF_TVARIABLE, 8) != NULL) {
    /*  Is already defined, so no need to insert  */
    isDefined = FALSE;
  } else {
    /*  Create pseudo variable definition for the expression. That is
     *   the easiest way which also avoids having to trace it all over
     *   again if we need to determine its type.
     */
    decl.JPACN_Next = NULL;
    decl.JPACN_Type = JPACT_DECLARATOR;
    if((decl.JPACN_Sub = pType->JSPE_Declarator) != NULL) {
      decl.JPACN_Start = pType->JSPE_Declarator->JPACN_Start;
      decl.JPACN_End = pType->JSPE_Declarator->JPACN_End;
    } else
      decl.JPACN_Start = decl.JPACN_End = -1;

    if(jsparsdefinevar2(jpa, vTab, JVDF_FDEFINED | JVDF_TVARIABLE,
			pType->JSPE_Queue, expression,
			&decl, pType->JSPE_Specifier, 0) == NULL)
      EGOTO(error);

    isDefined = TRUE;
  }

  /* ***  Recursive call  *** */
  if(pushExpandStack(paux, indx) == -1)
    EGOTO(error);

  j = genTraceStmt(paux, jpa, vTab, sbuffTypes, sbuffArgs, subDesc,
		   nSubDesc, expression, defLabel, mode);

  popExpandStack(paux);

  if(isDefined) {
    if(jsparsdelvardef(jpa, vTab, expression, JVDF_TVARIABLE, 8) != 1)
      EGOTO(error);
  }

  if(j == -1)
    EGOTO(error);

  /*  Signal that we did something  */
  return 1;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int pushExpandStack(traceaux_t *paux, int indx)

/************************************************************************/
{
  if(paux->IExpandMapStack >= paux->MExpandMapStack) {
    paux->MExpandMapStack += 20;
    if((paux->ExpandMapStack = (int *)
	realloc(paux->ExpandMapStack, paux->MExpandMapStack *
		sizeof(*(paux->ExpandMapStack)))) == NULL)
      return -1;
  }

  paux->ExpandMapStack[paux->IExpandMapStack++] = indx;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getExprType(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		       jsqueuestream_t *jsqs, jsparsexprtype_t *pType,
		       const char *name, int mode)

/************************************************************************/
{
static jctokfifo_t CopyQueue;
static int Init = 0;
  jsppstream_t *jps;
  jsparscnode_t *node;
  jctokfifo_t *queue;

  queue = jsqsgetqueue(jsqs);

  /*  Preprocess the expression before parsing it
   *   Also, replace '@' by name if given.
   */
  tokseqrewind((tokenseq_t *)jsqs->JSQS_User);

  if(Init == 0) {
    jctokcreate(&CopyQueue, 0);
    Init = 1;
  }

  jctokclear(&CopyQueue);

  /*  Prepare the preprocessor input stream  */
  jps = jsqsgetppstream(jpa->JSPA_Stream);
  if(jspppushinplevel(jps) == -1 ||
     jspppushppstream(jps, jsqs, 0) == -1)
    EGOTO(error);

  for(;;) {
    jsctoken_t token;

    /*  Do preprocessing when building the expression  */
    token = jsppgettoken(jps, NULL, 0);
    if(token.TK_Value == CTK_ERROR)
      EGOTO(error);

    if(token.TK_Value == CTK_EOF)
      break;

    if(token.TK_Value == '@' && name != NULL) {
      token.TK_Value  = CTK_IDENT;
      token.TK_String = (char *)name;
    }

    if(jctokinsert(&CopyQueue, token, 0) == -1)
      EGOTO(error);
  }

  if(jspppopinplevel(jps) == -1)
    EGOTO(error);

  /* ******************************************************* */

  if((jsqs = jsqscreatequeuestream(&CopyQueue, NULL, NULL)) == NULL)
    EGOTO(error);

  queue = &CopyQueue;

  if((node = jsparsgetexpression(jpa, jsqs, 0)) == NULL)
    EGOTO(error);

  /*  Determine the type in the current context. Mode is set to '1'
   *   to bail out if there are side-effects in the expression.
   */
  if(jsparsgetexprtype(pType, jpa, vTab, node, queue, 1) == NULL)
    EGOTO(error);

  jsparsfreecnode(node);

  if(queue == &CopyQueue)
    jsqsfree(jsqs);

  return 0;
error:
  if(queue == &CopyQueue)
    jsqsfree(jsqs);

  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static jsctoken_t nameconverter(void *expression, jsctoken_t token)

/************************************************************************/
{
  if(token.TK_Value == '@') {
    token.TK_Value = CTK_VANILLA;
    token.TK_String = (char *)expression;
  }

  return token;
}
/*JS*********************************************************************
*
*************************************************************************/

static int getTypeMapping(traceaux_t *paux, jsparsstream_t *jpa,
			  jsparsvartab_t *vTab, jsparsexprtype_t *pType,
			  argdesc_t **pDescs, int *pNSubDesc)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < paux->NTypeMappings ; ++i) {
    typemapping_t *tMap;
    int ret;

    tMap = &(paux->TypeMappings[i]);

    if((ret = jsparsisvartype(jpa, vTab, pType, &(tMap->ExprType), 0)) == -1)
      EGOTO(error);

    if(ret != 0) {
      *pDescs = tMap->Descriptions;
      *pNSubDesc = tMap->NDescriptions;
      return 1 + i;
    }
  }

  /*  nothing found  */
  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

const char *getTraceFormat(traceaux_t *paux, jsparsstream_t *jpa,
			   jsparsvartab_t *vTab,
			   jsparscnode_t *decl,
			   jsparscnode_t *spec,
			   jctokfifo_t *queue,
			   int *pUseLabel, int mode)

/************************************************************************/
{
  jsparsexprtype_t type;
  const char *format;
  int i, indGeneric;

  type.JSPE_Flags = 0;
  type.JSPE_Declarator = decl;
  type.JSPE_Specifier = spec;
  type.JSPE_SubTable = NULL;
  type.JSPE_Queue = queue;

  indGeneric = -1; /* this points to the __POINTER__ type format if existant */
  for(i = 0 ; i < paux->NTypeFormats ; ++i) {
    int ret;

    if(paux->TypeFormats[i].FFlag & 1) {
      indGeneric = i;
      continue;
    }

    if((ret = jsparsisvartype(jpa, vTab, &type,
			      &(paux->TypeFormats[i].ExprType), 0)) == -1)
      EGOTO(error);

    if(ret != 0)
      break;
  }

  if(i == paux->NTypeFormats) {
    long flag;
    int ptrcnt;

    /*  No matching format found, try pointer or resort to internal
     *   mapping
     */
    if((paux->Mode & JSCPP_TRACE_INTTYPEMAPPING) || indGeneric != -1)
      /*  Determine pointer nature. If we might use the internal formats,
       *   we recognize "wchar_t" as special type and not as a pointer.
       */
      if((flag = jsparsgetvartype(jpa, vTab, decl, spec, queue,
				  &ptrcnt,
				  (paux->Mode & JSCPP_TRACE_INTTYPEMAPPING) ?
				  2 : 0)) == -1)
	EGOTO(error);

    /*  If we have no pointer format or no pointer, use internal mapping
     *   or give up
     */
    if(indGeneric == -1 || ptrcnt == 0) {
      if(paux->Mode & JSCPP_TRACE_INTTYPEMAPPING)
	return getTypeFormat(flag, ptrcnt, *pUseLabel);

      return "";
    }

    /*  if we reach here, we have a pointer (ptrcnt != 0) and a generic
     *   pointer format (indGeneric != -1)
     */
    i = indGeneric;
  }

  assert(0 <= i && i < paux->NTypeFormats);

  if(paux->TypeFormats[i].FFlag & 2)
    /*  Force no label  */
    return paux->TypeFormats[i].Format[*pUseLabel = 0];

  if(paux->TypeFormats[i].FFlag & 4)
    /*  Force label  */
    return paux->TypeFormats[i].Format[*pUseLabel = 1];

  if((format = paux->TypeFormats[i].Format[*pUseLabel]) != NULL)
    return format;

  /*  This string was empty, so try the other one  */
  *pUseLabel = !*pUseLabel;

  return paux->TypeFormats[i].Format[*pUseLabel];
error:
  return NULL;
}

/*JS*********************************************************************
*   Here we map a given type mask + a indirection counter to a single
*    letter our routine _functiontrace can understand
*************************************************************************/

static const char *getTypeFormat(long flag, int ptrcnt, int useLabel)

/************************************************************************/
{
static const char * const StdFormats[][2] = {
  /*  simple types  */
  { "",  "" },  /*  0: default  */
  { "c", "C" }, /*  1  */
  { "h", "H" }, /*  2  */
  { "i", "I" }, /*  3  */
  { "l", "L" }, /*  4  */
  { "f", "F" }, /*  5  */
  { "d", "D" }, /*  6  */

  /*  pointer types  */
  { "s", "S" }, /*  7  */
  { "w", "W" }, /*  8  */
  { "p", "P" }  /*  9  */
};
  int indx;

  switch(flag) {
  /*case  JPA_VARTYPE_VOID: type = 'v'; break;*/
  case JPA_VARTYPE_CHAR:   indx = (ptrcnt == 1) ? 7 : 1; break;
  case JPA_VARTYPE_SHORT:  indx = 2; break;
  case JPA_VARTYPE_WCHAR:  indx = (ptrcnt == 1) ? 8 : 3; break;
  case JPA_VARTYPE_ENUM: /*  enumeration  */
  case JPA_VARTYPE_INT:    indx = 3; break;
  case JPA_VARTYPE_LONG:   indx = 4; break;
  case JPA_VARTYPE_FLOAT:  indx = 5; break;
  case JPA_VARTYPE_DOUBLE: indx = 6; break;
  default:                 indx = 0; break; /*  struct, union etc.  */
  }

  if(indx < 7 && ptrcnt != 0)
    indx = 9; /*  generic pointer  */

  return StdFormats[indx][useLabel ? 1 : 0];
}
