/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Add tracing calls to functions.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            readPatterns(patchdesc_t *pdesc, int start, int end,
				      jctokfifo_t *queue);
Prototype int            readArgs(traceaux_t *paux, jsparsstream_t *jpa,
				  argdesc_t **pList, int *pNList,
				  int start, int end, jctokfifo_t *queue,
				  int mode);
Prototype jsparsvartab_t*readExprType(jsparsstream_t *jpa, jctokfifo_t *queue,
				      int start, int end,
				      jsparsexprtype_t *eType);
static int               addExpression(traceaux_t *paux, jsparsstream_t *jpa,
				       argdesc_t **pList, int *pNList,
				       const char *label, jctokfifo_t *queue,
				       int start, int end, int type);
static int               dogetsequence(void *user, jctokfifo_t *queue);

static int               addArgDescription(argdesc_t **ptrdesc, int *pn,
					   const char *label,
					   const char *expression,
					   jsqueuestream_t *jsqs);
Prototype jctokfifo_t   *copyToks(jsparsstream_t *jpa, jctokfifo_t *queue,
				  int *pStart, int *pEnd);

/*JS*********************************************************************
 *
 *************************************************************************/

int readPatterns(patchdesc_t *pdesc, int start, int end,
		 jctokfifo_t *queue)

/************************************************************************/
{
    int wantcomma, m;

    m = 0;

    wantcomma = 0;
    for( ; start <= end ; start += queue->TKFF_Step) {
	const char *pattern;
	int tokval;

	tokval = queue->TKFF_Token[start].TK_Value;
	if(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END)
	    continue;

	if(wantcomma) {
	    if(tokval != ',')
		/*  Remainder is not for us  */
		break;

	    wantcomma = 0;
	    continue;
	}

	if(tokval == '*') {
	    /*  A useful shortcut  */
	    pattern = "*";
	} else {
	    if(tokval != CTK_IDENT && tokval != CTK_STRING)
		EGOTO(invalid);

	    pattern = queue->TKFF_Token[start].TK_String;
	}

	if(pdesc->NPatterns >= m) {
	    m += 10;
	    if((pdesc->Patterns = (const char **)
		realloc((void *)pdesc->Patterns, m * sizeof(*(pdesc->Patterns)))) == NULL)
		EGOTO(error);
	}

	if((pdesc->Patterns[pdesc->NPatterns++] = strdup(pattern)) == NULL)
	    EGOTO(error);

	wantcomma = 1;
    }

    return start;
 error:
    return -1;
 invalid:
    return -2;
}

/*JS*********************************************************************
 *   mode bit 0: replace '@' by name of return variable.
 *        bit 1: do not expect surrounding brackets, read until the end.
 *************************************************************************/

int readArgs(traceaux_t *paux, jsparsstream_t *jpa,
	     argdesc_t **pList, int *pNList,
	     int start, int end,
	     jctokfifo_t *queue, int mode)

/************************************************************************/
{
    const char *label;
    int state, bracket, startExpr, endExpr;

    bracket = 0;

    if(mode & 2) {
	*pNList = 0;
	state = 15;
    } else
	state = 0;

    for( ; start <= end && state >= 0 ; start += queue->TKFF_Step) {
	int tokval;

	tokval = queue->TKFF_Token[start].TK_Value;

	if(CTK_BLANK_START <= tokval && tokval <= CTK_BLANK_END)
	    continue;

#ifdef XDEBUG
	printf("    %d %d l%d: %d %c\n", state, start, bracket, tokval,
	       tokval < 256 ? tokval : '?');
#endif

	switch(state) {
	case 0: /*  look for '('  */
	    if(tokval != '(') {
		state = -1;
		/*  make sure we do not leave out the token we ignore  */
		start -= queue->TKFF_Step;
		break;
	    }

	    *pNList = 0;

	    state = 10;
	    break;

	case 10: /*  very first expression  */
	    if(tokval == ')') { /*  we have the case "( )"  */
		state = -1;
		break;
	    }
	    /*FALLTHROUGH*/
	case 15: /*  memorize begin of expression  */
	    if(tokval == CTK_STRING) {
		if((label = strdup(queue->TKFF_Token[start].TK_String)) == NULL)
		    EGOTO(error);

		state = 16;
		break;
	    }

	    label = NULL;
	    /*FALLTHROUGH*/
	case 16:
	    startExpr = endExpr = start;
	    state = 18;
	    /*FALLTHROUGH*/
	case 18: /*  read expression until trailing comma  */
	    if(tokval == '(')
		++bracket;
	    else if(bracket == 0 && (tokval == ',' ||
				     (!(mode & 2) && tokval == ')'))) {
		int ret;

		if((ret = addExpression(paux, jpa, pList, pNList, label, queue,
					startExpr, endExpr, mode)) == -1)
		    EGOTO(error);

		if(ret == -2)
		    goto invalid;

		state = tokval == ')' ? -1 : 15;
	    } else if(tokval == ')')
		--bracket;

	    endExpr = start;
	    break;
	}
    }

    if(mode & 2) {
	int ret;

	if(bracket < 0)
	    goto invalid;

	if((ret = addExpression(paux, jpa, pList, pNList, label, queue,
				startExpr, endExpr, mode)) == -1)
	    EGOTO(error);

	if(ret == -2)
	    goto invalid;
    } else if(state > 0)
	goto invalid;

    return start;
 error:
 invalid:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsvartab_t *readExprType(jsparsstream_t *jpa, jctokfifo_t *queue,
			     int start, int end,
			     jsparsexprtype_t *eType)

/************************************************************************/
{
    jsqueuestream_t *jsqs;
    jsparscnode_t *node;
    tokenseq_t sequence;

    sequence.Curr = sequence.Start = start;
    sequence.End  = end;

    if((jsqs = jsqscreatequeuestream(queue, dogetsequence,
				     &sequence)) == NULL ||
       (node = jsparsgettype(jpa, jsqs, 0)) == NULL)
	EGOTO(error);

    if(jsparsnode2type(eType, node, queue) == NULL)
	EGOTO(error);

    return jsparsfindvartab(jpa, node);
 error:
    return NULL;
}

/*JS*********************************************************************
 *   mode bit 0: replace '@' by name of return variable
 *   mode bit 2: leave 'Expression' field empty.
 *   mode bit 3: Do not generate label field if empty.
 *************************************************************************/

static int addExpression(traceaux_t *paux, jsparsstream_t *jpa,
			 argdesc_t **pList, int *pNList,
			 const char *label, jctokfifo_t *queue,
			 int start, int end, int mode)

/************************************************************************/
{
    jsqueuestream_t *jsqs;
    tokenseq_t *psequence;
    const char *expression;

#ifdef CONFIG_JSPARSE_ARGLABELS
    /*  The label will still contain the '@', but not the expression  */
    if(label == NULL && !(mode & 8) &&
       (label = jctoktostring2(queue, start, end, NULL, NULL)) == NULL)
	EGOTO(error);
#endif

    if(mode & 1) { /*  return statement patches  */
	int i;

	for(i = start ; i <= end ; i += queue->TKFF_Step) {
	    if(queue->TKFF_Token[i].TK_Value == '@') {
		queue->TKFF_Token[i].TK_Value = CTK_IDENT;
		queue->TKFF_Token[i].TK_String = (char *)(paux->NameRetVariable);
	    }
	}
    }

    if(mode & 4)
	expression = NULL;
    else if((expression = jctoktostring2(queue, start, end,
					 NULL, NULL)) == NULL)
	EGOTO(error);

    if(queue != &(jpa->JSPA_GlobalQueue) &&
       (queue = copyToks(jpa, queue, &start, &end)) == NULL)
	EGOTO(error);

    /*  Create the queue-stream from which we will read the tokens later
     *   on
     */
    if((psequence = (tokenseq_t *)malloc(sizeof(*psequence))) == NULL)
	EGOTO0(error);

    psequence->Curr = psequence->Start = start;
    psequence->End  = end;

    if((jsqs = jsqscreatequeuestream(queue, dogetsequence, psequence)) == NULL)
	EGOTO(error);

    /*  Now add it to the global list  */
    return addArgDescription(pList, pNList, label, expression, jsqs);
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int dogetsequence(void *user, jctokfifo_t *queue)

/************************************************************************/
{
    int ret;

    if(((tokenseq_t *)user)->Curr > ((tokenseq_t *)user)->End)
	return -2; /*  EOF  */

    ret = ((tokenseq_t *)user)->Curr;
    ((tokenseq_t *)user)->Curr += queue->TKFF_Step;

    return ret;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int addArgDescription(argdesc_t **ptrdesc, int *pn, const char *label,
			     const char *expression,
			     jsqueuestream_t *jsqs)

/************************************************************************/
{
    if((*ptrdesc = (argdesc_t *)realloc(*ptrdesc, (*pn + 1) *
					sizeof(**ptrdesc))) == NULL)
	return -1;

    (*ptrdesc)[*pn].Label      = label;
    (*ptrdesc)[*pn].Expression = expression;
    (*ptrdesc)[*pn].QStream    = jsqs;

    ++(*pn);

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jctokfifo_t *copyToks(jsparsstream_t *jpa, jctokfifo_t *queue,
		      int *pStart, int *pEnd)

/************************************************************************/
{
    jctokfifo_t *dqueue;
    int dstart;

    /*  We have to copy it, otherwise we can't use it later
     *   to parse the tokens into an expression - except we
     *   go to the whole hassle of tokenizing it from the
     *   strings again.
     */
    dqueue = &(jpa->JSPA_GlobalQueue);

    if((dstart = jctokwriten(dqueue, &(queue->TKFF_Token[*pStart]),
			     *pEnd + queue->TKFF_Step - *pStart,
			     -1, JCTOK_MODE_INSERT)) == -1)
	EGOTO(error);

    /*  All indexes are relative to the new queue  */
    *pEnd += dstart - *pStart;
    *pStart = dstart;

    return dqueue;
 error:
    return NULL;
}
