/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Determine which global variables are used
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_GLOBVARS     "globvars"
/*  END-DEFINITIONS  */

/*  We have two additional flags where we mark usage and assignment  */
#define FLAG_FREADGLOB          JVDF_FUSER
#define FLAG_FWRITEGLOB        (JVDF_FUSER << 1)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doGlobVars(jsparsenviron_t *ppenv,
				    const char *names[],
				    int nnames, FILE *fpOut, int mode);
static int               checkFile(jsparsstream_t *jpa, FILE *fpOut);

static int               doProcGlobVarsMain(jsparsstream_t *jpa,
					    jsparscnode_t *node,
					    jctokfifo_t *queue,
					    FILE *fpOut);
static int               doVariable(jsparsstream_t *jpa, jsparscnode_t *node,
				    jctokfifo_t *queue, jsparsvartab_t *vTab,
				    jsparsvardef_t *vDef,
				    long flags, void *user);

/*********     GLOBAL VARIABLES                                 *********/

/*JS*********************************************************************
 *
 *************************************************************************/

int doGlobVars(jsparsenviron_t *ppenv, const char *names[], int nnames,
	       FILE *fpOut, int mode)

/************************************************************************/
{
    int i;

    for(i = 0 ; i < nnames ; ++i) {
	jsparsstream_t *jpa;
	jsppstream_t *jps;
	int ret, pmode;

	pmode = JPAM_FVARTABLES;

	/*  Get meaningful warning messages  */
	pmode |= JPAM_FGETFPOS;

	if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
	    EGOTO(error);

	jps = jsqsgetppstream(jpa->JSPA_Stream);
	jps->JSPP_Mode |= JPPM_PRESSPACE;

	ret = checkFile(jpa, fpOut);

	if(closeParserStream(jpa, ret != 0 ? 2 : 0) == -1)
	    EGOTO(error);

	if(ret == -1)
	    EGOTO(error);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int checkFile(jsparsstream_t *jpa, FILE *fpOut)

/************************************************************************/
{
    jctokfifo_t *queue;

    queue = jsparsgetqueue(jpa);
    for(;;) {
	jsparscnode_t *node;

	if(jsparsget(jpa, &node, 0) == -1)
	    EGOTO(error);

	if(node == NULL)
	    break;

	if(doProcGlobVarsMain(jpa, node, queue, fpOut) == -1)
	    EGOTO(error);

	if(jsparsflush(jpa, node, NULL, 0) == -1)
	    EGOTO(error);

	/*  Clear used flags, but not globally used flags  */
	jsparsclearvarflags(jpa->JSPA_VarTables[0],
			    JVDF_FREAD | JVDF_FWRITE);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int doProcGlobVarsMain(jsparsstream_t *jpa, jsparscnode_t *node,
			      jctokfifo_t *queue, FILE *fpOut)

/************************************************************************/
{
    if(jsparsprocessvars(jpa, node, queue, doVariable, NULL, 1) == -1)
	return -1;

    /*  Now look for usage of global variables  */
    if(node->JPACN_Type == JPACT_FUNCDEF) {
	jsparscnode_t *n;
	const char *name;
	int k, prName;

	n = node->JPACN_Sub; /*  DECLARATION  */
	n = n->JPACN_Sub; /*  DECLARATOR  */
	n = n->JPACN_Sub; /*  IDENTIFIER  */
	name = queue->TKFF_Token[n->JPACN_IndexPP].TK_String;

	prName = FALSE;
	for(k = 0 ; k < jpa->JSPA_NVarTables ; ++k) {
	    jsparsvartab_t *vTab;
	    int i;

	    vTab = jpa->JSPA_VarTables[k];

	    for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i) {
		jsparsvardef_t *vDef;

		for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next) {
		    long flags;

		    flags = vDef->JPAVD_Flags;

		    /*  Only for global variables  */
		    if(k == 0 && (flags & (JVDF_FREAD | JVDF_FWRITE))) {
			char *readchar, *writechar;

			readchar = writechar = NULL;
			switch(flags & JVDF_TMASK) {
			case JVDF_TSTRUCT: case JVDF_TUNION:
			case JVDF_TENUM:   case JVDF_TENUMIDENT:
			case JVDF_TTYPEDEF:
			    /*  I don't want to see these  */
			    break;
			case JVDF_TVARIABLE:
			    readchar = "<";
			    writechar = ">";
			    break;
			case JVDF_TFUNCTION:
			    /*  neither these  */
			    break;
			case JVDF_TLABEL:
			    break;
			}

			if(readchar) {
			    if(!prName) {
				prName = TRUE;
				fprintf(fpOut, "%s:", name);
			    }

			    fprintf(fpOut, " %s%s", (flags & JVDF_FWRITE) ? writechar :
				    readchar, vDef->JPAVD_Name);
			}
		    }
		}
	    }
	}

	if(prName)
	    putc('\n', fpOut);
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int doVariable(jsparsstream_t *jpa, jsparscnode_t *node,
		      jctokfifo_t *queue, jsparsvartab_t *vTab,
		      jsparsvardef_t *vDef,
		      long flags, void *user)

/************************************************************************/
{
    if(flags & JVDF_FREAD)
	flags |= FLAG_FREADGLOB;

    if(flags & JVDF_FWRITE)
	flags |= FLAG_FWRITEGLOB;

    vDef->JPAVD_Flags |= flags;

    return 0;
}
