/*JS*********************************************************************
 *
 *    Program : JSPARSE
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : ANSI-C parser.
 *    Part    : Routines to scan
 *
 *              Copyright (C) 2004-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/
/*#define NDEBUG*/
#include <assert.h>

#include <stddef.h> /*  for "offsetof" macro  */
#include <stdlib.h>
#include <stdio.h>

#include <wchar.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
/*  To avoid having to pass around a lot of stuff, we define our
 *   own auxiliary structure
 */
typedef struct {
    int Mode;
    jsparsstream_t *JPA;
    jsparsvartab_t *VarTable;
} expraux_t;

/*  I don't know if enum type is "int" or "long", so I let the
 *   compiler determine it.
 */
enum dummy_e {
    ONE, TWO, THREE
};

#define DETERMINE_ALIGNMENT(name, type)			\
    struct detalign_ ## name {				\
	char dummy;					\
	type variable;					\
    };							\
    static const int ALIGNMENT_ ## name =		\
	offsetof(struct detalign_ ## name, variable)

DETERMINE_ALIGNMENT(SHORT,    short);
DETERMINE_ALIGNMENT(INT,      int);
DETERMINE_ALIGNMENT(LONG,     long);
#ifdef CONFIG_HAS_LONGLONG
DETERMINE_ALIGNMENT(LONGLONG, long long);
#endif

DETERMINE_ALIGNMENT(FLOAT,    float);
DETERMINE_ALIGNMENT(DOUBLE,   double);

DETERMINE_ALIGNMENT(WCHAR,    wchar_t);

DETERMINE_ALIGNMENT(ENUM,     enum dummy_e);
DETERMINE_ALIGNMENT(VOIDP,    void *);

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype jsparsexprtype_t *jsparsgetexprtype(jsparsexprtype_t *pType,
					      jsparsstream_t *jpa,
					      jsparsvartab_t *vTab,
					      jsparscnode_t *expr,
					      jctokfifo_t *queue,
					      int mode);
Prototype int            jsparsfreeexprtype(jsparsstream_t *jpa,
					    jsparsexprtype_t *eType,
					    int flag);
Prototype unsigned long  jsparsgetvarsize(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					  const char *varName);
Prototype unsigned long  jsparsgetvarsize2(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					   jsparsvardef_t *vDef);
Prototype unsigned long  jsparsgetvarsize3(jsparsstream_t *jpa, jsparsvartab_t *vTab,
					   jctokfifo_t *queue, jsparscnode_t *decl,
					   jsparscnode_t *spec);

static int               getExpressionType(expraux_t *paux,
					   jsparscnode_t *expr,
					   jctokfifo_t *queue,
					   jsparsexprtype_t *pType);
static int               getUnaryExpressionType(expraux_t *paux,
						jsparscnode_t *expr,
						jctokfifo_t *queue,
						jsparsexprtype_t *pType);
static int               handleCast(expraux_t *paux, jsparscnode_t *expr,
				    jctokfifo_t *queue, jsparsexprtype_t *pType);
static int               genSubTable(expraux_t *paux, jsparsexprtype_t *pType,
				     jctokfifo_t *queue, jsparscnode_t *node);

static int               checkBadCast(expraux_t *paux, jctokfifo_t *queue,
				      jsparscnode_t *node,
				      jsparsexprtype_t *type1,
				      jsparsexprtype_t *type2);

static int               mergeTypes(expraux_t *paux, jctokfifo_t *queue,
				    jsparsexprtype_t *pType, jsparscnode_t *expr);
static unsigned long     getSizeOfType(expraux_t *paux, jctokfifo_t *queue,
				       jsparscnode_t *decl, jsparscnode_t *spec,
				       int *pAlign);
static unsigned long     getSizeOfType2(expraux_t *paux, jsparsexprtype_t *pType,
					int *pAlign);

static jsparscnode_t    *findStructUnionDef(expraux_t *paux, jsparsexprtype_t *pType);

/*********     GLOBAL / STATIC VARIABLES       			*********/
static const jsparscnode_t TokenChar   = { NULL, JPACT_CHAR,    NULL, -1, -1 };
static const jsparscnode_t TokenShort  = { NULL, JPACT_SHORT,   NULL, -1, -1 };
static const jsparscnode_t TokenInt    = { NULL, JPACT_INT,     NULL, -1, -1 };
static const jsparscnode_t TokenLong   = { NULL, JPACT_LONG,    NULL, -1, -1 };
static const jsparscnode_t TokenDouble = { NULL, JPACT_DOUBLE,  NULL, -1, -1 };

static const jsparscnode_t TokenStar   = { NULL, JPACT_POINTER, NULL, -1, -1 };

static const jsparsexprtype_t TypeChar   = {
    0, 0L, NULL, (jsparscnode_t *)&TokenChar,   NULL, NULL
};
static const jsparsexprtype_t TypeShort  = {
    0, 0L, NULL, (jsparscnode_t *)&TokenShort,  NULL, NULL
};
static const jsparsexprtype_t TypeInt    = {
    0, 0L, NULL, (jsparscnode_t *)&TokenInt,    NULL, NULL
};
static const jsparsexprtype_t TypeLong   = {
    0, 0L, NULL, (jsparscnode_t *)&TokenLong,   NULL, NULL
};
static const jsparsexprtype_t TypeDouble = {
    0, 0L, NULL, (jsparscnode_t *)&TokenDouble, NULL, NULL
};

static const jsparsexprtype_t TypeString = {
    0, 0L, (jsparscnode_t *)&TokenStar, (jsparscnode_t *)&TokenChar, NULL, NULL
};

static const jsparsexprtype_t TypeWString = {
    0, 0L, (jsparscnode_t *)&TokenStar, (jsparscnode_t *)&TokenInt, NULL, NULL
};

static const jsparsexprtype_t TypeUString = {
    0, 0L, (jsparscnode_t *)&TokenStar, (jsparscnode_t *)&TokenShort, NULL, NULL
};

/*JS*********************************************************************
 *   mode bit 0: If set, the expression must not contain anything with
 *               side effects (e. g. "++"), otherwise an error is returned.
 *        bit 1: resolve typedefs.
 *************************************************************************/

jsparsexprtype_t *jsparsgetexprtype(jsparsexprtype_t *pType,
				    jsparsstream_t *jpa,
				    jsparsvartab_t *vTab,
				    jsparscnode_t *expr,
				    jctokfifo_t *queue, int mode)

/************************************************************************/
{
    expraux_t aux;
    int alloc;

    aux.Mode = mode;
    aux.JPA = jpa;
    aux.VarTable = vTab;

    alloc = 0;
    if(pType == NULL) {
	if((pType = (jsparsexprtype_t *)malloc(sizeof(*pType))) == NULL)
	    EGOTO0(error);

	alloc = 1;
    }

    if(getExpressionType(&aux, expr, queue, pType) == -1)
	/*  The routine only fills the structure and also
	 *   frees the content in case of failure,
	 *   so we have to free the structure only
	 */
	EGOTO(error);

    if(mode & 2) {
	int ret;

	do {
	    if((ret = _jsparsres1type(jpa, vTab, pType)) == -1)
		EGOTO(error);
	} while(ret != 0); /*  repeat it until we reach a basic type  */
    }

    return pType;
 error:
    if(alloc) free(pType);
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsfreeexprtype(jsparsstream_t *jpa, jsparsexprtype_t *eType, int flag)

/************************************************************************/
{
    if(eType->JSPE_Flags & JPA_VARTYPE_FCOPYDECL) {
	jsparscnode_t *node;

	for(node = eType->JSPE_Declarator ; node ; )
	    node = jsparsfreecnode(node);
    } else if(eType->JSPE_Flags & JPA_VARTYPE_FALLOCDECL)
	/*  We had added an allocated '*' to the declarator  */
	jsparsfreecnode(eType->JSPE_Declarator);

    if(eType->JSPE_Flags & JPA_VARTYPE_FFREESUBTAB)
	/*  The subtable stems from a cast like "(struct { int i; })"  */
	jsparsfreevartab(eType->JSPE_SubTable);

    if(flag == 0) free(eType);

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

unsigned long jsparsgetvarsize(jsparsstream_t *jpa, jsparsvartab_t *vTab,
			       const char *varName)

/************************************************************************/
{
    jsparsvardef_t *vDef;

    if((vDef = jsparsgetvardef(jpa, vTab, varName, JVDF_TVARIABLE, 1)) == NULL)
	EGOTO(error);

    return jsparsgetvarsize2(jpa, vTab, vDef);
 error:
    return (unsigned long)-1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

unsigned long jsparsgetvarsize2(jsparsstream_t *jpa, jsparsvartab_t *vTab,
				jsparsvardef_t *vDef)

/************************************************************************/
{
    jsparsexprtype_t eType;
    expraux_t aux;
    unsigned long size;

    _jsparssetuptype(&eType, vDef);

    aux.Mode = 2;
    aux.JPA = jpa;
    aux.VarTable = vTab;

    size = getSizeOfType2(&aux, &eType, NULL);

    jsparsfreeexprtype(jpa, &eType, 1);

    return size;
}

/*JS*********************************************************************
 *
 *************************************************************************/

unsigned long jsparsgetvarsize3(jsparsstream_t *jpa, jsparsvartab_t *vTab,
				jctokfifo_t *queue, jsparscnode_t *decl,
				jsparscnode_t *spec)

/************************************************************************/
{
    expraux_t aux;

    aux.Mode = 2;
    aux.JPA = jpa;
    aux.VarTable = vTab;

    return getSizeOfType(&aux, queue, decl, spec, NULL);
}

/* ERROR-DEFINITIONS from jsparsgetexprtype label _ERR_JSPARSGETEXPRTYPE
   Illegal expression encountered
*/

/*JS*********************************************************************
 *
 *************************************************************************/

static int getExpressionType(expraux_t *paux, jsparscnode_t *expr,
			     jctokfifo_t *queue,
			     jsparsexprtype_t *pType)

/************************************************************************/
{
    switch(expr->JPACN_Type) {
	/*  Assignments like "int i = 1;" have a declarator instead of
	 *  a unary expression
	 */
    case JPACT_DECLARATOR:
    case JPACT_UNARYEXPR:
	return getUnaryExpressionType(paux, expr->JPACN_Sub, queue, pType);

    case JPACT_CONDEVAL:
    case JPACT_ADD: case JPACT_SUB: case JPACT_TIMES: case JPACT_DIV:
    case JPACT_LOR: case JPACT_LAND: case JPACT_EQUALS: case JPACT_NEQUALS:
    case JPACT_LOWER: case JPACT_GREATER: case JPACT_LOEQUAL: case JPACT_GREQUAL:
    case JPACT_OR: case JPACT_EXOR: case JPACT_AND:
    case JPACT_LSHIFT: case JPACT_RSHIFT:
    case JPACT_MODULO:
	return mergeTypes(paux, queue, pType, expr);

    case JPACT_SIZEOF: /*  this yields a long integer  */
	{
	    unsigned long size;
	    int alignment;

	    if((size = getSizeOfType(paux, queue, NULL, expr->JPACN_Sub, &alignment)) ==
	       (unsigned long)-1)
		EGOTO(error);

	    *pType = TypeLong;
	    pType->JSPE_Queue = queue;

	    if(size != 0) { /*  At leat deal graciously with variable things  */
		pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;
		pType->JSPE_LongConst = size;
	    }
	}
	return 0;

    case JPACT_ASS:
    case JPACT_ADDASS: case JPACT_SUBASS:
    case JPACT_MULASS: case JPACT_DIVASS: case JPACT_MODASS:
    case JPACT_LSHIFTASS: case JPACT_RSHIFTASS:
    case JPACT_ANDASS: case JPACT_ORASS:
    case JPACT_EXORASS:
	/*  the left hand side determines the type since the right hand side
	 *   is casted automatically by the compiler
	 */
	return getExpressionType(paux, expr->JPACN_Sub, queue, pType);

    default:
	/*  Anything else gives an error  */
	jsparsseterrcode(_ERR_JSPARSGETEXPRTYPE, 0, queue, expr);
	EGOTO0(error);
    }

 error:
    return -1;
}

/* ERROR-DEFINITIONS from jsparsgetunaryexprtype label _ERR_JSPARSGETUEXPRTYPE
   Illegal unary expression
   Cannot dereference non-pointer expression
   Cannot call non-function expression
   Level of dereferencing does not permit member expression
   Not a struct/union
   Expression contains illegal side effect
   Cannot take address twice
*/

/*JS*********************************************************************
 *
 *************************************************************************/

static int getUnaryExpressionType(expraux_t *paux, jsparscnode_t *expr,
				  jctokfifo_t *queue,
				  jsparsexprtype_t *pType)

/************************************************************************/
{
    /*  Treat first part of unary expression  */
    switch(expr->JPACN_Type) {
    case JPACT_IDENT:
	{
	    jsparsvardef_t *vDef;
	    const char *name;

	    name = queue->TKFF_Token[expr->JPACN_IndexPP].TK_String;
	    if((vDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
				       JVDF_TVARIABLE, 1)) != NULL) {
		_jsparssetuptype(pType, vDef);

	    } else if((vDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
					      JVDF_TFUNCTION, 1)) != NULL) {
		_jsparssetuptype(pType, vDef);

#if 0 /*  The parser already generated special tokens for these  */
	    } else if((vDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
					      JVDF_TENUMIDENT, 0)) != NULL) {
		*pType = TypeInt;
#endif
	    } else
		EGOTO(error);
	}
	break;

    case JPACT_ENUMCONST:
	{
	    jsparsvardef_t *vDef;
	    const char *name;

	    name = queue->TKFF_Token[expr->JPACN_IndexPP].TK_String;
	    if((vDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
				       JVDF_TENUMIDENT, 0)) == NULL)
		EGOTO(error);

	    *pType = TypeInt;

	    pType->JSPE_LongConst = vDef->JPAVD_LongValue;
	    pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;
	}

	break;

    case JPACT_INTCONST:
	*pType = TypeInt;
	if((pType->JSPE_LongConst =
	    strtoul(queue->TKFF_Token[expr->JPACN_IndexPP].TK_String,
		    NULL, 0)) == 0)
	    pType->JSPE_Flags |= JPA_VARTYPE_FISINTNULL;

	pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;

	break;

    case JPACT_FLOATCONST:
	*pType = TypeDouble;
	break;

    case JPACT_CHARCONST:
	*pType = TypeChar;
	break;

    case JPACT_WCHARCONST:
	*pType = TypeInt;
	break;

    case JPACT_UCHARCONST:
	*pType = TypeShort;
	break;

    case JPACT_STRING:
	*pType = TypeString;
	break;

    case JPACT_WSTRING:
	*pType = TypeWString;
	break;

    case JPACT_USTRING:
	*pType = TypeUString;
	break;

    default:
	if(getExpressionType(paux, expr, queue, pType) == -1)
	    EGOTO(error);

	/*
	  jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 0);
	  EGOTO0(error);
	*/

	break;
    }

    if(pType->JSPE_Queue == NULL)
	pType->JSPE_Queue = queue;

    /*  Now scan the remaining things  */
    for(expr = expr->JPACN_Next ; expr ; expr = expr->JPACN_Next) {
	jsparscnode_t *node;

	switch(expr->JPACN_Type) {
	case JPACT_POINTTO: case JPACT_POINTER: case JPACT_VECTOR:
	    while((node = pType->JSPE_Declarator) == NULL) {
		int ret;

		/*  We cannot dereference if there is no pointer. Well, but
		 *   there is still hope - maybe we have a typedef?
		 *   Just resolve *one* typedef and hope that things are
		 *   better then.
		 */
		if((ret = _jsparsres1type(paux->JPA, paux->VarTable, pType)) == -1)
		    EGOTO(error);

		if(ret == 0)
		    /*  That didn't help, so we give up  */
		    break;
	    }

	    /*  Skip type qualifiers  */
	    for( ; node != NULL ; node = node->JPACN_Next)
		if(node->JPACN_Type != JPACT_CONST && node->JPACN_Type != JPACT_VOLATILE)
		    break;

	    if(node == NULL || (node->JPACN_Type != JPACT_POINTER &&
				node->JPACN_Type != JPACT_VECTOR)) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 1, queue, node);
		EGOTO0(error);
	    }

	    /*  Remove the pointer thing from the declarator  */
	    if(pType->JSPE_Flags & (JPA_VARTYPE_FCOPYDECL |
				    JPA_VARTYPE_FALLOCDECL)) {
		pType->JSPE_Declarator = jsparsfreecnode(node);
		pType->JSPE_Flags &= ~JPA_VARTYPE_FALLOCDECL;
	    } else
		pType->JSPE_Declarator = node->JPACN_Next;

	    if(expr->JPACN_Type != JPACT_POINTTO)
		break;

	    /*FALLTHROUGH*/
	case JPACT_MEMBER:
	    while(pType->JSPE_SubTable == NULL) {
		int ret;

		if((ret = _jsparsres1type(paux->JPA, paux->VarTable, pType)) == -1)
		    EGOTO(error);

		if(ret == 0)
		    /*  That didn't help, so we give up  */
		    break;
	    }

	    if(pType->JSPE_SubTable == NULL) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 4, queue, expr);
		EGOTO0(error);
	    }

	    if(pType->JSPE_Declarator != NULL) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 3, queue, expr);
		EGOTO(error);
	    }

	    node = expr->JPACN_Sub;

	    {
		jsparsvardef_t *vDef;
		const char *name;

		name = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

		if((vDef = jsparsgetvardef(paux->JPA, pType->JSPE_SubTable, name,
					   JVDF_TVARIABLE, 1)) == NULL)
		    EGOTO(error);

		_jsparssetuptype(pType, vDef);
	    }
	    break;

	case JPACT_FUNCCALL:
	    if(paux->Mode & 1) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 5, queue, expr);
		EGOTO0(error);
	    }

	    node = pType->JSPE_Declarator;

	    if(node && node->JPACN_Type == JPACT_POINTER)
		/*  ANSI relaxed syntax to call a function  */
		node = node->JPACN_Next;

	    if(node == NULL || node->JPACN_Type != JPACT_FUNCCALL) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 2, queue, expr);
		EGOTO0(error);
	    }

	    pType->JSPE_Declarator = node->JPACN_Next;
	    break;

	case JPACT_ADDRESS:
	    if(pType->JSPE_Flags & JPA_VARTYPE_FALLOCDECL) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 6, queue, expr);
		EGOTO0(error);
	    }

	    if((pType->JSPE_Declarator =
		jsparsmkcnode(pType->JSPE_Declarator, JPACT_POINTER, NULL,
			      -1, -1, NULL)) == NULL)
		EGOTO(error);

	    if(!(pType->JSPE_Flags & JPA_VARTYPE_FCOPYDECL))
		pType->JSPE_Flags |= JPA_VARTYPE_FALLOCDECL;

	    break;

	case JPACT_CAST:
	    if(handleCast(paux, expr, queue, pType) == -1)
		EGOTO(error);

	    break;

	case JPACT_UPLUS: case JPACT_UMINUS: case JPACT_NOT: case JPACT_LNOT:
	    /*  Skip it  */
	    break;

	case JPACT_POSTINC: case JPACT_POSTDEC:
	case JPACT_PREINC:  case JPACT_PREDEC:
	    if(paux->Mode & 1) {
		jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 5, queue, expr);
		EGOTO0(error);
	    }

	    break;

	default:
	    jsparsseterrcode(_ERR_JSPARSGETUEXPRTYPE, 0, queue, expr);
	    EGOTO0(error);
	}
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int handleCast(expraux_t *paux, jsparscnode_t *expr, jctokfifo_t *queue,
		      jsparsexprtype_t *pType)

/************************************************************************/
{
    jsparscnode_t *node;
    jsparsexprtype_t newType;

    /*  A cast means we basically start all over again with
     *   a new expression type
     */
    newType.JSPE_Flags = 0;

    /*  We have to copy this since it might live longer than the original
     *   node. This happens e. g. when patching in "jsftrace": We generate
     *   a node temporarily and determine its type, then we delete the node.
     *   So the copy below makes sure that this type or sub table lives long
     *   enough.
     */
    if((node = jsparscopycnode(expr, queue, queue)) == NULL)
	EGOTO(error);

    node = node->JPACN_Sub;
    if(node->JPACN_Type == JPACT_DECLARATOR) {
	jsparscnode_t *sub;

	sub = node->JPACN_Sub;
	if(sub && sub->JPACN_Type == JPACT_IDENT)
	    sub = sub->JPACN_Next;
	newType.JSPE_Declarator = sub;

	node = node->JPACN_Next;
    } else
	newType.JSPE_Declarator = NULL;

    newType.JSPE_Specifier  = node;
    newType.JSPE_Queue      = queue;

    /*  Scan specifiers ("node") for struct/union */
    for( ; node ; node = node->JPACN_Next)
	if(node->JPACN_Type == JPACT_STRUCT || node->JPACN_Type == JPACT_UNION)
	    break;

    newType.JSPE_SubTable = NULL;
    if(node != NULL) { /*  Found a struct/union  */
	int type;

	type = node->JPACN_Type == JPACT_STRUCT ? JVDF_TSTRUCT : JVDF_TUNION;
	node = node->JPACN_Sub;
	if(node->JPACN_Type == JPACT_IDENT) {
	    jsparsvardef_t *subDef;
	    const char *name;

	    name = queue->TKFF_Token[node->JPACN_IndexPP].TK_String;

	    /*  Find the definition of this compound type  */
	    if((subDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
					 type, 0)) != NULL)
		newType.JSPE_SubTable = subDef->JPAVD_SubTable;

	    node = node->JPACN_Next;
	}

	if(newType.JSPE_SubTable == NULL) {
	    /*  It is actually not a good programming style to define a structure
	     *   only within a cast, but we support it anyway
	     */
	    if(node && node->JPACN_Type == JPACT_BLOCK &&
	       genSubTable(paux, &newType, queue, node) == -1)
		EGOTO(error);
	}
    }

    /*  Special mode: Find possibly illegal casts, e. g.
     *    char (*list)[10];
     *    char **ptr = (char **)list;
     */
    if((paux->JPA->JSPA_WarnFlags & JPAM_WARNBADCAST) &&
       checkBadCast(paux, queue, expr, pType, &newType) == -1)
	EGOTO(error);

    /*  Check if this is actually casting an integer to another integer and
     *   preserve integer constant flags. This is to allow something like
     *       enum {
     *           ENUMIDENT = (int)1;
     *       };
     */
    {
	long tflags;
	int ptrcnt;

	if((tflags = jsparsgetvartype(paux->JPA, paux->VarTable, newType.JSPE_Declarator,
				      newType.JSPE_Specifier, queue,
				      &ptrcnt, 0)) == -1)
	    EGOTO(error);

	if(!(tflags & ~(JPA_VARTYPE_CHAR | JPA_VARTYPE_SHORT |
			JPA_VARTYPE_INT | JPA_VARTYPE_LONG |
			JPA_VARTYPE_LONGLONG)) && ptrcnt == 0) {
	    unsigned long iflags;

	    if((iflags = (pType->JSPE_Flags & (JPA_VARTYPE_FLONGCONST |
					       JPA_VARTYPE_FARRAYSIZE))) != 0) {
		newType.JSPE_Flags |= iflags;
		newType.JSPE_LongConst = pType->JSPE_LongConst;
	    }
	}
    }

    /*  Free old stuff  */
    jsparsfreeexprtype(paux->JPA, pType, 1);

    *pType = newType;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int genSubTable(expraux_t *paux, jsparsexprtype_t *pType,
		       jctokfifo_t *queue, jsparscnode_t *node)

/************************************************************************/
{
    jsparsvartab_t *vTab;
    int ret;

    if((vTab = _jsparsgetvartab(paux->VarTable, JVDT_FSTRUCTUNION,
				node->JPACN_Start, node->JPACN_End)) == NULL)
	EGOTO(error);

    ret = _jsparsscanvars(paux->JPA, node, queue, vTab, 0);

    pType->JSPE_Flags |= JPA_VARTYPE_FFREESUBTAB;

    /*  We need this to be able to resolve definitions from within
     *   structures that reference other typedefs properly
     */
#if 0
    /*  This was needed so enumeration within structures were defined in
     *   the right (=global) scope. We can now get rid of it.
     */
    vTab->JPAVT_Next = NULL;
#endif

    pType->JSPE_SubTable = vTab;

    return ret;
 error:
    return -1;
}

#ifdef CONFIG_JSCPPLIB_DEBUG
/*JS*********************************************************************
 *
 *************************************************************************/

static int checkBadCast(expraux_t *paux, jctokfifo_t *queue,
			jsparscnode_t *node,
			jsparsexprtype_t *type1, jsparsexprtype_t *type2)

/************************************************************************/
{
    jsparsexprtype_t copy2;
    jsparscnode_t *decl1, *decl2;
    int ret;

    /*  We may do whatever we want to type1, but not type2  */
    if(jsparsmakeetype(&copy2, type2->JSPE_Declarator, type2->JSPE_Specifier,
		       queue) == NULL)
	EGOTO(error);
    type2 = &copy2;

    do {
	if((ret = _jsparsres1type(paux->JPA, paux->VarTable, type1)) == -1)
	    EGOTO(error);
    } while(ret != 0); /*  repeat it until we reach a basic type  */

    do {
	if((ret = _jsparsres1type(paux->JPA, paux->VarTable, type2)) == -1)
	    EGOTO(error);
    } while(ret != 0); /*  repeat it until we reach a basic type  */

    decl1 = type1->JSPE_Declarator;
    decl2 = type2->JSPE_Declarator;

    /*  Check first part of declarator: Has to be a pointer/array type  */
    if(decl1 != NULL && (decl1->JPACN_Type == JPACT_POINTER ||
			 decl1->JPACN_Type == JPACT_VECTOR) &&
       decl2 != NULL && (decl2->JPACN_Type == JPACT_POINTER ||
			 decl2->JPACN_Type == JPACT_VECTOR)) {
	/*  OK, check second part  */
	decl1 = decl1->JPACN_Next;
	decl2 = decl2->JPACN_Next;

	if(decl1 != NULL && decl2 != NULL &&
	   ((decl1->JPACN_Type == JPACT_POINTER &&
	     decl2->JPACN_Type == JPACT_VECTOR) ||
	    (decl2->JPACN_Type == JPACT_POINTER &&
	     decl1->JPACN_Type == JPACT_VECTOR))) {
	    jsparsprinterr(paux->JPA, queue, node,
			   "Possibly bad cast of pointer and array", 2);
	}
    }

    jsparsfreeexprtype(paux->JPA, &copy2, 1);

    return 0;
 error:
    return -1;
}
#endif

/* ERROR-DEFINITIONS from jsparsmergetypes label _ERR_JSPARSMERGETYPES
   Illegal expression of two pointers
   Illegal expression with one pointer
   Illegal pointer offset type
   Illegal operation with struct/union
   Illegal operands for integer operation
*/

/*JS*********************************************************************
 *
 *************************************************************************/

static int mergeTypes(expraux_t *paux, jctokfifo_t *queue,
		      jsparsexprtype_t *pType, jsparscnode_t *expr)

/************************************************************************/
{
    jsparsexprtype_t left, right;
    long result;
    int type, isConst, flag1, flag2, ptrCnt1, ptrCnt2;

    type = expr->JPACN_Type;

    expr = expr->JPACN_Sub;

    isConst = 0;
    if(type == JPACT_CONDEVAL) {
	/*  Check if condition is constant  */
	if(getExpressionType(paux, expr, queue, &left) == -1)
	    EGOTO(error2);

	if(left.JSPE_Flags & JPA_VARTYPE_FLONGCONST)
	    isConst = left.JSPE_LongConst ? 1 : 2;
	else
	    /*  Prevent coding below to fold "var ? 1 : 2" into a constant  */
	    isConst = -1;

	jsparsfreeexprtype(paux->JPA, &left, 1);

	/*  Skip condition  */
	expr = expr->JPACN_Next;
    }

    if(getExpressionType(paux, expr, queue, &left) == -1)
	EGOTO(error2);

    expr = expr->JPACN_Next;
    if(getExpressionType(paux, expr, queue, &right) == -1) {
	jsparsfreeexprtype(paux->JPA, &left, 1);
	EGOTO(error2);
    }

    if((flag1 = jsparsgetvartype2(paux->JPA, paux->VarTable, &left,
				  &ptrCnt1, 0)) == -1 ||
       (flag2 = jsparsgetvartype2(paux->JPA, paux->VarTable, &right,
				  &ptrCnt2, 0)) == -1)
	EGOTO(error);

    /*  Map enum to int  */
    if(flag1 & JPA_VARTYPE_ENUM)
	flag1 = (flag1 & ~JPA_VARTYPE_ENUM) | JPA_VARTYPE_INT;
    if(flag2 & JPA_VARTYPE_ENUM)
	flag2 = (flag2 & ~JPA_VARTYPE_ENUM) | JPA_VARTYPE_INT;

    /*  Recognize short form "x" for "x != 0" in logical and and or  */
    if(type == JPACT_LAND || type == JPACT_LOR) {
	*pType = TypeInt;

	/*  Check for constant logical expression, i. e. "... && 0"
	 *   as well as "... || 1"
	 */
	if(((left.JSPE_Flags & JPA_VARTYPE_FLONGCONST) &&
	    (type == JPACT_LAND ? left.JSPE_LongConst == 0 :
	     left.JSPE_LongConst != 0)) ||
	   ((right.JSPE_Flags & JPA_VARTYPE_FLONGCONST) &&
	    (type == JPACT_LAND ? right.JSPE_LongConst == 0 :
	     right.JSPE_LongConst != 0))) {
	    pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;
	    pType->JSPE_LongConst = type == JPACT_LAND ? 0 : 1;
	}

	jsparsfreeexprtype(paux->JPA, &left, 1);
	jsparsfreeexprtype(paux->JPA, &right, 1);

	return 0;
    }

    /*  Special treatment for "0" being a pointer: We have one side
     *   being a pointer, the other not. If one side is "0", we can
     *   treat it as a pointer to make the syntax correct.
     */
    if(((ptrCnt1 > 0) ^ (ptrCnt2 > 0)) &&
       (type == JPACT_CONDEVAL || type == JPACT_EQUALS || type == JPACT_NEQUALS ||
	type == JPACT_LOWER || type == JPACT_GREATER ||
	type == JPACT_LOEQUAL || type == JPACT_GREQUAL)) {
	if(left.JSPE_Flags & JPA_VARTYPE_FISINTNULL)
	    ++ptrCnt1;

	if(right.JSPE_Flags & JPA_VARTYPE_FISINTNULL)
	    ++ptrCnt2;
    }

    if(ptrCnt1 > 0 && ptrCnt2 > 0) {
	/*  We could check here if the two pointer targets are compatible,
	 *   but we leave it to a decent compiler :)
	 */
	if(type == JPACT_SUB || type == JPACT_EQUALS || type == JPACT_NEQUALS ||
	   type == JPACT_LOWER || type == JPACT_GREATER ||
	   type == JPACT_LOEQUAL || type == JPACT_GREQUAL) {
	    jsparsfreeexprtype(paux->JPA, &left, 1);
	    jsparsfreeexprtype(paux->JPA, &right, 1);
	    *pType = TypeInt;

	} else if(type == JPACT_CONDEVAL) {
	    /*  We need the if because of the special treatment for "0"
	     *   from above
	     */
	    *pType = left;
	    jsparsfreeexprtype(paux->JPA, &right, 1);

	} else {
	    jsparsseterrcode(_ERR_JSPARSMERGETYPES, 0, queue, expr);
	    EGOTO(error);
	}

	return 0;
    }

    if(ptrCnt1 > 0 || ptrCnt2 > 0) {
	int flag;

	/*  Definitely only *one* side is a pointer, check
	 *   for valid syntax.
	 */
	if(type != JPACT_ADD && type != JPACT_SUB) {
	    jsparsseterrcode(_ERR_JSPARSMERGETYPES, 1, queue, expr);
	    EGOTO(error);
	}

	if(ptrCnt1 > 0) {
	    *pType = left;
	    jsparsfreeexprtype(paux->JPA, &right, 1);
	    flag = flag2;
	} else {
	    *pType = right;
	    jsparsfreeexprtype(paux->JPA, &left, 1);
	    flag = flag1;
	}

	if(flag & ~(JPA_VARTYPE_CHAR | JPA_VARTYPE_SHORT |
		    JPA_VARTYPE_INT | JPA_VARTYPE_LONG |
		    JPA_VARTYPE_LONGLONG)) {
	    jsparsseterrcode(_ERR_JSPARSMERGETYPES, 2, queue, expr);
	    jsparsfreeexprtype(paux->JPA, pType, 1);
	    EGOTO(error2);
	}

	return 0;
    }

    /*assert(ptrCn1 == && ptrCnt2 == 0);*/

    /*  Check if this is a constant  */
    if(isConst >= 0) {
	if((left.JSPE_Flags & JPA_VARTYPE_FLONGCONST) &&
	   (right.JSPE_Flags & JPA_VARTYPE_FLONGCONST)) {
	    long lvalue, rvalue;

	    lvalue = left.JSPE_LongConst;
	    rvalue = right.JSPE_LongConst;
	    switch(type) {
	    case JPACT_CONDEVAL: result = isConst == 1 ? lvalue : rvalue;   break;
	    case JPACT_ADD:      result = lvalue +  rvalue; isConst = TRUE; break;
	    case JPACT_SUB:      result = lvalue -  rvalue; isConst = TRUE; break;
	    case JPACT_TIMES:    result = lvalue *  rvalue; isConst = TRUE; break;
	    case JPACT_DIV:      result = lvalue /  rvalue; isConst = TRUE; break;

	    case JPACT_LOR:      result = lvalue || rvalue; isConst = TRUE; break;
	    case JPACT_LAND:     result = lvalue && rvalue; isConst = TRUE; break;
	    case JPACT_EQUALS:   result = lvalue == rvalue; isConst = TRUE; break;
	    case JPACT_NEQUALS:  result = lvalue != rvalue; isConst = TRUE; break;
	    case JPACT_LOWER:    result = lvalue <  rvalue; isConst = TRUE; break;
	    case JPACT_GREATER:  result = lvalue >  rvalue; isConst = TRUE; break;
	    case JPACT_LOEQUAL:  result = lvalue <= rvalue; isConst = TRUE; break;
	    case JPACT_GREQUAL:  result = lvalue >= rvalue; isConst = TRUE; break;
	    case JPACT_OR:       result = lvalue |  rvalue; isConst = TRUE; break;
	    case JPACT_EXOR:     result = lvalue ^  rvalue; isConst = TRUE; break;
	    case JPACT_AND:      result = lvalue &  rvalue; isConst = TRUE; break;
	    case JPACT_LSHIFT:   result = lvalue << rvalue; isConst = TRUE; break;
	    case JPACT_RSHIFT:   result = lvalue >> rvalue; isConst = TRUE; break;
	    case JPACT_MODULO:   result = lvalue %  rvalue; isConst = TRUE; break;
#ifdef CONFIG_JSCPPLIB_DEBUG
	    default:
		EGOTO(error);
#endif
	    }
	} else
	    /*  Got "? : " operator, but one of the expression wasn't constant  */
	    isConst = 0;
    }

    if((flag1 & (JPA_VARTYPE_STRUCT | JPA_VARTYPE_UNION)) ||
       (flag2 & (JPA_VARTYPE_STRUCT | JPA_VARTYPE_UNION))) {
	jsparsseterrcode(_ERR_JSPARSMERGETYPES, 3, queue, expr);
	EGOTO(error);
    }

    /*  The "bigger" one wins  */
    if(flag1 >= flag2) {
	*pType = left;
	jsparsfreeexprtype(paux->JPA, &right, 1);
    } else {
	*pType = right;
	jsparsfreeexprtype(paux->JPA, &left, 1);
	flag1 = flag2;
    }

    /*  These are pure integer operators. They must be applied to integer
     *   operands only and yield an integer.
     */
    if((
	/*  FIXME: Why did I put here, it is rubbish??
	    type == JPACT_EQUALS  || type == JPACT_NEQUALS ||
	    type == JPACT_LOWER   || type == JPACT_GREATER ||
	    type == JPACT_LOEQUAL || type == JPACT_GREQUAL ||
	*/
	type == JPACT_OR      || type == JPACT_EXOR    || type == JPACT_AND ||
	type == JPACT_LSHIFT  || type == JPACT_RSHIFT  ||
	type == JPACT_MODULO) &&
       flag1 > JPA_VARTYPE_LONGLONG) {
	jsparsseterrcode(_ERR_JSPARSMERGETYPES, 4, queue, expr);
	EGOTO(error);
    }

    if(isConst > 0) {
	pType->JSPE_Flags |= JPA_VARTYPE_FLONGCONST;
	pType->JSPE_LongConst = result;

    } else {
	/*  Make sure constness doesn't survive illegally (from left or right
	 *   above
	 */
	pType->JSPE_Flags &= ~JPA_VARTYPE_FLONGCONST;
    }

    return 0;

 error:
    jsparsfreeexprtype(paux->JPA, &left, 1);
    jsparsfreeexprtype(paux->JPA, &right, 1);
 error2:
    return -1;
}

/* ERROR-DEFINITIONS from jsparsgetsizeoftype label _ERR_JSPARSGETSIZEOFTYPE
   Empty dimension in array
   Array dimension is not constant
   Cannot determine size of type
*/

/*JS*********************************************************************
 *
 *************************************************************************/

static unsigned long getSizeOfType(expraux_t *paux, jctokfifo_t *queue,
				   jsparscnode_t *decl, jsparscnode_t *spec,
				   int *pAlign)

/************************************************************************/
{
    jsparsexprtype_t eType;
    unsigned long size;

#if 0
    fprintf(stderr,"\nSIZEOF-----------------\n");
    jsparsprctree(stderr, queue, decl, 0, 0);
    fprintf(stderr,"\n-----------------\n");
    jsparsprctree(stderr, queue, spec, 0, 0);
    fprintf(stderr,"\n-----------------\n");
#endif

    if(spec->JPACN_Type == JPACT_UNARYEXPR) {
	/*  "sizeof(EXPRESSION)"  */
	if(getExpressionType(paux, spec, queue, &eType) == -1)
	    EGOTO(error);

    } else {
	if(spec->JPACN_Type == JPACT_DECLARATOR) {
	    decl = spec;
	    spec = spec->JPACN_Next;
	}

	/*  "sizeof(TYPE)"  */
	if(decl && decl->JPACN_Type == JPACT_DECLARATOR) {
	    decl = decl->JPACN_Sub;
	    if(decl && decl->JPACN_Type == JPACT_IDENT)
		decl = decl->JPACN_Next;
	}

	if(jsparsmakeetype(&eType, decl, spec, queue) == NULL)
	    EGOTO(error);
    }

    size = getSizeOfType2(paux, &eType, pAlign);

    jsparsfreeexprtype(paux->JPA, &eType, 1);

    return size;
 error:
    return (unsigned long)-1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static unsigned long getSizeOfType2(expraux_t *paux, jsparsexprtype_t *pType,
				    int *pAlign)

/************************************************************************/
{
    jsparscnode_t *node;
    unsigned long size, flags;
    int ret;

    do {
	if((ret = _jsparsres1type(paux->JPA, paux->VarTable, pType)) == -1)
	    EGOTO(error);
    } while(ret != 0); /*  repeat it until we reach a basic type  */

    /*  Look at declarators first - if it is a pointer, we don't have to
     *   look at the specifiers
     */
    size = 1;
    flags = pType->JSPE_Flags;
    for(node = pType->JSPE_Declarator ; node ; node = node->JPACN_Next) {
        unsigned int nodeType;

        nodeType = node->JPACN_Type;
        if(nodeType == JPACT_VECTOR) {
	    jsparsexprtype_t aType;

	    if(flags & JPA_VARTYPE_FARRAYSIZE) {
		/*  Only use it for first dimension  */
		flags &= ~JPA_VARTYPE_FARRAYSIZE;

		size *= pType->JSPE_LongConst;
		continue;
	    }

	    if(node->JPACN_Sub == NULL) {
		/*  Empty dimension  */
		jsparsseterrcode(_ERR_JSPARSGETSIZEOFTYPE, 0, pType->JSPE_Queue, node);
		EGOTO(error);
	    }

	    /*  Get size of array  */
	    if(getExpressionType(paux, node->JPACN_Sub, pType->JSPE_Queue, &aType) == -1)
		EGOTO(error);

	    if(!(aType.JSPE_Flags & (JPA_VARTYPE_FARRAYSIZE | JPA_VARTYPE_FLONGCONST))) {
#if 0 /*  FIXME: How shall we deal with extensions to the C language like
       *   dynamic array sizing?
       */
		jsparsfreeexprtype(paux->JPA, &eType, 1);
		return 0;
#else
		jsparsseterrcode(_ERR_JSPARSGETSIZEOFTYPE, 1, pType->JSPE_Queue, node);
		EGOTO(error);
#endif
	    }

	    size *= aType.JSPE_LongConst;
	    jsparsfreeexprtype(paux->JPA, &aType, 1);

        } else if(nodeType == JPACT_POINTER) {
	    /*  FIXME: I require that all pointers are the same  */
	    CASSERT(sizeof(void *) == sizeof(char *) &&
		    sizeof(void *) == sizeof(int *) &&
		    sizeof(void *) == sizeof(long *) &&
		    sizeof(void *) == sizeof(double *) &&
		    sizeof(void *) == sizeof(void **) &&
		    sizeof(void *) == sizeof(jsparscnode_t *));

	    size *= sizeof(void *);
	    if(pAlign) *pAlign = ALIGNMENT_VOIDP;
	    break;

        } else if(nodeType != JPACT_CONST && nodeType != JPACT_VOLATILE &&
                  nodeType != JPACT_SMALLINT) {
            jsparsprctree(stderr, pType->JSPE_Queue, pType->JSPE_Declarator, 0, 0);
            jsparsprctree(stderr, pType->JSPE_Queue, pType->JSPE_Specifier, 0, 0);
	    EGOTO(error);
	}
    }

    if(node == NULL) {
	unsigned long basesize;
	int varFlag;

	/*  This is not a pointer, so we have to look at the specifiers as well  */
	if((varFlag = jsparsgetvartype(paux->JPA, paux->VarTable, NULL, pType->JSPE_Specifier,
				       pType->JSPE_Queue, NULL, 0)) == -1)
	    EGOTO(error);

	switch(varFlag) {
        case JPA_VARTYPE_VOID:
            basesize = 0;
            if(pAlign) *pAlign = 1;
            break;

	case JPA_VARTYPE_CHAR:
	    basesize = sizeof(char);
	    if(pAlign) *pAlign = 1;
	    break;

	case JPA_VARTYPE_SHORT:
	    basesize = sizeof(short);
	    if(pAlign) *pAlign = ALIGNMENT_SHORT;
	    break;

	case JPA_VARTYPE_INT:
	    basesize = sizeof(int);
	    if(pAlign) *pAlign = ALIGNMENT_INT;
	    break;

	case JPA_VARTYPE_LONGLONG:
#ifdef CONFIG_HAS_LONGLONG
	    basesize = sizeof(long long);
	    if(pAlign) *pAlign = ALIGNMENT_LONGLONG;
	    break;
#endif

	case JPA_VARTYPE_LONG:
	    basesize = sizeof(long);
	    if(pAlign) *pAlign = ALIGNMENT_LONG;
	    break;

	case JPA_VARTYPE_FLOAT:
	    basesize = sizeof(float);
	    if(pAlign) *pAlign = ALIGNMENT_FLOAT;
	    break;

	case JPA_VARTYPE_DOUBLE:
	    basesize = sizeof(double);
	    if(pAlign) *pAlign = ALIGNMENT_DOUBLE;
	    break;

	case JPA_VARTYPE_WCHAR:
	    basesize = sizeof(wchar_t);
	    if(pAlign) *pAlign = ALIGNMENT_WCHAR;
	    break;

	case JPA_VARTYPE_STRUCT:
	case JPA_VARTYPE_UNION:
	    {
		expraux_t straux;

		straux = *paux;

		if((node = findStructUnionDef(&straux, pType)) == NULL)
		    EGOTO(error);

		straux.VarTable = pType->JSPE_SubTable;
		assert(straux.VarTable != NULL);

		basesize = (varFlag == JPA_VARTYPE_STRUCT) ? 0 : 1;
		if(pAlign) *pAlign = 1;

		for( ; node ; node = node->JPACN_Next) {
		    jsparscnode_t *subDecl, *subSpec;
		    int isList;

		    assert(node->JPACN_Type == JPACT_DECLARATION);
		    subDecl = node->JPACN_Sub;
		    subSpec = subDecl->JPACN_Next;
		    if(subDecl->JPACN_Type == JPACT_LIST) {
			isList = TRUE;
			subDecl = subDecl->JPACN_Sub;

		    } else
			isList = FALSE;

		    for( ; subDecl ; subDecl = subDecl->JPACN_Next) {
			unsigned long subsize;
			int subalign;

			if((subsize = getSizeOfType(&straux, pType->JSPE_Queue,
						    subDecl, subSpec,
						    &subalign)) == (unsigned long)-1)
			    EGOTO(error);

#ifndef __linux__ /*  we have sometimes padding arrays of size 0  */
			/*  This cannot happen unless we have "dynamic" structure sizing  */
			assert(subsize > 0);
#endif

			if(varFlag == JPA_VARTYPE_STRUCT) {
			    int off;

			    /*  First align field  */
			    if((off = basesize % subalign) != 0)
				basesize += subalign - off;

			    basesize += subsize;
			} else { /*  UNIONS  */
			    if(subsize > basesize)
				basesize = subsize;
			}

			/*  Maintain alignment for structure/union as a whole  */
			if(pAlign && subalign > *pAlign)
			    *pAlign = subalign;

			if(!isList)
			    break;
		    }
		}
	    }

	    break;

	case JPA_VARTYPE_ENUM:
	    basesize = sizeof(enum dummy_e);
	    if(pAlign) *pAlign = ALIGNMENT_ENUM;
	    break;

#ifdef CONFIG_JSCPPLIB_DEBUG
	default:
	    jsparsseterrcode(_ERR_JSPARSGETSIZEOFTYPE, 2, pType->JSPE_Queue, pType->JSPE_Declarator);
	    EGOTO(error);
#endif
	}

	size *= basesize;
    }

#ifdef TRACETYPE_JSCPP
    jstrace(TRACETYPE_JSCPP, 6,
	    "    Determine size %ld alignment %d.\n", size, pAlign ? *pAlign : -1);
#endif

    return size;
 error:
    return (unsigned long)-1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jsparscnode_t *findStructUnionDef(expraux_t *paux,
					 jsparsexprtype_t *pType)

/************************************************************************/
{
    for(;;) {
	jsparsvardef_t *vDef;
	jsparscnode_t *node;
	const char *name;
	int type;

	node = pType->JSPE_Specifier;

	for( ; node ; node = node->JPACN_Next)
	    if(node->JPACN_Type == JPACT_STRUCT || node->JPACN_Type == JPACT_UNION)
		break;
	assert(node != NULL);

	type = node->JPACN_Type;

	name = NULL;
	for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next) {
	    if(node->JPACN_Type == JPACT_BLOCK) {
		/*  An anonymous structure, generate a sub table on the fly  */
		if(pType->JSPE_SubTable == NULL &&
		   genSubTable(paux, pType, pType->JSPE_Queue, node) == -1)
		    EGOTO(error);

		return node->JPACN_Sub;
	    }

	    if(node->JPACN_Type == JPACT_IDENT)
		name = pType->JSPE_Queue->TKFF_Token[node->JPACN_IndexPP].TK_String;
	}

	/*  if the structure doesn't have the block with the definitions itself,
	 *   it must at least have a name
	 */
	assert(name != NULL);

	if((vDef = jsparsgetvardef(paux->JPA, paux->VarTable, name,
				   type == JPACT_STRUCT ? JVDF_TSTRUCT :
				   JVDF_TUNION, 0)) == NULL)
	    EGOTO(error);

	jsparsfreeexprtype(paux->JPA, pType, 1);
	_jsparssetuptype(pType, vDef);
    }

 error:
    return NULL;
}
