/*JS*********************************************************************
*
*    Program : JSPARSE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : ANSI-C parser.
*    Part    : Routines to analyze the execution flow
*
*              Copyright (C) 2004-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/
#include <stdio.h>

/*#define NDEBUG*/
#include <assert.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

#include <jsbitutil.h>

/*********     DEFINES						*********/
#ifndef JPA_FLOW_SINGLESTMT
/*  START-DEFINITIONS 2 */
typedef struct {
  long JPAFLN_Flags;

  jsparscnode_t *JPAFLN_Node;

  unsigned int JPAFLN_Counter; /*  counter of flow nodes pointing to this node  */
  unsigned int JPAFLN_Next1, JPAFLN_Next2;

} jsparsflownode_t;

/* ***  For JPAFLN_Flags field  *** */
/*  Single statement of a if-else or while/do-while/for loop which
 *   has to be surrounded by curly brackets if to be extended to
 *   more than one statement. This helps us avoid generating too
 *   many curly brackets when patching code.
 */
#define JPA_FLOW_SINGLESTMT      (1 << 0)

/*  Statement is at the start of a block, so we can insert
 *   declarations.
 */
#define JPA_FLOW_BLOCKSTART      (1 << 1)

/*  Not a statement, only an expression (i. e. condition of an if
 *   statement or a loop).
 */
#define JPA_FLOW_EXPRESSION      (1 << 4)

/*  Only a pseudo statement to link other statements, should be
 *   skipped. This is used for statements that only connect the
 *   various switch labels.
 */
#define JPA_FLOW_PSEUDOSTMT      (1 << 5)

/*  The statement has subnodes that are recorded somewhere else as
 *   statements. Only the first sub node must be looked at.
 */
#define JPA_FLOW_FIRSTSUBNODE    (1 << 6)

/*  The statement was reached after a branch in the flow
 */
#define JPA_FLOW_BRANCHED        (1 << 7)

typedef struct {
  unsigned int JPAFL_Start;

  unsigned int JPAFL_NStatements;

  jsparsflownode_t *JPAFL_Statements;

  unsigned long **JPAFL_Bitmaps;
  unsigned int JPAFL_MBitmaps;

} jsparsflow_t;

/*  The index in the JPAFLN_NextX fields to indicate that the surrounding
 *   routine is left
 */
#define JPA_FLOW_INDEX_RETURN    -2

#define jsparsflowclear(jpa)     jsparsflowclear2((jpa)->JSPA_StmtFlow)
#define jsparsflowmark(jpa, indx, opFunction, user, mode) \
	jsparsflowmark2((jpa)->JSPA_StmtFlow, indx, opFunction, user, mode)
/*  END-DEFINITIONS  */
#endif

typedef struct {
  jsparsstream_t *JPA;
  jctokfifo_t *Queue;
  jsparsflow_t *JFlow;
  unsigned int MStatements;

  unsigned int GotoList;

} flowaux_t;

/*  Add the node with the given index 'ind' to the list. Returns -1 on error
 *   (no list)
 */
#define addNode(jFlow, pList, ind)					\
  (pList == NULL ? (unsigned int)-1 :					\
   (((jFlow)->JPAFL_Statements[ind].JPAFLN_Next1 = *(pList),		\
     *(pList) = (ind))))

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
IPrototype jsparsflow_t  *_jsparsscanflow(jsparsstream_t *jpa,
					  jsparscnode_t *node,
					  jctokfifo_t *queue);
IPrototype void           _jsparsfreeflow(jsparsflow_t *jFlow);

Prototype void            jsparsflowclear2(jsparsflow_t *jFlow);
Prototype int             jsparsflowmark2(jsparsflow_t *jFlow, int indx,
					  int (*opFunction)(jsparscnode_t *,
							    int, void *),
					  void *user, int mode);
static int                doflowmark(jsparsflow_t *jFlow, int indx, int level,
				     int (*opFunction)(jsparscnode_t *, int, void *),
				     void *user, int mode);

static int                scanNode(flowaux_t *faux, jsparscnode_t *node,
				   unsigned int *iNext,
				   unsigned int *(iLoop[2]),
				   unsigned int *iSwitch,
				   long flags);
static int                getStmt(flowaux_t *faux, jsparscnode_t *node,
				  unsigned int *list, long flags);
static void               connectList(jsparsflow_t *jFlow, int list,
				      int nr);

static int                checkLoopCondition(flowaux_t *faux,
					     jsparscnode_t *node);

static int                connectGotos(flowaux_t *faux, jsparscnode_t *top);

Prototype int             jsparsfindflow(jsparsstream_t *jpa,
					 jsparscnode_t *node);
static int                findNodeIndex(jsparsflownode_t *stmt,
					unsigned int n,
					unsigned int indx);

#ifdef TRACETYPE_JSCPP
static void               showStmtFlow(FILE *fp, jsparsflow_t *jFlow,
				       jctokfifo_t *queue, int mode);
#endif

/* ERROR-DEFINITIONS from jsparsscanflow label _ERR_JSPARSSCANFLOW
   'break' statement outside loop or switch statement
   'continue' statement outside loop
   'case' or 'default' label outside switch statement
   Label used but not defined
*/

/*JS*********************************************************************
*
*************************************************************************/

jsparsflow_t *_jsparsscanflow(jsparsstream_t *jpa, jsparscnode_t *node,
			      jctokfifo_t *queue)

/************************************************************************/
{
  jsparscnode_t *body;
  jsparsflow_t *jFlow;
  flowaux_t faux;
  unsigned int next, i;

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & (3 << (14 + CONFIG_JSCPPLIB_DEBUG)))
    jpa->JSPA_Mode |= JPAM_FFLOWBLOCKEND;
#endif

  /*  Unset flag  */
  jpa->JSPA_Mode &= ~JPAM_FFALLTOEND;

  assert(node->JPACN_Type == JPACT_FUNCDEF);
  body = node->JPACN_Sub;
  assert(body->JPACN_Type == JPACT_DECLARATION);

  body = body->JPACN_Next;
  if(body->JPACN_Type == JPACT_LIST ||
     body->JPACN_Type == JPACT_DECLARATION) {
    /*  K&R-type function definition  */
    body = body->JPACN_Next;
  }

  assert(body->JPACN_Type == JPACT_COMPSTMT);

  if((jFlow = (jsparsflow_t *)malloc(sizeof(*jFlow))) == NULL)
    return NULL;

  jFlow->JPAFL_Statements = NULL;
  jFlow->JPAFL_NStatements = 0;

  faux.JPA   = jpa;
  faux.Queue = queue;
  faux.JFlow = jFlow;
  faux.MStatements = 0;
  faux.GotoList = -1;

  /*  Returns the start node  */
  next = -1;
  if((jFlow->JPAFL_Start = scanNode(&faux, body, &next,
				    NULL, NULL, 0)) == (unsigned int)-1)
    EGOTO(error);

  jFlow->JPAFL_Statements[jFlow->JPAFL_Start].JPAFLN_Counter++;

  if(next != (unsigned int)-1) {
    unsigned int list, count;

    /*  At the end of the function body, there is an implicit 'return',
     *   so the nodes that fall to the end implicitly do 'return'. If
     *   the function is non-void, this is actually an error which we
     *   detect here.
     */

    /*  Check if these statements that have an implicit return are
     *   reached at all
     */
    count = 0;
    for(list = next ; list != (unsigned int)-1 ;
	list = jFlow->JPAFL_Statements[list].JPAFLN_Next1)
      count += jFlow->JPAFL_Statements[list].JPAFLN_Counter;

    if(count > 0) {
      jpa->JSPA_Mode |= JPAM_FFALLTOEND;

      if((jpa->JSPA_WarnFlags & JPAM_WARNFLOW) && jpa->JSPA_VarTables != NULL) {
	jsparscnode_t *sub;
	jsparsvardef_t *vDef;
	const char *funcname;
	long type;
	int ptrcnt;

	sub = node->JPACN_Sub;
	assert(sub->JPACN_Type == JPACT_DECLARATION);
	sub = sub->JPACN_Sub;
	assert(sub->JPACN_Type == JPACT_DECLARATOR);
	sub = sub->JPACN_Sub;

	funcname = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;

	if((vDef = jsparsgetvardef(jpa, jpa->JSPA_VarTables[0], funcname,
				   JVDF_TFUNCTION, 0)) == NULL ||
	   (type = jsparsgetvartype3(jpa, jpa->JSPA_VarTables[0],
				     vDef, &ptrcnt, 1)) == -1)
	  EGOTO(error);

	if(!(type == JPA_VARTYPE_VOID && ptrcnt == 0)) {
	  jsparsprinterr(jpa, queue, node,
			 "Control reaches end of non-void function", 0);
	}
      }
    }

    connectList(jFlow, next, JPA_FLOW_INDEX_RETURN);
  }

  if(faux.GotoList != (unsigned int)-1 &&
     connectGotos(&faux, body) == -1)
    EGOTO(error);

  jFlow->JPAFL_MBitmaps = 1;
  if((jFlow->JPAFL_Bitmaps = (unsigned long **)
      malloc(jFlow->JPAFL_MBitmaps * sizeof(*(jFlow->JPAFL_Bitmaps)))) == NULL)
    EGOTO(error);

  {
    int mapSize;

    mapSize = BITS_NRBYTES(jFlow->JPAFL_Bitmaps[0], jFlow->JPAFL_NStatements);
    for(i = 0 ; i < jFlow->JPAFL_MBitmaps ; ++i)
      if((jFlow->JPAFL_Bitmaps[i] = (unsigned long *)malloc(mapSize)) == NULL)
	EGOTO(error);
  }

#ifdef CONFIG_JSCPPLIB_DEBUG
  if(Debug & (3 << (12 + CONFIG_JSCPPLIB_DEBUG)))
    showStmtFlow(stdout, jFlow, queue,
		 (Debug >> (12 + CONFIG_JSCPPLIB_DEBUG)) & 3);
#endif

#ifdef TRACETYPE_JSCPP
  {
    FILE *tfp;

    if((tfp = jstracegetfp(TRACETYPE_JSCPP, 6)) != NULL)
      showStmtFlow(tfp, jFlow, queue, 0);
  }
#endif

  if(jpa->JSPA_WarnFlags & JPAM_WARNFLOW) {
    /* ***  Look for unreachable code  *** */
    for(i = 0 ; i < jFlow->JPAFL_NStatements ; ++i) {
      if(jFlow->JPAFL_Statements[i].JPAFLN_Counter > 0 ||
	 (jFlow->JPAFL_Statements[i].JPAFLN_Flags & JPA_FLOW_PSEUDOSTMT))
	continue;

      jsparsprinterr(jpa, queue, jFlow->JPAFL_Statements[i].JPAFLN_Node,
		     "Unreachable code", 0);
    }
  }

  if(jpa->JSPA_WarnFlags & JPAM_WARNFLOW) {
    const jsparsvartab_t *vTab;

    /*  Look for unused labels  */
    if((vTab = jsparsfindvartab(jpa, node)) != NULL) {
      for(i = 0 ; i < JPA_VARIABLE_TABSIZE ; ++i) {
	const jsparsvardef_t *vDef;

	for(vDef = vTab->JPAVT_Vars[i] ; vDef ; vDef = vDef->JPAVD_Next) {
	  long flags;

	  flags = vDef->JPAVD_Flags;
	  if((flags & JVDF_TMASK) == JVDF_TLABEL && !(flags & JVDF_FREAD)) {
	    jsparsprinterr(jpa, queue, vDef->JPAVD_Declarator,
			   "Unused label '%s'", 0, vDef->JPAVD_Name);
	  }
	}
      }
    }
  }

#ifndef NDEBUG /*  only makes sense when assertions are not disabled  */
  /*   Check that the assertion stmt[i].JPACN_Start < stmt[j].JPACN_Start
   *    if i < j is fullfilled. This is used in routine findNodeIndex
   */
  for(i = 1 ; i < jFlow->JPAFL_NStatements ; ++i) {
    const jsparscnode_t *prev, *curr;

    prev = jFlow->JPAFL_Statements[i - 1].JPAFLN_Node;
    curr = jFlow->JPAFL_Statements[i].JPAFLN_Node;

    /*printf("%d: %d %d\n",i,prev->JPACN_Start,curr->JPACN_Start);*/
    assert(prev->JPACN_Start < curr->JPACN_Start);
  }
#endif

  return jFlow;
error:
  free(jFlow->JPAFL_Statements);
  free(jFlow);
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

void _jsparsfreeflow(jsparsflow_t *jFlow)

/************************************************************************/
{
  if(jFlow->JPAFL_Statements)
    free(jFlow->JPAFL_Statements);

  if(jFlow->JPAFL_Bitmaps) {
    unsigned int i;

    for(i = 0 ; i < jFlow->JPAFL_MBitmaps ; ++i)
      free(jFlow->JPAFL_Bitmaps[i]);

    free(jFlow->JPAFL_Bitmaps);
  }

  free(jFlow);
}

/*JS*********************************************************************
*
*************************************************************************/

void jsparsflowclear2(jsparsflow_t *jFlow)

/************************************************************************/
{
  memset(jFlow->JPAFL_Bitmaps[0], 0,
	 BITS_NRBYTES(jFlow->JPAFL_Bitmaps[0], jFlow->JPAFL_NStatements));
}

/*JS*********************************************************************
*   If mode bit 1 is set, we use separate bitmaps when we traverse each
*    branch.
*************************************************************************/

int jsparsflowmark2(jsparsflow_t *jFlow, int indx,
		    int (*opFunction)(jsparscnode_t *, int, void *),
		    void *user, int mode)

/************************************************************************/
{
  return doflowmark(jFlow, indx, 0, opFunction, user, mode);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doflowmark(jsparsflow_t *jFlow, int indx, int level,
		      int (*opFunction)(jsparscnode_t *, int, void *),
		      void *user, int mode)

/************************************************************************/
{
  do {
    unsigned int next, ret;

    if(BIT_ISSET(jFlow->JPAFL_Bitmaps[level], indx))
      return 0; /*  We already saw this  */

    BIT_SET(jFlow->JPAFL_Bitmaps[level], indx);

    if(opFunction != NULL &&
       (ret = opFunction(jFlow->JPAFL_Statements[indx].JPAFLN_Node,
			 level, user)) != 0)
      return ret;

    next = jFlow->JPAFL_Statements[indx].JPAFLN_Next2;
    if(next != (unsigned int)-1) {
      unsigned int nextLevel;

      nextLevel = level;
      if(mode & 1) {
	unsigned int mapSize;

	mapSize = BITS_NRBYTES(jFlow->JPAFL_Bitmaps[0], jFlow->JPAFL_NStatements);

	if(++nextLevel >= jFlow->JPAFL_MBitmaps) {
	  unsigned int i;

	  i = jFlow->JPAFL_MBitmaps;
	  REALLOC_RESIZE(jFlow->JPAFL_MBitmaps, 10);

	  for( ;  i < jFlow->JPAFL_MBitmaps ; ++i)
	    if((jFlow->JPAFL_Bitmaps[i] = (unsigned long *)malloc(mapSize)) == NULL)
	      EGOTO(error);
	}

	memcpy(jFlow->JPAFL_Bitmaps[nextLevel], jFlow->JPAFL_Bitmaps[level], mapSize);
      }

      /*  Recursion on JPAFLN_Next2 field, iteration on JPAFLN_Next1  */
      if((ret = doflowmark(jFlow, next, nextLevel, opFunction, user, mode)) != 0)
	return ret;
    }

    indx = jFlow->JPAFL_Statements[indx].JPAFLN_Next1;

  } while(indx >= 0);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*   Creates a statement flow node for the given node and returns its
*    index or -1 on error. iNext, iLoop[0], iLoop[1], iSwitch are lists
*    where the resulting node might be added.
*
*   iNext, iLoop[0] ... iLoop[2], and iSwitch point to the head of linked
*    lists of nodes. The nodes are linked through the JPAFLN_Next1 field.
*    -1 indicates the end of the list. For each node in the lists, the
*    JPAFLN_Next1 field is supposed to be set by the caller (via routine
*    'connectList').
*
*    The nodes in the various lists want the following to be put in their
*    JPAFLN_Next1 field:
*
*   iNext:    Index of the node that comes after the given node in
*             execution sequence.
*   iLoop[0]: Index of the next node for loop or switch abortion (from
*             "break" statements).
*             loop[0] should be set to NULL when calling this routine
*             if we are not within a loop or switch statement.
*   iLoop[1]: Index of the next node for loop continuation (from
*             "continue" statements). loop[1] should be NULL when
*             calling this function if we are not within a loop.
*   iSwitch:  Index of the next node in sequence of 'case'/'default'
*             statement in a switch. There are pseudo statements for
*             each case label that are linked in a list. Execution
*             flow proceeds along this list from label to label and
*             branches into the various case blocks.
*             iSwitch should be set to NULL if we are not within
*             a switch statement.
*             If the switch statement has a 'default' clause, it can not
*             just fall-through if none of the 'case' labels is met.
*             Statement flow will always enter the switch body. We
*             memorize having seen a 'default' statement with
*             iSwitch[1] == TRUE.
*************************************************************************/

static int scanNode(flowaux_t *faux, jsparscnode_t *node,
		    unsigned int *iNext, unsigned int *(iLoop[2]),
		    unsigned int *iSwitch, long flags)

/************************************************************************/
{
  jsparsflow_t *jFlow;
  jsparscnode_t *sub;
  unsigned int ind;

  /*  We cannot cache jFlow->JPAFL_Statements, because every time
   *   we call getStmt or ourselves recursively, it might change due
   *   to reallocation
   */
  jFlow = faux->JFlow;

  switch(node->JPACN_Type) {
  case JPACT_COMPSTMT:
    {
      unsigned int flag, lNext[2], blockInd, startFlag;

      /*  Everything that comes now is properly wrapped  */
      flags &= ~JPA_FLOW_SINGLESTMT;

      if(!(faux->JPA->JSPA_Mode & JPAM_FFLOWBLOCKEND))
	blockInd = -1;
	/*   We have to allocate it beforehand, so the assertion
	 *    stmt[i].JPACN_Start < stmt[j].JPACN_Start if i < j is
	 *    fullfilled. This is used in routine findNodeIndex
	 */
      else if((blockInd = getStmt(faux, node, iNext,
				  JPA_FLOW_PSEUDOSTMT | flags)) == (unsigned int)-1)
	EGOTO(error);

      startFlag = JPA_FLOW_BLOCKSTART;
      ind = lNext[0] = lNext[1] = -1;
      for(flag = 0, sub = node->JPACN_Sub ; sub ;
	  flag = !flag, sub = sub->JPACN_Next,
	    flags &= ~JPA_FLOW_BRANCHED) {
	unsigned int newInd, *currNext;

	lNext[flag] = -1;

	if(sub->JPACN_Next || blockInd != (unsigned int)-1)
	  currNext = &lNext[flag];
	else
	  currNext = iNext;

	if((newInd = scanNode(faux, sub, currNext, iLoop, iSwitch,
			      flags)) == (unsigned int)-1)
	  EGOTO(error);

	/*  first loop for return value  */
	if(ind == (unsigned int)-1)
	  ind = newInd;

	if(startFlag != 0 && sub->JPACN_Type != JPACT_DECLARATION &&
	   sub->JPACN_Type != JPACT_TYPEDEF) {
	  /*  This is the first statement after the declarations. It
	   *   is the only one that may be prefixed with declarations.
	   */
	  if(sub->JPACN_Type != JPACT_LABELEDSTMT) {
#if 0
	    /*  FIXME: We better wrap "for()" etc as pseudo statements */
	    if(!(jFlow->JPAFL_Statements[newInd].JPAFLN_Flags &
		 JPA_FLOW_EXPRESSION))
#endif
	    jFlow->JPAFL_Statements[newInd].JPAFLN_Flags |= startFlag;
	  }

	  startFlag = 0;
	}

	/*  Anything to connect from previous iteration?  */
	if(lNext[!flag] != (unsigned int)-1)
	  /*  second or following iteration, so connect previous ones  */
	  connectList(jFlow, lNext[!flag], newInd);
      }

      if(ind == (unsigned int)-1) {
	/*  Empty compound statement, so return the statement itself   */
	ind = blockInd != (unsigned int)-1 ? blockInd :
	  getStmt(faux, node, iNext, flags);

      } else if(blockInd != (unsigned int)-1) {
	if(lNext[!flag] != (unsigned int)-1)
	  connectList(jFlow, lNext[!flag], blockInd);
      }
    }
    break;

  case JPACT_IFELSE:
    {
      jsparscnode_t *cond;
      unsigned int condType, ind1;

      cond = node->JPACN_Sub;
      sub = cond->JPACN_Next; /*  the true part  */

      /*  The condition  */
      if((condType = checkLoopCondition(faux, cond)) == (unsigned int)-1)
	EGOTO(error);

      if((ind = getStmt(faux, (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			node : cond,
			/*  Without 'else' part, the statement can continue
			 *   with what follows after it - unless it is always
			 *   true!
			 */
			sub->JPACN_Next == NULL && condType != 2 ?
			iNext : NULL,
			(faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			JPA_FLOW_FIRSTSUBNODE | flags :
			JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	EGOTO(error);

      /*  The true part  */
      if((ind1 = scanNode(faux, sub, iNext, iLoop, iSwitch,
			  JPA_FLOW_BRANCHED | JPA_FLOW_SINGLESTMT |
			  flags)) == (unsigned int)-1)
	EGOTO(error);

      /* ************************************************** */

      /*  Connect the condition with the true part - unless the condition is
       *   always false
       */
      if(condType != 1) {
	jFlow->JPAFL_Statements[ind].JPAFLN_Next2 = ind1;
	jFlow->JPAFL_Statements[ind1].JPAFLN_Counter++;
      }

      /*  The (optional) else part  */
      if((sub = sub->JPACN_Next) != NULL) {
	if((ind1 = scanNode(faux, sub, iNext, iLoop, iSwitch,
			    JPA_FLOW_BRANCHED | JPA_FLOW_SINGLESTMT |
			    flags)) == (unsigned int)-1)
	  EGOTO(error);

	/*  Connect the condition with the else part  */
	if(condType != 2) {
	  jFlow->JPAFL_Statements[ind].JPAFLN_Next1 = ind1;
	  jFlow->JPAFL_Statements[ind1].JPAFLN_Counter++;
	}
      }
    }
    break;

  case JPACT_WHILE: case JPACT_DOWHILE:
    {
      unsigned int condType, indBody, next, *(loop[2]);

      next = -1;

      /*  Any break statements are looking for the statements after the loop,
       *   the continue statements are looking for the loop start
       */
      loop[0] = iNext;
      loop[1] = &next;

      sub = node->JPACN_Sub;

      if(node->JPACN_Type == JPACT_WHILE) {
	/* ***  WHILE LOOP  *** */
	/*  The loop condition  */
	if((condType = checkLoopCondition(faux, sub)) == (unsigned int)-1)
	  EGOTO(error);

	/*  If the condition is not always true, the next statement
	 *   after the loop can be what follows the loop
	 */
	if((ind = getStmt(faux, (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			  node : sub, condType != 2 ? iNext : NULL,
			  (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			  JPA_FLOW_FIRSTSUBNODE | flags :
			  JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	  EGOTO(error);

	/*  The loop body  */
	sub = sub->JPACN_Next;
	if((indBody = scanNode(faux, sub, &next, loop, iSwitch,
			       JPA_FLOW_BRANCHED | JPA_FLOW_SINGLESTMT |
			       flags)) == (unsigned int)-1)
	  EGOTO(error);
      } else {
	/* ***  DO-WHILE LOOP  *** */
	/*  The loop body  */
	if((indBody = scanNode(faux, sub, &next, loop, iSwitch,
			       JPA_FLOW_BRANCHED | JPA_FLOW_SINGLESTMT |
			       flags)) == (unsigned int)-1)
	  EGOTO(error);

	/*  The loop condition  */
	sub = sub->JPACN_Next;
	if((condType = checkLoopCondition(faux, sub)) == (unsigned int)-1)
	  EGOTO(error);

	/*  If the condition is not always true, the next statement
	 *   after the loop can be what follows the loop
	 */
	if((ind = getStmt(faux, (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			  node : sub, condType != 2 ? iNext : NULL,
			  (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			  JPA_FLOW_FIRSTSUBNODE | flags :
			  JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	  EGOTO(error);
      }

      /*  After the body comes the loop condition  */
      connectList(jFlow, next, ind);

      /*  Connect the condition with the body if condition permits  */
      if(condType != 1) {
	jFlow->JPAFL_Statements[ind].JPAFLN_Next2 = indBody;
	jFlow->JPAFL_Statements[indBody].JPAFLN_Counter++;

      } else if(node->JPACN_Type == JPACT_DOWHILE) {
	/*  Recognize "do { ... } while(0)"  */
	jFlow->JPAFL_Statements[ind].JPAFLN_Flags |= JPA_FLOW_PSEUDOSTMT;
      }

      if(node->JPACN_Type == JPACT_DOWHILE)
	/*  The loop entry is the body  */
	ind = indBody;
    }
    break;

  case JPACT_FOR:
    {
      jsparscnode_t *pre, *cond, *cont;
      unsigned int condType, indCond, indBody, indReinit, next, *(loop[2]);

      /*  Any break statements are looking for the statements after the loop,
       *   the continue statements are looking for the loop start
       */
      loop[0] = iNext;
      loop[1] = &next;

      pre  = node->JPACN_Sub;  /*  loop initialization  */
      cond = pre->JPACN_Next;  /*  loop condition  */
      cont = cond->JPACN_Next; /*  loop reinitialization  */
      sub = cont->JPACN_Next;  /*  body  */

      /*  The initialization  */
      if((ind = getStmt(faux, (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			node : pre, NULL,
			(faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			JPA_FLOW_FIRSTSUBNODE | flags :
			JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	EGOTO(error);

      /*  The condition. If loop can be left, connect the condition with
       *   what follows after the loop
       */
      if((condType = checkLoopCondition(faux, cond)) == (unsigned int)-1)
	EGOTO(error);

      if((indCond = getStmt(faux, cond, condType == 2 ? NULL : iNext,
			    JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	EGOTO(error);

      /*  The continuation  */
      if((indReinit = getStmt(faux, cont, NULL,
			      JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	EGOTO(error);

      /*  The body  */
      next = -1;
      if((indBody = scanNode(faux, sub, &next, loop, iSwitch,
			     JPA_FLOW_BRANCHED | JPA_FLOW_SINGLESTMT |
			     flags)) == (unsigned int)-1)
	EGOTO(error);

      /* ************************************* */

      /*  Connect initialization to the condition  */
      jFlow->JPAFL_Statements[ind].JPAFLN_Next1 = indCond;
      jFlow->JPAFL_Statements[indCond].JPAFLN_Counter++;

      /*  Connect condition with body if it can be true  */
      if(condType != 1) {
	jFlow->JPAFL_Statements[indCond].JPAFLN_Next2 = indBody;
	jFlow->JPAFL_Statements[indBody].JPAFLN_Counter++;
      }

      connectList(jFlow, next, indReinit);

      /*  connect re-initialization with condition  */
      jFlow->JPAFL_Statements[indReinit].JPAFLN_Next2 = indCond;
      jFlow->JPAFL_Statements[indCond].JPAFLN_Counter++;
    }
    break;

  case JPACT_BREAK:
    if(iLoop == NULL || iLoop[0] == NULL) {
      jsparsseterrcode(_ERR_JSPARSSCANFLOW, 0, faux->Queue, node);
      EGOTO(error);
    }

    ind = getStmt(faux, node, iLoop[0], flags);
    break;

  case JPACT_CONTINUE:
    if(iLoop == NULL || iLoop[1] == NULL) {
      jsparsseterrcode(_ERR_JSPARSSCANFLOW, 1, faux->Queue, node);
      EGOTO(error);
    }

    ind = getStmt(faux, node, iLoop[1], flags);
    break;

    /* *************************************************************** */
  case JPACT_SWITCH:
    {
      unsigned int ind2, indSwitch[2], *(loop[2]);

      sub = node->JPACN_Sub;

      /*  The switch expression itself  */
      if((ind = getStmt(faux, (faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			node : sub, NULL,
			(faux->JPA->JSPA_Mode & JPAM_FFLOWREALSTMT) ?
			JPA_FLOW_FIRSTSUBNODE | flags :
			JPA_FLOW_EXPRESSION | flags)) == (unsigned int)-1)
	EGOTO(error);

      sub = sub->JPACN_Next;

      /*  We temporarily overwrite the destination for 'break' statements,
       *   but leave those for 'continue' statements
       */
      loop[0] = iNext;
      loop[1] = iLoop ? iLoop[1] : NULL;

      /*
       *   We generate a linked list of statements nodes chained through
       *   the JPAFLN_Next1 field, starting from the switch expression.
       *   In the chain, the JPAFLN_Next2 field points to the actual
       *   nodes that follow the various labels.
       */
      indSwitch[0] = ind;
      indSwitch[1] = 0; /*  set to one if a 'default' is seen  */
      if((ind2 = scanNode(faux, sub, iNext, loop,
			  indSwitch, flags)) == (unsigned int)-1)
	EGOTO(error);

      /*  The only way to enter the switch body is through the 'case' or
       *   'default' label, so we don't connect anything to the body.
       */

      if(!indSwitch[1]) {
	/*  This switch has no 'default' clause, so if none of the case labels
	 *   applies, the whole thing will be skipped, so we connect the
	 *   switch expression also with what follows the switch.
	 *   The link to the various case labels now goes through the Next2
	 *   field.
	 */
	ind2 = jFlow->JPAFL_Statements[ind].JPAFLN_Next1;

	jFlow->JPAFL_Statements[ind].JPAFLN_Next2 = ind2;
	jFlow->JPAFL_Statements[ind2].JPAFLN_Counter++;

	addNode(jFlow, iNext, ind);
      }
    }
    break;

  case JPACT_LABELEDSTMT:
    {
      jsparscnode_t *label, *stmt;
      unsigned int hasSwitchLabel, ind2;

      label = node->JPACN_Sub;  /*  the labels  */
      stmt  = label->JPACN_Next; /*  the actual statement  */

      /*  Check for any switch relevant stuff in the labels  */
      hasSwitchLabel = FALSE;
      for(sub = label->JPACN_Type == JPACT_LIST ? label->JPACN_Sub : label ;
	  sub ; sub = sub->JPACN_Next) {
	if(sub->JPACN_Type == JPACT_CASE)
	  hasSwitchLabel = TRUE;
	else if(sub->JPACN_Type == JPACT_DEFAULT) {
	  hasSwitchLabel = TRUE;
	  if(iSwitch)
	    /*  Memorize that this switch has a 'default' clause  */
	    iSwitch[1] = TRUE;
	}

	if(label->JPACN_Type != JPACT_LIST)
	  break;
      }

      if(hasSwitchLabel) {
	if(iSwitch == NULL) {
	  jsparsseterrcode(_ERR_JSPARSSCANFLOW, 2, faux->Queue, node);
	  EGOTO(error);
	}

	/*  Generate a statement for connection only  */
	if((ind2 = getStmt(faux, node, NULL,
			   JPA_FLOW_PSEUDOSTMT | flags)) == (unsigned int)-1)
	  EGOTO(error);

	jFlow->JPAFL_Statements[*iSwitch].JPAFLN_Next1 = ind2;
	jFlow->JPAFL_Statements[ind2].JPAFLN_Counter++;
	jFlow->JPAFL_Statements[ind2].JPAFLN_Flags |= JPA_FLOW_BRANCHED;
	*iSwitch = ind2;
      }

      /*  We return a node that points to the statement without labels  */
      if(stmt != NULL) {
	if((ind = scanNode(faux, stmt, iNext, iLoop,
			   iSwitch, flags)) == (unsigned int)-1)
	  EGOTO(error);
      } else {
	/*  This can happen since we allow constructs like
	 *   "{ ... trailing_label: }" which is not ANSI
	 *   conformant but appears anyway.
	 */
	if((ind = getStmt(faux, label, iNext,
			  JPA_FLOW_PSEUDOSTMT | flags)) == (unsigned int)-1)
	  EGOTO(error);
      }

      if(hasSwitchLabel) {
	jFlow->JPAFL_Statements[ind2].JPAFLN_Next2 = ind;
	jFlow->JPAFL_Statements[ind].JPAFLN_Counter++;
      }
    }
    break;

  case JPACT_RETURN:
    /*  Nothing follows after a return statement...  */
    ind = getStmt(faux, node, NULL, flags);

    /*  ... it leaves the function  */
    jFlow->JPAFL_Statements[ind].JPAFLN_Next1 = JPA_FLOW_INDEX_RETURN;
    break;

  case JPACT_GOTO:
    /*  Add it to the list of goto's  */
    ind = getStmt(faux, node, &(faux->GotoList), flags);
    break;

  default: /*  Simple statements, expressions etc.  */
    ind = getStmt(faux, node, iNext, flags);
    break;
  }

  return ind;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getStmt(flowaux_t *faux, jsparscnode_t *node,
		   unsigned int *list,
		   long flags)

/************************************************************************/
{
  jsparsflow_t *jFlow;
  jsparsflownode_t *flNode;
  int ind;

  jFlow = faux->JFlow;

  if(jFlow->JPAFL_NStatements >= faux->MStatements) {
    REALLOC_RESIZE(faux->MStatements, 100);
    if((jFlow->JPAFL_Statements = (jsparsflownode_t *)
	realloc(jFlow->JPAFL_Statements, faux->MStatements *
		sizeof(*(jFlow->JPAFL_Statements)))) == NULL)
      return -1;
  }

  ind = jFlow->JPAFL_NStatements++;
  flNode = &(jFlow->JPAFL_Statements[ind]);

  flNode->JPAFLN_Flags   = flags;

  flNode->JPAFLN_Counter = 0;
  flNode->JPAFLN_Node    = node;

  if(list) {
    flNode->JPAFLN_Next1 = *list;
    *list = ind;

  } else
    flNode->JPAFLN_Next1 = -1;

  flNode->JPAFLN_Next2 = -1;

  return ind;
}

/*JS*********************************************************************
*   list points to the head of a linked list of nodes. The nodes are
*    connected through their JPAFLN_Next1 field.
*    This routine connects all nodes to field nr.
*************************************************************************/

static void connectList(jsparsflow_t *jFlow, int list, int nr)

/************************************************************************/
{
  jsparsflownode_t *stmt;
  int count;

  stmt = jFlow->JPAFL_Statements;
  for(count = 0 ; list != -1 ; ++count) {
    int next;

    next = stmt[list].JPAFLN_Next1;
    stmt[list].JPAFLN_Next1 = nr;

    list = next;
  }

  if(nr >= 0)
    stmt[nr].JPAFLN_Counter += count;
}

/*JS*********************************************************************
*  Check for constant conditions of loops or if-statements. This is to
*   optimize the statement flow in cases like "while(1)" etc.
*   0: condition may be both.
*   1: condition is always false.
*   2: condition is always true.
*************************************************************************/

static int checkLoopCondition(flowaux_t *faux, jsparscnode_t *node)

/************************************************************************/
{
  int ret;

  ret = 0; /*  I don't know  */
  if(node->JPACN_Type == JPACT_UNARYEXPR) {
    jsparsvartab_t *vTab;
    jsparsexprtype_t eType;

    if(node->JPACN_Sub == NULL) /*  "for(;;)"  */
      return 2;

    /*  Check if the expression is an integer constant  */
    if((vTab = jsparsfindvartab(faux->JPA, node)) == NULL)
      EGOTO(error);

    if(jsparsgetexprtype(&eType, faux->JPA, vTab, node,
			 faux->Queue, 0) == NULL)
      EGOTO(error);

    if(eType.JSPE_Flags & JPA_VARTYPE_FLONGCONST)
      ret = eType.JSPE_LongConst ? 2 : 1;

    jsparsfreeexprtype(faux->JPA, &eType, 1);
  }

  return ret;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int connectGotos(flowaux_t *faux, jsparscnode_t *top)

/************************************************************************/
{
  jsparsstream_t *jpa;
  jctokfifo_t *queue;
  jsparsvartab_t *vTab;
  jsparsflownode_t *stmt;
  int list, next;

  jpa = faux->JPA;
  queue = faux->Queue;

  vTab = jsparsfindvartab(jpa, top);

  stmt = faux->JFlow->JPAFL_Statements;
  for(list = faux->GotoList ; list != -1 ; list = next) {
    jsparsvardef_t *vDef;
    jsparscnode_t *node;
    const char *name;
    int ind;

    next = stmt[list].JPAFLN_Next1;

    /*  Get label name  */
    node = stmt[list].JPAFLN_Node; /*  the "goto" statement  */
    ind = node->JPACN_Sub->JPACN_Start;
    name = queue->TKFF_Token[ind].TK_String;

    if((vDef = jsparsgetvardef(jpa, vTab, name, JVDF_TLABEL, 0)) == NULL) {
      jsparsseterrcode2(_ERR_JSPARSSCANFLOW, 3, name, faux->Queue, top);
      EGOTO(error);
    }

    /*  To find unused labels  */
    vDef->JPAVD_Flags |= JVDF_FREAD;

    node = vDef->JPAVD_Declarator;

    /*  Locate the label in the statement flow  */
    if((ind = findNodeIndex(stmt, faux->JFlow->JPAFL_NStatements,
			    node->JPACN_Start)) == -1)
      EGOTO(error);

    stmt[list].JPAFLN_Next1 = ind;

    if(ind >= 0)
      stmt[ind].JPAFLN_Counter++;
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jsparsfindflow(jsparsstream_t *jpa, jsparscnode_t *node)

/************************************************************************/
{
  jsparsflow_t *jFlow;

  jFlow = jpa->JSPA_StmtFlow;
  return findNodeIndex(jFlow->JPAFL_Statements, jFlow->JPAFL_NStatements,
		       node->JPACN_Start);
}

/*JS*********************************************************************
*
*************************************************************************/

static int findNodeIndex(jsparsflownode_t *stmt, unsigned int n, unsigned int indx)

/************************************************************************/
{
  unsigned int lower, upper;

  lower = 0; upper = n;
  while(lower < upper) {
    unsigned int m = (lower + upper) >> 1;

    if(stmt[m].JPAFLN_Node->JPACN_Start < indx)
      lower = m + 1;
    else
      upper = m;
  }

  if(lower >= n)
    return -1;

  return lower;
}

#ifdef TRACETYPE_JSCPP
/*JS*********************************************************************
*
*************************************************************************/

static void showStmtFlow(FILE *fp, jsparsflow_t *jFlow, jctokfifo_t *queue, int mode)

/************************************************************************/
{
  unsigned int i;

  for(i = 0 ; i < jFlow->JPAFL_NStatements ; ++i) {
    jsparsflownode_t *flNode;
    jsparscnode_t *node;
    long flags;

    flNode = &(jFlow->JPAFL_Statements[i]);
    node   = flNode->JPAFLN_Node;
    flags  = flNode->JPAFLN_Flags;

    fprintf(fp,
	    "STMT %3d: FLAGS %c%c%c%c%c%c CNT %2d N1 %3d N2 %3d NODE [%d,%d]\n",
	    i,
	    (flags & JPA_FLOW_SINGLESTMT)   ? 'S' : '-',
	    (flags & JPA_FLOW_BLOCKSTART)   ? 'B' : '-',
	    (flags & JPA_FLOW_EXPRESSION)   ? 'E' : '-',
	    (flags & JPA_FLOW_PSEUDOSTMT)   ? 'P' : '-',
	    (flags & JPA_FLOW_FIRSTSUBNODE) ? '1' : '-',
	    (flags & JPA_FLOW_BRANCHED)     ? 'C' : '-',
	    flNode->JPAFLN_Counter,
	    flNode->JPAFLN_Next1, flNode->JPAFLN_Next2,
	    node->JPACN_Start / queue->TKFF_Step,
	    node->JPACN_End / queue->TKFF_Step);

    if(mode == 2 && !(flags & JPA_FLOW_PSEUDOSTMT)) {
      putc('\t', fp);
      jsparsprnode(fp, queue, node, 0);
      putc('\n', fp);
    }

    if(mode == 3)
      jsparsprctree(fp, queue, node, 2, 3);
  }
}
#endif
