/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : C preprocessor and ANSI-C parser.
*    Part    : ANSI-C Grammar for C.
*              Taken from: K&R "The C programming language", second
*              german edition and almost completely rewritten and
*              simplified. The grammar therein contains many errors
*              and is NOT ANSI compliant. Problems that have been
*              resolved by this grammar include
*               - does now accept the redefinition of a typedef name in
*                 an inner block.
*               - different name class for goto labels exists.
*               - a lot of bad constructions are recognized within the
*                 grammar.
*              Some non-ANSI'sm are accepted and warning messages can be
*              produced on demand.
*
*              Copyright (C) 1994-2007 Joerg Schoen
*
*************************************************************************/

/*
 *  Debugging bits (assuming that CONFIG_JSCPPLIB_DEBUG == 8)
 *   Bit 12: Show tokens as processed by lexical analyzer.
 *   Bit 13: Show parser tree.
 *   Bit 14: Show parser tree with start/end indexes of each
 *           node. If bit 13 is set as well, show the queue.
 *   Bit 15: Do not process tree.
 *
 *   Bit 16: Show generated variable tables. Also, do not exit
 *           on parser error.
 *   Bit 17: Show declarator and specifier for each variable.
 *   Bit 18: Show also sub variable tables of struct and unions.
 *   Bit 19: Switch on yacc debugging if compiled with debugging.
 *
 *   Bit 20: Generate and show statement flow.

 *   Bit 23: Report assignments where right hand side is a boolean
 *           expression.
 */

%{

/*********     INCLUDES 					*********/
#ifdef DEBUG_ALLOC
 /*  Some versions of YACC includes <malloc.h>
  *   which interferes with <dbgalloc.h>,
  *   so fool YACC.
  */
# include <malloc.h>
#endif

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
#ifndef JPA_KEYWORD_TABSIZE
/*  START-DEFINITIONS 3 */
/*  Hash table for keywords  */
#define JPA_KEYWORD_TABSIZE  251   /*  prime number  */

/*  For a keyword or an identifier  */
typedef struct _jsparsident_t _jsparsident_t;
struct _jsparsident_t {
  _jsparsident_t *JPAID_Next;
  int             JPAID_Value;

  const char     *JPAID_File;
  int             JPAID_Line;

  char	          JPAID_Name[1];
};

typedef struct _jsparstypedef_t _jsparstypedef_t;
struct _jsparstypedef_t {
  _jsparstypedef_t *JPATD_Next;
  int               JPATD_Flags;
  _jsparsident_t   *JPATD_HashTable[JPA_KEYWORD_TABSIZE];
};

typedef struct {
  int JPABC_Flag;
  const char *JPABC_Name;
  int JPABC_ArrayArg, JPABC_IndexArg;
} _jsparsboundchk_t;

/* **  For JSPA_Mode  ** */
/*  To detect empty files and thus emit a warning (ANSI forbids empty files)  */
#define JPAM_SEENNODE               (1 << 0)

/*  The parser should return original tokens along with the preprocessed ones  */
#define JPAM_FGETORIG               (1 << 1)
/*  The parser should return file position information along with the tokens  */
#define JPAM_FGETFPOS               (1 << 2)
/*  When flushing, print the original ones instead of the preprocessed ones  */
#define JPAM_FPRINTORIG             (1 << 3)
/*  When flushing, print preprocessed ones except when in system include
 *   files or when expanding tokens from system include files: then print
 *   original ones.
 */
#define JPAM_FEXPANDUSERFILES       (1 << 4)

/*  When parsing, generate variable tables  */
#define JPAM_FVARTABLES             (1 << 5)

/*  When parsing, generate a statement flow  */
#define JPAM_FSTMTFLOW              (1 << 6)

/*  Generate a complete flow for each node  */
#define JPAM_FSTMTFULLFLOW          (1 << 7)

/*  When generating a statement flow chart, treat the end of a compound
 *   statement "}" as a separate "statement". All statements that flow to
 *   the end of the block will "tip" on it instead of going directly to
 *   the follower.
 */
#define JPAM_FFLOWBLOCKEND          (1 << 8)

/*  When encountering "if", "while", "for", the statement itself
 *   should become the node instead of the expression with
 *   JPA_FLOW_EXPRESSION set
 */
#define JPAM_FFLOWREALSTMT          (1 << 9)

/*  This flag is set by the statement flow generator if the end of the
 *   block can be reached
 */
#define JPAM_FFALLTOEND             (1 << 10)

/*  Make some things a warning rather than an error  */
#define JPAM_FMIXEDDECLS            (1 << 11)
#define JPAM_FALLOWREDECL           (1 << 12)

#ifdef JSPARSE_INTERNAL
#define JPAM_FCLOSESTREAM           (1 << 16)
#endif

#define JPAM_FUSERFLAGS             (1 << 24)

/* ****  For JSPA_WarnFlags   **** */
/*  Show warnings pertaining the preprocessor  */
#define JPAM_WARNPREPROC            (1 << 0)

/*  Warn about missing 'static' in function definition where static
 *   prototype was seen and function calls without prototype.
 */
#define JPAM_WARNPROTO              (1 << 1)
/*  Show other warnings like bad code style (better use a decent compiler for that) */
#define JPAM_WARNSTYLE              (1 << 2)
/*  Check for unused variables, redefinitions etc.  */
#define JPAM_WARNVARS               (1 << 3)
/*  Do bounds checking  */
#define JPAM_WARNBOUNDS             (1 << 4)
/*  Find bad casts  */
#define JPAM_WARNBADCAST            (1 << 5)
/*  Warn about unreachable code or unused labels  */
#define JPAM_WARNFLOW               (1 << 6)
/*  Warn about functions with high usage of stack memory, i. e. local variables  */
#define JPAM_WARNSTACKUSE           (1 << 7)

typedef struct {
  long              JSPA_Mode;

  long              JSPA_WarnFlags;

  jsqueuestream_t  *JSPA_Stream;

  _jsparsident_t   *JSPA_KeyWordTable[JPA_KEYWORD_TABSIZE];

  /*  The following three are needed by the parser to handle
   *   typedefs correctly.
   */
  _jsparstypedef_t *JSPA_TypeDefs;
  _jsparstypedef_t *JSPA_DeferBlockDef;
  _jsparstypedef_t *JSPA_SavedBlockDef;

  jsparsvartab_t  **JSPA_VarTables;
  unsigned int      JSPA_NVarTables, JSPA_MVarTables;

  _jsparsboundchk_t *JSPA_BoundChkPragma;
  unsigned int JSPA_NBoundChkPragma, JSPA_MBoundChkPragma;

  jsparsflow_t     *JSPA_StmtFlow;

#define JSPA_Statements     JSPA_StmtFlow->JPAFL_Statements
#define JSPA_FlowStart      JSPA_StmtFlow->JPAFL_Start

  jctokfifo_t       JSPA_GlobalQueue;

  /*  Called to print error messages  */
  void (*JSPA_PrintError)(void *user, const char *message);

  void *JSPA_User;
} jsparsstream_t;

#define jsparsgetqueue(jpa)       jsqsgetqueue((jpa)->JSPA_Stream)

/*  First free token for parser  */
#define PTK_PARSERTOKEN   (PTK_USERTOKEN + 10)

#ifdef JSPARSE_INTERNAL
#define BTD_CMPSTMT             1
#endif
/*  END-DEFINITIONS  */
#endif

/*  If defined, redefine all global symbols to avoid name clash  */
#ifdef CONFIG_JSPARSE_YNAME
# define CONCAT(a, b)         CONCATX(a, b)
# define CONCATX(a, b)        a ## b

# define yyparse    CONCAT(CONFIG_JSPARSE_YNAME, parse)
# define yylex      CONCAT(CONFIG_JSPARSE_YNAME, lex)
# define yyerror    CONCAT(CONFIG_JSPARSE_YNAME, error)
# define yylval     CONCAT(CONFIG_JSPARSE_YNAME, lval)
# define yychar     CONCAT(CONFIG_JSPARSE_YNAME, char)
# define yydebug    CONCAT(CONFIG_JSPARSE_YNAME, debug)
/*  The bison man page omitted yynerrs  */
# define yynerrs    CONCAT(CONFIG_JSPARSE_YNAME, nerrs)
#endif

#ifdef CONFIG_JSCPPLIB_DEBUG
# define YYDEBUG   1
#endif

/*  This is needed in conjunction with Jim Roskind's yacc debugger.
 *   Note to set the environment variable "YYDEBUG" to 6.
 */
#define YYDEBUG_LEXER_TEXT  (PA_TOKEN((yylval).TV_Index).TK_String ? \
			PA_TOKEN((yylval).TV_Index).TK_String : "empty")

#define PA_TOKEN(indx)          ParserQueue->TKFF_Token[indx]

#define STMT_IS_EMPTY(stmt)    ((stmt)->JPACN_Type == JPACT_UNARYEXPR && \
				(stmt)->JPACN_Sub == NULL)

/* ***  Common abbreviations  *** */
#define MKCNODE(next, type, sub, start, end)  \
		jsparsmkcnode(next, type, sub, start, end, ParserQueue)

#define MKUNARYEXPR(node, sub) \
		_jsparsmkunary(node, sub, ParserQueue)

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES					*********/
static int               yylex(void);

static _jsparsident_t   *putNewIdent(jsparsstream_t *jpa, jctokfifo_t *queue,
				     jsparscnode_t *node, int value);

static void              yyerror(char *string);

Prototype jsparscnode_t *jsparsgetexpression(jsparsstream_t *jpa,
					     jsqueuestream_t *jsqs,
					     int mode);
Prototype jsparscnode_t *jsparsgettype(jsparsstream_t *jpa,
				       jsqueuestream_t *jsqs,
				       int mode);

IPrototype int           _jsparsgetnode(jsparsstream_t *jpa,
					jsqueuestream_t *jsqs,
					jsparscnode_t **pNode,
					int starttoken);
static int               exprIsConstant(jctokfifo_t *queue, jsparscnode_t *node);

Prototype const char    *jsparstok2string(jsctoken_t token);

IPrototype const jsctokops_t JSParserCOperators[];
IPrototype const jsctokops_t JSParserCKeyWords[];

extern int yydebug;

/*********     GLOBAL / STATIC VARIABLES       			*********/
static jsparsstream_t  *ParserJPA;

/*  We ensure that ParserQueue == jsqgetqueue(ParserStream)
 *   We have it in an extra variable for sake of efficiency.
 */
static jsqueuestream_t *ParserStream;
static jctokfifo_t     *ParserQueue;

static int ParserPrefixToken = -1;
static int ParserIndex;
static jsparscnode_t *ParserNode;

%}

%start program

%union
{
  int            TV_Index; /* Index in ParserQueue of this token  */
  jsparscnode_t *TV_CNode; /* A tree returned by "MKCNODE" of this construct */
}

/*  Generic input tokens  */
%token <TV_Index> TK_ERROR
%token <TV_Index> TK_IDENTIFIER   TK_INTCONST     TK_FLOATCONST
%token <TV_Index> TK_CHARCONST    TK_WCHARCONST   TK_UCHARCONST
%token <TV_Index> TK_STRING       TK_WSTRING      TK_USTRING

%token <TV_Index> TK_TYPEDEFNAME  TK_ENUMCONST

/*  Token for reserved words of C  */
%token <TV_Index> TK_KW_AUTO      TK_KW_BREAK     TK_KW_CASE     TK_KW_CHAR
%token <TV_Index> TK_KW_CONST     TK_KW_CONTINUE  TK_KW_DEFAULT  TK_KW_DO
%token <TV_Index> TK_KW_DOUBLE    TK_KW_ELSE      TK_KW_ENUM     TK_KW_EXTERN
%token <TV_Index> TK_KW_FLOAT     TK_KW_FOR       TK_KW_GOTO     TK_KW_IF
%token <TV_Index> TK_KW_INT       TK_KW_LONG      TK_KW_REGISTER TK_KW_RETURN
%token <TV_Index> TK_KW_SHORT     TK_KW_SIGNED    TK_KW_SIZEOF   TK_KW_TYPEOF
%token <TV_Index> TK_KW_STATIC    TK_KW_STRUCT    TK_KW_SWITCH   TK_KW_TYPEDEF
%token <TV_Index> TK_KW_UNION     TK_KW_UNSIGNED  TK_KW_VOID     TK_KW_VOLATILE
%token <TV_Index> TK_KW_WHILE

/*  Token for C operators  */
%token <TV_Index> TK_OP_INC       TK_OP_DEC       TK_OP_POINT     TK_OP_EQUALS
%token <TV_Index> TK_OP_NEQUALS   TK_OP_LOEQUAL   TK_OP_GREQUAL   TK_OP_LSHIFT
%token <TV_Index> TK_OP_RSHIFT    TK_OP_LAND      TK_OP_LOR       TK_OP_ADDASS
%token <TV_Index> TK_OP_SUBASS    TK_OP_MULASS    TK_OP_DIVASS    TK_OP_MODASS
%token <TV_Index> TK_OP_LSHIFTASS TK_OP_RSHIFTASS TK_OP_ORASS     TK_OP_ANDASS
%token <TV_Index> TK_OP_EXORASS   TK_ELLIPSIS

/*  Special tokens to use parser with different lhs  */
%token TK_START_EXPRESSION

/*  Special token to read a type  */
%token TK_START_TYPE

/*  When using bison, this suppresses the warning for
 *   our only shift/reduce conflict due to the
 *   if-else ambiguity.
 */
/*%expect 1*/

/*  Alternate method to get rid of the shift/reduce conflict:
 *   Look for "LOWER_THAN_ELSE" in remainder of file.
 */
%nonassoc LOWER_THAN_ELSE
%nonassoc TK_KW_ELSE

/*  We collect the sequence of tokens and return the pragma line as a whole  */
%token <TV_CNode> TK_PRAGMALINE

%type <TV_Index>  '=' ':' ';' '&' '+' '-' '*' '/' '~' '!' '.'
%type <TV_Index>  '{' '}' '(' ')' '[' ']'

/* ***  Complicated tokens return a tree structure  *** */
%type <TV_CNode>  program program_unit special_expression
%type <TV_CNode>  external_declaration function_definition function_declarator
%type <TV_CNode>  function_declarator_kr global_declaration declaration
%type <TV_CNode>  declaration_list typedef_declaration
%type <TV_CNode>  typedef_declaration_start_withtype
%type <TV_CNode>  typedef_declaration_start_notype
%type <TV_CNode>  struct_or_union_specification

%type <TV_CNode>  declaration_specifiers declaration_specifiers_notype
%type <TV_CNode>  declaration_specifiers_withtype storage_class_specifier
%type <TV_CNode>  simple_type_specifier complex_type_specifier
%type <TV_CNode>  specifier_qualifier_list specifier_qualifier_list_simple
%type <TV_CNode>  specifier_qualifier_list_complex
%type <TV_CNode>  specifier_qualifier_list2

%type <TV_CNode>  type_qualifier_list type_qualifier_list_opt
%type <TV_CNode>  type_qualifier

%type <TV_CNode>  struct_or_union_specifier struct_or_union
%type <TV_CNode>  struct_declaration_list init_declarator_list
%type <TV_CNode>  init_any_declarator_list
%type <TV_CNode>  typedef_any_declarator_list typedef_any_declarator_basic
%type <TV_CNode>  typedef_declarator_list typedef_declarator_basic
%type <TV_CNode>  init_declarator init_any_declarator
%type <TV_CNode>  any_declarator2
%type <TV_CNode>  struct_declaration
%type <TV_CNode>  specifier_qualifier_list_notype
%type <TV_CNode>  specifier_qualifier_list_withtype
%type <TV_CNode>  struct_declarator_list_opt struct_declarator_list
%type <TV_CNode>  struct_declarator

%type <TV_CNode>  struct_any_declarator_list_opt struct_any_declarator_list
%type <TV_CNode>  struct_any_declarator
%type <TV_CNode>  struct_any_declarator2
%type <TV_CNode>  smallint_length
%type <TV_CNode>  enum_specifier
%type <TV_CNode>  enumerator_list init_enumerator enumerator
%type <TV_CNode>  typeof_construct

%type <TV_CNode>  declarator2 declarator
%type <TV_CNode>  declarator_noparen direct_declarator_noparen
%type <TV_CNode>  direct_declarator_noparen_id
%type <TV_CNode>  direct_declarator_noparen_noid
%type <TV_CNode>  declarator_noparen2

%type <TV_CNode>  declarator_paren direct_declarator_paren
%type <TV_CNode>  direct_declarator_paren_id
%type <TV_CNode>  direct_declarator_paren_noid
%type <TV_CNode>  declarator_paren2

%type <TV_CNode>  typedef_declarator
%type <TV_CNode>  typedef_declarator_noparen typedef_direct_declarator_noparen
%type <TV_CNode>  typedef_direct_declarator_noparen_id
%type <TV_CNode>  typedef_direct_declarator_noparen_noid
%type <TV_CNode>  typedef_declarator_noparen2

%type <TV_CNode>  typedef_declarator_paren typedef_direct_declarator_paren
%type <TV_CNode>  typedef_direct_declarator_paren_id
%type <TV_CNode>  typedef_direct_declarator_paren_noid
%type <TV_CNode>  typedef_declarator_paren2
%type <TV_CNode>  declarator_app

%type <TV_CNode>  declarator_app_nokr declarator_app_func declarator_app_nofunc
%type <TV_CNode>  any_declarator
%type <TV_CNode>  pointer
%type <TV_CNode>  parameter_type_list
%type <TV_CNode>  parameter_list parameter_declaration
%type <TV_CNode>  any_declarator_param2 any_declarator_param identifier_list
%type <TV_CNode>  kr_declaration initializer initializer_list type_name
%type <TV_CNode>  abstract_declarator direct_abstract_declarator

%type <TV_CNode>  statement labeled_statement unlabeled_statement
%type <TV_CNode>  label_list statement_label expression_statement
%type <TV_CNode>  compound_statement compound_statement2
%type <TV_CNode>  compound_statement_content statement_or_declaration_list
%type <TV_CNode>  statement_or_declaration
%type <TV_CNode>  label_list_opt selection_statement
%type <TV_CNode>  iteration_statement expression_opt jump_statement
%type <TV_CNode>  expression unlabeled_expression
%type <TV_CNode>  assignment_expression assignment_operator
%type <TV_CNode>  conditional_expression constant_expression

%type <TV_CNode>  logical_or_expression logical_and_expression
%type <TV_CNode>  inclusive_or_expression exclusive_or_expression
%type <TV_CNode>  and_expression equality_expression relational_expression
%type <TV_CNode>  shift_expression
%type <TV_CNode>  additive_expression multiplicative_expression

%type <TV_CNode>  cast_expression unary_expression unary_operator
%type <TV_CNode>  postfix_expression primary_expression
%type <TV_CNode>  argument_expression_list argument_expression
%type <TV_CNode>  constant integer_constant character_constant
%type <TV_CNode>  floating_constant enumeration_constant
%type <TV_CNode>  any_identifier identifier_or_typedefname
%type <TV_CNode>  identifier_or_enumconst
%type <TV_CNode>  identifier string cstring wstring ustring typedef_name

%type <TV_CNode>  error_stmt

%type <TV_CNode>  pragma_line

%%  /* ***  SECTION RULES  *** */

/* ***************************** */
/* ***  Rules for a program  *** */
/* ***************************** */
program:
  program_unit
  {
    ParserNode = $$;

    /*  That's enough for now, so stop parsing. Otherwise, we'll
     *   parse the *whole* input file at once.
     */
    YYACCEPT;
  }
  ;

program_unit:
  /*  empty  */ { $$ = NULL; }
  | external_declaration
  | special_expression
  ;

special_expression:
  TK_START_EXPRESSION expression
  {
    $$ = $2;
    /*  We do not "YYACCEPT", since the *whole* input has to fit  */
  }
  | TK_START_TYPE type_name
  {
    $$ = MKCNODE(NULL, JPACT_TYPEOF, $2, $2->JPACN_Start, $2->JPACN_End);
    /*  We do not "YYACCEPT", since the *whole* input has to fit  */
  }
  ;

external_declaration:
  function_definition
  | global_declaration
  | pragma_line /*  Allow pragma lines in global space  */
  | error_stmt  /*  error handling is very crude  */
  ;

function_definition:
  declaration_specifiers function_declarator
    {
      /*  Reuse old block definition (if existent)  */
      if(_jsparspushbtype(ParserJPA, ParserJPA->JSPA_SavedBlockDef,
			  BTD_CMPSTMT) == -1)
	YYABORT;

      ParserJPA->JSPA_SavedBlockDef = NULL;
    }
    compound_statement2 POP_BLOCK
    {
      int start;

      $2->JPACN_Next = $1;
      start = _jsparsgettail($1)->JPACN_Start;

      $$ = MKCNODE(NULL, JPACT_FUNCDEF,
		   MKCNODE($4, JPACT_DECLARATION, $2, start,
			   $2->JPACN_End), start, $4->JPACN_End);
    }
  | function_declarator
    {
      /*  Reuse old block definition (if existent)  */
      if(_jsparspushbtype(ParserJPA, ParserJPA->JSPA_SavedBlockDef,
			  BTD_CMPSTMT) == -1)
	YYABORT;

      ParserJPA->JSPA_SavedBlockDef = NULL;

      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		     "Return value of function defaults to \"int\"", 0);
    }
    compound_statement2 POP_BLOCK
    {
      $$ = MKCNODE(NULL, JPACT_FUNCDEF,
		   MKCNODE($3, JPACT_DECLARATION, $1,
			   $1->JPACN_Start, $1->JPACN_End),
		   $1->JPACN_Start, $3->JPACN_End);
    }
  | declaration_specifiers function_declarator_kr
    {
      /*  Reuse old block definition (if existent)  */
      if(_jsparspushbtype(ParserJPA, ParserJPA->JSPA_SavedBlockDef,
			  BTD_CMPSTMT) == -1)
	YYABORT;

      ParserJPA->JSPA_SavedBlockDef = NULL;
    }
    declaration_list compound_statement2 POP_BLOCK
    {
      jsparscnode_t *n;
      int start;

      $2->JPACN_Next = $1;
      start = _jsparsgettail($1)->JPACN_Start;

      if($4->JPACN_Next == NULL) {
	n = $4;
	n->JPACN_Next = $5;
      } else
	n = MKCNODE($5, JPACT_LIST, $4, $4->JPACN_Start,
		    _jsparsgettail($4)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_FUNCDEF,
		   MKCNODE(n, JPACT_DECLARATION, $2, start,
			   $2->JPACN_End), start, $5->JPACN_End);
    }
  | function_declarator_kr
    {
      /*  Reuse old block definition (if existent)  */
      if(_jsparspushbtype(ParserJPA, ParserJPA->JSPA_SavedBlockDef,
			  BTD_CMPSTMT) == -1)
	YYABORT;

      ParserJPA->JSPA_SavedBlockDef = NULL;

      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		     "Return value of function defaults to \"int\"", 0);
    }
    declaration_list compound_statement2 POP_BLOCK
    {
      jsparscnode_t *n;

      if($3->JPACN_Next == NULL) {
	n = $3;
	n->JPACN_Next = $4;
      } else
	n = MKCNODE($4, JPACT_LIST, $3, $3->JPACN_Start,
		    _jsparsgettail($3)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_FUNCDEF,
		   MKCNODE(n, JPACT_DECLARATION, $1,
			   $1->JPACN_Start, $1->JPACN_End),
		   $1->JPACN_Start, $4->JPACN_End);
    }
  ;

function_declarator:
  declarator2 /* can be no typedef_name here, because we are in global block */
    {
      jsparscnode_t *n;

      /*  Check if this is a valid function declarator  */
      if($1->JPACN_Sub->JPACN_Next == NULL ||
	 $1->JPACN_Sub->JPACN_Next->JPACN_Type != JPACT_FUNCCALL) {
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Illegal external declaration", 1);
	YYERROR;
      }

      if((n = $1->JPACN_Sub->JPACN_Next->JPACN_Sub) == NULL) { /*  "f()"  */
	if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	  jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Missing \"void\" in function arguments", 0);
      } else if(n->JPACN_Type != JPACT_PARDECL) { /*  "f(a,b)"  */
	if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	  jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Untyped function arguments defaults to \"int\"", 0);
      } else if(n->JPACN_Sub->JPACN_Type == JPACT_VOID) {
	if(n->JPACN_Next || n->JPACN_Sub->JPACN_Next) {
	  jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Illegal \"void\" in parameter list", 1);
	  YYERROR;
	}
      } else {
	/*  All arguments must contain identifiers  */
	do {
	  if(n->JPACN_Sub->JPACN_Type != JPACT_ELLIPSIS &&
	     (n->JPACN_Sub->JPACN_Type != JPACT_DECLARATOR ||
	      n->JPACN_Sub->JPACN_Sub->JPACN_Type != JPACT_IDENT)) {
	    jsparsprinterr(ParserJPA, ParserQueue, n,
			   "Untyped argument in function definition", 1);
	    YYERROR;
	  }
	} while((n = n->JPACN_Next) != NULL);
      }
    }
  ;

function_declarator_kr:
  declarator2 /* can be no typedef_name here, because we are in global block */
    {
      /*  Check if this is a valid function declarator  */
      if($1->JPACN_Sub->JPACN_Next == NULL ||
	 $1->JPACN_Sub->JPACN_Next->JPACN_Type != JPACT_FUNCCALL) {
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Illegal external declaration", 1);
	YYERROR;
      }

      /*  If "f()", than no declarations before '{' are allowed,
       *   case "f()" will be handled by "function_declarator"
       */
      if($1->JPACN_Sub->JPACN_Next->JPACN_Sub == NULL ||
	 /*  No mixing of K&R and ANSI typing of parameters allowed  */
	 $1->JPACN_Sub->JPACN_Next->JPACN_Sub->JPACN_Type != JPACT_IDENT) {
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Illegal K&R function declarator", 1);
	YYERROR;
      }

      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		     "Old K & R style function definition encountered", 0);
    }
  ;

global_declaration:
  typedef_declaration
  | declaration_specifiers init_declarator_list ';'
    {
      jsparscnode_t *n;

      if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1;
      } else
	n = MKCNODE($1, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_DECLARATION, n,
		   _jsparsgettail($1)->JPACN_Start, $3);
    }
  | init_declarator_list ';'
  {
    jsparscnode_t *n;

    /*  A declaration with missing 'int'  */
    if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
      jsparsprinterr(ParserJPA, ParserQueue, $1,
		     "Untyped declaration defaults to \"int\"", 0);

    if($1->JPACN_Next == NULL) {
      n = $1;
    } else
      n = MKCNODE(NULL, JPACT_LIST, $1, $1->JPACN_Start,
		  _jsparsgettail($1)->JPACN_End);

    $$ = MKCNODE(NULL, JPACT_DECLARATION, n, $1->JPACN_Start, $2);
  }
  | struct_or_union_specification
  ;

/* ******************************** */
/* ***  Rules for declarations  *** */
/* ******************************** */
declaration:
  typedef_declaration
  | declaration_specifiers_notype init_declarator_list ';'
    {
      jsparscnode_t *n;

      if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1;
      } else
	n = MKCNODE($1, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_DECLARATION, n,
		   _jsparsgettail($1)->JPACN_Start, $3);
    }
  /*  If declaration specifiers contain an explicit type name, allow
   *   variables to be a typedefname. We than have a redefinition
   *   of a typedefname.
   */
  | declaration_specifiers_withtype init_any_declarator_list ';'
    {
      jsparscnode_t *n;

      if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1;
      } else {
	n = MKCNODE($1, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);
      }

      $$ = MKCNODE(NULL, JPACT_DECLARATION, n,
		   _jsparsgettail($1)->JPACN_Start, $3);
    }
  | struct_or_union_specification
  ;

declaration_list: /* **  list  ** */
  declaration
  | declaration_list declaration
    { $$ = _jsparslinktail($1, $2); }
  ;

typedef_declaration:
  typedef_declaration_start_withtype typedef_any_declarator_list ';'
    {
      jsparscnode_t *n;

      if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1->JPACN_Sub;
      } else {
	n = MKCNODE($1->JPACN_Sub, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);
      }

      $$->JPACN_Sub = n;
      $$->JPACN_End = $3;
    }
  | typedef_declaration_start_notype typedef_declarator_list ';'
    {
      jsparscnode_t *n;

      if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1->JPACN_Sub;
      } else {
	n = MKCNODE($1->JPACN_Sub, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);
      }

      $$->JPACN_Sub = n;
      $$->JPACN_End = $3;
    }
  | typedef_declaration_start_notype ';'
    {
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Useless 'typedef' construct", 0);
    }
  | typedef_declaration_start_withtype ';'
    {
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Useless 'typedef' construct", 0);
    }
  ;

typedef_declaration_start_withtype: /* ** not a list!  ** */
  /*  Type specifier right to "typedef"  */
  type_qualifier_list_opt TK_KW_TYPEDEF specifier_qualifier_list
    {
      $$ = MKCNODE(NULL, JPACT_TYPEDEF, _jsparslinktail($3, $1),
		   $1 ? _jsparsgettail($1)->JPACN_Start : $2, $3->JPACN_End);
    }
  /*  Type specifier left to "typedef"  */
  | specifier_qualifier_list_simple TK_KW_TYPEDEF type_qualifier_list_opt
    {
      $$ = MKCNODE(NULL, JPACT_TYPEDEF, _jsparslinktail($3, $1),
		   _jsparsgettail($1)->JPACN_Start, $3 ? $3->JPACN_End : $2);
    }
  /*  Type specifier left to "typedef"  */
  | specifier_qualifier_list_complex TK_KW_TYPEDEF type_qualifier_list_opt
    {
      $$ = MKCNODE(NULL, JPACT_TYPEDEF, _jsparslinktail($3, $1),
		   _jsparsgettail($1)->JPACN_Start, $3 ? $3->JPACN_End : $2);
    }
  /*  Simple type specifier spread around "typedef"  */
  | specifier_qualifier_list_simple TK_KW_TYPEDEF
    specifier_qualifier_list_simple
    {
      $$ = MKCNODE(NULL, JPACT_TYPEDEF, _jsparslinktail($3, $1),
		   _jsparsgettail($1)->JPACN_Start, $3->JPACN_End);
    }
  ;

typedef_declaration_start_notype:
  type_qualifier_list_opt TK_KW_TYPEDEF type_qualifier_list_opt
    {
      $$ = MKCNODE(NULL, JPACT_TYPEDEF, _jsparslinktail($3, $1),
		   $1 ? _jsparsgettail($1)->JPACN_Start : $2,
		   $3 ? $3->JPACN_End : $2);
    }
  ;

struct_or_union_specification:
  declaration_specifiers ';'
    {
      if(_jsparschkstrunion(ParserJPA, ParserQueue, $1) == -1)
        YYERROR;

      $$ = MKCNODE(NULL, JPACT_DECLARATION, $1,
		   _jsparsgettail($1)->JPACN_Start, $2);
    }
  ;

declaration_specifiers: /* **  list  ** */
  declaration_specifiers_notype
  | declaration_specifiers_withtype
  ;

declaration_specifiers_notype: /* **  list  ** */
  type_qualifier_list_opt storage_class_specifier type_qualifier_list_opt
    {
      /*  No explicit type -> means 'int'  */
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $2,
		       "No explicit type, int assumed", 0);

      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  | type_qualifier_list
    {
      /*  No explicit type -> means 'int'  */
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "No explicit type, int assumed", 0);
    }
  ;

declaration_specifiers_withtype: /* **  list  ** */
  specifier_qualifier_list_simple storage_class_specifier
  type_qualifier_list_opt
    {
      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  | specifier_qualifier_list_simple storage_class_specifier
    specifier_qualifier_list_simple
    {
      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  | specifier_qualifier_list_complex storage_class_specifier
    type_qualifier_list_opt
    {
      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  | type_qualifier_list_opt storage_class_specifier
    specifier_qualifier_list
    {
      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  | specifier_qualifier_list
  ;

storage_class_specifier:
  TK_KW_AUTO       { $$ = MKCNODE(NULL, JPACT_AUTO, NULL, $1, $1); }
  | TK_KW_REGISTER { $$ = MKCNODE(NULL, JPACT_REGISTER, NULL, $1, $1); }
  | TK_KW_STATIC   { $$ = MKCNODE(NULL, JPACT_STATIC, NULL, $1, $1); }
  | TK_KW_EXTERN   { $$ = MKCNODE(NULL, JPACT_EXTERN, NULL, $1, $1); }
  ;

simple_type_specifier:
  TK_KW_VOID       { $$ = MKCNODE(NULL, JPACT_VOID, NULL, $1, $1); }
  | TK_KW_CHAR     { $$ = MKCNODE(NULL, JPACT_CHAR, NULL, $1, $1); }
  | TK_KW_SHORT    { $$ = MKCNODE(NULL, JPACT_SHORT, NULL, $1, $1); }
  | TK_KW_INT      { $$ = MKCNODE(NULL, JPACT_INT, NULL, $1, $1); }
  | TK_KW_LONG     { $$ = MKCNODE(NULL, JPACT_LONG, NULL, $1, $1); }
  | TK_KW_SIGNED   { $$ = MKCNODE(NULL, JPACT_SIGNED, NULL, $1, $1); }
  | TK_KW_UNSIGNED { $$ = MKCNODE(NULL, JPACT_UNSIGNED, NULL, $1, $1); }
  | TK_KW_FLOAT    { $$ = MKCNODE(NULL, JPACT_FLOAT, NULL, $1, $1); }
  | TK_KW_DOUBLE   { $$ = MKCNODE(NULL, JPACT_DOUBLE, NULL, $1, $1); }
  ;

complex_type_specifier:
  struct_or_union_specifier
  | enum_specifier
  | typedef_name
  | typeof_construct
  ;

specifier_qualifier_list: /* **  list  ** */
  specifier_qualifier_list_simple
  | specifier_qualifier_list_complex
  ;

/*  An arbitrary mixing of simple types like "int" and
 *   qualifiers like "const".
 */
specifier_qualifier_list_simple:
  specifier_qualifier_list2 type_qualifier_list_opt
  {
    $$ = _jsparslinktail($2, $1);
  }
  ;

/*  Only one complex type specifier may appear. This is
 *   important to recognize redefinitions of typedef
 *   names in inner blocks.
 */
specifier_qualifier_list_complex: /* **  list  ** */
  type_qualifier_list_opt complex_type_specifier type_qualifier_list_opt
  {
    $2->JPACN_Next = $1;
    $$ = _jsparslinktail($3, $2);
  }
  ;

specifier_qualifier_list2: /* **  list  ** */
  type_qualifier_list_opt simple_type_specifier
    { $$ = _jsparslinktail($2, $1); }
  | specifier_qualifier_list2 type_qualifier_list_opt simple_type_specifier
  {
    $3->JPACN_Next = _jsparslinktail($2, $1);
    $$ = $3;
  }
  ;

type_qualifier_list: /* **  list  ** */
  type_qualifier
  | type_qualifier_list type_qualifier
  {
    $$ = _jsparslinktail($2, $1);
  }
  ;

type_qualifier_list_opt:
  /*  empty  */ { $$ = NULL; }
  | type_qualifier_list
  ;

type_qualifier:
  TK_KW_CONST      { $$ = MKCNODE(NULL, JPACT_CONST, NULL, $1, $1); }
  | TK_KW_VOLATILE { $$ = MKCNODE(NULL, JPACT_VOLATILE, NULL, $1, $1); }
  ;

/* ********************************* */
/* ***  Rules for struct/unions  *** */
/* ********************************* */
struct_or_union_specifier:
  struct_or_union any_identifier '{' PUSH_BLOCK
    struct_declaration_list '}' POP_BLOCK
    {
      $1->JPACN_Sub = $2;
      $2->JPACN_Type = JPACT_IDENT;
      $2->JPACN_Next = MKCNODE(NULL, JPACT_BLOCK, $5, $3, $6);
      $1->JPACN_End = $6;
    }
  | struct_or_union '{' PUSH_BLOCK struct_declaration_list '}' POP_BLOCK
    {
      $1->JPACN_Sub = MKCNODE(NULL, JPACT_BLOCK, $4, $2, $5);
      $1->JPACN_End = $5;
    }
  | struct_or_union any_identifier
    {
      $1->JPACN_Sub = $2;
      $2->JPACN_Type = JPACT_IDENT;
      $1->JPACN_End = $2->JPACN_End;
    }
  ;

struct_or_union:
  TK_KW_STRUCT  { $$ = MKCNODE(NULL, JPACT_STRUCT, NULL, $1, $1); }
  | TK_KW_UNION { $$ = MKCNODE(NULL, JPACT_UNION,  NULL, $1, $1); }
  ;

struct_declaration_list: /* **  list  ** */
  struct_declaration
  | struct_declaration_list struct_declaration
    { $$ = _jsparslinktail($1, $2); }
  ;

init_declarator_list: /* **  list  ** */
  init_declarator
  | init_declarator_list ',' init_any_declarator
    { $$ = _jsparslinktail($1, $3); }
  ;

init_any_declarator_list: /* **  list  ** */
  init_any_declarator
  | init_any_declarator_list ',' init_any_declarator
    { $$ = _jsparslinktail($1, $3); }
  ;

typedef_any_declarator_list: /* **  list  ** */
  typedef_any_declarator_basic
  | typedef_any_declarator_list ',' typedef_any_declarator_basic
    { $$ = _jsparslinktail($1, $3); }
  ;

typedef_any_declarator_basic:
  any_declarator
    {
      /*  Recognize typedefs  */
      if(putNewIdent(ParserJPA, ParserQueue, $1->JPACN_Sub,
		     TK_TYPEDEFNAME) == NULL)
	YYABORT;

      /*  in case we have a redefinition  */
      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT;
    }
  ;

typedef_declarator_list: /* **  list  ** */
  typedef_declarator_basic
  | typedef_declarator_list ',' typedef_declarator_basic
    { $$ = _jsparslinktail($1, $3); }
  ;

typedef_declarator_basic:
  declarator
    {
      /*  Recognize typedefs  */
      if(putNewIdent(ParserJPA, ParserQueue, $1->JPACN_Sub,
		     TK_TYPEDEFNAME) == NULL)
	YYABORT;

      /*  in case we have a redefinition  */
      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT;
    }
  ;

init_declarator:
  declarator2
  | declarator2 '=' initializer
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_ASS, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

init_any_declarator:
  any_declarator2
  | any_declarator2 '=' initializer
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_ASS, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

any_declarator2:
  any_declarator
    {
      if(putNewIdent(ParserJPA, ParserQueue, $1->JPACN_Sub,
		     TK_IDENTIFIER) == NULL)
	YYABORT;

      /*  in case we have a redefinition  */
      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT;
    }
  ;

struct_declaration:
  specifier_qualifier_list_notype struct_declarator_list_opt ';'
    {
      jsparscnode_t *n;

      if($2 == NULL) {
	n = $1;

	if(_jsparschkstrunion(ParserJPA, ParserQueue, n) == -1)
	  YYERROR;
      } else if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1;
      } else
	n = MKCNODE($1, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_DECLARATION, n,
		   _jsparsgettail($1)->JPACN_Start, $3);
    }
  | specifier_qualifier_list_withtype struct_any_declarator_list_opt ';'
    {
      jsparscnode_t *n;

      if($2 == NULL) {
	n = $1;

	if(_jsparschkstrunion(ParserJPA, ParserQueue, n) == -1)
	  YYERROR;
      } else if($2->JPACN_Next == NULL) {
	n = $2;
	n->JPACN_Next = $1;
      } else
	n = MKCNODE($1, JPACT_LIST, $2, $2->JPACN_Start,
		    _jsparsgettail($2)->JPACN_End);

      $$ = MKCNODE(NULL, JPACT_DECLARATION, n,
		   _jsparsgettail($1)->JPACN_Start, $3);
    }
  ;

specifier_qualifier_list_notype:
  type_qualifier_list
  ;

specifier_qualifier_list_withtype:
  specifier_qualifier_list_simple
  | type_qualifier_list_opt complex_type_specifier type_qualifier_list_opt
    {
      $2->JPACN_Next = $1;
      $$ = _jsparslinktail($3, $2);
    }
  ;

struct_declarator_list_opt:
  /*  empty  */ { $$ = NULL; }
  | struct_declarator_list
  ;

struct_declarator_list: /* **  list  ** */
  struct_declarator
  | struct_declarator_list ',' struct_any_declarator
    { $$ = _jsparslinktail($1, $3); }
  ;

struct_declarator:
  declarator2
  | declarator2 ':' smallint_length
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub,
			    MKCNODE(NULL, JPACT_SMALLINT, $3, $2,
				    $3->JPACN_End));
      $1->JPACN_End = $3->JPACN_End;
    }
  | ':' smallint_length
    {
      $$ = MKCNODE(NULL, JPACT_DECLARATOR,
		   MKCNODE(NULL, JPACT_SMALLINT, $2, $1, $2->JPACN_End),
		   $1, $2->JPACN_End);
    }
  ;

struct_any_declarator_list_opt:
  /*  empty  */ { $$ = NULL; }
  | struct_any_declarator_list
  ;

struct_any_declarator_list: /* **  list  ** */
  struct_any_declarator
  | struct_any_declarator_list ',' struct_any_declarator
    { $$ = _jsparslinktail($1, $3); }
  ;

struct_any_declarator:
  struct_any_declarator2
  | struct_any_declarator2 ':' smallint_length
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub,
				      MKCNODE(NULL, JPACT_SMALLINT, $3, $2,
					      $3->JPACN_End));
      $1->JPACN_End = $3->JPACN_End;
    }
  | ':' smallint_length
    {
      $$ = MKCNODE(NULL, JPACT_DECLARATOR,
		   MKCNODE(NULL, JPACT_SMALLINT, $2, $1, $2->JPACN_End),
		   $1, $2->JPACN_End);
    }
  ;

/*
 *  Allow typedef name for struct member, but do
 *   not recognize further on in sub-inner
 *   blocks as a variable definition.
 */
struct_any_declarator2:
  any_declarator
    {
      /*  Do not warn if struct member is identical to typedef name  */
      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT; /*  make always "CT_IDENT"  */
    }
  ;

smallint_length:
  constant_expression
  ;

/*  Do not start a new block, since enumeration constants are valid
 *   in surrounding compound statement block
 */
enum_specifier:
  TK_KW_ENUM any_identifier '{' enumerator_list trailing_comma '}'
    {
      $2->JPACN_Type = JPACT_IDENT;
      $2->JPACN_Next = MKCNODE(NULL, JPACT_BLOCK, $4, $3, $6);
      $$ = MKCNODE(NULL, JPACT_ENUM, $2, $1, $6);
    }
  | TK_KW_ENUM '{' enumerator_list trailing_comma '}'
    {
      $$ = MKCNODE(NULL, JPACT_ENUM,
		   MKCNODE(NULL, JPACT_BLOCK, $3, $2, $5), $1, $5);
    }
  | TK_KW_ENUM any_identifier
    {
      $2->JPACN_Type = JPACT_IDENT;
      $$ = MKCNODE(NULL, JPACT_ENUM, $2, $1, $2->JPACN_End);
    }
  ;

enumerator_list: /* **  list  ** */
  init_enumerator
  | enumerator_list ',' init_enumerator
    { $$ = _jsparslinktail($1, $3); }
  ;

init_enumerator:
  enumerator
  | enumerator '=' constant_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_ASS, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

enumerator:
  any_identifier
    {
      if(putNewIdent(ParserJPA, ParserQueue, $1,
		     TK_ENUMCONST) == NULL)
	YYABORT;

      $1->JPACN_Type = JPACT_IDENT; /*  in case we have a redefinition  */
    }
  ;

trailing_comma:
  /*  empty  */
  | ','
  {
    if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
      jsparsprinterr(ParserJPA, ParserQueue, NULL,
		   "Trailing comma at end of enumerator definition", 0);
  }
  ;

/*  A nice extension: "__typeof__()" can be
 *   used where a type name is expected.
 */
typeof_construct:
  TK_KW_TYPEOF '(' expression ')'
    { $$ = MKCNODE(NULL, JPACT_TYPEOF, $3, $1, $4); }
  | TK_KW_TYPEOF '(' type_name ')'
    { $$ = MKCNODE(NULL, JPACT_TYPEOF, $3, $1, $4); }
  ;

declarator2:
  declarator
    {
      if(putNewIdent(ParserJPA, ParserQueue, $1->JPACN_Sub,
		     TK_IDENTIFIER) == NULL)
	YYABORT;

      /*  in case we have a redefinition  */
      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT;
    }
  ;

declarator:
  declarator_noparen
  | declarator_paren
  ;

/* ****  To disambiguite ...  **** */
declarator_noparen:
  direct_declarator_noparen
  | pointer direct_declarator_noparen
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

direct_declarator_noparen:
  direct_declarator_noparen_id
  | direct_declarator_noparen_noid
  ;

direct_declarator_noparen_id:
  identifier_or_enumconst
  {
    $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1, $1->JPACN_Start, $1->JPACN_End);
  }
  | direct_declarator_noparen_id declarator_app
    {
      /*  Check if this function call is directly applied to the identifier  */
      if($2->JPACN_Type == JPACT_FUNCCALL &&
	 $1->JPACN_Sub->JPACN_Next == NULL) {
	/*  Anything left over?  */
	if(ParserJPA->JSPA_SavedBlockDef) {
	  _jsparsfreebtype(ParserJPA->JSPA_SavedBlockDef);

	  ParserJPA->JSPA_SavedBlockDef = NULL;
	}

	/*  Check if this is really a function definition in ANSI style  */
	if($2->JPACN_Sub && $2->JPACN_Sub->JPACN_Type == JPACT_PARDECL) {
	  ParserJPA->JSPA_SavedBlockDef = ParserJPA->JSPA_DeferBlockDef;
	  ParserJPA->JSPA_DeferBlockDef = NULL;
	}
      }

      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  ;

direct_declarator_noparen_noid:
  direct_declarator_noparen_noid declarator_app
    {
      /*  Check if this function call is directly applied to the identifier  */
      if($2->JPACN_Type == JPACT_FUNCCALL &&
	 $1->JPACN_Sub->JPACN_Next == NULL) {
	/*  Anything left over?  */
	if(ParserJPA->JSPA_SavedBlockDef) {
	  _jsparsfreebtype(ParserJPA->JSPA_SavedBlockDef);
	  ParserJPA->JSPA_SavedBlockDef = NULL;
	}

	/*  Check if this is really a function definition in ANSI style  */
	if($2->JPACN_Sub && $2->JPACN_Sub->JPACN_Type == JPACT_PARDECL) {
	  ParserJPA->JSPA_SavedBlockDef = ParserJPA->JSPA_DeferBlockDef;
	  ParserJPA->JSPA_DeferBlockDef = NULL;
	}
      }

      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  | '(' declarator_noparen2 ')'
    {
      $$ = $2;
      $$->JPACN_Start = $1;
      $$->JPACN_End = $3;
      _jsparschknode(ParserJPA, $$);
    }
  ;

declarator_noparen2:
  /*  Case "direct_declarator_noparen_id" has to be dropped  */
  direct_declarator_noparen_noid
  | pointer direct_declarator_noparen_id
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  | pointer direct_declarator_noparen_noid
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

/* ****  To disambiguite ...  **** */
declarator_paren:
  direct_declarator_paren
  | pointer direct_declarator_paren
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

direct_declarator_paren:
  /*  Case "direct_declarator_paren_id" has to be dropped  */
  direct_declarator_paren_noid
  ;

direct_declarator_paren_id:
  identifier_or_enumconst
  {
    $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1, $1->JPACN_Start, $1->JPACN_End);
  }
  | direct_declarator_paren_id declarator_app
    {
      /*  Check if this function call is directly applied to the identifier  */
      if($2->JPACN_Type == JPACT_FUNCCALL &&
	 $1->JPACN_Sub->JPACN_Next == NULL) {
	/*  Anything left over?  */
	if(ParserJPA->JSPA_SavedBlockDef) {
	  _jsparsfreebtype(ParserJPA->JSPA_SavedBlockDef);
	  ParserJPA->JSPA_SavedBlockDef = NULL;
	}

	/*  Check if this is really a function definition in ANSI style  */
	if($2->JPACN_Sub && $2->JPACN_Sub->JPACN_Type == JPACT_DECLARATOR) {
	  ParserJPA->JSPA_SavedBlockDef = ParserJPA->JSPA_DeferBlockDef;
	  ParserJPA->JSPA_DeferBlockDef = NULL;
	}
      }

      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  ;

direct_declarator_paren_noid:
  direct_declarator_paren_noid declarator_app
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  | '(' declarator_paren2 ')'
    {
      $$ = $2;
      $$->JPACN_Start = $1;
      $$->JPACN_End = $3;
      _jsparschknode(ParserJPA, $$);
    }
  ;

declarator_paren2:
  direct_declarator_paren_id
  | direct_declarator_paren_noid
  /*  Case "pointer direct_declarator_paren_id" has to be dropped  */
  | pointer direct_declarator_paren_noid
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

/*  The same as "declarator", but with a typedef_name as the identifier  */
typedef_declarator:
  typedef_declarator_noparen
  | typedef_declarator_paren
  ;

/* ****  To disambiguite ...  **** */
typedef_declarator_noparen:
  typedef_direct_declarator_noparen
  | pointer typedef_direct_declarator_noparen
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

typedef_direct_declarator_noparen:
  typedef_direct_declarator_noparen_id
  | typedef_direct_declarator_noparen_noid
  ;

typedef_direct_declarator_noparen_id:
  typedef_name
     {
       /*  We have a redefinition of a typedef. If old stems from
	*   the same block, this is an error.
	*/
       $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1,
		    $1->JPACN_Start, $1->JPACN_End);
     }
  | typedef_direct_declarator_noparen_id declarator_app
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  ;

typedef_direct_declarator_noparen_noid:
  typedef_direct_declarator_noparen_noid declarator_app
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  | '(' typedef_declarator_noparen2 ')'
    {
      $$ = $2;
      $$->JPACN_Start = $1;
      $$->JPACN_End = $3;
      _jsparschknode(ParserJPA, $$);
    }
  ;

typedef_declarator_noparen2:
  /*  Case "typedef_direct_declarator_noparen_id" has to be dropped  */
  typedef_direct_declarator_noparen_noid
  | pointer typedef_direct_declarator_noparen_id
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  | pointer typedef_direct_declarator_noparen_noid
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

/* ****  To disambiguite ...  **** */
typedef_declarator_paren:
  typedef_direct_declarator_paren
  | pointer typedef_direct_declarator_paren
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

typedef_direct_declarator_paren:
  /*  Case "typedef_direct_declarator_paren_id" has to be dropped  */
  typedef_direct_declarator_paren_noid
  ;

typedef_direct_declarator_paren_id:
  typedef_name
     {
       /*  We have a redefinition of a typedef. If old stems from
	*   the same block, this is an error.
	*/
       $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1,
		    $1->JPACN_Start, $1->JPACN_End);
     }
  | typedef_direct_declarator_paren_id declarator_app
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
      _jsparschknode(ParserJPA, $$);
    }
  ;

typedef_direct_declarator_paren_noid:
  typedef_direct_declarator_paren_noid declarator_app
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  | '(' typedef_declarator_paren2 ')'
    {
      $$ = $2;
      $$->JPACN_Start = $1;
      $$->JPACN_End = $3;
      _jsparschknode(ParserJPA, $$);
    }
  ;

typedef_declarator_paren2:
  typedef_direct_declarator_paren_id
  | typedef_direct_declarator_paren_noid
  /*  Case "pointer typedef_direct_declarator_paren_id" has to be dropped  */
  | pointer typedef_direct_declarator_paren_noid
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

declarator_app:
  declarator_app_nokr
  /*  Should only appear when defining a function in K&R style  */
  | '(' PUSH_CMPBLOCK identifier_list POP_BLOCK ')'
    {
      $$ = MKCNODE(NULL, JPACT_FUNCCALL, $3, $1, $5);
    }
  ;

declarator_app_nokr:
  declarator_app_nofunc
  | declarator_app_func
  ;

declarator_app_nofunc:
  '[' constant_expression ']'
    { $$ = MKCNODE(NULL, JPACT_VECTOR, $2, $1, $3); }
  |  '[' ']'
    { $$ = MKCNODE(NULL, JPACT_VECTOR, NULL, $1, $2); }
  ;

declarator_app_func:
  '(' PUSH_CMPBLOCK parameter_type_list POP_BLOCK ')'
    { $$ = MKCNODE(NULL, JPACT_FUNCCALL, $3, $1, $5); }
  |  '(' ')'
    { $$ = MKCNODE(NULL, JPACT_FUNCCALL, NULL, $1, $2); }
  ;

pointer: /* **  list  ** */
  '*' type_qualifier_list_opt
    { $$ = _jsparslinktail($2, MKCNODE(NULL, JPACT_POINTER, NULL, $1, $1)); }
  | '*' type_qualifier_list_opt pointer
    {
      $$ = _jsparslinktail($3, _jsparslinktail($2, MKCNODE(NULL, JPACT_POINTER,
							   NULL, $1, $1)));
    }
  ;

any_declarator:
  declarator
  | typedef_declarator
  ;

parameter_type_list: /* **  list  ** */
  parameter_list
  /*  Allow ellipsis only at the end of an argument list  */
  | parameter_list ',' TK_ELLIPSIS
    {
      $$ = _jsparslinktail($1, MKCNODE(NULL, JPACT_PARDECL,
				       MKCNODE(NULL, JPACT_ELLIPSIS,
					       NULL, $3, $3), $3, $3));
    }
  ;

parameter_list: /* **  list  ** */
  parameter_declaration
  | parameter_list ',' parameter_declaration
    { $$ = _jsparslinktail($1, $3); }
  ;

parameter_declaration:
  declaration_specifiers_notype declarator
    {
      $2->JPACN_Next = $1;
      $$ = MKCNODE(NULL, JPACT_PARDECL, $2,
		   _jsparsgettail($1)->JPACN_Start, $2->JPACN_End);
    }
  | declaration_specifiers_withtype any_declarator_param2
    {
      $2->JPACN_Next = $1;
      $$ = MKCNODE(NULL, JPACT_PARDECL, $2,
		   _jsparsgettail($1)->JPACN_Start, $2->JPACN_End);
    }
  | declaration_specifiers_notype abstract_declarator
    {
      $2->JPACN_Next = $1;
      $$ = MKCNODE(NULL, JPACT_PARDECL, $2,
		   _jsparsgettail($1)->JPACN_Start, $2->JPACN_End);
    }
  | declaration_specifiers_withtype abstract_declarator
    {
      $2->JPACN_Next = $1;
      $$ = MKCNODE(NULL, JPACT_PARDECL, $2,
		   _jsparsgettail($1)->JPACN_Start, $2->JPACN_End);
    }
  | declaration_specifiers
    {
      $$ = MKCNODE(NULL, JPACT_PARDECL, $1, _jsparsgettail($1)->JPACN_Start, $1->JPACN_End);
    }
  ;

any_declarator_param2:
  any_declarator_param
    {
      if(putNewIdent(ParserJPA, ParserQueue, $1->JPACN_Sub,
		     TK_IDENTIFIER) == NULL)
	YYABORT;

      $1->JPACN_Sub->JPACN_Type = JPACT_IDENT; /*  in case we have a redefinition  */
    }
  ;

any_declarator_param:
  declarator
  | typedef_declarator_noparen
  ;

identifier_list: /* **  list  ** */
  kr_declaration
  | identifier_list ',' kr_declaration
    { $$ = _jsparslinktail($1, $3); }
  ;

kr_declaration:
  identifier_or_enumconst
    {
      if(putNewIdent(ParserJPA, ParserQueue, $1, TK_IDENTIFIER) == NULL)
	  YYABORT;

      $1->JPACN_Type = JPACT_IDENT; /*  in case we have a redefinition  */
    }
  ;

initializer:
  assignment_expression
  | '{' initializer_list '}'
    { $$ = MKCNODE(NULL, JPACT_BLOCK, $2, $1, $3); }
  | '{' initializer_list ',' '}'
    {
      $$ = MKCNODE(NULL, JPACT_BLOCK, $2, $1, $4);

      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $$,
		     "Superfluous comma at the end of initializers", 0);
    }
  ;

initializer_list: /* **  list  ** */
  initializer
  | initializer_list ',' initializer
    { $$ = _jsparslinktail($1, $3); }
  ;

/*  For casts and "sizeof"  */
type_name: /* **  list  ** */
  specifier_qualifier_list abstract_declarator
    { $$ = _jsparslinktail($2, $1); }
  | specifier_qualifier_list
  ;

abstract_declarator:
  pointer
    {
      $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1, _jsparsgettail($1)->JPACN_Start, $1->JPACN_End);
    }
  | pointer direct_abstract_declarator
    {
      $2->JPACN_Sub = _jsparslinktail($2->JPACN_Sub, $1);
      $2->JPACN_Start = _jsparsgettail($1)->JPACN_Start;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  | direct_abstract_declarator
  ;

direct_abstract_declarator:
  direct_abstract_declarator declarator_app /*_nofunc*/
    {
      $1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $2);
      $1->JPACN_End = $2->JPACN_End;
    }
  | declarator_app_func
    {
      $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1, $1->JPACN_Start, $1->JPACN_End);
    }
  | declarator_app_nofunc
    {
      $$ = MKCNODE(NULL, JPACT_DECLARATOR, $1, $1->JPACN_Start, $1->JPACN_End);
    }
  | '(' abstract_declarator ')'
     {
       $2->JPACN_Start = $1;
       $2->JPACN_End = $3;
       $$ = $2;
      _jsparschknode(ParserJPA, $$);
     }
  ;

/* ****************************** */
/* ***  Rules for statements  *** */
/* ****************************** */
statement:
  labeled_statement
  | unlabeled_statement
  | error_stmt
  ;

/*  This version with left associative labels prevents deeply nested trees
 *   in case of many "case" statements in front of an executable one. On
 *   some systems, yacc produces stack overflow if we use right-associative.
 */
labeled_statement:
  label_list unlabeled_statement
    {
      jsparscnode_t *n;

      if($1->JPACN_Next) {
	n = MKCNODE($2, JPACT_LIST, $1, $1->JPACN_Start, _jsparsgettail($1)->JPACN_End);
      } else {
	n = $1;
	n->JPACN_Next = $2;
      }

      $$ = MKCNODE(NULL, JPACT_LABELEDSTMT, n, n->JPACN_Start, $2->JPACN_End);
    }
  ;

label_list: /* **  list  ** */
  statement_label
  | label_list statement_label
    { $$ = _jsparslinktail($1, $2); }
  ;

statement_label:
  identifier_or_enumconst ':'
    {
      $1->JPACN_Type = JPACT_IDENT;
      $$ = MKCNODE(NULL, JPACT_LABEL, $1, $1->JPACN_Start, $1->JPACN_End);
    }
  /*  Is used instead of "typedef_name" to allow parser to uniquely
   *   parse a compound statement, otherwise we get a lot of
   *   shift/reduce conflicts. The problem is if the first statement
   *   in a compound statement after the declarations is labeled and
   *   the labeled statement is a "typedef_name". yacc needs more
   *   than one lookahead token.
   */
  | declaration_specifiers ':'
    {
      /*  "declaration_specifiers" must consist of a single typedef_name
       *   or an identifier.
       */
      if($1->JPACN_Next || $1->JPACN_Type != JPACT_TYPEDEFNAME) {
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Illegal labeled statement", 1);
	YYABORT;
      }

      $1->JPACN_Type = JPACT_IDENT;
      $$ = MKCNODE(NULL, JPACT_LABEL, $1, $1->JPACN_Start, $1->JPACN_End);
    }
  | TK_KW_CASE constant_expression ':'
    { $$ = MKCNODE(NULL, JPACT_CASE, $2, $1, $3); }
  | TK_KW_DEFAULT ':'
    { $$ = MKCNODE(NULL, JPACT_DEFAULT, NULL, $1, $2); }
  | pragma_line /*  pragma lines that prefix statements  */
  ;

unlabeled_statement:
  expression_statement
  | compound_statement
  | selection_statement
  | iteration_statement
  | jump_statement
  ;

expression_statement:
  unlabeled_expression ';'
    {
      $1->JPACN_End = $2;
    }
  | ';' /*  empty statement  */
    {
      /*
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, NULL,
		       "Empty statement seen", 0);
      */

      $$ = MKCNODE(NULL, JPACT_UNARYEXPR, NULL, $1, $1);
    }
  ;

compound_statement:
  '{' PUSH_CMPBLOCK compound_statement_content '}' POP_BLOCK
    {
      $3->JPACN_Start = $1;
      $3->JPACN_End = $4;
      $$ = $3;
      _jsparschknode(ParserJPA, $$);
    }
  ;

/*  Is used in function definition to include parameters
 *   in variable list as well.
 */
compound_statement2:
  '{' compound_statement_content '}'
    {
      $2->JPACN_Start = $1;
      $2->JPACN_End = $3;
      $$ = $2;
      _jsparschknode(ParserJPA, $$);
    }
  ;

compound_statement_content:
  statement_or_declaration_list label_list_opt
    {
      /*  Unless C99 extensions are on, we check strict ANSI-C
       *   syntax: declarations come *before* executable statements
       */
      if(!(ParserJPA->JSPA_Mode & JPAM_FMIXEDDECLS)) {
	jsparscnode_t *node;
	int stmtseen = FALSE;

	for(node = $1 ; node ; node = node->JPACN_Next) {
	  if(!(node->JPACN_Type == JPACT_DECLARATION ||
	       node->JPACN_Type == JPACT_TYPEDEF)) {
	    /*  Found first executable statement  */
	    stmtseen = TRUE;
	  } else if(stmtseen) {
	    jsparsprinterr(ParserJPA, ParserQueue, node,
			   "Forbidden mixing of declarations and code", 1);
	    YYERROR;
	  }
	}
      }

      $$ = MKCNODE(NULL, JPACT_COMPSTMT, _jsparslinktail($1, $2), 0, 0);
    }
    | /*  empty compound statement  */ label_list_opt
    {
      $$ = MKCNODE(NULL, JPACT_COMPSTMT, $1, 0, 0);
    }
  ;

statement_or_declaration_list: /* **  list  ** */
  statement_or_declaration
  | statement_or_declaration_list statement_or_declaration
    { $$ = _jsparslinktail($1, $2); }
  ;

statement_or_declaration:
  statement
  | declaration
  ;

label_list_opt:
    /*  empty  */
    { $$ = NULL; }
  | label_list
    {
      jsparscnode_t *n;

      /*  A simple extension:  Labels at the end of a compound statement,
       *   without trailing ';'
       */
      if(ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE)
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		     "ANSI C forbids label at end of compound statement", 0);

      if($1->JPACN_Next) {
	n = MKCNODE(NULL, JPACT_LIST, $1, $1->JPACN_Start,
		    _jsparsgettail($1)->JPACN_End);
      } else {
	n = $1;
      }

      $$ = MKCNODE(NULL, JPACT_LABELEDSTMT, n, n->JPACN_Start, n->JPACN_End);
    }
  ;

selection_statement:
  /*  The next two rules generate our famous "if-else" shift/reduce
   *   conflict. This is fortunately our only conflict :)
   */
  TK_KW_IF '(' expression ')' statement
  /*  The following line in conjunction with the "%nonassoc" lines
   *   can be used to remove the remaining shift/reduce conflict.
   */
    %prec LOWER_THAN_ELSE
    {
      $3->JPACN_Next = $5;
      $$ = MKCNODE(NULL, JPACT_IFELSE, $3, $1, $5->JPACN_End);

      if((ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE) &&
	 STMT_IS_EMPTY($5))
        jsparsprinterr(ParserJPA, ParserQueue, $5,
		       "Empty 'if' statement body", 0);
    }
  | TK_KW_IF '(' expression ')' statement TK_KW_ELSE statement
    {
      $3->JPACN_Next = $5;
      $5->JPACN_Next = $7;
      $$ = MKCNODE(NULL, JPACT_IFELSE, $3, $1, $7->JPACN_End);

      if((ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE) &&
	 STMT_IS_EMPTY($5))
        jsparsprinterr(ParserJPA, ParserQueue, $5,
		       "Empty 'if' statement body", 0);

      if((ParserJPA->JSPA_WarnFlags & JPAM_WARNSTYLE) &&
	 STMT_IS_EMPTY($7))
        jsparsprinterr(ParserJPA, ParserQueue, $7,
		       "Empty 'else' statement body", 0);
    }
  | TK_KW_SWITCH '(' expression ')' statement
    {
      $3->JPACN_Next = $5;
      $$ = MKCNODE(NULL, JPACT_SWITCH, $3, $1, $5->JPACN_End);
    }
  ;

iteration_statement:
  TK_KW_WHILE '(' expression ')' statement
    {
      $3->JPACN_Next = $5;
      $$ = MKCNODE(NULL, JPACT_WHILE, $3, $1, $5->JPACN_End);
    }
  | TK_KW_DO statement TK_KW_WHILE '(' expression ')' ';'
    {
      $2->JPACN_Next = $5;
      $$ = MKCNODE(NULL, JPACT_DOWHILE, $2, $1, $7);
    }
  | TK_KW_FOR '(' expression_opt ';' expression_opt ';' expression_opt ')'
    statement
    {
      $3->JPACN_Next = $5;
      if($3->JPACN_Start == -1) {
	$3->JPACN_Start = $3->JPACN_End = $4;
	_jsparschknode(ParserJPA, $3);
      }

      $5->JPACN_Next = $7;
      if($5->JPACN_Start == -1) {
	$5->JPACN_Start = $5->JPACN_End = $6;
	_jsparschknode(ParserJPA, $5);
      }

      $7->JPACN_Next = $9;
      if($7->JPACN_Start == -1) {
	$7->JPACN_Start = $7->JPACN_End = $8;
	_jsparschknode(ParserJPA, $7);
      }

      $$ = MKCNODE(NULL, JPACT_FOR, $3, $1, $9->JPACN_End);
    }
  ;

expression_opt: /*  empty  */
    { $$ = MKCNODE(NULL, JPACT_UNARYEXPR, NULL, -1, -1); }
  | expression
  ;

jump_statement:
  TK_KW_GOTO any_identifier ';'
    {
      $2->JPACN_Type = JPACT_IDENT;
      $$ = MKCNODE(NULL, JPACT_GOTO, $2, $1, $3);
    }
  | TK_KW_CONTINUE ';'          { $$ = MKCNODE(NULL, JPACT_CONTINUE, NULL, $1, $2); }
  | TK_KW_BREAK ';'             { $$ = MKCNODE(NULL, JPACT_BREAK, NULL, $1, $2); }
  | TK_KW_RETURN expression ';' { $$ = MKCNODE(NULL, JPACT_RETURN, $2, $1, $3); }
  | TK_KW_RETURN ';'            { $$ = MKCNODE(NULL, JPACT_RETURN, NULL, $1, $2); }
  ;

/* ******************************* */
/* ***  Rules for expressions  *** */
/* ******************************* */
expression:
  unlabeled_expression
  /*
  | pragma_prefix unlabeled_expression
  {
    $1->JPACN_Next = $2;
    $$ = MKCNODE(NULL,JPACT_LABELEDSTMT,$1,$1->JPACN_Start,$2->JPACN_End);
  }
  */
  ;

unlabeled_expression:
  assignment_expression
  | unlabeled_expression ',' assignment_expression
    {
      if($1->JPACN_Type == JPACT_LIST) {
	$1->JPACN_Sub = _jsparslinktail($1->JPACN_Sub, $3);
	$1->JPACN_End = $3->JPACN_End;
      } else {
	$1->JPACN_Next = $3;
	$$ = MKCNODE(NULL, JPACT_LIST, $1, $1->JPACN_Start, $3->JPACN_End);
      }
    }
  ;

assignment_expression:
  conditional_expression
  | unary_expression assignment_operator assignment_expression
    {
      $2->JPACN_Sub = $1;
      $1->JPACN_Next = $3;
      $2->JPACN_Start = $1->JPACN_Start;
      $2->JPACN_End = $3->JPACN_End;
      $$ = $2;

#ifdef CONFIG_JSCPPLIB_DEBUG
      /*  Special mode: Find assignments where right hand side is
       *   a boolean expression and report it
       */
      if(Debug & (32768 << CONFIG_JSCPPLIB_DEBUG)) {
	int type;

	type = $3->JPACN_Type;

	if(type == JPACT_EQUALS  || type == JPACT_NEQUALS ||
	   type == JPACT_LOWER   || type == JPACT_GREATER ||
	   type == JPACT_LOEQUAL || type == JPACT_GREQUAL ||
	   type == JPACT_LAND    || type == JPACT_LOR     || type == JPACT_NOT)
	  jsparsprinterr(ParserJPA, ParserQueue, $3,
			 "Assignment of boolean expression", 2);
      }
#endif

      _jsparschknode(ParserJPA, $$);
    }
  ;

assignment_operator:
  '='               { $$ = MKCNODE(NULL, JPACT_ASS, NULL, $1, $1); }
  | TK_OP_MULASS    { $$ = MKCNODE(NULL, JPACT_MULASS, NULL, $1, $1); }
  | TK_OP_DIVASS    { $$ = MKCNODE(NULL, JPACT_DIVASS, NULL, $1, $1); }
  | TK_OP_MODASS    { $$ = MKCNODE(NULL, JPACT_MODASS, NULL, $1, $1); }
  | TK_OP_ADDASS    { $$ = MKCNODE(NULL, JPACT_ADDASS, NULL, $1, $1); }
  | TK_OP_SUBASS    { $$ = MKCNODE(NULL, JPACT_SUBASS, NULL, $1, $1); }
  | TK_OP_LSHIFTASS { $$ = MKCNODE(NULL, JPACT_LSHIFTASS, NULL, $1, $1); }
  | TK_OP_RSHIFTASS { $$ = MKCNODE(NULL, JPACT_RSHIFTASS, NULL, $1, $1); }
  | TK_OP_ANDASS    { $$ = MKCNODE(NULL, JPACT_ANDASS, NULL, $1, $1); }
  | TK_OP_EXORASS   { $$ = MKCNODE(NULL, JPACT_EXORASS, NULL, $1, $1); }
  | TK_OP_ORASS     { $$ = MKCNODE(NULL, JPACT_ORASS, NULL, $1, $1); }
  ;

conditional_expression:
  logical_or_expression
  | logical_or_expression '?' expression ':' conditional_expression
    {
      $3->JPACN_Next = $5;
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_CONDEVAL, $1, $1->JPACN_Start, $5->JPACN_End);
    }
  ;

constant_expression:
  conditional_expression
    {
      if(!exprIsConstant(ParserQueue, $1)) {
	jsparsprinterr(ParserJPA, ParserQueue, $1,
		       "Constant expression expected", 1);
	YYERROR;
      }
    }
  ;

logical_or_expression:
  logical_and_expression
  | logical_or_expression TK_OP_LOR logical_and_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_LOR, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

logical_and_expression:
  inclusive_or_expression
  | logical_and_expression TK_OP_LAND inclusive_or_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_LAND, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

inclusive_or_expression:
  exclusive_or_expression
  | inclusive_or_expression '|' exclusive_or_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_OR, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

exclusive_or_expression:
  and_expression
  | exclusive_or_expression '^' and_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_EXOR, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

and_expression:
  equality_expression
  | and_expression '&' equality_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_AND, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

equality_expression:
  relational_expression
  | equality_expression TK_OP_EQUALS relational_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_EQUALS, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | equality_expression TK_OP_NEQUALS relational_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_NEQUALS, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

relational_expression:
  shift_expression
  | relational_expression '<' shift_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_LOWER, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | relational_expression '>' shift_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_GREATER, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | relational_expression TK_OP_LOEQUAL shift_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_LOEQUAL, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | relational_expression TK_OP_GREQUAL shift_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_GREQUAL, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

shift_expression:
  additive_expression
  | shift_expression TK_OP_LSHIFT additive_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_LSHIFT, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | shift_expression TK_OP_RSHIFT additive_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_RSHIFT, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

additive_expression:
  multiplicative_expression
  | additive_expression '+' multiplicative_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_ADD, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | additive_expression '-' multiplicative_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_SUB, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

multiplicative_expression:
  cast_expression
  | multiplicative_expression '*' cast_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_TIMES, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | multiplicative_expression '/' cast_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_DIV, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  | multiplicative_expression '%' cast_expression
    {
      $1->JPACN_Next = $3;
      $$ = MKCNODE(NULL, JPACT_MODULO, $1, $1->JPACN_Start, $3->JPACN_End);
    }
  ;

cast_expression:
  unary_expression
  | '(' type_name ')' cast_expression
    {
      $$ = MKUNARYEXPR($4, MKCNODE(NULL, JPACT_CAST, $2, $1, $3));
      $$->JPACN_Start = $1;
      _jsparschknode(ParserJPA, $$);
    }
  ;

unary_expression:
  postfix_expression
  | TK_OP_INC unary_expression
    {
      $$ = MKUNARYEXPR($2, MKCNODE(NULL, JPACT_PREINC, NULL, $1, $1));
      $$->JPACN_Start = $1;
      _jsparschknode(ParserJPA, $$);
    }
  | TK_OP_DEC unary_expression
    {
      $$ = MKUNARYEXPR($2, MKCNODE(NULL, JPACT_PREDEC, NULL, $1, $1));
      $$->JPACN_Start = $1;
      _jsparschknode(ParserJPA, $$);
    }
  | unary_operator cast_expression
    {
      $$ = MKUNARYEXPR($2, $1);
      $$->JPACN_Start = $1->JPACN_Start;
      _jsparschknode(ParserJPA, $$);
    }
  | TK_KW_SIZEOF unary_expression
    { $$ = MKCNODE(NULL, JPACT_SIZEOF, $2, $1, $2->JPACN_End); }
  | TK_KW_SIZEOF '(' type_name ')'
    { $$ = MKCNODE(NULL, JPACT_SIZEOF, $3, $1, $4); }
  ;

unary_operator:
  '&'   { $$ = MKCNODE(NULL, JPACT_ADDRESS, NULL, $1, $1); }
  | '*' { $$ = MKCNODE(NULL, JPACT_POINTER, NULL, $1, $1); }
  | '+' { $$ = MKCNODE(NULL, JPACT_UPLUS,   NULL, $1, $1); }
  | '-' { $$ = MKCNODE(NULL, JPACT_UMINUS,  NULL, $1, $1); }
  | '~' { $$ = MKCNODE(NULL, JPACT_NOT,     NULL, $1, $1); }
  | '!' { $$ = MKCNODE(NULL, JPACT_LNOT,    NULL, $1, $1); }
  ;

postfix_expression:
  primary_expression
  | postfix_expression '[' expression ']'
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_VECTOR, $3, $2, $4));
      $$->JPACN_End = $4;
    }
  | postfix_expression '(' argument_expression_list ')'
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_FUNCCALL, $3, $2, $4));
      $$->JPACN_End = $4;
    }
  | postfix_expression '(' ')'
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL,
					   JPACT_FUNCCALL, NULL, $2, $3));
      $$->JPACN_End = $3;
    }
  | postfix_expression '.' any_identifier
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_MEMBER, $3,
					   $2, $3->JPACN_End));
      $$->JPACN_End = $3->JPACN_End;
    }
  | postfix_expression TK_OP_POINT any_identifier
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_POINTTO, $3,
					   $2, $3->JPACN_End));
      $$->JPACN_End = $3->JPACN_End;
    }
  | postfix_expression TK_OP_INC
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_POSTINC, NULL, $2, $2));
      $$->JPACN_End = $2;
    }
  | postfix_expression TK_OP_DEC
    {
      $$ = MKUNARYEXPR($1, MKCNODE(NULL, JPACT_POSTDEC, NULL, $2, $2));
      $$->JPACN_End = $2;
    }
  ;

primary_expression:
  identifier
    { $$ = MKCNODE(NULL, JPACT_UNARYEXPR, $1, $1->JPACN_Start, $1->JPACN_End); }
  | constant
    { $$ = MKCNODE(NULL, JPACT_UNARYEXPR, $1, $1->JPACN_Start, $1->JPACN_End); }
  | string
    { $$ = MKCNODE(NULL, JPACT_UNARYEXPR, $1, $1->JPACN_Start, $1->JPACN_End); }
  | '(' expression ')'
    {
	/*  Solely another pair of brackets surrounding this
         *   expression, include them but don't generate new
	 *   unary expression. This will be done if necessary
	 *   by the calls to makeUnaryExpression above
         */
	$2->JPACN_Start = $1;
	$2->JPACN_End   = $3;
	$$ = $2;
	_jsparschknode(ParserJPA, $$);
    }
  ;

argument_expression_list: /* **  list  ** */
  argument_expression
  | argument_expression_list ',' argument_expression
    { $$ = _jsparslinktail($1, $3); }
  ;

argument_expression:
  assignment_expression
  /*
  | pragma_prefix assignment_expression
  {
    $1->JPACN_Next = $2;
    $$ = MKCNODE(NULL,JPACT_LABELEDSTMT,$1,$1->JPACN_Start,$2->JPACN_End);
  }
  */
  ;

constant:
  integer_constant
  | character_constant
  | floating_constant
  | enumeration_constant
  ;

any_identifier:
  identifier_or_typedefname
  | enumeration_constant
  ;

identifier_or_typedefname:
  identifier
  | typedef_name
  ;

identifier_or_enumconst:
  identifier
  | enumeration_constant
  ;

/* *********************** */
/* ***  Input symbols  *** */
/* *********************** */
integer_constant:
  TK_INTCONST { $$ = MKCNODE(NULL, JPACT_INTCONST, NULL, $1, $1); }
  ;

character_constant:
  TK_CHARCONST    { $$ = MKCNODE(NULL, JPACT_CHARCONST, NULL, $1, $1); }
  | TK_WCHARCONST { $$ = MKCNODE(NULL, JPACT_WCHARCONST, NULL, $1, $1); }
  | TK_UCHARCONST { $$ = MKCNODE(NULL, JPACT_UCHARCONST, NULL, $1, $1); }
  ;

floating_constant:
  TK_FLOATCONST { $$ = MKCNODE(NULL, JPACT_FLOATCONST, NULL, $1, $1); }
  ;

identifier:
  TK_IDENTIFIER { $$ = MKCNODE(NULL, JPACT_IDENT, NULL, $1, $1); }
  ;

/*  Allow multiple strings so that string concatentation
 *   can be avoided by preprocessor. Reject mixing of
 *   normal strings with wide character strings.
 */
string:
  cstring
  | wstring
  | ustring
  ;

cstring:
  TK_STRING           { $$ = MKCNODE(NULL, JPACT_STRING, NULL, $1, $1);  }
  | cstring TK_STRING { $$->JPACN_End = $2; }
  ;

wstring:
  TK_WSTRING           { $$ = MKCNODE(NULL, JPACT_WSTRING, NULL, $1, $1); }
  | wstring TK_WSTRING { $$->JPACN_End = $2; }
  ;

ustring:
  TK_USTRING           { $$ = MKCNODE(NULL, JPACT_USTRING, NULL, $1, $1); }
  | ustring TK_USTRING { $$->JPACN_End = $2; }
  ;

typedef_name:
  TK_TYPEDEFNAME
  { $$ = MKCNODE(NULL, JPACT_TYPEDEFNAME, NULL, $1, $1); }
  ;

enumeration_constant:
  TK_ENUMCONST
  { $$ = MKCNODE(NULL, JPACT_ENUMCONST, NULL, $1, $1); }
  ;

/*  Start/end new name space  */
PUSH_BLOCK:    { if(_jsparspushbtype(ParserJPA, NULL, 0) == -1) YYABORT; }  ;
PUSH_CMPBLOCK: { if(_jsparspushbtype(ParserJPA, NULL, BTD_CMPSTMT) == -1) YYABORT; }  ;
POP_BLOCK:     { if(_jsparspopbtype(ParserJPA) == -1) YYABORT;  }  ;

/* ***  Error recovery  *** */
error_stmt:
  error ';'
  {
#ifdef CONFIG_JSCPPLIB_DEBUG
    if(Debug & (256 << CONFIG_JSCPPLIB_DEBUG)) {
      /*  Simple error recovery. Ignore till next ';'
       *   This program is NOT intended to parse
       *   invalid C code.
       */
      yyerrok;
      $$ = MKCNODE(NULL, JPACT_EMPTY, NULL, $2, $2);
    } else
#endif
    YYABORT;
  }
  | error TK_ERROR
  {
    /*  Error in preprocessing step, too serious to continue  */
    YYABORT;
  }
  ;

/* ***************************************** */
/* ***  Rules to recognize pragma lines  *** */
/* ***************************************** */
pragma_line:
  TK_PRAGMALINE
  ;

%%  /* ***  SECTION SUBROUTINES  *** */

/*********     GLOBAL / STATIC VARIABLES       			*********/

/* **  Known operators for routine getCToken  ** */
/*  They are assigned values greater than PTK_PARSERTOKEN,
 *   otherwise there may be conflicts between
 *   getCToken and the parser.
 */
const jsctokops_t JSParserCOperators[] = {
    { "++",   PTK_PARSERTOKEN + TK_OP_INC       },
    { "--",   PTK_PARSERTOKEN + TK_OP_DEC       },
    { "->",   PTK_PARSERTOKEN + TK_OP_POINT     },

    { "<<=",  PTK_PARSERTOKEN + TK_OP_LSHIFTASS },
    { ">>=",  PTK_PARSERTOKEN + TK_OP_RSHIFTASS },

    { "==",   PTK_PARSERTOKEN + TK_OP_EQUALS    },
    { "!=",   PTK_PARSERTOKEN + TK_OP_NEQUALS   },
    { "<=",   PTK_PARSERTOKEN + TK_OP_LOEQUAL   },
    { ">=",   PTK_PARSERTOKEN + TK_OP_GREQUAL   },

    { "<<",   PTK_PARSERTOKEN + TK_OP_LSHIFT    },
    { ">>",   PTK_PARSERTOKEN + TK_OP_RSHIFT    },
    { "&&",   PTK_PARSERTOKEN + TK_OP_LAND      },
    { "||",   PTK_PARSERTOKEN + TK_OP_LOR       },

    { "+=",   PTK_PARSERTOKEN + TK_OP_ADDASS    },
    { "-=",   PTK_PARSERTOKEN + TK_OP_SUBASS    },
    { "*=",   PTK_PARSERTOKEN + TK_OP_MULASS    },
    { "/=",   PTK_PARSERTOKEN + TK_OP_DIVASS    },
    { "%=",   PTK_PARSERTOKEN + TK_OP_MODASS    },
    { "|=",   PTK_PARSERTOKEN + TK_OP_ORASS     },
    { "&=",   PTK_PARSERTOKEN + TK_OP_ANDASS    },
    { "^=",   PTK_PARSERTOKEN + TK_OP_EXORASS   },
    { "...",  PTK_PARSERTOKEN + TK_ELLIPSIS     },
    { NULL, 0 }
};

/* ***  Table of reserved words  *** */
const jsctokops_t JSParserCKeyWords[] = {
    { "auto",     TK_KW_AUTO     }, { "break",    TK_KW_BREAK    },
    { "case",     TK_KW_CASE     }, { "char",     TK_KW_CHAR     },
    { "const",    TK_KW_CONST    }, { "continue", TK_KW_CONTINUE },
    { "default",  TK_KW_DEFAULT  }, { "do",       TK_KW_DO       },
    { "double",   TK_KW_DOUBLE   }, { "else",     TK_KW_ELSE     },
    { "enum",     TK_KW_ENUM     }, { "extern",   TK_KW_EXTERN   },
    { "float",    TK_KW_FLOAT    }, { "for",      TK_KW_FOR      },
    { "goto",     TK_KW_GOTO     }, { "if",       TK_KW_IF       },
    { "int",      TK_KW_INT      }, { "long",     TK_KW_LONG     },
    { "register", TK_KW_REGISTER }, { "return",   TK_KW_RETURN   },
    { "short",    TK_KW_SHORT    }, { "signed",   TK_KW_SIGNED   },
    { "sizeof",   TK_KW_SIZEOF   }, { "static",   TK_KW_STATIC   },
    { "struct",   TK_KW_STRUCT   }, { "switch",   TK_KW_SWITCH   },
    { "typedef",  TK_KW_TYPEDEF  }, { "union",    TK_KW_UNION    },
    { "unsigned", TK_KW_UNSIGNED }, { "void",     TK_KW_VOID     },
    { "volatile", TK_KW_VOLATILE }, { "while",    TK_KW_WHILE    },
#ifdef CONFIG_JSPARSE_TYPEOF
    { "__typeof__", TK_KW_TYPEOF },
#endif
#ifdef CONFIG_JSPARSE_CONSTANT
    { "__constant__", TK_INTCONST },
#endif
    { NULL, 0 }
};

/*JS*********************************************************************
 *   The lexical analyzer.
 *************************************************************************/

static int yylex(void)

/************************************************************************/
{
    jsparsstream_t *jpa;
    jctokfifo_t *queue;
    jsctoken_t token;
    int ignLevel, specLevel, ignBlock;
    int startSpecial, parserToken, indx;
    static const struct {
	int Old, New;
    } TransList[] = {
	/*  Basic input tokens that have different values
	 *   for the preprocessor and the parser.
	 */
	{ CTK_NUMBER,     TK_INTCONST   },
	{ CTK_INUMBER,    TK_INTCONST   },
	{ CTK_DNUMBER,    TK_FLOATCONST },

	{ CTK_CHARCONST,  TK_CHARCONST  },
	{ CTK_WCHARCONST, TK_WCHARCONST },
	{ CTK_UCHARCONST, TK_UCHARCONST },

	{ CTK_STRING,     TK_STRING     },
	{ CTK_WSTRING,    TK_WSTRING    },
	{ CTK_USTRING,    TK_USTRING    },

	{ CTK_ERROR,      TK_ERROR      }
    };

    if(ParserPrefixToken != -1) {
	parserToken = ParserPrefixToken;
	ParserPrefixToken = -1;
	return parserToken;
    }

    jpa = ParserJPA;
    queue = ParserQueue; /*  == jsqsgetqueue(ParserStream)  */

    startSpecial = -1;
    ignLevel = specLevel = ignBlock = 0;
    for( ; ; ) {
	const char *parserString;
	int jsqsmode;

	jsqsmode = (jpa->JSPA_Mode & JPAM_FEXPANDUSERFILES) ? 1 : 0;

	if((indx = jsqsget(ParserStream, jsqsmode)) == -1)
	    EGOTO(error);

	if(indx == -2) {
	    parserToken = CTK_EOF;

	} else {
	    token = queue->TKFF_Token[indx];

	    /*  Get token value  */
	    parserToken  = token.TK_Value;
	    parserString = token.TK_String;
	}

	/*  Skip blanks and empty tokens  */
	if(parserToken == PTK_EMPTY ||
	   (CTK_BLANK_START <= parserToken && parserToken <= CTK_BLANK_END))
	    continue;

	/* **************************************** */
	if(ignBlock > 0) {
	    /*  ignBlock = 0: outside special treatment.
	     *  ignBlock = 1: saw "__ignoreblock__" macro.
	     *  ignBlock = 2: saw "__ignoreblock__ {".
	     *  ignBlock > 2: within inner blocks.
	     */
	    if(ignBlock == 1 && parserToken != '{') {
		jsparsprinterr(jpa, queue, NULL,
			       "\"__ignoreblock__\" not followed by opening brace", 1);
		EGOTO(error);
	    }

	    if(parserToken == '{') {
		++ignBlock;

	    } else if(parserToken == '}') {
		if(--ignBlock == 1)
		    ignBlock = 0; /*  special treatment finished  */
	    }

	    continue;
	} else if(parserToken == PTK_IGNBLOCK) {
	    /*  start recognizing next block  */
	    ignBlock = 1;
	    continue;
	}

	/* **************************************** */
	/*  Ignore any special token sequences  */
	if(parserToken == PTK_STARTIGN) {
	    ++ignLevel;
	    continue;
	}

	if(parserToken == PTK_STARTSPC) {
	    if(++specLevel == 1)
		/*  Memorize where it started  */
		startSpecial = indx;
	    continue;
	}

	if(parserToken == PTK_ENDIGN) {
	    --ignLevel;
	    continue;
	}
	/* **************************************** */

	if(parserToken == PTK_ENDSPC) {
	    if(--specLevel == 0) {
		/*  We have seen a complete sequence, so generate a pragma line  */
		if((yylval.TV_CNode = _jsparsmkpragmaline(queue,
							  startSpecial, indx)) == NULL)
		    EGOTO(error);

		parserToken = TK_PRAGMALINE;
		goto ende;
	    }
	    continue;
	}

	/* Skip blanks and tokens enclosed in STARTIGN/ENDIGN and ignore blocks */
	if(ignLevel == 0 && specLevel == 0 && ignBlock == 0) {
	    /* **  Recognize C keywords and typedefs  ** */
	    if(parserToken == CTK_IDENT) {
		_jsparstypedef_t *bDef;
		_jsparsident_t *pIdent;

		/*  Look in table of C key words first (highest priority)  */
		if((pIdent = _jsparsgetident(parserString,
					     jpa->JSPA_KeyWordTable)) != NULL) {
		    parserToken = pIdent->JPAID_Value;
		    goto stopthebus;
		}

		/*  Scan all block definitions and look in table of typedef names  */
		for(bDef = jpa->JSPA_TypeDefs ; bDef ; bDef = bDef->JPATD_Next) {
		    if((pIdent = _jsparsgetident(parserString,
						 bDef->JPATD_HashTable)) != NULL) {
			parserToken = pIdent->JPAID_Value;
			goto stopthebus;
		    }
		}

		/*  Cast to correct parser token  */
		parserToken = TK_IDENTIFIER;

	    } else if(parserToken >= PTK_PARSERTOKEN) {
		parserToken -= PTK_PARSERTOKEN;

	    } else if(parserToken >= CTK_MINIMUM) {
		unsigned int i;

		/*  Look in list of tokens to transfer  */
		for(i = 0 ; i < (sizeof(TransList) / sizeof(TransList[0])) ; ++i) {
		    if(parserToken == TransList[i].Old) {
			parserToken = TransList[i].New; /*  cast to new value  */
			goto stopthebus;
		    }
		}
	    }

	    break;
	}
    }

 stopthebus:
    /*  Set value for parser and for yyerror handling  */
    ParserIndex = yylval.TV_Index = indx;

 ende:

#ifdef CONFIG_JSCPPLIB_DEBUG
    if(parserToken != CTK_EOF && (Debug & (128 << CONFIG_JSCPPLIB_DEBUG))) {
	jsctoken_t ptoken;
	const char *tokString;

	if((tokString = jsparstok2string(token)) != NULL)
	    fprintf(stdout, "YYLEX: %s ", tokString);
	else
	    fprintf(stdout, "YYLEX: %d ", token.TK_Value);

	ptoken.TK_Value = parserToken;
	if(parserToken >= 256)
	    /*  'parserToken' is a different namespace, so make sure
	     *   jsparstok2string interprets it correctly
	     */
	    ptoken.TK_Value += PTK_PARSERTOKEN;

	ptoken.TK_String = token.TK_String; /*  not used, but who knows?  */

	if((tokString = jsparstok2string(ptoken)) != NULL)
	    fprintf(stdout, " (-> %s):", tokString);
	else
	    fprintf(stdout, " (-> %d):", parserToken);

	if(token.TK_String)
	    fprintf(stdout, " \"%s\"\n", token.TK_String);
	else
	    fprintf(stdout, "\n");
    }
#endif

    return parserToken;
 error:
    ParserIndex = yylval.TV_Index = -1;
    return TK_ERROR;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static _jsparsident_t *putNewIdent(jsparsstream_t *jpa, jctokfifo_t *queue,
				   jsparscnode_t *node, int value)

/************************************************************************/
{
    const char *name;
    _jsparstypedef_t *b;
    _jsparsident_t *pIdent;
    int indx;

    indx = node->JPACN_IndexPP;
    name = queue->TKFF_Token[indx].TK_String;

    b = jpa->JSPA_TypeDefs;
    if(value == TK_ENUMCONST) {
	/*  Recognize enumeration constants. Must be visible in
	 *   surrounding compound statement.
	 */
	for( ; b ; b = b->JPATD_Next)
	    if(b->JPATD_Flags & BTD_CMPSTMT) break;
    }

#ifdef XDEBUG
    {
	int count;
	_jsparstypedef_t *bDef;

	/*  Count nesting  */
	for(count = 0, bDef = jpa->JSPA_TypeDefs ; bDef ; bDef = bDef->JPATD_Next)
	    count++;
	if(value == TK_TYPEDEFNAME)
	    printf("%d: NEW typedef \"%s\"\n", count, name);
	else if(value == TK_IDENTIFIER)
	    printf("%d: NEW identifier \"%s\"\n", count, name);
	else if(value == TK_ENUMCONST)
	    printf("%d: NEW enum const \"%s\"\n", count, name);
    }
#endif

    pIdent = _jsparsputident(name, value, b->JPATD_HashTable);

    if(pIdent) {
	int ifpos;

	if((ifpos = jctokindfpos(queue)) != -1) {
	    /*  Memorize where this thing was defined  */
	    pIdent->JPAID_File = queue->TKFF_Token[indx + ifpos].TK_String;
	    pIdent->JPAID_Line = queue->TKFF_Token[indx + ifpos].TK_Value;
	} else {
	    pIdent->JPAID_File = NULL;
	    pIdent->JPAID_Line = 0;
	}
    }

    if(jpa->JSPA_WarnFlags & JPAM_WARNVARS) {
	/*  Look if this identifier is defined in an outer scope  */
	for(b = b->JPATD_Next ; b ; b = b->JPATD_Next) {
	    _jsparsident_t *oIdent;

	    if((oIdent = _jsparsgetident(name, b->JPATD_HashTable)) != NULL) {
		const char *oldType, *newType;

		oldType = oIdent->JPAID_Value == TK_TYPEDEFNAME ? "typedef" :
		    oIdent->JPAID_Value == TK_ENUMCONST ? "enumeration constant" :
		    "variable";
		newType = value == TK_TYPEDEFNAME ? "Typedef name" :
		    value == TK_ENUMCONST ? "Enumeration constant" :
		    "Variable";

		if(oIdent->JPAID_File)
		    jsparsprinterr(jpa, queue, node,
				   "%s \"%s\" shadows previous %s defined at l. %d '%s'", 0,
				   newType, name, oldType,
				   oIdent->JPAID_Line, oIdent->JPAID_File);
		else
		    jsparsprinterr(jpa, queue, node,
				   "%s \"%s\" shadows previous %s", 0,
				   newType, name, oldType);

		break;
	    }
	}
    }

    return pIdent;
}

/*JS*********************************************************************
 *   Error routine for the parser.
 *************************************************************************/

static void yyerror(char *string)

/************************************************************************/
{
    jsparsprinterr2(ParserJPA, ParserQueue, ParserIndex, "%s", 1, string);
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *jsparsgetexpression(jsparsstream_t *jpa, jsqueuestream_t *jsqs,
				   int mode)

/************************************************************************/
{
    jsparscnode_t *node;

    if(_jsparsgetnode(jpa, jsqs, &node, TK_START_EXPRESSION) == -1)
	return NULL;

    return node;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *jsparsgettype(jsparsstream_t *jpa, jsqueuestream_t *jsqs,
			     int mode)

/************************************************************************/
{
    jsparscnode_t *node;

    if(_jsparsgetnode(jpa, jsqs, &node, TK_START_TYPE) == -1)
	return NULL;

    return node;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparsgetnode(jsparsstream_t *jpa, jsqueuestream_t *jsqs,
		   jsparscnode_t **pNode, int starttoken)

/************************************************************************/
{
    int ret;

#if defined(CONFIG_JSCPPLIB_DEBUG) && YYDEBUG
    if(Debug & (2048 << CONFIG_JSCPPLIB_DEBUG))
	yydebug = 1;
#endif

    ParserJPA    = jpa;
    ParserStream = jsqs;
    ParserQueue  = jsqsgetqueue(ParserStream);

    ParserPrefixToken = starttoken;
    ret = yyparse();
    ParserPrefixToken = -1;

    *pNode = ParserNode;

    return ret != 0 ? -1 : 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int exprIsConstant(jctokfifo_t *queue, jsparscnode_t *node)

/************************************************************************/
{
#if 0 /*  FIXME: Determine if expression is a constant  */
    printf("CONSTANT\n");
    jsparsprctree(stderr, queue, node, 0, 0);
#endif

    switch(node->JPACN_Type) {
    case JPACT_IDENT:
	return FALSE;

    case JPACT_ENUMCONST: case JPACT_INTCONST: case JPACT_FLOATCONST:
    case JPACT_CHARCONST: case JPACT_WCHARCONST: case JPACT_UCHARCONST:
    case JPACT_STRING: case JPACT_WSTRING: case JPACT_USTRING:
    case JPACT_SIZEOF:
	return TRUE;

    default:
	break;
    }

    for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
	if(!exprIsConstant(queue, node))
	    return FALSE;

    return TRUE;
}

#ifdef CONFIG_JSCPPLIB_DEBUG
/*JS*********************************************************************
 *
 *************************************************************************/

const char *jsparstok2string(jsctoken_t token)

/************************************************************************/
{
#define GEN_CASEVALUE(name)    case name: return #name

    if(token.TK_Value > PTK_PARSERTOKEN)
	switch(token.TK_Value - PTK_PARSERTOKEN) {
	    GEN_CASEVALUE(TK_CHARCONST);
	    GEN_CASEVALUE(TK_ELLIPSIS);
	    GEN_CASEVALUE(TK_ENUMCONST);
	    GEN_CASEVALUE(TK_ERROR);
	    GEN_CASEVALUE(TK_FLOATCONST);
	    GEN_CASEVALUE(TK_IDENTIFIER);
	    GEN_CASEVALUE(TK_INTCONST);
	    GEN_CASEVALUE(TK_KW_AUTO);
	    GEN_CASEVALUE(TK_KW_BREAK);
	    GEN_CASEVALUE(TK_KW_CASE);
	    GEN_CASEVALUE(TK_KW_CHAR);
	    GEN_CASEVALUE(TK_KW_CONST);
	    GEN_CASEVALUE(TK_KW_CONTINUE);
	    GEN_CASEVALUE(TK_KW_DEFAULT);
	    GEN_CASEVALUE(TK_KW_DO);
	    GEN_CASEVALUE(TK_KW_DOUBLE);
	    GEN_CASEVALUE(TK_KW_ELSE);
	    GEN_CASEVALUE(TK_KW_ENUM);
	    GEN_CASEVALUE(TK_KW_EXTERN);
	    GEN_CASEVALUE(TK_KW_FLOAT);
	    GEN_CASEVALUE(TK_KW_FOR);
	    GEN_CASEVALUE(TK_KW_GOTO);
	    GEN_CASEVALUE(TK_KW_IF);
	    GEN_CASEVALUE(TK_KW_INT);
	    GEN_CASEVALUE(TK_KW_LONG);
	    GEN_CASEVALUE(TK_KW_REGISTER);
	    GEN_CASEVALUE(TK_KW_RETURN);
	    GEN_CASEVALUE(TK_KW_SHORT);
	    GEN_CASEVALUE(TK_KW_SIGNED);
	    GEN_CASEVALUE(TK_KW_SIZEOF);
	    GEN_CASEVALUE(TK_KW_STATIC);
	    GEN_CASEVALUE(TK_KW_STRUCT);
	    GEN_CASEVALUE(TK_KW_SWITCH);
	    GEN_CASEVALUE(TK_KW_TYPEDEF);
	    GEN_CASEVALUE(TK_KW_TYPEOF);
	    GEN_CASEVALUE(TK_KW_UNION);
	    GEN_CASEVALUE(TK_KW_UNSIGNED);
	    GEN_CASEVALUE(TK_KW_VOID);
	    GEN_CASEVALUE(TK_KW_VOLATILE);
	    GEN_CASEVALUE(TK_KW_WHILE);
	    GEN_CASEVALUE(TK_OP_ADDASS);
	    GEN_CASEVALUE(TK_OP_ANDASS);
	    GEN_CASEVALUE(TK_OP_DEC);
	    GEN_CASEVALUE(TK_OP_DIVASS);
	    GEN_CASEVALUE(TK_OP_EQUALS);
	    GEN_CASEVALUE(TK_OP_EXORASS);
	    GEN_CASEVALUE(TK_OP_GREQUAL);
	    GEN_CASEVALUE(TK_OP_INC);
	    GEN_CASEVALUE(TK_OP_LAND);
	    GEN_CASEVALUE(TK_OP_LOEQUAL);
	    GEN_CASEVALUE(TK_OP_LOR);
	    GEN_CASEVALUE(TK_OP_LSHIFT);
	    GEN_CASEVALUE(TK_OP_LSHIFTASS);
	    GEN_CASEVALUE(TK_OP_MODASS);
	    GEN_CASEVALUE(TK_OP_MULASS);
	    GEN_CASEVALUE(TK_OP_NEQUALS);
	    GEN_CASEVALUE(TK_OP_ORASS);
	    GEN_CASEVALUE(TK_OP_POINT);
	    GEN_CASEVALUE(TK_OP_RSHIFT);
	    GEN_CASEVALUE(TK_OP_RSHIFTASS);
	    GEN_CASEVALUE(TK_OP_SUBASS);
	    GEN_CASEVALUE(TK_STRING);
	    GEN_CASEVALUE(TK_TYPEDEFNAME);
	    GEN_CASEVALUE(TK_UCHARCONST);
	    GEN_CASEVALUE(TK_USTRING);
	    GEN_CASEVALUE(TK_WCHARCONST);
	    GEN_CASEVALUE(TK_WSTRING);
	}

    return jspptok2string(token);
}
#endif
