/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Routines to build tree from grammar.
 *
 *              Copyright (C) 1994-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
/*#define NDEBUG*/
#include <assert.h>

#include <stdarg.h>

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES                                          *********/
#if 0
/*  START-DEFINITIONS 1 */
/*   Type of tree nodes  */
typedef enum {
    JPACT_EMPTY = 0,

    /* **  For declarations  ** */

    /*  Declaration types  */
    JPACT_DECLARATION, JPACT_FUNCDEF, JPACT_PARDECL, JPACT_DECLARATOR,

    /*  Storage class specifiers  */
    JPACT_AUTO, JPACT_REGISTER, JPACT_STATIC, JPACT_EXTERN, JPACT_TYPEDEF,

    /*  Type specifiers  */
    JPACT_VOID, JPACT_CHAR, JPACT_SHORT, JPACT_INT, JPACT_LONG, JPACT_SIGNED, JPACT_UNSIGNED,
    JPACT_FLOAT, JPACT_DOUBLE, JPACT_STRUCT, JPACT_UNION, JPACT_ENUM,

    /*  Type qualifiers  */
    JPACT_CONST, JPACT_VOLATILE,

    /*  Complex types  */
    JPACT_POINTER, JPACT_VECTOR, JPACT_FUNCCALL, JPACT_ELLIPSIS,
    JPACT_SMALLINT, JPACT_TYPEOF,

    /*  Comma lists and blocks  */
    JPACT_LIST, JPACT_BLOCK,

    /* **  Statements  ** */
    JPACT_COMPSTMT, JPACT_LABELEDSTMT,
    JPACT_SWITCH, JPACT_CASE, JPACT_DEFAULT, JPACT_CONTINUE, JPACT_BREAK,
    JPACT_IFELSE, JPACT_WHILE, JPACT_DOWHILE, JPACT_FOR, JPACT_RETURN,
    JPACT_LABEL, JPACT_GOTO,

    /* **  Expressions  ** */
    /*  Basic constants  */
    JPACT_IDENT, JPACT_INTCONST, JPACT_FLOATCONST,
    JPACT_CHARCONST, JPACT_WCHARCONST, JPACT_UCHARCONST,
    JPACT_STRING, JPACT_WSTRING, JPACT_USTRING,

    /*  Special node for type names and enumeration constants  */
    JPACT_TYPEDEFNAME, JPACT_ENUMCONST,

    /*  Simple expression  */
    JPACT_UNARYEXPR,

    /*  Assignment expression  */
    JPACT_ASS, JPACT_MULASS, JPACT_DIVASS, JPACT_MODASS, JPACT_ADDASS, JPACT_SUBASS,
    JPACT_LSHIFTASS, JPACT_RSHIFTASS, JPACT_ANDASS, JPACT_EXORASS, JPACT_ORASS,

    /*  Arithmetric/logical expressions  */
    JPACT_CONDEVAL, JPACT_LOR, JPACT_LAND, JPACT_EQUALS, JPACT_NEQUALS,
    JPACT_LOWER, JPACT_GREATER, JPACT_LOEQUAL, JPACT_GREQUAL,
    JPACT_OR, JPACT_EXOR, JPACT_AND, JPACT_LSHIFT, JPACT_RSHIFT,
    JPACT_ADD, JPACT_SUB, JPACT_TIMES, JPACT_DIV, JPACT_MODULO,

    /*  Unary operators  */
    JPACT_PREINC, JPACT_PREDEC, JPACT_POSTINC, JPACT_POSTDEC,
    JPACT_ADDRESS, JPACT_UPLUS, JPACT_UMINUS, JPACT_NOT, JPACT_LNOT,
    JPACT_MEMBER, JPACT_POINTTO,
    JPACT_SIZEOF, JPACT_CAST,

    /* **  Pragma lines  ** */
    JPACT_PRAGMALINE, /* general pragma line  */
    JPACT_PRAGMALINE2, /* pragma line with correct pragma introducer seen  */
    JPACT_PRAGMACONT

#ifdef CONFIG_JSCPPLIB_DEBUG
    /*  Highest allowed type  */
    , JPACT_MAXTYPE
#endif
} jsparscnodetype_e;

/* **  Tree to store C program  ** */
typedef struct jsparscnode_t jsparscnode_t;
struct jsparscnode_t {
    jsparscnode_t    *JPACN_Next;

    jsparscnodetype_e JPACN_Type;
    jsparscnode_t    *JPACN_Sub;

    /*  Start and end index in queue  */
    unsigned int      JPACN_Start, JPACN_End;

    /*  We send the original tokens first, then the preprocessed
     *   ones. So if we have a single token like IDENT where original
     *   and preprocessed one is different, JPACN_Start points to the
     *   original one, JPACN_End to the preprocessed one
     */
#define JPACN_IndexOrig  JPACN_Start
#define JPACN_IndexPP    JPACN_End
};

#define _jsparschknode(jpa, node)		\
    (((jpa)->JSPA_Mode & JPAM_FGETORIG) ?	\
     _jsparschknode2(jpa, node) : 0)

#define jsparsflushtokens(jpa, fp) jsparsflushtokens2(jpa,		\
						      (jpa)->JSPA_Stream, fp)

typedef struct {
    jsppenviron_t JPAE_PPEnviron;
    unsigned long JPAE_PAMode;
    unsigned long JPAE_WarnOpts;

} jsparsenviron_t;

#define jsparsenvgetpp(paenv)            (&((paenv)->JPAE_PPEnviron))
#define jsparsenvdefine(paenv, string)   jsppenvdefine(jsparsenvgetpp(paenv), string)

/*  The "W;" overwrites the single "W" later on  */
#define JSPARSE_OPTSTRING         "W;" JSCPP_OPTSTRING

#define jsparsseterrcode(errnr, erroff, queue, node)	\
    jsparsseterrcode2(errnr, erroff, NULL, queue, node)
/*  END-DEFINITIONS  */
#endif

#define JSCPP_PRAGMANAME_KEYWORD           "keyword"
#define JSCPP_PRAGMANAME_BOUNDARYCHECK     "boundarycheck"

#define Prototype extern
#define IPrototype extern
/*********     PROTOTYPES                                       *********/
IPrototype int           _jsparsdopredefine(jsparsstream_t *jpa);
Prototype int            jsparsflushtokens2(jsparsstream_t *jpa,
                                            jsqueuestream_t *jsqs,
                                            FILE *fp);

Prototype void           jsparsseterrcode2(int errnr, int erroff,
                                           const char *string,
                                           jctokfifo_t *queue,
                                           jsparscnode_t *node);
Prototype void           jsparsprinterr(jsparsstream_t *jpa,
                                        jctokfifo_t *queue,
                                        jsparscnode_t *node,
                                        const char *message, int mode,
                                        ...);
Prototype void           jsparsprinterr2(jsparsstream_t *jpa,
                                         jctokfifo_t *queue,
                                         unsigned int indx,
                                         const char *message,
                                         int mode, ...);
static void              _jsparsprinterr(jsparsstream_t *jpa,
                                         jctokfifo_t *queue,
                                         unsigned int indx,
                                         const char *message,
                                         int mode, va_list ap);

static jsparscnode_t    *alloccnode(void);
static void              freecnode(jsparscnode_t *node);
/*
  static void              freecnodemem(void);
  static void              cleanupcnodemem(void);
*/
Prototype jsparscnode_t *jsparsmkcnode(jsparscnode_t *next,
                                       jsparscnodetype_e type,
                                       jsparscnode_t *sub,
                                       unsigned int start,
                                       unsigned int end,
                                       jctokfifo_t *queue);
Prototype jsparscnode_t *jsparsfreecnode(jsparscnode_t *node);

IPrototype int           _jsparschknode2(jsparsstream_t *jpa,
                                         jsparscnode_t *node);

IPrototype int           _jsparsdopragma(jsppstream_t *jps,
                                         jsctoken_t *pToken, int n,
                                         void *user);

Prototype jsparscnode_t *jsparscopycnode(jsparscnode_t *node,
                                         jctokfifo_t *queue,
                                         jctokfifo_t *destQ);
IPrototype jsparscnode_t *jsparscopyncnode(jsparscnode_t *node,
                                           jctokfifo_t *queue,
                                           jctokfifo_t *destQ);
static jsparscnode_t    *subCopy(jsparscnode_t *node, int offset,
                                 jctokfifo_t *queue);

IPrototype jsparscnode_t *_jsparslinktail(jsparscnode_t *node,
                                          jsparscnode_t *next);
IPrototype jsparscnode_t *_jsparsgettail(jsparscnode_t *node);

Prototype int            jsparsinserttokstr(jsparsstream_t *jpa,
                                            jsparscnode_t *tree,
                                            jctokfifo_t *queue,
                                            const char *strings[], int where,
                                            int mode);
Prototype int            jsparsinserttoks(jsparsstream_t *jpa,
                                          jsparscnode_t *tree,
                                          jctokfifo_t *queue,
                                          jsctoken_t *tokens, int n,
                                          int where, int mode);
static void              scanNodes(jsparscnode_t *node,
                                   unsigned int limit, int diff);

Prototype int            jsparsmatchorigtoks(jctokfifo_t *queue, int *pStart,
                                             int *pEnd);

Prototype void           jsparsprctree(FILE *fp, jctokfifo_t *queue,
                                       jsparscnode_t *node,
                                       int ident, int mode);
Prototype void           jsparsprnode(FILE *fp, jctokfifo_t *queue,
                                      jsparscnode_t *node, int mode);

Prototype void           jsparsenvclear(jsparsenviron_t *jpaenv);
Prototype int            jsparsenvhandle(jsparsenviron_t *jpaenv, int opt, const char *string);
Prototype void           jsparsenvfree(jsparsenviron_t *jpaenv);


Prototype const char     JSParsStdOptsHelp[];

/*********     GLOBAL VARIABLES                                 *********/
const char JSParsStdOptsHelp[] =
    "    -W<type>  Enable warnings. Use '-Whelp' to see a list of available\n"
    "              suboptions.\n"
    ;

static const dbgopts_t JSParseWarnOptions[] = {
    { "preproc", JPAM_WARNPREPROC, NULL,
      "Warn about garbage in preprocessor lines and empty macro arguments." },

    { "proto", JPAM_WARNPROTO, NULL,
      "Warn about missing prototype and differences in scope between\n\
declaration and definition (missing 'static')." },
    { "style", JPAM_WARNSTYLE, NULL,
      "Warn about bad style like implicit 'int' return value, K&R stuff,\n\
trailing commas in structure initializations etc." },
    { "vars", JPAM_WARNVARS, NULL,
      "Warn about redefinitions, shadowed definitions, unused variables or\n\
localizable variables." },
    { "stackuse", JPAM_WARNSTACKUSE, NULL,
      "Warn about functions with high usage of stack memory (> 100KB), i. e.\n\
too many local variables." },
    { "flow", JPAM_WARNFLOW, NULL,
      "Warn about unreachable code or unused labels." },
    { "bounds", JPAM_WARNBOUNDS, NULL,
      "Do bounds checking for fixed size arrays." },

    { "badcast", JPAM_WARNBADCAST, NULL,
      "Find bad casts like in\n\
\"char (*list)[10]; char **ptr = (char **)list;\"." },

    { NULL,          0, NULL,  NULL }
};

static const dbgopts_t JSParseExtensionOptions[] = {
    { "mixeddecls", JPAM_FMIXEDDECLS, NULL,
      "Allow C99 mixing of declarations and statements." },
    { "allowredecl", JPAM_FALLOWREDECL, NULL,
      "Allow redeclarations of struct/union/typedefs." },
    { NULL,          0, NULL,  NULL }
};

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparsdopredefine(jsparsstream_t *jpa)

/************************************************************************/
{
    jsppstream_t *jps;

    jps = jsqsgetppstream(jpa->JSPA_Stream);

#ifdef JSPARSE_USE_PREDEFINES
    if(_jsparspredefine(jpa, JSPPLanguageLevel - 2) == -1)
	EGOTO(error);
#endif

    /*  Special macro to ignore next block  */
    {
	static const jsctoken_t IgnBlockTokens[] = {
	    CTOKEN_GENSTATIC(PTK_IGNBLOCK, NULL)
	};

	if(jsppdefundefmac(jps, "__ignoreblock__", IgnBlockTokens,
			   sizeof(IgnBlockTokens) / sizeof(IgnBlockTokens[0]),
			   MAFLAG_INTERNAL) == -1)
	    EGOTO(error);
    }


#ifdef CONFIG_JSPARSE_IGNORE
    {
	/*  Definition for "__ignore__": Enclose its argument
	 *   in "STARTIGN" ... "ENDIGN". Multiple arguments
	 *   are treated as one.
	 */
	static const jsctoken_t IgnoreTokens[] = {
	    CTOKEN_GENSTATIC('(',            NULL),
	    CTOKEN_GENSTATIC(CTK_IDENT,      "x" ),
	    CTOKEN_GENSTATIC(')',            NULL),

	    CTOKEN_GENSTATIC(CTK_BLANKS,     " " ),
	    CTOKEN_GENSTATIC(PTK_STARTIGN,   NULL),
	    CTOKEN_GENSTATIC(CTK_IDENT,      "x" ),
	    CTOKEN_GENSTATIC(PTK_ENDIGN,     NULL)
	};

	/*  Special macro to protect tokens from parsing  */
	if(jsppdefundefmac(jps, "__ignore__", IgnoreTokens,
			   sizeof(IgnoreTokens) / sizeof(IgnoreTokens[0]),
			   MAFLAG_ONEARGUMENT | MAFLAG_INTERNAL) == -1)
	    EGOTO(error);
    }
#endif

#ifdef CONFIG_JSPARSE_MACRO  /*  Make ourselves visible  */
    {
	jsctoken_t token;
	char buffer[30];

	token.TK_Flags = 0;
	token.TK_Value = CTK_NUMBER;
	sprintf(buffer, "%d", CONFIG_JSPARSE_VERSION);
	token.TK_String = buffer;

	if(jsppdefundefmac(jps, CONFIG_JSPARSE_MACRO, &token, 1,
			   MAFLAG_INTERNAL) == -1)
	    EGOTO(error);
    }
#endif

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Flush all tokens from the queue.
 *************************************************************************/

int jsparsflushtokens2(jsparsstream_t *jpa, jsqueuestream_t *jsqs, FILE *fp)

/************************************************************************/
{
    jsppstream_t *jps;
    jctokfifo_t *queue;

    jps = jsqsgetppstream(jsqs);
    queue = jsqsgetqueue(jsqs);

    return jsppflushtokens(jps, queue, fp,
			   (jpa->JSPA_Mode & JPAM_FPRINTORIG) ? 1 : 0);
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsseterrcode2(int errnr, int erroff, const char *string,
                       jctokfifo_t *queue, jsparscnode_t *node)

/************************************************************************/
{
    jsppseterrcode(errnr, erroff, string, queue,
		   node ? node->JPACN_IndexPP : (unsigned int)-1);
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsprinterr(jsparsstream_t *jpa, jctokfifo_t *queue,
                    jsparscnode_t *node, const char *message, int mode, ...)

/************************************************************************/
{
    va_list ap;
    unsigned int indx;

    indx = node != NULL ? node->JPACN_IndexPP : (unsigned int)-1;

    va_start(ap, mode);
    _jsparsprinterr(jpa, queue, indx, message, mode, ap);
    va_end(ap);
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsprinterr2(jsparsstream_t *jpa, jctokfifo_t *queue,
                     unsigned int indx, const char *message, int mode, ...)

/************************************************************************/
{
    va_list ap;

    va_start(ap, mode);
    _jsparsprinterr(jpa, queue, indx, message, mode, ap);
    va_end(ap);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static void _jsparsprinterr(jsparsstream_t *jpa, jctokfifo_t *queue,
                            unsigned int indx, const char *message, int mode,
                            va_list ap)

/************************************************************************/
{
    char buffer[2048];
    unsigned int off, pos;

    off = sprintf(buffer, "%s", (mode & 3) == 0 ? "WARNING" :
		  (mode & 3) == 1 ? "ERROR" : "DEBUG");

    if(queue != NULL && indx != (unsigned int)-1 &&
       (pos = jctokindfpos(queue)) != (unsigned int)-1) {
	pos += indx;

	off += sprintf(buffer + off, " l. %d \"%s\"",
		       queue->TKFF_Token[pos].TK_Value,
		       queue->TKFF_Token[pos].TK_String);
    }

    buffer[off++] = ':';
    buffer[off++] = ' ';

    /*  Print error message  */
    off += vsprintf(buffer + off, message, ap);

    buffer[off++] = '.';
    buffer[off++] = '\n';
    buffer[off++] = '\0';

    if(jpa->JSPA_PrintError)
	jpa->JSPA_PrintError(jpa->JSPA_User, buffer);

    return;
}

/*JS*********************************************************************
 *   Generate allocation routine
 *************************************************************************/

#define GENMALLOC_TYPE          jsparscnode_t
#define GENMALLOC_NAMEALLOC     alloccnode
#define GENMALLOC_NAMEFREE      freecnode
/*
  #define GENMALLOC_NAMEFREEMEM   freecnodemem
  #define GENMALLOC_NAMECLEANMEM  cleanupcnodemem
*/
#define GENMALLOC_MEMSIZE       200000
#define GENMALLOC_STATIC

#include "genmalloc.h"

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *jsparsmkcnode(jsparscnode_t *next,
                             jsparscnodetype_e type,
                             jsparscnode_t *sub,
                             unsigned int start, unsigned int end,
                             jctokfifo_t *queue)

/************************************************************************/
{
    jsparscnode_t *node;

    if((node = alloccnode()) == NULL)
	return NULL;

    if(queue && start != (unsigned int)-1 &&
       jctokindorig(queue) >= 0) {
	unsigned int step = jctokgetstep(queue);

	/*  Make sure we include original tokens as well  */
	while(start >= step &&
	      queue->TKFF_Token[start - step].TK_Value == PTK_EMPTY)
	    start -= step;
    }

    /*  Initialize node  */
    node->JPACN_Type  = type;
    node->JPACN_Next  = next;
    node->JPACN_Sub   = sub;
    node->JPACN_Start = start;
    node->JPACN_End   = end;

    return node;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *jsparsfreecnode(jsparscnode_t *node)

/************************************************************************/
{
    jsparscnode_t *sub;

    if(node->JPACN_Type == JPACT_EMPTY)
	/*  Was deleted before ??  */
	return node->JPACN_Next;

    /*  Clear node type  */
    node->JPACN_Type = JPACT_EMPTY;

    /*  Free subnodes  */
    sub = node->JPACN_Sub;
    while(sub != NULL)
	sub = jsparsfreecnode(sub);

    sub = node->JPACN_Next;
    freecnode(node);

    return sub;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparschknode2(jsparsstream_t *jpa, jsparscnode_t *node)

/************************************************************************/
{
    jctokfifo_t *queue;
    unsigned int step;

    queue = jsqsgetqueue(jpa->JSPA_Stream);
    step = jctokgetstep(queue);

    /*  Make sure we include original tokens as well  */
    while(node->JPACN_Start >= step &&
	  queue->TKFF_Token[node->JPACN_Start - step].TK_Value == PTK_EMPTY)
	node->JPACN_Start -= step;

    return 0;
}

/* ERROR-DEFINITIONS from _jsparsdopragma label _ERR_JSPARSDOPRAGMA
   Invalid syntax in 'keyword' pragma line
   Invalid syntax in 'boundarycheck' pragma line
   Missing 'array' or 'index' clause in 'boundarycheck' pragma line
*/

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparsdopragma(jsppstream_t *jps, jsctoken_t *pToken, int n, void *user)

/************************************************************************/
{
#define JPA         ((jsparsstream_t *)user)

    /*  Some pragmas are recognized here since it is easier  */
    if(tkStrcmp(pToken->TK_String, JSCPP_PRAGMANAME_KEYWORD) == 0) {
	const char *save[2];
	unsigned int i;

	/*  Find two identifiers  */
	for(i = 0 ; i < 2 ; i++) {
	    /*  Skip blank token  */
	    do {
		pToken++;
		n--;

	    } while(n > 0 && CTK_BLANK_START <= pToken->TK_Value &&
		    pToken->TK_Value <= CTK_BLANK_END);

	    if(n <= 0 || pToken->TK_Value != CTK_IDENT) {
		jsparsseterrcode(_ERR_JSPARSDOPRAGMA, 0, NULL, NULL);
		EGOTO(error);
	    }

	    save[i] = pToken->TK_String;
	}

	if(jsparsaddkeyword(JPA, save[0], save[1]) == -1)
	    EGOTO(error);

	/*  Eat up  */
	return 0;
    }

    if(tkStrcmp(pToken->TK_String, JSCPP_PRAGMANAME_BOUNDARYCHECK) == 0) {
	_jsparsboundchk_t *pboundchk;
	const char *name;
	unsigned int indexes[2], state, flag;

	flag = 0;
	name = NULL;
	indexes[0] = indexes[1] = -1;
	for(state = 0 ; ; ) {
	    /*  Skip blank token  */
	    do {
		pToken++;
		n--;

	    } while(n > 0 && CTK_BLANK_START <= pToken->TK_Value &&
		    pToken->TK_Value <= CTK_BLANK_END);

	    if(n <= 0)
		break;

	    switch(state) {
	    case 0: /* **  Looking for function name  ** */
		if(pToken->TK_Value != CTK_IDENT)
		    EGOTO(syntaxError);

		name = pToken->TK_String;
		state = 1;
		break;

	    case 1: /* **  Looking for "array" or "index"  ** */
		if(pToken->TK_Value != CTK_IDENT)
		    EGOTO(syntaxError);

		if(strcmp(pToken->TK_String, "array") == 0)
		    state = 2;
		else if(strcmp(pToken->TK_String, "index") == 0)
		    state = 3;
		else if(strcmp(pToken->TK_String, "reserve1") == 0)
		    flag = 1;
		else
		    EGOTO(syntaxError);

		break;
	    case 2: case 3:
		if(pToken->TK_Value != CTK_NUMBER && pToken->TK_Value != CTK_INUMBER)
		    EGOTO(syntaxError);

		indexes[state == 2 ? 0 : 1] = strtol(pToken->TK_String, NULL, 10);
		state = 1;
		break;
	    }
	}

	if(name == NULL ||
	   indexes[0] == (unsigned int)-1 ||
	   indexes[1] == (unsigned int)-1) {
	    jsparsseterrcode(_ERR_JSPARSDOPRAGMA, 2, NULL, NULL);
	    EGOTO(error);
	}

	if(JPA->JSPA_NBoundChkPragma >= JPA->JSPA_MBoundChkPragma) {
	    REALLOC_RESIZE(JPA->JSPA_MBoundChkPragma, 10);
	    if((JPA->JSPA_BoundChkPragma = (_jsparsboundchk_t *)
		realloc(JPA->JSPA_BoundChkPragma, JPA->JSPA_MBoundChkPragma *
			sizeof(*(JPA->JSPA_BoundChkPragma)))) == NULL)
		EGOTO(error);
	}

	pboundchk = &(JPA->JSPA_BoundChkPragma[JPA->JSPA_NBoundChkPragma]);
	pboundchk->JPABC_Flag = flag;
	if((pboundchk->JPABC_Name = strdup(name)) == NULL) EGOTO(error);
	pboundchk->JPABC_ArrayArg = indexes[0];
	pboundchk->JPABC_IndexArg = indexes[1];

	JPA->JSPA_NBoundChkPragma++;

	/*  Eat up  */
	return 0;
    }

#undef JPA

    /*   Preserve in output, bracketized in PTK_STARTSPC, PTK_ENDSPC and
     *    without interpreting macros
     */
    return 4;
 syntaxError:
    jsparsseterrcode(_ERR_JSPARSDOPRAGMA, 1, NULL, NULL);
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparscnode_t *jsparscopycnode(jsparscnode_t *node,
                               jctokfifo_t *queue, jctokfifo_t *destQ)

/************************************************************************/
{
    unsigned int newStart, step;

    step = jctokgetstep(queue);

    if(destQ == queue)
	/*  Only copying nodes on the same queue, no need to copy the tokens  */
	newStart = node->JPACN_Start;
    /*  Copy all needed tokens to destination queue  */
    else if((newStart = jctokwriten(destQ, &queue->TKFF_Token[node->JPACN_Start],
				    node->JPACN_End + step - node->JPACN_Start,
				    -1, JCTOK_MODE_INSERT)) == (unsigned int)-1)
	return NULL;

    return subCopy(node, newStart - node->JPACN_Start, destQ);
}

/*JS*********************************************************************
 *   Similar to copyCNode, but traverses also CN_Next field to copy a
 *    sequence of nodes.
 *************************************************************************/

jsparscnode_t *jsparscopyncnode(jsparscnode_t *node,
                                jctokfifo_t *queue, jctokfifo_t *destQ)

/************************************************************************/
{
    unsigned int start, end, newStart, step;
    jsparscnode_t *res, *o, *n;

    step = jctokgetstep(queue);

    /*  Compute token range  */
    start = node->JPACN_Start;
    end = node->JPACN_End;
    if(start == (unsigned int)-1) {
	/*  This is a generated node that has no tokens  */
	newStart = start;
	destQ = NULL;

    } else {
	for(o = node->JPACN_Next ; o ; o = o->JPACN_Next)
	    if(o->JPACN_Start < start)
		start = o->JPACN_Start;
	    else if(o->JPACN_End > end)
		end = o->JPACN_End;

	if(destQ == queue)
	    /*  Only copying nodes on the same queue, no need to copy the tokens  */
	    newStart = start;
	/*  Copy all needed tokens to destination queue  */
	else if((newStart = jctokwriten(destQ, &queue->TKFF_Token[start],
					end + step - start,
					-1, JCTOK_MODE_INSERT)) == (unsigned int)-1)
	    return NULL;
    }

    /*  Copy head node and loop  */
    res = subCopy(node, newStart - start, destQ);

    for(o = node->JPACN_Next, n = res ; o ; o = o->JPACN_Next, n = n->JPACN_Next)
	n->JPACN_Next = subCopy(o, newStart - start, destQ);

    return res;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jsparscnode_t *subCopy(jsparscnode_t *node, int offset,
                              jctokfifo_t *queue)

/************************************************************************/
{
    jsparscnode_t *newNode, *node2, *new2;

#ifdef CONFIG_JSCPPLIB_DEBUG
    if(node->JPACN_Type == JPACT_EMPTY) {
	fprintf(stderr, "ERROR subCopy: Empty node encountered!\n");
	return NULL;
    }
#endif

    /*  Copy first subnode, then main one  */
    if((node2 = node->JPACN_Sub) == NULL)
	new2 = NULL;
    else if((new2 = subCopy(node2, offset, queue)) == NULL)
	EGOTO(error);

    if((newNode = jsparsmkcnode(NULL, node->JPACN_Type, new2,
				node->JPACN_Start + offset,
				node->JPACN_End + offset,
				queue)) == NULL)
	EGOTO(error);

    if(node2)
	while((node2 = node2->JPACN_Next)) {
	    if((new2->JPACN_Next = subCopy(node2, offset, queue)) == NULL)
		EGOTO(error);

	    new2 = new2->JPACN_Next;
	}

    return newNode;
 error:
    return NULL;
}

/*JS*********************************************************************
 *   Link node 'next' at the end of the list starting with the node 'node'.
 *    Returns the new head.
 *************************************************************************/

jsparscnode_t *_jsparslinktail(jsparscnode_t *node, jsparscnode_t *next)

/************************************************************************/
{
    jsparscnode_t *n;

    if(node == NULL) return next;

    n = node;
    while(n->JPACN_Next) n = n->JPACN_Next;
    n->JPACN_Next = next;

    return node;
}

/*JS*********************************************************************
 *   Find the last node in the linked list.
 *************************************************************************/

jsparscnode_t *_jsparsgettail(jsparscnode_t *node)

/************************************************************************/
{
    if(node == NULL) {
#ifdef CONFIG_JSCPPLIB_DEBUG
	fprintf(stderr, "ERROR in getTail: Empty list!\n");
#endif
	return NULL;
    }

    while(node->JPACN_Next)
	node = node->JPACN_Next;

    return node;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsinserttokstr(jsparsstream_t *jpa, jsparscnode_t *tree,
                       jctokfifo_t *queue, const char *strings[],
                       int where, int mode)

/************************************************************************/
{
    jsctoken_t *tokens;
    unsigned int i, step;

    step = jctokgetstep(queue);
    if((tokens = (jsctoken_t *)malloc(step * sizeof(*tokens))) == NULL)
	EGOTO(error);

    for(i = 0 ; i < step ; ++i) {
	tokens[i].TK_Flags = 0;
	if((tokens[i].TK_String = strings[i]) != NULL)
	    tokens[i].TK_Value = CTK_VANILLA;
	else
	    tokens[i].TK_Value = PTK_EMPTY;
    }

    i = jsparsinserttoks(jpa, tree, queue, tokens, step, where, mode);

    free(tokens);

    return i;
 error:
    return -1;
}

/*JS*********************************************************************
 *  Insert the tokens into the queue and also adapt the node Start and
 *   End pointers accordingly.
 *   If bit 0 of mode is set, nodes pointing to position where are not
 *   shifted upwards.
 *************************************************************************/

int jsparsinserttoks(jsparsstream_t *jpa, jsparscnode_t *tree,
                     jctokfifo_t *queue, jsctoken_t *tokens, int n,
                     int where, int mode)

/************************************************************************/
{
    if(n > 0) {
	if(jctokwriten(queue, tokens, n, where, JCTOK_MODE_INSERT) == -1)
	    EGOTO(error);

	if(where == -1)
	    /*  At the end, so there is nothing to do for us.  */
	    return 0;

    } else {
	if(where == -1)
	    EGOTO(error);

	where += -n;
	if(jctokdelete(queue, where + n, where, 0) == -1)
	    EGOTO(error);
    }

    if(mode & 1) where += n;

    if(tree && n != 0)
	scanNodes(tree, where, n);

    if(jpa && jpa->JSPA_VarTables) {
	unsigned int i;

	/*  Do not change global variable table  */
	for(i = 1 ; i < jpa->JSPA_NVarTables ; ++i) {
	    if(jpa->JSPA_VarTables[i]->JPAVT_StartIndex >= where)
		jpa->JSPA_VarTables[i]->JPAVT_StartIndex += n;

	    if(jpa->JSPA_VarTables[i]->JPAVT_EndIndex >= where)
		jpa->JSPA_VarTables[i]->JPAVT_EndIndex += n;
	}
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static void scanNodes(jsparscnode_t *node, unsigned int limit, int diff)

/************************************************************************/
{
    if(node->JPACN_Start >= limit) node->JPACN_Start += diff;
    if(node->JPACN_End >= limit)   node->JPACN_End += diff;

    /*  Scan recursively  */
    for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
	scanNodes(node, limit, diff);
}


/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsmatchorigtoks(jctokfifo_t *queue, int *pStart, int *pEnd)

/************************************************************************/
{
    int start, step, indorig;

    step = jctokgetstep(queue);
    indorig = jctokindorig(queue);
    assert(indorig != -1);

    if(pStart != NULL) {
	int ostart;

	start = *pStart;

	/*  Check for any preprocessed space tokens to the left we
	 *   might swallow
	 */
	for( ; start >= step &&
		 queue->TKFF_Token[start - step + indorig].TK_Value == PTK_EMPTY ;
	     start -= step) {
	    int ppValue = queue->TKFF_Token[start - step].TK_Value;

	    if(PTK_FIRSTINTERN <= ppValue && ppValue <= PTK_LASTINTERN)
		break;

	    /*  There can be no PTK_EMPTY tokens in the preprocessed stream,
	     *   so we only check for space tokens
	     */
	    if(!(CTK_BLANK_START <= ppValue && ppValue <= CTK_BLANK_END))
		return FALSE;
	}

	/*  Properly walk to the beginning of the preprocessed ones     */
	ostart = start;
	while(start >= step &&
	      queue->TKFF_Token[start - step].TK_Value == PTK_EMPTY)
	    start -= step;

	if(start == ostart &&
	   queue->TKFF_Token[start + indorig].TK_Value == PTK_EMPTY)
	    /*  There were no original tokens left of us (i. e. with PTK_EMPTY in
	     *   the preprocessed stream) and they are also not parallel to us.
	     */
	    return FALSE;
    }

    if(pEnd != NULL) {
	int end = *pEnd;

	for( ; (end + step) < queue->TKFF_Index &&
		 queue->TKFF_Token[end + step + indorig].TK_Value == PTK_EMPTY ;
	     end += step) {
	    int ppValue = queue->TKFF_Token[end + step].TK_Value;

	    if(PTK_FIRSTINTERN <= ppValue && ppValue <= PTK_LASTINTERN)
		break;

	    if(!(CTK_BLANK_START <= ppValue && ppValue <= CTK_BLANK_END))
		return FALSE;
	}

	*pEnd = end;
    }

    if(pStart != NULL) *pStart = start;

    return TRUE;
}

/*JS*********************************************************************
 *   mode bit 0:
 *   mode bit 1: show Start/End field divided by stepsize of queue.
 *   mode bit 2: continue printing with node->JPACN_Next field.
 *   mode bit 3: Do not print sub nodes.
 *************************************************************************/

void jsparsprctree(FILE *fp, jctokfifo_t *queue, jsparscnode_t *node,
                   int ident, int mode)

/************************************************************************/
{
    static const char *const Names[] = {
	"EMPTY",

	"DECLARATION", "FUNCDEF", "PARDECL", "DECLARATOR",
	"auto", "register", "static", "extern", "typedef",
	"void", "char", "short", "int", "long", "signed", "unsigned",
	"float", "double", "struct", "union", "enum",
	"const", "volatile",
	"pointer", "vector", "funccall", "ellipsis",
	"smallint", "typeof",

	"LIST", "BLOCK",

	"COMPSTMT", "LABELEDSTMT",
	"switch", "case", "default", "continue", "break",
	"ifelse", "while", "dowhile", "for", "return",
	"label", "goto",

	"IDENTIFIER", "intconst", "floatconst",
	"charconst", "wcharconst", "ucharconst",
	"string", "wstring", "ustring",
	"TYPEDEFNAME", "ENUMCONST",

	"UNARYEXPR",
	"ass", "mulass", "divass", "modass", "addass", "subass",
	"lshiftass", "rshiftass", "andass", "exorass", "orass",
	"condeval", "lor", "land", "equals", "nequals",
	"lower", "greater", "loequal", "grequal",
	"or", "exor", "and", "lshift", "rshift",
	"add", "sub", "times", "div", "idiv",
	"preinc", "predec", "postinc", "postdec",
	"address", "uplus", "uminus", "not", "lnot",
	"member", "pointto",
	"sizeof", "cast",
	"PRAGMA_LINE", "PRAGMA_LINE2", "PRAGMA_CONT"
    };
    int step;

    step = queue ? jctokgetstep(queue) : 1;

    do {
	if(node == NULL) {
	    fprintf(fp, "%*s%2d Empty node\n", ident, "", ident);
	    break;
	}

	if(node->JPACN_Type >= JPACT_MAXTYPE) {
	    fprintf(fp, "ILLEGAL node %d!!\n", node->JPACN_Type);

	} else {
	    const char *string;
	    int fromi, toi;

	    fprintf(fp, "%*s%2d %s", ident, "", ident, Names[node->JPACN_Type]);

	    fromi = node->JPACN_Start / step;
	    toi   = node->JPACN_End / step;
	    if(mode & 2)
		fprintf(fp, "[%d,%d]", fromi, toi);

	    if(node->JPACN_Start != (unsigned int)-1 &&
	       (fromi == toi ||
		/*  Strings may consist of multiple tokens due to string
		 *  concatenation.
		 */
		node->JPACN_Type == JPACT_STRING || node->JPACN_Type == JPACT_WSTRING) &&
	       queue &&
	       (string = queue->TKFF_Token[node->JPACN_Start].TK_String)) {
		int len;

		assert(node->JPACN_Start < queue->TKFF_Index && node->JPACN_End < queue->TKFF_Index);

		fputs(" \"", fp);
		for(len = 40 ; len-- > 0 && *string ; string++) {
		    if(*string == '\n')
			putc(' ', fp);
		    else if(*string == '\\' && string[1] == '\n') {
			string++;
			len++; /*  count correctly  */
		    } else
			putc(*string, fp);
		}
		fputs("\"\n", fp);
	    } else
		putc('\n', fp);

	    /*  Process subnodes  */
	    if(node->JPACN_Sub && !(mode & 8))
		jsparsprctree(fp, queue, node->JPACN_Sub, ident + 1, mode | 4);
	}

	node = node->JPACN_Next;
    } while(node && (mode & 4));
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsprnode(FILE *fp, jctokfifo_t *queue, jsparscnode_t *node,
                  int mode)

/************************************************************************/
{
    unsigned int i, step;

    step = jctokgetstep(queue);

    for(i = node->JPACN_Start ; i <= node->JPACN_End ; i += step)
	printCToken(fp, queue->TKFF_Token[i]);
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsenvclear(jsparsenviron_t *jpaenv)

/************************************************************************/
{
    jpaenv->JPAE_PAMode = jpaenv->JPAE_WarnOpts = 0;

    jsppenvclear(jsparsenvgetpp(jpaenv));
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jsparsenvhandle(jsparsenviron_t *jpaenv, int opt, const char *string)

/************************************************************************/
{
    switch(opt)
	{
	case 'e': /*  Switch on extensions  */
	    if(string == NULL)
		jpaenv->JPAE_PAMode = JPAM_FMIXEDDECLS | JPAM_FALLOWREDECL;
	    else if(parsedbgopts(JSParseExtensionOptions, string,
				 (long *)&(jpaenv->JPAE_PAMode), 7) == -1)
		EGOTO(error);

	    break;

	case 'W':
	    if(parsedbgopts(JSParseWarnOptions, string,
			    (long *)&(jpaenv->JPAE_WarnOpts), 3) == -1)
		EGOTO(error);

	    break;
	case 'V':
	    {
		jsparsstream_t jpa;

		if(jsparsopen2(&jpa, NULL, jpaenv, NULL, 0, 0) == NULL)
		    EGOTO(error);

		jsppshowmac(jsqsgetppstream(jpa.JSPA_Stream), stdout, -1);

		jsppshowinc(jsqsgetppstream(jpa.JSPA_Stream), stdout, 0);

		jsparsclose(&jpa, 1);
	    }
	    exit(0);
	}

    /*  Delegate  */
    if(jsppenvhandle(jsparsenvgetpp(jpaenv), opt, string) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jsparsenvfree(jsparsenviron_t *jpaenv)

/************************************************************************/
{
    jsppenvfree(jsparsenvgetpp(jpaenv));
}
