/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C preprocessor and ANSI-C parser.
 *    Part    : Predefined keywords
 *
 *              Copyright (C) 2004-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES 					*********/

#define JSPARSE_INTERNAL
#include <jsparse.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSPARSE_USE_PREDEFINES
/*  END-DEFINITIONS  */

#define IPrototype extern
/*********     PROTOTYPES					*********/
IPrototype int           _jsparspredefine(jsparsstream_t *jpa, int level);

/*********     GLOBAL /STATIC DEFINITIONS			*********/

/*
 * If bit 16 of mask is set, a new keyword definition is added,
 *  otherwise a preprocessor definition is added. If bit 17 is set,
 *  the macro definition is deferred.
 */
static const struct {
    const char *Name;
    const char *Value;
    const int   Mask;
} PreDefines[] = {

    /* *************************** GNU (i386) ***************************** */
#ifdef __GNUC__
    /*  ***  New keywords  *** */
    { "_Bool",                  "int",                             0x10007 },
    { "__signed__",             "signed",                          0x10007 },
    { "__signed",               "signed",                          0x10007 },
    { "__const__",              "const",                           0x10007 },
    { "__const",                "const",                           0x10007 },
    { "__volatile__",           "volatile",                        0x10007 },
    { "__volatile",             "volatile",                        0x10007 },
    { "__inline__",             "volatile",                        0x10007 },
    { "__inline",               "volatile",                        0x10007 },
    { "__FUNCTION__",           "__constant__",                    0x10007 },
    { "__PRETTY_FUNCTION__",    "__constant__",                    0x10007 },
    { "__func__",               "__constant__",                    0x10007 },
    { "__builtin_va_list",      "int",                             0x10007 },

    /* ***  Additional macros  *** */
    { "__extension__=__ignore__(__extension__)",           NULL,   0x04 },
    { "__extension=__ignore__(__extension)",               NULL,   0x04 },
    { "__restrict=__ignore__(__restrict)",                 NULL,   0x04 },
    { "__asm__((arg))=__asm__(__ignore__(arg))",           NULL,   0x04 },
    { "__asm((arg))=__asm(__ignore__(arg))",               NULL,   0x04 },
    { "__volatile__((arg))=__ignore__(__volatile__(arg))", NULL,   0x04 },
    { "__volatile((arg))=__ignore__(__volatile(arg))",     NULL,   0x04 },
#endif

#ifndef __GNUC__
    /*  Other platforms also use __attribute__   */
    { "__attribute__(arg)=__ignore__(__attribute__(arg))", NULL,   0x07 },
#endif

    /* *****************************  Windows  ***************************** */
#ifdef _WIN32
    /*  It doesn't really matter to what we define this, at least our
     *   parser should swallow the code
     */

    /* ***  Additional macros  *** */
    { "__asm=__ignore__(__asm) __ignoreblock__",           NULL,   0x07    },
    { "_asm=__ignore__(_asm)  __ignoreblock__",            NULL,   0x07    },
    { "__based(arg)=__ignore__(__based(arg))",             NULL,   0x07    },
    { "__cdecl=__ignore__(__cdecl)",                       NULL,   0x07    },
    { "_cdecl=__cdecl",                                    NULL,   0x07    },
    { "__declspec(arg)=__ignore__(__declspec(arg))",       NULL,   0x07    },
    { "_declspec=__declspec",                              NULL,   0x07    },
    { "__except(arg)=__ignore__(__except(arg))",           NULL,   0x07    },
    { "__far=__ignore__(__far)",                           NULL,   0x07    },
    { "__finally=__ignore__(__finally)",                   NULL,   0x07    },
    { "__loadds=__ignore__(__loadds)",                     NULL,   0x07    },
    { "__pascal=__ignore__(__pascal)",                     NULL,   0x07    },
    { "__stdcall=__ignore__(__stdcall)",                   NULL,   0x07    },
    { "_stdcall=__stdcall",                                NULL,   0x07    },
    { "__try=__ignore__(__try)",                           NULL,   0x07    },
    { "__w64=__ignore__(__w64)",                           NULL,   0x07    },

    /*  ***  New keywords  *** */
    { "__export",               "volatile",                        0x10007 },
    { "__extern",               "extern",                          0x10007 },
    { "__forceinline",          "volatile",                        0x10007 },
    { "__huge",                 "volatile",                        0x10007 },
    { "__inline",               "volatile",                        0x10007 },
    { "_inline",                "volatile",                        0x10007 },
    { "__int8",                 "char",                            0x10007 },
    { "__int16",                "short",                           0x10007 },
    { "__int32",                "int",                             0x10007 },
    { "__int64",                "long",                            0x10007 },
    { "__restrict",             "volatile",                        0x10007 },
    { "__unaligned",            "volatile",                        0x10007 },
#endif

    { NULL, NULL, 0 }
};

/*JS*********************************************************************
 *
 *************************************************************************/

int _jsparspredefine(jsparsstream_t *jpa, int level)

/************************************************************************/
{
    jsppstream_t *jps;
    int i, bmask;

    if(level < 0)
	return 0;

    bmask = 1 << level;

    jps = jsqsgetppstream(jpa->JSPA_Stream);

    for(i = 0 ; PreDefines[i].Name ; ++i) {
	if((PreDefines[i].Mask & bmask) == 0)
	    continue;

	if(PreDefines[i].Mask & 0x10000) {
	    if(jsparsaddkeyword(jpa, PreDefines[i].Name,
				PreDefines[i].Value) == -1)
		EGOTO(error);
	} else {
	    jsctoken_t token;
	    int mflags;

	    mflags = (PreDefines[i].Mask & 0x20000) ? MAFLAG_DEFEREXPANSION : 0;

	    if((token.TK_String = PreDefines[i].Value) != NULL) {
		if(jsppdefundefmac(jps, PreDefines[i].Name,
				   &token, 1, mflags | MAFLAG_INTERNAL) == -1)
		    EGOTO(error);

	    } else {
		/*  More complex definition  */
		if(jsppdefinemacro(jps, PreDefines[i].Name, mflags) == -1)
		    EGOTO(error);
	    }
	}
    }

    return 0;
 error:
    return -1;
}
