/*JS*********************************************************************
 *
 *    Program : JSCPP
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Application for preprocessor library.
 *
 *    Part    : Preprocessing routines.
 *
 *              Copyright (C) 2003-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include "jscpp.h"

/*********     DEFINES                                          *********/

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype jsppstream_t  *getPreProcStream(jsppenviron_t *jppenv,
                                          const char *name, int mode);
Prototype jsparsstream_t *getParserStream(jsparsenviron_t *ppenv,
                                          const char *name, int omode, int mode);
static void              printerror(void *user, const char *message);

Prototype int            doPreProcess(jsppenviron_t *ppenv, const char *names[],
                                      int nnames, const char *outName, int mode);
static int               doPProc(jsppstream_t *jps, FILE *fpOut);
static int               doPPOrig(jsppstream_t *jps, FILE *fpOut);

#ifdef CONFIG_JSCPPLIB_DEBUG
static int               debugPP(jsppstream_t *jps, FILE *fpOut);
static int               showPreProcess(jsppstream_t *jps, FILE *fpOut);
#endif

/*JS*********************************************************************
 *
 *************************************************************************/

jsppstream_t *getPreProcStream(jsppenviron_t *jppenv, const char *name,
                               int mode)

/************************************************************************/
{
    jsppstream_t *jps;

    if(name && strcmp(name, "-") == 0) name = NULL;

    if((jps = jsppopen2(NULL, jppenv, name, NULL, NULL, 0, mode)) == NULL)
	return NULL;

    jps->JSPP_PrintError = printerror;

    return jps;
}

/*JS*********************************************************************
 *
 *************************************************************************/

jsparsstream_t *getParserStream(jsparsenviron_t *ppenv, const char *name,
                                int omode, int mode)

/************************************************************************/
{
    jsparsstream_t *jpa;

    if((jpa = jsparsopen2(NULL, NULL, ppenv,
			  name == NULL || strcmp(name, "-") == 0 ?
			  NULL : name, omode, mode)) == NULL)
	EGOTO(error);

    jsparssetprinterr(jpa, printerror);

    return jpa;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static void printerror(void *user, const char *message)

/************************************************************************/
{
    fputs(message, stderr);
}

/*JS*********************************************************************
 *   Bit 0 of mode: outName is the output directory.
 *   Bit 1 of mode: preprocess to original tokens.
 *************************************************************************/

int doPreProcess(jsppenviron_t *ppenv, const char *names[],
                 int nnames, const char *outName, int mode)

/************************************************************************/
{
    FILE *fpOut;
    int i;

    if(outName == NULL)
	fpOut = stdout;
    else if(mode & 1)
	fpOut = NULL;
    else if((fpOut = fopen(outName, "w")) == NULL) {
	fprintf(stderr, "ERROR: Cannot open '%s' - Break.\n", outName);
	EGOTO(error);
    }

    for(i = 0 ; i < nnames ; ++i) {
	jsppstream_t *jps;

	if((jps = getPreProcStream(ppenv, names[i], 0)) == NULL)
	    EGOTO(error);

	if(mode & 2)
	    jps->JSPP_Mode |= JPPM_PRESSPACE;

	if(fpOut == NULL &&
	   (fpOut = openFile(outName, names[i])) == NULL)
	    EGOTO(error);

#ifdef CONFIG_JSCPPLIB_DEBUG
	if(Debug & 1) {
	    if(showPreProcess(jps, fpOut) == -1)
		EGOTO(error);

	} else if(Debug & 2) {
	    if(debugPP(jps, fpOut) == -1)
		EGOTO(error);

	} else
#endif
	    if(!(mode & 2)) {
		if(doPProc(jps, fpOut) == -1)
		    EGOTO(error);
	    } else {
		if(doPPOrig(jps, fpOut) == -1)
		    EGOTO(error);
	    }

#ifdef CONFIG_JSCPPLIB_DEBUG
	if(Debug & 0xf8)
	    jsppshowmac(jps, stdout, (Debug >> 3) & 31);
#endif

	if(closePreProcStream(jps, 0) == -1)
	    EGOTO(error);

	if(outName && (mode & 1)) {
	    if(fclose(fpOut) == EOF)
		EGOTO(error);

	    fpOut = NULL;
	}
    }

    if(fpOut && fclose(fpOut) == EOF)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int doPProc(jsppstream_t *jps, FILE *fpOut)

/************************************************************************/
{
    jsctoken_t token;

    while((token = jsppgettoken(jps, NULL, 0)).TK_Value != CTK_EOF) {
	if(PTK_FIRSTINTERN <= token.TK_Value && token.TK_Value <= PTK_LASTINTERN)
	    continue;

	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	/*  Ignore empty tokens  */
	if(token.TK_Value == PTK_EMPTY) continue;

	/*  Write token to output  */
	if(jspptkprint(jps, fpOut, token) == -1)
	    EGOTO(error);
    }

    if(!(jps->JSPP_Mode & JPPM_PRESSPACE))
	/*  Print newline at EOF. The internal logic deferred output
	 *   of spaces, so we have to write it here. This deletes any
	 *   unnecessary trailing spaces and fixes files that end
	 *   without a newline.
	 */
	putc('\n', fpOut);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Preprocess, but output original tokens.
 *************************************************************************/

static int doPPOrig(jsppstream_t *jps, FILE *fpOut)

/************************************************************************/
{
    jsctoken_t token, otoken;

    while((token = jsppgettoken(jps, &otoken, 0)).TK_Value != CTK_EOF) {
	/*  Ignore empty tokens (i. e. preprocessed ones, but no original ones) */
	if(otoken.TK_Value == PTK_EMPTY)
	    continue;

	/*  Write token to output  */
	if(jspptkprint(jps, fpOut, otoken) == -1)
	    EGOTO(error);
    }

    /*  Print final spaces  */
    if(jspptkprint(jps, fpOut, otoken) == -1)
	EGOTO(error);

    return 0;
 error:
    return -1;
}

#ifdef CONFIG_JSCPPLIB_DEBUG
/*JS*********************************************************************
 *
 *************************************************************************/

static int debugPP(jsppstream_t *jps, FILE *fpOut)

/************************************************************************/
{
    jsctoken_t token, otoken;

    while((token = jsppgettoken(jps, &otoken, 0)).TK_Value != CTK_EOF) {
	int len;

	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	len = fprintf(fpOut, "%3d %p ", token.TK_Value, token.TK_String);
	if(token.TK_String) {
	    fputc('\'', fpOut);
	    len += 2 + cStrFPrint(fpOut, token.TK_String, strlen(token.TK_String));
	    fputc('\'', fpOut);
	}

	fprintf(fpOut, "%*s %3d %p ", 40 - len, "", otoken.TK_Value, token.TK_String);
	if(otoken.TK_String) {
	    fputc('\'', fpOut);
	    len += 2 + cStrFPrint(fpOut, otoken.TK_String, strlen(otoken.TK_String));
	    fputc('\'', fpOut);
	}
	fputc('\n', fpOut);
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int showPreProcess(jsppstream_t *jps, FILE *fpOut)

/************************************************************************/
{
    jsctoken_t token;
    int flag;

#define S_HL_SPECIAL    "\33[35m"
#define S_HL_ORIG       "\33[34m"
#define S_HL_PREPROC    "\33[31m"
#define E_HL            "\33[0m"

    flag = 2;
    do {
	jsctoken_t oToken;

	token = jsppgettoken(jps, &oToken, 0);

	if(token.TK_Value == CTK_ERROR)
	    EGOTO(error);

	if(token.TK_Value == PTK_STARTIGN) {
	    fprintf(fpOut, S_HL_SPECIAL "STARTIGN%s!" E_HL, token.TK_String ? token.TK_String :
		    "");
	} else if(token.TK_Value == PTK_ENDIGN) {
	    fprintf(fpOut, S_HL_SPECIAL "ENDIGN%s!" E_HL, token.TK_String ? token.TK_String :
		    "");
	} else if(token.TK_Value == PTK_FIRSTINTERN) {
	    fprintf(fpOut, S_HL_SPECIAL "FIRSTINTERN%s!" E_HL, token.TK_String ? token.TK_String :
		    "");
	} else if(token.TK_Value == PTK_LASTINTERN) {
	    fprintf(fpOut, S_HL_SPECIAL "LASTINTERN%s!" E_HL, token.TK_String ? token.TK_String :
		    "");
	} else if(token.TK_Value == PTK_EMPTY) {
	    /*  0: Only original token  */
	    if(flag != 0) {
		fprintf(fpOut, S_HL_ORIG); /*  highlighting  */
		flag = 0;
	    }

	    if(oToken.TK_Value == PTK_EMPTY) {
		fprintf(fpOut, "<EMPTY>");
	    } else {
		printCToken(stdout, oToken);
	    }
	} else if(oToken.TK_Value == PTK_EMPTY) {
	    /*  1: Only preprocessed token  */
	    if(flag != 1) {
		fprintf(fpOut, S_HL_PREPROC); /*  highlighting  */
		flag = 1;
	    }

	    if(jspptkprint(jps, stdout, token) == -1)
		EGOTO(error);

	} else {
	    if(token.TK_Value != oToken.TK_Value ||
	       token.TK_String != oToken.TK_String) {
		fprintf(fpOut, S_HL_SPECIAL "???" E_HL);
	    }

	    /*  2: Both tokens are non-empty and identical  */
	    if(flag != 2) {
		fprintf(fpOut, E_HL); /*  no highlighting  */
		flag = 2;
	    }
	    printCToken(stdout, token);
	}
    } while(token.TK_Value != CTK_EOF);

    if(flag != 2)
	fprintf(fpOut, E_HL); /*  no highlighting  */

#undef S_HL_SPECIAL
#undef S_HL_ORIG
#undef S_HL_PREPROC
#undef E_HL

    return 0;
 error:
    return -1;
}
#endif
