/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Determine dependencies between pointer variables.
*
*              Copyright (C) 2004-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_PTRDEPS     "ptrdeps"

/*  Do also try to add a const for the parameter itself (not yet
 *   working)
 */
#define JSCPP_PTRDEP_ARGCONST         (1 << 0)

/*  Only try to add const to the specifier  */
#define JSCPP_PTRDEP_SPECONLY         (1 << 1)

/*  Print changed routines on stdout  */
#define JSCPP_PTRDEP_VERBOSE          (1 << 2)

#define JSCPP_PTRDEP_NUMFORMAT        (1 << 3)
/*  END-DEFINITIONS  */

typedef struct {
  jsparsvardef_t *VarDef;
  unsigned long ConstFlags;
  int PtrDepth;
} fctparam_t;

typedef struct {
  fctparam_t *Parameter;
  int MinIndirection;
} fctparref_t;

typedef struct {
  jsparsstream_t *JPA;
  jctokfifo_t *Queue;
} flowaux_t;

typedef struct {
  int Mode;
  int Depth;
} assignaux_t;

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doPtrDeps(jsparsenviron_t *ppenv,
				   const char *names[],
				   int nnames, FILE *fpOut, int mode);
static int               procFile(jsparsstream_t *jpa, FILE *fpOut, int mode);

static int               doProcPtrDepsMain(jsparsstream_t *jpa,
					   jsparscnode_t *node,
					   jctokfifo_t *queue,
					   FILE *fpOut, int mode);

static fctparref_t      *allocparref(void);
static void              freeparref(fctparref_t *pr);
static void              cleanparref(void);

static int               scanNode(jsparscnode_t *node, int level, void *user);
static int               doAssignment(jsparsstream_t *jpa, jsparscnode_t *node,
				      jctokfifo_t *queue);
static int               doUnary(jsparsstream_t *jpa, jsparscnode_t *node,
				 jctokfifo_t *queue);
static int               doFuncCall(jsparsstream_t *jpa,
				    jsparsvartab_t *vTab,
				    jsparsvardef_t *vDef,
				    jsparscnode_t *node,
				    jctokfifo_t *queue);
static int               getrhsvariable(jsparsstream_t *jpa,
					jsparsvartab_t *vTab,
					jsparscnode_t *node,
					jctokfifo_t *queue,
					jsparsvardef_t **pvDef);

/*JS*********************************************************************
*
*************************************************************************/

int doPtrDeps(jsparsenviron_t *ppenv, const char *names[], int nnames,
	      FILE *fpOut, int mode)

/************************************************************************/
{
  int i;

  for(i = 0 ; i < nnames ; ++i) {
    jsparsstream_t *jpa;
    jsppstream_t *jps;
    int ret, pmode;

    pmode = JPAM_FVARTABLES | JPAM_FSTMTFLOW;
    pmode |= JPAM_FGETORIG | JPAM_FPRINTORIG;

    /*  Get meaningful warning messages  */
    pmode |= JPAM_FGETFPOS;

    if((jpa = getParserStream(ppenv, names[i], 0, pmode)) == NULL)
      EGOTO(error);

    jps = jsqsgetppstream(jpa->JSPA_Stream);
    jps->JSPP_Mode |= JPPM_PRESSPACE;

    ret = procFile(jpa, fpOut, mode);

    if(closeParserStream(jpa, ret != 0 ? 2 : 0) == -1)
      EGOTO(error);

    if(ret == -1) EGOTO(error);
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int procFile(jsparsstream_t *jpa, FILE *fpOut, int mode)

/************************************************************************/
{
  jctokfifo_t *queue;

  queue = jsparsgetqueue(jpa);
  for(;;) {
    jsparscnode_t *node;

    if(jsparsget(jpa, &node, 0) == -1)
      EGOTO(error);

    if(node == NULL)
      break;

    if(doProcPtrDepsMain(jpa, node, queue, fpOut, mode) == -1)
      EGOTO(error);

    if(jsparsflush(jpa, node, fpOut, 0) == -1)
      EGOTO(error);
  }

  /*  Flush any trailing tokens from last jsparsget
   *   (e. g. trailing comments)
   */
  if(jsparsflushtokens(jpa, fpOut) == -1)
    EGOTO(error);

  cleanparref();

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doProcPtrDepsMain(jsparsstream_t *jpa, jsparscnode_t *node,
			     jctokfifo_t *queue, FILE *fpOut, int mode)

/************************************************************************/
{
  jsparsvartab_t *vTab;
  jsparscnode_t *head, *body, *n, *n2;
  const char *funcname;
  fctparam_t *params;
  int i, nparams, prFlag;

  if(node->JPACN_Type != JPACT_FUNCDEF)
    /*  We don't bother about global typedefs  */
    return 0;

  head = node->JPACN_Sub;
  body = head->JPACN_Next;
  if(body->JPACN_Type == JPACT_LIST ||
     body->JPACN_Type == JPACT_DECLARATION)
    /*  K&R-type function definition  */
    body = body->JPACN_Next;

  /*  To be able to include the function arguments, search for the declarator,
   *   step down to the actual function arguments
   */
  for(n = head->JPACN_Sub ; n ; n = n->JPACN_Next)
    if(n->JPACN_Type == JPACT_DECLARATOR)
      break;

  assert(n != NULL);

  if((vTab = jsparsfindvartab(jpa, n)) == NULL)
    EGOTO(error);

  n = n->JPACN_Sub;
  assert(n->JPACN_Type == JPACT_IDENT);
  funcname = queue->TKFF_Token[n->JPACN_IndexPP].TK_String;

  n = n->JPACN_Next;
  assert(n->JPACN_Type == JPACT_FUNCCALL);

  /*  Walk through all the parameters  */
  nparams = 0;
  for(n2 = n->JPACN_Sub ; n2 ; n2 = n2->JPACN_Next)
    ++nparams;

  if((params = (fctparam_t *)malloc(nparams * sizeof(*params))) == NULL)
    EGOTO(error);

  for(i = 0, n2 = n->JPACN_Sub ; n2 ; ++i, n2 = n2->JPACN_Next) {
    const char *name;
    fctparref_t *parref;

    /*  Initialize parameter description  */
    params[i].ConstFlags = 0;

    /*  Connect this variable definition to it  */
    if(doGetName(NULL, n2, &name, queue, 0) == -1)
      EGOTO(error);

    if(name == NULL)
      break; /*  ellipsis  */

    if((params[i].VarDef = jsparsgetvardef(jpa, vTab, name,
					   JVDF_TVARIABLE, 0)) == NULL)
      EGOTO(error);

    if(jsparsgetvartype3(jpa, vTab, params[i].VarDef,
			 &(params[i].PtrDepth), 0) == -1)
      EGOTO(error);

    if((parref = allocparref()) == NULL)
      EGOTO(error);

    parref->Parameter = &params[i];
    parref->MinIndirection = 0;

    params[i].VarDef->JPAVD_UserPtr = parref;
  }
  nparams = i;

  /*  Now "flow" through the code  */
  {
    flowaux_t faux;

    faux.JPA = jpa;
    faux.Queue = queue;

    jsparsflowclear(jpa);
    if(jsparsflowmark(jpa, jpa->JSPA_FlowStart, scanNode, &faux, 0) == -1)
      EGOTO(error);
  }

  /*  Check for parameters that are constant in some respects  */
  prFlag = 0;
  for(i = 0, n2 = n->JPACN_Sub ; i < nparams ; ++i, n2 = n2->JPACN_Next) {
    jsparsvardef_t *vDef;
    jsparscnode_t *n3;
    int j, hasConstSpec;

    vDef = params[i].VarDef;

    hasConstSpec = FALSE;
    for(n3 = vDef->JPAVD_Specifier ; n3 ; n3 = n3->JPACN_Next)
      if(n3->JPACN_Type == JPACT_CONST) {
	hasConstSpec = TRUE;
	break;
      }

    prFlag &= ~2;

    if(mode & JSCPP_PTRDEP_SPECONLY) {
      n3 = NULL;

      j = params[i].PtrDepth;
    } else {
      /*  Go to identifier node  */
      n3 = vDef->JPAVD_Declarator->JPACN_Sub;

      j = 0;
    }

    for( ; j <= params[i].PtrDepth ; ++j) {
      jsparscnode_t *pnode;

      /*  node before which to put the "const" keyword  */
      if((pnode = n3) != NULL) {
	int hasConst;

	/*  Advance to next node and check for following "const" keyword  */
	hasConst = FALSE;
	for(n3 = n3->JPACN_Next ; n3 ; n3 = n3->JPACN_Next) {
	  int type = n3->JPACN_Type;

	  if(type == JPACT_VECTOR || type == JPACT_POINTER ||
	     type == JPACT_IDENT)
	    /*  Next level reached  */
	    break;

	  if(type == JPACT_CONST)
	    hasConst = TRUE;
	}

	if(hasConst || pnode->JPACN_Type == JPACT_VECTOR)
	  continue;
      }

      if(!(mode & JSCPP_PTRDEP_ARGCONST) && j == 0)
	continue;

      if((params[i].ConstFlags & (1UL << j)))
	continue;

      if(j == params[i].PtrDepth) {
	if(hasConstSpec)
	  break;

	/*  Last iteration, place "const" before specifier (looks nicer)  */
	for(pnode = vDef->JPAVD_Specifier ; pnode->JPACN_Next ; )
	  pnode = pnode->JPACN_Next;
      }

      if(mode & JSCPP_PTRDEP_VERBOSE) {
	if(!(prFlag & 1))
	  printf("%s%c", funcname, (mode & JSCPP_PTRDEP_NUMFORMAT) ?
		 ' ' : '(');

	if(!(prFlag & 2)) {
	  if(mode & JSCPP_PTRDEP_NUMFORMAT) {
	    if(prFlag & 1) putchar(' ');

	    printf((mode & JSCPP_PTRDEP_SPECONLY) ? "%d" : "%d[", i);
	  } else {
	    if(prFlag & 1) putchar(',');
	    printf((mode & JSCPP_PTRDEP_SPECONLY) ? "%s" : "%s[",
		   vDef->JPAVD_Name);
	  }
	} else
	  putchar(' ');

	if(!(mode & JSCPP_PTRDEP_SPECONLY))
	  printf("%d", params[i].PtrDepth - j);

	prFlag |= 3;
      }

      if(pnode) {
	int start, end;

	start = pnode->JPACN_Start;
	end   = pnode->JPACN_End;

	if(jpa->JSPA_Mode & JPAM_FGETORIG) { ++start; ++end; }

	if(jctoksubstitute(queue, start, end, "const %o") == -1)
	  EGOTO(error);
      }
    }

    if((prFlag & 2) && !(mode & JSCPP_PTRDEP_SPECONLY))
      putchar(']');
  }

  if(prFlag & 1) {
    if(!(mode & JSCPP_PTRDEP_NUMFORMAT))
      putchar(')');
    putchar('\n');
  }

  free(params);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

#define GENMALLOC_TYPE          fctparref_t
#define GENMALLOC_NAMEALLOC     allocparref
#define GENMALLOC_NAMEFREE      freeparref
#define GENMALLOC_NAMECLEANMEM  cleanparref
#define GENMALLOC_MEMSIZE       100
#define GENMALLOC_STATIC

#include "genmalloc.h"

/*JS*********************************************************************
*
*************************************************************************/

static int scanNode(jsparscnode_t *node, int level, void *user)

/************************************************************************/
{
  int flag;

  flag = 1; /*  Scan recursively  */

#define JPA       (((flowaux_t *)user)->JPA)
#define QUEUE     (((flowaux_t *)user)->Queue)

  switch(node->JPACN_Type) {

    /*  Don't look at them again - we already did it when creating
     *   the variable definitions
     */
  case JPACT_TYPEDEF: case JPACT_PARDECL:

    /*  We are not looking at struct/union/enum names  */
  case JPACT_STRUCT: case JPACT_UNION: case JPACT_ENUM:

  case JPACT_LABEL: case JPACT_GOTO:

  case JPACT_MEMBER: case JPACT_POINTTO:
    /*  Don't look at the member names  */
    flag = 0;
    break;

  case JPACT_ASS:       case JPACT_MULASS:    case JPACT_DIVASS:
  case JPACT_MODASS:    case JPACT_ADDASS:    case JPACT_SUBASS:
  case JPACT_LSHIFTASS: case JPACT_RSHIFTASS:
  case JPACT_ANDASS:    case JPACT_EXORASS:   case JPACT_ORASS:
    if(doAssignment(JPA, node, QUEUE) == -1)
      EGOTO(error);

    break;

  case JPACT_UNARYEXPR:
    if(doUnary(JPA, node, QUEUE) == -1)
      EGOTO(error);

    break;

  default:
    break;
  }

  if(flag) {
    /*  Scan recursively  */
    for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next)
      if(scanNode(node, level, user) == -1)
	EGOTO(error);
  }

#undef JPA

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*  Determines the depth of the expression. Check that we stay in reach of
*   the current variable, i. e. "a->b->c" changes something in "a->b", not
*   in "a".
*************************************************************************/

static int processOps(void *user, jsparscnode_t *node)

/************************************************************************/
{
#define paaux ((assignaux_t *)user)

  if(paaux->Mode & 2) {
    /* We reach to something else, not pertaining our current variable */
    paaux->Mode |= 4;
    return 0;
  }

  paaux->Mode |= 1;
  if(node->JPACN_Type == JPACT_POINTER || node->JPACN_Type == JPACT_VECTOR)
    paaux->Depth++;
  else if(node->JPACN_Type == JPACT_MEMBER)
    paaux->Mode |= 2; /*  mark it if we find something else  */
  else if(node->JPACN_Type == JPACT_POINTTO) {
    paaux->Mode |= 2; /*  mark it if we find something else  */
    paaux->Depth++;
  }

#undef paaux

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doAssignment(jsparsstream_t *jpa, jsparscnode_t *node,
			jctokfifo_t *queue)

/************************************************************************/
{
  jsparsvartab_t *vTab;
  const char *name;
  jsparsvardef_t *lvDef;
  jsparscnode_t *sub;
  fctparam_t *param;
  fctparref_t *parref;
  assignaux_t aaux;
  int ret;

  /*  Check left hand side if it is a pointer or something pointed
   *   to.
   */
  if((vTab = jsparsfindvartab(jpa, node)) == NULL)
    EGOTO(error);

  sub = node->JPACN_Sub;
  assert(sub->JPACN_Type == JPACT_UNARYEXPR ||
	 sub->JPACN_Type == JPACT_DECLARATOR);

  aaux.Mode = 0;
  aaux.Depth = 0;
  if((ret = jsparsgetvarname(sub, &sub, processOps, &aaux)) == -1)
    EGOTO(error);

  name = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;

  if(ret == 0)
    return 0;

  /*  Check for simple assignments "a = ..."   */
  if(aaux.Mode == 0) {
    jsparsvardef_t *rvDef;
    fctparref_t *oparref;

    /*  Only normal assignments are interesting for direct pointer
     *   assignment
     */
    if(node->JPACN_Type != JPACT_ASS)
      return 0;

    if((lvDef = jsparsgetvardef(jpa, vTab, name, JVDF_TVARIABLE,
				0)) == NULL)
      EGOTO(error);

    sub = sub->JPACN_Next;
    if((ret = getrhsvariable(jpa, vTab, sub, queue, &rvDef)) == -1)
      EGOTO(error);

    if(ret == 0)
      /*  maybe a constant  */
      return 0;

    if((oparref = (fctparref_t *)rvDef->JPAVD_UserPtr) == NULL) {
      /*  FIXME: The following code is wrong, because we are not properly
       *   taking into account flow control: We might have something like
       *     if(<cond>) param = value;
       *   After the if statement, param might or might not refer to the
       *   parameter anymore, depending on the condition.
       */
#if 0
      /*  Left hand side no longer refers to a parameter  */
      if(lvDef->JPAVD_UserPtr) {
	freeparref((fctparref_t *)lvDef->JPAVD_UserPtr);
	lvDef->JPAVD_UserPtr = NULL;
      }
#endif
    } else {
      int ptrcnt;

      param = oparref->Parameter;

      if((parref = allocparref()) == NULL)
	EGOTO(error);

      if(jsparsgetvartype3(jpa, vTab, lvDef, &ptrcnt, 0) == -1)
	EGOTO(error);

      parref->Parameter = param;

      /*  The left hand side might have a lower pointer depth like in
       *      void dummy(char ***a)
       *      {
       *        char *p;
       *
       *        p = **a;
       *      }
       */
      parref->MinIndirection =
	MAX(MAX(oparref->MinIndirection, 1), param->PtrDepth - ptrcnt);

      /*  FIXME: We are doing too much here since we ignore flow control  */
      lvDef->JPAVD_UserPtr = parref;
    }

    return 0;
  }

#ifdef XDEBUG
  jsparsprctree(stdout, queue, sub, 0, 0);
#endif

  if(aaux.Mode & 4)
    /* We reached to something else, not pertaining our current variable */
    return 0;

  if((lvDef = jsparsgetvardef(jpa, vTab, name, JVDF_TVARIABLE, 0)) == NULL)
    EGOTO(error);

  if((parref = (fctparref_t *)lvDef->JPAVD_UserPtr) == NULL)
    /*  This assignment is not relevant for us  */
    return 0;

  param = parref->Parameter;

#ifdef XDEBUG
  printf("ASSIGN %d %d\n", aaux.Depth , parref->MinIndirection);
#endif

  if(aaux.Depth >= parref->MinIndirection)
    param->ConstFlags |= (1 << aaux.Depth);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doUnary(jsparsstream_t *jpa, jsparscnode_t *node,
		   jctokfifo_t *queue)

/************************************************************************/
{
  jsparsvartab_t *vTab;
  jsparsvardef_t *vDef;
  jsparscnode_t *sub;
  const char *name;
  fctparam_t *param;
  fctparref_t *parref;
  int ptrcnt;

  if((vTab = jsparsfindvartab(jpa, node)) == NULL)
    EGOTO(error);

  sub = node->JPACN_Sub;
  if(sub == NULL || sub->JPACN_Type != JPACT_IDENT)
    return 0; /*  a constant expression  */

  name = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;
  if((vDef = jsparsgetvardef(jpa, vTab, name, JVDF_TVARIABLE,
			     0)) == NULL) {
    /*  Variable not found, check for function call  */
    sub = sub->JPACN_Next;

    if((vDef = jsparsgetvardef(jpa, vTab, name,
			       JVDF_TFUNCTION, 0)) == NULL) {
      if(sub == NULL || sub->JPACN_Type != JPACT_FUNCCALL)
	/*  Unknown variable - better stop  */
	EGOTO(error);

      /*  Missing prototype - this is fatal for our task at hand  */
      if(jpa->JSPA_WarnFlags & JPAM_WARNPROTO)
	jsparsprinterr(jpa, queue, sub,
		       "Missing prototype for function '%s'.", 0,
		       name);

      return 0;
    }

    if(sub == NULL || sub->JPACN_Type != JPACT_FUNCCALL)
      /*  Using a function pointer - nothing for us  */
      return 0;

    return doFuncCall(jpa, vTab, vDef, sub, queue);
  }

  if((parref = (fctparref_t *)vDef->JPAVD_UserPtr) == NULL)
    /*  This assignment is not relevant for us  */
    return 0;

  param = parref->Parameter;

  if(jsparsgetvartype3(jpa, vTab, vDef, &ptrcnt, 0) == -1)
    EGOTO(error);

  if(ptrcnt == 0)
    /*  This isn't a pointer, nothing to worry about  */
    return 0;

  /*  Check for any side effects  */
  for(sub = sub->JPACN_Next ; sub ; sub = sub->JPACN_Next)
    if(sub->JPACN_Type == JPACT_PREINC ||
       sub->JPACN_Type == JPACT_POSTINC ||
       sub->JPACN_Type == JPACT_PREDEC ||
       sub->JPACN_Type == JPACT_POSTDEC) {
      int level;

      /*  A side effect to what the pointer points to  */
      level = param->PtrDepth - ptrcnt;
#ifdef XDEBUG
      printf("ASSIGN %d %d\n", level, parref->MinIndirection);
#endif
      if(level >= parref->MinIndirection)
	param->ConstFlags |= (1 << level);
    } else if(sub->JPACN_Type == JPACT_POINTER ||
	      sub->JPACN_Type == JPACT_VECTOR) {
      --ptrcnt;
    } else if(sub->JPACN_Type == JPACT_MEMBER ||
	      sub->JPACN_Type == JPACT_POINTTO) {
      /*  This leads us to something else which won't affect our
       *   current variable
       */
      break;
    }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doFuncCall(jsparsstream_t *jpa, jsparsvartab_t *vTab,
		      jsparsvardef_t *vDef, jsparscnode_t *node,
		      jctokfifo_t *queue)

/************************************************************************/
{
  jsparscnode_t *pardecl;

  pardecl = vDef->JPAVD_Declarator;
#ifdef XDEBUG
  jsparsprctree(stdout, queue, pardecl, 0, 0);
#endif

  pardecl = pardecl->JPACN_Sub;
  assert(pardecl->JPACN_Type == JPACT_IDENT);
  pardecl = pardecl->JPACN_Next;
  assert(pardecl->JPACN_Type == JPACT_FUNCCALL);
  pardecl = pardecl->JPACN_Sub;

  for(node = node->JPACN_Sub ; node ; node = node->JPACN_Next,
	pardecl = pardecl->JPACN_Next) {
    jsparsvardef_t *rvDef;
    jsparscnode_t *decl, *sub2;
    fctparam_t *param;
    fctparref_t *parref;
    long constFlag;
    int i, n, off;

    if(pardecl == NULL) {
      /*  Actual call has more arguments than prototype  */
      break;
    }

    if((i = getrhsvariable(jpa, vTab, node, queue, &rvDef)) == -1)
      EGOTO(error);

    if(i == 0)
      continue;

    if((parref = (fctparref_t *)rvDef->JPAVD_UserPtr) == NULL)
      /*  This assignment is not relevant for us  */
      continue;

    param = parref->Parameter;

    off = 0;
    for(sub2 = node->JPACN_Sub ; sub2 ; sub2 = sub2->JPACN_Next)
      if(sub2->JPACN_Type == JPACT_POINTER || sub2->JPACN_Type == JPACT_VECTOR)
	++off;
      else if(sub2->JPACN_Type == JPACT_ADDRESS)
	--off;

    assert(pardecl->JPACN_Type == JPACT_PARDECL);
    decl = pardecl->JPACN_Sub;

    /*  Check for "const" in the declarator  */
    constFlag = 0;
    for(sub2 = decl->JPACN_Sub, n = 0 ; sub2 ; sub2 = sub2->JPACN_Next)
      if(sub2->JPACN_Type == JPACT_POINTER || sub2->JPACN_Type == JPACT_VECTOR)
	++n;
      else if(sub2->JPACN_Type == JPACT_CONST)
	constFlag |= (1 << n);

    /*  Check for const in the specifier  */
    for(sub2 = decl->JPACN_Next ; sub2 ; sub2 = sub2->JPACN_Next)
      if(sub2->JPACN_Type == JPACT_CONST) {
	constFlag |= (1UL << n);
	break;
      }

#ifdef XDEBUG
    printf("\n---- %#x n %d off %d\n", constFlag, n, off);
#endif

    /*  Must start at "1" - parameters are passed as value, so the actual
     *   argument cannot be changed
     */
    for(i = 1 ; i <= n ; ++i) {
      int level;

      if(constFlag & (1UL << i))
	continue;

      level = i + off;
      if(level >= parref->MinIndirection)
	param->ConstFlags |= (1UL << level);
    }
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int getrhsvariable(jsparsstream_t *jpa, jsparsvartab_t *vTab,
			  jsparscnode_t *node, jctokfifo_t *queue,
			  jsparsvardef_t **pvDef)

/************************************************************************/
{
  jsparscnode_t *varNode;

  if(jsparsgetvarname(node, &varNode, NULL, NULL) == 0)
    return 0;

  if((*pvDef = jsparsgetvardef(jpa, vTab,
			       queue->TKFF_Token[varNode->JPACN_IndexPP].TK_String,
			       JVDF_TVARIABLE,
			       0)) == NULL)
    return 0;

  return 1;
}

#if 0
/* EXAMPLE CODE  */
char *strcpy(char *, const char *);
int dummy(char *arg1, char **arg2)
{
  char *p;

  p = arg1; /*  no const to arg1 anymore!  */
  p = arg2[1]; /*  no const to arg2 anymore!  */
  p = arg1 + 3;
  *p = 'b';
  strcpy(p, ""); /*  no const to p anymore!  */
  strcpy(arg1, ""); /*  no const to arg1 anymore!  */
  *arg1 = 'a'; /*  no const to arg1 anymore!  */
}

/*
  Method:
  - walk along flow control in a routine.
  - have an array for each parameter to the routine, representing
    the value that was passed to this parameter. The array contains
    a bitmap for the parameter. The bits represent if the value was
    changed on each level of indirection. Bit 0 is for direct change
    (i. e. parameter was assigned something). Bit 1 if a level-one
    indirection was done ("char *s" and "*s = 'X'" or "char **a" and
    "*a = ptr") and so on.
  - If at the end of scanning a bit is still '0', the "const" modifier
    can be added, e. g. "char **s" might in the best case change to
    "const char * const *s".
  - The JPAVD_UserPtr field is used to let a variable/parameter refer
    to one entry in the array. Assigning variables copies the JPAVD_UserPtr
    field if the right hand side is another variable is assigned to one
    variable
  - There are problems if indirect assignments are done like in
    "char **par" and "char *local = par[10];"
*/
#endif
