/*JS*********************************************************************
 *
 *    Program : JSSHOWTRACE
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Show 'jscpp' trace files as a tree for easy navigation
 *
 *              Copyright (C) 2007-2009 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jssubs.h>

#include <jedlib.h>

#include <jsdllist.h>

#define DEBUG

/*********     DEFINES                                          *********/
#ifdef DEBUG
#define OPTSTRING      "hd:as:r:lCD:"
#else
#define OPTSTRING      "hd:as:r:lC"
#endif

/* ******************************************************************** */
#define PROGBAR_HEIGHT                  1

#define TRACELINE_CHARS_START           "{["
#define TRACELINE_CHARS_END             "}]"
#define TRACELINE_CHARS_COMM            "#"

#define SKIP_LINE_NUMBERS

/* ******************************************************************** */
/*  Only build up the tree up to this initial level. It is later expanded
 *   dynamically when the user opens nodes.
 */
#define DEFAULT_MAXDEPTH                5

#define DEFAULT_MAXOPENDEPTH            2

/* ******************************************************************** */
/*  Initially, we read 100 lines. We dynamically adapt this to reach
 *   the given time slice below between screen updates
 */
#define DEFAULT_LINES2READ            100

/*  Try to use up ~ 0.2 seconds for scanning the file before checking
 *   for keystrokes
 */
#define DEFAULT_READLINESLICE      200000 /*  microsec  */

/* ******************************************************************** */
/* ******************************************************************** */

/*  Auxiliary structure accompanying one screen line with sub children  */
typedef struct  {
    /*
     *  Bit 0: Children are not yet scanned.
     *  Bit 1: Node is currently being set up.
     */
    int Flags;
    int Depth;

    long StartPos;
    long EndPos;
} tracecall_t;

#define tcallinit(tcall, flags, depth, start, end)	\
    ((tcall)->Flags = (flags),				\
     (tcall)->Depth = (depth),				\
     (tcall)->StartPos = (start),			\
     (tcall)->EndPos = (end))

/*  Describe one work package, i. e. a range of lines to scan and build
 *   up as a tree
 */
typedef struct {
    _jsdllnode_t Connect;

    /*  This work load is to expand the following node  */
    tracecall_t *Reference;
    long Offset;

    jedtreenode_t *Root;

    jedtreenode_t **Stack;
    int NStack, MStack;
} tracecalltree_t;

/*  Main work structure  */
typedef struct {
    FILE *File;

    int *Ranges, NRanges;
    jstrbuff_t RangeBuffer;

    long FileOffset;

    jedtree_t *Tree;
    int MaxDepth, MaxOpenDepth;

    _jsdll_t WorkList;
    tracecalltree_t *Last;
    double WorkDone;

    long StartTime[2];
    long TotalLines, TotalMillis;
    long Lines2Read;
    double ReadLineSlice;

    fstralloc_t StringPool;

    char SearchArray[1000];

} workenviron_t;

#define tclistinit(pwork)    _jsdllinit(&((pwork)->WorkList))
#define tclistempty(pwork)   _jsdllisempty(&((pwork)->WorkList))

#define tcaddtail(pwork, tctree)				\
    _jsdlladdtail(&((pwork)->WorkList), &((tctree)->Connect))
#define tcaddhead(pwork, tctree)				\
    _jsdlladdhead(&((pwork)->WorkList), &((tctree)->Connect))
#define tcaddafter(tcnode, tctree)				\
    _jsdlladdafter(&((tcnode)->Connect), &((tctree)->Connect))

#define tclistaddtail(pwork, tclist2)			\
    _jsdlllistaddtail(&((pwork)->WorkList), tclist2)
#define tclistfirstnode(pwork)						\
    _jsdllnode2utype(_jsdllgetfirstnode(&((pwork)->WorkList)), tracecalltree_t, Connect)
#define tclistremove(tctree)			\
    _jsdllremove(&((tctree)->Connect))

#define tclistloop(pwork, tctree)					\
    _jsdllloop(&((pwork)->WorkList), tctree, tracecalltree_t, Connect)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
static int               readLineRanges(int **pRanges, int *pNRanges, const char *string);
static int               getFileRange(FILE *fp, int *pFlags, long *pStart, long *pEnd);

static int               pushwork(workenviron_t *pwork, jedtreenode_t *root,
				  tracecall_t *tcall, int mode);
static tracecalltree_t  *popwork(workenviron_t *pwork);
static int               dotriggernode(void *user, jedtree_t *edtr,
				       jedtreenode_t *node);
static int               triggernode(workenviron_t *pwork, jedtreenode_t *node,
				     int mode);
static int               readLines(workenviron_t *pwork);

static jedtreenode_t    *addNode(jedtree_t *edtr, jedtreenode_t *root,
				 const char *content, int flags,
				 long startPos, long endPos,
				 int mode);
static int               searchNode(workenviron_t *pwork, int mode);
static int               searchString(void *user, jedtreenode_t *node,
				      int linenr, int level);

static tracecall_t      *tracecallalloc(void);
/*static void              tracecallfree(tracecall_t *tcall);*/
static tracecalltree_t  *tracecalltreealloc(void);
static void              tracecalltreefree(tracecalltree_t *tcall);

/*********     GLOBAL VARIABLES                                 *********/
#ifdef DEBUG
static long ReadBytes, ReadLines;
#endif

/*JS*********************************************************************
 *
 *************************************************************************/

int main(int argc,char *argv[])

/************************************************************************/
{
    const char *tracefile, *string;
    workenviron_t work;
    tracecall_t tcall;
    long start, end;
    int extraLine, searchIndex, searchLen;
    int c, flags, edinitmode, tmode;

    edinitmode = 1;
    work.Ranges = NULL;
    work.NRanges = 0;
    jstrbuffclear(&(work.RangeBuffer));

#ifdef SHOWTRACE_DEFAULT_RANGES
    if(readLineRanges(&(work.Ranges), &(work.NRanges), SHOWTRACE_DEFAULT_RANGES) == -1)
	EGOTO(error3);
#endif

    work.MaxDepth = DEFAULT_MAXDEPTH;
    work.MaxOpenDepth = DEFAULT_MAXOPENDEPTH;
    work.ReadLineSlice = DEFAULT_READLINESLICE;

    tmode = JEDT_TREEWIN_FFRAME | JEDT_TREEWIN_FSCROLLBAR;
    while((c = jsGetOpt(&argc, argv, OPTSTRING, &string)) != EOF)
	switch(c) {
	default:
	    EGOTO(error);

	case 'h':
	    goto help;

	case 'd':
	    work.MaxDepth = atoi(string);
	    break;

	case 'a':
	    work.MaxDepth = 100000; /*  that should be enough for all future displays :)  */
	    break;

	case 's':
	    work.ReadLineSlice = atof(string) * 1000000;
	    break;

	case 'r':
	    if(readLineRanges(&(work.Ranges), &(work.NRanges), string) == -1)
		EGOTO(error3);
	    break;

	case 'l':
	    tmode |= JEDT_TREEWIN_FLINENR;
	    break;

	case 'C':
	    edinitmode |= 2 | 4 | 8 | 16 | 32;
	    break;

#ifdef DEBUG
	case 'D':
	    parsedbgopts(NULL, string, &Debug, 0);
	    break;
#endif
	}

    if(argc != 2)
	goto help;

    /*  Open trace file  */
    tracefile = argv[1];
    if((work.File = fopen(tracefile, "r")) == NULL)
	EGOTO(error2);

    if(getFileRange(work.File, &flags, &start, &end) == -1)
	EGOTO(error);

    work.FileOffset = 0;
    fstrinit(&(work.StringPool));

    /*  Initialize curses stuff  */
    if(jedinitialize(edinitmode) == -1)
	EGOTO(error2);

    c = edinitmode != 1 ? 1 : 0;
    extraLine = LINES - 2 * c - PROGBAR_HEIGHT - 1;
    if((work.Tree = jedtreecreate(NULL, tracefile, c, c, extraLine, COLS - 2 * c,
				  tmode)) == NULL)
	EGOTO(error);

    /*  This will dynamically open nodes  */
    work.Tree->JEDT_Setup = dotriggernode;
    work.Tree->JEDT_User = &work;

    tclistinit(&work);
    work.Last = NULL;
    work.WorkDone = 0.0;

    tcallinit(&tcall, flags, 0, start, end);
    if(pushwork(&work, NULL, &tcall, 0) == -1)
	EGOTO(error);

    /*  Preset start time to estimate read speed  */
    if(getMicroTime(&(work.StartTime[0]), &(work.StartTime[1])) == -1)
	EGOTO(error);

    /* *********************************************************** */
    /*
     *  bit 0: If set, nodelay is set, i. e. wgetch does not hang.
     *  bit 1: If set, print text in extra line.
     */
    flags = 2;
    searchIndex = -1;
    work.SearchArray[0] = '\0';
    searchLen = 0;
    for(;;) {
	int rc, key;

	if(flags & 2) {
	    mvprintw(extraLine, 0, "+|-|' ' open|close|toggle; O|C open|close all; C-F|C-B page up|down; /|n|N search|up|down\n");
	    refresh();
	    flags &= ~2;
	}

	if(!tclistempty(&work)) {
	    if((flags & 1) == 0) {
		nodelay(work.Tree->JEDT_Window, TRUE);
		flags |= 1;
	    }

	    key = wgetch(work.Tree->JEDT_Window);
	    if(key == ERR) {
		if((rc = readLines(&work)) == -1)
		    EGOTO(error);

		if(rc == 0) {
		    nodelay(work.Tree->JEDT_Window, FALSE);
		    flags &= ~1;

		    jedprogbarclear(stdscr, extraLine + 1, 0,
				    PROGBAR_HEIGHT, COLS);
		    wrefresh(stdscr);

		} else {
		    tracecalltree_t *curr;
		    double have, total;
		    int cnt;

		    /*  Sometimes the progress bar flips back and forth. This is
		     *   due to the fact that EndPos is only set when a ']' line
		     *   is found which then increases the workload.
		     */
		    have = total = work.WorkDone;
		    cnt = 0;
		    tclistloop(&work, curr) {
			have  += curr->Offset            - curr->Reference->StartPos;
			total += curr->Reference->EndPos - curr->Reference->StartPos;
			++cnt;
		    }

		    if(total > 0.0) {
#ifdef DEBUG
			if(Debug & 1)
			    mvprintw(LINES - PROGBAR_HEIGHT, 0, "%-6d ", cnt);
#endif

			jedprogbarshow(stdscr, extraLine + 1,
#ifdef DEBUG
				       (Debug & 1) ? 6 :
#endif
				       0,
				       PROGBAR_HEIGHT, COLS, have / total,
				       GET_COLORPAIR(4, 7) | (1 << 9) | (1 << 10));
			wrefresh(stdscr);
		    }
		}
	    }

	    jedtreedokey(work.Tree, 0, 0, 0); /*  redraw  */

	} else {
#ifdef DEBUG
	    if(Debug & 1) {
		mvprintw(LINES - PROGBAR_HEIGHT, 0,
			 "TOPSC %d CURR %d NLINES %d LREAD %ld BREAD %ld         ",
			 work.Tree->JEDT_TopScreenLine, work.Tree->JEDT_CurrentLine,
			 work.Tree->JEDT_NLines, ReadLines, ReadBytes);
		wrefresh(stdscr);
	    }
#endif

	    jedtreedokey(work.Tree, 0, 0, 0); /*  redraw  */
	    key = wgetch(work.Tree->JEDT_Window);
	}

	if(searchIndex != -1 && key != ERR) {
	    switch(key) {
	    case '\n': case '\r': case KEY_ENTER:
		work.SearchArray[searchLen] = '\0';

		searchIndex = -1;
		flags |= 2;

		if(searchNode(&work, 0) == 0) {
		    beep();
		}

		jedtreedokey(work.Tree, 0, 0, 0); /*  redraw  */

		continue;

	    case KEY_LEFT:
		if(searchIndex == 0)
		    beep();
		else
		    --searchIndex;
		break;

	    case KEY_RIGHT:
		if(searchIndex >= searchLen)
		    beep();
		else
		    ++searchIndex;
		break;

	    case KEY_BACKSPACE:
		if(searchIndex == 0) {
		    beep();
		    continue;
		}

		--searchIndex;
		--searchLen;
		memmove(&work.SearchArray[searchIndex], &work.SearchArray[searchIndex + 1],
			searchLen - searchIndex + 1);

		break;

	    case KEY_DC:
		if(searchIndex >= searchLen) {
		    beep();
		    continue;
		}

		--searchLen;
		memmove(&work.SearchArray[searchIndex], &work.SearchArray[searchIndex + 1],
			searchLen - searchIndex + 1);
		break;

	    default:
		if(!isprint(key)) {
		    beep();
		    continue;
		}

		memmove(&work.SearchArray[searchIndex + 1], &work.SearchArray[searchIndex],
			searchLen - searchIndex + 1);
		work.SearchArray[searchIndex++] = key;
		++searchLen;
		break;
	    }

	    mvaddnstr(extraLine, 8, work.SearchArray, searchIndex);
	    if(searchIndex < searchLen) {
		addch(A_REVERSE | work.SearchArray[searchIndex]);
		addstr(work.SearchArray + searchIndex + 1);

	    } else
		addch(A_REVERSE | ' ');

	    clrtoeol();
	    refresh();

	    continue;
	}

	if(key == '/') {
	    searchIndex = 0;
	    searchLen = 0;
	    work.SearchArray[0] = '\0';
	    mvprintw(extraLine, 0, "SEARCH: ");
	    addch(A_REVERSE | ' ');
	    clrtoeol();
	    refresh();
	    continue;
	}

	if(key == 'n' || key == 'N') {
	    if(work.SearchArray[0] == '\0') {
		beep();
	    } else if(searchNode(&work, key == 'N' ? 2 : 0) == 0) {
		beep();
	    }

	    jedtreedokey(work.Tree, 0, 0, 0); /*  redraw  */
	    continue;
	}

	if(key == 'q' || key == '\033')
	    break;

#ifdef DEBUG
	if(key == 'D') {
	    Debug ^= 1;
	    continue;
	}
#endif

	if(key == 'l') {
	    jedtreedokey(work.Tree, 0, 0, 0); /*  redraw  */
	    continue;
	}

	if(key != ERR && jedtreedokey(work.Tree, jeddokeytrans(key, NULL), 1, 1) == 0) {
	    beep();
	}
    }

    fclose(work.File);

    if(jeddeinitialize() == -1)
	EGOTO(error2);

    return 0;
 help:
    fprintf(stderr, "\
Usage: jsshowtrace [<opts>] <tracefile>             (C) Joerg Schoen 2007-2009\n\
  Display trace files from 'jscpp' as a tree for easy navigation. The file is\n\
  read in the background, so one can immediately start navigating.\n\
  To limit memory consumption, nodes are initially only built up with a\n\
  maximum depth of %d and will later be rebuilt when opened.\n\
  Options:\n\
    -h         Show this help information\n\
    -d<depth>  Build up nodes to given depth.\n\
    -a         Build up tree completely during initial read.\n\
    -l         Show line numbers.\n\
    -s<slice>  Interval for reading chunks before looking for user input.\n\
               Defaults to %.2f sec.\n\
    -C         Do not use special curses stuff, e. g. colors etc.\n\
    -r<cols>   Cut each line according to the given columns. <cols>\n\
               consists of one or multiple comma separated\n\
                 '<start>-<end>' or '<start>-' or '<column>'.\n"
#ifdef SHOWTRACE_DEFAULT_RANGES
	    "               Default is '-r" SHOWTRACE_DEFAULT_RANGES "'.\n"
#endif
	    ,
	    DEFAULT_MAXDEPTH, DEFAULT_READLINESLICE * 0.000001);

    return 5;

 error:
    jeddeinitialize();
 error2:
    jsperror("jsshowtrace");
 error3:
    return 20;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int readLineRanges(int **pRanges, int *pNRanges, const char *string)

/************************************************************************/
{
    int *ranges, nranges, mranges;

    ranges = NULL;
    nranges = mranges = 0;
    for( ; ; ) {
	int flags;

	if(nranges >= mranges) {
	    mranges += 5;
	    if((ranges = (int *)realloc(ranges, 2* mranges * sizeof(*ranges))) == NULL)
		return -1;
	}

	flags = 0;
	if(isdigit(*string)) {
	    ranges[2 * nranges] = strtol(string, (char **)&string, 10);
	    flags |= 1;

	} else
	    ranges[2 * nranges] = 0; /*  "-<to>"  */

	if(*string == '-') {
	    ++string;

	    if(isdigit(*string)) {
		ranges[2 * nranges + 1] = strtol(string, (char **)&string, 10); /*  "<from>-<to>"  */
		flags |= 2;

	    } else
		ranges[2 * nranges + 1] = -1;  /*  "<from>-" means til end of line  */

	} else
	    ranges[2 * nranges + 1] = ranges[2 * nranges]; /*  "<col>" means single column  */

	if(flags == 0) /*  do not allow "-"  */
	    EGOTO(syntError);

	++nranges;

	if(*string == '\0')
	    break;

	if(*string != ',')
	    EGOTO(syntError);

	++string;
    }

    *pRanges = ranges;
    *pNRanges = nranges;

    return 0;
 syntError:
    fprintf(stderr, "ERROR: Illegal range '%s'!\n", string);
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int getFileRange(FILE *fp, int *pFlags, long *pStart, long *pEnd)

/************************************************************************/
{
    long fpLen;
#if 0
    char *line;
    long ordinal0;

    ordinal0 = -1;
    while((line = jsgetline(fp, GETLINE_RMEOL)) != NULL) {
	long number;

	if(!isdigit(*line))
	    continue;

	number = strtol(line, &line, 10);
	if(*line == ':') {
	    ordinal0 = number;
	    break;
	}
    }
#endif

    if((fpLen = fpFileLength(fp)) == -1)
	EGOTO(error);

#if 0
    if(ordinal0 != -1) {
	long lower, upper;

	/*  This was generated with limited size of the trace file, so
	 *   locate the actual position of the transition where
	 *     ordinal[line - 1] >= ordinal[0]
	 *   and
	 *     ordinal[line]     <  ordinal[0]
	 */
	lower = 0;    upper = fpLen;
	while(lower < upper) {
	    int c;
	    long m, currord, llen;

	    m = (lower + upper) >> 1;
	    if(fseek(fp, m, SEEK_SET) == -1)
		EGOTO(error);

	    /*  Find start of next line and correct m  */
	    while((c = fgetc(fp)) != EOF) {
		++m;

		if(c == '\n')
		    break;
	    }

	    currord = -1;
	    llen = 0;
	    while((line = jsgetline(fp, GETLINE_RMEOL)) != NULL) {
		llen = strlen(line) + 1;
		if(*line != '#') {
		    if(isdigit(*line))
			currord = strtol(line, &line, 10);

		    break;
		}

		m += llen;
	    }

#if 0
	    dbgprintf("  [%ld, %ld] M %ld -> %ld ORD %ld CURR %ld\n", lower, upper,
		      (lower + upper) >> 1, m, ordinal0, currord);
#endif

	    if(currord >= ordinal0 || m == upper)
		lower = m + llen;
	    else
		upper = m;
	}

	if(lower < fpLen) {
	    *pFlags = 4;
	    *pStart = upper;
	    *pEnd = upper;
	    abort();
	} else {
	    *pFlags = 0;
	    *pStart = 0;
	    *pEnd = fpLen;
	}

    } else
#endif
	{
	    *pFlags = 0;
	    *pStart = 0;
	    *pEnd = fpLen;
	}

#if 0
    dbgprintf("GOT %ld [%ld, %ld]\n", fpLen, *pStart, *pEnd);
#endif

    rewind(fp);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Add another work package. If bit 0 of mode is set, add this at the
 *    end.
 *************************************************************************/

static int pushwork(workenviron_t *pwork, jedtreenode_t *root,
		    tracecall_t *tcall, int mode)

/************************************************************************/
{
    tracecalltree_t *tctree;

#ifdef XDEBUG
    dbgprintf("PUSH %ld %ld\n", tcall->StartPos,tcall->EndPos);
#endif

    if((tctree = tracecalltreealloc()) == NULL)
	return -1;

    tctree->Reference = tcall;
    tctree->Offset = tcall->StartPos;
    tctree->Root  = root;

    tctree->Stack = NULL;
    tctree->NStack = tctree->MStack = 0;

    if(tclistempty(pwork)) {
	pwork->WorkDone = 0.0;
	pwork->TotalLines = pwork->TotalMillis = 0;
	pwork->Lines2Read = DEFAULT_LINES2READ;
    }

    if(mode & 1) {
	tcaddtail(pwork, tctree);

    } else {
	if(pwork->Last != NULL)
	    tcaddafter(pwork->Last, tctree);
	else
	    tcaddhead(pwork, tctree);

	pwork->Last = tctree;
    }

    return 0;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static tracecalltree_t *popwork(workenviron_t *pwork)

/************************************************************************/
{
    tracecalltree_t *tctree;

    tctree = tclistfirstnode(pwork);
    tclistremove(tctree);

    if(!tclistempty(pwork))
	pwork->WorkDone += tctree->Reference->EndPos - tctree->Reference->StartPos;

    free(tctree->Stack);
    tracecalltreefree(tctree);

    if(tclistempty(pwork))
	return NULL;

    return tclistfirstnode(pwork);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int dotriggernode(void *user, jedtree_t *edtr, jedtreenode_t *node)

/************************************************************************/
{
    return triggernode((workenviron_t *)user, node, 0);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int triggernode(workenviron_t *pwork, jedtreenode_t *node, int mode)

/************************************************************************/
{
    tracecall_t *tcall;

    tcall = (tracecall_t *)node->JEDTN_User;

    if(tcall && (tcall->Flags & 1)) {
	tcall->Flags &= ~1;

	/*  Nodes that are still being built up are collected at the END
	 *   of the current working list to ensure they are not scanned
	 *   unless start and end point is known.
	 */
	if(pushwork(pwork, node, tcall,
		    (tcall->Flags & 2) ? 1 : 0) == -1)
	    return -1;

	if(mode & 1) {
	    while(!tclistempty(pwork))
		if(readLines(pwork) == -1)
		    EGOTO(error);
	}
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int readLines(workenviron_t *pwork)

/************************************************************************/
{
    tracecalltree_t *tctree;
    long iline, nlines;

    nlines = pwork->Lines2Read; /*  -1: Read whole file  */

    tctree = tclistfirstnode(pwork);
    pwork->Last = NULL;

    if(pwork->FileOffset != tctree->Offset) {
	if(fseek(pwork->File, tctree->Offset, SEEK_SET) == -1)
	    EGOTO(error);

	pwork->FileOffset = tctree->Offset;
    }

    for(iline = 0 ; iline != nlines ; ) {
	char *line, *string;
	jedtreenode_t *node;
	long fileOffset;
	int llen;
	char c;

	fileOffset = pwork->FileOffset;

	/*  We either have
	 *  ---------======================---------------------
	 *           ^StartPos             ^EndPos
	 *  or
	 *  =========----------------------=====================
	 *           ^EndPos               ^StartPos
	 */
	if(tctree->Reference->EndPos <= fileOffset &&
	   (tctree->Reference->StartPos <= tctree->Reference->EndPos ||
	    fileOffset < tctree->Reference->StartPos)) {
	    line = NULL;

	} else if((line = jsgetline(pwork->File, GETLINE_RMEOL)) == NULL &&
		  tctree->Reference->StartPos > tctree->Reference->EndPos) {
	    /*  Wrap around  */
	    if(fseek(pwork->File, 0, SEEK_SET) == -1)
		EGOTO(error);

	    /*tctree->Flags |= 8;*/
	    pwork->FileOffset = 0;
	    continue;
	}

#ifdef XDEBUG
	dbgprintf("  POS %ld N %ld ST %ld E %ld: %ld\n",
		  fileOffset, pwork->FileOffset,
		  tctree->Reference->StartPos, tctree->Reference->EndPos,
		  line ? strlen(line) : -1);
#endif

	if(line == NULL) {
	    if((tctree = popwork(pwork)) == NULL)
		/*  EOF reached, so work package is done  */
		return 0;

	    if(fileOffset != tctree->Offset) {
		if(fseek(pwork->File, tctree->Offset, SEEK_SET) == -1)
		    EGOTO(error);

		pwork->FileOffset = tctree->Offset;
	    }

	    continue;
	}

	llen = strlen(line);

	/*  Analyze line  */
	++iline;
	pwork->FileOffset += llen + 1;
#ifdef _WIN32
	pwork->FileOffset++;
#endif

#ifdef DEBUG
	ReadBytes += llen + 1;
	ReadLines++;
#endif

	while(llen > 0 && isspace(line[llen - 1]))
	    --llen;

	if(pwork->NRanges > 0) {
	    int i;

	    jstrbuffempty(&(pwork->RangeBuffer));
	    for(i = 0 ; i < pwork->NRanges ; ++i) {
		int start, end;

		start = pwork->Ranges[2 * i];
		end   = pwork->Ranges[2 * i + 1];

		if(start > llen)
		    start = llen;

		if(end == -1 || end >= llen)
		    end = llen - 1;

		if(start >= end)
		    continue;

		if(jstrbuffstore(&(pwork->RangeBuffer), line + start, end - start + 1) == -1)
		    EGOTO(error);
	    }

	    line = jstrbuffer(&(pwork->RangeBuffer));
	    llen = jstrbufflen(&(pwork->RangeBuffer));
	}

	/*  Now cut off trailing spaces  */
	while(llen > 0 && isspace(line[llen - 1]))
	    --llen;

	line[llen] = '\0';

#ifdef XDEBUG
	if(tctree->NStack > 0)
	    dbgprintf("%d %10p %10p %10p %10p %10p %10p %10p '%s'\n",
		      tctree->NStack, tctree->Stack[0], tctree->Stack[1], tctree->Stack[2],
		      tctree->Stack[3], tctree->Stack[4],
		      tctree->Stack[5], tctree->Stack[6],
		      line);
#endif

#ifdef SKIP_LINE_NUMBERS
	if(isdigit(line[0])) {
	    int ord;

	    ord = strtol(line, &line, 10);
	    if(*line == ':')
		++line;
	}
#endif

	while(*line != '\0' && isspace(*line))
	    ++line;

	if((c = *line) == '\0')
	    c = 'X'; /*  not any of TRACELINE_CHARS_*  */

	if(strchr(TRACELINE_CHARS_START, c) != NULL) { /* ***  Push on stack  *** */
	    if(tctree->NStack < pwork->MaxDepth) {
		jedtreenode_t *parent;
		int currDepth;

		++line;
		if((string = fstralloc(&(pwork->StringPool), strlen(line) + 1)) == NULL)
		    EGOTO(error);

		strcpy(string, line);

		parent = tctree->NStack > 0 ? tctree->Stack[tctree->NStack - 1] : tctree->Root;
		currDepth = tctree->Reference->Depth + tctree->NStack;

		if((node = addNode(pwork->Tree, parent, string, 0,
				   pwork->FileOffset, pwork->FileOffset,
				   currDepth < pwork->MaxOpenDepth - 1 ?
				   1 : 0)) == NULL)
		    EGOTO(error);

		((tracecall_t *)(node->JEDTN_User))->Flags |=
		    (tctree->NStack == pwork->MaxDepth - 1) ? 3 : 2;

		/*  Push it on the stack  */
		if(tctree->NStack >= tctree->MStack) {
		    REALLOC_RESIZE(tctree->MStack, 50);
		    if((tctree->Stack = (jedtreenode_t **)
			realloc(tctree->Stack, (tctree->MStack + 2) *
				sizeof(*(tctree->Stack)))) == NULL)
			EGOTO(error);
		}

		tctree->Stack[tctree->NStack] = node;

	    } else if(tctree->NStack == pwork->MaxDepth) {
		node = tctree->Stack[tctree->NStack - 1];

		node->JEDTN_Flags |= JEDT_TREE_FSUBNODES | JEDT_TREE_FCALLBACK;
	    }

	    tctree->NStack++;

	} else if(strchr(TRACELINE_CHARS_END, c) != NULL && tctree->NStack > 0) {
	    tctree->NStack--;

	    if(tctree->NStack < pwork->MaxDepth) {
		tracecall_t *tcall;
		const char *sep;
		int len;

		node = tctree->NStack >= 0 ? tctree->Stack[tctree->NStack] : tctree->Root;
		assert(node != NULL);
		tcall = (tracecall_t *)(node->JEDTN_User);

		assert(tcall->Flags & 2);

		++line;
		len = jstrtoid(node->JEDTN_Name);
		if(strncmp(node->JEDTN_Name, line, len) == 0) {
		    line += len;
		    sep = " =>";
		} else
		    sep = " #>";

		len = strlen(node->JEDTN_Name) + strlen(line) + strlen(sep) + 1;

#ifdef DEBUG
		if(Debug & 1) len +=25;
#endif

		if((string = fstralloc(&(pwork->StringPool), len)) == NULL)
		    EGOTO(error);

#ifdef DEBUG
		if(Debug & 1)
		    sprintf(string, "%s%s (%ld,%ld)", node->JEDTN_Name, line,
			    tcall->StartPos, fileOffset);
		else
#endif
		    sprintf(string, "%s%s%s", node->JEDTN_Name, sep, line);

		if(jedtreesetname(node, string, 0) == -1)
		    EGOTO(error);

		tcall->EndPos = fileOffset - llen;

		tcall->Flags &= ~2;
	    }

	} else if(strchr(TRACELINE_CHARS_COMM, c) == NULL &&
		  tctree->NStack < pwork->MaxDepth) {
	    if((string = fstralloc(&(pwork->StringPool), strlen(line) + 1)) == NULL)
		EGOTO(error);

	    strcpy(string, line);

	    if((node = addNode(pwork->Tree, tctree->NStack > 0 ?
			       tctree->Stack[tctree->NStack - 1] : tctree->Root,
			       string, 0, -1, -1, 0)) == NULL)
		EGOTO(error);

	}
    }

    tctree->Offset = pwork->FileOffset;

#if XDEBUG
    assert(ftell(pwork->File) == tctree->Offset);
#endif

    {
	long tEnd[2], diff;

	/*  Get end time  */
	if(getMicroTime(tEnd, tEnd + 1) == -1)
	    EGOTO(error);

	/*  Milliseconds needed  */
	if((diff = (tEnd[0] - pwork->StartTime[0]) * 1000000 +
	    (tEnd[1] - pwork->StartTime[1])) == 0)
	    diff = 1;

	pwork->TotalLines += nlines;
	pwork->TotalMillis += diff;

	nlines = ((double)pwork->TotalLines * (double)pwork->ReadLineSlice) /
	    (double)pwork->TotalMillis;
	if(nlines < 1)
	    nlines = 1;

	pwork->Lines2Read = nlines;

	/*  Set start time for next round  */
	pwork->StartTime[0] = tEnd[0];
	pwork->StartTime[1] = tEnd[1];
    }

    return 1;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static jedtreenode_t *addNode(jedtree_t *edtr, jedtreenode_t *parent,
			      const char *content, int flags,
			      long startPos, long endPos,
			      int mode)

/************************************************************************/
{
    jedtreenode_t *ednode;

    if((ednode = jedtreeaddnode(edtr, parent, content,
				0/*JEDT_TREEADD_FKEEPDIRTY*/)) == NULL)
	EGOTO(error);

    if(mode & 1)
	ednode->JEDTN_Flags |= JEDT_TREE_FOPENED;

    if(startPos != -1) {
	tracecall_t *tcall;
	int depth;

	if((tcall = tracecallalloc()) == NULL)
	    return NULL;

	depth = parent ? ((tracecall_t *)parent->JEDTN_User)->Depth + 1 : 1;
	tcallinit(tcall, flags, depth, startPos, endPos);

	ednode->JEDTN_User = tcall;
    }

    return ednode;
 error:
    return NULL;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int searchNode(workenviron_t *pwork, int mode)

/************************************************************************/
{
    return jedtreesearch(pwork->Tree, searchString, pwork, mode);
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int searchString(void *user, jedtreenode_t *node, int linenr, int level)

/************************************************************************/
{
    if(triggernode((workenviron_t *)user, node, 1) == -1)
	EGOTO(error);

    if(strstr(node->JEDTN_Name, ((workenviron_t *)user)->SearchArray) != NULL) {
#if 0
	dbgprintf("Found '%s' in l.%d '%s'\n",
		  ((workenviron_t *)user)->SearchArray, linenr,
		  node->JEDTN_Name);
#endif
	return 2 | 4;
    }

#if 0
    dbgprintf("LINE %d '%s'\n", linenr, node->JEDTN_Name);
#endif

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

#define GENMALLOC_TYPE          tracecall_t
#define GENMALLOC_NAMEALLOC     tracecallalloc
/*#define GENMALLOC_NAMEFREE      tracecallfree*/

#define GENMALLOC_MEMSIZE       20000
#define GENMALLOC_STATIC

#include "genmalloc.h"

/*JS*********************************************************************
 *
 *************************************************************************/

#define GENMALLOC_TYPE          tracecalltree_t
#define GENMALLOC_NAMEALLOC     tracecalltreealloc
#define GENMALLOC_NAMEFREE      tracecalltreefree

#define GENMALLOC_MEMSIZE       20000
#define GENMALLOC_STATIC

#include "genmalloc.h"
