/*JS*********************************************************************
*
*    Program : JSCPP
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Application for preprocessor library.
*
*    Part    : Instrument code to trace every single statement processed
*
*              Copyright (C) 2005-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include "jscpp.h"

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define JSCPP_MODE_STMTSTEP        "stmtstep"
#define JSCPP_MODE_STMTSTEP0       "stmtstep0"
#define JSCPP_MODE_CODECOVER       "codecov"

#define STMTTRACE_FUNCTION         "_jstracestmt"
#define STMT_FILENAME_VAR          "_jstracefname"

#define CODECOVER_REGFUNCTION      "_jscovregister"
#define CODECOVER_COUNTARRAY       "_jscovcounts"
#define CODECOVER_DESCRIPTARRAY    "_jscovdescript"
/*  END-DEFINITIONS  */

#define STMT_TRACE_FUNCTION_CALL1     STMTTRACE_FUNCTION \
					"(" STMT_FILENAME_VAR ",__LINE__)"
#define STMT_TRACE_FUNCTION_CALL2     STMTTRACE_FUNCTION \
					"(" STMT_FILENAME_VAR ",%d)"

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            doStmtTracing(traceaux_t *paux,
				       const patchdesc_t *pdesc,
				       jsparsstream_t *jpa,
				       jsparscnode_t *tree,
				       jctokfifo_t *queue,
				       const char *funcName);
static int               patchStatement(traceaux_t *paux,
					const patchdesc_t *pdesc,
					jsparsstream_t *jpa,
					jsparscnode_t *tree,
					long flags, jsparscnode_t *node,
					jctokfifo_t *queue,
					const char *funcName);
Prototype int            checkAssignment(jsparsstream_t *jpa,
					 jctokfifo_t *queue,
					 jsparscnode_t *node);
Prototype int            tokInsertString(jsparsstream_t *jpa,
					 jsparscnode_t *tree,
					 jctokfifo_t *queue, int pos,
					 const char *string,
					 int mode);

/*JS*********************************************************************
*
*************************************************************************/

int doStmtTracing(traceaux_t *paux, const patchdesc_t *pdesc,
		  jsparsstream_t *jpa, jsparscnode_t *tree,
		  jctokfifo_t *queue, const char *funcName)

/************************************************************************/
{
  jsparsflow_t *jFlow;
  int i, lastPos;

  /*  Process all statements   */
  jFlow = jpa->JSPA_StmtFlow;

  /*  We are lazy here: We process statements in increasing order in
   *   the array. After skipping some declarations, we assume that the
   *   first statement we hit is the first that is really executed.
   *   This will generate the function name entry in the log.
   */
  lastPos = -1;
  for(i = 0 ; i < jFlow->JPAFL_NStatements ; ++i) {
    jsparscnode_t *curr;
    long flags;
    int ret;

    flags = jFlow->JPAFL_Statements[i].JPAFLN_Flags;
    curr = jFlow->JPAFL_Statements[i].JPAFLN_Node;

    /*  Only trace statements where a branch occurs or we are after a branch
     *   (except for the first statement)
     */
    if(!(pdesc->Type & PDESC_TYPE_FALLSTMT) &&
       !(flags & (JPA_FLOW_BRANCHED | JPA_FLOW_BLOCKSTART))) {
      int fpos;

      if((fpos = jctokindfpos(queue)) != -1)
	  lastPos = queue->TKFF_Token[curr->JPACN_Start + fpos].TK_Value;

      continue;
    }

    if(flags & JPA_FLOW_PSEUDOSTMT)
      continue;

    if(curr->JPACN_Type == JPACT_TYPEDEF ||
       ((curr->JPACN_Type == JPACT_UNARYEXPR ||
	 curr->JPACN_Type == JPACT_COMPSTMT) && curr->JPACN_Sub == NULL))
      continue;

    if(curr->JPACN_Type == JPACT_DECLARATION) {
      jsparscnode_t *node2;

      /*  We check if this contains something like an assignment like in
       *     void dummy() { int i = callee(); printf("Hello, world!\n"); }
       */
      curr = curr->JPACN_Sub;

      /*  Make sure that we do not try to patch static or const definitions like
       *   "static int X = 1;"
       */
      for(node2 = curr->JPACN_Next ; node2 ; node2 = node2->JPACN_Next)
	if(node2->JPACN_Type == JPACT_CONST || node2->JPACN_Type == JPACT_STATIC)
	  break;

      if(node2 != NULL)
	continue;

      /*  Pretend an expression statement in what follows  */
      flags |= JPA_FLOW_EXPRESSION;

      if(curr->JPACN_Type == JPACT_LIST) {
	for(curr = curr->JPACN_Sub ; curr ; curr = curr->JPACN_Next) {
	  if(curr->JPACN_Type != JPACT_ASS)
	    continue;

	  if((ret = checkAssignment(jpa, queue, curr)) == -1)
	    EGOTO(error);
	  if(ret == 0)
	    continue;

	  node2 = curr->JPACN_Sub->JPACN_Next;

	  if(patchStatement(paux, pdesc, jpa, tree, flags, node2, queue,
			    funcName) == -1)
	    EGOTO(error);

	  if(!(pdesc->Type & PDESC_TYPE_FALLSTMT) &&
	     !(flags & (JPA_FLOW_BRANCHED | JPA_FLOW_BLOCKSTART)))
	    break;
	}

	continue;
      }

      if(curr->JPACN_Type != JPACT_ASS)
	continue;

      if((ret = checkAssignment(jpa, queue, curr)) == -1)
	EGOTO(error);

      if(ret == 0)
	continue;

      curr = curr->JPACN_Sub->JPACN_Next;
    }

    if((ret = patchStatement(paux, pdesc, jpa, tree, flags, curr, queue,
			     funcName)) == -1)
      EGOTO(error);

    if(ret && !(flags & JPA_FLOW_EXPRESSION))
      /*  This flag is no longer true  */
      jFlow->JPAFL_Statements[i].JPAFLN_Flags &= ~JPA_FLOW_BLOCKSTART;

    if(pdesc->Type & PDESC_TYPE_FCODECOVERAGE) {
	int fpos, n;

	n = paux->NCovRange;
	if(paux->NCovRange >= paux->MCovRange) {
	    paux->MCovRange += 100;
	    if((paux->CovRange = (int *)
		realloc(paux->CovRange, 2 * paux->MCovRange * sizeof(*(paux->CovRange)))) == NULL)
		EGOTO(error);
	}

	if(n > 0)
	    paux->CovRange[2 * (n - 1) + 1] = lastPos;

	if((fpos = jctokindfpos(queue)) != -1)
	    lastPos = queue->TKFF_Token[curr->JPACN_Start + fpos].TK_Value;

	paux->CovRange[2 * n] = lastPos;

	paux->NCovRange++;
    }
  }

  if((pdesc->Type & PDESC_TYPE_FCODECOVERAGE) && paux->NCovRange > 0)
      paux->CovRange[2 * (paux->NCovRange - 1) + 1] = lastPos;

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int patchStatement(traceaux_t *paux, const patchdesc_t *pdesc,
			  jsparsstream_t *jpa, jsparscnode_t *tree,
			  long flags, jsparscnode_t *node,
			  jctokfifo_t *queue, const char *funcName)

/************************************************************************/
{
  const char *formatStart, *formatEnd, *string;
  int start, end;
  const char *prefix[2];

  if(flags & JPA_FLOW_EXPRESSION) {
    prefix[0] = "(";
    prefix[1] = ", ";
    formatEnd = ")";

  } else if(flags & JPA_FLOW_SINGLESTMT) {
    prefix[0] = "{ ";
    prefix[1] = "; ";
    formatEnd = " }";

  } else {
    prefix[0] = NULL;
    prefix[1] = "; ";
    formatEnd = NULL;
  }

  if(!(pdesc->Type & PDESC_TYPE_FCODECOVERAGE))
      string = paux->TraceFormatStmt;
  else if((flags & JPA_FLOW_BLOCKSTART) && strcmp(funcName, "main") == 0)
      string = CODECOVER_REGFUNCTION "(" CODECOVER_COUNTARRAY ", " CODECOVER_DESCRIPTARRAY ");"
	  CODECOVER_COUNTARRAY"[%C]++;";
  else
      string = CODECOVER_COUNTARRAY"[%C]++;";

  if((formatStart = genTraceFctCall(paux, pdesc, queue, node,
				    string, "", funcName, "", "", prefix)) == NULL)
    EGOTO(error);

  start = node->JPACN_Start;
  end = node->JPACN_End;

  if(jctokindorig(queue) != -1 && !jsparsmatchorigtoks(queue, &start, &end)) {
    /*  This case happens frequently, so only print it if extended
     *   warnings are on
     */
    if(jpa->JSPA_WarnFlags & JPAM_WARNSTYLE)
      jsparsprinterr(jpa, queue, node,
		     "Cannot patch statement in function '%s'.", 0,
		     funcName);

    /*  We cannot patch macro expansions  */
    return 0;
  }

  /*  'end' has to come first because inserting at position 'start'
   *   moves everything
   */
  if(formatEnd != NULL &&
     tokInsertString(jpa, tree, queue, end + jctokgetstep(queue),
		     formatEnd, 0) == -1)
    EGOTO(error);

  if(tokInsertString(jpa, tree, queue, start, formatStart, 0) == -1)
    EGOTO(error);

  return 1;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int checkAssignment(jsparsstream_t *jpa, jctokfifo_t *queue, jsparscnode_t *node)

/************************************************************************/
{
  jsparscnode_t *sub;
  jsparsvartab_t *vTab;
  jsparsvardef_t *vDef;
  jsparsexprtype_t eType;
  const char *varname;
  int ret;

  /*assert(node->JPACN_Type == JPACT_ASS);*/

  node = node->JPACN_Sub;

  assert(node->JPACN_Type == JPACT_DECLARATOR);

  sub = node->JPACN_Sub;
  node = node->JPACN_Next;

  /*  Avoid patching something like
   *   struct { int a,b; } x = { 1, 2 };
   */
  if(node->JPACN_Type == JPACT_BLOCK)
    return 0;

  assert(sub->JPACN_Type == JPACT_IDENT);
  varname = queue->TKFF_Token[sub->JPACN_IndexPP].TK_String;

  if((vTab = jsparsfindvartab(jpa, sub)) == NULL ||
     (vDef = jsparsgetvardef(jpa, vTab, varname, JVDF_TVARIABLE, 0)) == NULL ||
     jsparsvdef2type(jpa, vTab, &eType, vDef, 2) == NULL)
    EGOTO(error);

  /*jsparsprctree(stdout,queue, eType.JSPE_Declarator,0,4);*/

  ret = 1;
  if((sub = eType.JSPE_Declarator) != NULL) {
    if(sub->JPACN_Type == JPACT_DECLARATOR) sub = sub->JPACN_Sub;
    if(sub && sub->JPACN_Type == JPACT_IDENT) sub = sub->JPACN_Next;

    if(sub && sub->JPACN_Type == JPACT_VECTOR)
      ret = 0;
  }

  jsparsfreeexprtype(jpa, &eType, 1);

  return ret;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int tokInsertString(jsparsstream_t *jpa, jsparscnode_t *tree,
		    jctokfifo_t *queue, int pos,
		    const char *string, int mode)

/************************************************************************/
{
  jsctoken_t tokens[3];
  int indfpos, indOff;

  /*  Both tokens are "vanilla" ones  */
  tokens[0].TK_Value = CTK_VANILLA;
  tokens[0].TK_String = NULL;
  tokens[1] = tokens[0];

  if((indOff = jctokindorig(queue)) == -1)
    indOff = 0;

  if((indfpos = jctokindfpos(queue)) != -1)
    tokens[indfpos] = queue->TKFF_Token[pos + indfpos];

  tokens[indOff].TK_String = (char *)string;

  return jsparsinserttoks(jpa, tree, queue, tokens,
			  jctokgetstep(queue), pos, mode);
}
