/*JS*********************************************************************
*
*    Program : CRCCALC
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Calculate a CRC checksum.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*              Hint: To get a final checksum of 0, use the bytes of
*                    the checksum in the order most significant to
*                    least significant as input bytes for crcCalc.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <limits.h>

#include <jsconfig.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS  */
#include <jsendian.h>

/*  Should be CHAR_BIT, but we do not wan't to export this
 *   into the definition of CRC_TABSIZE.
 */
#define CRC_CHAR_BIT    8

/* **  Needed length of a CRC table  ** */
#define CRC_TABSIZE     (1 << CRC_CHAR_BIT)

/*  These macros can be used to write a calculated checksum at the
 *   end of the area to get a total checksum of zero
 */
/*  For 16 bits  */
#define CHKSUM_TO_CHARS1(value, array)  JS_INT162CHAR_BE(value, array)
#define CHKSUM_LEN1                     JS_INT162CHAR_LEN

/*  For 32 bits  */
#define CHKSUM_TO_CHARS2(value, array)  JS_INT322CHAR_BE(value, array)
#define CHKSUM_LEN2                     JS_INT322CHAR_LEN

/*  For 64 bits  */
#define CHKSUM_TO_CHARS3(value, array)  JS_INT642CHAR_BE(value, array)
#define CHKSUM_LEN3                     JS_INT642CHAR_LEN

/*  We can get the polynomial back from the generated table  */
#define CRC_GETPOLY(tab)   (tab[1])
/*  END-DEFINITIONS  */

/*  To cut off bits  */
#define CHKSUM_MASK1                0xffffUL
#define CHKSUM_MASK2            0xffffffffUL
#define CHKSUM_MASK3    0xffffffffffffffffUL

#define BIT(i)          (1UL << (i))
#define BIT_MASK        (BIT(CRC_CHAR_BIT) - 1)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype void           crcGenTab(unsigned long *table, unsigned long poly,
				   int flag);
Prototype unsigned long  crcCalc(unsigned long start, const unsigned char *buf,
				 long bufLen, const unsigned long *crcTab,
				 int flag);

/*JS*********************************************************************
*   Fill a CRC table for the generating polynomial poly. flag determines
*    the size of the CRC:
*      flag = 1: 16Bit CRC.
*      flag = 2: 32Bit CRC.
*      flag = 3: 64Bit CRC, if the machine has 64 Bit longs.
*************************************************************************/

void crcGenTab(unsigned long *table, unsigned long poly, int flag)

/************************************************************************/
{
  unsigned long bitMask[CRC_CHAR_BIT], highBit;
  int i;

  switch(flag) {
  case 1: /* ***  16 Bit CRC  *** */
    highBit = BIT(15);
    poly &= CHKSUM_MASK1;
    break;
  default: /*  Default is 32 Bit CRC  */
  case 2: /* ***  32 Bit CRC  *** */
    highBit = BIT(31);
    poly &= CHKSUM_MASK2;
    break;
#ifdef CONFIG_64BIT_LONG
  case 3: /* ***  64 Bit CRC  *** */
    highBit = BIT(63);
    poly &= CHKSUM_MASK3;
    break;
#endif
  }

  /*  Set up table of masks for every bit  */
  bitMask[0] = poly;
  for(i = 1 ; i < CRC_CHAR_BIT ; i++)
    bitMask[i] = (bitMask[i - 1] & highBit) ?
      (((bitMask[i - 1] & ~highBit) << 1) ^ poly) : (bitMask[i - 1] << 1);

  for(i = 0 ; i < CRC_TABSIZE ; i++) {
    unsigned long tabVal;
    int j, k;

    tabVal = 0L;
    for(j = i, k = 0 ; j ; k++, j >>= 1)
      if(j & 1) tabVal ^= bitMask[k];

    table[i] = tabVal;
  }

  return;
}

/*JS*********************************************************************
*   Update a CRC with the buffer using the crcTab. flag determines the
*    size of the CRC analogous as in routine crcGenTab.
*************************************************************************/

unsigned long crcCalc(unsigned long start, const unsigned char *buf, long bufLen,
		      const unsigned long *crcTab, int flag)

/************************************************************************/
{
  long i;

  /*  Update checksum. Repeat code to take advantage of constant right shift  */
#define UPDATE_LOOP(cksize) \
    for(i = 0 ; i < bufLen ; i++) \
      start = (start << CRC_CHAR_BIT) ^ \
	crcTab[((start >> (cksize - CRC_CHAR_BIT)) ^ buf[i]) & BIT_MASK]

#define UPDATE_LOOP0(cksize) \
    for(i = 0 ; i < bufLen ; i++) \
      start = (start << CRC_CHAR_BIT) ^ \
	crcTab[(start >> (cksize - CRC_CHAR_BIT)) & BIT_MASK]

  switch(flag) {
  case 1: /* ***  16 Bit CRC  *** */
    if(buf)
      UPDATE_LOOP(16);
    else
      UPDATE_LOOP0(16);
    start &= CHKSUM_MASK1;
    break;

  default: /*  Default is 32 Bit as well  */
  case 2: /* ***  32 Bit CRC  *** */
    if(buf)
      UPDATE_LOOP(32);
    else
      UPDATE_LOOP0(32);
    start &= CHKSUM_MASK2;
    break;

#ifdef CONFIG_64BIT_LONG
  case 3: /* ***  64 Bit CRC  *** */
    if(buf)
      UPDATE_LOOP(64);
    else
      UPDATE_LOOP0(64);
    start &= CHKSUM_MASK3;
    break;
#endif
  }

#undef UPDATE_LOOP
#undef UPDATE_LOOP0

  return start;
}

#ifdef GENCRCTAB
/* ***  Generate a CRC table  *** */
#include <stdlib.h>

#include <jssubs.h>

#define NPERLINE    6

unsigned long CrcTab[CRC_TABSIZE];

int main(int argc, char *argv[])
{
  unsigned long poly;
  int flag, size, i;

  if(argc != 3) {
    fprintf(stderr, "Usage: gencrctab <crcsize> <polynomial>\n\
  Outputs a CRC table with the given generating polynomial\n\
  crcsize may be "
#ifdef CONFIG_64BIT_LONG
	    "16, 32 or 64"
#else
	    "16 or 32"
#endif
	    "bits.\n");
    return 5;
  }

  size = strtol(argv[1], NULL, 0);
  switch(size) {
  case 16: flag = 1; break;
  default:
    fprintf(stderr, "WARNING: Illegal size for CRC, assuming 32.\n");
    size = 32;
  case 32: flag = 2; break;
#ifdef CONFIG_64BIT_LONG
  case 64: flag = 3; break;
#endif
  }
  poly = jstrtoul2(argv[2], NULL, 0);

  crcGenTab(CrcTab, poly, flag);

  printf("\
/* ***  CRC Tab for a %d bit CRC with generating polynomial 0x%08lx  *** */\n\
const unsigned long CrcTab[%ld] = {\n ", size, poly, CRC_TABSIZE);
  for(i = 0 ; i < CRC_TABSIZE ; i++) {
    printf(" 0x%08lx", CrcTab[i]);
    if(i < (CRC_TABSIZE - 1)) {
      if((i % NPERLINE) == (NPERLINE-1))
	printf(",\n ");
      else
	printf(",");
    }
  }
  printf("\n};\n");

  return 0;
}

#elif defined(CHKSUM)
/**********************************************************************
 *  Simple application: Calculate checksum of files on command line
 *   or stdin. Similar to command "cksum"
 **********************************************************************/

/*  The following polynomial is
    X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0
*/
#define GEN_POLY    0x04c11db7
#define CRC_SIZE    2
#define START_CRC   0xffffffff

#define BUFSIZE     8192

static unsigned long CrcTab[CRC_TABSIZE];

/* **  Prototype  ** */
static int cksum(char *fileName);

int main(int argc, char *argv[])
{
  /*  Generate table  */
  crcGenTab(CrcTab, GEN_POLY, CRC_SIZE);

  if(argc == 1) {
    /*  stdin  */
    if(cksum(NULL) < 0)
      goto error;

  } else {
    unsigned int i;

    for(i = 1 ; i < argc ; i++)
      if(cksum(argv[i]) < 0)
	goto error;
  }

  return 0;
error:
  perror("cksum");
}

static unsigned char Buffer[BUFSIZE];

static int cksum(char *fileName)
{
  FILE *fp;
  unsigned long crc;
  long len, i;

  if(fileName == NULL)
    fp = stdin;
  else if((fp = fopen(fileName, "r")) == NULL)
    EGOTO(error);

  crc = START_CRC;
  len = 0;
  while((i = fread(Buffer, 1, BUFSIZE, fp)) > 0) {
    len += i;
    crc = crcCalc(crc, Buffer, i, CrcTab, CRC_SIZE);
  }

  if(ferror(fp))
    EGOTO(error);

  /*  Print checksum, file length and name  */
  printf("%-40s\t%#010lx %8lu\n", fileName ? fileName : "", crc, len);

  if(fileName)
    fclose(fp);

  return 0;
error:
  if(fileName && fp)
    fclose(fp);
  return -1;
}
#endif
