/*JS*********************************************************************
 *
 *    Program : CTOKFIFO
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : Store C tokens in FIFO queue.
 *
 *              Copyright (C) 1994-2018 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>

#include <ctype.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#ifndef JCTOK_MODE_INSERT
/*  START-DEFINITIONS 1 */
/* **  To defer a sequence of tokens for later output  ** */
typedef struct {
    const char        *TKFPM_String;
    unsigned int       TKFPM_Offset;
} jctokfifoptrmap_t;

typedef struct {
    unsigned int       TKFF_Mode, TKFF_Step;

    unsigned int       TKFF_Start, TKFF_Index, TKFF_Max;
    jsctoken_t        *TKFF_Token;

    char              *TKFF_Mem;
    unsigned int       TKFF_MIndex, TKFF_MMax;

    jctokfifoptrmap_t *TKFF_PtrMap;
    unsigned int       TKFF_NPtrMap, TKFF_MPtrMap;

} jctokfifo_t;

#define JCTOK_TYPE_FORIGTOK       1
#define JCTOK_TYPE_FFILEPOS       2

/*  TK_String field is not monotonically increasing  */
#define JCTOK_TYPE_FNOSTRORDER    256

#define JCTOK_MAXSTEP             3

#define jctokgetstep(queue)      ((queue)->TKFF_Step)
#define jctokindorig(queue)      (((queue)->TKFF_Mode &			\
                                   JCTOK_TYPE_FORIGTOK) ? 1 : -1U)
#define jctokindfpos(queue)      (((queue)->TKFF_Mode &		\
                                   JCTOK_TYPE_FFILEPOS) ?	\
                                  (queue)->TKFF_Step - 1 : -1U)

/*  Empty the queue  */
#define jctokclear(queue)   ((queue)->TKFF_Start =	\
			     (queue)->TKFF_Index = 0,	\
			     jctokclearptrmap(queue))

#define jctokclearptrmap(queue)   ((queue)->TKFF_NPtrMap = 0)

#define jctoktostring(queue, start, end)		\
    jctoktostring2(queue, start, end, NULL, NULL)

/*  Modes for routine jctokwriten  */
#define JCTOK_MODE_INSERT    (1 << 0)
#define JCTOK_MODE_CHKPTR    (1 << 1)
/*  END-DEFINITIONS  */
#endif

#define FIFO_TOKINC    1000
#define FIFO_CHARINC  10000

#define INSTRINGMEM(queue, ptr)   ((queue)->TKFF_Mem <= (ptr) &&	\
				   (ptr) < ((queue)->TKFF_Mem +		\
					    (queue)->TKFF_MIndex))

/*  For jctokshowqueue  */
#define SCREENWIDTH              120

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype jctokfifo_t   *jctokcreate(jctokfifo_t *queue, int mode);
Prototype void           jctokfree(jctokfifo_t *queue, int flag);

Prototype int            jctokinsert(jctokfifo_t *queue, jsctoken_t token, int mode);
Prototype int            jctokwriten(jctokfifo_t *queue, jsctoken_t *tokens,
                                     int n, int where, int mode);
static int               _jctokinc(jctokfifo_t *queue,
                                   int nTokens, int stringLen);
static int               _jctokaddptr(jctokfifo_t *queue,
				      const char *string, unsigned int offset);
static int               _jctokfindptr(jctokfifo_t *queue, const char *string);

Prototype int            jctokdelete(jctokfifo_t *queue, int start, int end,
                                     int mode);

Prototype jsctoken_t     jctokget(jctokfifo_t *queue);
Prototype int            jctokgetfpos(jctokfifo_t *queue, const char **pName);
Prototype int            jctokgetindex(jctokfifo_t *queue);

Prototype void           jctokblank(jctokfifo_t *queue, int last);

Prototype int            jctoksubstitute(jctokfifo_t *queue, int start,
                                         int end, const char *format, ...);
static int               countNewLines(const char *string);

Prototype int            jctoksubststring(jctokfifo_t *queue, int start,
                                          int end, const char *string);


Prototype char          *jctoktostring2(jctokfifo_t *queue, int start, int end,
                                        jsctoken_t (*converter)(void *user,
                                                                jsctoken_t token),
                                        void *user);

Prototype int            jctoktostrbuff1(jstrbuff_t *sbuffer,
                                         jsctoken_t token);
Prototype int            jctoktostrbuff(jstrbuff_t *sbuffer, jctokfifo_t *queue,
                                        int start, int end,
                                        jsctoken_t (*converter)(void *user,
                                                                jsctoken_t token),
                                        void *user);

Prototype int            jctokgetcolumn(jctokfifo_t *queue, int indx,
                                        int mode);
Prototype void           jctokshowqueue(FILE *fp, jctokfifo_t *queue);
Prototype void           jctokshowqueue2(FILE *fp, jctokfifo_t *queue,
                                         int start, int end);

/*JS*********************************************************************
 *
 *************************************************************************/

jctokfifo_t *jctokcreate(jctokfifo_t *queue, int mode)

/************************************************************************/
{
    int step;

    if(queue == NULL &&
       (queue = (jctokfifo_t *)malloc(sizeof(*queue))) == NULL)
	return NULL;

    step = 1;
    if(mode & JCTOK_TYPE_FORIGTOK) ++step;
    if(mode & JCTOK_TYPE_FFILEPOS) ++step;

    queue->TKFF_Mode = mode;
    queue->TKFF_Step = step;

    queue->TKFF_Token = NULL;
    queue->TKFF_Mem = NULL;
    queue->TKFF_Max = queue->TKFF_MMax = 0;
    queue->TKFF_MIndex = 0;

    queue->TKFF_PtrMap = NULL;
    queue->TKFF_MPtrMap = 0;

    jctokclear(queue);

    return queue;
}

/*JS*********************************************************************
 *   Frees memory of a first-in-first-out queue.
 *************************************************************************/

void jctokfree(jctokfifo_t *queue, int flag)

/************************************************************************/
{
    free(queue->TKFF_Token);
    free(queue->TKFF_Mem);

    free(queue->TKFF_PtrMap);

    if(flag) free(queue);
}

/*JS*********************************************************************
 *   Implement a first-in-first-out queue for tokens: Insert a token on
 *    the queue. Returns the index in the queue where the token is stored
 *    or <0 on error.
 *************************************************************************/

int jctokinsert(jctokfifo_t *queue, jsctoken_t token, int mode)

/************************************************************************/
{
    const char *tokString;
    unsigned int len;

    tokString = token.TK_String;

    len = tokString != NULL && !(mode & 1) ? strlen(tokString) + 1 : 0;

    if(_jctokinc(queue, 1, len) == -1)
	EGOTO(error);

    queue->TKFF_Token[queue->TKFF_Index].TK_Flags = token.TK_Flags;
    queue->TKFF_Token[queue->TKFF_Index].TK_Value = token.TK_Value;

    if(len > 0) {
	char *dest;

	/*  Set up new pointer  */
	dest = &queue->TKFF_Mem[queue->TKFF_MIndex];
	queue->TKFF_Token[queue->TKFF_Index].TK_String = dest;

	strcpy(dest, tokString);
	queue->TKFF_MIndex += len;

    } else {
	queue->TKFF_Token[queue->TKFF_Index].TK_String = tokString;
    }

    return queue->TKFF_Index++;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Write/insert n tokens in queue. Property that TK_String field is
 *    identical to previous or preprevious is preserved. Returns the index
 *    in the queue where the first token is stored or <0 on error. If where
 *    is -1, the tokens are inserted at the end of the queue.
 *    String memory is - if necessary - inserted at the end of the current
 *    string memory, so TK_String of the token array does no longer
 *    monotonically increase! This means, JCTOK_MODE_INSERT is not possible for
 *    routine jctokdelete.
 *************************************************************************/

int jctokwriten(jctokfifo_t *queue, jsctoken_t *tokens, int n, int where,
                int mode)

/************************************************************************/
{
    int len, i, curr, step, indfpos, flag;
    char *dest;

    curr = queue->TKFF_Index;
    step = jctokgetstep(queue);
    indfpos = jctokindfpos(queue);

    if(where < 0 || where > curr)
	where = curr;

    /*  Calculate needed string space  */
    for(len = i = 0 ; i < n ; ++i) {
	const char *tokString;

	tokString = tokens[i].TK_String;
	if(tokString == NULL)
	    continue;

	if(i >= 1 && tokens[i - 1].TK_String == tokString)
	    continue;

	if(i >= 2 && tokens[i - 2].TK_String == tokString)
	    continue;

	if(indfpos != -1 && (where + i) % step == indfpos)
	    /*  The file names are shared, so no need to copy  */
	    continue;

	if((mode & JCTOK_MODE_CHKPTR) &&
	   _jctokfindptr(queue, tokString) != -1)
	    continue;

	len += strlen(tokString) + 1;
    }

    /*  How many new tokens are added  */
    i = (mode & JCTOK_MODE_INSERT) ? n : MAX(where + n - curr, 0);

    if(_jctokinc(queue, i, len) == -1)
	EGOTO(error);

    /*  Anything to move?  */
    flag = 0;
    if(mode & JCTOK_MODE_INSERT) {
	if(where < curr) {
	    memmove(&queue->TKFF_Token[where + n], &queue->TKFF_Token[where],
		    (curr - where) * sizeof(*(queue->TKFF_Token)));
	    flag = 1; /*  we insert tokens *before* other tokens  */
	}

    } else if((where + n) < curr)
	flag = 1; /*  we overwrite, but there are still some after us  */

    dest = &(queue->TKFF_Mem[queue->TKFF_MIndex]);
    queue->TKFF_MIndex += len;
    queue->TKFF_Index += i;
    for(i = 0 ; i < n ; i++) {
	const char *tokString;
	unsigned int pos;

	pos = where + i;

	/*  Copy token at i to pos  */
	queue->TKFF_Token[pos].TK_Flags = tokens[i].TK_Flags;
	queue->TKFF_Token[pos].TK_Value = tokens[i].TK_Value;
	tokString = tokens[i].TK_String;

	if(tokString == NULL) {
	    queue->TKFF_Token[pos].TK_String = NULL;
	    continue;
	}

	if(i >= 1 && tokens[i - 1].TK_String == tokString) {
	    /*  Preserve property that string pointers are identical  */
	    queue->TKFF_Token[pos].TK_String = queue->TKFF_Token[pos - 1].TK_String;
	    continue;
	}

	if(i >= 2 && tokens[i - 2].TK_String == tokString) {
	    /*  Preserve property that string pointers are identical  */
	    queue->TKFF_Token[pos].TK_String = queue->TKFF_Token[pos - 2].TK_String;
	    continue;
	}

	if(indfpos != -1 && (pos % step) == indfpos) {
	    /*  The file names are shared, so no need to copy  */
	    queue->TKFF_Token[pos].TK_String = tokString;
	    continue;
	}

	if(mode & JCTOK_MODE_CHKPTR) {
	    unsigned int ptroff;

	    if((ptroff = _jctokfindptr(queue, tokString)) != (unsigned int)-1) {
		queue->TKFF_Token[pos].TK_String = &(queue->TKFF_Mem[ptroff]);
		continue;
	    }

	    ptroff = dest - queue->TKFF_Mem;
	    if(_jctokaddptr(queue, tokString, ptroff) == -1)
		EGOTO(error);
	}

	/*  Use new space  */
	queue->TKFF_Token[pos].TK_String = dest;

	len = strlen(tokString) + 1;
	strcpy(dest, tokString);

	dest += len;

	if(flag)
	    /*  We inserted tokens before other tokens and used new
	     *  string memory, so the monotonicity is no longer preserved
	     */
	    queue->TKFF_Mode |= JCTOK_TYPE_FNOSTRORDER;
    }

    /*  Check that not too much string memory is used  */
    assert(dest == queue->TKFF_Mem + queue->TKFF_MIndex);

    return where;
 error:
    return -1;
}

/*JS*********************************************************************
 *   Do the memory allocation stuff for FIFO's.
 *************************************************************************/

static int _jctokinc(jctokfifo_t *queue, int nTokens, int stringLen)

/************************************************************************/
{
    /*  Check if space is available */
    if((queue->TKFF_Index + nTokens) > queue->TKFF_Max) {
	queue->TKFF_Max = MAX(queue->TKFF_Index + nTokens, queue->TKFF_Max +
			      FIFO_TOKINC);

	if((queue->TKFF_Token = (jsctoken_t *)
	    realloc(queue->TKFF_Token, queue->TKFF_Max *
		    sizeof(*queue->TKFF_Token))) == NULL)
	    EGOTO(error);
    }

    if(stringLen > 0 && (queue->TKFF_MIndex + stringLen) > queue->TKFF_MMax) {
	char *oldPtr = queue->TKFF_Mem;
	unsigned int i;

	/*  Increase memory for token strings  */
	queue->TKFF_MMax = MAX(queue->TKFF_MIndex + stringLen,
			       queue->TKFF_MMax + FIFO_CHARINC);
	if((queue->TKFF_Mem = (char*)
	    realloc(queue->TKFF_Mem, queue->TKFF_MMax)) == NULL) {
	    queue->TKFF_MMax = 0;
	    EGOTO(error);
	}

	/*  Change all pointers  */
	for(i = queue->TKFF_Start ; i < queue->TKFF_Index ; ++i)
	    if(queue->TKFF_Token[i].TK_String) {
		ptrdiff_t off;

		off = queue->TKFF_Token[i].TK_String - oldPtr;
		if(0 <= off && off < queue->TKFF_MIndex)
		    queue->TKFF_Token[i].TK_String = &queue->TKFF_Mem[off];
	    }
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int _jctokaddptr(jctokfifo_t *queue,
			const char *string, unsigned int offset)

/************************************************************************/
{
    if(queue->TKFF_NPtrMap >= queue->TKFF_MPtrMap) {
	queue->TKFF_MPtrMap += 20;
	if((queue->TKFF_PtrMap = (jctokfifoptrmap_t *)
	    realloc(queue->TKFF_PtrMap, queue->TKFF_MPtrMap *
		    sizeof(queue->TKFF_PtrMap[0]))) == NULL)
	    EGOTO(error);
    }

    queue->TKFF_PtrMap[queue->TKFF_NPtrMap].TKFPM_String = string;
    queue->TKFF_PtrMap[queue->TKFF_NPtrMap].TKFPM_Offset = offset;
    queue->TKFF_NPtrMap++;

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int _jctokfindptr(jctokfifo_t *queue, const char *string)

/************************************************************************/
{
    unsigned int i, n;

    n = queue->TKFF_NPtrMap;
    for(i = 0 ; i < n ; ++i) {
        unsigned int offset;

        if(queue->TKFF_PtrMap[i].TKFPM_String != string)
            continue;

        offset = queue->TKFF_PtrMap[i].TKFPM_Offset;
        if(strcmp(queue->TKFF_Mem + offset, string) == 0)
            return offset;
    }

    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jctokdelete(jctokfifo_t *queue, int start, int end, int mode)

/************************************************************************/
{
    if(end == -1)
	end = queue->TKFF_Index;

    if(start < end) {
	if((mode & 1) && !(queue->TKFF_Mode & JCTOK_TYPE_FNOSTRORDER)) {
	    /*  Optimize usage of string memory  */
	    char *s, *starts, *ends;
	    unsigned int i;

	    /*  Find start and end of string memory used by the region to be
	     *   deleted
	     */
	    starts = NULL;
	    ends = NULL;
	    for(i = start ; i < end ; ++i)
		if((s = (char *)queue->TKFF_Token[i].TK_String) != NULL &&
		   INSTRINGMEM(queue, s)) {
		    if(starts == NULL) starts = s;
		    ends = s;
		}

	    if(starts != NULL) {
		int shift;

		ends += strlen(ends) + 1;

		shift = ends - starts;

		if((i = &(queue->TKFF_Mem[queue->TKFF_MIndex]) - ends) > 0) {
		    memmove(starts, ends, i);

		    for(i = end ; i < queue->TKFF_Index ; ++i)
			if((s = (char *)queue->TKFF_Token[i].TK_String) != NULL &&
			   INSTRINGMEM(queue, s) && ends <= s)
			    queue->TKFF_Token[i].TK_String -= shift;
		}

		queue->TKFF_MIndex -= shift;
	    }
	}

	if(end < queue->TKFF_Index)
	    memmove(&(queue->TKFF_Token[start]), &(queue->TKFF_Token[end]),
		    (queue->TKFF_Index - end) * sizeof(queue->TKFF_Token[start]));

	queue->TKFF_Index -= end - start;
    }

    return queue->TKFF_Index;
}

/*JS*********************************************************************
 *   Implement a first-in-first-out queue for tokens: Get a token from the
 *    queue, if the queue is empty, CTK_EOF is returned.
 *************************************************************************/

jsctoken_t jctokget(jctokfifo_t *queue)

/************************************************************************/
{
    jsctoken_t token;

    if(queue->TKFF_Start < queue->TKFF_Index) {
	token = queue->TKFF_Token[queue->TKFF_Start];
	queue->TKFF_Start += queue->TKFF_Step;

	if(token.TK_Value == CTK_EOF) /*    last token read, clear queue  */
	    jctokclear(queue);

    } else {
	/*  FIFO is empty, so reset and return EOF token  */
	jctokclear(queue);

	token.TK_Flags  = 0;
	token.TK_Value  = CTK_EOF;
	token.TK_String = NULL;
    }

    return token;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jctokgetfpos(jctokfifo_t *queue, const char **pName)

/************************************************************************/
{
    int ind;

    if((ind = jctokindfpos(queue)) == -1)
	return -1;

    if(pName != NULL)
	*pName = queue->TKFF_Token[queue->TKFF_Start + ind].TK_String;

    return queue->TKFF_Token[queue->TKFF_Start + ind].TK_Value;
}

/*JS*********************************************************************
 *   Implement a first-in-first-out queue for tokens: Returns the index
 *    of the next token in the queue, -2 on EOF.
 *************************************************************************/

int jctokgetindex(jctokfifo_t *queue)

/************************************************************************/
{
    int ind;

    if(queue->TKFF_Start < queue->TKFF_Index) {
	ind = queue->TKFF_Start;
	queue->TKFF_Start += queue->TKFF_Step;

	if(queue->TKFF_Token[ind].TK_Value == CTK_EOF) {
	    jctokclear(queue);
	    ind = -2;
	}
    } else {
	/*  FIFO is empty, so reset and return EOF token  */
	jctokclear(queue);
	ind = -2;
    }

    return ind;
}

/*JS*********************************************************************
 *   Cast tokens up to (last-1) in a FIFO queue to blanks, so that only
 *    newlines survive.
 *************************************************************************/

void jctokblank(jctokfifo_t *queue, int last)

/************************************************************************/
{
    int i, j, n;

    n = queue->TKFF_Index;
    if(last < 0) last = n;

    for(i = j = queue->TKFF_Start ; i < n ; i++) {
	if(i < last) {
	    char *s;

	    if((s = (char *)queue->TKFF_Token[i].TK_String)) {
		char *d;
		int flag = FALSE;

		for(d = s ; *s ; s++)
		    if(*s == '\n') {
			*d++ = *s;
			flag = TRUE;
		    }

		if(flag) {
		    *d = '\0';
		    queue->TKFF_Token[j].TK_Flags = 0;
		    queue->TKFF_Token[j].TK_Value = CTK_BLANKSEOL;
		    queue->TKFF_Token[j].TK_String = queue->TKFF_Token[i].TK_String;
		    j++;
		}
		continue;

	    } else if(0 <= queue->TKFF_Token[i].TK_Value &&
		      queue->TKFF_Token[i].TK_Value < CTK_MINIMUM)
		continue;
	}

	queue->TKFF_Token[j++] = queue->TKFF_Token[i];
    }
    queue->TKFF_Index = j;

    return;
}

/*JS*********************************************************************
 *   General format: '%[<n1>[.<n2>]]X' where n1 and n2 are optional
 *    numbers. Known format specifiers:
 *     %o Substitute by the original tokens in [start,end]. If n1 is given,
 *        only the given token (1 == start, 2 == next one after start etc.)
 *        is substituted.
 *     %s Substitute it by the given string. If n1 is given, the string
 *        has at least this length, filled up with spaces.
 *     %d Substitute by the integer.
 *     %c Substitute by the given single character.
 *     %n Substitute by as many newlines as the original strings contains.
 *     %i Generate a newline + indentation. The indentation is the one
 *        of the original token in the line, at least n1. If n2 is present,
 *        it is added to it.
 *     %I The same as %i, but the indentation of the first non-space
 *        token in the line of the original token is used.
 *************************************************************************/

int jctoksubstitute(jctokfifo_t *queue, int start, int end,
                    const char *format, ...)

/************************************************************************/
{
    va_list ap;
    const char *s;
    char *dest;
    int len, indent, step;

    indent = -1;
    step = queue->TKFF_Step;

    /*   Initialize argument list    */
    va_start(ap, format);

    /*  Determine needed string memory  */
    for(len = 0, s = format ; *s ; ++s) {
	if(*s == '%') {
	    int n1, n2;

	    ++s;
	    if(isdigit(*s))
		n1 = strtol(s, (char **)&s, 10);
	    else
		n1 = -1;

	    if(*s == '.') {
		++s;
		if(isdigit(*s))
		    n2 = strtol(s, (char **)&s, 10);
		else
		    n2 = -1;
	    } else
		n2 = -1;

	    switch(*s) {
	    case 'o':
		if(n1 == -1) {
		    if(end >= queue->TKFF_Index)
			EGOTO(error);

		    for(n1 = start ; n1 <= end ; n1 += step)
			if(queue->TKFF_Token[n1].TK_Value >= 0)
			    len += CTK_ESTLENGTH(queue->TKFF_Token[n1]);
		} else {
		    if((n1 = start + (n1 - 1) * step) >= queue->TKFF_Index)
			EGOTO(error);

		    if(queue->TKFF_Token[n1].TK_Value >= 0)
			len += CTK_ESTLENGTH(queue->TKFF_Token[n1]);
		}
		break;

	    case 'n':
		for(n1 = start ; n1 <= end ; n1 += step)
		    if(queue->TKFF_Token[n1].TK_String)
			len += countNewLines(queue->TKFF_Token[n1].TK_String);
		    else if(queue->TKFF_Token[n1].TK_Value == '\n')
			++len;
		break;

	    case '%':
		++len;
		break;

	    case 's':
		{
		    const char *string;
		    int l;

		    string = va_arg(ap, const char *);
		    if((l = strlen(string)) < n1) /*  also works for n1 == -1  */
			l = n1;

		    len += l;
		}
		break;

	    case 'd':
		/*  That should be enough for integers, even on 64 bit machines  */
		len += sizeof(long) * CHAR_BIT;
		(void)va_arg(ap, int);
		break;

	    case 'c':
		++len;
		(void)va_arg(ap, int);
		break;

	    case 'i': case 'I':
		{
		    int l;

		    if(indent == -1)
			indent = jctokgetcolumn(queue, start, *s == 'I' ? 2 : 0);

		    l = 1 + MAX(indent, n1); /*  newline + indentation  */
		    if(n2 != -1) l += n2;

		    len += l;
		}
		break;

	    default:
		EGOTO(error);
	    }
	} else
	    ++len;
    }

    va_end(ap);

    va_start(ap, format);

    if(_jctokinc(queue, 0, len + 1) == -1)
	EGOTO(error);

    /*  Now print tokens  */
    dest = &queue->TKFF_Mem[queue->TKFF_MIndex];
    for(len = 0, s = format ; *s ; ++s) {
	if(*s == '%') {
	    int n1, n2;

	    ++s;
	    if(isdigit(*s))
		n1 = strtol(s, (char **)&s, 10);
	    else
		n1 = -1;

	    if(*s == '.') {
		++s;
		if(isdigit(*s))
		    n2 = strtol(s, (char **)&s, 10);
		else
		    n2 = -1;
	    } else
		n2 = -1;

	    switch(*s) {
	    case 'o':
		if(n1 == -1) {
		    if(end >= queue->TKFF_Index)
			EGOTO(error);

		    for(n1 = start ; n1 <= end ; n1 += step) {
			/*len+= sprintf(dest+len,"[%d %d %d]",off,end,step);*/
			if(queue->TKFF_Token[n1].TK_Value >= 0)
			    len += sprintCToken(dest + len, queue->TKFF_Token[n1]);
		    }
		} else {
		    n1 = start + (n1 - 1) * step;
		    if(n1 >= queue->TKFF_Index)
			EGOTO(error);

		    if(queue->TKFF_Token[n1].TK_Value >= 0)
			len += sprintCToken(dest + len, queue->TKFF_Token[n1]);
		}
		break;

	    case 'n':
		{
		    int count;

		    count = 0;
		    for(n1 = start ; n1 <= end ; n1 += step)
			if(queue->TKFF_Token[n1].TK_String)
			    count += countNewLines(queue->TKFF_Token[n1].TK_String);
			else if(queue->TKFF_Token[n1].TK_Value == '\n')
			    ++count;

		    while(count-- > 0)
			dest[len++] = '\n';
		}
		break;

	    case '%':
		dest[len++] = *s;
		break;

	    case 's':
		{
		    const char *string;
		    int l;

		    string = va_arg(ap, const char *);

		    strcpy(dest + len, string);
		    l = strlen(string);
		    len += l;

		    while(l++ < n1)  /*  need to fill up until length n1 + 1  */
			dest[len++] = ' ';
		}
		break;

	    case 'd':
		len += sprintf(dest + len, "%d", va_arg(ap, int));
		break;

	    case 'c':
		dest[len++] = (char)va_arg(ap, int);
		break;

	    case 'i': case 'I':
		{
		    int count;

		    dest[len++] = '\n';
		    count = MAX(indent, n1);
		    if(n2 != -1) count += n2;

		    while(count-- > 0)
			dest[len++] = ' ';
		}
		break;

	    default:
		EGOTO(error);
	    }
	} else
	    dest[len++] = *s;
    }
    dest[len] = '\0';

    va_end(ap);

    queue->TKFF_MIndex += len + 1;

    queue->TKFF_Token[start].TK_Flags  = 0;
    queue->TKFF_Token[start].TK_Value  = CTK_VANILLA;
    queue->TKFF_Token[start].TK_String = dest;

    for(start += step ; start <= end ; start += step) {
	queue->TKFF_Token[start].TK_Flags  = 0;
	queue->TKFF_Token[start].TK_Value  = CTK_VANILLA;
	queue->TKFF_Token[start].TK_String = "";
    }

    return 0;
 error:
    va_end(ap);
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

static int countNewLines(const char *string)

/************************************************************************/
{
    int count;

    count = 0;
    while((string = strchr(string, '\n'))) {
	++string;
	++count;
    }

    return count;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jctoksubststring(jctokfifo_t *queue, int start, int end,
		     const char *string)

/************************************************************************/
{
    int step, len;
    char *dest;

    len = strlen(string);
    if(_jctokinc(queue, 0, len + 1) == -1)
	EGOTO(error);

    dest = &queue->TKFF_Mem[queue->TKFF_MIndex];
    strcpy(dest, string);
    queue->TKFF_MIndex += len + 1;

    queue->TKFF_Token[start].TK_Flags  = 0;
    queue->TKFF_Token[start].TK_Value  = CTK_VANILLA;
    queue->TKFF_Token[start].TK_String = dest;

    step = queue->TKFF_Step;

    for(start += step ; start <= end ; start += step) {
	queue->TKFF_Token[start].TK_Flags  = 0;
	queue->TKFF_Token[start].TK_Value  = CTK_VANILLA;
	queue->TKFF_Token[start].TK_String = "";
    }

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

char *jctoktostring2(jctokfifo_t *queue, int start, int end,
                     jsctoken_t (*converter)(void *user, jsctoken_t token),
                     void *user)

/************************************************************************/
{
    char *dest;
    jsctoken_t token;
    int off, len, step;

    step = queue->TKFF_Step;

    len = 1;
    for(off = start ; off <= end ; off += step) {
	token = queue->TKFF_Token[off];

	if(converter) token = converter(user, token);

	if(token.TK_Value >= 0)
	    len += CTK_ESTLENGTH(token);
    }

    if((dest = (char *)malloc(len)) == NULL)
	return NULL;

    len = 0;
    for(off = start ; off <= end ; off += step) {
	token = queue->TKFF_Token[off];

	if(converter) token = converter(user, token);

	if(token.TK_Value >= 0)
	    len += sprintCToken(dest + len, token);
    }

    return dest;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jctoktostrbuff1(jstrbuff_t *sbuffer, jsctoken_t token)

/************************************************************************/
{
    int len;

    if(token.TK_Value < 0)
	return 0;

    if(jstrbuffstore(sbuffer, NULL, CTK_ESTLENGTH(token)) == -1)
	EGOTO(error);

    if((len = sprintCToken(jstrbuffer(sbuffer) + jstrbufflen(sbuffer),
			   token)) == -1)
	EGOTO(error);

    jstrbuffinc(sbuffer, len);

    return 0;
 error:
    return -1;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int jctoktostrbuff(jstrbuff_t *sbuffer, jctokfifo_t *queue,
                   int start, int end,
                   jsctoken_t (*converter)(void *user, jsctoken_t token),
                   void *user)

/************************************************************************/
{
    char *dest;
    jsctoken_t token;
    int off, len, step;

    step = queue->TKFF_Step;

    len = 0;
    for(off = start ; off <= end ; off += step) {
	token = queue->TKFF_Token[off];

	if(converter) token = converter(user, token);

	if(token.TK_Value >= 0)
	    len += CTK_ESTLENGTH(token);
    }

    /*  Make sure we have enough space left  */
    if(jstrbuffstore(sbuffer, NULL, len) == -1)
	return -1;

    dest = jstrbuffer(sbuffer);
    len = jstrbufflen(sbuffer);
    for(off = start ; off <= end ; off += step) {
	token = queue->TKFF_Token[off];

	if(converter) token = converter(user, token);

	if(token.TK_Value >= 0)
	    len += sprintCToken(dest + len, token);
    }

    assert(len <= sbuffer->SB_MBuffer);

    jstrbuffset(sbuffer, len);

    return 0;
}

/*JS*********************************************************************
 *   Find column of the token at position 'indx' in the queue.
 *    mode bit 0: If set, return -1 if there are non-space tokens between
 *                the line start and the token.
 *         bit 1: If set, return the column of the first non-space token
 *                in the line. Tokens that extend across lines (e. g.
 *                multiline strings) are not considered.
 *         bit 2: Same as bit 1, but return '-1 - newcolumn' in case the
 *                first non-space token is not us and 'newcolumn' in case
 *                it is us.
 *************************************************************************/

int jctokgetcolumn(jctokfifo_t *queue, int indx, int mode)

/************************************************************************/
{
    int step, curri, column;

    step = jctokgetstep(queue);

    /*  First search for the last token before us that contains a newline  */
    column = 0; /*  the default if we are at the first token  */
    for(curri = indx - step ; curri >= 0 ; curri -= step) {
	int value;
	const char *string, *t;

	value = queue->TKFF_Token[curri].TK_Value;
	string = queue->TKFF_Token[curri].TK_String;

	if((mode & 1) &&
	   !(CTK_BLANK_START <= value && value <= CTK_BLANK_END))
	    return -1;  /*  there is something non-space adhead of us  */

	if(value == CTK_CPPCOMMENT)
	    break;  /*  column is already 0  */

	if(string && (t = strchr(string, '\n')) != NULL) {
	    /*  Little optimization: Calculate starting from the newline  */
	    column = updateColumn(0, t + 1);
	    break;
	}
    }

    /*  We already looked at 'curri', so include the tokens that
     *   follow. This also works if curri < 0, i. e. we hit the
     *   start of the queue
     */
    for(curri += step ; curri < indx ; curri += step) {
	if(mode & 6) {
	    int value = queue->TKFF_Token[curri].TK_Value;

	    if(!(CTK_BLANK_START <= value && value <= CTK_BLANK_END))
		/*  This is a non-space token, so stop  */
		return (mode & 4) ? -1 - column : column;
	}

	column = updcolCToken(queue->TKFF_Token[curri], column);
    }

    return column;
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jctokshowqueue(FILE *fp, jctokfifo_t *queue)

/************************************************************************/
{
    jctokshowqueue2(fp, queue, 0, queue->TKFF_Index);
}

/*JS*********************************************************************
 *
 *************************************************************************/

void jctokshowqueue2(FILE *fp, jctokfifo_t *queue, int start, int end)

/************************************************************************/
{
    int i, cols, step;

    step = queue->TKFF_Step;

    cols = (SCREENWIDTH - 5) / step;

    for(i = start ; i < end ; ++i) {
	const char *s;
	char a[2];
	int l;

	if((s = queue->TKFF_Token[i].TK_String) == NULL) {
	    int value;

	    value = queue->TKFF_Token[i].TK_Value;
	    if(0 <= value && value < 256) {
		a[0] = value;
		a[1] = '\0';
		s = a;
	    }
	}

	if((i % step) == 0)
	    fprintf(fp, "%3d: ", i / step);

	l = fprintf(fp, "%3d %16p ", queue->TKFF_Token[i].TK_Value, s);
	if(s) {
	    fputc('\'', fp);
	    l += 2 + cStrFPrint(fp, s, strlen(s));
	    fputc('\'', fp);
	}

	if((i % step) == (step - 1))
	    fprintf(fp, "\n");
	else if(l < cols)
	    fprintf(fp, "%*s", cols - l, "");
    }
}
