/*JS*********************************************************************
*
*    Program : EXTFILE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Extension to ANSI "FILE":
*      - An arbitrary amount of characters may be unget.
*      - Reading from strings also possible.
*      - May push modules on the input queue, so that preprocessing of
*        stream is possible.
*        A module routine
*            int module(void *uPtr,int *pC);
*        is expected to process its input character *pC and return
*        it in *pC. The return value of the routine may have the
*        following values:
*          EMOD_IDLE:
*            Nothing has been done, so this module will be called again.
*            Only necessary if one wants to avoid "goto" in the module.
*          EMOD_IN:
*            A one to none processing has been done: *pC -> -, i.e. the input
*             character is simply discarded.
*          EMOD_OUT:
*            A none to one processing - -> *pC has been done, i.e. the input
*             character has not been used. It is saved and will be re-input
*             the next time the module is called.
*          EMOD_THROUGH:
*            A one to one processing *pC -> *pC has been done.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <jsconfig.h>

#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jsllio.h>

#include <jssubs.h>

#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

/*********     DEFINES                                          *********/
#ifndef EMOD_IDLE
/*  START-DEFINITIONS  */
#include <stdio.h>   /*	 For "FILE" definition    */

/*  Basic definition for extended file pointer  */
typedef struct {
  int FM_UngetChar;
  int (*FM_Module)(void *uPtr, int *pC);
  void *FM_User;
} efilemod_t;

/*  Return values for modules  */
#define EMOD_IDLE  0  /*  nothing done, call module again  */
#define EMOD_IN    1  /*  input char used  */
#define EMOD_OUT   2  /*  output char produced  */
#define EMOD_THROUGH  (EMOD_IN | EMOD_OUT)

typedef struct {
  long          EF_Flags;

  /*   > 0: current linen number in file.
   *  <= 0: current negated index in string
   */
  int           EF_LineNr;
  union {
    FILE       *ef_FPtr;
    void       *ef_Ptr;
    const char *ef_String;
  } ef_Cont;
  int          *EF_UngetBuff;
  unsigned int  EF_UngetIndex, EF_UngetMax;

  efilemod_t  *EF_ModStack; /*  Stack of file modules  */
  int           EF_ModMax;
  union {
    int       (*ef_PPut)(int c, void *p);
    int       (*ef_PGet)(void *p, int flag);
  } ef_Cont2;
  void         *EF_User;
} extfile_t;

#define EF_FPtr    ef_Cont.ef_FPtr
#define EF_Ptr     ef_Cont.ef_Ptr
#define EF_String  ef_Cont.ef_String

#define EF_PPut    ef_Cont2.ef_PPut
#define EF_PGet    ef_Cont2.ef_PGet

/*  If reading from a string, get the current position  */
#define efstrpos(efp)    ((efp)->EF_String - (efp)->EF_LineNr - \
			  (efp)->EF_UngetIndex)

/*  For EF_Flags field  */
#define EFFLAG_EOF       (1<<0)
#define EFFLAG_ERROR     (1<<1)
#define EFFLAG_PIPE      (1<<2)
#define EFFLAG_FREESTR   (1<<3)
#define EFFLAG_BUFFLINE  (1<<4)
#define EFFLAG_WRITE     (1<<5)
#define EFFLAG_INPUTEOF  (1<<6)

#define EFFLAG_USER      (1<<16)

#define efeof(efp)       ((efp)->EF_PGet ? \
			  ((efp)->EF_Flags & EFFLAG_INPUTEOF) : \
			  (efp)->EF_LineNr > 0 ? \
                          (efp->EF_PGet == NULL && feof((efp)->EF_FPtr)) : \
                          ((efp)->EF_String[-(efp)->EF_LineNr]) == '\0')
#define efcleareof(efp)  ((efp)->EF_Flags &= ~(EFFLAG_EOF|EFFLAG_INPUTEOF))

#define eferror(efp)     ((efp)->EF_Flags & EFFLAG_ERROR)

#define eputc(c, efp)     (efwrite(&c, 1, efp) == 1 ? (c) : EOF)
#define efputs(s, efp)    efwrite(s, strlen(s), efp)

/*  Close an extended file descriptor without closing the underlying
 *   stdio file descriptor
 */
#define efclose2(efp)     ((efp)->EF_FPtr = NULL, efclose(efp))
/*  END-DEFINITIONS  */
#endif

/*  We may supply our own opener routines that open special kind of
 *   files, recognized by a prefix
 */
typedef struct efopener_t efopener_t;
struct efopener_t {
  efopener_t *EFO_Next;

  const char *EFO_Prefix;
  extfile_t *(*EFO_Open)(void *user, const char *postfix,
			 const char *mode);
  void *EFO_User;
};

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE  1
#endif

/*  I'm not sure if 'EOF' is always defined to -1  */
#define MYEOF   (-1)
#define MYEMPTY (-2)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype extfile_t     *efopen(const char *name, const char *mode);
Prototype extfile_t     *effopen(FILE *fp, int iswrite);
Prototype extfile_t     *efmkdest(int (*pPut)(int c, void *p), void *p);
Prototype extfile_t     *efmksrc(int (*pGet)(void *, int), void *p);

Prototype int            efclose(extfile_t *efp);
Prototype FILE          *efileptr(extfile_t *efp);

Prototype int            efread(void *ptr, long size, extfile_t *efp);
Prototype int            efwrite(const void *ptr, long size, extfile_t *efp);

Prototype int            efgetc(extfile_t *efp);
Prototype int            eungetc(int c, extfile_t *efp);

Prototype int            efputc(int c, extfile_t *efp);

Prototype int            epushmod(extfile_t *efp, int (*module)(void *, int *pC),
				  void *user, int where);
Prototype int            epopmod(extfile_t *efp, int (*module)(void *, int *pC),
				 void **puser);

Prototype int            epushopener(const char *prefix,
				     extfile_t *(*opener)(void *user,
							  const char *postfix,
							  const char *mode),
				     void *user);
Prototype int            epopopener(const char *prefix);


Prototype const char    *EFilePath;

/*********     GLOBAL VARIABLES                                 *********/
const char *EFilePath;

static efopener_t *EOpener;

/*JS*********************************************************************
*   Open a extented file. If mode is NULL, name is a string to read from.
*    If name is NULL, standard input is used.
*************************************************************************/

extfile_t *efopen(const char *name, const char *mode)

/************************************************************************/
{
  extfile_t *efp;

  if((efp = (extfile_t *)malloc(sizeof(*efp))) == NULL)
    return(NULL);

  /*  Preset structure  */
  efp->EF_Flags = 0;

  efp->EF_UngetBuff = NULL;
  efp->EF_UngetIndex = efp->EF_UngetMax = 0;

  efp->EF_ModStack = NULL;
  efp->EF_ModMax = 0;

  efp->EF_PPut = NULL;

  efp->EF_User = NULL;

  if(mode && mode[0] != 's') {
    const char *mode2;

    efp->EF_LineNr = 1;

    /*  FIXME: This is too simple, but for now it is sufficient  */
    if(strchr(mode, 'r')) {
      mode2 = "r";
    } else {
      efp->EF_Flags |= EFFLAG_WRITE;
      mode2 = "w";
    }

    if(name == NULL)
      efp->EF_FPtr = (efp->EF_Flags & EFFLAG_WRITE) ? stdout : stdin;
      /*  If there is a 'f' in the mode, force it to be a file
       *   and prevent executing commands
       */
    else if(!strchr(mode, 'f') && (name[0] == '|' || strchr(mode, 'p'))) {
      if((efp->EF_FPtr = popen(name[0] == '|' ? name + 1 : name,
			       mode2)) == NULL)
	EGOTO(error);
      efp->EF_Flags |= EFFLAG_PIPE;

    } else {
      efopener_t *efo;
      const char *s;

      for(efo = EOpener ; efo ; efo = efo->EFO_Next)
	if((s = jstrstartswith(name, efo->EFO_Prefix)) != NULL) {
	  free(efp);
	  return efo->EFO_Open(efo->EFO_User, s, mode2);
	}

      if((efp->EF_FPtr = fopenPath2(name, mode2, EFilePath)) == NULL)
	EGOTO(error);
    }
  } else {
    efp->EF_LineNr = 0;
    if(mode) {
      if((efp->EF_String = strdup(name)) == NULL)
	EGOTO(error);

      efp->EF_Flags |= EFFLAG_FREESTR;
    } else
      efp->EF_String = name;
  }

  return efp;
error:
  free(efp);
  return NULL;
}

/*JS*********************************************************************
*   Makes a EFile from a FILE structure.
*************************************************************************/

extfile_t *effopen(FILE *fp, int iswrite)

/************************************************************************/
{
  extfile_t *efp;

  if((efp = (extfile_t *)malloc(sizeof(*efp))) == NULL) return(NULL);

  /*  Preset structure  */
  efp->EF_Flags = iswrite ? EFFLAG_WRITE : 0;

  efp->EF_UngetBuff = NULL;
  efp->EF_UngetIndex = efp->EF_UngetMax = 0;

  efp->EF_ModStack = NULL;
  efp->EF_ModMax = 0;

  efp->EF_PPut = NULL;

  efp->EF_User = NULL;

  efp->EF_LineNr = 1;
  efp->EF_FPtr = fp;

  return(efp);
}

/*JS*********************************************************************
*   Creates an EFile opened for writing. The characters are written using
*    routine pPut, p is an arbitrary pointer that is passed to the routine
*    when it is called. May be used to implement a program internal pipe
*    where modules may be pushed on.
*************************************************************************/

extfile_t *efmkdest(int (*pPut)(int c, void *p), void *p)

/************************************************************************/
{
  extfile_t *efp;

  if((efp = (extfile_t *)malloc(sizeof(*efp))) == NULL) return(NULL);

  /*  Preset structure  */
  efp->EF_Flags = EFFLAG_WRITE;
  efp->EF_LineNr = 1;
  efp->EF_Ptr = p;

  efp->EF_UngetBuff = NULL;
  efp->EF_UngetIndex = efp->EF_UngetMax = 0;

  efp->EF_ModStack = NULL;
  efp->EF_ModMax = 0;

  efp->EF_PPut = pPut;

  efp->EF_User = NULL;

  return efp;
}

/*JS*********************************************************************
*   Creates an EFile opened for reading. The characters are read using
*    routine pGet, p is an arbitrary pointer that is passed to the
*    routine when it is called. May be used to implement a program
*    internal pipe where modules may be pushed on.
*************************************************************************/

extfile_t *efmksrc(int (*pGet)(void *, int), void *p)

/************************************************************************/
{
  extfile_t *efp;

  if((efp = (extfile_t *)malloc(sizeof(*efp))) == NULL)
    return NULL;

  /*  Preset structure  */
  efp->EF_Flags = 0;
  efp->EF_LineNr = 1;
  efp->EF_Ptr = p;

  efp->EF_UngetBuff = NULL;
  efp->EF_UngetIndex = efp->EF_UngetMax = 0;

  efp->EF_ModStack = NULL;
  efp->EF_ModMax = 0;

  efp->EF_PGet = pGet;

  efp->EF_User = NULL;

  return efp;
}

/*JS*********************************************************************
*   Closes an extended file.
*************************************************************************/

int efclose(extfile_t *efp)

/************************************************************************/
{
  int i, ret;

  ret = 0;
  for(i = efp->EF_ModMax - 1 ; i >= 0 ; --i)
    if(efp->EF_ModStack[i].FM_Module != NULL &&
       efp->EF_ModStack[i].FM_Module(efp->EF_ModStack[i].FM_User, NULL) == -1)
      ret = -1;

  if(efp->EF_Flags & EFFLAG_WRITE) {
    if(efp->EF_PPut)
      /*  Let the underlying stream clean up  */
      (void)efp->EF_PPut(-1, efp->EF_Ptr);
  } else {
    if(efp->EF_PGet)
      /*  Let the underlying stream clean up  */
      (void)efp->EF_PGet(efp->EF_Ptr, -1);
  }

  if(efp->EF_LineNr <= 0 || efp->EF_PGet) {
    if(efp->EF_Flags & EFFLAG_FREESTR)
      free((void *)efp->EF_String);
  } else if(efp->EF_FPtr && efp->EF_FPtr != stdin) {
    /*  Don't close standard input  */
    if(efp->EF_Flags & EFFLAG_PIPE) {
      if((i = pclose(efp->EF_FPtr)) == -1)
	ret = -1;
      else if(ret != -1)
	/*  Communicate the return value of pclose  */
	ret = i;
    } else if(fclose(efp->EF_FPtr) == EOF)
      ret = -1;
  }

  free(efp->EF_ModStack);
  free(efp->EF_UngetBuff);
  free(efp);

  return ret;
}

/*JS*********************************************************************
*   Gets the underlying FILE pointer of an extented file.
*************************************************************************/

FILE *efileptr(extfile_t *efp)

/************************************************************************/
{
  /*  Make sure that extented file has not been created via
   *   efmkdest or efmksrc
   */
  if(efp->EF_PPut)
    return NULL;

  return efp->EF_FPtr;
}

/*JS*********************************************************************
*
*************************************************************************/

int efread(void *ptr, long size, extfile_t *efp)

/************************************************************************/
{
  long i;
  int c;

  if(efp->EF_Flags & EFFLAG_EOF)
    return 0;

  if(efp->EF_Flags & (EFFLAG_WRITE|EFFLAG_ERROR))
    return -1;

  i = 0;

  /*  Any unget character?  */
  for( ; i < size && efp->EF_UngetIndex > 0 ; ++i)
    ((char *)ptr)[i] = efp->EF_UngetBuff[--(efp->EF_UngetIndex)];

  if(i < size) {
    if(efp->EF_ModMax > 0) {
      while(i < size) {
	int start;

	start = efp->EF_ModMax - 1;
	c = efp->EF_ModStack[start].FM_UngetChar;
#if MYEOF != EOF
	if(c == MYEOF) c = EOF;
#endif

	do {
	  int save, flag;

	  if(start < 0) {
	    if(efp->EF_Flags & EFFLAG_INPUTEOF) {
	      c = EOF;
	    } else {
	      /* **  Process real input stream  ** */
	      if(efp->EF_PGet) {
		c = efp->EF_PGet(efp->EF_Ptr, 0);

	      } else if(efp->EF_LineNr > 0) {
		c = getc(efp->EF_FPtr);

		if(c == EOF && !feof(efp->EF_FPtr))
		  EGOTO(error);
	      } else {
		/*  Read from string (prevent c from becoming negative)  */
		if((c = ((unsigned char *)efp->EF_String)[-efp->EF_LineNr]))
		  efp->EF_LineNr--;
		else
		  c = EOF;
	      }

	      /*  Avoid calling the input functions again if EOF
	       *   has been reached
	       */
	      if(c == EOF)
		efp->EF_Flags |= EFFLAG_INPUTEOF;
	    }

	    /*  Go to first module  */
	    start = 0;
	    continue;
	  }

	  /*  No input character for current module  */
	  if(c == MYEMPTY) {
	    start--;
	    if(start >= 0) {
	      c = efp->EF_ModStack[start].FM_UngetChar;
#if MYEOF != EOF
	      if(c == MYEOF) c = EOF;
#endif
	    }
	    continue;
	  }

	  /* **  Process module start  ** */
	  save = c;
	  if(efp->EF_ModStack[start].FM_Module == NULL) {
	    /*  Simulate module not doing anything */
	    flag = EMOD_IN | EMOD_OUT;

	  } else if((flag = efp->EF_ModStack[start].
	      FM_Module(efp->EF_ModStack[start].FM_User, &c)) == -1)
	    EGOTO(error);

	  /*  Store unused input value */
	  if((flag & EMOD_IN) && save != EOF) {
	    /* -- Input value used and was not EOF -- */
	    efp->EF_ModStack[start].FM_UngetChar = MYEMPTY;

	  } else {
	    /* -- Input value not used -- */
	    efp->EF_ModStack[start].FM_UngetChar =
	      (save == EOF) ? MYEOF : save;
	  }

	  if(flag & EMOD_OUT) {
	    /* --  Output value generated, next module  -- */
	    start++;

	  } else {
	    c = efp->EF_ModStack[start].FM_UngetChar;
#if MYEOF != EOF
	    if(c == MYEOF) c = EOF;
#endif
	  }
	} while(start < efp->EF_ModMax);

	/* Keep EOF, so next call will yield EOF without calling any module */
	if(c == EOF) {
	  efp->EF_Flags |= EFFLAG_EOF;
	  break;
	}

	((char *)ptr)[i++] = c;
	if(c == '\n' && (efp->EF_Flags & EFFLAG_BUFFLINE))
	  break;
      }

    } else if(efp->EF_PGet) {
      while(i < size) {
	if(efp->EF_Flags & EFFLAG_INPUTEOF)
	  break;

	if((c = efp->EF_PGet(efp->EF_Ptr, 0)) == EOF) {
	  efp->EF_Flags |= EFFLAG_INPUTEOF;
	  break;
	}

	((char *)ptr)[i++] = c;

	if(c == '\n' && (efp->EF_Flags & EFFLAG_BUFFLINE))
	  break;
      }

    } else if(efp->EF_LineNr > 0) {
      int j = size - i;

      if(j == 1) {
	if((c = getc(efp->EF_FPtr)) == EOF)
	  i = 0;
	else
	  ((char *)ptr)[i++] = c;
      } else {
	if((c = fread((char *)ptr + i, 1, j, efp->EF_FPtr)) == -1)
	  EGOTO(error);

	i += c;
	if(i == 0 && !feof(efp->EF_FPtr))
	  EGOTO(error);
      }

    } else {
      c = strlen(efp->EF_String - efp->EF_LineNr);

      if(c > size - i) c = size - i;
      memcpy((char *)ptr + i, efp->EF_String - efp->EF_LineNr, c);
      efp->EF_LineNr -= c;
      i += c;
    }
  }

  if(efp->EF_LineNr > 0) { /*  Update line counter  */
    const char *s, *e;

    for(s = (const char *)ptr, e = s + i ; s < e &&
	  (s = (const char *)memchr(s, '\n', e - s)) ; ++s)
      efp->EF_LineNr++;
  }

  return i;
error:
  efp->EF_Flags |= EFFLAG_ERROR;
  return -1;
}

/*JS*********************************************************************
*   Write characters on an extented file stream, processing any modules if
*    necessary.
*************************************************************************/

int efwrite(const void *ptr, long size, extfile_t *efp)

/************************************************************************/
{
  long i;

  if(!(efp->EF_Flags & EFFLAG_WRITE))
    return -1;

  if(efp->EF_ModMax == 0 && efp->EF_PPut == NULL)
    return fwrite(ptr, 1, size, efp->EF_FPtr);

  for(i = 0 ; i < size ; ++i) {
    int start, c;

    /*  If some modules have stopped processing, break immediately  */
    if(efp->EF_Flags & EFFLAG_EOF) break;

    c = ((char *)ptr)[i];

    /*  Invariant condition is:
     *    forall i > start: efp->EF_ModStack[i].FM_UngetChar == MYEMPTY
     */
    start = 0;
    do {
      int save, flag;

      if(c == EOF) {
	efp->EF_Flags |= EFFLAG_EOF;
	break;
      }

      if(start == efp->EF_ModMax) {
	/* **  Process real output stream  ** */
	if((efp->EF_PPut ? efp->EF_PPut(c, efp->EF_Ptr) :
	    putc(c, efp->EF_FPtr)) == EOF)
	  EGOTO(error);

	c = MYEMPTY; /*  Fall through to next if condition  */
      }

      /*  No input character for current module  */
      if(c == MYEMPTY) {
	start--;
	if(start >= 0) c = efp->EF_ModStack[start].FM_UngetChar;
	continue;
      }

      /* **  Process module start  ** */
      save = c;
      if(efp->EF_ModStack[start].FM_Module == NULL) {
	/*  Simulate module not doing anything */
	flag = EMOD_IN | EMOD_OUT;

      } else if((flag = efp->EF_ModStack[start].
		 FM_Module(efp->EF_ModStack[start].FM_User, &c)) == -1)
	EGOTO(error);

      if(flag & EMOD_IN)
	/* -- Input value used -- */
	efp->EF_ModStack[start].FM_UngetChar = MYEMPTY;
      else
	/* -- Input value not used -- */
	efp->EF_ModStack[start].FM_UngetChar = save;

      if(flag & EMOD_OUT) {
	start++; /*  Go to next module  */
      } else {
	c = efp->EF_ModStack[start].FM_UngetChar;
      }
    } while(start >= 0);
  }

  return i;
error:
  efp->EF_Flags |= EFFLAG_ERROR;
  return -1;
}

/*JS*********************************************************************
*   Gets a single character from an extended file processing any modules
*    if necessary.
*************************************************************************/

int efgetc(extfile_t *efp)

/************************************************************************/
{
  unsigned char c;

  /*  Treat most common case more quickly. I can't pack this into
   *   a macro due to the EF_LineNr increment :(
   */
  if(efp->EF_UngetIndex == 0 && efp->EF_ModMax == 0 &&
     efp->EF_PGet == NULL && efp->EF_LineNr > 0) {
    int ic;

    if((ic = getc(efp->EF_FPtr)) == '\n')
      efp->EF_LineNr++;

    return ic;
  }

  /*  Slow, but full mode :( */
  if(efread(&c, 1, efp) <= 0)
    return EOF;

  return c;
}

/*JS*********************************************************************
*   Puts a character back. Returns c or EOF on error.
*************************************************************************/

int eungetc(int c, extfile_t *efp)

/************************************************************************/
{
  int i;

  if(c == EOF) {
    efp->EF_Flags |= EFFLAG_EOF;
    return c;
  }

  i = efp->EF_UngetIndex;
  if(i >= efp->EF_UngetMax) {
    /*  Increase unget buffer  */
    if((efp->EF_UngetBuff = (int *)
	realloc(efp->EF_UngetBuff, (efp->EF_UngetMax + 20) *
		sizeof(*(efp->EF_UngetBuff)))) == NULL)
      EGOTO(error);
    efp->EF_UngetMax += 20;
  }

  efp->EF_UngetBuff[i] = c;
  efp->EF_UngetIndex++;

  /*  Decrement line counter. It must not become zero!  */
  if(efp->EF_LineNr > 1 && c == '\n') efp->EF_LineNr--;

  return(c);
error:
  efp->EF_Flags |= EFFLAG_ERROR;
  return(EOF);
}

/*JS*********************************************************************
*
*************************************************************************/

int efputc(int c, extfile_t *efp)

/************************************************************************/
{
  if(efwrite(&c, 1, efp) != 1) return(EOF);
  return(c);
}

/*JS*********************************************************************
*   Push a module on the module queue. Modules pushed with where == 0 are
*    inserted at the head of the module queue and process the input stream
*    first (output stream last). Modules pushed with where == -1 process
*    the input stream last (output stream first).
*************************************************************************/

int epushmod(extfile_t *efp, int (*module)(void *uPtr, int *pC),
	     void *user, int where)

/************************************************************************/
{
  int i;

  /*  Increase module stack  */
  if((efp->EF_ModStack = (efilemod_t *)
      realloc(efp->EF_ModStack, (efp->EF_ModMax + 1) *
	      sizeof(*(efp->EF_ModStack)))) == NULL)
    return(-1);

  if(where > efp->EF_ModMax || where < 0) where = efp->EF_ModMax;

  for(i = efp->EF_ModMax ; i > where ; i--)
    efp->EF_ModStack[i] = efp->EF_ModStack[i - 1];

  efp->EF_ModStack[where].FM_Module    = module;
  efp->EF_ModStack[where].FM_UngetChar = MYEMPTY;
  efp->EF_ModStack[where].FM_User      = user;

  (efp->EF_ModMax)++;

  return(0);
}

/*JS*********************************************************************
*   Removes a module from the input queue. If a module is not present,
*    return 1, < 0 in case of error.
*************************************************************************/

int epopmod(extfile_t *efp, int (*module)(void *uPtr, int *pC), void **puser)

/************************************************************************/
{
  int ret, i, oldc;

  ret = 0;
  for(i = efp->EF_ModMax - 1 ; i >= 0 ; i--)
    if(efp->EF_ModStack[i].FM_Module == module)
      break;

  if(i < 0)
    return 1; /*  module not found  */

  oldc = efp->EF_ModStack[i].FM_UngetChar;
  if(puser) *puser = efp->EF_ModStack[i].FM_User;

  if(oldc != MYEMPTY && oldc != MYEOF) {
    if(efp->EF_Flags & EFFLAG_WRITE) {
      if(i == (efp->EF_ModMax - 1)) {
	/* **  Process real output stream  ** */
	if((efp->EF_PPut ? efp->EF_PPut(oldc, efp->EF_Ptr) :
	    putc(oldc, efp->EF_FPtr)) == EOF)
	  ret = -1;
      } else if(efp->EF_ModStack[i + 1].FM_UngetChar == MYEMPTY) {
	efp->EF_ModStack[i + 1].FM_UngetChar = oldc;
      } else {
	/*  We can't move this character to another module, so
	 *   simulate an empty module
	 */
	efp->EF_ModStack[i].FM_Module = NULL;
	return ret;
      }
    } else {
      if(i == 0) {
	if(eungetc(oldc, efp) == EOF)
	  ret = -1;
      } else if(efp->EF_ModStack[i - 1].FM_UngetChar == MYEMPTY) {
	efp->EF_ModStack[i - 1].FM_UngetChar = oldc;
      } else {
	/*  We can't move this character to another module, so
	 *   simulate an empty module
	 */
	efp->EF_ModStack[i].FM_Module = NULL;
	return ret;
      }
    }
  }

  /*  Delete this module  */
  (efp->EF_ModMax)--;

  while(i < efp->EF_ModMax) {
    efp->EF_ModStack[i] = efp->EF_ModStack[i + 1];
    i++;
  }

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int epushopener(const char *prefix, extfile_t *(*opener)(void *user,
							const char *postfix,
							const char *mode),
		void *user)

/************************************************************************/
{
  efopener_t *efo;

  if((efo = (efopener_t *)malloc(sizeof(*efo))) == NULL)
    return -1;

  efo->EFO_Prefix = prefix;
  efo->EFO_Open   = opener;
  efo->EFO_User   = user;

  efo->EFO_Next = EOpener;
  EOpener = efo;

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int epopopener(const char *prefix)

/************************************************************************/
{
  efopener_t *efo, *prev;

  for(prev = NULL, efo = EOpener ; efo ; prev = efo, efo = efo->EFO_Next) {
    if(strcmp(efo->EFO_Prefix, prefix) == 0) {
      if(prev)
	prev->EFO_Next = efo->EFO_Next;
      else
	EOpener = efo->EFO_Next;

      free(efo);

      return 0;
    }
  }

  /*  not found  */
  return -1;
}
