/*JS*********************************************************************
*
*    Program : FLOCKS
*    Language: ANSI-C with UNIX fcntl style locking functions
*    Author  : Joerg Schoen
*    Purpose : Routines to lock regions of files.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <string.h>

#include <jsconfig.h>

#ifndef CONFIG_NO_POSIX
#include <jsllio.h>
#endif

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS */
#include <jsconfig.h>  /*  For configurational definitions  */
#include <stdio.h>   /*	 For "FILE" definition    */

/*  Definitions for routine lockRegion and fopenLock  */
/*  Don't change these values!  */
#define JSLOCK_READ     0
#define JSLOCK_READW    1
#define JSLOCK_WRITE    2
#define JSLOCK_WRITEW   3
#define JSLOCK_UNLOCK   4

#ifndef CONFIG_NO_POSIX
/*  Macros for usage of routine lockRegion2  */
#define lockRegion(fp, type, offset, whence, len)  \
	lockRegion2(fileno(fp), type, offset, whence, len)
#else
extern int lockRegion(FILE *fp, int type, long offset, int whence, long len);
#endif

#define lockRead(fp, o, w, l)   lockRegion(fp, JSLOCK_READ, o, w, l)
#define lockReadW(fp, o, w, l)  lockRegion(fp, JSLOCK_READW, o, w, l)

#define lockWrite(fp, o, w, l)  lockRegion(fp, JSLOCK_WRITE, o, w, l)
#define lockWriteW(fp, o, w, l) lockRegion(fp, JSLOCK_WRITEW, o, w, l)

#define lockRead2(fd, o, w, l)   lockRegion2(fd, JSLOCK_READ, o, w, l)
#define lockReadW2(fd, o, w, l)  lockRegion2(fd, JSLOCK_READW, o, w, l)

#define lockWrite2(fd, o, w, l)  lockRegion2(fd, JSLOCK_WRITE, o, w, l)
#define lockWriteW2(fd, o, w, l) lockRegion2(fd, JSLOCK_WRITEW, o, w, l)

#define unLock(fp, o, w, l)      lockRegion(fp, JSLOCK_UNLOCK, o, w, l)
#define unLock2(fd, o, w, l)     lockRegion2(fd, JSLOCK_UNLOCK, o, w, l)
/*  END-DEFINITIONS */

/*  Undo any effects of "dbgflock.h"  */
#undef lockRegion2

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            lockRegion2(int fd, int type, long offset, int whence,
				     long len);
Prototype FILE          *fopenLock(const char *name, const char *mode, int type);

/*********     GLOBAL / STATIC VARIABLES                        *********/
#ifndef CONFIG_NO_POSIX
static const int LockTypes[] = {
  F_RDLCK, F_RDLCK, F_WRLCK, F_WRLCK, F_UNLCK
};

/*JS*********************************************************************
*   Auxiliary routine to lock a region in a file.
*************************************************************************/

int lockRegion2(int fd, int type, long offset, int whence, long len)

/************************************************************************/
{
  struct flock lock;

#if 0
  /*  A little bit type checking  */
  if(type < 0 || type >= 5) return(-1);
#endif

  lock.l_type   = LockTypes[type];
  lock.l_start  = offset;
  lock.l_whence = whence;
  lock.l_len    = len;

  return(fcntl(fd, (type & 1) ? F_SETLKW : F_SETLK, &lock));
}

/*JS*********************************************************************
*   Same as fopen, but locks the file too. Simulates MANDATORY locking,
*    so it prevents the file from beeing truncated if another process has
*    opened it and has a lock on it.
*************************************************************************/

FILE *fopenLock(const char *name, const char *mode, int type)

/************************************************************************/
{
  FILE *fp;
  struct flock lock;
  int oFlag;
  int fd;
  char mode2[5];

  fd = -1; /*  Preset for error case  */

  /*  Preset oFlag for open call  */
  if(strchr(mode, '+') == NULL) {
    /*  Only one kind of access mode  */
    switch(mode[0]) {
    case 'r': oFlag = O_RDONLY;                  break;
      /*  In case of 'w' do not truncate now  */
    case 'w': oFlag = O_WRONLY|O_CREAT;          break;
    case 'a': oFlag = O_WRONLY|O_APPEND|O_CREAT; break;
    default: goto error;
    }
  } else {
    /*  Read and write access  */
    switch(mode[0]) {
    case 'r': oFlag = O_RDWR;                  break;
      /*  In case of 'w' do not truncate now  */
    case 'w': oFlag = O_RDWR|O_CREAT;          break;
    case 'a': oFlag = O_RDWR|O_APPEND|O_CREAT; break;
    default: goto error;
    }
  }

  /*  A little bit type checking  */
  if(type < 0 || type >= 5) goto error;
  lock.l_type   = LockTypes[type];
  lock.l_start  = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len    = 0;

  /*  Open file descriptor and try to lock  */
  if((fd = open(name,
#ifdef _WIN32
		O_BINARY |
#endif
		oFlag, 0644)) < 0 ||
     fcntl(fd, (type & 1) ? F_SETLKW : F_SETLK, &lock))
    goto error;

  /*  If open and lock suceeded, truncate file in case of write!  */
  if(mode[0] == 'w') {
    if(ftruncate(fd, 0) == -1)
      goto error;
  }

  /*  Now copy mode to mode2, but skip any 'b' characters  */
  /*   which are disallowed in our unix version.           */
  {
    const char *string;
    char *string2;

    string = mode;
    string2 = mode2;

    do {
      if(*string != 'b') *string2++ = *string;
    } while(*string++);
  }

  /*  Transfer to standard library file pointer  */
  if((fp = fdopen(fd, mode2)) == NULL) goto error;

  return(fp);
error:
  if(fd >= 0) close(fd);
  return(NULL);
}

#else

/* ***  No POSIX functions available  *** */
int lockRegion(FILE *fp, int type, long offset, int whence, long len)
{  return(0); }
FILE *fopenLock(const char *name, const char *mode, int type)
{  return(fopen(name, mode)); }
#endif

#ifdef TEST
int main(int argc, char **argv)
{
  FILE *fp;
  int lucky;

  if(argc == 1) lucky = JSLOCK_READ;
  else lucky = JSLOCK_WRITE;

  if((fp = fopenLock("testfile", "r+", lucky)) == NULL) {
    printf("ERROR: Cannot open and lock file.\n");

    if((fp = fopen("testfile", "r+")) == NULL) {
      printf("ERROR: Cannot open file.\n");
      exit(20);
    }
  }

  printf("Successfully opened...\n\
 Now waiting 200 secs\n");

  sleep(200);

  printf("Good bye\n");

  fclose(fp);
  return(0);
}
#endif
