/*JS*********************************************************************
*
*    Program : FOPENCREATE
*    Language: ANSI-C with POSIX directory creating function.
*    Author  : Joerg Schoen
*    Purpose : Build up needed directory structure when opening file
*              for writing or change permissions on the fly and set
*              them back.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <jsconfig.h>

#include <jsllio.h>

#include <jssubs.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS */
#include <stdio.h>   /*	 For "FILE" definition    */
/*  END-DEFINITIONS */

#ifdef TESTX
# define DEBUG  /*  Use verbosity when compiling test program  */
#endif

/*  Create directories with all permissions (will be modified by umask)  */
#define DIRCREATEMODE   0777

/*  For routine doPermit  */
#define MODE_CHPERM       (1<<0)
#define MODE_CHPERMFILE   (1<<1)  /*  includes MODE_CHPERM  */
#define MODE_MKDIR        (1<<2)

/*  Wrap critical regions to prevent CNTRL-C leaving a changed structure
 *   behind
 */
#define USE_CRITICAL

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype FILE          *fopenPermit(const char *name, const char *mode);
Prototype FILE          *fopenCreate(const char *name, const char *mode);
static int               doFOpen(const char *name, void *p);

Prototype int            statPermit(const char *name, void *bufPtr);
Prototype int            lstatPermit(const char *name, void *bufPtr);
Prototype int            utimePermit(const char *name, void *bufPtr);

Prototype int            symlinkPermit(const char *oldpath,
				       const char *newpath);
static int               doSymLink(const char *name, void *p);

Prototype int            chmodPermit(const char *name, unsigned long mode);
static int               doChMod(const char *name, void *p);

static int		 doPermit(const char *fileName,
				  int (*func)(const char *name, void *p),
				  void *arg, int mode);

Prototype int            FOpCreateVerbose;

/*********     GLOBAL/STATIC VARIABLES                          *********/
/*  If >= 0 , print on stderr, when a directory is created  */
int FOpCreateVerbose = -1;

static FILE *FilePtr;

/*JS*********************************************************************
*   Similar to 'fopen', but creates the underlying directory structure if
*    the file is opened for writing -- if the file is opened solely for
*    reading, this would not make sense. Also, if any access permissions
*    of intermediate directories are bad, they are changed for the open
*    and restored at the end.
*************************************************************************/

FILE *fopenPermit(const char *name, const char *mode)

/************************************************************************/
{
  if(doPermit(name, doFOpen, (void *)mode, strchr(mode, 'r') ?
	      MODE_CHPERMFILE : (MODE_CHPERMFILE | MODE_MKDIR)) == -1)
    return NULL;

  return FilePtr;
}

/*JS*********************************************************************
*   Similar to 'fopen', but creates the underlying directory structure if
*    the file is opened for writing -- if the file is opened solely for
*    reading, this would not make sense. Does not, however, change any
*    permissions.
*************************************************************************/

FILE *fopenCreate(const char *name, const char *mode)

/************************************************************************/
{
  if(doPermit(name, doFOpen, (void *)mode, strchr(mode, 'r') ?
	      0 : MODE_MKDIR) == -1)
    return NULL;

  return FilePtr;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doFOpen(const char *name, void *p)

/************************************************************************/
{
  if((FilePtr = fopen(name, (const char *)p)) == NULL)
    return -1;

  return 0;
}

/*JS*********************************************************************
*   Similar to "stat", but changes permissions of intermediate directories
*    on the fly and sets them back. Does not create directories, since
*    that wouldn't make sense. Since the prototype of this function might
*    be visible in code where "struct stat" is not known, we use a
*    "void *" pointer instead.
*************************************************************************/

int statPermit(const char *name, void *bufPtr)

/************************************************************************/
{
  /*  Do not change permissions of file itself, since that
   *   would give wrong results of 'stat' call,
   */
  return doPermit(name, (int (*)(const char *, void *))stat, bufPtr,
		  MODE_CHPERM);
}

/*JS*********************************************************************
*   Similar to "lstat", but changes permissions of intermediate directories
*    on the fly and sets them back. Does not create directories, since that
*    wouldn't make sense. Since the prototype of this function might be
*    visible in code where "struct stat" is not known, we use a "void *"
*    pointer instead.
*************************************************************************/

int lstatPermit(const char *name, void *bufPtr)

/************************************************************************/
{
  /*  Do not change permissions of file itself, since that
   *   would give wrong results of 'lstat' call,
   */
  return doPermit(name, (int (*)(const char *, void *))lstat, bufPtr,
		  MODE_CHPERM);
}

/*JS*********************************************************************
*   Similar to "utime", but changes permissions of intermediate directories
*    on the fly and sets them back. Does not create directories, since that
*    wouldn't make sense. Since the prototype of this function might be
*    visible in code where "struct timbuf" is not known, we use a "void *"
*    pointer instead.
*************************************************************************/

int utimePermit(const char *name, void *bufPtr)

/************************************************************************/
{
  return doPermit(name, (int (*)(const char *, void *))utime, bufPtr,
		  MODE_CHPERMFILE);
}

#ifndef _WIN32
/*JS*********************************************************************
*
*************************************************************************/

int symlinkPermit(const char *oldpath, const char *newpath)

/************************************************************************/
{
  return doPermit(newpath, doSymLink, (void *)oldpath,
		  MODE_CHPERMFILE | MODE_MKDIR);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doSymLink(const char *name, void *p)

/************************************************************************/
{
  return symlink((const char *)p, name);
}
#endif

/*JS*********************************************************************
*   Similar to "chmod", but changes permissions of intermediate directories
*    on the fly and sets them back. Does not create directories, since
*    that wouldn't make sense. Since the prototype of this function might
*    be visible in code where "mode_t" is not known, we use "unsigned long"
*    instead.
*************************************************************************/

int chmodPermit(const char *name, unsigned long mode)

/************************************************************************/
{
  /*  If changing permissions on file itself, 'chmod' call
   *   would be senseless.
   */
  return doPermit(name, doChMod, &mode, MODE_CHPERM);
}

/*JS*********************************************************************
*
*************************************************************************/

static int doChMod(const char *name, void *p)

/************************************************************************/
{
  return chmod(name, *(unsigned long *)p);
}

/*JS*********************************************************************
*   Do the required stuff of changing permissions of intermediate
*    directories (if (mode & (MODE_CHPERM|MODE_CHPERMFILE))), changing
*    permissions on file itself (if (mode & MODE_CHPERMFILE)), and creating
*    directories (if (mode & MODE_MKDIR)) and call user supplied function
*    on file.
*************************************************************************/

static int doPermit(const char *fileName, int (*func)(const char *n, void *p),
		    void *arg, int mode)

/************************************************************************/
{
  char *string, *name;
  int ret, count, nOk, saveErrno;
  struct stat stat_buf;
  struct table {
    int    Pos;
    mode_t Mode;
  } *table;

  /*  Save error code to restore from unsuccessfull stat's etc  */
  saveErrno = errno;

  /*  First try directly to open file (saves a lot of stuff)  */
  if((ret = (*func)(fileName, arg)) >= 0)
    return ret;

  /*  Prepare: make a copy of the file name, add "./" if necessary.
   *   Count directory delimiters and allocate table.
   */
  table = NULL;
  if((name = (char *)malloc(strlen(fileName) + 3)) == NULL)
    EGOTO(error);

  /*  Check for absolute path name or already a "./" at the start  */
  if(fileName[0] == CONFIG_DIR_DELIMITER ||
     (fileName[0] == '.' && fileName[1] == CONFIG_DIR_DELIMITER)) {
    strcpy(name, fileName);
  } else {
    name[0] = '.';
    name[1] = CONFIG_DIR_DELIMITER;
    strcpy(name + 2, fileName);
  }

  for(string = name, count = 1 ;
      (string = strchr(string, CONFIG_DIR_DELIMITER)) ; string++) {
    count++;
    while(string[1] == CONFIG_DIR_DELIMITER) /* skip multiple '/' */
      string++;
  }

  if((table = (struct table *)malloc(count * sizeof(*table))) == NULL)
    EGOTO(error);

  /* ***  First loop: Find longest part of file where "stat" succeeds  *** */
  string = &name[strlen(name)];
  for(nOk = 0 ; ; nOk++) {
    table[nOk].Mode = (mode_t)-1; /*  be clean  */

    if(stat(name, &stat_buf) != -1)
      break;

    /*  Remove next path component  */
    do
      string--;
    while(*string != CONFIG_DIR_DELIMITER && string > name);

    if(string == name)
      /*  no more parts, we are *really* in lack of
       *   the appropriate permissions
       */
      EGOTO(error);

    while(string[-1] == CONFIG_DIR_DELIMITER) /* skip multiple '/' */
      string--;

    table[nOk].Pos = string - name;
    *string = '\0';
  }

#ifdef DEBUG
  printf("STAT #%d \"%s\" IS OK.\n", nOk, name);
  errno = 0;
#endif

#ifdef USE_CRITICAL
  /*  Prevent CNTRL-C in region where things are changed.  */
  criticalStart();
#endif

  {
    mode_t perm;
    uid_t myUid = getuid();
    gid_t myGid = getgid();

    /* ***  Second loop: Change permissions starting from current part
     * ***   or create directory structure until we reach the file
     */
    /*  Preset perm  */
    perm = stat_buf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO);
    for(count = nOk ; ; ) {
      mode_t needed;

      /*  What permissions do we need to succeed? We give ourselves
       *   "rwx" for directories, "rw" for files.
       */
      if(stat_buf.st_uid == myUid)
	needed = (count == 0) ? (S_IRUSR|S_IWUSR) : S_IRWXU;
      else if(stat_buf.st_gid == myGid)
	needed = (count == 0) ? (S_IRGRP|S_IWGRP) : S_IRWXG;
      else
	needed = (count == 0) ? (S_IROTH|S_IWOTH) : S_IRWXO;

      /*  Only change permissions if we are allowed to. We only need
       *   to change permissions of the file itself if we are going
       *   to open the file.
       */
      if((perm & needed) != needed &&
	 (count || (mode & MODE_CHPERMFILE)) &&
	 (mode & (MODE_CHPERM|MODE_CHPERMFILE))) {
#ifdef DEBUG
	printf("  %d CHMOD \"%s\" to %04o | %04o\n", count, name, perm, needed);
#endif
	table[count].Mode = perm;
	if(chmod(name, perm | needed)) {
	  saveErrno = errno; /*  Now keep error code  */
	  break;
	}
      }

      if(--count < 0)
	break;

      /*  Restore previous file name  */
      name[table[count].Pos] = CONFIG_DIR_DELIMITER;

#ifdef DEBUG
      printf("  %d STAT \"%s\"\n", count, name);
#endif
      if(stat(name, &stat_buf) >= 0) {
	/*  Get permissions for next loop  */
	perm = stat_buf.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO);
      } else if(count == 0) {
	/*  File does not exist, so fopen will create it  */
	count--;
	break;
      } else {
	/*  Maybe next part of directory didn't exist  */

	/*  Only create directory if we are allowed to.
	 *   In case we do, we make no assumption about the permissions
	 *   of the newly created directory, since we might have an ugly
	 *   umask value so that the permissions are not ok. Therefore
	 *   call "stat" on new directory.
	 */
	if(!(mode & MODE_MKDIR) ||
	   (FOpCreateVerbose >= 0 &&
	    fprintf(stderr, "%*sCreating \"%s\"\n", FOpCreateVerbose, "", name)
	    < 0) ||
	   mkdir(name, DIRCREATEMODE) == -1 || stat(name, &stat_buf) == -1) {
	  saveErrno = errno; /*  Now keep error code  */
	  break;
	}

	perm = stat_buf.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO);
      }
    }
  }

  if(count < 0) {
    /*  Now call user function  */
    ret = (*func)(name, arg);

    /*  Keep error code in case of failure  */
    if(ret == -1)
      saveErrno = errno;

#ifdef DEBUG
    printf("USERFUNCTION \"%s\" %s.\n", name, ret >= 0 ? "SUCCEEDED" : "FAILED");
#endif
  } else {
#ifdef DEBUG
    printf("PROBLEM at %d \"%s\"\n", count, name);
#endif
    /*  Previous loop failed, so restore file name for final loop  */
    while(--count >= 0)
      name[table[count].Pos] = CONFIG_DIR_DELIMITER;
  }

  /* ***  Final loop: Restore any changed permissions  *** */
  if(mode & (MODE_CHPERM|MODE_CHPERMFILE)) {
    for(count = 0 ; ; count++) {
      if(table[count].Mode != (mode_t)-1) {
#ifdef DEBUG
	printf("  %d BACK CHMOD \"%s\" to %04o\n", count, name, table[count].Mode);
#endif
	/*  Try hard to finish our job  */
	(void)chmod(name, table[count].Mode);
      }
      if(count >= nOk)
	break;

      name[table[count].Pos] = '\0';
    }
  }

#ifdef USE_CRITICAL
  /*  We are through, now allow CNTRL-C  */
  criticalEnd();
#endif

  /*  Clean up  */
  free(name);
  free(table);
  errno = saveErrno;

  return ret;
error:
  if(name) free(name);
  if(table) free(table);
  return -1;
}

#if defined(TEST) || defined(TESTX)
/* ***  Small test program  *** */
int main(int argc, char *argv[])
{
  FILE *fp1, *fp2;
  int c;

  if(argc != 3) {
    printf("Usage: createcopy <file1> <file2>\n\
  Is able to open the files even if the underlying directory structure does not\n\
  exist or permissions for opening the files are bad. In the first case,\n\
  directories will be created, in the second, permissions are changed and\n\
  restored at the end.\n");
    return 5;
  }

  FOpCreateVerbose = 1;

  if((fp1 = fopenCreate(argv[1], "rb")) == NULL ||
     (fp2 = fopenCreate(argv[2], "wb")) == NULL)
    goto error;

  while((c = getc(fp1)) != EOF) putc(c, fp2);

  if(ferror(fp1) || ferror(fp2) || fclose(fp1) || fclose(fp2)) goto error;

  return 0;
error:
  perror("createcopy");
  return 30;
}
#endif
