/*JS*********************************************************************
*
*    Program : FOPENPATH
*    Language: ANSI-C with POSIX directory scanning functions
*    Author  : Joerg Schoen
*    Purpose : Seek in path and directories before fopen.
*
*              Copyright (C) 1994-2004 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <jsconfig.h>

#include <jssubs.h>

#include <jsllio.h>

#include <errno.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS */
#include <stdio.h>   /*	 For "FILE" definition    */
#define fopenPath(name, mode)          fopenPath2(name, mode, FOpenPath)
#define fopenPath2(name, mode, path)   fopenPath3(name, mode, path, 0)
/*  END-DEFINITIONS */

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype FILE		*fopenPath3(const char *name, const char *mode,
				    const char *path, int flag);
Prototype FILE		*fopenScan(const char *name, const char *mode);
static int               scandirectories(char *current, int start,
					 const char *pattern);
static int               docompare(const char *name, const char *pattern);

Prototype const char	*FOpenName, *FOpenPath;

/*********     GLOBAL VARIABLES 				*********/
/*  Input for fopenPath: the path  */
const char *FOpenPath = NULL;

/*  Output of fopenPath and fopenScan: the 'real' file name  */
const char *FOpenName = NULL;

#define LEN_DELIMITSTRING      2

static const char PDelimitString[LEN_DELIMITSTRING + 1] = {
  CONFIG_DIR_DELIMITER,  CONFIG_DIR_DELIMITER, '\0'
};

/*JS*********************************************************************
*   Tries to find the file in the path. This has to be a ':' separated
*    list of directories. If path is NULL, this call is equivalent to a
*    normal fopen call.
*    The current directory is not searched for the file automatically,
*    so it must be put explicitly in the path. An empty entry in the path
*    is equivalent to the current directory. If the file was found anywhere
*    in the path, the global variable 'FOpenName' contains a pointer to a
*    static field containing the file name. This field may be equal to 'name'.
*    "//" sequences in the path lets the routine search recursively
*    through the directory tree: between two '/' an arbitrary path is
*    tried. Symbolic links are followed.
*
*    Example:
*      path = ":/home//input//"     and   name = "test"
*
*    Possible results may be:
*      "test" or
*      "/home/input/test" or "/home/SPECIAL/input/GENERAL/test" or
*      "/home/AA/BB/input/XX/YY/test".
*    If flag is 0, we behave a bit relaxed: A '/' somewhere in 'name' is OK
*    (we are more like the C preprocessor who looks up <sys/types.h> in the
*    include path). Only an ABSOLUTE pathname forces us to resort to
*    simple fopen.
*    If bit 0 of flag is non-zero, we become strict: Any '/' in the path
*    lets us resort to fopen. If bit 1 is set, we even try the path when
*    we open a file for writing.
*************************************************************************/

FILE *fopenPath3(const char *name, const char *mode, const char *path,
		 int flag)

/************************************************************************/
{
  const char *string;
static char nameBuffer[JS_FILENAME_MAX + 1];
  FILE *fp;
  int saveErrno;

  /*  We do a normal fopen call IF:
   *  - path is not set
   *  - file name is absolute or with bit 0 of flag set contains
   *    a '/' somewhere.
   *  - we open for writing and bit 1 of flag is cleared
   */
  if(path == NULL ||
     ((flag & 1) ? strchr(name, CONFIG_DIR_DELIMITER) != NULL :
      name[0] == CONFIG_DIR_DELIMITER) ||
     (!(flag & 2) && strchr(mode, 'r') == NULL)) {
    if((fp = fopen(name, mode)) != NULL)
      FOpenName = name;

    return fp;
  }

  /*  Scan through path to find file  */
  saveErrno = errno;
  for(string = path ; *string ; ) {
    char *string2;

    /*	Copy next path name  */
    for(string2 = nameBuffer ; *string ; ) {
      if(*string == CONFIG_LIST_DELIMITER) {
	/*  Skip delimiter and break  */
	string++;
	break;
      }
      *string2++ = *string++;
    }

    /*	Insert '/' if necessary  */
    if(string2 > nameBuffer && string2[-1] != CONFIG_DIR_DELIMITER)
      *string2++ = CONFIG_DIR_DELIMITER;

    /*	And append name  */
    strcpy(string2, name);

    /*	Try to open file  */
    if((fp = fopenScan(nameBuffer, mode))) {
      /*  Reset error code from unsuccessful opens  */
      errno = saveErrno;

      return fp;
    }
  }

  /*  File not found in path, so return NULL  */
  return NULL;
}

/* ***	Static buffer for seekroutines	*** */
static char NameBuffer[JS_FILENAME_MAX + 1];

/*JS*********************************************************************
*   Sets up for call of subOpen, which searchs through directory structure.
*************************************************************************/

FILE *fopenScan(const char *name, const char *mode)

/************************************************************************/
{
  FILE *fp;
  char *current;

  /*  Preset found name  */
  FOpenName = NULL;

  /*  Look if path contains special delimiter  */
  if((current = strstr(name, PDelimitString)) == NULL) {
    /*	Just call fopen  */
    if((fp = fopen(name, mode)))
      FOpenName = name;
  } else {
    int offset, ret;

    /*	Set up for call of sub routine  */
    offset = (current - name);

    /*	Generate first part of filename  */
    strncpy(NameBuffer, name, offset);
    NameBuffer[offset] = '\0';

    /*	 Now try to get file anywhere in directory structure  */
    if((ret = scandirectories(&NameBuffer[offset], offset, name)) == -1)
      return NULL;

    if(ret == 0)
      fp = NULL;
    else if((fp = fopen(NameBuffer, mode)) != NULL)
      FOpenName = NameBuffer;
  }

  return fp;
}

/*JS*********************************************************************
*   Assumes, that 'NameBuffer' already contains the first fixed part of
*    the file name which ends at offset 'start'. 'current' points within
*    'NameBuffer' to the start of the next part which starts with "//".
*    The routine will recursively scan through the directory in
*    Namebuffer and check the pattern.
*    On the first matching file found, '1' is returned, otherwise '0'.
*    On error, '-1' is returned.
*
*    The routine is less than perfect, because the number of file pointers
*    open (from 'opendir' calls), increases with the depth of the directory
*    structure. An alternative would be to use a combination of 'telldir'
*    and 'seekdir' to memorize the current scanning position within a
*    directory, but neither function is POSIX. Another alternative is to
*    first memorize all names before continuing.
*************************************************************************/

static int scandirectories(char *current, int start, const char *pattern)

/************************************************************************/
{
  DIR *dp;
  struct dirent *dirp;
  int remLength, ret;

  if((dp = opendir(NameBuffer)) == NULL)
    return -1;

  *current++ = CONFIG_DIR_DELIMITER;

  /*	Get remaining space for filename length overflow check	*/
  remLength = &NameBuffer[sizeof(NameBuffer) - 1] - current;

  ret = 0;

  while((dirp = readdir(dp)) != NULL) {
    struct stat stBuf;
    int len;

    /*  Reject "." and ".." entry  */
    if(dirp->d_name[0] == '.' && (dirp->d_name[1] == '.' ||
				  dirp->d_name[1] == '\0'))
      continue;

    len = strlen(dirp->d_name);

    /*  Check file name length overflow  */
    if(len > remLength)
      EGOTO(error);

    strcpy(current, dirp->d_name);

    /*  FIXME: If we use 'stat', we might follow symbolic links and even
     *   stop on broken ones
     */
    if(lstat(NameBuffer, &stBuf) == -1)
      EGOTO(error);

    if(S_ISDIR(stBuf.st_mode)) {
      if((ret = scandirectories(current + len, start, pattern)) == -1)
	EGOTO(error);

      if(ret != 0)
	break;
    } else {
      /*  A file, try a match  */
      if(docompare(NameBuffer + start, pattern + start))
	return 1;
    }
  }

  (void)closedir(dp);

  return ret;
error:
  (void)closedir(dp);
  return -1;
}

/*JS*********************************************************************
*   Compare the name with the pattern. In string 'pattern', a "//"
*    sequence basically works as a "*" shell wildcard which
*    also matches "/".
*************************************************************************/

static int docompare(const char *name, const char *pattern)

/************************************************************************/
{
  const char *next;
  int len;

  /*  Determine length of string that has to match exactly  */
  if((next = strstr(pattern, PDelimitString)) != NULL)
    len = next - pattern;
  else
    len = strlen(pattern);

  if(len > 0 && memcmp(name, pattern, len) != 0)
    return FALSE;

  name += len;
  pattern += len;

  if(*pattern == '\0')
    /*  Pattern exhausted, so we genuinely have a match  */
    return TRUE;

  /*  Pattern starts with "//". Skip it  */
  pattern += LEN_DELIMITSTRING;

  while(*name == CONFIG_DIR_DELIMITER) {
    ++name;

    /*  Here is where recursion comes in (which is natural for pattern
     *   matching)
     */
    if(docompare(name, pattern))
      return TRUE;

    if((name = strchr(name, CONFIG_DIR_DELIMITER)) == NULL)
      break;
  }

  return FALSE;
}

#ifdef TEST
int main(int argc, char *argv[])
{
  if(argc != 3) {
    printf("Usage: testfopenpath <path> <file>\n");
    return(20);
  }

  /*  Set path	*/
  FOpenPath = argv[1];

  if(fopenPath(argv[2], "r"))
    printf("%s: %s\n", argv[2], FOpenName);
  else
    printf("ERROR: File \"%s\" not found!\n", argv[2]);

  return(0);
}
#endif
