/*JS*********************************************************************
*
*    Program : FSTRALLOC
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Fast string allocator
*
*              Copyright (C) 1994-2011 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

#include <assert.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#ifndef fstrmarkinvalidate
/*  START-DEFINITIONS  */
typedef struct fstrallocchunk_t fstrallocchunk_t;

struct fstrallocchunk_t {
  fstrallocchunk_t *FSAC_Next;

  unsigned int FSAC_Index, FSAC_Max;
  char FSAC_Mem[1];
};

typedef struct {
  fstrallocchunk_t *FSA_Chain;
  unsigned int FSA_AllocSize;

} fstralloc_t;

#define fstrinit(fstr)				\
  ((fstr)->FSA_Chain = NULL,			\
   (fstr)->FSA_AllocSize = FStringAllocSize)

/* ******************************************************* */
/*  Memorize current watermark of a pool to be restored later  */
typedef struct {
  const fstrallocchunk_t *FSAM_Chunk;
  unsigned int FSAM_Index;

} fstrallocmark_t;

#define fstrmarkinvalidate(pmark)		\
  ((pmark)->FSAM_Chunk = NULL,			\
   (pmark)->FSAM_Index = -1)

#define fstrmarkisvalid(pmark)			\
  ((pmark)->FSAM_Index != (unsigned int)-1)

/*  END-DEFINITIONS  */
#endif

#ifndef FSTRALLOC_SIZE
# define FSTRALLOC_SIZE         (1 << 12)
#endif

#define FSTRALLOC_FILLBYTE      0xff

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype char          *fstralloc(fstralloc_t *fstr, unsigned int len);
Prototype char          *fstrdup(fstralloc_t *fstr, const char *string);

Prototype void           fstrclear(fstralloc_t *fstr);
Prototype void           fstrfree(fstralloc_t *fstr);

Prototype void           fstrmarksave(const fstralloc_t *fstr,
				      fstrallocmark_t *pmark);
Prototype void           fstrmarkrestore(fstralloc_t *fstr,
					 const fstrallocmark_t *pmark);

Prototype unsigned int   FStringAllocSize;

unsigned int FStringAllocSize = FSTRALLOC_SIZE;

/*JS*********************************************************************
*
*************************************************************************/

char *fstralloc(fstralloc_t *fstr, unsigned int len)

/************************************************************************/
{
  fstrallocchunk_t *pool;
  unsigned int maxsize;

  for(pool = fstr->FSA_Chain ; pool ; pool = pool->FSAC_Next)
    if(pool->FSAC_Index + len <= pool->FSAC_Max) {
      char *str;

      str = &(pool->FSAC_Mem[pool->FSAC_Index]);
      pool->FSAC_Index += len;
      return str;
    }

  /*  Allocate a new chunk for the pool  */
  maxsize = MAX(len, fstr->FSA_AllocSize);
  if((pool = (fstrallocchunk_t *)malloc(sizeof(*pool) + maxsize - 1)) == NULL)
    EGOTO(error);

  pool->FSAC_Index = len;
  pool->FSAC_Max = maxsize;

  /*  Hang new pools in front of list, so they are faster found.
   *   This is also used below in fstrmarkrestore
   */
  pool->FSAC_Next = fstr->FSA_Chain;
  fstr->FSA_Chain = pool;

  return pool->FSAC_Mem;
error:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

char *fstrdup(fstralloc_t *fstr, const char *string)

/************************************************************************/
{
  char *dest;

  if((dest = fstralloc(fstr, strlen(string) + 1)) == NULL)
    return NULL;

  strcpy(dest, string);
  return dest;
}

/*JS*********************************************************************
*
*************************************************************************/

void fstrclear(fstralloc_t *fstr)

/************************************************************************/
{
  fstrallocchunk_t *pool;

  for(pool = fstr->FSA_Chain ; pool ; pool = pool->FSAC_Next) {
#if !defined NDEBUG && defined FSTRALLOC_FILLBYTE
    if(pool->FSAC_Index > 0)
      memset(pool->FSAC_Mem, FSTRALLOC_FILLBYTE, pool->FSAC_Index);
#endif

    pool->FSAC_Index = 0;
  }
}

/*JS*********************************************************************
*
*************************************************************************/

void fstrfree(fstralloc_t *fstr)

/************************************************************************/
{
  fstrallocchunk_t *pool, *next;

  for(pool = fstr->FSA_Chain ; pool ; pool = next) {
    next = pool->FSAC_Next;
    free(pool);
  }

  fstr->FSA_Chain = NULL; /*  make it reusable  */
}

/*JS*********************************************************************
*
*************************************************************************/

void fstrmarksave(const fstralloc_t *fstr, fstrallocmark_t *pmark)

/************************************************************************/
{
  fstrallocchunk_t *pool;

  for(pool = fstr->FSA_Chain ; pool ; pool = pool->FSAC_Next)
    /*  Memorize the first one not being empty. Usually the first
     *   one is non-empty. But in case a pool was cleared, we might
     *   find the first non-empty one later in the chain.
     */
    if(pool->FSAC_Index > 0) {
      pmark->FSAM_Chunk = pool;
      pmark->FSAM_Index = pool->FSAC_Index;
      return;
    }

  /*  Complete pool is empty. Create a valid mark that effectively
   *   does a 'fstrclear'.
   */
  pmark->FSAM_Chunk = NULL;
  pmark->FSAM_Index = 0;
}

/*JS*********************************************************************
*
*************************************************************************/

void fstrmarkrestore(fstralloc_t *fstr, const fstrallocmark_t *pmark)

/************************************************************************/
{
  fstrallocchunk_t *pool;

  /*  Should not be called with an invalid marker - it would then
   *   clear the complete pool!
   */
  assert(fstrmarkisvalid(pmark));

  for(pool = fstr->FSA_Chain ; pool ; pool = pool->FSAC_Next) {
    if(pool == pmark->FSAM_Chunk) {
#if !defined NDEBUG && defined FSTRALLOC_FILLBYTE
      if(pool->FSAC_Index > pmark->FSAM_Index)
        memset(pool->FSAC_Mem + pmark->FSAM_Index, FSTRALLOC_FILLBYTE,
	       pool->FSAC_Index - pmark->FSAM_Index);
#endif

      pool->FSAC_Index = pmark->FSAM_Index;
      break;
    }

    /*  All other pools in front of the one from the mark were added
     *   later and thus must be empty.
     */
#if !defined NDEBUG && defined FSTRALLOC_FILLBYTE
    if(pool->FSAC_Index > 0)
      memset(pool->FSAC_Mem, FSTRALLOC_FILLBYTE, pool->FSAC_Index);
#endif

    pool->FSAC_Index = 0;
  }
}
