/*JS*********************************************************************
*
*    Program : GETCTOKEN
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Get token of C language.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jssubs.h>

#include <jsconfig.h>

#include <jsllio.h>

/*********     DEFINES                                          *********/
#ifdef CONFIG_NO_STRDUP
# define strdup  mystrdup  /*  strdup is not POSIX conformant  */
#endif

/*#define NEWNEW*/

#ifndef GCTK_NOLINEESC
/*  START-DEFINITIONS  */
#include <stdio.h>   /*  For "FILE" definition    */

#define GCTK_NOLINEESC     (1 <<  0) /*  Drop "\\\n" in identifiers and numbers */
#define GCTK_NOLINEESCTOT  (1 <<  1) /*  Drop "\\\n"  completely */
#define GCTK_UNIQUE        (1 <<  2) /*  Generate unique token string memory  */
#define GCTK_PPINCLUDELINE (1 <<  3) /*  Recognize <..> like strings  */
#define GCTK_CPPCOMMENT    (1 <<  4) /*  Recognize C++ comments  */
#define GCTK_READTOTERM    (1 <<  5) /*  Read till GCTKTerminator is found  */
/*  Escape character does not work in single quoted strings  */
#define GCTK_SHELLSTRING   (1 <<  6)

#define GCTK_SKIPSPACETOK  (1 <<  7) /*  Don't return blanks, just skip them  */
#define GCTK_SKIPSPACETOK2 (1 <<  8) /*  Don't return blanks, except newline */
#define GCTK_INCSTARTSTOP  (1 <<  9) /*  Return complete tokens  */
#define GCTK_CBRASTRING    (1 << 10) /*  Recognize string with curly brackets  */
#define GCTK_SQBRASTRING   (1 << 11) /*  Recognize string with square brackets  */

/*  Switch on extensions: Allow '$' in identifiers and newline in string */
#define GCTK_EXTENSIONS    (1 << 12)
/*  Extension2: Return " in case of \" sequence  */
#define GCTK_EXTENSIONS2   (1 << 13)

#define GCTK_SQLSTRING     (1 << 14) /*  '' and "" are escapes for quotes  */
#define GCTK_SQLCOMMENT    (1 << 15) /*  -- introduces comment  */

#define GCTK_FLAG_USER     (1 << 16) /*  Start of user definable flags  */

/*  Basic definition for C token  */
typedef struct {
#define CTOKEN_FALLOCATED         (1 << 0)
#define CTOKEN_FORIGTOKEN         (1 << 8)
    int TK_Flags;
    int TK_Value;

    union {
        const char  *tk_String;
        void        *tk_Ptr; /*  to store an arbitrary pointer  */
    } tk_cont;
#define TK_String  tk_cont.tk_String
#define TK_Ptr     tk_cont.tk_Ptr

} jsctoken_t;

#define CTOKEN_GENSTATIC(value, string)    { 0, value, { string } }

#define copyCToken(token, mode)     copyCToken2(NULL, token, mode)

/*  Definition for operator table  */
typedef struct {
  char *KD_String;
  int   KD_Token;
} jsctokops_t;

/*  Basic return tokens of routine getCToken  */
#define CTK_EOF            (-1) /*  My own EOF value  */
#define CTK_MINIMUM        257  /*  Lowest special token value  */

#define CTK_ERROR          (CTK_MINIMUM + 0) /* Error in getCToken  */

#define CTK_BLANK_START   (CTK_MINIMUM + 1)  /* Blank tokens  */
#define CTK_BLANKSEOL      (CTK_MINIMUM + 1)
#define CTK_BLANK_START2  (CTK_MINIMUM + 2)  /* Blank tokens without newline */
#define CTK_BLANKS         (CTK_MINIMUM + 2)
#define CTK_COMMENT        (CTK_MINIMUM + 3)
#define CTK_CPPCOMMENT     (CTK_MINIMUM + 4)
#define CTK_SQLCOMMENT     (CTK_MINIMUM + 5)
#define CTK_BLANK_END     (CTK_MINIMUM + 5)

#define CTK_STRING         (CTK_MINIMUM + 10)  /*  String tokens  */
#define CTK_WSTRING        (CTK_MINIMUM + 11)
#define CTK_USTRING        (CTK_MINIMUM + 12)
#define CTK_PPSTRING       (CTK_MINIMUM + 13)
#define CTK_CHARCONST      (CTK_MINIMUM + 14)
#define CTK_WCHARCONST     (CTK_MINIMUM + 15)
#define CTK_UCHARCONST     (CTK_MINIMUM + 16)

#define CTK_IDENT          (CTK_MINIMUM + 20) /*  Identifier  */

#define CTK_NUMBER         (CTK_MINIMUM + 25) /*  Number tokens:  */
#define CTK_INUMBER        (CTK_MINIMUM + 26)
#define CTK_DNUMBER        (CTK_MINIMUM + 27)

#define CTK_VANILLA        (CTK_MINIMUM + 30) /*  Solely string  */
#define CTK_CBRASTRING     (CTK_MINIMUM + 31) /*  String in curly brackets  */
#define CTK_SQBRASTRING    (CTK_MINIMUM + 32) /*  String in square brackets  */

#define CTK_USERTOKEN      (CTK_MINIMUM + 43) /*  First user defined token == 300  */

/*  Estimate needed space to print a token  */
#define CTK_ESTLENGTH(token)   ((token).TK_String ? \
                                (strlen((token).TK_String) + 5) : 1)
/*  END-DEFINITIONS  */
#endif

/* ERROR-DEFINITIONS from getCToken label _ERR_GETCTOKEN
   Unfinished token
   Unfinished string
   Unfinished character constant
   Unfinished preprocessor string
   Wrong number found
*/

/* ****  Private definitions for getToken routine  *** */
/*  Stepsize for increasing the length of the token buffer  */
#define BUFFINCREMENT        80

/* ***  Internal flags for routine getToken  *** */
#define FLAG_BLANKS_EOL      (1<<0)
#define FLAG_NUMISINT        (1<<1)
#define FLAG_NUMISFLOAT      (1<<2)
#define FLAG_WIDESTRING      (1<<3)
#define FLAG_UWIDESTRING     (1<<4)
#define FLAG_ISSPECIAL       (1<<5)
#define FLAG_EOFSEEN         (1<<6)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
/*Prototype int            initCToken(extfile_t *eFp);*/

Prototype jsctoken_t     getCToken(extfile_t *eFp, const jsctokops_t *opDef,
                                   int mode);
Prototype int            printCToken(FILE *fp, jsctoken_t token);
Prototype int            sprintCToken(char *s, jsctoken_t token);

Prototype jsctoken_t     copyCToken2(fstralloc_t *fstr, jsctoken_t token, int mode);
Prototype void           freeCToken(jsctoken_t token);

Prototype const char    *tkStrcompare(const char *input, const char *string);
Prototype int            tkStrcmp(const char *input, const char *string);

Prototype int            countnlCToken(const char *string);
Prototype int            updcolCToken(jsctoken_t token, int column);

Prototype void           tkDelNewLines(char *string);

Prototype const char    *dbgNameCToken(jsctoken_t token);

Prototype int            GCTKTerminator;

/*********     GLOBAL VARIABLES                                 *********/
int GCTKTerminator = '\n';

#ifdef NEWNEW
static char *TokBuffer = NULL; /*  Will be reused if not requested UNIQUE  */
static int   TokBuffLen = 0;

/*JS*********************************************************************
*   Initialize the getCToken routine.
*************************************************************************/

static int initCToken(extfile_t *eFp)

/************************************************************************/
{
    FILE *fp;
    long size;
    struct stat stbuf;

    if((fp = efileptr(eFp)) == NULL || fstat(fileno(fp), &stbuf)) EGOTO(error);

    size = stbuf.st_size;

    size += 2;
    /*  Increase token buffer  */
    if(size >= TokBuffLen) {
        TokBuffLen = size;
        if((TokBuffer = (char *)realloc(TokBuffer, TokBuffLen)) == NULL)
            EGOTO(error);
    }

    return 0;
error:
    return -1;
}
#endif

/*JS*********************************************************************
*   Get a C token from file fp. Recognizes operators from table opDef.
*************************************************************************/

jsctoken_t getCToken(extfile_t *eFp, const jsctokops_t *opDef, int mode)

/************************************************************************/
{
    jsctoken_t token;
    int flag, c, indx, cbranest;
    const char *cbradef;
#ifndef NEWNEW
    static char *TokBuffer = NULL; /*  Will be reused if not requested UNIQUE  */
    static int   TokBuffLen = 0;
#endif
    enum {
        ST_START,
        ST_ESCAPED,
        ST_BLANK,
        ST_COMMENT, ST_COMMENT2, ST_COMMENT3, ST_COMMENTPP,
        ST_COMMENTSQL0, ST_COMMENTSQL,
        ST_WSTRING, ST_USTRING,
        ST_STRING, ST_STRING2, ST_STRING3,
        ST_CHCONST, ST_CHCONST2, ST_CHCONST3,
        ST_PPSTRING, ST_PPSTRING2,
        ST_IDENT,
        ST_NUM, ST_NUMFRAC, ST_NUMFRAC2, ST_NUMEXP, ST_NUMEXP2,
        ST_NUMHEX, ST_NUMHEX2, ST_NUMOCT, ST_NUMTRAIL,
        ST_READTOTERM,
        ST_SPECIAL,
        ST_BRASTRING, ST_BRADQUOTE, ST_BRADQUOTEESC,
        ST_BRASQUOTE, ST_BRASQUOTEESC,
        ST_BRACOMMENT, ST_BRACOMMENT2, ST_BRACOMMENT3, ST_BRACOMMENTPP
    } state;

#ifdef NEWNEW
    /*  Just a hack... */
    if(!(eFp->EF_Flags & 2)) {
        if(initCToken(eFp)) EGOTO(error);
        eFp->EF_Flags |= 2;
    }
#endif

    indx  = 0;
    flag  = 0;
    state = ST_START;
    cbranest = 0; cbradef = NULL;
    token.TK_Value = CTK_ERROR;
    token.TK_Flags = 0;
    while(token.TK_Value == CTK_ERROR) {
        if(flag & FLAG_EOFSEEN) {
            JSSETERROR(_ERR_GETCTOKEN, 0);
            EGOTO(error);
        }

        /*  Get next char to look at  */
        if((c = efgetc(eFp)) == EOF)
            flag |= FLAG_EOFSEEN;

        /* ***  Save character in buffer  *** */
#ifndef NEWNEW
        if((indx + 2) >= TokBuffLen) {
            if((TokBuffer = (char *)
                realloc(TokBuffer, TokBuffLen + BUFFINCREMENT)) == NULL) /*DBG*/
                EGOTO(error);

            TokBuffLen += BUFFINCREMENT;
        }
#endif
        TokBuffer[indx++] = c;

        if(c == '\\' && !(flag & FLAG_ISSPECIAL)) {
            if((c = efgetc(eFp)) == '\n') {
                /*  Handle "\\\n" here: Save it in buffer or drop it   */
                if((mode & GCTK_NOLINEESCTOT)) {
                    indx--; /*  Drop "\\\n"  */

                } else {
                    TokBuffer[indx++] = c; /*  save it  */
                }

                continue;
            }

            eungetc(c, eFp);
            c = '\\';
        }

    redocurrent:
#if 0
        printf("    %d(%d,%d): '%c' (%d)\n", state, flag, indx, c, c);
#endif

        switch(state) {
        case ST_START: /* ***  Searching for new token  *** */
            if((mode & GCTK_READTOTERM) && c != GCTKTerminator) {
                state = ST_READTOTERM;
                goto redocurrent;

            } else if(isspace(c)) {
                if(((mode & GCTK_SKIPSPACETOK2) && c != '\n') ||
                   (mode & GCTK_SKIPSPACETOK)) {
                    indx--;
                    continue;
                }
                state = ST_BLANK; /*  => blank sequence  */
                goto redocurrent;

            } else if(c == 'L') {
                state = ST_WSTRING; /*  maybe wide char or string constant  */

            } else if(c == 'u' && (mode & GCTK_EXTENSIONS)) {
                state = ST_USTRING; /*  maybe wide char or string constant  */

            } else if(isalpha(c) || c == '_' ||
                      (c == '$' && (mode & GCTK_EXTENSIONS))) {
                state = ST_IDENT; /*  => identifier  */

            } else if(isdigit(c)) {
                if(c == '0')
                    state = ST_NUMHEX; /*  => start of octal or hexadecimal number  */
                else
                    state = ST_NUM; /*  => start of ordinary number  */

            } else if(c == '.') {
                state = ST_NUMFRAC2; /*  => may be floating point number  */

            } else if(c == '/') {
                state = ST_COMMENT; /*  => may be start of comment  */

            } else if(c == '\"') {/* " */
                state = ST_STRING; /*  => string  */
                if(!(mode & GCTK_INCSTARTSTOP))
                    indx--; /*  Exclude start char  */

            } else if(c == '\'') {
                state = ST_CHCONST; /*  => character constant  */
                if(!(mode & GCTK_INCSTARTSTOP))
                    indx--; /*  Exclude start char  */

            } else if((mode & GCTK_PPINCLUDELINE) && c == '<') {
                state = ST_PPSTRING; /*  => <..> string in preprocessor line  */
                if(!(mode & GCTK_INCSTARTSTOP))
                    indx--; /*  Exclude start char  */

            } else if((mode & GCTK_EXTENSIONS2) && c == '\\') {
                state = ST_ESCAPED;
                flag |= FLAG_ISSPECIAL;
                indx--; /*  Exclude '\\'  */

            } else if((mode & GCTK_CBRASTRING) && c == '{') {
                state = ST_BRASTRING;
                cbradef = "{}";
                cbranest = 1;
                if(!(mode & GCTK_INCSTARTSTOP)) indx--; /*  Exclude start char  */

            } else if((mode & GCTK_SQBRASTRING) && c == '[') {
                state = ST_BRASTRING;
                cbradef = "[]";
                cbranest = 1;
                if(!(mode & GCTK_INCSTARTSTOP)) indx--; /*  Exclude start char  */

            } else if((mode & GCTK_SQLCOMMENT) && c == '-') { /*  maybe SQL comment  */
                state = ST_COMMENTSQL0;

            } else if(ispunct(c)) {
                state = ST_SPECIAL; /*  => specials  */
                goto redocurrent;

            } else if(c == EOF) {
                token.TK_Value = CTK_EOF;

            } else if(iscntrl(c) || c >= 128) {
                indx--; /*  Ignore control characters and non-ASCII ones  */

            } else {
                token.TK_Value = c;
            }
            break;

            /* ---------------------------------------------------------- */
        case ST_ESCAPED: /* ***  Escaped character  *** */
            token.TK_Value = c;
            break;

            /* ---------------------------------------------------------- */
        case ST_BLANK: /* ***  Blank sequence  *** */
            if(!isspace(c)) {
                /*  End of blank sequence  */
                eungetc(c, eFp); indx--;
                token.TK_Value = (flag & FLAG_BLANKS_EOL) ? CTK_BLANKSEOL : CTK_BLANKS;

            } else if(c == '\n') {
                /*  Mark kind of sequence  */
                flag |= FLAG_BLANKS_EOL;
            }
            break;

            /* ---------------------------------------------------------- */
        case ST_COMMENT: /* ***  Found first char of comment start  *** */
            if(c == '*') {
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Drop comment start  */
                state = ST_COMMENT2; /*  => really comment  */

            } else if((mode & GCTK_CPPCOMMENT) && c == '/') {
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Drop comment start  */
                state = ST_COMMENTPP; /*  => really C++ comment  */

            } else {
                /*  No comment, put lost character back  */
                eungetc(c, eFp); indx--;
                state = ST_SPECIAL; /*  => specials  */
                c = '/';
                goto redocurrent;
            }

            break;

        case ST_COMMENT2: /* ***  Within comment  *** */
            if(c == '*') state = ST_COMMENT3; /*  => may be comment end  */
            break;

        case ST_COMMENT3: /* ***  Found first char of comment end  *** */
            if(c != '/') {
                state = ST_COMMENT2; /*  => May be still '*'  */
                goto redocurrent;

            } else {
                if(!(mode & GCTK_INCSTARTSTOP)) indx -= 2; /*  Drop comment end  */
                token.TK_Value = CTK_COMMENT;
            }
            break;

        case ST_COMMENTPP:  /* ***  Within C++ comment  */
        case ST_COMMENTSQL: /* ***  Within SQL comment  */
            if(c == '\n' || c == EOF) {
                if(c == '\n') eungetc(c, eFp); /*  Return '\n', too  */
                /*  Drop comment end. We NEVER let it be a part of the comment,
                 *   otherwise we have problems, i. e. that the preprocessor
                 *   relies on seeing the newline to recognize a preprocessor
                 *   command on the next line.
                 */
                indx--;

                token.TK_Value = state == ST_COMMENTSQL ? CTK_SQLCOMMENT : CTK_CPPCOMMENT;
            }
            break;

            /* ---------------------------------------------------------- */
        case ST_COMMENTSQL0:
            if(c != '-') {
                /*  No comment, put lost character back  */
                eungetc(c, eFp); indx--;
                state = ST_SPECIAL;
                c = '-';
                goto redocurrent;
            }

            if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Drop comment start  */
            state = ST_COMMENTSQL;
            break;

            /* ---------------------------------------------------------- */
        case ST_WSTRING:
            if(c == '\"') {
                flag |= FLAG_WIDESTRING;
                state = ST_STRING;
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Exclude L and "  */

            } else if(c == '\'') {
                flag |= FLAG_WIDESTRING;
                state = ST_CHCONST;
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Exclude L and '  */

            } else {
                state = ST_IDENT;
                goto redocurrent;
            }
            break;

            /* ---------------------------------------------------------- */
        case ST_USTRING:
            if(c == '\"') {
                flag |= FLAG_UWIDESTRING;
                state = ST_STRING;
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Exclude u and "  */

            } else if(c == '\'') {
                flag |= FLAG_UWIDESTRING;
                state = ST_CHCONST;
                if(!(mode & GCTK_INCSTARTSTOP)) indx = 0; /*  Exclude u and '  */

            } else {
                state = ST_IDENT;
                goto redocurrent;
            }

            break;

            /* ---------------------------------------------------------- */
        case ST_STRING: /* ***  Within string  *** */
            if(c == '\\') {
                if(!(mode & GCTK_SQLSTRING)) {
                    state = ST_STRING2; /*  => Escaped character  */
                    flag |= FLAG_ISSPECIAL;
                }

            } else if(c == '\"') { /* " */
                if(mode & GCTK_SQLSTRING) {
                    state = ST_STRING3;
                    flag |= FLAG_ISSPECIAL;

                } else {
                    if(!(mode & GCTK_INCSTARTSTOP))
                        indx--; /*  Drop string finisher  */

                    /*  => string finished  */
                    token.TK_Value = (flag & FLAG_WIDESTRING) ? CTK_WSTRING :
                        (flag & FLAG_UWIDESTRING) ? CTK_USTRING : CTK_STRING;
                }

            } else if((c == '\n' && !(mode & GCTK_EXTENSIONS)) || c == EOF) {
                JSSETERROR(_ERR_GETCTOKEN, 1); /*  Unfinished string  */
                EGOTO(error);
            }

            break;

        case ST_STRING2: /* ***  Escape char found within string  *** */
            state = ST_STRING; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;
            break;

        case ST_STRING3:
            if(c != '\"') {
                eungetc(c, eFp); indx--;

                /*  Was finisher...  */
                if(!(mode & GCTK_INCSTARTSTOP))
                    indx--; /*  Drop string finisher  */

                /*  => string finished  */
                token.TK_Value = CTK_STRING;
                break;
            }

            state = ST_STRING; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;

            break;

            /* ---------------------------------------------------------- */
        case ST_CHCONST: /* ***  Within character constant  *** */
            if(c == '\\') {
                if(!(mode & (GCTK_SHELLSTRING | GCTK_SQLSTRING))) {
                    state = ST_CHCONST2; /*  => Escaped character  */
                    flag |= FLAG_ISSPECIAL;
                }

            } else if(c == '\'') {
                if(mode & GCTK_SQLSTRING) {
                    state = ST_CHCONST3;
                    flag |= FLAG_ISSPECIAL;

                } else {
                    if(!(mode & GCTK_INCSTARTSTOP))
                        indx--; /*  Drop char const finisher */

                    /*  character constant finished  */
                    token.TK_Value = (flag & FLAG_WIDESTRING) ? CTK_WCHARCONST :
                        (flag & FLAG_UWIDESTRING) ? CTK_UCHARCONST : CTK_CHARCONST;
                }

            } else if((c == '\n' && !(mode & GCTK_EXTENSIONS)) || c == EOF) {
                JSSETERROR(_ERR_GETCTOKEN, 2); /*  Unfinished character constant  */
                EGOTO(error);
            }
            break;

        case ST_CHCONST2: /* ***  Escape char within character constant  *** */
            state = ST_CHCONST; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;
            break;

        case ST_CHCONST3:
            if(c != '\'') {
                eungetc(c, eFp); indx--;

                /*  Was finisher...  */
                if(!(mode & GCTK_INCSTARTSTOP)) indx--; /*  Drop string finisher  */

                /*  => character constant finished  */
                token.TK_Value = CTK_CHARCONST;
                break;
            }

            state = ST_CHCONST; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;

            break;

        case ST_PPSTRING: /* ***  Within preprocessor string  *** */
            if(c == '\\') {
                state = ST_PPSTRING2;
                flag |= FLAG_ISSPECIAL;

            } else if(c == '>') {
                if(!(mode & GCTK_INCSTARTSTOP)) indx--; /*  Drop finisher  */
                token.TK_Value = CTK_PPSTRING; /*  finished  */

            } else if(c == '\n' || c == EOF) {
                JSSETERROR(_ERR_GETCTOKEN, 3); /*  Unfinished preprocessor string  */
                EGOTO(error);
            }
            break;

        case ST_PPSTRING2: /* ***  Escape char within preprocessor string  *** */
            state = ST_PPSTRING; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;
            break;

            /* ---------------------------------------------------------- */
        case ST_IDENT: /* ***  Within identifier  *** */
            if(isalnum(c) || c == '_' || (c == '$' && (mode & GCTK_EXTENSIONS)))
                break;

            eungetc(c, eFp); indx--;
            token.TK_Value = CTK_IDENT; /*  => identifier  */
            break;

            /* ---------------------------------------------------------- */
        case ST_NUM: /* ***  Within number: integer part  *** */
            if(c == '.')
                state = ST_NUMFRAC; /*  => may be fractional part  */
            else if(c == 'e' || c == 'E')
                state = ST_NUMEXP; /*  => start of exponential part  */
            else if(!isdigit(c)) {
                state = ST_NUMTRAIL; /*  => suffixes  */
                goto redocurrent;
            }
            break;

        case ST_NUMFRAC: /* ***  Within number: fractional part  *** */
            if(c == 'e' || c == 'E')
                state = ST_NUMEXP; /* => exponential part  */
            else if(!isdigit(c)) {
                flag |= FLAG_NUMISFLOAT;
                state = ST_NUMTRAIL; /*  => suffixes  */
                goto redocurrent;
            }
            break;

        case ST_NUMFRAC2: /* ***  Decimal point without digits in front  *** */
            if(isdigit(c)) {
                state = ST_NUMFRAC; /*  => fractional part  */
            } else {
                state = ST_SPECIAL; /*  => this is not a number  */
                eungetc(c, eFp); indx--;
                c = '.'; /*  restart with '.'  */
                goto redocurrent;
            }
            break;
        case ST_NUMEXP: /* ***  Within number: sign of exponent  *** */
            if(c != '+' && c != '-' && !isdigit(c)) {
#if 1 /*  If we encounter "1e", don't throw an error, return "1" as number and continue with "e"  */
                eungetc(c, eFp); --indx;
                eungetc('e', eFp); --indx;
                token.TK_Value = CTK_NUMBER;
                break;
#else
                JSSETERROR(_ERR_GETCTOKEN, 4); /*  Wrong number found  */
                EGOTO(error);
#endif
            }

                state = ST_NUMEXP2;
            break;

        case ST_NUMEXP2: /* ***  Within number: number of exponent  *** */
            if(!isdigit(c)) {
                flag |= FLAG_NUMISFLOAT; /*  mark floating point number  */
                state = ST_NUMTRAIL;
                goto redocurrent;
            }
            break;
        case ST_NUMHEX: /* ***  Number starting with '0'  *** */
            if(c == 'x' || c == 'X')
                state = ST_NUMHEX2;
            else if(isdigit(c) && c != '8' && c != '9')
                state = ST_NUMOCT; /*  => octal number  */
            else if(c == '.')
                state = ST_NUMFRAC;
            else if(c == 'e' || c == 'E')
                state = ST_NUMEXP; /*  => floating point number  */
            else {
                state = ST_NUMTRAIL; /*  => suffixes  */
                goto redocurrent;
            }
            break;
        case ST_NUMHEX2: /* ***  Within hexadecimal number  *** */
            if(!isxdigit(c)) {
                flag |= FLAG_NUMISINT; /*  mark integer number  */
                state = ST_NUMTRAIL; /*  => suffixes  */
                goto redocurrent;
            }
            break;
        case ST_NUMOCT: /* ***  Within octal number  *** */
            if(!isdigit(c) || c == '8' || c == '9') {
                flag |= FLAG_NUMISINT; /*  mark integer number  */
                state = ST_NUMTRAIL; /*  => suffixes  */
                goto redocurrent;
            }
            break;
        case ST_NUMTRAIL: /* ***  Check for trailer of numbers  *** */
            if((c == 'f' || c == 'F') && !(flag & FLAG_NUMISINT)) {
                flag |= FLAG_NUMISFLOAT;
            } else if((c == 'u' || c == 'U') && !(flag & FLAG_NUMISFLOAT)) {
                flag |= FLAG_NUMISINT;
            } else if(!(c == 'l' || c == 'L')) {
                eungetc(c, eFp); indx--;

                /*  Choose correct number token  */
                if(flag & FLAG_NUMISINT)
                    token.TK_Value = CTK_INUMBER;
                else if(flag & FLAG_NUMISFLOAT)
                    token.TK_Value = CTK_DNUMBER;
                else
                    token.TK_Value = CTK_NUMBER;
            }
            break;
            /* ---------------------------------------------------------- */
        case ST_SPECIAL: /* ***  Specials: Operators etc.  *** */
            if(opDef) {
                const char *string;
                int lowerMark, i;

                TokBuffer[indx] = '\0';

                for(i = 0 ; opDef[i].KD_String ; i++)
                    if((string = tkStrcompare(opDef[i].KD_String, TokBuffer))) {
                        if(*string == '\0') {
                            /*  Complete operator found  */
                            token.TK_Value = opDef[i].KD_Token;
                        }
                        goto stopthebus;
                    }

                /*  Anything that matches found?  */
                if(opDef[i].KD_String)
                    goto stopthebus;

                /*  Set lower mark, cause buffer may start with "\\\n"  */
                lowerMark = (TokBuffer[0] == '\\' && TokBuffer[1] == '\n') ? 3 : 1;

                while(indx > lowerMark) {
                    /*  unget last character and try full matching  */
                    indx--;
                    eungetc(((unsigned char *)TokBuffer)[indx], eFp);
                    TokBuffer[indx] = '\0';

                    for(i = 0 ; opDef[i].KD_String ; i++)
                        if(tkStrcmp(opDef[i].KD_String, TokBuffer) == 0) {
                            token.TK_Value = opDef[i].KD_Token;
                            goto stopthebus;
                        }
                }
            }

            /*  Return single characters  */
            if(TokBuffer[0] == '\\' && TokBuffer[1] == '\n')
                token.TK_Value = ((unsigned char *)TokBuffer)[2];
            else
                token.TK_Value = ((unsigned char *)TokBuffer)[0];

        stopthebus:
            break;
            /* ---------------------------------------------------------- */
        case ST_READTOTERM: /* ***  Read until terminator  *** */
            if(c == GCTKTerminator) {
                eungetc(c, eFp); indx--;
                token.TK_Value = CTK_BLANKS;
            }
            break;
            /* ---------------------------------------------------------- */
        case ST_BRASTRING:
            if(c == cbradef[0])
                ++cbranest;
            else if(c == cbradef[1]) {
                if(--cbranest == 0) {
                    if(!(mode & GCTK_INCSTARTSTOP)) indx--; /*  Drop string finisher  */
                    token.TK_Value = cbradef[0] == '{' ? CTK_CBRASTRING : CTK_SQBRASTRING;
                }
            } else if(c == '\"')
                state = ST_BRADQUOTE;
            else if(c == '\'')
                state = ST_BRASQUOTE;
            else if(c == '/')
                state = ST_BRACOMMENT;
            break;
        case ST_BRADQUOTE:
            if(c == '\\') {
                state = ST_BRADQUOTEESC;
                flag |= FLAG_ISSPECIAL;
            } else if(c == '\"') state = ST_BRASTRING;
            break;
        case ST_BRADQUOTEESC:
            state = ST_BRADQUOTE; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;
            break;
        case ST_BRASQUOTE:
            if(c == '\\') {
                if(!(mode & GCTK_SHELLSTRING)) {
                    state = ST_BRASQUOTEESC;
                    flag |= FLAG_ISSPECIAL;
                }
            } else if(c == '\'') state = ST_BRASTRING;
            break;
        case ST_BRASQUOTEESC:
            state = ST_BRASQUOTE; /*  => Just ignore meaning of this character  */
            flag &= ~FLAG_ISSPECIAL;
            break;
        case ST_BRACOMMENT:
            if(c == '*')
                state = ST_BRACOMMENT2;
            else if((mode & GCTK_CPPCOMMENT) && c == '/')
                state = ST_BRACOMMENTPP;
            else
                state = ST_BRASTRING;
            break;
        case ST_BRACOMMENT2:
            if(c == '*') state = ST_BRACOMMENT3;
            break;
        case ST_BRACOMMENT3:
            if(c == '/') state = ST_BRASTRING;
            else if(c != '*') state = ST_BRACOMMENT2;
            break;
        case ST_BRACOMMENTPP:
            if(c == '\n') state = ST_BRASTRING;
            break;
        }
    } /* **  END OF MAIN LOOP  ** */

    /*  Set return token  */
    if(token.TK_Value == CTK_EOF) {
        token.TK_String = NULL;   /*  No string  */

    } else {
        /*  Set string correctly  */
        if(token.TK_Value >= CTK_MINIMUM) {
            token.TK_String = TokBuffer;
            TokBuffer[indx] = '\0';

            if(mode & GCTK_UNIQUE) {
                TokBuffer = NULL;
                TokBuffLen = 0;
            }
        } else {
            /*  No string if token is a single character  */
            token.TK_String = NULL;
        }
    }

    return token;
error:
    token.TK_Flags = 0;
    token.TK_Value = CTK_ERROR;
    token.TK_String = NULL;
    return token;
}

/*JS*********************************************************************
*
*************************************************************************/

int printCToken(FILE *fp, jsctoken_t token)

/************************************************************************/
{
    const char *string;
    int len;

    string = token.TK_String;

    switch(token.TK_Value) {
    case CTK_COMMENT:    len = fprintf(fp, "/*%s*/",  string);  break;
    case CTK_CPPCOMMENT: len = fprintf(fp, "//%s",    string);    break;
    case CTK_STRING:     len = fprintf(fp, "\"%s\"",  string);  break;
    case CTK_WSTRING:    len = fprintf(fp, "L\"%s\"", string); break;
    case CTK_USTRING:    len = fprintf(fp, "u\"%s\"", string); break;
    case CTK_PPSTRING:   len = fprintf(fp, "<%s>",    string);    break;
    case CTK_CHARCONST:  len = fprintf(fp, "'%s'",    string);    break;
    case CTK_WCHARCONST: len = fprintf(fp, "L'%s'",   string);   break;
    case CTK_UCHARCONST: len = fprintf(fp, "u'%s'",   string);   break;
    case CTK_EOF:        len = 0;  break;

    case CTK_BLANKSEOL:
    {
        int c;

        for(len = 0 ; (c = *string) ; ++string) {
            assert(isspace(c));

            if(c != '\n') {
                unsigned int off;

                for(off = 1 ; (c = string[off]) ; ++off)
                    if(c == '\n')
                        break;

                if(c == '\n') {
                    /*  skip spaces before newline  */
                    string += off - 1;
                }
                else
                    c = *string;
            }

            putc(c, fp);
        }
    }
    break;

    default:
        if(token.TK_Value < CTK_MINIMUM) {
            if(token.TK_Value >= 0)
                putc(token.TK_Value, fp);
            else
                putc('?', fp);

            len = 1;
            break;
        }
        /*  fall through  */

    case CTK_BLANKS:
    case CTK_IDENT:   case CTK_VANILLA:
    case CTK_NUMBER:  case CTK_INUMBER:    case CTK_DNUMBER:
        if(string == NULL) {
            len = fprintf(fp, "<NULL#%d>", token.TK_Value);
            break;
        }

        len = fwrite(string, strlen(string), 1, fp);

        break;
    }

    return len;
}

/*JS*********************************************************************
*
*************************************************************************/

int sprintCToken(char *dest, jsctoken_t token)

/************************************************************************/
{
    const char *string;
    int len;

    string = token.TK_String;

    switch(token.TK_Value) {
    case CTK_COMMENT:    len = sprintf(dest, "/*%s*/",  string);  break;
    case CTK_CPPCOMMENT: len = sprintf(dest, "//%s",    string);    break;
    case CTK_STRING:     len = sprintf(dest, "\"%s\"",  string);  break;
    case CTK_WSTRING:    len = sprintf(dest, "L\"%s\"", string); break;
    case CTK_USTRING:    len = sprintf(dest, "u\"%s\"", string); break;
    case CTK_PPSTRING:   len = sprintf(dest, "<%s>",    string);    break;
    case CTK_CHARCONST:  len = sprintf(dest, "'%s'",    string);    break;
    case CTK_WCHARCONST: len = sprintf(dest, "L'%s'",   string);   break;
    case CTK_UCHARCONST: len = sprintf(dest, "u'%s'",   string);   break;
    case CTK_EOF:
        len = 0;  dest[0] = '\0';
        break;

    default:
        if(token.TK_Value < CTK_MINIMUM) {
            dest[0] = (token.TK_Value >= 0) ? token.TK_Value : '?';
            dest[1] = '\0';

            len = 1;
            break;
        }

    case CTK_BLANKSEOL:
    case CTK_BLANKS:
    case CTK_IDENT:   case CTK_VANILLA:
    case CTK_NUMBER:  case CTK_INUMBER:    case CTK_DNUMBER:
        if(string != NULL) {
            len = strlen(string);
            memcpy(dest, string, len + 1);

        } else {
            memcpy(dest, "<NULL>", 7);
            len = 6;
        }

        break;
    }

    return len;
}

/*JS*********************************************************************
*
*************************************************************************/

const char *dbgNameCToken(jsctoken_t token)

/************************************************************************/
{
    if(0 <= token.TK_Value && token.TK_Value < CTK_MINIMUM) {
        static char Buffer[5];

        if(isprint(token.TK_Value))
            sprintf(Buffer, "'%c'", (char)token.TK_Value);
        else
            sprintf(Buffer, "0x%02x", (unsigned int)token.TK_Value);

        return Buffer;
    }

#define GEN_CASEVALUE(name)    case name: return #name

    switch(token.TK_Value) {
        GEN_CASEVALUE(CTK_EOF);

        GEN_CASEVALUE(CTK_ERROR);
        GEN_CASEVALUE(CTK_BLANKSEOL);
        GEN_CASEVALUE(CTK_BLANKS);
        GEN_CASEVALUE(CTK_COMMENT);
        GEN_CASEVALUE(CTK_CPPCOMMENT);

        GEN_CASEVALUE(CTK_STRING);
        GEN_CASEVALUE(CTK_WSTRING);
        GEN_CASEVALUE(CTK_USTRING);
        GEN_CASEVALUE(CTK_PPSTRING);
        GEN_CASEVALUE(CTK_CHARCONST);
        GEN_CASEVALUE(CTK_WCHARCONST);
        GEN_CASEVALUE(CTK_UCHARCONST);

        GEN_CASEVALUE(CTK_IDENT);
        GEN_CASEVALUE(CTK_NUMBER);
        GEN_CASEVALUE(CTK_INUMBER);
        GEN_CASEVALUE(CTK_DNUMBER);

        GEN_CASEVALUE(CTK_VANILLA);
        GEN_CASEVALUE(CTK_CBRASTRING);
        GEN_CASEVALUE(CTK_SQBRASTRING);
    }
#undef GEN_CASEVALUE

    return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

jsctoken_t copyCToken2(fstralloc_t *fstr, jsctoken_t token, int mode)

/************************************************************************/
{
    jsctoken_t token2;
    const char *tokString;

    token2.TK_Flags = token.TK_Flags & ~CTOKEN_FALLOCATED;
    token2.TK_Value = token.TK_Value;

    tokString = token.TK_String;

    if(mode & 1) {
        token2.TK_String = tokString;

    } else if(tokString == NULL) {
        token2.TK_String = NULL;

    } else if(fstr != NULL) {
	if((token2.TK_String = fstrdup(fstr, tokString)) == NULL)
	    EGOTO(error);

    } else {
	if((token2.TK_String = strdup(token.TK_String)) == NULL)
	    EGOTO(error);

	token2.TK_Flags |= CTOKEN_FALLOCATED;
    }

    return token2;
 error:
    token2.TK_Value = CTK_ERROR;
    return token2;
}

/*JS*********************************************************************
*
*************************************************************************/

void freeCToken(jsctoken_t token)

/************************************************************************/
{
    if((token.TK_Flags & CTOKEN_FALLOCATED) && token.TK_String != NULL)
        free((void *)token.TK_String);
}

/*JS*********************************************************************
*   Compares 'string' totally with the beginning of 'input'. Returns NULL
*    when the comparison fails, otherwise returns the first position after
*    the string in 'input'.
*    Special version: Ignore "\\\n" sequence in string.
*************************************************************************/

const char *tkStrcompare(const char *input, const char *string)

/************************************************************************/
{
    while(*string) {
        if(string[0] == '\\' && string[1] == '\n') {
            string += 2;
            continue; /*  *string might be '\0', so check for it  */
        }

        if(*string++ != *input++)
            return NULL;
    }

    return (char *)input;
}

/*JS*********************************************************************
*   Same as strcmp, but ignores "\\\n" sequences in string and input.
*************************************************************************/

int tkStrcmp(const char *input, const char *string)

/************************************************************************/
{
    for( ; ; input++, string++) {
        if(string[0] == '\\' && string[1] == '\n') string += 2;
        if(input[0] == '\\' && input[1] == '\n') input += 2;

        if(*string == '\0' || *string != *input) break;
    }

    return *input - *string;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int countnlCToken(const char *string)

/************************************************************************/
{
    int count;

    count = 0;
    while((string = strchr(string, '\n'))) {
        ++string;
        ++count;
    }

    return count;
}

/*JS*********************************************************************
 *
 *************************************************************************/

int updcolCToken(jsctoken_t token, int column)

/************************************************************************/
{
    int value;
    const char *string;

    if((value = token.TK_Value) == CTK_ERROR || value == CTK_EOF || value < 0)
        return column;

    if(value == CTK_CPPCOMMENT)
        return 0; /*  The comment ends with an implicit new line  */

    if((string = token.TK_String) != NULL)
        return updateColumn(column, string);

    return column + 1; /*  single character token  */
}

/*JS*********************************************************************
 *   Removes "\\\n" completely and sets '\n' and '\t' to ' '.
 *************************************************************************/

void tkDelNewLines(char *string)

/************************************************************************/
{
    char *dest;

    for(dest = string ; *string ; ) {
        if(*string == '\n' || *string == '\t') {
            *dest++ = ' ';
            string++;
        } else if(*string == '\\' && string[1] == '\n') {
            string += 2;
        } else
            *dest++ = *string++;
    }
    *dest = '\0';
}

/************************************************************************/

#ifdef TEST
int main(int argc, char *argv[])
{
    extfile_t *fp;
    jsctoken_t token;
    int mode;

    if(argc != 2) {
        fprintf(stderr, "Usage: testctoken <file>\n");
        return 5;
    }

    if((fp = efopen(argv[1], "r")) == NULL) {
        fprintf(stderr, "ERROR: Cannot open \"%s\" -- Break.\n", argv[1]);
        return 10;
    }

    mode = GCTK_CPPCOMMENT|GCTK_SQLCOMMENT|GCTK_EXTENSIONS|GCTK_EXTENSIONS2;

#if 0
    mode |= GCTK_CBRASTRING;
#endif

    do {
        token = getCToken(fp, NULL, mode);

#ifdef THROUGH
        /*  Simply printout token  */
        printCToken(stdout, token);
#else
        printf("%d: ", fp->EF_LineNr);

        switch(token.TK_Value) {
        case CTK_ERROR:
            EGOTO(error);

        case CTK_BLANKS:
        case CTK_BLANKSEOL:
            printf("Token BLANKS%s: \"%s\"\n",
                   token.TK_Value == CTK_BLANKSEOL ? "EOL" : "", token.TK_String);
            break;
        case CTK_COMMENT:
            printf("Token COMMENT: \"%s\"\n", token.TK_String);
            break;
        case CTK_CPPCOMMENT:
            printf("Token C++COMMENT: \"%s\"\n", token.TK_String);
            break;
        case CTK_SQLCOMMENT:
            printf("Token SQLCOMMENT: \"%s\"\n", token.TK_String);
            break;
        case CTK_STRING:
            printf("Token STRING: \"%s\"\n", token.TK_String);
            break;
        case CTK_PPSTRING:
            printf("Token PPSTRING: \"%s\"\n", token.TK_String);
            break;
        case CTK_CHARCONST:
            printf("Token CHARCONST: \"%s\"\n", token.TK_String);
            break;
        case CTK_IDENT:
            printf("Token IDENTIFIER: \"%s\"\n", token.TK_String);
            break;
        case CTK_NUMBER:
            printf("Token NUMBER: %s\n", token.TK_String);
            break;
        case CTK_INUMBER:
            printf("Token INUMBER: %s\n", token.TK_String);
            break;
        case CTK_DNUMBER:
            printf("Token DNUMBER: %s\n", token.TK_String);
            break;
        case CTK_CBRASTRING:
            printf("Token CBRASTRING: {%s}\n", token.TK_String);
            break;
        case CTK_EOF:
            break;
        default:
            if(token.TK_Value < CTK_MINIMUM)
                printf("Token %d: '%c'\n", token.TK_Value, token.TK_Value);
            else
                printf("Token %d: \"%s\"\n", token.TK_Value, token.TK_Value);
            break;
        }
#endif
    } while(token.TK_Value != CTK_EOF);

    efclose(fp);

    return 0;
error:
    jsperror("testgetctoken");
    return 30;
}
#endif
