/*JS*********************************************************************
*
*    Program : IFSTACK
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Generic if/else/endif-stack handling routines.
*
*              Copyright (C) 2001 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#ifndef IFSTK_TYPE_ASK
/*  START-DEFINITIONS  */
typedef struct {
  int *IS_Mem, *IS_Stack, IS_Max;
} ifstack_t;

/*  For "op" argument below  */
#define IFSTK_TYPE_ASK       0 /*  Only return current condition  */
#define IFSTK_TYPE_CANBETRUE 1 /*  Only return if 'elif' or 'else' could switch to true  */

#define IFSTK_TYPE_IF        5 /*  Handle 'if <cond>'  */
#define IFSTK_TYPE_ELIF      6 /*  Handle 'elif <cond>'  */
#define IFSTK_TYPE_ELSE      7 /*  Handle 'else'   */
#define IFSTK_TYPE_ENDIF     8 /*  Handle 'endif'  */

#define IDSTK_ELSESEEN   (1 << 2) /*  'else' part already encountered  */

#define ifstacklevel(istk)     ((istk)->IS_Mem == NULL ? 0 : \
				1 + ((istk)->IS_Stack - (istk)->IS_Mem))
#define ifstackhadelse(istk)   ((istk)->IS_Stack && \
				(*((istk)->IS_Stack) & IDSTK_ELSESEEN))
/*  END-DEFINITIONS  */
#endif

#define IDSTK_MASKED	 (1 << 0) /*  Current part is FALSE part   */
#define IDSTK_SKIPPING	 (1 << 1) /*  Remaining parts are skipped  */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype ifstack_t     *ifstackcreate(ifstack_t *istk);
Prototype int            ifstackfree(ifstack_t *istk, int flag);
Prototype int            ifstackop(ifstack_t *istk, int op, int cond);

/* ERROR-DEFINITIONS from ifstack label _ERR_IFSTACK
   Unmatched 'else', 'elif' or 'endif' part
   Duplicate 'else' part
   Illegal 'elif' part
   Unfinished 'if' construct
*/

/*JS*********************************************************************
*
*************************************************************************/

ifstack_t *ifstackcreate(ifstack_t *istk)

/************************************************************************/
{
  if(istk || (istk = (ifstack_t *)malloc(sizeof(*istk)))) {
    istk->IS_Mem = istk->IS_Stack = NULL;
    istk->IS_Max = 0;
  }

  return istk;
}

/*JS*********************************************************************
*   If bit 0 of flag is set, free istk. If bit 1 is set, do not check
*    for errors (and thus pollute the error code).
*************************************************************************/

int ifstackfree(ifstack_t *istk, int flag)

/************************************************************************/
{
  int ret;

  if(!(flag & 2) && istk->IS_Stack && istk->IS_Stack >= istk->IS_Mem) {
    JSSETERROR(_ERR_IFSTACK, 3);
    ret = -1;
  } else
    ret = 0;

  if(istk->IS_Mem) free(istk->IS_Mem);

  if(flag & 1) free(istk);

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int ifstackop(ifstack_t *istk, int op, int cond)

/************************************************************************/
{
  if(op != IFSTK_TYPE_ASK && op != IFSTK_TYPE_CANBETRUE) {
    /*  'if' is the only part where we push something  */
    if(op == IFSTK_TYPE_IF) {
      int indx;

      indx = istk->IS_Mem ? istk->IS_Stack - istk->IS_Mem : -1;

      /*  Enough memory?  */
      if(indx >= (istk->IS_Max - 1)) {
	istk->IS_Max += 10;
	if((istk->IS_Mem = (int *)
	    realloc(istk->IS_Mem, istk->IS_Max * sizeof(*(istk->IS_Mem))))
	   == NULL)
	  return -1;

	/*  Set value in new memory area  */
	istk->IS_Stack = istk->IS_Mem + indx;
      }

      if(indx >=0 && (*istk->IS_Stack & (IDSTK_SKIPPING|IDSTK_MASKED)))
	*++(istk->IS_Stack) = IDSTK_SKIPPING; /*  skip whole construction  */
      else
	*++(istk->IS_Stack) = IDSTK_MASKED; /*  simulate a previous FALSE part  */
    } else {
      /*  Any other part expects something on the stack  */
      if(istk->IS_Mem == NULL || istk->IS_Stack < istk->IS_Mem) {
	JSSETERROR(_ERR_IFSTACK, 0);
	return -1;
      }

      if(op == IFSTK_TYPE_ELSE) {
	if(*(istk->IS_Stack) & IDSTK_ELSESEEN) {
	  JSSETERROR(_ERR_IFSTACK, 1);
	  return -1;
	}
	*istk->IS_Stack |= IDSTK_ELSESEEN;
      } else if(op == IFSTK_TYPE_ELIF) {
	if(*(istk->IS_Stack) & IDSTK_ELSESEEN) {
	  JSSETERROR(_ERR_IFSTACK, 2);
	  return -1;
	}
      }
    }

    if(op == IFSTK_TYPE_ENDIF) {
      /*  We already know that there is something on the stack
       *   which we can pop (see some lines ago)
       */
      istk->IS_Stack--;
    } else if(!(*istk->IS_Stack & IDSTK_SKIPPING)) {
      if(!(*(istk->IS_Stack) & IDSTK_MASKED))
	/*  Go from TRUE to FALSE part  */
	*(istk->IS_Stack) |= IDSTK_SKIPPING; /*  keep on FALSE  */
      else if(op == IFSTK_TYPE_ELSE || cond)
	/*  May go from FALSE to TRUE part depending on condition  */
	*(istk->IS_Stack) &= ~IDSTK_MASKED;
    }
  }

  if(istk->IS_Stack == NULL || istk->IS_Stack < istk->IS_Mem)
    return TRUE;

  if(op == IFSTK_TYPE_CANBETRUE)
    return (*(istk->IS_Stack) & (IDSTK_MASKED | IDSTK_SKIPPING))
	   == IDSTK_MASKED;

  return (*(istk->IS_Stack) & (IDSTK_MASKED | IDSTK_SKIPPING)) == 0;
}

#ifdef TEST
#include <stdio.h>

static int showstack(ifstack_t *istk)
{
  int i, level, state;

  state = ifstackop(istk, IFSTK_TYPE_ASK, 0);
  level = ifstacklevel(istk);
  printf("%*s|%s[", level, "", state ? "ACTIVE LINES " : "");

  if(istk->IS_Mem)
    for(i = 0 ; i < level ; ++i)
      printf(" %d", istk->IS_Mem[i]);
  printf("]\n");

  return(state ? 1 : 0);
}

static int dostack(ifstack_t *istk, int op)
{
static const int Ops[] = {
  IFSTK_TYPE_IF, IFSTK_TYPE_ELIF, IFSTK_TYPE_ELSE, IFSTK_TYPE_ENDIF
};
  int depth, cond = op / 10;
  op = Ops[op % 10 - 1];

  depth = istk->IS_Mem ? istk->IS_Stack - istk->IS_Mem + 1 : 0;
  printf("%*s#", op == IFSTK_TYPE_IF ? depth : depth - 1, "");

  if(op == IFSTK_TYPE_IF) printf("if %d\n", cond);
  else if(op == IFSTK_TYPE_ELIF) printf("elif %d\n", cond);
  else if(op == IFSTK_TYPE_ELSE) printf("else\n");
  else if(op == IFSTK_TYPE_ENDIF) printf("endif\n");

  return(ifstackop(istk, op, cond));
}

static struct {
  int State; /*  The expected state  */
  int Op;    /*  the operation to perform 10digit specifies conditional value  */
  int Ret;   /*  the expected return value of ifstackop  */
} Tests[] = {
  { 1, 11,  1 }, /* #if 1   */
  { 1,  3,  0 }, /* #else   */
  { 0,  4,  1 }, /* #endif  */
  { 0, -1,  0 },

  { 1, 11,  1 }, /* #if 1   */
  { 1,  2,  0 }, /* #elif 0 */
  { 0,  2,  0 }, /* #elif 1 */
  { 0,  3,  0 }, /* #else   */
  { 0,  4,  1 }, /* #endif  */
  { 0, -1,  0 },

  { 1,  1,  0 }, /* #if 0   */
  { 0,  2,  0 }, /* #elif 0 */
  { 0, 12,  1 }, /* #elif 1 */
  { 1,  3,  0 }, /* #else   */
  { 0,  4,  1 }, /* #endif  */
  { 0, -1,  0 },

  { 1,  1,  0 }, /* #if 0   */
  { 0,  2,  0 }, /* #elif 0 */
  { 0,  2,  0 }, /* #elif 0 */
  { 0,  3,  1 }, /* #else   */
  { 1,  4,  1 }, /* #endif  */
  { 0, -1,  0 },

  { 1, 11,  1 }, /* #if 1   */
  { 1, 11,  1 }, /*  #if 1   */
  { 1,  3,  0 }, /*  #else   */
  { 0,  4,  1 }, /*  #endif  */
  { 1,  3,  0 }, /* #else   */
  { 0, 11,  0 }, /*  #if 1   */
  { 0,  3,  0 }, /*  #else   */
  { 0,  4,  0 }, /*  #endif  */
  { 0,  4,  1 }, /* #endif  */
  { 0, -1,  0 },

  { 1,  2, -1 }, /*  unmatched #elif   */
  { 1,  3, -1 }, /*  unmatched #else   */
  { 1,  4, -1 }, /*  unmatched #endif  */
  { 0, -1,  0 },

  { 1, 11,  1 }, /*  #if 1   */
  { 1,  3,  0 }, /*  #else   */
  { 0,  3, -2 }, /*  duplicate #else   */
  { 0,  4,  1 }, /*  illegal #elif   */
  { 0, -1,  0 },

  { 1, 11,  1 }, /*  #if 1   */
  { 1,  3,  0 }, /*  #else   */
  { 0,  2, -3 }, /*  illegal #elif   */
  { 0,  4,  1 }, /*  #endif  */
  { 0, -1,  0 },
};

int main(int argc, char *argv[])
{
  ifstack_t *istk;
  int i;

  if((istk = ifstackcreate(NULL)) == NULL) goto error;

  if(argc > 1) {
    for(i = 1 ; i < argc ; ++i) {
      showstack(istk);
      if(dostack(istk, atoi(argv[i])) == -1) goto error;
    }
    showstack(istk);
  } else {
    int j, errcnt = 0;

    for(i = 0 ; i < sizeof(Tests) / sizeof(Tests[0]) ; ++i) {
      if(Tests[i].Op == -1) {
	printf("------------------------------\n");
	continue;
      }

      /*  Check state  */
      if(showstack(istk) != Tests[i].State) {
	printf("  ERROR: Expected %d\n", Tests[i].State);
	++errcnt;
      }

      j = dostack(istk, Tests[i].Op);
      if(j == -1) {
	if(Tests[i].Ret >= 0 || JSErrNo != (_ERR_IFSTACK - 1 - Tests[i].Ret)) {
	  printf("  ERROR: 'ifstackop' did not fail as expected (%d != %d)!\n",
		 JSErrNo - _ERR_IFSTACK, -Tests[i].Ret - 1);
	  ++errcnt;
	}
      } else {
	if(Tests[i].Ret < 0 || (Tests[i].Ret && !j) || (!Tests[i].Ret && j)) {
	  printf("  ERROR: State did not change as expected!\n");
	  ++errcnt;
	}
      }
    }

    if(errcnt) printf("Found %d errors\n", errcnt);
  }

  if(ifstackfree(istk, 1) == -1) goto error;

  return 0;
error:
  jsperror("testifstack");
  return 20;
}
#endif
