/*JS*********************************************************************
*
*    Program : JSGETLINE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Get a line of arbitrary length with some additional
*              options.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>

/*********     DEFINES                                          *********/
#define COMMENT_INTRODUCER  '#'  /*  Comments start with this character */

/*  START-DEFINITIONS  */
#include <stdio.h>   /*	 For "FILE" definition    */

/*  For parameter 'mode' of jsgetline  */
#define GETLINE_RMEOL     (1<<0)  /*  Remove '\n'  */
#define GETLINE_CONCAT    (1<<1)  /*  Allow line concatenation  */
#define GETLINE_RMCONCAT  (1<<2)  /*  Remove "\\\n" sequences  */
#define GETLINE_COMMENT   (1<<3)  /*  Allow shell like comments (" .. # ..") */
#define GETLINE_UNIQUE    (1<<4)  /*  Make returned line unique  */
#define GETLINE_RMBSLASH  (1<<5)  /*  Leave '\n', but remove '\\'  */
/*  END-DEFINITIONS  */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype char          *jsgetline(FILE *fp, int mode);
Prototype void           jsgetlineclear(void);

/*********     STATIC VARIABLES                                 *********/
static char *Buffer;
static long  BuffLen;

/*JS*********************************************************************
*   Read a line of arbitrary length into a static buffer.
*************************************************************************/

char *jsgetline(FILE *fp, int mode)

/************************************************************************/
{
  long i, c, state;

  for(state = 0, i = 0 ; state >= 0 && (c = getc(fp)) != EOF ; ) {
    if((i + 2) >= BuffLen) {
      BuffLen += 100;
      if((Buffer = (char *)realloc(Buffer, BuffLen)) == NULL) {
	BuffLen = 0;
	return NULL;
      }
    }

#ifdef _WIN32
    /*  Swallow '\r' in front of '\n'  */
    if(c == '\r') {
      if((c = getc(fp)) != '\n') {
	ungetc(c, fp);
	c = '\r';
      }
    }
#endif

    switch(state) {
    case 0:
      if(c == '\n') {
	state = -1;
	if(mode & GETLINE_RMEOL)
	  continue;

      } else if(c == '\\' && (mode & GETLINE_CONCAT)) {
	state = 1;

      } else if(c == COMMENT_INTRODUCER && (mode & GETLINE_COMMENT)) {
	state = 2;
	continue;
      }

      break;

    case 1: /* **  Ignore special meaning  ** */
      state = 0;
      if(c == '\n' && (mode & (GETLINE_RMCONCAT|GETLINE_RMBSLASH))) {
	i--;
	if(!(mode & GETLINE_RMBSLASH))
	  continue;
      }
      break;

    case 2: /* **  Scan comment 'til EOL  ** */
      if(c == '\n') {
	state = -1;
	if(!(mode & GETLINE_RMEOL)) break;
      }

      continue; /*  don't save  */
    }

    Buffer[i++] = c;
  }

  if(c == EOF && i == 0)
    return NULL;

  Buffer[i] = '\0';

  if(mode & GETLINE_UNIQUE) {
    char *save = Buffer;

    Buffer = NULL;
    BuffLen = 0;
    return save;
  }

  return Buffer;
}

/*JS*********************************************************************
*   Clear the memory used by jsgetline (like a 'late' GETLINE_UNIQUE).
*************************************************************************/

void jsgetlineclear(void)

/************************************************************************/
{
  Buffer = NULL;
  BuffLen = 0;
}

#ifdef TEST
int main(int argc, char *argv[])
{
  char *line;
  int mode;

  if(argc == 2)
    mode = atoi(argv[1]);
  else
    mode = GETLINE_RMEOL | GETLINE_CONCAT | GETLINE_RMCONCAT | GETLINE_COMMENT;

  while((line = jsgetline(stdin, mode))) {
    printf("\"%s\"\n", line);
  }

  return(0);
}
#endif
