/*JS*********************************************************************
*
*    Program : JSTRACE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Tracing infrastructure
*
*              Copyright (C) 2010 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <assert.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>

#include <errno.h>

#include <jssubs.h>

typedef struct {
  int Flags;
  FILE *File;

} tracecntrl_t;

#define JSTRACE_ENVVAR       "JSTRACELEVEL"

#define _jstraceinit()				\
  (TraceControl == NULL ? _jstraceinit0() : 0)

#define _jstraceopenfile(type)					\
  (TraceControl[type].File == NULL ? _jstraceopenfile0(type) : 0)

/*  If trace code is included, generate at least one for the main program  */
/*  TRACE-DEFINITION symbol TRACETYPE_MAIN name MAIN */

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype void           jstrace(int type, int level,
				 const char *format, ...);
Prototype FILE          *jstracegetfp(int type, int level);
Prototype int            jstracelevel(int type);
Prototype int            jstracemask(int type);

Prototype int            jstraceset(const char *type, int level);
Prototype int            jstraceseti(int ttype, int level);
Prototype int            jstracesets(const char *string);

static int               _jstraceinit0(void);
static int               _jstraceopenfile0(int type);
static int               _jstracesets(const char *string);

Prototype char           JSTraceDirectory[JS_FILENAME_MAX];
Prototype int            JSTraceFlags;

/*********     GLOBAL / STATIC VARIABLES                        *********/
static char TraceBuffer[4096];

static tracecntrl_t *TraceControl;

char JSTraceDirectory[JS_FILENAME_MAX];
int JSTraceFlags;

static int Trace2Stderr;

/*JS*********************************************************************
*
*************************************************************************/

void jstrace(int type, int level, const char *format, ...)

/************************************************************************/
{
  FILE *fp;

  if((fp = jstracegetfp(type, level)) == NULL)
    return;

  {
    va_list ap;

    va_start(ap, format);
    if(vsprintf(TraceBuffer, format, ap) == -1)
      EGOTO(error);

    va_end(ap);
  }

  if(fp != stderr) {
    if((JSTraceFlags & 1) && !(TraceControl[type].Flags & 1)) {
      while(level-- > 0)
	fputc(' ', fp);
    }

    level = strlen(TraceBuffer) - 1;
    if(level >= 0 && TraceBuffer[level] == '\n')
      TraceControl[type].Flags &= ~1;
    else
      TraceControl[type].Flags |= 1;

  }

  if(fputs(TraceBuffer, fp) == -1)
    fprintf(stderr, "ERROR writing trace: %s\n", strerror(errno));

  fflush(fp);

  return;
error:
  fprintf(stderr, "ERROR in trace routine: %s\n", strerror(errno));
  return;
}

/*JS*********************************************************************
*
*************************************************************************/

FILE *jstracegetfp(int type, int level)

/************************************************************************/
{
  FILE *fp;

  assert(0 <= type && type < JSNTraceTypes);

  _jstraceinit();

  if(level >= 0) {
    int trcLevel;

    if((trcLevel = JSTraceLevels[type]) < 0)
      trcLevel = 0;

    if(level > trcLevel)
      return NULL;

  } else {
    int trcMask;

    level = -level;
    if((trcMask = -JSTraceLevels[type]) < 0)
      trcMask = 0;

    if(!(level & trcMask))
      return NULL;
  }

  if(_jstraceopenfile(type) == -1) {
    fprintf(stderr, "ERROR opening trace file: %s\n", strerror(errno));
  }

  if((fp = TraceControl[type].File) == NULL)
    fp = stderr;

  return fp;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstracelevel(int type)

/************************************************************************/
{
  _jstraceinit();

  return JSTraceLevels[type] > 0 ? JSTraceLevels[type] : 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstracemask(int type)

/************************************************************************/
{
  _jstraceinit();

  return JSTraceLevels[type] < 0 ? JSTraceLevels[type] : 0;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstraceset(const char *type, int level)

/************************************************************************/
{
  int itype;

  for(itype = 0 ; itype < JSNTraceTypes ; ++itype)
    if(strcmp(JSTraceTypes[itype], type) == 0)
      return jstraceseti(itype, level);

  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstraceseti(int type, int level)

/************************************************************************/
{
  int oldLevel;

  _jstraceinit();

  if(!(0 <= type && type < JSNTraceTypes))
    return -1;

  oldLevel = JSTraceLevels[type];
  JSTraceLevels[type] = level;

  return oldLevel;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstracesets(const char *string)

/************************************************************************/
{
  _jstraceinit();

  return _jstracesets(string);
}

/*JS*********************************************************************
*
*************************************************************************/

static int _jstraceinit0(void)

/************************************************************************/
{
  const char *string;
  int i;

  if((TraceControl = (tracecntrl_t *)
      malloc((JSNTraceTypes + 1) * sizeof(*TraceControl))) == NULL) { /*DBG*/
    fprintf(stderr, "FATAL: Low on memory!\n");
    abort();
  }

  for(i = 0 ; i < JSNTraceTypes ; ++i) {
    TraceControl[i].Flags = 0;
    TraceControl[i].File = NULL;
  }

  if((string = getenv(JSTRACE_ENVVAR)) != NULL)
    return _jstracesets(string);

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int _jstraceopenfile0(int type)

/************************************************************************/
{
  FILE *fp;
  int ret;

  ret = 0;
  if((fp = TraceControl[type].File) != NULL &&
     fp != stderr &&
     fclose(fp) == EOF)
    ret = -1;

  if(Trace2Stderr) {
    fp = stderr;

  } else {
    char fileName[JS_FILENAME_MAX];

    if(JSTraceDirectory[0] != '\0')
      sprintf(fileName, "%s/%s.trc", JSTraceDirectory, JSTraceTypes[type]);
    else
      sprintf(fileName, "%s.trc", JSTraceTypes[type]);

    if((fp = fopen(fileName, "a")) == NULL)
      ret = -1;
  }

  TraceControl[type].File = fp;

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

static int _jstracesets(const char *string)

/************************************************************************/
{
  while(*string) {
    const char *string2;
    int level, type;

    if(isdigit(*string)) {
      level = atoi(string);
      for(type = 0 ; type < JSNTraceTypes ; ++type)
	JSTraceLevels[type] = level;

      break;
    }

    if((string2 = jstrstartswith(string, "stderr")) != NULL) {
      Trace2Stderr = 1;
      string = string2;

    } else {
      for(type = 0 ; type < JSNTraceTypes ; ++type) {
	if((string2 = jstrstartswith(string, JSTraceTypes[type])) != NULL &&
	   !isalnum(*string2)) {
	  string = string2;
	  break;
	}
      }

      if(type == JSNTraceTypes) {
	string2 = string;
	while(isalnum(*string))
	  ++string;

	fprintf(stderr, "WARNING: Unknown trace type '%.*s' - ignored!\n",
		(int)(string - string2), string2);
      }

      if(*string == '=' || *string == ':') {
	++string;
	level = strtol(string, (char **)&string2, 0);
	if(string2 == string) {
	  fprintf(stderr, "WARNING: Illegal trace definition '%s' - ignoring remainder!\n",
		  string);
	  return 0;
	}

	string = string2;

      } else
	level = 3;

      if(type < JSNTraceTypes)
	JSTraceLevels[type] = level;
    }

    if(*string && strchr(",;+ ", *string) != NULL)
      ++string;
  }

  return 0;
}
