/*JS*********************************************************************
*
*    Program : PIGGYBACKFILE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Add files 'piggyback' in executable
*
*              Copyright (C) 2003 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdlib.h>
#include <string.h>

#include <jsconfig.h>

#include <jsendian.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS 1 */
#define PGBACK_MAGIC        "PGO1"
#define PGBACK_MAGIC_INV    "1OGP"

#define PGBACK_MAGIC_LEN    4

#define PGBACK_FINAL        PGBACK_MAGIC "\377\377"

#define PGBACK_FLEN4BYTE       (1 << 0)
#define PGBACK_FBWCOMP1        (1 << 1)
#define PGBACK_FBWENCRYPT      (1 << 2)

#ifdef CONFIG_HAS_BWCOMPRESSION
#define CONFIG_PIGGYBACK_COMPRESS
/*  This is rather big, but should get the best out of it  */
#define BWCOMP1_BLOCKSIZE        500000
#endif
/*  END-DEFINITIONS  */

/*
 *   Format of a chunk: 4 bytes magic value. 2 bytes flags (-1 to denote
 *    end), then 2 or 4 bytes length, then the file name, terminated by
 *    a '\0', then the actual data.
 */

typedef struct {
  union {
    unsigned char *FDATA_data;
    const char *FDATA_name;
  } c;
#define FDATA_Data    c.FDATA_data
#define FDATA_Name    c.FDATA_name
  long FDATA_Length;

#ifdef CONFIG_PIGGYBACK_COMPRESS
  unsigned char *FDATA_BWBuffer;
  long FDATA_BWLeft, FDATA_BWTotal;

  long FDATA_CryptPwd;
#endif
} filedata_t;

/*  To encrypt entries, we calculate the checksum of the password
 *   and use this to generate a random sequence with which we
 *   "scramble" the compressed values.
 */
#define CHKSUM_START    0xffffffff
#define CHKSUM_SIZE     2

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype extfile_t     *piggybackopener(void *data, const char *name,
 					 const char *mode);
static int               piggybackopen(void *user, int flags, long len,
				       const char *name,
				       unsigned char *data);
static int               piggybackget(void *user, int flag);

Prototype void           piggybackcrypt(long pwd, unsigned char *data,
					unsigned long len);
Prototype unsigned long  piggystring2pwd(const char *name);
Prototype int            piggybackscan(unsigned char *data,
				       int (*routine)(void *user,
						      int flags, long len,
						      const char *name,
						      unsigned char *d),
				       void *user,
				       unsigned char **pend, int mode);

Prototype unsigned long (*PiggyBackGetPwd)(const char *name);

/*********     GLOBAL VARIABLES                                 *********/
unsigned long (*PiggyBackGetPwd)(const char *name);

/* ERROR-DEFINITIONS from piggybackopener label _ERR_PIGGYBACKOP
   Cannot write files
   File not found in load
   Compression not supported
   Cannot read encrypted entry
*/

/*JS*********************************************************************
*
*************************************************************************/

extfile_t *piggybackopener(void *data, const char *name, const char *mode)

/************************************************************************/
{
  filedata_t *fdata;
  int ret;

  if(mode[0] != 'r') {
    JSSETERROR(_ERR_PIGGYBACKOP, 0);
    EGOTO(error2);
  }

  if((fdata = (filedata_t *)malloc(sizeof(*fdata))) == NULL)
    EGOTO(error);

  if(*name == CONFIG_DIR_DELIMITER)
    ++name;

  fdata->FDATA_Name = name;

  if((ret = piggybackscan((unsigned char *)data, piggybackopen, fdata,
			  NULL, 0)) == -1)
    EGOTO(error);

  if(ret == 0) {
    JSSETERROR2(_ERR_PIGGYBACKOP, 1, name);
    EGOTO(error);
  }

  return efmksrc(piggybackget, fdata);
error:
  free(fdata);
error2:
  return NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

static int piggybackopen(void *user, int flags, long len,
			 const char *name,
			 unsigned char *data)

/************************************************************************/
{
#define filedata    ((filedata_t *)user)
  if(strcmp(name, filedata->FDATA_Name) != 0)
    return 0;

  filedata->FDATA_Data = data;
  filedata->FDATA_Length = len;

#ifdef CONFIG_PIGGYBACK_COMPRESS
  if(flags & PGBACK_FBWCOMP1) {
    if((filedata->FDATA_BWBuffer =
	malloc(BWCOMP1_BLOCKSIZE + bwestimateaux(BWCOMP1_BLOCKSIZE))) == NULL)
      return -1;

    filedata->FDATA_BWLeft = filedata->FDATA_BWTotal = 0;

    if(!(flags & PGBACK_FBWENCRYPT))
      filedata->FDATA_CryptPwd = -1;
    else {
      if(PiggyBackGetPwd == NULL) {
	JSSETERROR2(_ERR_PIGGYBACKOP, 3, name);
	return -1;
      }

      /*  Ask provider for a password  */
      if((filedata->FDATA_CryptPwd = (*PiggyBackGetPwd)(filedata->FDATA_Name)) == -1)
	return -1;
    }

  } else {
    filedata->FDATA_BWBuffer = NULL;

    /*  Crypting is only done together with compression  */
    if(flags & PGBACK_FBWENCRYPT) {
      JSSETERROR2(_ERR_PIGGYBACKOP, 3, name);
      return -1;
    }
  }
#else
  if(flags & PGBACK_FBWCOMP1) {
    /*  Not supported, but compressed file found!  */
    JSSETERROR2(_ERR_PIGGYBACKOP, 2, name);
    return -1;
  }

  if(flags & PGBACK_FBWENCRYPT) {
    /*  No compression, no encryption  */
    JSSETERROR2(_ERR_PIGGYBACKOP, 3, name);
    return -1;
  }
#endif

  return 1;
#undef filedata
}

/*JS*********************************************************************
*
*************************************************************************/

static int piggybackget(void *user, int flag)

/************************************************************************/
{
#define filedata    ((filedata_t *)user)

  if(flag == -1) {
#ifdef CONFIG_PIGGYBACK_COMPRESS
    /*  Clean up  */
    if(filedata->FDATA_BWBuffer)
      free(filedata->FDATA_BWBuffer);
#endif

    free(user);
    return 0;
  }

#ifdef CONFIG_PIGGYBACK_COMPRESS
  if(filedata->FDATA_BWBuffer) {
    /*  COMPRESSED FILE  */
    if(filedata->FDATA_BWLeft >= filedata->FDATA_BWTotal) {
      unsigned long len;

      if((len = filedata->FDATA_Length) == 0)
	return EOF;

      if((filedata->FDATA_BWTotal =
	  bwdecompress(filedata->FDATA_Data, &len,
		       filedata->FDATA_BWBuffer,
		       filedata->FDATA_BWBuffer + BWCOMP1_BLOCKSIZE,
		       0)) == -1)
	return -1;

      if(filedata->FDATA_CryptPwd != -1)
	piggybackcrypt(filedata->FDATA_CryptPwd,
		       filedata->FDATA_Data + BWTRANS_BLOCKHDR_SIZE,
		       filedata->FDATA_BWTotal - BWTRANS_BLOCKHDR_SIZE);

      filedata->FDATA_Data   += len;
      filedata->FDATA_Length -= len;

      filedata->FDATA_BWLeft = 0;
    }

    return filedata->FDATA_BWBuffer[filedata->FDATA_BWLeft++];
  }
#endif

  /*    UNCOMPRESSED FILE  */
  if(filedata->FDATA_Length == 0)
    return EOF;

  filedata->FDATA_Length--;

  return *(filedata->FDATA_Data)++;

#undef filedata
}

#ifdef CONFIG_PIGGYBACK_COMPRESS
/*JS*********************************************************************
*
*************************************************************************/

void piggybackcrypt(long pwd, unsigned char *data,
		    unsigned long len)

/************************************************************************/
{
  unsigned char hashs[JS_INT322CHAR_LEN];
  unsigned long pos;

  for(pos = 0 ; pos < len ; ++pos) {
    int hpos;

    if((hpos = pos % JS_INT322CHAR_LEN) == 0) {
      long hash;

      hash = lran0(pos == 0 ? &pwd : NULL);
      JS_INT322CHAR_LE(hash, hashs);
    }

    data[pos] ^= hashs[hpos];
  }
}
#endif

/*JS*********************************************************************
*
*************************************************************************/

unsigned long  piggystring2pwd(const char *password)

/************************************************************************/
{
  return crcCalc(CHKSUM_START, (unsigned char *)password, strlen(password),
		 JCrcTab, CHKSUM_SIZE);
}

/*JS*********************************************************************
*
*************************************************************************/

int piggybackscan(unsigned char *data,
		  int (*routine)(void *u,
				 int flags, long len,
				 const char *name,
				 unsigned char *d),
		  void *user,
		  unsigned char **pend, int mode)

/************************************************************************/
{
  int ret = 0;

  if(pend) *pend = NULL; /*  preset  */

  for(;;) {
    int i, flags;
    long len;

    /*  Every section has to start with the magic value  */
    for(i = 0 ; i < PGBACK_MAGIC_LEN ; ++i)
      /*  We have this reversed to make sure we are not generating the
       *   magic tag ourselves (if an external scanner scans our
       *   executable)
       */
      if(data[i] != PGBACK_MAGIC_INV[PGBACK_MAGIC_LEN - 1 - i]) {
	JSSETERROR(_ERR_PIGGYBACKOP, 0);
	EGOTO(error);
      }

    data += PGBACK_MAGIC_LEN;

    flags = JS_CHAR2INT16_LE(data);
    data += JS_INT162CHAR_LEN;

    if(flags == 0xffff) {
      if(pend) *pend = data;
      break;
    }

    if(flags & PGBACK_FLEN4BYTE) {
      len = JS_CHAR2INT32_LE(data);
      data += JS_INT322CHAR_LEN;
    } else {
      len = JS_CHAR2INT16_LE(data);
      data += JS_INT162CHAR_LEN;
    }

    i = strlen((char *)data) + 1;

    if((ret = routine(user, flags, len, (const char *)data, data + i)) == -1)
      EGOTO(error);

    if(ret != 0)
      break;

    /*  Skip name and data  */
    data += i + len;
  }

  return ret;
error:
  return -1;
}

/************************************************************************/
#if defined(PGBACK_FILE)

#include PGBACK_FILE

int main(int argc, char *argv[])
{
  int i;

  if(epushopener("test:", piggybackopener, (void *)Data) == -1)
    goto error;

  for(i = 1 ; i < argc ; ++i) {
    extfile_t *efp;
    int c;

    if((efp = efopen(argv[i], "r")) == NULL)
      goto error;

    printf("---'%s'\n", argv[i]);

    while((c = efgetc(efp)) != EOF)
      putchar(c);

    if(efclose(efp) == -1)
      goto error;
  }

  return 0;
error:
  jsperror("testpiggyback");
  return 30;
}
#endif
