/*JS*********************************************************************
*
*    Program : RANDOM
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Some random number generators, taken from Numerical
*              Recipes in C, second edition.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>

/*********     DEFINES                                          *********/

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype double         ran0(const long *pseed);
Prototype long           lran0(const long *pseed);

Prototype double         ran1(const long *pseed);
Prototype long           lran1(const long *pseed);

/*JS*********************************************************************
*
*************************************************************************/

double ran0(const long *pseed)

#define RND_IA     16807
#define RND_IM     2147483647
#define RND_AM     (1.0 / RND_IM)
#define RND_IQ     127773
#define RND_IR     2836
#define RND_MASK   123459876

/************************************************************************/
{
  long lvalue;

  lvalue = lran0(pseed);

  /*  Compute Seed to a floating point result  */
  return lvalue * RND_AM;
}

#undef RND_AM

/*JS*********************************************************************
*   "Minimal" random number generator of Park and Miller. Returns a
*    uniform random deviate between 0.0 and 1.0. If pseed != NULL, the
*    start value is set. Set or reset seed to any integer value (except
*    the unlikely value MASK) to initialize the sequence. Seed must not
*    be altered between calls for successive deviates in a sequence.
*************************************************************************/

long lran0(const long *pseed)

/************************************************************************/
{
  long k;
static long Seed = RND_MASK;

  if(pseed && *pseed != RND_MASK) {
    /*  XORing with MASK allows use of zero and other simple
     *   bit patterns for Seed.
     */
    Seed = *pseed ^ RND_MASK;
  }

  k = Seed / RND_IQ;

  /*  Compute Seed = (IA * Seed) % IM without overflow by Schrage's method  */
  Seed = RND_IA * (Seed - k * RND_IQ) - RND_IR * k;
  if(Seed < 0) Seed += RND_IM;

  return Seed;
}
#undef RND_IA
#undef RND_IM
#undef RND_IQ
#undef RND_IR
#undef RND_MASK

/*JS*********************************************************************
*
*************************************************************************/

double ran1(const long *pseed)

#define RND_IA     16807
#define RND_IM     2147483647
#define RND_AM     (1.0 / RND_IM)
#define RND_IQ     127773
#define RND_IR     2836
#define RND_NTAB   32
#define RND_NDIV   (1 + (RND_IM - 1 ) / RND_NTAB)

#define RND_EPS    1.0e-15
#define RND_RNMX   (1.0 - RND_EPS)

/************************************************************************/
{
  long lvalue;
  double temp;

  lvalue = lran1(pseed);

  /*  Because users don't expect endpoint values  */
  if((temp = RND_AM * lvalue) > RND_RNMX)
    return RND_RNMX;

  return temp;
}

#undef RND_RNMX
#undef RND_EPS

/*JS*********************************************************************
*   "Minimal" random number generator of Park and Miller with Bays-Durham
*    shuffle and added safeguards. Returns a uniform random deviate
*    between 0.0 and 1.0 (exclusive the endpoint values). If pseed != NULL,
*    the start value is set. Set or reset seed to a negative integer value
*    to initialize the sequence. Seed must not be altered between calls
*    for successive deviates in a sequence.
*************************************************************************/

long lran1(const long *pseed)

/************************************************************************/
{
  long k;
  int j;
static long Seed = -1;
static long Iy = 0;
static long Iv[RND_NTAB];

  if(pseed)
    Seed = *pseed;

  if(Seed <= 0 || Iy == 0) { /*  Initialize  */
    if(-Seed < 1) Seed = 1; /*  Be sure to prevent Seed = 0  */
    else Seed = -Seed;

    /*  Load the shuffle table (after 8 warm ups)  */
    for(j = RND_NTAB + 7 ; j >= 0 ; j--) {
      k = Seed / RND_IQ;
      Seed = RND_IA * (Seed - k * RND_IQ) - RND_IR * k;
      if(Seed < 0 ) Seed += RND_IM;
      if(j < RND_NTAB) Iv[j] = Seed;
    }
    Iy = Iv[0];
  }

  /*  Start here when not initializing  */
  k = Seed / RND_IQ;

  /*  Compute Seed = (IA * Seed) % IM without overflow by Schrage's method  */
  Seed = RND_IA * (Seed - k * RND_IQ) - RND_IR * k;
  if(Seed < 0) Seed += RND_IM;

  /*  Will be in the range 0 .. NTAB-1  */
  j = Iy / RND_NDIV;

  /*  Output previously stored value and refill the shuffle table  */
  Iy = Iv[j];
  Iv[j] = Seed;

  return Iy;
}
#undef RND_IA
#undef RND_IM
#undef RND_AM
#undef RND_IQ
#undef RND_IR
#undef RND_NTAB
#undef RND_NDIV
