/*JS*********************************************************************
*
*    Program : SAXPARSER
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : SAX XML parser for C
*
*              Copyright (C) 2005 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jssubs.h>

#define DEBUG

/*********     DEFINES                                          *********/
#ifndef SAXTYPE_FSKIPSPACES
/*  START-DEFINITIONS  */
/*  Trim heading and trailing spaces from character sequences  */
#define SAXTYPE_FTRIMSPACES        1
/*  Do not return character sequences consisting solely of spaces  */
#define SAXTYPE_FSKIPSPACES        2
/*  Keep escape sequences like &lt; in attribute values and characters  */
#define SAXTYPE_FKEEPESCCHAR       4
/*  Convert <element/> into <element></element>  */
#define SAXTYPE_FNOSINGLEEL        8
/*  Use pLineNr[1] to update current column  */
#define SAXTYPE_FUPDCOLUMN        16

typedef enum {
  SAXTYPE_EOF,

  SAXTYPE_XMLVERSION,    /*  <?xml version="1.0" ... ?>  */
  SAXTYPE_PROCINSTR,     /*  <?process ... >  */
  SAXTYPE_DOCTYPE,       /*  <!DOCTYPE ... >  */
  SAXTYPE_CHARACTER,
  SAXTYPE_COMMENT,       /*  <!-- ... -->  */

  SAXTYPE_STARTELEMENT,  /*  <element ...>  */
  SAXTYPE_STOPELEMENT,   /*  </element>  */
  SAXTYPE_ELEMENT        /*  <element .../>  */
} saxtype_e;

typedef struct {
  char *SAPA_Buffer;
  const char **SAPA_Attributes;
  unsigned int SAPA_NAttributes;

  unsigned int SAPA_Mode;
  unsigned int SAPA_State;
  unsigned int SAPA_MAttributes, SAPA_MBuffer;

} saxparse_t;

#define saxparsegetname(sapa)    ((sapa)->SAPA_Buffer)
#define saxparsegetattrs(sapa)   ((sapa)->SAPA_Attributes)
#define saxparsegetnattrs(sapa)  ((sapa)->SAPA_NAttributes)
/*  END-DEFINITIONS  */
#endif

#define IS_NAMECHAR(c)    (isalnum(c) || c == '_' || c == '-')

#define CHARBUFF_INC     200
#define ATTRBUFF_INC      20

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype saxparse_t    *saxparsecreate(int mode);
Prototype void           saxparsefree(saxparse_t *sapa);

Prototype int            saxparse(saxparse_t *sapa, FILE *fp,
				  unsigned int *pLineNr);
static int               trimCharacters(char *string, unsigned int len, int mode);
static int               convertEscapes(char **strings, int n, int mode);

/*********     GLOBAL VARIABLES                                 *********/
static const struct {
  const char *Name;
  char Character;

} Escapes[] = {
  { "lt",    '<'  },
  { "gt",    '>'  },
  { "amp",   '&'  },
  { "quot",  '"'  },
  { "apos",  '\'' },

  { NULL,    '\0' }
};

/* ERROR-DEFINITIONS from saxparse label _ERR_SAXPARSE
   Invalid tag name
   Invalid character in tag
   Missing closing '>' in complete tag
   Invalid attribute/value separator
   Missing opening double quotes
   Invalid comment introducer
   Invalid character
   Unterminated escape sequence
   Invalid escape sequence
   Premature EOF
   Syntax error
   Multibyte representation too long
   Illegal '<![CDATA[' section
*/

/*JS*********************************************************************
*
*************************************************************************/

saxparse_t *saxparsecreate(int mode)

/************************************************************************/
{
  saxparse_t *sapa;

  if((sapa = (saxparse_t *)malloc(sizeof(*sapa))) == NULL)
    return NULL;

  sapa->SAPA_Buffer = NULL;
  sapa->SAPA_Attributes = NULL;
  sapa->SAPA_NAttributes = 0;

  sapa->SAPA_Mode = mode;
  sapa->SAPA_State = 0;
  sapa->SAPA_MAttributes = 0;
  sapa->SAPA_MBuffer = 0;

  return sapa;
}

/*JS*********************************************************************
*
*************************************************************************/

void saxparsefree(saxparse_t *sapa)

/************************************************************************/
{
  free((void *)sapa->SAPA_Attributes);
  free(sapa->SAPA_Buffer);
  free(sapa);
}

/*JS*********************************************************************
*
*************************************************************************/

int saxparse(saxparse_t *sapa, FILE *fp, unsigned int *pLineNr)

/************************************************************************/
{
  char *buffer;
  unsigned int iBuffer, state, ret;
  int c;

  ret = SAXTYPE_EOF;
  buffer = sapa->SAPA_Buffer;
  iBuffer = 0;

  sapa->SAPA_NAttributes = 0;

  state = sapa->SAPA_State;
  while((c = getc(fp)) != EOF) {
    const char *attr;
    unsigned int saveAttr;

    if(pLineNr != NULL) {
      if(sapa->SAPA_Mode & SAXTYPE_FUPDCOLUMN) {
	if(c == '\n') {
	  /*  next line  */
	  pLineNr[0]++;
	  pLineNr[1] = 1;

	} else
	  /*  next column  */
	  pLineNr[1]++;

      } else if(c == '\n')
	/*  next line  */
	pLineNr[0]++;
    }

/*  Make this a proper statement so it can be ended by a ';'  */
#define SKIP_SPACES(c)  if(isspace(c)) continue

#ifdef DEBUG
    if(sapa->SAPA_Mode & 1024)
      /*  For debugging  */
      printf("STATE %3d IBUFF %2d IATTR %2d CHAR '%c'\n",
 	     state, iBuffer, sapa->SAPA_NAttributes, c);
#endif

    saveAttr = FALSE;
    switch(state) {
    case 0: /* *** looking for tags *** */
      if(c != '<')
	/*  continue collecting characters */
	break;

      state = 5;

      /*  Any characters seen so far?  */
      if(iBuffer > 0) {
	if((c = trimCharacters(buffer, iBuffer, sapa->SAPA_Mode)) == -1)
	  EGOTO(error);

	if(c) {
	  ret = SAXTYPE_CHARACTER;
	  goto ende;
	}
      }

      iBuffer = 0;
      c = -1;
      break;

    case 5:  /* *** scan tag name, look for first char  *** */
      if(c == '!') {
	/*  comment or CDATA  */
	state = 50;
	c = -1;
	break;
      }

      if(c == '/') {
	state = 110;
	c = -1;
	break;
      }

      if(c == '?') {
	state = 210;
	c = -1;
	break;
      }

      state = 10;
      /*FALLTHROUGH*/
    case  10: /* *** scan start or complete tag, look for first char *** */
    case 110: /* *** scan end tag, look for first char *** */
    case 210: /* *** scan processing instructions, look for first char *** */
      SKIP_SPACES(c);

      if(!isalpha(c)) {
	JSSETERROR(_ERR_SAXPARSE, 0);
	EGOTO(error);
      }

      state += 5;
      break;

    case  15: /* *** scan start or complete tag name, second plus char  *** */
    case 115: /* *** scan end tag, second plus char  *** */
    case 215: /* *** scan processing instructions, second plus char *** */
      if(c == '>') {
        /*  Element without attributes  */
	buffer[iBuffer] = '\0';

	ret = state == 15 ? SAXTYPE_STARTELEMENT :
	  state == 115 ? SAXTYPE_STOPELEMENT : SAXTYPE_PROCINSTR;
	state = 0;

	goto ende;
      }

      if(c == '/') {
	if(state != 15) {
	  JSSETERROR(_ERR_SAXPARSE, 1);
	  EGOTO(error);
	}

	++state;
	/*  this element has no attributes, so no need to bother in
	 *   calling convertEscapes
	 */

	if(sapa->SAPA_Mode & SAXTYPE_FNOSINGLEEL) {
	  buffer[iBuffer] = '\0';
	  ret = SAXTYPE_STARTELEMENT;

	  /*  if we get back and reach state 16, we still have the
	   *  element name in the buffer and can reuse it.
	   */
	  goto ende;
	}

	c = 0; /*  terminate tag name */
      } else if(isspace(c)) {
	state += 5;
	c = 0; /*  terminate tag name */
      }

      break;

    case 16: /* *** scan complete tag, looking for closing > *** */
      if(c != '>') {
	JSSETERROR(_ERR_SAXPARSE, 2);
	EGOTO(error);
      }

      ret = (sapa->SAPA_Mode & SAXTYPE_FNOSINGLEEL) ? SAXTYPE_STOPELEMENT :
        SAXTYPE_ELEMENT;
      state = 0;

      goto ende;

    case 20: /* *** scan start or complete tag, look for attributes  *** */
      if(c == '/') {
	state = 16;

        if(!(sapa->SAPA_Mode & SAXTYPE_FKEEPESCCHAR) &&
           convertEscapes((char **)sapa->SAPA_Attributes + 1,
			  sapa->SAPA_NAttributes, 1) == -1)
	  EGOTO(error);

        if(sapa->SAPA_Mode & SAXTYPE_FNOSINGLEEL) {
	  buffer[iBuffer] = '\0';
          ret = SAXTYPE_STARTELEMENT;

	  /*  if we get back and reach state 16, we still have the
	   *  element name in the buffer and can reuse it.
	   */
          goto ende;
        }

	c = -1; /*  whatever came before is already terminated  */
	break;
      }

    case 120: case 220: /* *** scan for attributes  *** */
      SKIP_SPACES(c);

      if(c == '>') {
	if(!(sapa->SAPA_Mode & SAXTYPE_FKEEPESCCHAR) &&
	   convertEscapes((char **)sapa->SAPA_Attributes + 1,
			  sapa->SAPA_NAttributes, 1) == -1)
	  EGOTO(error);

	ret = state == 20 ? SAXTYPE_STARTELEMENT :
	  state == 120 ? SAXTYPE_STOPELEMENT : SAXTYPE_PROCINSTR;

	state = 0;

	goto ende;
      }

      if(isalpha(c)) {
	state += 5;
	attr = buffer + iBuffer;
	saveAttr = TRUE;

      } else if(state == 220 && c == '?') {
        state = 222;
        c = -1;

      } else
	EGOTO(syntaxError);

      break;

    case 222: /* *** scan "<?" tag, having seen closing '?'  *** */
      if(c != '>') EGOTO(syntaxError);

      ret = SAXTYPE_XMLVERSION;
      state = 0;
      goto ende;

    case 25: case 125: case 225: /* *** scan attribute name  *** */
      if(IS_NAMECHAR(c) || c == ':')
	break;

      state += (c == '=') ? 2 : 1;

      c = 0; /*  terminate attribute name  */
      break;

    case 26: case 126:
    case 226: /* *** look for '=' between attribute and content *** */
      SKIP_SPACES(c);

      if(c == '=') {
	c = -1;
	++state;
#if 0 /*  FIXME: Is this XML?  */
      } else if(isalpha(c)) {
	attr = NULL; /*  empty content  */
	saveAttr = TRUE;
	--state;
#endif
      } else {
	JSSETERROR(_ERR_SAXPARSE, 3);
	EGOTO(error);
      }

      break;

    case 27: case 127: case 227: /* *** look for attribute value  *** */
      SKIP_SPACES(c);
      if(c == '"') {
	++state;
      } else if(c == '\'') {
	state += 2;
      } else {
	JSSETERROR(_ERR_SAXPARSE, 4);
	EGOTO(error);
      }

      attr = buffer + iBuffer;
      saveAttr = TRUE;

      c = -1;
      break;

    case 28: case 128:
    case 228: /* *** scan attribute value in double quotes *** */
      if(c == '"') {
	c = 0; /*  terminate attribute value  */
	state -= 8;
	break;
      }
      break;

    case 29: case 129:
    case 229: /* *** scan attribute value in single quotes *** */
      if(c == '\'') {
	c = 0; /*  terminate attribute value  */
	state -= 9;
	break;
      }
      break;

    /* ************************************************************** */

    case 50: /* *** scan comments or <![CDATA[, having seen <! so far  *** */
      if(c == '[') {
	state = 70;
	c = -1;
	break;
      }

      if(isalpha(c)) {
	state = 60;
	break;
      }
    case 51: /* *** scan comments, having seen <!- so far  *** */
      if(c != '-') {
	JSSETERROR(_ERR_SAXPARSE, 5);
	EGOTO(error);
      }
      c = -1;
      ++state;

      break;

    case 52: /* *** scanning comment, looking for end  *** */
    case 53: /* *** scanning comment, having seen - *** */
      if(c == '-')
	++state;
      else
	state = 52;

      break;

    case 54: /* *** scanning comment, having seen -- *** */
      if(c == '>') {
	/*  Delete the two hyphens  */
	buffer[iBuffer - 2] = '\0';

	ret = SAXTYPE_COMMENT;
	state = 0;

	goto ende;
      }

      state = 53;
      break;

    /* ************************************************************** */

    case 60: /* ***  scanning DOCTYPE etc.  *** */
      if(c == '[') {
	state = 65;
	break;
      }

      if(c == '>') {
	buffer[iBuffer - 1] = '\0';

	ret = SAXTYPE_DOCTYPE;
	state = 0;

	goto ende;
      }

      break;

    case 65:
      if(c == ']')
	state = 60;

      break;

    /* ************************************************************** */
    case 70: /* ***  scanning CDATA section, looking for 'C'  *** */
    case 71: /* 'D' */
    case 72: /* 'A' */
    case 73: /* 'T' */
    case 74: /* 'A' */
    case 75: /* '[' */
      if(c != "CDATA["[state - 70]) {
	JSSETERROR(_ERR_SAXPARSE, 12);
	EGOTO(error);
      }

      ++state;
      c = -1;
      break;

    case 76: /* ***  within <![CDATA[ ... ]]> section  *** */
    case 77: /*  having seen first ']'  */
      if(c == ']')
	++state; /*  still add the ']' to the buffer  */
      else
	state = 76;

      break;

    case 78:
      if(c == '>') {
	buffer[iBuffer - 2] = '\0';

	ret = SAXTYPE_CHARACTER;
	state = 0;

	goto ende;
      }

      state = 76; /*  that was something else  */
      break;

    /* ************************************************************** */

#ifdef DEBUG
    default:
      fprintf(stderr, "ERROR: Invalid state %d!\n", state);
      EGOTO(syntaxError);
#endif
    }

    /*  Must come first to ensure that the pointer adaption below
     *   is applied to attr as well!
     */
    if(saveAttr) {
      if(sapa->SAPA_NAttributes >= sapa->SAPA_MAttributes) {
	REALLOC_RESIZE(sapa->SAPA_MAttributes, ATTRBUFF_INC);
	if((sapa->SAPA_Attributes = (const char **)
	    realloc((void *)sapa->SAPA_Attributes, sapa->SAPA_MAttributes *
		    sizeof(*(sapa->SAPA_Attributes)))) == NULL)
	  EGOTO(error);
      }

      sapa->SAPA_Attributes[sapa->SAPA_NAttributes++] = attr;
    }

    /*  Save a character? Is also done for c == '\0'  */
    if(c != -1) {
      if((iBuffer + 1) >= sapa->SAPA_MBuffer) {
	unsigned int i;

	REALLOC_RESIZE(sapa->SAPA_MBuffer, CHARBUFF_INC);
	if((buffer = (char *)realloc(buffer, sapa->SAPA_MBuffer)) == NULL)
	  EGOTO(error);

	/*  Adapt pointers  */
	for(i = 0 ; i < sapa->SAPA_NAttributes ; ++i)
	  sapa->SAPA_Attributes[i] = buffer +
	    (sapa->SAPA_Attributes[i] - sapa->SAPA_Buffer);

	sapa->SAPA_Buffer = buffer;
      }

      buffer[iBuffer++] = c;
    }
  }

  /*  Check that we are in the correct state at EOF  */
  if(state != 0) {
    JSSETERROR(_ERR_SAXPARSE, 9);
    EGOTO(error);
  }

  /*  Properly return a character sequence after the last token.
   *   This might not be allowed, but we'll return at at least and
   *   leave it to our caller to complain about it.
   */
  if(iBuffer > 0) {
    if((c = trimCharacters(buffer, iBuffer, sapa->SAPA_Mode)) == -1)
      EGOTO(error);

    if(c)
      ret = SAXTYPE_CHARACTER;
  }

ende:
  sapa->SAPA_State = state;

#undef SKIP_SPACES

  return ret;
syntaxError:
  JSSETERROR(_ERR_SAXPARSE, 10);
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int trimCharacters(char *string, unsigned int len, int mode)

/************************************************************************/
{
  if(mode & SAXTYPE_FTRIMSPACES) {
    unsigned int i;

    for(i = 0 ; i < len ; ++i)
      if(!isspace(string[i]))
	break;

    if(i > 0) {
      memmove(string, string + i, len - i);
      len -= i;
    }

    while(len > 0 && isspace(string[len - 1]))
      --len;
  }
  string[len] = '\0';

  if(mode & SAXTYPE_FSKIPSPACES) {
    if(mode & SAXTYPE_FTRIMSPACES) {
      if(len == 0)
	return 0;
    } else {
      char *s;

      for(s = string ; *s ; ++s)
	if(!isspace(*s)) break;

      if(*s == '\0')
	return 0;
    }
  }

  if(!(mode & SAXTYPE_FKEEPESCCHAR) &&
     convertEscapes(&string, 1, 0) == -1)
    return -1;

  return 1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int convertEscapes(char **strings, int n, int mode)

/************************************************************************/
{
  for( ; n > 0 ; strings += 2, n -= 2) {
    char *src, *dest;
    int c;

    for(src = dest = *strings ; (c = *src) ; ++src) {
      if(c != '&') {
	/*  Check for illegal characters  */
	if(c == '<') {
	  JSSETERROR(_ERR_SAXPARSE, 6);
	  EGOTO(error0);
	}

	*dest++ = c;
      } else if(*++src == '#') {
	wchar_t wc;

	++src;
	wc = *src == 'x' ? strtol(src + 1, &src, 16) : strtol(src, &src, 10);

	if(*src != ';')
	  EGOTO(error);

	if((c = wctomb(dest, wc)) == -1)
	  /*  Seems as if this character cannot be represented  */
	  EGOTO(error0);

	dest += c;

	/*  Check if the UTF-8 or whatever representation overwrote the
	 *   string. Since "&#1;" is at least four bytes long and
         *   more would be needed to represent any reasonable character,
	 *   this should work almost everywhere (according to Lebesgue).
	 *   UTF-8 is guaranteed to need not more than 6 bytes for all
	 *   possible 32 bit values and with current UNICODE characters
	 *   only up to four bytes.
	 */
	if(dest > (src + 1)) {
	  JSSETERROR(_ERR_SAXPARSE, 11);
	  EGOTO(error0);
	}
      } else {
	char *end;
	unsigned int i;

	if((end = strchr(src, ';')) == NULL)
	  EGOTO(error);

	*end = '\0';
	for(i = 0 ; Escapes[i].Name != NULL ; ++i)
	  if(strcmp(src, Escapes[i].Name) == 0)
	    break;

	if(Escapes[i].Name == NULL) {
	  JSSETERROR(_ERR_SAXPARSE, 8);
	  EGOTO(error0);
	}

	*dest++ = Escapes[i].Character;
	src = end;
      }
    }
    *dest = '\0';
  }

  return 0;
error:
  JSSETERROR(_ERR_SAXPARSE, 7);
error0:
  return -1;
}
