/*JS*********************************************************************
*
*    Program : STRBUFFER
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Collect strings in a buffer that will be reallocated.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <jsconfig.h>  /*  because of stdarg.h  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#ifndef jstrbuffclear
/*  START-DEFINITIONS  */
#include <stdarg.h>  /*  For "va_list" definition */

typedef struct {
  char *SB_Buffer;
  int SB_MBuffer, SB_NBuffer;
} jstrbuff_t;

#define jstrbuffer(sb)         ((sb)->SB_Buffer)
#define jstrbuffmaxlen(sb)     ((sb)->SB_MBuffer)
#define jstrbufflen(sb)        ((sb)->SB_NBuffer)
#define jstrbuffputc(sb, c)    ((sb)->SB_Buffer[((sb)->SB_NBuffer)++] = (c))
#define jstrbuffclose(sb)      ((sb)->SB_Buffer[(sb)->SB_NBuffer] = '\0')

#define jstrbuffinc(sb, len)    ((sb)->SB_NBuffer += (len))
#define jstrbuffset(sb, len)    ((sb)->SB_NBuffer = (len))

#define jstrbuffclear(sb)      ((sb)->SB_Buffer = NULL, \
			       (sb)->SB_MBuffer = (sb)->SB_NBuffer = 0)
#define jstrbuffempty(sb)      (jstrbuffset(sb, 0), (sb)->SB_Buffer ? \
				(sb)->SB_Buffer[0] = '\0' : 0)
#define jstrbufffree(sb)       free(((sb)->SB_Buffer))
/*  END-DEFINITIONS  */
#endif

/*  How many additional space should be allocated in strbuffstore?  */
#define ADDSPACE    4

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype int            jstrbuffstore(jstrbuff_t *sb, const char *s, long len);

Prototype int            jstrbuffprint(jstrbuff_t *sb, const char *s);
Prototype int            jstrbuffprintf(jstrbuff_t *sb, const char *format, ...);
Prototype int            jstrbuffvprintf(jstrbuff_t *sb, const char *format,
					va_list ap1, va_list ap2);

Prototype int            jstrbuffins(jstrbuff_t *sb, const char *s, long off,
				    long len);
Prototype void           jstrbuffdel(jstrbuff_t *sb, long off, long len);

/*JS*********************************************************************
*
*************************************************************************/

int jstrbuffstore(jstrbuff_t *sb, const char *s, long len)

/************************************************************************/
{
  int i;

  i = sb->SB_NBuffer;
  if((i + len + ADDSPACE) >= sb->SB_MBuffer) {
    REALLOC_RESIZE(sb->SB_MBuffer, i + len + ADDSPACE);
    if((sb->SB_Buffer = (char *)realloc(sb->SB_Buffer,
					sb->SB_MBuffer)) == NULL)
      return -1;
  }

  if(s != NULL) {
    memcpy(sb->SB_Buffer + i, s, len);
    sb->SB_NBuffer += len;
  }

  return len;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstrbuffprint(jstrbuff_t *sb, const char *s)

/************************************************************************/
{
  int ret;

  /*  Store only the characters, the trailing '\0' is stored, but
   *   does not go into the length.
   */
  if((ret = jstrbuffstore(sb, s, strlen(s))) != -1)
    sb->SB_Buffer[sb->SB_NBuffer] = '\0';

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstrbuffprintf(jstrbuff_t *sb, const char *format, ...)

/************************************************************************/
{
  va_list ap1, ap2;
  int ret;

  va_start(ap1, format);
  va_start(ap2, format);
  ret = jstrbuffvprintf(sb, format, ap1, ap2);
  va_end(ap1);
  va_end(ap2);

  return ret;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstrbuffvprintf(jstrbuff_t *sb, const char *format, va_list ap1,
		    va_list ap2)

/************************************************************************/
{
  int len;

  /*  Estimate length  */
  if((len = vlprintf(format, ap1)) < 0)
    EGOTO(error);

  len += sb->SB_NBuffer;

  if((len + 1 + ADDSPACE) >= sb->SB_MBuffer) {
    REALLOC_RESIZE(sb->SB_MBuffer, len + 1 + ADDSPACE);
    if((sb->SB_Buffer = (char *)
	realloc(sb->SB_Buffer, sb->SB_MBuffer)) == NULL)
      EGOTO(error);
  }

  if((len = vsprintf(sb->SB_Buffer + sb->SB_NBuffer, format, ap2)) < 0)
    EGOTO(error);

  sb->SB_NBuffer += len;

  return len;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

int jstrbuffins(jstrbuff_t *sb, const char *s, long off, long len)

/************************************************************************/
{
  char *dest;
  long n;

  /*  First create space  */
  if(jstrbuffstore(sb, NULL, len) == -1)
    EGOTO(error);

  n = sb->SB_NBuffer;
  if(off < 0) {
    if((off += n) < 0)
      off = 0;

  } else if(off >= n)
    off = n;

  dest = sb->SB_Buffer + off;

  if(off < n)
    memmove(dest + len, dest, n - off);

  if(s)
    memcpy(dest, s, len);
  sb->SB_NBuffer += len;

  return len;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

void jstrbuffdel(jstrbuff_t *sb, long off, long len)

/************************************************************************/
{
  long n;

  n = sb->SB_NBuffer - off;
  if(len < n) {
    char *s = sb->SB_Buffer + off;

    memmove(s, s + len, n - len + 1);

  } else if(len > n)
    len = n;

  sb->SB_NBuffer -= len;
}
