/*JS*********************************************************************
*
*    Program : VAPRINTF
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Similar to the 'sprintf' family functions, but allocates
*              string automatically with 'malloc' and returns it.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <jsconfig.h> /*  because of stdarg.h  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include <math.h>

#include <limits.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
/*  START-DEFINITIONS */
#include <stdarg.h>  /*  For "va_list" definition */
/*  END-DEFINITIONS */

/*#define XDEBUG*/

/*  Default upper length for integers and pointers. Should be
 *   enough even for 64 bit machines and if pointers are
 *   printed as bits.
 */
#define MAX_INTPOINT   (sizeof(long) * CHAR_BIT)

/*  Default upper length for floating point values (very crude)  */
#define MAX_FLOATVAL   (sizeof(double) * CHAR_BIT + 4)

#define Prototype extern
/*********     PROTOTYPES                                       *********/
Prototype char          *aprintf(const char *format, ...);
Prototype char          *vaprintf(const char *format, va_list ap1,
				  va_list ap2);

Prototype long           vlprintf(const char *format, va_list ap);

/*********     GLOBAL VARIABLES                                 *********/

/*JS*********************************************************************
*   Similar to 'sprintf', but no buffer argument is supplied. Space will
*    be allocated automatically using 'malloc' and returned.
*************************************************************************/

char *aprintf(const char *format, ...)

/************************************************************************/
{
  va_list ap1, ap2;
  char *s;

  va_start(ap1, format);
  va_start(ap2, format);
  s = vaprintf(format, ap1, ap2);
  va_end(ap1);
  va_end(ap2);

  return s;
}

/*JS*********************************************************************
*   Similar to 'vsprintf', but no buffer argument is supplied. Space will
*    be allocated automatically using 'malloc' and returned.
*************************************************************************/

char *vaprintf(const char *format, va_list ap1, va_list ap2)

/************************************************************************/
{
  char *string;
  long len, len2;

  /*  Estimate length  */
  if((len = vlprintf(format, ap1)) < 0)
    EGOTO(error);

  /*  Now allocate buffer and call vsprintf  */
  if((string = (char *)malloc(len + 1)) == NULL)
    EGOTO(error);

  if((len2 = vsprintf(string, format, ap2)) < 0 || len2 > len) {
#ifdef XDEBUG
    if(len2 >= 0) {
      fprintf(stderr, "\
FATAL: Wrong estimation in vaprintf (%d > %d) -- Break.\n", len2, len);
      abort();
    }
#endif
    free(string);
    EGOTO(error);
  }

#if 1
  /*  Minimize memory usage (might be dropped to speed up)  */
  string = (char *)realloc(string, len2 + 1);
#endif

  return string;
error:
  return NULL;
}

/*JS*********************************************************************
*   Estimates the needed length for the printf format with arguments.
*************************************************************************/

long vlprintf(const char *format, va_list ap)

/************************************************************************/
{
  long len;

  /*  Scan format and estimate needed length  */
  for(len = 0 ; *format ; ) {
    const char *string;
    long minField, precField, flag, i;
    union {
      long          IVal;
      unsigned char UVal;
      char         *SVal;
      double        DVal;
      void         *PVal;
      short        *NSVal;
      int          *NIVal;
      long         *NLVal;
    } value;

    /*  Find next command  */
    string = strchr(format, '%');

    if(string == NULL) {
      /* ***  Nothing more to come  *** */
      len += strlen(format);
      break;
    } else if(string != format) {
      len += string - format;
      format = string;
      continue;
    }

    /* ***  Handle '%' formats  *** */
    format++;
    while(strchr("-+ 0#", *format)) format++; /*  Skip flags  */

    /*  Read 'min[.prec]' */
    if(isdigit(*(unsigned char *)format)) {
      /*  Note: strtoul cannot fail since we have already seen a digit  */
      minField = strtoul(format, (char **)&format, 10);
    } else if(*format == '*') {
      minField = va_arg(ap, int);
      format++;
    } else
      minField = -1;

    if(*format != '.') {
      precField = -1;
    } else if(format++, isdigit(*(unsigned char *)format)) {
      /* **  Read prec  ** */
      precField = strtoul(format, (char **)&format, 10);
    } else if(*format == '*') {
      precField = va_arg(ap, int);
      format++;
    } else
      precField = -1;

    /*  Skip size specifier  */
    if(*format == 'h') {
      flag = 0;
      format++;
    } else if(*format == 'l' || *format == 'L') {
      flag = 2;
      format++;
    } else
      flag = 1;

    /*  Read conversion specifier and estimate length  */
    switch(*format) {
    case 'd': case 'i': case 'o': /* **  integers  ** */
    case 'x': case 'X': case 'u':
      value.IVal = (flag == 0) ? va_arg(ap, int) :
	((flag == 1) ? va_arg(ap, int) : va_arg(ap, long));
      i = MAX_INTPOINT;
      break;
    case 'c': /* **  character  ** */
      value.UVal = (unsigned char)va_arg(ap, int);
      i = 1;
      break;
    case 's': /* **  string  ** */
      value.SVal = va_arg(ap, char *);

      /*  We do *not* use strlen in case precision is given, since
       *   string might not be properly terminated with a '\0'.
       */
      i = (precField >= 0) ? precField : (int)strlen(value.SVal);
      break;
    case 'f':
      value.DVal = va_arg(ap, double);

      /*  That's tricky and might result in *very* long printouts  */
      if(value.DVal != 0.0) {
#if 1
	int bdigit;

	/*  This method should be faster: get the number of binary digits
	 *   and multiply them by 4 (ln(10)/ln(2)=3.32).
	 */
	(void)frexp(fabs(value.DVal), &bdigit);
	if(bdigit < 0) bdigit = -bdigit; /*  something like 0.000000000000123 */
	i = (bdigit << 2) + (precField < 0 ? 10 : (precField + 4));
#else
	/*  Estimate the length in digits  */
        i = (int)log10(fabs(value.DVal));
	if(i < 0) i = -i; /*  something like  "0.000000000000123"  */
        i += (precField < 0 ? 10 : (precField + 4));
#endif
      } else
        i = MAX_FLOATVAL;
      break;
    case 'e': case 'E': case 'g': case 'G': /* **  double  ** */
      value.DVal = va_arg(ap, double);
      i = MAX_FLOATVAL;
      if(precField >= 0) precField += (MAX_FLOATVAL - 6);

      break;
    case 'p': /* **  pointer  ** */
      value.PVal = va_arg(ap, void *);
      i = MAX_INTPOINT;
      break;
    case 'n': /* **  store number of chars  ** */
      if(flag == 0)
	value.NSVal = va_arg(ap, short *);
      else if(flag == 1)
	value.NIVal = va_arg(ap, int *);
      else
	value.NLVal = va_arg(ap, long *);
      i = 0; /*  ignore  */
      break;
    case '%':
      i = 1;
      break;
    default:
      /*  Undefined, copy symbol  */
#ifdef XDEBUG
      fprintf(stderr, "WARNING: Unknown format specifier '%c'!\n", *format);
#endif
      goto error;
    }
    if(*format) format++;

    if(i < minField) i = minField;
    if(i < precField) i = precField;

    len += i;
  }

  return(len);
error:
  return(-1);
}

#ifdef TEST
int main(int argc, char *argv[])
{
  char *string;
  int i1, i2;
  double d1, d2;

  i1 = 1342423424;
  i2 = 17;
  d1 = 1.9876532424242e-50;
  d2 = 1.9876532424242E60;

  string = aprintf("hello %d %020i--\n\
>>%50o\n\
--%#x%*.*d--\n\
%s %f\n\
!%10.10g%%\n",
		    i1, i2, i1, i2, 30, 5, i2, "WATISDAT",
		    d2, d1);

  if(string == NULL) {
    perror("testvaprintf");
    exit(30);
  }

  fputs(string, stdout);

  return(0);
}
#endif
