/*JS*********************************************************************
*
*    Program : WILDCMP
*    Language: ANSI-C
*    Author  : Joerg Schoen (algorithm taken from old code of Matt Dillon)
*    Purpose : Compares wild cards with '*', '?' characters and '[]'
*              classes. A prefix of '~' negates the wildcard.
*
*              Copyright (C) 1994-2002 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES 					*********/
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <jsconfig.h>

#include <jssubs.h>

/*********     DEFINES						*********/
/*  START-DEFINITIONS  */
#define WILD_STRICT	  (1<<0)
#define WILD_STARTNAME	  (1<<1)
#define WILD_IGNORECASE   (1<<2)
#define WILD_PARTIALLY	  (1<<3)
/*  END-DEFINITIONS  */

#define MAXSAVE   50 /*  Nobody makes complicated wild cards (hopefully)  */

#define NEGATE    '~'

/* ERROR-DEFINITIONS from wildcmp label _ERR_WILDCMP
   '*' nesting too deep
   Wildcard ended prematurely
*/

#define Prototype extern
/*********     PROTOTYPES					*********/
Prototype int		 wildcmp(const char *name, const char *wild, int mode);
Prototype int		 jiswildcard(const char *name);

/*JS*********************************************************************
*   Compares name with wildcard. Returns boolean value. If the WILD_STRICT
*    bit in mode is set, a wildcard '*' or '?' can not substitute a
*    directory delimiting value '/' so that for example '*' '/' '*' stands
*    for any file in any sub directory of current directory (not below) and
*    '*' stands for files (or dir names) only in current directory.
*    If the WILD_STARTNAME bit is set, than the name is treated as the begin
*    of a real name and TRUE is returned even if name ended and wildcard not.
*    If the WILD_PARTIALLY bit is set, TRUE is returned even if the wild
*    card matches only the start of name.
*    A '~' as start character of the wild card negates it.
*************************************************************************/

int wildcmp(const char *name, const char *wild, int mode)

/************************************************************************/
{
  int saveIndex, trueVal;
  struct SavePos {
    const char *SaveWild, *SaveName;
  } Save[MAXSAVE];

  /*   Check if wild card is negated	*/
  if(*wild == NEGATE) {
    wild++;
    trueVal = FALSE;
  } else {
    trueVal = TRUE;
  }

#if 0
  /*  Check if number of directory delimiters match   */
  if(mode & WILD_STRICT) {
    const char *string;

    /*	Count number of '/' in wildcard   */
    saveIndex = 0;
    for(string = wild ; *string ; string++)
      if(*string == CONFIG_DIR_DELIMITER) saveIndex++;

    /*	Count number of '/' in name   */
    for(string = name ; *string ; string++)
      if(*string == CONFIG_DIR_DELIMITER) saveIndex--;

    /*	Check if numbers are the same	*/
    if((mode & WILD_STARTNAME) ? (saveIndex < 0) : (saveIndex != 0))
      goto failed;
  }
#endif

  saveIndex = 0;

  while(*name || *wild) {
    /*	Name already ended in case of WILD_STARTNAME ?	*/
    if((mode & WILD_STARTNAME) && trueVal && *name == '\0') break;

    if((mode & WILD_PARTIALLY) && trueVal && *wild == '\0') break;

    switch(*wild) {
    case '\\':
      /*  Next character has no special meaning  */
      if(wild[1] == '\0') goto endofwild;
      wild++;
    default:
      /*  Two character compare  */
      if((mode & WILD_IGNORECASE) ?
	 tolower(*(unsigned char *)name) != tolower(*(unsigned char *)wild) :
	 *name != *wild)
	goto goback;
      break;  /*   Both characters are the same    */
    case '?':
      if(*name == '\0' ||
	 ((mode & WILD_STRICT) && *name == CONFIG_DIR_DELIMITER))
	goto goback;

      break;  /*   Just don't look at next character    */
    case '[':
      ++wild;
      {
	int reversed, matched;

	/*  Check if reversed pattern  */
	if(*wild == '^') {
	  reversed = TRUE;
	  wild++;
	} else
	  reversed = FALSE;

	for(matched = FALSE ; *wild && *wild != ']' ; ++wild) {
	  if(*wild == '\\') {
	    /*  Check if wild card ends after '\'  */
	    if(*++wild == '\0') goto endofwild;

	    /*	Special character like ']' or '-' in class matched?  */
	    if(*wild == *name) matched = TRUE;
	  } else if(wild[1] == '-') {
	    /*  Check if wild card ends after '-'  */
	    if(wild[2] == '\0') goto endofwild;

	    /*	Match class?  */
	    if(*wild <= *name && *name <= wild[2]) matched = TRUE;

	    /*	Two more characters to skip  */
	    wild += 2;
	  } else if(*wild == *name)
	    matched = TRUE;
	}

	/*  Any class matched resp. no class matched?  */
	if(matched == reversed) goto goback;
      }
      break;
    case '*':
      if(saveIndex == sizeof(Save)/sizeof(Save[0])) {
	JSSETERROR(_ERR_WILDCMP, 0);
	goto error;
      }

      /*  Save position after '*' on stack and pass '*'  */
      Save[saveIndex].SaveWild = ++wild;

      /*  Save position in name string on stack 	 */
      Save[saveIndex].SaveName = name;
      saveIndex++;
      continue;
goback:
      /*  Go back to last '*' wildcard where name was not used up    */
      do {
	saveIndex--;
      } while(saveIndex >= 0 && (*Save[saveIndex].SaveName == '\0' ||
				 ((mode & WILD_STRICT) &&
				  *Save[saveIndex].SaveName ==
				  CONFIG_DIR_DELIMITER)));

      /*   Check if end reached   */
      if(saveIndex < 0) goto failed;

      /*   Restart at last wildcard    */
      wild = Save[saveIndex].SaveWild;

      /*   At least one character must have been used up from this '*',  */
      /*    so increment saved position (is still saved!).               */
      name = ++Save[saveIndex].SaveName;
      saveIndex++;
      continue;
    }

    /*	Go to next characters if not already last reached    */
    if(*name) ++name;
    if(*wild) ++wild;
  }

  return(trueVal);
failed:
  return(!trueVal);

endofwild:
  JSSETERROR(_ERR_WILDCMP, 1);
error:
  return(FALSE);
}

/*JS*********************************************************************
*
*************************************************************************/

int jiswildcard(const char *name)

/************************************************************************/
{
  /*  Negate prefix?  */
  if(*name == NEGATE)
    return TRUE;

  while(*name) {
    /*  Escaped character?  */
    if(*name == '\\' && name[1])
      name++; /*  Skip escaped character  */
    else if(strchr("*?[", *name))
      return TRUE;

    name++;
  }

  return FALSE;
}

#ifdef TEST

#define LENGTH  500

int main(int argc, char *argv[])
{
  char wild[LENGTH], text[LENGTH];

  printf("Enter wildcard, than strings to test. A blank line gets prompts\n\
 for a new pattern and a blank pattern exits the program.\n");

  for(;;) {
    printf("Enter wildcard: ");

    if(gets(wild) == NULL || *wild == '\0') break;

    for(;;) {
      printf("Enter text:  ");

      if(gets(text) == NULL) goto ende;

      if(text[0] == '\0') break;

      printf("<%s> matches %s <%s>.\n", text, wildcmp(text, wild, 0) ? "" : "not",
	wild);
    }
  }

ende:
  return(0);
}
#endif
