/*JS*********************************************************************
 *
 *    Program : JSTOKMAT
 *    Language: ANSI-C
 *    Author  : Joerg Schoen
 *    Purpose : C-token matcher
 *
 *              Copyright (C) 2005-2007 Joerg Schoen
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 *
 *************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include <jssubs.h>

/*********     DEFINES                                          *********/
#define DEBUG

#define OPTSTRING    "hs:ignmvC"

#ifdef DEBUG
#define OPTSTRING_DBG  "D:"
#else
#define OPTSTRING_DBG  ""
#endif

/*********     PROTOTYPES                                       *********/
static void usage(int level);

/*JS*********************************************************************
 *
 *************************************************************************/

static void usage(int level)

/************************************************************************/
{
    printf("\
Usage: jstokmat [<opts>] <xml> [<in> [<out>]]      (C) Joerg Schoen 2005-2007\n\
  Do token based pattern matching and replacement controlled by <xmlfile>.\n\
  Read from <input> (default to standard in) and write to <output> (defaults\n\
  to standard out).\n");

    if(level > 0) {
	printf("\
  The structure of the XML file is\n\
    <?xml version=\"1.0\"?>\n\
    <commands>\n\
      <matcher>\n\
        <pattern>PATTERN</pattern>\n\
        <substitute>SUBSTITUTE</substitute>\n\
      </matcher>\n\
      ...\n\
    </commands>\n\
  with one or more <matcher> tags. Matchers can be named like\n\
    <matcher name=\"myrule\">...</matcher>\n\
  and later be selected using command line option '-s' (see below).\n");

	printf("\
  The following tags are supported in PATTERN:\n\
    <identifier name=\"NAME\"/> OR <number name=\"NAME\"/> OR\n\
    <string name=\"NAME\"/>\n\
      Matches a single C-like identifier, number, or string.\n\
    <expression name=\"NAME\"/> OR <statement name=\"NAME\"/>\n\
      Matches a C-like expression or statement. A C-like expression is\n\
      terminated by a ',' or ';' outside of all brackets. A statement\n\
      is only terminated by ';'.\n\
\n\
      The attribute 'name' in these tags is optional. Using it, one can\n\
      refer to this part within the same PATTERN or reference it in the\n\
      substitute.\n\
    <expression name=\"NAME\">PATTERN</expression> OR\n\
    <statement name=\"NAME\">PATTERN</statement>\n\
      Wraps a part of a pattern to be referred to somewhere else or in the\n\
      substitute.\n\
    <choice>\n\
      <branch>PATTERN1</branch>\n\
      <branch>PATTERN2</branch>\n\
      ...\n\
    </choice>\n\
      Choose between alternative patterns.\n\
    <repeat>PATTERN1</repeat>\n\
      Repeat a pattern zero or more times. The repeat tag can have optional\n\
      attributes 'min' or 'max' or 'count' to specify the lower/upper limit or\n\
      the number of occurence of the pattern.\n\
    <optional>PATTERN</repeat>\n\
      Identical to <repeat max=\"1\">PATTERN</repeat>\n\
  The following tags are supported in SUBSTITUTE:\n\
    Besides clear text, any named tag from the PATTERN can be used in\n\
    SUBSTITUTE.\n\
    Additionally, the following tags are valid:\n\
    <line/>\n\
      Expands to the line number where the pattern was matched.\n\
    <counter name=\"NAME\"/>\n\
      Expands to a number that is incremented after each usage.\n");
    }

    printf("  Options:\n");
    if(level > 0)
	printf("    -h           Show this help information.\n");
    else
	printf("    -h           Show more detailed help information.\n");

    printf("\
    -i           Change files in place, keeping the old one under <name>~.\n\
                 In this mode, all arguments after <xmlfile> are treated as\n\
                 files to be changed.\n\
    -g           Only print matched and converted patterns to standard output.\n\
                 All file arguments are treated as input files.\n\
    -n           Same as '-g', but precede results by file and line number.\n\
    -m           Match longest possible token sequence.\n\
    -v           Show when files are processed.\n\
    -C           Preserve comments in the output. They are printed after the\n\
                 substituted pattern. If the option is given twice, they are\n\
                 printed before the substituted pattern.\n\
    -s<wildcard> Only process matchers from the XML file whose name matches\n\
                 the given wild card in shell style. This option can be used\n\
                 multiple times.\n\
                 If this options is used, unnamed patterns are not processed.\n");
}

/*JS*********************************************************************
 *
 *************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
    tokpattmatch_t *topa;
    const char *string;
    const char **matchers;
    int nMatchers, mMatchers;
    extfile_t *efp;
    FILE *fp;
    int c, i, mode, tmode;

    mode = 0;
    tmode = 0;
    matchers = NULL;
    nMatchers = mMatchers = 0;
    while((c = jsGetOpt(&argc, argv, OPTSTRING OPTSTRING_DBG,
			(const char **)&string)) != EOF) {
	switch(c) {
	default:
	case 'h':
	    usage(1);
	    goto help;

	case 's':
	    if(nMatchers >= mMatchers) {
		mMatchers += 10;
		if((matchers = (const char **)realloc((void *)matchers, mMatchers *
						      sizeof(*matchers))) == NULL)
		    EGOTO(error);
	    }
	    matchers[nMatchers++] = string;
	    break;

	case 'v':
	    mode |= 1;
	    break;

	case 'i':
	    mode |= 2;
	    break;

	case 'g':
	    mode |= 256;
	    break;

	case 'n':
	    mode |= 256 | 512;
	    break;

	case 'm':
	    tmode |= 16;
	    break;

	case 'C':
	    if(mode & 4)
		mode = (mode & ~4) | 8;
	    else
		mode |= 4;
	    break;

#ifdef DEBUG
	case 'D':
	    parsedbgopts(NULL, string, &Debug, 0);
	    break;
#endif
	}
    }

    if((mode & (256|2)) ? argc < 3 : (argc < 2 || argc > 4)) {
	usage(0);
	goto help;
    }

    if((topa = topacreate(argv[1])) == NULL)
	EGOTO(error);

    if(matchers != NULL) {
	for(i = 0 ; i < topa->TOPA_NDefinitions ; ++i) {
	    const char *matName;

	    if((matName = topa->TOPA_Definitions[i].TOPD_Name) == NULL) {
		/*  If patterns are given, do not process unnamed matchers  */
		topa->TOPA_Definitions[i].TOPD_StartState = NULL;
	    } else {
		int j;

		for(j = 0 ; j < nMatchers ; ++j)
		    if(wildcmp(matName, matchers[j], 0))
			break;

		if(j == nMatchers)
		    topa->TOPA_Definitions[i].TOPD_StartState = NULL;
	    }
	}
    }

    /* ************************** */
    tmode |= (mode & 512) ? 8|4 : (mode & 256) ? 4 : (mode & 8) ? 2 : (mode & 4) ? 1 : 0;
    if(mode & (256|2)) {
	if(argc < 3) {
	    usage(0);
	    goto help;
	}

	for(i = 2 ; i < argc ; ++i) {
	    char tmpname[JS_FILENAME_MAX];

	    string = argv[i];

	    if(!(mode & 256))
		sprintf(tmpname, "%s~1", string);

	    if((efp = efopen(string, "r")) == NULL)
		EGOTO(error);

	    if(mode & 256)
		fp = stdout;
	    else if((fp = fopen(tmpname, "w")) == NULL)
		EGOTO(error);

	    if((c = topaprocess(topa, efp, string, fp, NULL, tmode)) == -1)
		EGOTO(error);

	    if(fp != stdout && fclose(fp) == EOF)
		EGOTO(error);

	    if(efclose(efp) == -1)
		EGOTO(error);

	    if(mode & 256) {
		if(mode & 1)
		    printf("Got %d matches in file '%s'.\n", c, string);

		continue;
	    }

	    if(c) {
		char bckname[JS_FILENAME_MAX];

		if(mode & 1)
		    printf("File %s got %d changes.\n", string, c);

		sprintf(bckname, "%s~", string);

		if(rename(string, bckname) == -1 ||
		   rename(tmpname, string) == -1)
		    EGOTO(error);
	    } else {
		if(mode & 1)
		    printf("File %s did not change.\n", string);

		if(remove(tmpname) == -1)
		    EGOTO(error);
	    }
	}
    } else {
	if((efp = efopen(argc > 2 ? argv[2] : NULL, "r")) == NULL)
	    EGOTO(error);

	if(argc <= 3)
	    fp = stdout;
	else if((fp = fopen(argv[3], "w")) == NULL)
	    EGOTO(error);

	if((c = topaprocess(topa, efp, argc > 2 ? argv[2] : "-", fp, NULL, tmode)) == -1)
	    EGOTO(error);

	if(fp != stdout && fclose(fp) == EOF)
	    EGOTO(error);

	if(efclose(efp) == -1)
	    EGOTO(error);
    }

    topadelete(topa);

    return 0;
 help:
    return 5;
 error:
    jsperror("jstokmat");
    return 20;
}
