/*JS*********************************************************************
*
*    Program : PGBACK
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Generate / extract piggyback files.
*
*              Copyright (C) 2003-2012 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <jssubs.h>

#include <jsendian.h>

/*********     DEFINES                                          *********/
#ifdef CONFIG_HAS_BWCOMPRESSION
#define OPTSTRING         "hvl:ncC:B:p:b:s"
#else
#define OPTSTRING         "hvl:np:b:s"
#endif

typedef struct {
  const char *Directory;
  unsigned char *Data;
} auxlistunpack_t;

#define Prototype extern
/*********     PROTOTYPES                                       *********/
static unsigned long     getloadpwd(const char *name);
static int               doPack1(FILE *fp, const char *basedir, const char *fileName,
				 const char *prefix, unsigned long lengths[], int mode);
static int               doPack(FILE *fp, const char *list[], unsigned int n,
				const char *basedir, const char *strip, int mode);
static int               writePause(FILE *fp, int quoteflag);
static int               writeChar(FILE *fp, int c, int mode);
static int               writeMagic(FILE *fp, int strflag);

static int               doShowUnPack(FILE *fp, const char *dest, int mode);
static int               listUnPackLoad(void *user, int flags, long len,
					const char *name, unsigned char *data);

/*********     GLOBAL VARIABLES                                 *********/
static int Verbose;

static const char *Label = "Data";

static char Magic[PGBACK_MAGIC_LEN];

static int Pos = -1;
static long CryptPwd = -1;

#ifdef CONFIG_HAS_BWCOMPRESSION
static unsigned char *CompBuffer;
static int CompBlockSize = BWCOMP1_BLOCKSIZE;
#endif

/*JS*********************************************************************
*
*************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
  FILE *fp;
  const char *string, *prefix, *basedir;
  int mode, c;

  /*  We avoid having the magic value directly in our executable, so
   *   a scanner won't find any garbage
   */
  for(c = 0 ; c < PGBACK_MAGIC_LEN ; ++c)
    Magic[c] = PGBACK_MAGIC_INV[PGBACK_MAGIC_LEN - 1 - c];

  /*
   * bit 0: Generate string output.
   * bit 1: Generate static variable.
   * bit 2: Generate binary output.
   * bit 8: Compress files in load.
   * bit 9: Encrypt files.
   */
  mode = 1;
  prefix = NULL;
  basedir = NULL;

  while((c = jsGetOpt(&argc, argv, OPTSTRING, (const char **)&string))
	!= EOF) {
    switch(c) {
    default:

    case 'h':
      goto usage;

    case 'v':
      ++Verbose;
      break;

    case 'l':
      Label = string;
      break;

    case 'n':
      mode &= ~1;
      break;

    case 's':
      mode |= 2;
      break;

#ifdef CONFIG_HAS_BWCOMPRESSION
    case 'c':
      mode |= 256;
      break;

    case 'C':
      CryptPwd = piggystring2pwd(string);
      mode |= 512;
      break;

    case 'B':
      if((CompBlockSize = atoi(string)) > BWCOMP1_BLOCKSIZE) {
	fprintf(stderr, "ERROR: Blocksize %d is larger than piggyback's max block size %d!\n",
		CompBlockSize, BWCOMP1_BLOCKSIZE);
	EGOTO(error2);
      }
      break;
#endif

    case 'b':
      basedir = string;
      break;

    case 'p':
      prefix = string;
      break;
    }
  }

  if(argc < 2)
    goto usage;

  PiggyBackGetPwd = getloadpwd; /*  in case unpacking needs a password  */

  if(strcmp(argv[1], "pack") == 0 ||
     strcmp(argv[1], "packbin") == 0) {
    if(argc < 4)
      goto usage;

    if(strcmp(argv[1], "packbin") == 0)
      mode |= 4;

#ifdef CONFIG_HAS_BWCOMPRESSION
    if((mode & 256) &&
       (CompBuffer = (unsigned char *)
	malloc(3 * CompBlockSize + bwestimateaux(CompBlockSize))) == NULL)
      EGOTO(error);
#endif

    if((fp = fopen(argv[2], mode & 4 ? "wb" : "w")) == NULL)
      EGOTO(error);

    if(doPack(fp, (const char **)argv + 3, argc - 3, basedir, prefix, mode) == -1)
      EGOTO(error);

    if(fclose(fp) == EOF)
      EGOTO(error);
  } else if(strcmp(argv[1], "show") == 0) {
    if(argc != 3)
      goto usage;

    if((fp = fopen(argv[2], "rb")) == NULL)
      EGOTO(error);

    if(doShowUnPack(fp, NULL, mode) == -1)
      EGOTO(error);

    if(fclose(fp) == EOF)
      EGOTO(error);
  } else if(strcmp(argv[1], "unpack") == 0) {
    if(argc != 4)
      goto usage;

    if((fp = fopen(argv[2], "rb")) == NULL)
      EGOTO(error);

    if(doShowUnPack(fp, argv[3], mode) == -1)
      EGOTO(error);

    if(fclose(fp) == EOF)
      EGOTO(error);
  } else
    goto usage;

  return 0;
error:
  jsperror("pgback");
error2:
  return 30;
usage:
  printf("\
Usage: pgback [<opts>] <mode>...\n\
  Mode can be:\n\
    pack|packbin <dest_file> <file> ...\n\
      Pack the given files, generating a C-source file containing them.\n\
      If <file> is \"-\", file names are read from stdin, one per line.\n\
      Mode 'packbin' generates a binary file.\n\
    show <bin_file>\n\
      Scan the given binary file for packed files, listing them.\n\
    unpack <bin_file> <directory>\n\
      Scan the given binary file for packed files and unpack them relative\n\
      to the given directory. Subdirectories are created on demand.\n\
  Available options:\n\
    -h            Show this help information.\n\
    -l<label>     Change generated label name in source file.\n\
    -s            Generate static variable.\n\
    -n            Use numeric character field instead of concatenated\n\
                  string.\n");

#ifdef CONFIG_HAS_BWCOMPRESSION
  printf("\
    -c            Generate compressed files during packing.\n\
    -C<pwd>       In pack mode, encrypt entries with given password.\n\
                  In unpack mode, decrypt entries if necessary.\n\
    -B<blocksize> Set blocksize for encryption. Maximum value supported\n\
                  by piggyback reader is %d which is also the default.\n",
	 BWCOMP1_BLOCKSIZE);
#endif
	 printf("\
    -b<basedir>   Files to pack are relative to <basedir>\n\
    -p<prefix>    Remove prefix from file names in generated load.\n");

  return 5;
}

/*JS*********************************************************************
*
*************************************************************************/

static unsigned long getloadpwd(const char *name)

/************************************************************************/
{
  return CryptPwd;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doPack1(FILE *fp, const char *basedir, const char *fileName,
		   const char *prefix, unsigned long lengths[], int mode)

/************************************************************************/
{
  char fileBuffer[JS_FILENAME_MAX];
  unsigned char buffer[JS_INT162CHAR_LEN + JS_INT322CHAR_LEN];
  const char *name;
  FILE *fpin;
  unsigned long origLen, len;
  int flags, c, j;

  if(basedir != NULL) {
    sprintf(fileBuffer, "%s/%s", basedir, fileName);
    fileName = fileBuffer;
  }

  if((fpin = fopen(fileName, "rb")) == NULL)
    EGOTO(error);

  if(prefix && (name = jstrstartswith(fileName, prefix)) != NULL) {
    while(*name == '/'
#ifdef _WIN32
	  || *name == '\\'
#endif
	  )
      ++name;
  } else
    name = fileName;

  if(Verbose > 0)
    printf("Packing '%s' to '%s'.\n", fileName, name);

  flags = 0;

#ifdef CONFIG_HAS_BWCOMPRESSION
  if(mode & 256) {
    FILE *compFp;
    unsigned long rlen;
    unsigned char *cdest;
    void *caux;

    if((compFp = tmpfile()) == NULL)
      EGOTO(error);

    /*
     * Buffer usage:
     *  |<-- uncompressed data -->|<-- compressed data -->|<-- auxiliary space -->|
     *         block_size               2*block_size           bwestimateaux()
     */
    cdest = &CompBuffer[CompBlockSize];
    caux = (void *)&cdest[2 * CompBlockSize];

    for(origLen = 0 ;;) {
      unsigned long clen;

      if((rlen = fread(CompBuffer, 1, CompBlockSize, fpin)) == (unsigned long)-1)
	EGOTO(error);

      if(rlen == 0)
	break;

      origLen += rlen;
      if((clen = bwcompress(CompBuffer, rlen, cdest, caux,
			    0)) == (unsigned long)-1)
	EGOTO(error);

      if(mode & 512) {
	piggybackcrypt(CryptPwd, CompBuffer + BWTRANS_BLOCKHDR_SIZE,
		       clen - BWTRANS_BLOCKHDR_SIZE);
      }

      if(fwrite(cdest, 1, clen, compFp) != clen)
	EGOTO(error);
    }

    if(fclose(fpin) == EOF)
      EGOTO(error);

    fflush(compFp);
    rewind(compFp);
    fpin = compFp;

    flags |= PGBACK_FBWCOMP1;
  } else
    origLen = -1;
#endif

  if((len = fpFileLength(fpin)) == (unsigned long)-1)
    EGOTO(error);

#ifdef CONFIG_HAS_BWCOMPRESSION
  if(Verbose > 0 && origLen != (unsigned long)-1)
    printf("  Compressed %10ld -> %10ld (%3.0f%%).\n", origLen, len,
	   origLen > 0 ? 100.0 * len / origLen : 100.0);
#endif

  if(lengths) {
    lengths[0] += origLen;
    lengths[1] += len;
  }

  if(writeMagic(fp, mode & 5) == -1)
    EGOTO(error);

  if(len > 0xffff) {
    flags |= PGBACK_FLEN4BYTE;

    JS_INT322CHAR_LE(len, buffer + JS_INT162CHAR_LEN);
    c = JS_INT162CHAR_LEN + JS_INT322CHAR_LEN;
  } else {
    JS_INT162CHAR_LE(len, buffer + JS_INT162CHAR_LEN);
    c = JS_INT162CHAR_LEN + JS_INT162CHAR_LEN;
  }

  JS_INT162CHAR_LE(flags, buffer);

  /*  When writing binary data (flags + length), make
   *   sure, we don't write e. g. 'A' when we actually
   *   want hex value 65 - the machine we compile it
   *   on later might not use ASCII!
   */
  for(j = 0 ; j < c ; ++j)
    if(writeChar(fp, buffer[j], 2 | (mode & 5)) == -1)
      EGOTO(error);

  /*  Write file name  */
  for(j = 0 ; name[j] != '\0' ; ++j)
    if(writeChar(fp, name[j], mode & 5) == -1)
      EGOTO(error);

  if(writeChar(fp, '\0', mode & 5) == -1)
    EGOTO(error);

  /*  Start a new line  */
  if(writePause(fp, mode & 5) == -1)
    EGOTO(error);

  /*  Write the actual data  */
  while((c = getc(fpin)) != EOF)
    if(writeChar(fp, c, mode & 5) == EOF)
      EGOTO(error);

  if(fclose(fpin) == EOF)
    EGOTO(error);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doPack(FILE *fp, const char *list[], unsigned int n, const char *basedir,
		  const char *prefix, int mode)

/************************************************************************/
{
  unsigned int i, j, flags;
  unsigned char buffer[JS_INT162CHAR_LEN + JS_INT322CHAR_LEN];
  unsigned long lengths[2];

  if(!(mode & 4)) {
    fprintf(fp,
	    "/* ***  Machine generated by 'pgback' -- do not edit  *** */\n\n");

    if(!(mode & 2))
      fprintf(fp, "\
#define Prototype extern\n\
/*********     PROTOTYPES                                       *********/\n\
Prototype const unsigned char %s[];\n\n", Label);

    fprintf(fp, "\
%sconst unsigned char %s[] =%s\n", (mode & 2) ? "static " : "",
	    Label, (mode & 1) ? "" : " {");
  }

  lengths[0] = lengths[1] = 0;

  for(i = 0 ; i < n ; ++i) {
    if(strcmp(list[i], "-") == 0) {
      char line[1000];

      while(fgets(line, sizeof(line), stdin) != NULL) {
	j = strlen(line);
	while(j > 0 && isspace(line[j - 1]))
	  --j;

	line[j] = '\0';

	if(doPack1(fp, basedir, line, prefix, lengths, mode) == -1)
	  EGOTO(error);
      }

    } else if(doPack1(fp, basedir, list[i], prefix, lengths, mode) == -1)
      EGOTO(error);
  }

  /*  Write trailer: an entry with flags == -1  */
  if(writeMagic(fp, mode & 5) == -1)
    EGOTO(error);

  flags = -1;
  JS_INT162CHAR_LE(flags, buffer);

  for(j = 0 ; j < JS_INT162CHAR_LEN ; ++j)
    if(writeChar(fp, buffer[j], 2 | (mode & 5)) == -1)
      EGOTO(error);

  if(!(mode & 4) && fprintf(fp, (mode & 1) ? "\";\n" : "\n};\n") == -1)
    EGOTO(error);

  if(Verbose > 0) {
#ifdef CONFIG_HAS_BWCOMPRESSION
    printf("Total: %10ld -> %10ld (%3.0f%%).\n", lengths[0], lengths[1],
      lengths[0] > 0 ? 100.0 * lengths[1] / lengths[0] : 100.0);
#else
    printf("Total: %10ld\n", lengths[1]);
#endif
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int writePause(FILE *fp, int mode)

/************************************************************************/
{
  if(Pos == 3 || (mode  & 4))
    return 0;

  if(Pos != -1) {
    if((mode & 1) && putc('\"', fp) == EOF)
      return -1;

    if(putc('\n', fp) == EOF)
      return -1;
  }

  return (Pos = fprintf(fp, mode & 1 ? "  \"" : "  "));
}

/*JS*********************************************************************
*   mode bit 0: If set, write character string
*        bit 1: If set, write character as octal escape sequence
*        bit 2: If set, write binary
*************************************************************************/

static int writeChar(FILE *fp, int c, int mode)

/************************************************************************/
{
static const char EscChar1[] = "\n\t\f\\\"";
static const char EscChar2[] = "ntf\\\"";
  char *s;

  if(mode & 4) {
    if(c == EOF)
      return 0;

    return fputc(c, fp) == -1 ? -1 : 0;
  }

  if((Pos > 76 || Pos == -1) && writePause(fp, mode & 1) == -1)
    return -1;

  if(c == EOF) {
    Pos = -1;

    if((mode & 1) && putc('\"', fp) == EOF)
      return -1;

    if(putc('\n', fp) == EOF)
      return -1;

    return 0;
  }

  if(mode & 1) {
    if(c != '\0' && !(mode & 2) && (s = strchr(EscChar1, c)) != NULL) {
      if(putc('\\', fp) == EOF || putc(EscChar2[s - EscChar1], fp) == EOF)
	return -1;

      Pos += 2;
    } else if(!(mode & 2) && isprint(c)) {
      if(putc(c, fp) == EOF)
	return -1;

      Pos++;
    } else {
      int i;

      if((i = fprintf(fp, "\\%03o", c)) == -1)
	return -1;

      Pos += i;
    }

    if(c == '\n' && writePause(fp, mode & 1) == EOF)
      return -1;
  } else {
    int l;

    if((l = fprintf(fp, " %d,", c)) == -1)
      return -1;

    Pos += l;
  }

  return c;
}

/*JS*********************************************************************
*
*************************************************************************/

static int writeMagic(FILE *fp, int strflag)

/************************************************************************/
{
  int j;

  for(j = 0 ; j < PGBACK_MAGIC_LEN ; ++j)
    if(writeChar(fp, Magic[j], strflag) == EOF)
      return -1;

  return PGBACK_MAGIC_LEN;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doShowUnPack(FILE *fp, const char *dest, int mode)

/************************************************************************/
{
  unsigned char *buffer;
  unsigned long off, len;
  auxlistunpack_t aux;

  aux.Directory = dest;

  if((len = fpFileLength(fp)) == (unsigned long)-1 ||
     (buffer = (unsigned char *)malloc(len)) == NULL ||
     fread(buffer, 1, len, fp) != len)
    EGOTO(error);

  for(off = 0 ; off < len ; ) {
    unsigned char *s;
    int j;

    if((s = memchr(buffer + off, Magic[0], len - off))
       == NULL)
      break; /*  no more loads found  */

    off = s - buffer;
    for(j = 1 ; j < PGBACK_MAGIC_LEN ; ++j)
      if(buffer[off + j] != Magic[j])
	break;

    if(j != PGBACK_MAGIC_LEN) {
      ++off;
      continue;
    }

    if(dest == NULL)
      printf("Showing load at offset %ld:\n", off);
    else
      printf("Unpacking load at offset %ld to '%s'\n", off, dest);

    aux.Data = buffer + off;

    if(piggybackscan(buffer + off, listUnPackLoad, &aux, &s, 0) == -1)
      EGOTO(error);

    printf("Load length %ld\n", s - (buffer + off));

    off = s - buffer;
  }

  free(buffer);

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int listUnPackLoad(void *user, int flags, long len,
			  const char *name,
			  unsigned char *data)

/************************************************************************/
{
  if(((auxlistunpack_t *)user)->Directory) {
    char fileName[JS_FILENAME_MAX];
    char buffer[4096];
    extfile_t *efp;
    FILE *fp;
    unsigned long rlen;

    sprintf(fileName, "%s/%s", ((auxlistunpack_t *)user)->Directory, name);

    if((efp = piggybackopener(((auxlistunpack_t *)user)->Data, name, "r")) == NULL)
      return -1;

    if((fp = fopenCreate(fileName, "w")) == NULL)
      return -1;

    while((rlen = efread(buffer, sizeof(buffer), efp)) != 0) {
      if(rlen == (unsigned long)-1)
	return -1;

      if(fwrite(buffer, 1, rlen, fp) != rlen)
	return -1;
    }

    if(efclose(efp) == -1 || fclose(fp) == EOF)
      return -1;

  } else
    printf("  FLAGS 0x%04x LEN %6ld '%s'\n", flags, len, name);

  return 0;
}
