/*JS*********************************************************************
*
*    Program : SUPPTRACE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Auxiliary function for jsftrace.
*
*              Copyright (C) 2004-2007 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*************************************************************************
*  Supported settings through environment variable "FUNCTIONTRACE" (case
*  insensitive):
*   off            Switch tracing off (the default).
*   on             Switch tracing completely on.
*   function       Only trace function call and return.
*   variable       Only trace variable changes and access.
*   statement      Only trace statements processed.
*   fullstatement  Only trace statements processed, but in readable form.
*   user           Only trace user generated one with function "_traceprintf".
*   local          Switch tracing off for other programs we execute.
*
*   start=<funcname> Start tracing when the given function is entered
*                  for the first time.
*
*   output=<name>  Use <name> instead of "FUNCTION.TRC" for the tracing
*                  output. A single '$' is substituted by the process
*                  PID.
*   size=<size>    Maximum size of the trace file. The number can have
*                  an extension of 'k', 'K', 'm', or 'M'. When the trace
*                  file reaches this size, output continues at the start
*                  of the trace file.
*                  Lines are always written completely, so the actual
*                  size might be slightly bigger than demanded.
*   flush          Flushes every line individually. Useful in case the
*                  program crashes and one wants to know exactly when.
*
*   maxlen=<n>     Set the maximum output length of string arguments.
*   indent=<n>     Set the indentation in the call stack. Defaults to 1.
*
*   nosignal       Do not install signal handler to toggle tracing (default).
*   signal=<n>     Use the given signal number to toggle tracing.
*   sigusr1        Use SIGUSR1 to toggle tracing.
*   sigusr2        Use SIGUSR2 to toggle tracing.
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>

#ifdef _WIN32
# include <process.h>
#else
# include <sys/types.h>
# include <unistd.h>
#endif

#include <signal.h>
#include <time.h>

/*********     DEFINES                                          *********/
#define VERSION                "1.12.0"

/*#define DEFAULT_TRACING_ON*/

#define MAX(a, b)               ((a) > (b) ? (a) :  (b))

#define MAX_IDENTLEN           256

#define TRACEFILE              "FUNCTION.TRC"

#define ENV_FUNCTIONTRACE      "FUNCTIONTRACE"

/*  If defined, this file is checked (in the current directory) for
 *   trace settings
 */
#define FILE_FUNCTIONTRACE     "FUNCTIONTRACE"

#define DEFAULT_TRACESIZE      (16 * 1024 * 1024) /*  16MB  */

#define TRACE_FUNCTIONS         1  /*  for function entry/exit tracing  */
#define TRACE_STATEMENTS        2  /*  for statement tracing  */
#define TRACE_STATEMENTS_FULL   4  /*  for statement tracing in full form  */
#define TRACE_VARIABLE          8  /*  for variable tracing  */
#define TRACE_USER             16  /*  for user defined tracing  */

#define TRACE_ALL_ON           (TRACE_FUNCTIONS | TRACE_STATEMENTS | \
				TRACE_VARIABLE | TRACE_USER)

typedef void (*sigfunction_t)(int);

extern void                    _jstracefunc(const char *fname,
					    const char *format, ...);
extern void                    _jstracevar(const char *file, int linenr,
					   const char *format, ...);
extern void                    _jstracestmt(const char *file, int linenr);
extern void                    _jstraceprintf(const char *format, ...);
extern int                     _jstracecontrol(const char *setting);

static int                     setup(void);
static int                     flushTrace(void);
static int                     readSettings(const char *input, int flag);
static int                     handleSetting(char *input);
static int                     mystricmp(const char *s1, const char *s2);

static int                     openTraceFile(void);
static sigfunction_t           setupsignal(int signo);
static void                    sigHandler(int sig);

static int                     printFormat(FILE *fp, char type, va_list *ap);
static int                     printString(FILE *fp, const char *string,
					   int maxlen);

/*********     GLOBAL VARIABLES                                 *********/

/*  Has to be != 0 initially, otherwise we never enter the tracing stuff
 *   even once
 */
static unsigned char Tracing = TRACE_ALL_ON;
static unsigned char SaveTracing = TRACE_ALL_ON;

static int Init = 0;

static char TraceName[FILENAME_MAX] = TRACEFILE;
static long TraceSize = DEFAULT_TRACESIZE;
static long TraceCounter, TraceStartOffset, TraceOffset;

static char StartFunction[MAX_IDENTLEN];
static int StartFunctLevel = 0;

static FILE *FPTrace;
static int FPFlush; /*  != 0 -> flush after every line  */
static int Signal = -1;

static int IndentStep = 1, MaxStrLen = 100;
static char *StringBuffer;
static wchar_t *WStringBuffer;

/*  working variables  */
static int Indent;

/*  If somebody runs this through the tracer accidentally, it must
 *   not be traced itself!
 */
#ifdef __JSFTRACE__
#pragma __JSFTRACE__ trace    "*" off
#pragma __JSFTRACE__ vartrace "*" off
#endif

/*JS*********************************************************************
*
*************************************************************************/

void _jstracefunc(const char *fname, const char *format, ...)

/************************************************************************/
{
  const char *funcName;
  va_list ap;
  int line;
  char sep;

  /*  That is the quick end if there is nothing for us to do  */
  if((Tracing & TRACE_FUNCTIONS) == 0)
    return;

  if(!Init && setup() == 0)
    return;

  funcName = (fname[0] == '+' || fname[0] == '-') ? fname + 1 : fname;

  if(*funcName == '@') {
    ++funcName;
    line = 0;
  } else
    line = -1;

  /*  Function name given where to start tracing?  */
  if(StartFunction[0] != '\0') {
    /*  On level 0, check the name  */
    if(StartFunctLevel == 0 && strcmp(funcName, StartFunction) != 0)
      return;

    /*  If only function entry is traced (fname[0] neither '+' nor '-'),
     *   we start tracing if we enter the given function for the first
     *   time and never stop again (actually we do stop if an integer
     *   overflow occurs after about 2 billion function calls).
     */
    if(fname[0] != '-')
      ++StartFunctLevel;
    else if(StartFunctLevel > 0)
      --StartFunctLevel;
  }

  va_start(ap, format);
  if(line != -1)
    line = va_arg(ap, int);

  if(TraceSize != 0)
    TraceOffset += fprintf(FPTrace, "%ld:", ++TraceCounter);

  if(fname[0] == '+') {
    TraceOffset += fprintf(FPTrace, "%*s[%s", MAX(Indent, 0), "", funcName);

    Indent += IndentStep;
    sep = '(';
  } else if(fname[0] == '-') {
    Indent -= IndentStep;

    TraceOffset += fprintf(FPTrace, "%*s]%s return", MAX(Indent, 0),
			   "", funcName);
    sep = ' ';
  } else {
    TraceOffset += fprintf(FPTrace, "%*s|%s", MAX(Indent, 0), "", funcName);
    sep = ' ';
  }

  if(line >= 0)
    TraceOffset += fprintf(FPTrace, " l. %d", line);

  for( ; *format ; ++format, sep = ',') {
    fputc(sep, FPTrace); ++TraceOffset;

    if(isupper(*format)) {
      const char *label = va_arg(ap, const char *);

      TraceOffset += fprintf(FPTrace, "%s=", label);
    }

    TraceOffset += printFormat(FPTrace, (char)tolower(*format), &ap);
  }

  va_end(ap);

  if(fname[0] == '+' && sep != '(') {
    fputs(")\n", FPTrace);  TraceOffset += 2;
  } else if(fname[0] == '-') {
    fputs("\n", FPTrace);   TraceOffset += 2;
  } else {
    fputs("\n", FPTrace);   ++TraceOffset;
  }

  flushTrace();
}

/*JS*********************************************************************
*
*************************************************************************/

void _jstracestmt(const char *file, int linenr)

/************************************************************************/
{
static const char *OldFile;
static int OldLine;

  /*  That is the quick end if there is nothing for us to do  */
  if((Tracing & (TRACE_STATEMENTS | TRACE_STATEMENTS_FULL)) == 0)
    return;

  if(!Init && setup() == 0)
    return;

  if(TraceSize != 0)
    TraceOffset += fprintf(FPTrace, "%ld:", ++TraceCounter);

  if(Indent > 0)
    TraceOffset += fprintf(FPTrace, "%*s", MAX(Indent, 0), "");

  /*  It's enough to compare file pointers  */
  if(file != OldFile || (Tracing & TRACE_STATEMENTS_FULL)) {
    /*  File and line have changed  */
    TraceOffset += fprintf(FPTrace, "@%d %s\n", linenr, file);
    OldFile = file;
  } else {
    int diff = linenr - OldLine;

    if(linenr < (diff >= 0 ? diff : -diff)) {
      /*  It cheaper to display the absolute line  */
      TraceOffset += fprintf(FPTrace, "@%d\n", linenr);
    } else if(diff >= 0) {
      /*  Use the relative line diff  */
      TraceOffset += fprintf(FPTrace, "+%d\n", diff);
    } else {
      TraceOffset += fprintf(FPTrace, "-%d\n", -diff);
    }
  }
  OldLine = linenr;

  if(flushTrace())
    /*  Make sure we write the file name at least once if the file
     *   wraps around
     */
    OldFile = NULL;
}

/*JS*********************************************************************
*
*************************************************************************/

extern void _jstracevar(const char *file, int linenr,
			const char *format, ...)

/************************************************************************/
{
  va_list ap;

  /*  That is the quick end if there is nothing for us to do  */
  if((Tracing & TRACE_VARIABLE) == 0)
    return;

  if(!Init && setup() == 0)
    return;

  if(StartFunction[0] != '\0' && StartFunctLevel == 0)
    return;

#if 0
  {
    const char *s;

    /*  Skip the path  */
    if((s = strrchr(file, '/')) != NULL)
      file = s + 1;
  }
#endif

  va_start(ap, format);

  if(TraceSize != 0)
    TraceOffset += fprintf(FPTrace, "%ld:", ++TraceCounter);

  TraceOffset += fprintf(FPTrace, "%*s!'%s' l. %d: %s variable '%s' %s value ",
			 MAX(Indent, 0), "", file, linenr,
			 format[1] == '>' ? "Changing" :
			 format[1] == '&' ? "Getting address of" : "Reading",
			 format + 2, format[1] == '>' ? "to" : "with");

  TraceOffset += printFormat(FPTrace, *format, &ap);

  putc('\n', FPTrace);  TraceOffset++;

  va_end(ap);

  flushTrace();
}

/*JS*********************************************************************
*
*************************************************************************/

void _jstraceprintf(const char *format, ...)

/************************************************************************/
{
  va_list ap;

  /*  That is the quick end if there is nothing for us to do  */
  if((Tracing & TRACE_USER) == 0)
    return;

  if(!Init && setup() == 0)
    return;

  if(StartFunction[0] != '\0' && StartFunctLevel == 0)
    return;

  va_start(ap, format);

  if(TraceSize != 0)
    TraceOffset += fprintf(FPTrace, "%ld:", ++TraceCounter);

  if(*format == '\t') {
    ++format;
    TraceOffset += fprintf(FPTrace, "%*s", MAX(Indent, 0), "");
  }

  TraceOffset += vfprintf(FPTrace, format, ap);

  flushTrace();

  va_end(ap);
}

/*JS*********************************************************************
*
*************************************************************************/

int _jstracecontrol(const char *setting)

/************************************************************************/
{
  char *s;
  int ret;

  Init = 1;

  s = getenv(ENV_FUNCTIONTRACE);

  if(setting && (ret = readSettings(setting, s == NULL ? 1 : 0)) == -1)
    return -1;

  /*  Note: We want the environment to overwrite what the program
   *   specifies, so we handle it AFTER handling the argument.
   */
  if(s != NULL)
    return readSettings(s, 1);

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int setup(void)

/************************************************************************/
{
  char *env;
  int ret;

  Init = 1;

#ifndef DEFAULT_TRACING_ON
  /*  Default: Unless we see "on" in environment variable "FUNCTIONTRACE",
   *   we are quiet
   */
  Tracing = 0;
#endif

  ret = 0;

  env = getenv(ENV_FUNCTIONTRACE);

#ifdef FILE_FUNCTIONTRACE
  {
    FILE *fp;

    if((fp = fopen(FILE_FUNCTIONTRACE, "r")) != NULL) {
      char buffer[1000];

      if(fgets(buffer, sizeof(buffer) - 1, fp) != NULL) {
	if(readSettings(buffer, env == NULL ? 1 : 0) == -1)
	  goto error;
      }

      if(fclose(fp) == EOF)
	fprintf(stderr, "JSCPP: ERROR: Cannot read '%s'.\n",
		FILE_FUNCTIONTRACE);
    }
  }
#endif

  /*  Check for options via environment variables  */
  if(env != NULL) {
    if((ret = readSettings(env, 1)) == -1)
      goto error;
  }

  return ret <= 0 ? 0 : 1;
error:
  fprintf(stderr, "JSCPP: ERROR initializing tracing - disabling it.\n");
  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int flushTrace(void)

/************************************************************************/
{
  if(FPFlush)
    fflush(FPTrace);

  if(TraceSize != 0 && TraceOffset >= TraceSize) {
    if(!FPFlush)
      fflush(FPTrace);

    /*  Reuse trace file from beginning  */
    if(fseek(FPTrace, TraceStartOffset, SEEK_SET) == -1) {
      fprintf(stderr,
	      "JSCPP: FATAL: Cannot seek in trace file: '%s' - giving up\n",
	      strerror(errno));
      Tracing = 0;
    }

    TraceOffset = TraceStartOffset;

    return 1;
  }

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static int readSettings(const char *input, int flag)

/************************************************************************/
{
  char *mem, *temp, *s;

  if((mem = (char *)malloc(strlen(input) + 1)) == NULL)
    goto error;

  strcpy(mem, input);

  for(temp = mem ; ; ) {
    if((s = strchr(temp, ';')))
      *s = '\0';

    handleSetting(temp);

    if(s == NULL)
      break;

    temp = s + 1;
  }

  free(mem);

  if(StartFunction[0]) /*  make sure it works  */
    Tracing |= TRACE_FUNCTIONS;

  if(Tracing == 0)
    return 0;

  if(flag) {
    if(StringBuffer) free(StringBuffer);
    if(WStringBuffer) free(WStringBuffer);

    if((StringBuffer = (char *)malloc((MaxStrLen + 1) *
				      sizeof(*StringBuffer))) == NULL ||
       (WStringBuffer = (wchar_t *)malloc((MaxStrLen + 1) *
					  sizeof(*WStringBuffer))) == NULL)
      goto error;

    StringBuffer[MaxStrLen] = '\0'; /*  terminate it properly  */
    WStringBuffer[MaxStrLen] = '\0';

    if(openTraceFile() == -1)
      goto error;

    /*  Install signal handler  */
    if(Signal != -1 && setupsignal(Signal) == SIG_ERR)
      goto error;
  }

  return 1;
error:
  /*  Bark and give up tracing  */
  Tracing = 0;
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int handleSetting(char *input)

/************************************************************************/
{
  char *value;

  if((value = strchr(input, '=')))
    *value++ = '\0';

  if(strcmp(input, "off") == 0 && value == NULL) {
    Tracing = 0;
  } else if(mystricmp(input, "on") == 0 && value == NULL) {
    Tracing |= TRACE_ALL_ON;
  } else if(mystricmp(input, "function") == 0 && value == NULL) {
    Tracing |= TRACE_FUNCTIONS;
  } else if(mystricmp(input, "statement") == 0 && value == NULL) {
    Tracing |= TRACE_STATEMENTS;
  } else if(mystricmp(input, "fullstatement") == 0 && value == NULL) {
    Tracing |= TRACE_STATEMENTS_FULL;
  } else if(mystricmp(input, "variable") == 0 && value == NULL) {
    Tracing |= TRACE_VARIABLE;
  } else if(mystricmp(input, "user") == 0 && value == NULL) {
    Tracing |= TRACE_USER;
  } else if(mystricmp(input, "local") == 0 && value == NULL) {
    /*  Switch it off for our child processes  */
    putenv(ENV_FUNCTIONTRACE"=off");

  } else if(mystricmp(input, "start") == 0 && value != NULL) {
    strcpy(StartFunction, value);

  } else if(mystricmp(input, "output") == 0 && value) {
    int in, out;

    for(in = out = 0 ; value[in] ; ++in)
      if(value[in] == '$')
	out += sprintf(TraceName + out, "%d", getpid());
      else
	TraceName[out++] = value[in];

    TraceName[out] = '\0';
  } else if(mystricmp(input, "size") == 0 && value != NULL) {
    TraceSize = strtoul(value, &value, 10);
    if(*value == 'k' || *value == 'K')
      TraceSize <<= 10;
    else if(*value == 'm' || *value == 'M')
      TraceSize <<= 20;
  } else if(mystricmp(input, "flush") == 0 && value == NULL) {
    FPFlush = 1;

  } else if(mystricmp(input, "maxlen") == 0 && value) {
    MaxStrLen = atoi(value);
  } else if(mystricmp(input, "indent") == 0 && value) {
    IndentStep = atoi(value);

  } else if(mystricmp(input, "nosignal") == 0 && value == NULL) {
    Signal = -1;
  } else if(mystricmp(input, "signal") == 0 && value) {
    Signal = atoi(value);
#ifdef SIGUSR1
  } else if(mystricmp(input, "sigusr1") == 0 && value == NULL) {
    Signal = SIGUSR1;
#endif
#ifdef SIGUSR2
  } else if(mystricmp(input, "sigusr2") == 0 && value == NULL) {
    Signal = SIGUSR2;
#endif
  } else {
    fprintf(stderr, "JSCPP: WARNING: Syntax error '%s' in '%s'!\n",
	    input, ENV_FUNCTIONTRACE);
    goto error;
  }

  return 0;
error:
  return -1;
}

/*JS*********************************************************************
*
*************************************************************************/

static int mystricmp(const char *s1, const char *s2)

/************************************************************************/
{
  register char c1, c2;

  while((c1 = tolower(*(unsigned char *)s1)) ==
	(c2 = tolower(*(unsigned char *)s2))) {
    if(c1 == '\0')
      return 0;

    s1++;  s2++;
  }

  return c1 - c2;
}

/*JS*********************************************************************
*
*************************************************************************/

static int openTraceFile(void)

/************************************************************************/
{
  time_t tim;

  if(FPTrace != NULL)
    fclose(FPTrace);

  /*  Open trace file  */
  if((FPTrace = fopen(TraceName, TraceSize != 0 ? "w" : "a")) == NULL)
    return -1;

  tim = time(NULL);
  TraceStartOffset = TraceOffset =
    fprintf(FPTrace, "#*** JSCPP function tracing V%s started %s", VERSION,
	    ctime(&tim));

  /*  Flush it at least once to see the header lines  */
  fflush(FPTrace);

  return 0;
}

/*JS*********************************************************************
*
*************************************************************************/

static sigfunction_t setupsignal(int signo)

/************************************************************************/
{
#ifndef _WIN32
  struct sigaction act, oact;

  act.sa_handler = (void (*)(int))sigHandler;

  /*  No other signals (beside the generated) will be blocked  */
  sigemptyset(&act.sa_mask);

  act.sa_flags = 0;
  if(signo == SIGALRM) {
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;  /*  SunOS  */
#endif
  } else {
#ifdef SA_RESTART
    act.sa_flags |= SA_RESTART;    /*  SVR4,  4.3+BSD  */
#endif
  }

  /*  Install signal handler  */
  if(sigaction(signo, &act, &oact) == -1)
    return (void (*)(int))SIG_ERR;

  /*  Return old handler  */
  return (void (*)(int))oact.sa_handler;
#else
  return signal(signo, sigHandler);
#endif
}

/*JS*********************************************************************
*
*************************************************************************/

static void sigHandler(int sig)

/************************************************************************/
{
  if(Tracing != 0) {
    if(FPTrace != NULL) {
      fclose(FPTrace);
      FPTrace = NULL;
    }

    SaveTracing = Tracing;
    Tracing = 0;
  } else {
    openTraceFile();

    Tracing = SaveTracing;
  }

  signal(Signal, sigHandler);
}

/*JS*********************************************************************
*
*************************************************************************/

static int printFormat(FILE *fp, char type, va_list *ap)

/************************************************************************/
{
  int len;

  switch(type) {
  case 'c':
    len = fprintf(fp, "'%c'", va_arg(*ap, int));
    break;
  case 'h': case 'i':
    len = fprintf(fp, "%d", va_arg(*ap, int));
    break;
  case 'l':
    len = fprintf(fp, "%ld", va_arg(*ap, long));
    break;
  case 'f': case 'd':
    len = fprintf(fp, "%g", va_arg(*ap, double));
    break;
  case 's': /*  "char *"  */
    /*  This might produce garbage, but at least no core dump  */
    len = printString(fp, va_arg(*ap, char *), MaxStrLen);
    break;
  case 'w': /*  "wchar_t *"  */
  {
    wchar_t *wcs = va_arg(*ap, wchar_t *);

    if(wcs == NULL) {
      len = printString(fp, NULL, MaxStrLen);
    } else {
      wcstombs(StringBuffer, wcs, MaxStrLen);
      len = printString(fp, StringBuffer, MaxStrLen);
    }
  }
  break;

  case 'u': /*  "unsigned short *" as a multibyte string  */
  {
    unsigned short *us = va_arg(*ap, unsigned short *);

    if(us == NULL) {
      len = printString(fp, NULL, MaxStrLen);
    } else if(sizeof(*us) == sizeof(wchar_t)) {
      wcstombs(StringBuffer, (wchar_t *)us, MaxStrLen);
      len = printString(fp, StringBuffer, MaxStrLen);
    } else {
      int i;

      for(i = 0 ; us[i] != 0 && i < MaxStrLen ; ++i)
	WStringBuffer[i] = (wchar_t)us[i];
      WStringBuffer[i] = '\0';

      wcstombs(StringBuffer, WStringBuffer, MaxStrLen);
      len = printString(fp, StringBuffer, MaxStrLen);
    }
  }
  break;

  case 'p': /*  a generic pointer  */
    len = fprintf(fp, "%p", va_arg(*ap, void *));
    break;

  default: case '?':
    len = fprintf(fp, "?<%c>?", type);
    break;
  }

  return len;
}

/*JS*********************************************************************
*
*************************************************************************/

static int printString(FILE *fp, const char *string, int maxlen)

/************************************************************************/
{
  int outLen;

  if(string == NULL)
    return fprintf(fp, "NULL");

  putc('\"', fp);
  outLen = 2;

  for( ; *string && maxlen-- != 0 ; ++string) {
static const char InEsc[]  = "\\\"\n\t\r\f\v\b\a";
static const char OutEsc[] = "\\\"ntrfvba";
    const char *t;
    char c;

    c = *(unsigned char *)string;
    if(c && (t = strchr(InEsc, c))) {
      /*  Print "\n" in case of newline char etc.  */
      putc('\\', fp);    ++outLen;
      c = OutEsc[t - InEsc];
    } else if(!isprint(c)) { /*  Print octal representation  */
      int val, flag;

      putc('\\', fp);    ++outLen;

      /*  If a digit follows or this is the last char, make it a full
       *   length octal representation
       */
      flag = isdigit(string[1]) || string[1] == '\0';
      val = ((int)c & 0xc0) >> 6;
      if(val || flag) {
        putc('0' + val, fp);    ++outLen;
	flag = 1;
      }
      val = ((int)c & 0x38) >> 3;
      if(val || flag) {
        putc('0' + val, fp);    ++outLen;
	flag = 1;
      }

      c = '0' + ((int)c & 0x07);
    }

    putc(c, fp);    ++outLen;
  }

  putc('\"', fp);

  return outLen;
}
