 /*JS*********************************************************************
*
*    Program : TESTJSPARSE
*    Language: ANSI-C
*    Author  : Joerg Schoen
*    Purpose : Sample program to show usage of jscpp library
*
*              Copyright (C) 2006 Joerg Schoen
*
*   Licensed under the Apache License, Version 2.0 (the "License");
*   you may not use this file except in compliance with the License.
*   You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
*   Unless required by applicable law or agreed to in writing, software
*   distributed under the License is distributed on an "AS IS" BASIS,
*   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*   See the License for the specific language governing permissions and
*   limitations under the License.
*
*************************************************************************/

/*********     INCLUDES                                         *********/
#include <stdio.h>

#include <jsparse.h>

/*********     DEFINES                                          *********/
#define OPTSTRING         JSPARSE_OPTSTRING "hX:"

/*********     PROTOTYPES                                       *********/
static int doparse(jsparsenviron_t *jppenv, const char *name, FILE *fpOut);

/*JS*********************************************************************
*
*************************************************************************/

int main(int argc, char *argv[])

/************************************************************************/
{
  const char *string;
  jsparsenviron_t ppenv;
  FILE *fp;
  int c;

  if(jsparsinit(0) == -1)
    EGOTO(error);

  jsparsenvclear(&ppenv);

  /*  Parse options  */
  while((c = jsGetOpt(&argc, argv, OPTSTRING, (const char **)&string))
	!= EOF) {
    switch(c) {
    default:
      /*  Generic option handling for preprocessor  */
      if((c = jsparsenvhandle(&ppenv, c, string)) == -1)
	EGOTO(error);

      if(c != -2)
	/*  OK, option handled  */
	break;
      /*FALLTHROUGH*/

    case 'h':
      goto help;

    case 'X':
      if(parsedbgopts(JSParseDebugOptions, string, &Debug, 0) == -1)
	EGOTO(error);
      break;
    }
  }

  if(argc < 2) goto help;

  if(argc == 2)
    fp = stdout;
  else if((fp = fopen(argv[2], "w")) == NULL)
    EGOTO(error);

  if(doparse(&ppenv, argv[1], fp) == -1)
    EGOTO(error);

  jsparsenvfree(&ppenv);

  return 0;

help:
  fprintf(stderr, "Usage: testjscpp [<opts>] <infile> [<outfile>]\n\
  Sample application for JSCPP library that copies the input file\n\
  to the output file or stdout while parsing it.\n\
  Standard options:\n%s%s",
	  JSPPStdOptsHelp, JSParsStdOptsHelp);
  return 5;

error:
  jsperror("testjscpp");
  return 40;
}

/*JS*********************************************************************
*
*************************************************************************/

static int doparse(jsparsenviron_t *jppenv, const char *name,
		   FILE *fpOut)

/************************************************************************/
{
  jsparsstream_t *jpa;
  jctokfifo_t *queue;
  int pmode;

  pmode = 0;

  /*  Get meaningful error and warning messages by recording the current
   *   position
   */
  pmode |= JPAM_FGETFPOS;

  /*  The output is the *original* source  */
  pmode |= JPAM_FGETORIG | JPAM_FPRINTORIG;

  /*  Let's allow lazy style and use C99 extensions  */
  pmode |= JPAM_FMIXEDDECLS | JPAM_FALLOWREDECL;

  /*  Let the parser also analyze variables definitions and produce a flow graph  */
  pmode |= JPAM_FVARTABLES; /*  may use 'jsparsfindvariable' and friends  */
  pmode |= JPAM_FSTMTFLOW;  /*  member 'jpa->JSPA_StmtFlow' is set  */

  if((jpa = jsparsopen2(NULL, NULL, jppenv, name, 0, pmode)) == NULL)
    EGOTO(error);

  queue = jsparsgetqueue(jpa);

  for(;;) {
    jsparscnode_t *node;

    if(jsparsget(jpa, &node, 0) == -1)
      EGOTO(error);

    if(node == NULL)
      break;

    /* ***  Here we could do something with the node  *** */

    /*  For demonstration purposes, print the node tree on stdout  */
    jsparsprctree(stdout, queue, node, 0, 0);

    /*  Or we could modify the tokens and output the instrumented code  */
    /* ... */
    if(jsparsflush(jpa, node, fpOut, 0) == -1)
      EGOTO(error);
  }

  /*  Flush any trailing tokens from last jsparsget
   *   (e. g. trailing comments)
   */
  if(jsparsflushtokens(jpa, fpOut) == -1) EGOTO(error);

  if(jsparsclose(jpa, 0) == -1)
    EGOTO(error);

  return 0;
error:
  return -1;
}
